Package: testthat
Title: Unit Testing for R
Version: 3.3.2
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd")),
    person("R Core team", role = "ctb",
           comment = "Implementation of utils::recover()")
  )
Description: Software testing is important, but, in part because it is
    frustrating and boring, many of us avoid it. 'testthat' is a testing
    framework for R that is easy to learn and use, and integrates with
    your existing 'workflow'.
License: MIT + file LICENSE
URL: https://testthat.r-lib.org, https://github.com/r-lib/testthat
BugReports: https://github.com/r-lib/testthat/issues
Depends: R (>= 4.1.0)
Imports: brio (>= 1.1.5), callr (>= 3.7.6), cli (>= 3.6.5), desc (>=
        1.4.3), evaluate (>= 1.0.4), jsonlite (>= 2.0.0), lifecycle (>=
        1.0.4), magrittr (>= 2.0.3), methods, pkgload (>= 1.4.0),
        praise (>= 1.0.0), processx (>= 3.8.6), ps (>= 1.9.1), R6 (>=
        2.6.1), rlang (>= 1.1.6), utils, waldo (>= 0.6.2), withr (>=
        3.0.2)
Suggests: covr, curl (>= 0.9.5), diffviewer (>= 0.1.0), digest (>=
        0.6.33), gh, knitr, otel, otelsdk, rmarkdown, rstudioapi, S7,
        shiny, usethis, vctrs (>= 0.1.0), xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/testthat/start-first: watcher, parallel*
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-01-09 17:49:32 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Posit Software, PBC [cph, fnd],
  R Core team [ctb] (Implementation of utils::recover())
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-01-11 09:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 11:13:21 UTC; unix
Archs: testthat.so.dSYM

Package: outside
Title: What the Package Does (One Line, Title Case)
Version: 0.0.0.9000
Authors@R: 
    person(given = "First",
           family = "Last",
           role = c("aut", "cre"),
           email = "first.last@example.com",
           comment = c(ORCID = "YOUR-ORCID-ID"))
Description: What the package does (one paragraph).
License: `use_mit_license()`, `use_gpl3_license()` or friends to
    pick a license
Encoding: UTF-8
LazyData: true
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.1.1
Suggests: 
    testthat
Config/testthat/parallel: true
Config/testthat/edition: 3

Package: testConfigLoadAll
Title: What the Package Does (One Line, Title Case)
Version: 0.0.0.9000
Authors@R: 
    person("First", "Last", , "first.last@example.com", role = c("aut", "cre"),
           comment = c(ORCID = "YOUR-ORCID-ID"))
Description: What the package does (one paragraph).
License: `use_mit_license()`, `use_gpl3_license()` or friends to pick a
    license
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.2.0
Config/testthat/load-all: list(export_all = FALSE, helpers = FALSE)
Imports:
    rlang
Suggests: 
    testthat (>= 3.0.0)
Config/testthat/edition: 3

Package: teardown
Title: What the Package Does (One Line, Title Case)
Version: 0.0.0.9000
Authors@R: 
    person(given = "First",
           family = "Last",
           role = c("aut", "cre"),
           email = "first.last@example.com",
           comment = c(ORCID = "YOUR-ORCID-ID"))
Description: What the package does (one paragraph).
License: `use_mit_license()`, `use_gpl3_license()` or friends to
    pick a license
Encoding: UTF-8
LazyData: true
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.1.1
Suggests: 
    testthat
Config/testthat/parallel: true
Config/testthat/edition: 3

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,testthat_results)
S3method(as.expectation,default)
S3method(as.expectation,error)
S3method(as.expectation,expectation)
S3method(as.expectation,skip)
S3method(as.expectation,warning)
S3method(compare,POSIXt)
S3method(compare,character)
S3method(compare,default)
S3method(compare,numeric)
S3method(format,expectation)
S3method(format,expectation_success)
S3method(format,mismatch_character)
S3method(format,mismatch_numeric)
S3method(output_replay,character)
S3method(output_replay,error)
S3method(output_replay,message)
S3method(output_replay,recordedplot)
S3method(output_replay,source)
S3method(output_replay,warning)
S3method(print,comparison)
S3method(print,expectation)
S3method(print,mismatch_character)
S3method(print,mismatch_numeric)
S3method(print,testthat_hint)
S3method(print,testthat_results)
S3method(snapshot_replay,character)
S3method(snapshot_replay,condition)
S3method(snapshot_replay,source)
S3method(test_mock_method,integer)
S3method(testthat_print,default)
export("%>%")
export(CheckReporter)
export(CompactProgressReporter)
export(DebugReporter)
export(FailReporter)
export(JunitReporter)
export(ListReporter)
export(LlmReporter)
export(LocationReporter)
export(MinimalReporter)
export(MultiReporter)
export(ParallelProgressReporter)
export(ProgressReporter)
export(RStudioReporter)
export(Reporter)
export(SilentReporter)
export(SlowReporter)
export(StopReporter)
export(SummaryReporter)
export(TapReporter)
export(TeamcityReporter)
export(announce_snapshot_file)
export(auto_test)
export(auto_test_package)
export(capture_condition)
export(capture_error)
export(capture_expectation)
export(capture_message)
export(capture_messages)
export(capture_output)
export(capture_output_lines)
export(capture_warning)
export(capture_warnings)
export(check_reporter)
export(compare)
export(compare_file_binary)
export(compare_file_text)
export(context)
export(context_start_file)
export(default_compact_reporter)
export(default_reporter)
export(describe)
export(edition_get)
export(equals)
export(equals_reference)
export(evaluate_promise)
export(exp_signal)
export(expect)
export(expect_all_equal)
export(expect_all_false)
export(expect_all_true)
export(expect_condition)
export(expect_contains)
export(expect_cpp_tests_pass)
export(expect_disjoint)
export(expect_equal)
export(expect_equal_to_reference)
export(expect_equivalent)
export(expect_error)
export(expect_failure)
export(expect_false)
export(expect_gt)
export(expect_gte)
export(expect_identical)
export(expect_in)
export(expect_invisible)
export(expect_is)
export(expect_known_hash)
export(expect_known_output)
export(expect_known_value)
export(expect_length)
export(expect_less_than)
export(expect_lt)
export(expect_lte)
export(expect_mapequal)
export(expect_match)
export(expect_message)
export(expect_more_than)
export(expect_named)
export(expect_no_condition)
export(expect_no_error)
export(expect_no_failure)
export(expect_no_match)
export(expect_no_message)
export(expect_no_success)
export(expect_no_warning)
export(expect_null)
export(expect_output)
export(expect_output_file)
export(expect_r6_class)
export(expect_reference)
export(expect_s3_class)
export(expect_s4_class)
export(expect_s7_class)
export(expect_setequal)
export(expect_shape)
export(expect_silent)
export(expect_snapshot)
export(expect_snapshot_error)
export(expect_snapshot_failure)
export(expect_snapshot_file)
export(expect_snapshot_output)
export(expect_snapshot_value)
export(expect_snapshot_warning)
export(expect_success)
export(expect_that)
export(expect_true)
export(expect_type)
export(expect_vector)
export(expect_visible)
export(expect_warning)
export(expectation)
export(extract_test)
export(fail)
export(find_test_scripts)
export(get_reporter)
export(gives_warning)
export(has_names)
export(is.expectation)
export(is_a)
export(is_checking)
export(is_equivalent_to)
export(is_identical_to)
export(is_informative_error)
export(is_less_than)
export(is_more_than)
export(is_parallel)
export(is_snapshot)
export(is_testing)
export(it)
export(local_edition)
export(local_mock)
export(local_mocked_bindings)
export(local_mocked_r6_class)
export(local_mocked_s3_method)
export(local_mocked_s4_method)
export(local_on_cran)
export(local_reproducible_output)
export(local_snapshotter)
export(local_test_context)
export(local_test_directory)
export(make_expectation)
export(mock_output_sequence)
export(new_expectation)
export(not)
export(pass)
export(prints_text)
export(quasi_label)
export(run_cpp_tests)
export(set_max_fails)
export(set_reporter)
export(set_state_inspector)
export(setup)
export(show_failure)
export(shows_message)
export(simulate_test_env)
export(skip)
export(skip_if)
export(skip_if_not)
export(skip_if_not_installed)
export(skip_if_offline)
export(skip_if_translated)
export(skip_on_appveyor)
export(skip_on_bioc)
export(skip_on_ci)
export(skip_on_covr)
export(skip_on_cran)
export(skip_on_os)
export(skip_on_travis)
export(skip_unless_r)
export(snapshot_accept)
export(snapshot_download_gh)
export(snapshot_reject)
export(snapshot_review)
export(source_dir)
export(source_file)
export(source_test_helpers)
export(source_test_setup)
export(source_test_teardown)
export(succeed)
export(takes_less_than)
export(teardown)
export(teardown_env)
export(test_check)
export(test_dir)
export(test_env)
export(test_example)
export(test_examples)
export(test_file)
export(test_local)
export(test_package)
export(test_path)
export(test_rd)
export(test_that)
export(testing_package)
export(testthat_example)
export(testthat_examples)
export(testthat_print)
export(testthat_tolerance)
export(throws_error)
export(try_again)
export(use_catch)
export(verify_output)
export(watch)
export(with_mock)
export(with_mocked_bindings)
export(with_reporter)
import(rlang)
importFrom(R6,R6Class)
importFrom(brio,readLines)
importFrom(brio,writeLines)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(methods,new)
useDynLib(testthat, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# testthat <a href="https://testthat.r-lib.org"><img src="man/figures/logo.png" align="right" height="138" alt="" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/testthat)](https://cran.r-project.org/package=testthat)
[![R-CMD-check](https://github.com/r-lib/testthat/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/testthat/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/testthat/graph/badge.svg)](https://app.codecov.io/gh/r-lib/testthat)
<!-- badges: end -->

## Overview

Testing your code can be painful and tedious, but it greatly increases
the quality of your code. **testthat** tries to make testing as fun as
possible, so that you get a visceral satisfaction from writing tests.
Testing should be addictive, so you do it all the time. To make that
happen, testthat:

- Provides functions that make it easy to describe what you expect a
  function to do, including catching errors, warnings, and messages.

- Easily integrates in your existing workflow, whether it’s informal
  testing on the command line, building test suites, or using R CMD
  check.

- Displays test progress visually, showing a pass, fail, or error for
  every expectation. If you’re using the terminal or a recent version of
  RStudio, it’ll even colour the output.

testthat draws inspiration from the xUnit family of testing packages, as
well as from many of the innovative ruby testing libraries, like
[rspec](https://rspec.info/), [testy](https://github.com/ahoward/testy),
[bacon](https://github.com/leahneukirchen/bacon) and
[cucumber](https://cucumber.io).

testthat is the most popular unit testing package for R and is used by
thousands of CRAN packages.

If you’re not familiar with testthat, the [testing
chapter](https://r-pkgs.org/testing-basics.html) in [R
packages](https://r-pkgs.org) gives a good overview, along with workflow
advice and concrete examples.

## Installation

``` r
# Install the released version from CRAN
install.packages("testthat")

# Or the development version from GitHub:
# install.packages("pak")
pak::pak("r-lib/testthat")
```

## Usage

The easiest way to get started is with
[usethis](https://github.com/r-lib/usethis). Assuming you’re in a
package directory, just run `usethis::use_test("name")` to create a test
file, and set up all the other infrastructure you need. If you’re using
RStudio, press Cmd/Ctrl + Shift + T (or run `devtools::test()` if not)
to run all the tests in a package.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(testthat)

## -----------------------------------------------------------------------------
# # Only run test if a token file is available
# skip_if_not(file.exists("secure-token.json"))
# 
# # Only run test if R has memory profiling capabilities
# skip_if_not(capabilities("profmem"))
# 
# # Only run if we've opted-in to slow tests with an env var
# skip_if(Sys.getenv("RUN_SLOW_TESTS") == "true")

## -----------------------------------------------------------------------------
skip_if_dangerous <- function() {
  if (!identical(Sys.getenv("DANGER"), "")) {
    skip("Not run in dangerous environments.")
  } else {
    invisible()
  }
}

## ----eval = FALSE-------------------------------------------------------------
# convert_markdown_to_html <- function(in_path, out_path, ...) {
#   if (rmarkdown::pandoc_available("2.0")) {
#     from <- "markdown+gfm_auto_identifiers-citations+emoji+autolink_bare_uris"
#   } else if (rmarkdown::pandoc_available("1.12.3")) {
#     from <- "markdown_github-hard_line_breaks+tex_math_dollars+tex_math_single_backslash+header_attributes"
#   } else {
#     if (is_testing()) {
#       testthat::skip("Pandoc not available")
#     } else {
#       abort("Pandoc not available")
#     }
#   }
# 
#   ...
# }

## -----------------------------------------------------------------------------
is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

# This dummy function definition is included with the package to ensure that
# 'tools::package_native_routine_registration_skeleton()' generates the required
# registration info for the 'run_testthat_tests' symbol.
(function() {
  .Call("run_testthat_tests", FALSE, PACKAGE = "%s")
})

#' Determine testing status
#'
#' @description
#' These functions help you determine if you code is running in a particular
#' testing context:
#'
#' * `is_testing()` is `TRUE` inside a test.
#' * `is_snapshot()` is `TRUE` inside a snapshot test
#' * `is_checking()` is `TRUE` inside of `R CMD check` (i.e. by [test_check()]).
#' * `is_parallel()` is `TRUE` if the tests are run in parallel.
#' * `testing_package()` gives name of the package being tested.
#'
#' A common use of these functions is to compute a default value for a `quiet`
#' argument with `is_testing() && !is_snapshot()`. In this case, you'll
#' want to avoid an run-time dependency on testthat, in which case you should
#' just copy the implementation of these functions into a `utils.R` or similar.
#'
#' @export
is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

#' @export
#' @rdname is_testing
is_parallel <- function() {
  identical(Sys.getenv("TESTTHAT_IS_PARALLEL"), "true")
}

#' @export
#' @rdname is_testing
is_checking <- function() {
  identical(Sys.getenv("TESTTHAT_IS_CHECKING"), "true")
}

#' @export
#' @rdname is_testing
is_snapshot <- function() {
  identical(Sys.getenv("TESTTHAT_IS_SNAPSHOT"), "true")
}

#' @export
#' @rdname is_testing
testing_package <- function() {
  Sys.getenv("TESTTHAT_PKG")
}

#' Generate default testing environment.
#'
#' We use a new environment which inherits from [globalenv()] or a package
#' namespace. In an ideal world, we'd avoid putting the global environment on
#' the search path for tests, but it's not currently possible without losing
#' the ability to load packages in tests.
#'
#' @keywords internal
#' @export
test_env <- function(package = NULL) {
  if (is.null(package)) {
    env(globalenv())
  } else {
    # Must clone environment so that during R CMD check, it's not locked
    # preventing creation of S4 classes
    env_clone(asNamespace(package))
  }
}

test_that("local context is 'as promised' inside test_that()", {
  # high-level expectations or expectations that preceded testthat 3e
  expect_true(is_testing())
  expect_equal(testing_package(), "testthat")
  expect_false(is_interactive())
  expect_equal(Sys.getenv("R_TESTS"), "")

  # set in local_test_context()
  expect_equal(Sys.getenv("TESTTHAT"), "true")

  # testthat 3e, set in local_reproducible_output()
  expect_equal(edition_get(), 3L)
  expect_equal(getOption("width"), 80)
  expect_false(getOption("crayon.enabled"))
  expect_false(getOption("cli.dynamic"))
  expect_false(getOption("cli.unicode"))
  expect_equal(getOption("lifecycle_verbosity"), "warning")
  expect_equal(getOption("OutDec"), ".")
  expect_false(getOption("rlang_interactive"))
  expect_false(getOption("useFancyQuotes"))
  expect_equal(getOption("max.print"), 99999)

  expect_equal(Sys.getenv("RSTUDIO"), "")
  expect_equal(Sys.getenv("LANGUAGE"), "C")

  expect_equal(Sys.getlocale("LC_COLLATE"), "C")
})

test_that("can override usual options", {
  local_test_directory(tempdir(), "methods")
  expect_equal(testing_package(), "methods")
})

test_that("can override translation of error messages", {
  skip_on_cran()

  local_reproducible_output(lang = "fr")
  expect_error(mean[[1]], "objet de type")
  local_reproducible_output(lang = "es")
  expect_error(mean[[1]], "objeto de tipo")
})

test_that("can force cli to display RStudio style hyperlinks", {
  expect_snapshot({
    str(cli::ansi_hyperlink_types())
  })

  local_reproducible_output(crayon = TRUE, hyperlinks = TRUE, rstudio = TRUE)
  expect_snapshot({
    str(cli::ansi_hyperlink_types())
  })
})

test_that("environment has package name", {
  expect_equal(methods::getPackageName(test_env("testthat")), "testthat")
  expect_equal(methods::getPackageName(topenv()), "testthat")
})

setClass("MyClass")
test_that("Cannot create S4 class without special behaviour", {
  expect_error(setClass("MyClass2"), NA)
})

test_that("is_checking respects env var", {
  withr::local_envvar(TESTTHAT_IS_CHECKING = "true")
  expect_true(is_checking())

  withr::local_envvar(TESTTHAT_IS_CHECKING = "false")
  expect_false(is_checking())
})

test_that("is_snapshot() is true in snapshots", {
  local_edition(3) # why is this needed?
  expect_false(is_snapshot())

  expect_snapshot(is_snapshot())
  expect_snapshot_value(is_snapshot())
  expect_snapshot_output(is_snapshot())
  expect_snapshot_warning(if (is_snapshot()) warning("Is snapshotting!"))
  expect_snapshot_error(if (is_snapshot()) stop("Is snapshotting!"))
})

#' Source a file, directory of files, or various important subsets
#'
#' These are used by [test_dir()] and friends
#'
#' @param path Path to files.
#' @param pattern Regular expression used to filter files.
#' @param env Environment in which to evaluate code.
#' @param desc A character vector used to filter tests. This is used to
#'   (recursively) filter the content of the file, so that only the non-test
#'   code up to and including the matching test is run.
#' @param chdir Change working directory to `dirname(path)`?
#' @param wrap Automatically wrap all code within [test_that()]? This ensures
#'   that all expectations are reported, even if outside a test block.
#' @param shuffle If `TRUE`, randomly reorder the top-level expressions
#'   in the file.
#' @export
#' @keywords internal
source_file <- function(
  path,
  env = test_env(),
  chdir = TRUE,
  desc = NULL,
  wrap = TRUE,
  shuffle = FALSE,
  error_call = caller_env()
) {
  check_string(path, call = error_call)
  if (!file.exists(path)) {
    cli::cli_abort("{.arg path} does not exist.", call = error_call)
  }
  if (!is.environment(env)) {
    stop_input_type(env, "an environment", call = error_call)
  }
  check_character(desc, allow_null = TRUE)

  lines <- brio::read_lines(path)
  srcfile <- srcfilecopy(
    path,
    lines,
    file.info(path)[1, "mtime"],
    isFile = TRUE
  )

  ## We need to parse from a connection, because parse() has a bug,
  ## and converts the input to the native encoding, if the text arg is used
  con <- textConnection(lines, encoding = "UTF-8")
  withr::defer(try(close(con), silent = TRUE))
  exprs <- parse(con, n = -1, srcfile = srcfile, encoding = "UTF-8")
  if (shuffle) {
    exprs <- sample(exprs)
  }
  exprs <- filter_desc(exprs, desc, error_call = error_call)

  n <- length(exprs)
  if (n == 0L) {
    return(invisible())
  }

  if (chdir) {
    old_dir <- setwd(dirname(path))
    withr::defer(setwd(old_dir))
  }

  withr::local_options(testthat_topenv = env, testthat_path = path)
  if (wrap) {
    invisible(test_code(
      code = exprs,
      env = env,
      reporter = get_reporter() %||% StopReporter$new()
    ))
  } else {
    withCallingHandlers(
      invisible(eval(exprs, env)),
      error = function(err) {
        cli::cli_abort(
          "Failed to evaluate {.path {path}}.",
          parent = err,
          call = error_call
        )
      }
    )
  }
}

filter_desc <- function(exprs, descs, error_call = caller_env()) {
  if (length(descs) == 0) {
    return(exprs)
  }
  desc <- descs[[1]]

  subtest_idx <- which(unname(map_lgl(exprs, is_subtest)))

  matching_idx <- keep(subtest_idx, \(idx) exprs[[idx]][[2]] == desc)
  if (length(matching_idx) == 0) {
    cli::cli_abort(
      "Failed to find test with description {.str {desc}}.",
      call = error_call
    )
  } else if (length(matching_idx) > 1) {
    cli::cli_abort(
      "Found multiple tests with description {.str {desc}}.",
      call = error_call
    )
  }

  # Want all code up to and including the matching test, except for subtests
  keep_idx <- setdiff(seq2(1, matching_idx), setdiff(subtest_idx, matching_idx))
  # Recursively inspect the components of the subtest
  exprs[[matching_idx]][[3]] <- filter_desc(
    exprs[[matching_idx]][[3]],
    descs[-1],
    error_call = error_call
  )
  exprs[keep_idx]
}

is_subtest <- function(expr) {
  is_call(expr, c("test_that", "describe", "it"), n = 2) &&
    is_string(expr[[2]]) &&
    is_call(expr[[3]], "{")
}

#' @rdname source_file
#' @export
source_dir <- function(
  path,
  pattern = "\\.[rR]$",
  env = test_env(),
  chdir = TRUE,
  wrap = TRUE,
  shuffle = FALSE
) {
  files <- sort(dir(path, pattern, full.names = TRUE))

  error_call <- current_env()
  lapply(files, function(path) {
    source_file(
      path,
      env = env,
      chdir = chdir,
      wrap = wrap,
      shuffle = shuffle,
      error_call = error_call
    )
  })
}

#' @rdname source_file
#' @export
source_test_helpers <- function(path = "tests/testthat", env = test_env()) {
  source_dir(path, "^helper.*\\.[rR]$", env = env, wrap = FALSE)
}

#' @rdname source_file
#' @export
source_test_setup <- function(path = "tests/testthat", env = test_env()) {
  source_dir(path, "^setup.*\\.[rR]$", env = env, wrap = FALSE)
}

#' @rdname source_file
#' @export
source_test_teardown <- function(path = "tests/testthat", env = test_env()) {
  source_dir(path, "^teardown.*\\.[rR]$", env = env, wrap = FALSE)
}

#' Capture output to console
#'
#' Evaluates `code` in a special context in which all output is captured,
#' similar to [capture.output()].
#'
#' Results are printed using the `testthat_print()` generic, which defaults
#' to `print()`, giving you the ability to customise the printing of your
#' object in tests, if needed.
#'
#' @param code Code to evaluate.
#' @param print If `TRUE` and the result of evaluating `code` is
#'   visible, print the result using `testthat_print()`.
#' @param width Number of characters per line of output. This does not
#'   inherit from `getOption("width")` so that tests always use the same
#'   output width, minimising spurious differences.
#' @return `capture_output()` returns a single string. `capture_output_lines()`
#'   returns a character vector with one entry for each line
#' @keywords internal
#' @export
#' @examples
#' capture_output({
#'   cat("Hi!\n")
#'   cat("Bye\n")
#' })
#'
#' capture_output_lines({
#'   cat("Hi!\n")
#'   cat("Bye\n")
#' })
#'
#' capture_output("Hi")
#' capture_output("Hi", print = TRUE)
capture_output <- function(code, print = FALSE, width = 80) {
  output <- capture_output_lines(code, print, width = width)
  paste0(output, collapse = "\n")
}

#' @export
#' @rdname capture_output
capture_output_lines <- function(code, print = FALSE, width = 80) {
  eval_with_output(code, print = print, width = width)$out
}

eval_with_output <- function(code, print = FALSE, width = 80) {
  path <- withr::local_tempfile()

  if (!is.null(width)) {
    local_width(width)
  }

  result <- withr::with_output_sink(path, withVisible(code))
  if (result$visible && print) {
    withr::with_output_sink(path, testthat_print(result$value), append = TRUE)
  }

  # A sink() will always write in the native encoding, so we read with
  # base::readLines() then convert to UTF-8
  list(
    val = result$value,
    vis = result$visible,
    out = enc2utf8(base::readLines(path, warn = FALSE))
  )
}

#' @export
#' @rdname capture_output
testthat_print <- function(x) {
  UseMethod("testthat_print")
}

#' @export
testthat_print.default <- function(x) {
  print(x)
}

test_that("make_expectation returns and prints expectation", {

  x <- 1:5
  out <- capture_output(
    expect_equal(make_expectation(x), bquote(expect_equal(x, .(1:5))))
  )
  expect_equal(
    out,
    "expect_equal(x, 1:5)"
  )
})

test_that("multiplication works", {
  utf8 <- "M\u00e4chler"
  latin1 <- "M\xe4chler"
  Encoding(latin1) <- "latin1"

  expect_equal(capture_output_lines(cat(latin1)), utf8, ignore_encoding = FALSE)
})

test_that("capture output captures output", {
  out1 <- capture_output(print(1:5))
  out2 <- capture_output(1:5, print = TRUE)

  expect_equal(out1, "[1] 1 2 3 4 5")
  expect_equal(out2, "[1] 1 2 3 4 5")
})

test_that("capture output doesn't print invisible things", {
  out <- capture_output(invisible(1), print = TRUE)
  expect_equal(out, "")
})

find_edition <- function(path, package = NULL) {
  from_environment <- Sys.getenv("TESTTHAT_EDITION")
  if (nzchar(from_environment)) {
    return(as.integer(from_environment))
  }
  desc <- find_description(path, package)
  if (is.null(desc)) {
    return(2L)
  }

  as.integer(desc$get_field("Config/testthat/edition", default = 2L))
}

find_description <- function(path, package = NULL) {
  if (!is.null(package)) {
    return(desc::desc(package = package))
  } else {
    tryCatch(
      pkgload::pkg_desc(path),
      error = function(e) NULL
    )
  }
}

edition_deprecate <- function(in_edition, what, instead = NULL) {
  if (edition_get() < in_edition) {
    return()
  }

  warn(c(
    paste0("`", what, "` was deprecated in ", edition_name(in_edition), "."),
    i = instead
  ))
}

edition_require <- function(in_edition, what) {
  if (edition_get() >= in_edition || isTRUE(getOption("testthat.edition_ignore"))) {
    return()
  }

  stop(paste0("`", what, "` requires ", edition_name(in_edition), "."))
}

edition_name <- function(x) {
  if (x == 2) {
    "the 2nd edition"
  } else if (x == 3) {
    "the 3rd edition"
  } else {
    paste("edition ", x)
  }
}

#' Temporarily change the active testthat edition
#'
#' `local_edition()` allows you to temporarily (within a single test or
#' a single test file) change the active edition of testthat.
#' `edition_get()` allows you to retrieve the currently active edition.
#'
#' @export
#' @param x Edition Should be a single integer.
#' @param .env Environment that controls scope of changes. For expert use only.
local_edition <- function(x, .env = parent.frame()) {
  stopifnot(is_zap(x) || (is.numeric(x) && length(x) == 1))

  local_bindings(edition = x, .env = the, .frame = .env)
}


#' @export
#' @rdname local_edition
edition_get <- function() {
  the$edition %||% find_edition(".")
}

test_that("test_examples works with installed packages", {
  local_edition(2)

  local_mocked_bindings(test_rd = identity)
  expect_true(length(test_examples()) > 1)
})

test_that("test_examples fails if no examples", {
  withr::local_envvar(TESTTHAT_PKG = "")
  expect_error(test_examples("asdf"), "Could not find examples")
})

test_that("always returns a path",{
  withr::local_envvar(TESTTHAT = "true")
  withr::local_options(testthat_interactive = FALSE)

  expect_equal(test_path(), ".")
})

test_that("is vectorised", {
  withr::local_envvar(TESTTHAT = "true")
  withr::local_options(testthat_interactive = FALSE)

  expect_equal(test_path("x", c("a", "b")), c("x/a", "x/b"))
})

test_that("uses local path when called from test_file()/tools::testInstalledPackages()", {
  withr::local_envvar(TESTTHAT = "true")
  withr::local_options(testthat_interactive = FALSE)

  expect_equal(test_path("path"), "path")
})

test_that("returns local path during package loading", {
  withr::local_envvar(TESTTHAT = "false", DEVTOOLS_LOAD = "testthat")

  expect_equal(test_path("path"), "path")
})

test_that("returns full path when called interactively", {
  withr::local_envvar("TESTTHAT" = "false")
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "tests", "testthat"), recursive = TRUE)
  withr::local_dir(pkg)

  expect_equal(test_path("path"), "tests/testthat/path")
})

test_that("throws error if can't find tests/testthat", {
  withr::local_envvar("TESTTHAT" = "false")
  withr::local_dir(withr::local_tempdir())

  local_edition(3)
  local_reproducible_output()

  expect_snapshot(test_path("empty"), error = TRUE)
})

# test_dir() --------------------------------------------------------------

test_that("stops on failure", {
  withr::local_envvar(TESTTHAT_PARALLEL = "FALSE")
  expect_snapshot(error = TRUE, {
    test_dir(test_path("test_dir"), reporter = "silent")
  })
})

test_that("runs all tests and records output", {
  withr::local_envvar(TESTTHAT_PARALLEL = "FALSE")
  res <- test_dir(
    test_path("test_dir"),
    reporter = "silent",
    stop_on_failure = FALSE
  )
  df <- as.data.frame(res)
  df$user <- df$system <- df$real <- df$result <- NULL

  local_reproducible_output(width = 200)
  local_edition(3) # set to 2 in ./test_dir
  expect_snapshot_output(print(df))
})

test_that("complains if no files", {
  withr::local_envvar(TESTTHAT_PARALLEL = "FALSE")
  path <- withr::local_tempfile()
  dir.create(path)

  expect_snapshot(error = TRUE, test_dir(path))
})

test_that("can control if failures generate errors", {
  withr::local_envvar(TESTTHAT_PARALLEL = "FALSE")
  test_error <- function(...) {
    test_dir(test_path("test-error"), reporter = "silent", ...)
  }

  expect_snapshot(error = TRUE, test_error(stop_on_failure = TRUE))
  expect_no_error(test_error(stop_on_failure = FALSE))
})

test_that("can control if warnings errors", {
  withr::local_envvar(TESTTHAT_PARALLEL = "FALSE")
  test_warning <- function(...) {
    test_dir(test_path("test-warning"), reporter = "silent", ...)
  }

  expect_snapshot(error = TRUE, test_warning(stop_on_warning = TRUE))
  expect_no_error(test_warning(stop_on_warning = FALSE))
})

# test_file ---------------------------------------------------------------

test_that("can test single file", {
  out <- test_file(test_path("test_dir/test-basic.R"), reporter = "silent")
  expect_length(out, 5)
})

test_that("complains if file doesn't exist", {
  expect_snapshot(error = TRUE, test_file("DOESNTEXIST"))
})


# setup-teardown ----------------------------------------------------------

test_that("files created by setup still exist", {
  # These files should be created/delete by package-wide setup/teardown
  # We check that they exist here to make sure that they're not cleaned up
  # too early
  expect_true(file.exists("DELETE-ME"))
  expect_true(file.exists("DELETE-ME-2"))
})

# helpers -----------------------------------------------------------------

test_that("can filter test scripts", {
  x <- c("test-a.R", "test-b.R", "test-c.R")
  expect_equal(filter_test_scripts(x), x)
  expect_equal(filter_test_scripts(x, "a"), x[1])
  expect_equal(filter_test_scripts(x, "a", invert = TRUE), x[-1])

  # Strips prefix/suffix
  expect_equal(filter_test_scripts(x, "test"), character())
  expect_equal(filter_test_scripts(x, ".R"), character())
})

# ----------------------------------------------------------------------

test_that("can configure `load_all()` (#1636)", {
  path <- test_path("testConfigLoadAll")

  args <- find_load_all_args(path)
  expect_equal(args, list(export_all = FALSE, helpers = FALSE))

  results <- test_local(path, reporter = "silent")
  for (res in results) {
    expect_equal(sum(res[["failed"]]), 0)
  }
})

test_that("helpers are included in the testing environment", {
  expect_true("abcdefghi" %in% names(the$testing_env))
})

#' Do you expect this value?
#'
#' @description
#' These functions provide two levels of strictness when comparing a
#' computation to a reference value. `expect_identical()` is the baseline;
#' `expect_equal()` relaxes the test to ignore small numeric differences.
#'
#' In the 2nd edition, `expect_identical()` uses [identical()] and
#' `expect_equal` uses [all.equal()]. In the 3rd edition, both functions use
#' [waldo](https://github.com/r-lib/waldo). They differ only in that
#' `expect_equal()` sets `tolerance = testthat_tolerance()` so that small
#' floating point differences are ignored; this also implies that (e.g.) `1`
#' and `1L` are treated as equal.
#'
#' @param object,expected Computation and value to compare it to.
#'
#'   Both arguments supports limited unquoting to make it easier to generate
#'   readable failures within a function or for loop. See [quasi_label] for
#'   more details.
#' @param ...
#'   **3e**: passed on to [waldo::compare()]. See its docs to see other
#'   ways to control comparison.
#'
#'   **2e**: passed on to [testthat::compare()]/[identical()].
#' @param tolerance
#'   **3e**: passed on to [waldo::compare()]. If non-`NULL`, will
#'   ignore small floating point differences. It uses same algorithm as
#'   [all.equal()] so the tolerance is usually relative (i.e.
#'   `mean(abs(x - y) / mean(abs(y)) < tolerance`), except when the differences
#'   are very small, when it becomes absolute (i.e. `mean(abs(x - y) < tolerance`).
#'   See waldo documentation for more details.
#'
#'   **2e**: passed on to [testthat::compare()], if set. It's hard to
#'   reason about exactly what tolerance means because depending on the precise
#'   code path it could be either an absolute or relative tolerance.
#' @param label,expected.label Used to customise failure messages. For expert
#'   use only.
#' @seealso
#' * [expect_setequal()]/[expect_mapequal()] to test for set equality.
#' * [expect_reference()] to test if two names point to same memory address.
#' @inheritParams expect_that
#' @family expectations
#' @examples
#' a <- 10
#' expect_equal(a, 10)
#'
#' # Use expect_equal() when testing for numeric equality
#' \dontrun{
#' expect_identical(sqrt(2) ^ 2, 2)
#' }
#' expect_equal(sqrt(2) ^ 2, 2)
#' @name equality-expectations
NULL

#' @export
#' @rdname equality-expectations
expect_equal <- function(
  object,
  expected,
  ...,
  tolerance = if (edition_get() >= 3) testthat_tolerance(),
  info = NULL,
  label = NULL,
  expected.label = NULL
) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)
  check_number_decimal(tolerance, min = 0, allow_null = TRUE)

  if (edition_get() >= 3) {
    msg <- "Expected %s to equal %s."
    expect_waldo_equal_(msg, act, exp, info, ..., tolerance = tolerance)
  } else {
    if (!is.null(tolerance)) {
      comp <- compare(act$val, exp$val, ..., tolerance = tolerance)
    } else {
      comp <- compare(act$val, exp$val, ...)
    }

    if (comp$equal) {
      pass()
    } else {
      msg <- c(
        sprintf("Expected %s to equal %s.", act$lab, exp$lab),
        "Differences:",
        comp$message
      )
      fail(msg, info = info)
    }
  }
  invisible(act$val)
}


#' @export
#' @rdname equality-expectations
expect_identical <- function(
  object,
  expected,
  info = NULL,
  label = NULL,
  expected.label = NULL,
  ...
) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  if (edition_get() >= 3) {
    msg <- "Expected %s to be identical to %s."
    expect_waldo_equal_(msg, act, exp, info, ...)
  } else {
    if (identical(act$val, exp$val, ...)) {
      pass()
    } else {
      compare <- compare(act$val, exp$val)
      if (compare$equal) {
        msg_act <- "Objects equal but not identical"
      } else {
        msg_act <- compare$message
      }

      msg <- c(
        sprintf("Expected %s to be identical to %s.", act$lab, exp$lab),
        "Differences:",
        msg_act
      )
      fail(msg, info = info)
    }
  }
  invisible(act$val)
}

expect_waldo_equal_ <- function(
  msg,
  act,
  exp,
  info = NULL,
  ...,
  trace_env = caller_env()
) {
  comp <- waldo_compare(
    act$val,
    exp$val,
    ...,
    x_arg = "actual",
    y_arg = "expected"
  )
  if (length(comp) == 0) {
    pass()
  } else {
    msg <- c(
      sprintf(msg, act$lab, exp$lab),
      "Differences:",
      paste0(comp, "\n")
    )
    fail(msg, info = info, trace_env = trace_env)
  }
}

#' Is an object equal to the expected value, ignoring attributes?
#'
#' Compares `object` and `expected` using [all.equal()] and
#' `check.attributes = FALSE`.
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' `expect_equivalent()` is deprecated in the 3rd edition. Instead use
#' `expect_equal(ignore_attr = TRUE)`.
#'
#' @inheritParams expect_equal
#' @param ... Passed on to [compare()].
#' @keywords internal
#' @export
#' @examples
#' #' # expect_equivalent() ignores attributes
#' a <- b <- 1:3
#' names(b) <- letters[1:3]
#' \dontrun{
#' expect_equal(a, b)
#' }
#' expect_equivalent(a, b)
expect_equivalent <- function(
  object,
  expected,
  ...,
  info = NULL,
  label = NULL,
  expected.label = NULL
) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  edition_deprecate(
    3,
    "expect_equivalent()",
    "Use expect_equal(ignore_attr = TRUE)"
  )

  comp <- compare(act$val, exp$val, ..., check.attributes = FALSE)
  if (!comp$equal) {
    msg <- sprintf(
      "Expected %s to be equivalent to %s.\n%s",
      act$lab,
      exp$lab,
      comp$message
    )
    fail(msg, info = info)
  } else {
    pass()
  }
  invisible(act$val)
}

#' Temporarily set options for maximum reproducibility
#'
#' @description
#' `local_test_context()` is run automatically by `test_that()` but you may
#' want to run it yourself if you want to replicate test results interactively.
#' If run inside a function, the effects are automatically reversed when the
#' function exits; if running in the global environment, use
#' [withr::deferred_run()] to undo.
#'
#' `local_reproducible_output()` is run automatically by `test_that()` in the
#' 3rd edition. You might want to call it to override the the default settings
#' inside a test, if you want to test Unicode, coloured output, or a
#' non-standard width.
#'
#' @details
#' `local_test_context()` sets `TESTTHAT = "true"`, which ensures that
#' [is_testing()] returns `TRUE` and allows code to tell if it is run by
#' testthat.
#'
#' In the third edition, `local_test_context()` also calls
#' `local_reproducible_output()` which temporary sets the following options:
#'
#' * `cli.dynamic = FALSE` so that tests assume that they are not run in
#'   a dynamic console (i.e. one where you can move the cursor around).
#' * `cli.unicode` (default: `FALSE`) so that the cli package never generates
#'   unicode output (normally cli uses unicode on Linux/Mac but not Windows).
#'   Windows can't easily save unicode output to disk, so it must be set to
#'   false for consistency.
#' * `cli.condition_width = Inf` so that new lines introduced while
#'   width-wrapping condition messages don't interfere with message matching.
#' * `crayon.enabled` (default: `FALSE`) suppresses ANSI colours generated by
#'   the cli and crayon packages (normally colours are used if cli detects
#'   that you're in a terminal that supports colour).
#' * `cli.num_colors` (default: `1L`) Same as the crayon option.
#' * `lifecycle_verbosity = "warning"` so that every lifecycle problem always
#'   generates a warning (otherwise deprecated functions don't generate a
#'   warning every time).
#' * `max.print = 99999` so the same number of values are printed.
#' * `OutDec = "."` so numbers always uses `.` as the decimal point
#'   (European users sometimes set `OutDec = ","`).
#' * `rlang_interactive = FALSE` so that [rlang::is_interactive()] returns
#'   `FALSE`, and code that uses it pretends you're in a non-interactive
#'   environment.
#' * `useFancyQuotes = FALSE` so base R functions always use regular (straight)
#'   quotes (otherwise the default is locale dependent, see [sQuote()] for
#'   details).
#' * `width` (default: 80) to control the width of printed output (usually this
#'   varies with the size of your console).
#'
#' And modifies the following env vars:
#'
#' * Unsets `RSTUDIO`, which ensures that RStudio is never detected as running.
#' * Sets `LANGUAGE = "en"`, which ensures that no message translation occurs.
#'
#' Finally, it sets the collation locale to "C", which ensures that character
#' sorting the same regardless of system locale.
#'
#' @export
#' @param .env Environment to use for scoping; expert use only.
#' @examples
#' local({
#'   local_test_context()
#'   cat(cli::col_blue("Text will not be colored"))
#'   cat(cli::symbol$ellipsis)
#'   cat("\n")
#' })
local_test_context <- function(.env = parent.frame()) {
  withr::local_envvar(
    "_R_CHECK_BROWSER_NONINTERACTIVE_" = "true",
    TESTTHAT = "true",
    .local_envir = .env
  )
  if (edition_get() >= 3) {
    local_reproducible_output(.env = .env)
  }
}

#' @export
#' @param width Value of the `"width"` option.
#' @param crayon Determines whether or not crayon (now cli) colour
#'   should be applied.
#' @param unicode Value of the `"cli.unicode"` option.
#'   The test is skipped if `` l10n_info()$`UTF-8` `` is `FALSE`.
#' @param rstudio Should we pretend that we're inside of RStudio?
#' @param hyperlinks Should we use ANSI hyperlinks.
#' @param lang Optionally, supply a BCP47 language code to set the language
#'   used for translating error messages. This is a lower case two letter
#'   [ISO 639 country code](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes),
#'   optionally followed by "_" or "-" and an upper case two letter
#'   [ISO 3166 region code](https://en.wikipedia.org/wiki/ISO_3166-2).
#' @rdname local_test_context
#' @examples
#' test_that("test ellipsis", {
#'   local_reproducible_output(unicode = FALSE)
#'   expect_equal(cli::symbol$ellipsis, "...")
#'
#'   local_reproducible_output(unicode = TRUE)
#'   expect_equal(cli::symbol$ellipsis, "\u2026")
#' })
local_reproducible_output <- function(
  width = 80,
  crayon = FALSE,
  unicode = FALSE,
  rstudio = FALSE,
  hyperlinks = FALSE,
  lang = "C",
  .env = parent.frame()
) {
  if (unicode) {
    # If you force unicode display, you _must_ skip the test on non-utf8
    # locales; otherwise it's guaranteed to fail
    skip_if(!l10n_info()$`UTF-8`, "non utf8 locale")
  }

  local_width(width = width, .env = .env)
  withr::local_options(
    crayon.enabled = crayon,
    cli.hyperlink = hyperlinks,
    cli.hyperlink_run = hyperlinks,
    cli.hyperlink_help = hyperlinks,
    cli.hyperlink_vignette = hyperlinks,
    cli.dynamic = FALSE,
    cli.unicode = unicode,
    cli.condition_width = Inf,
    cli.num_colors = if (crayon) 8L else 1L,
    useFancyQuotes = FALSE,
    lifecycle_verbosity = "warning",
    OutDec = ".",
    rlang_interactive = FALSE,
    max.print = 99999,
    .local_envir = .env,
  )
  withr::local_envvar(
    RSTUDIO = if (rstudio) 1 else NA,
    RSTUDIO_SESSION_PID = if (rstudio) Sys.getpid() else NA,
    RSTUDIO_CHILD_PROCESS_PANE = if (rstudio) "build" else NA,
    RSTUDIO_CLI_HYPERLINKS = if (rstudio) 1 else NA,
    .local_envir = .env
  )

  withr::local_language(lang, .local_envir = .env)
  withr::local_collate("C", .local_envir = .env)
}

local_reporter_output <- function(.env = parent.frame()) {
  reporter <- get_reporter()
  if (!is.null(reporter)) {
    reporter$local_user_output(.env)
  }
}

waldo_compare <- function(x, y, ..., x_arg = "x", y_arg = "y") {
  # Need to very carefully isolate this change to this function - can not set
  # in expectation functions because part of expectation handling bubbles
  # up through calling handlers, which are run before on.exit()
  if (!is_snapshot()) {
    local_reporter_output()
  }
  waldo::compare(x, y, ..., x_arg = x_arg, y_arg = y_arg)
}

local_width <- function(width = 80, .env = parent.frame()) {
  withr::local_options(width = width, cli.width = width, .local_envir = .env)
  withr::local_envvar(RSTUDIO_CONSOLE_WIDTH = width, .local_envir = .env)
}


#' Locally set test directory options
#'
#' For expert use only.
#'
#' @param path Path to directory of files
#' @param package Optional package name, if known.
#' @export
#' @keywords internal
local_test_directory <- function(path, package = NULL, .env = parent.frame()) {
  # Set edition before changing working directory in case path is relative
  local_edition(find_edition(path, package), .env = .env)

  # Capture current working directory so we can use for relative paths
  wd <- getwd()

  withr::local_dir(path, .local_envir = .env)
  withr::local_envvar(
    R_TESTS = "",
    TESTTHAT = "true",
    TESTTHAT_PKG = package,
    TESTTHAT_WD = wd,
    .local_envir = .env
  )
}

local_interactive_reporter <- function(.env = parent.frame()) {
  local_assume_not_on_cran(.env)
  withr::local_options(testthat_interactive = TRUE, .local_envir = .env)

  # Use edition from working directory
  local_edition(find_edition("."), .env = .env)

  # Use StopReporter
  reporter <- StopReporter$new()
  old <- set_reporter(reporter)
  withr::defer(reporter$end_reporter(), envir = .env)
  withr::defer(set_reporter(old), envir = .env)

  reporter
}

CountReporter <- R6::R6Class(
  "CountReporter",
  inherit = Reporter,
  public = list(
    context_i = 0,
    context_count = 0,
    test_i = 0,
    test_count = 0,

    start_context = function(context) {
      self$context_count <- self$context_count + 1
      self$context_i <- self$context_i + 1
    },
    end_context = function(context) {
      self$context_i <- self$context_i - 1
      stopifnot(self$context_i >= 0)
    },
    start_test = function(context, test) {
      self$test_count <- self$test_count + 1
      self$test_i <- self$test_i + 1
    },
    end_test = function(context, test) {
      self$test_i <- self$test_i - 1
      stopifnot(self$test_i >= 0)
    }
  )
)

test_that("contexts are opened, then closed", {
  local_edition(2)
  report <- CountReporter$new()
  local_description_set()

  with_reporter(report, test_one_file("context.R"))
  expect_equal(report$context_count, 2)
  expect_equal(report$context_i, 0)
  expect_equal(report$test_count, 4)
  expect_equal(report$test_i, 0)
})

test_that("context_name strips prefix and extensions correctly", {
  expect_equal(context_name("test-metrics.R"), "metrics") # uppercase
  expect_equal(context_name("test-metrics.r"), "metrics") # lowercase
  expect_equal(context_name("test-check.Rfile.R"), "check.Rfile") # suffix only
  expect_equal(context_name("test-test-test.R"), "test-test") # 1st prefix only
  expect_equal(context_name("test_metrics.R"), "metrics")
})

local_edition(2)

context("Successes")

test_that("Success", {
  succeed()
})

context("Failures")

test_that("Failure:1", {
  expect_true(FALSE)
})

test_that("Failure:2a", {
  f <- function() expect_true(FALSE)
  f()
})

context("Errors")

test_that("Error:1", {
  stop("stop")
})

test_that("errors get tracebacks", {
  f <- function() g()
  g <- function() h()
  h <- function() stop("!")

  f()
})

context("Skips")

test_that("explicit skips are reported", {
  skip("skip")
})

test_that("empty tests are implicitly skipped", {
})

context("Warnings")

test_that("warnings get backtraces", {
  f <- function() {
    warning("def")
  }
  f()
})

test_that("can locally override edition", {
  local_edition(3)
  expect_equal(edition_get(), 3)

  local_edition(2)
  expect_equal(edition_get(), 2)
})

test_that("deprecation only fired for newer edition", {
  local_edition(2)
  expect_warning(edition_deprecate(3, "old stuff"), NA)

  local_edition(3)
  expect_snapshot(edition_deprecate(3, "old stuff"))
})

test_that("required only fired for older edition", {
  withr::local_options(testthat.edition_ignore = FALSE)
  local_edition(2)
  expect_error(edition_require(3, "new stuff"))

  withr::local_options(testthat.edition_ignore = FALSE)
  local_edition(3)
  expect_error(edition_require(3, "new stuff"), NA)
})

test_that("edition for testthat is 3", {
  expect_equal(find_edition(package = "testthat"), 3)
})

test_that("edition for non-package dir is 2", {
  withr::local_envvar(TESTTHAT_EDITION = NULL)
  expect_equal(find_edition(tempdir()), 2)
})

test_that("can set the edition via an environment variable", {
  local_edition(zap())
  withr::local_envvar(TESTTHAT_EDITION = 2)
  expect_equal(edition_get(), 2)

  withr::local_envvar(TESTTHAT_EDITION = 3)
  expect_equal(edition_get(), 3)
})

#' Describe the context of a set of tests.
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' Use of `context()` is no longer recommended. Instead omit it, and messages
#' will use the name of the file instead. This ensures that the context and
#' test file name are always in sync.
#'
#' A context defines a set of tests that test related functionality.  Usually
#' you will have one context per file, but you may have multiple contexts
#' in a single file if you so choose.
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' `context()` is deprecated in the third edition, and the equivalent
#' information is instead recorded by the test file name.
#'
#' @param desc description of context.  Should start with a capital letter.
#' @keywords internal
#' @export
#' @examples
#' context("String processing")
#' context("Remote procedure calls")
context <- function(desc) {
  edition_deprecate(3, "context()")
  context_start(desc)
}

context_start <- function(desc) {
  reporter <- get_reporter()
  if (!is.null(reporter)) {
    get_reporter()$.start_context(desc)
  }
}

#' Start test context from a file name
#'
#' For use in external reporters
#'
#' @param name file name
#' @keywords internal
#' @export
context_start_file <- function(name) {
  context_start(context_name(name))
}

context_name <- function(filename) {
  # Remove test- prefix
  filename <- sub("^test[-_]", "", filename)
  # Remove terminal extension
  filename <- sub("[.][Rr]$", "", filename)
  filename
}

context("First context.")

test_that("Logical equivalence", {
  x <- TRUE
  expect_equal(x, TRUE)
})

test_that("Numerical equivalence", {
  x <- 1
  expect_equal(x, 1)
})

context("Second context.")

test_that("A passing test", {
  expect_equal(TRUE, TRUE)
})

test_that("A failing test", {
  expect_equal(TRUE, FALSE)
})

#' Watches code and tests for changes, rerunning tests as appropriate.
#'
#' The idea behind `auto_test()` is that you just leave it running while
#' you develop your code.  Every time you save a file it will be automatically
#' tested and you can easily see if your changes have caused any test
#'  failures.
#'
#' The current strategy for rerunning tests is as follows:
#'
#' - if any code has changed, then those files are reloaded and all tests
#'   rerun
#' - otherwise, each new or modified test is run
#'
#' In the future, `auto_test()` might implement one of the following more
#' intelligent alternatives:
#'
#' - Use codetools to build up dependency tree and then rerun tests only
#'   when a dependency changes.
#' - Mimic ruby's autotest and rerun only failing tests until they pass,
#'   and then rerun all tests.
#
#' @seealso [auto_test_package()]
#' @export
#' @param code_path path to directory containing code
#' @param test_path path to directory containing tests
#' @param reporter test reporter to use
#' @param env environment in which to execute test suite.
#' @param hash Passed on to [watch()]. When FALSE, uses less accurate
#'   modification time stamps, but those are faster for large files.
#' @keywords debugging
auto_test <- function(code_path, test_path, reporter = default_reporter(),
                      env = test_env(),
                      hash = TRUE) {
  reporter <- find_reporter(reporter)
  code_path <- normalizePath(code_path)
  test_path <- normalizePath(test_path)

  # Start by loading all code and running all tests
  source_dir(code_path, env = env)
  test_dir(test_path, env = env, reporter = reporter$clone(deep = TRUE))

  # Next set up watcher to monitor changes
  watcher <- function(added, deleted, modified) {
    changed <- normalizePath(c(added, modified))

    tests <- changed[starts_with(changed, test_path)]
    code <- changed[starts_with(changed, code_path)]

    if (length(code) > 0) {
      # Reload code and rerun all tests
      cat("Changed code: ", paste0(basename(code), collapse = ", "), "\n")
      cat("Rerunning all tests\n")
      source_dir(code_path, env = env)
      test_dir(test_path, env = env, reporter = reporter$clone(deep = TRUE))
    } else if (length(tests) > 0) {
      # If test changes, rerun just that test
      cat("Rerunning tests: ", paste0(basename(tests), collapse = ", "), "\n")
      test_files(tests, env = env, reporter = reporter$clone(deep = TRUE))
    }

    TRUE
  }
  watch(c(code_path, test_path), watcher, hash = hash)
}

#' Watches a package for changes, rerunning tests as appropriate.
#'
#' @param pkg path to package
#' @export
#' @param reporter test reporter to use
#' @param hash Passed on to [watch()].  When FALSE, uses less accurate
#'   modification time stamps, but those are faster for large files.
#' @keywords debugging
#' @seealso [auto_test()] for details on how method works
auto_test_package <- function(pkg = ".", reporter = default_reporter(), hash = TRUE) {
  reporter <- find_reporter(reporter)

  path <- pkgload::pkg_path(pkg)
  package <- pkgload::pkg_name(path)

  code_path <- file.path(path, c("R", "src"))
  code_path <- code_path[file.exists(code_path)]
  code_path <- normalizePath(code_path)
  test_path <- normalizePath(file.path(path, "tests", "testthat"))

  # Start by loading all code and running all tests
  withr::local_envvar("NOT_CRAN" = "true")
  pkgload::load_all(path)
  test_dir(test_path, package = package, reporter = reporter$clone(deep = TRUE), stop_on_failure = FALSE)

  # Next set up watcher to monitor changes
  watcher <- function(added, deleted, modified) {
    changed <- normalizePath(c(added, modified))

    tests <- changed[starts_with(changed, test_path)]
    code <- changed[starts_with(changed, code_path)]

    # Remove helper from test and add it to code (if a helper changed,
    # like for code, reload all and rerun all tests)
    helper <- tests[starts_with(basename(tests), "helper-")]
    tests <- setdiff(tests, helper)
    code <- c(code, helper)

    if (length(code) > 0) {
      # Reload code and rerun all tests
      cat("Changed code: ", paste0(basename(code), collapse = ", "), "\n")
      cat("Rerunning all tests\n")
      pkgload::load_all(path, quiet = TRUE)
      test_dir(test_path, package = package, reporter = reporter$clone(deep = TRUE))
    } else if (length(tests) > 0) {
      # If test changes, rerun just that test
      cat("Rerunning tests: ", paste0(basename(tests), collapse = ", "), "\n")
      env <- env_clone(asNamespace(package))
      test_files(
        test_dir = test_path,
        test_package = package,
        test_paths = tests,
        env = env, 
        reporter = reporter$clone(deep = TRUE)
      )
    }

    TRUE
  }
  watch(c(code_path, test_path), watcher, hash = hash)
}

# Helpers -----------------------------------------------------------------

starts_with <- function(string, prefix) {
  substr(string, 1, nchar(prefix)) == prefix
}

#' Interactively debug failing tests
#'
#' This reporter will call a modified version of [recover()] on all
#' broken expectations.
#'
#' @export
#' @family reporters
DebugReporter <- R6::R6Class(
  "DebugReporter",
  inherit = Reporter,
  public = list(
    add_result = function(context, test, result) {
      if (!expectation_success(result) && !is.null(result$start_frame)) {
        if (sink_number() > 0) {
          sink(self$out)
          withr::defer(sink())
        }

        recover2(
          start_frame = result$start_frame,
          end_frame = result$end_frame
        )
      }
    }
  )
)

sink_number <- function() {
  sink.number(type = "output")
}

# recover2 ----------------------------------------------------------------

#  Modeled after utils::recover(), which is
#  part of the R package, https://www.R-project.org
#
#  Copyright (C) 1995-2016 The R Core Team
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.R-project.org/Licenses/
recover2 <- function(start_frame = 1L, end_frame = sys.nframe()) {
  calls <- sys.calls()

  if (.isMethodsDispatchOn()) {
    tState <- tracingState(FALSE)
    withr::defer(tracingState(tState))
  }
  from <- min(end_frame, length(calls))

  calls <- calls[start_frame:from]

  if (is_false(peek_option("testthat_format_srcrefs"))) {
    calls <- lapply(calls, zap_srcref)
  }
  calls <- utils::limitedLabels(calls)

  repeat {
    which <- show_menu(calls, "\nEnter a frame number, or 0 to exit  ")
    if (which) {
      frame <- sys.frame(start_frame - 2 + which)
      browse_frame(frame, skip = 7 - which)
    } else {
      break
    }
  }
}

# Helpers -----------------------------------------------------------------

zap_srcref <- function(x) {
  attr(x, "srcref") <- NULL
  x
}

show_menu <- function(choices, title = NULL) {
  utils::menu(choices, title = title)
}

browse_frame <- function(frame, skip) {
  eval(
    substitute(browser(skipCalls = skip), list(skip = skip)),
    envir = frame
  )
}

#' Do you expect a string to match this pattern?
#'
#' @details
#' `expect_match()` checks if a character vector matches a regular expression,
#' powered by [grepl()].
#'
#' `expect_no_match()` provides the complementary case, checking that a
#' character vector *does not* match a regular expression.
#'
#' @inheritParams expect_that
#' @inheritParams base::grepl
#' @param regexp Regular expression to test against.
#' @param all Should all elements of actual value match `regexp` (TRUE),
#'   or does only one need to match (FALSE).
#' @param fixed If `TRUE`, treats `regexp` as a string to be matched exactly
#'   (not a regular expressions). Overrides `perl`.
#' @inheritDotParams base::grepl -pattern -x -perl -fixed
#' @family expectations
#' @export
#' @examples
#' expect_match("Testing is fun", "fun")
#' expect_match("Testing is fun", "f.n")
#' expect_no_match("Testing is fun", "horrible")
#'
#' show_failure(expect_match("Testing is fun", "horrible"))
#' show_failure(expect_match("Testing is fun", "horrible", fixed = TRUE))
#'
#' # Zero-length inputs always fail
#' show_failure(expect_match(character(), "."))
expect_match <- function(
  object,
  regexp,
  perl = FALSE,
  fixed = FALSE,
  ...,
  all = TRUE,
  info = NULL,
  label = NULL
) {
  act <- quasi_label(enquo(object), label)

  check_character(object)
  check_string(regexp)
  check_bool(perl)
  check_bool(fixed)
  check_bool(all)

  if (length(object) == 0) {
    fail(
      sprintf("Expected %s to have at least one element.", act$lab),
      info = info
    )
  } else {
    expect_match_(
      act = act,
      regexp = regexp,
      perl = perl,
      fixed = fixed,
      ...,
      all = all,
      info = info,
      label = label,
      negate = FALSE
    )
  }

  invisible(act$val)
}

#' @describeIn expect_match Check that a string doesn't match a regular
#'   expression.
#' @export
expect_no_match <- function(
  object,
  regexp,
  perl = FALSE,
  fixed = FALSE,
  ...,
  all = TRUE,
  info = NULL,
  label = NULL
) {
  # Capture here to avoid environment-related messiness
  act <- quasi_label(enquo(object), label)
  check_character(object)
  check_string(regexp)
  check_bool(perl)
  check_bool(fixed)
  check_bool(all)

  expect_match_(
    act = act,
    regexp = regexp,
    perl = perl,
    fixed = fixed,
    ...,
    all = all,
    info = info,
    label = label,
    negate = TRUE
  )
  invisible(act$val)
}

expect_match_ <- function(
  act,
  regexp,
  perl = FALSE,
  fixed = FALSE,
  ...,
  all = TRUE,
  info = NULL,
  label = NULL,
  negate = FALSE,
  title = "text",
  trace_env = caller_env()
) {
  matches <- grepl(regexp, act$val, perl = perl, fixed = fixed, ...)
  condition <- if (negate) !matches else matches
  ok <- if (all) all(condition) else any(condition)

  if (!ok) {
    values <- show_text(act$val, condition)
    if (length(act$val) == 1) {
      which <- ""
    } else {
      which <- if (all) "every element of " else "some element of "
    }
    match <- if (negate) "not to match" else "to match"

    msg_exp <- sprintf(
      "Expected %s%s %s %s %s.",
      which,
      act$lab,
      match,
      if (fixed) "string" else "regexp",
      encodeString(regexp, quote = '"')
    )
    msg_act <- c(paste0("Actual ", title, ':'), values)
    fail(c(msg_exp, msg_act), info = info, trace_env = trace_env)
  } else {
    pass()
  }
}

# Adapted from print.ellmer_prompt
show_text <- function(x, matches = NULL, max_items = 20, max_lines = NULL) {
  matches <- matches %||% rep(TRUE, length(x))
  max_lines <- max_lines %||% (max_items * 25)

  n <- length(x)
  n_extra <- length(x) - max_items
  if (n_extra > 0) {
    x <- x[seq_len(max_items)]
    matches <- matches[seq_len(max_items)]
  }

  if (length(x) == 0) {
    return(character())
  }

  bar <- if (cli::is_utf8_output()) "\u2502" else "|"

  id <- ifelse(
    matches,
    cli::col_green(cli::symbol$tick),
    cli::col_red(cli::symbol$cross)
  )

  indent <- paste0(id, " ", bar, " ")
  exdent <- paste0("  ", cli::col_grey(bar), " ")

  x[is.na(x)] <- cli::col_red("<NA>")
  x <- paste0(indent, x)
  x <- gsub("\n", paste0("\n", exdent), x)

  lines <- strsplit(x, "\n")
  ids <- rep(seq_along(x), lengths(lines))
  first <- c(TRUE, ids[-length(ids)] != ids[-1])
  lines <- unlist(lines)

  if (length(lines) > max_lines) {
    if (first[max_lines + 1]) {
      max_lines <- max_lines - 1
    }

    lines <- lines[seq_len(max_lines)]
    lines <- c(lines, paste0(exdent, "..."))
    n_extra <- n - ids[max_lines - 1]
  }

  if (n_extra > 0) {
    lines <- c(lines, paste0("... and ", n_extra, " more.\n"))
  }
  lines
}

test_that("error outside of test_that()", {
  skip_on_covr()
  withr::local_envvar(TESTTHAT_PARALLEL = "TRUE")
  err <- tryCatch(
    capture.output(suppressMessages(testthat::test_local(
      test_path("test-parallel", "outside"),
      reporter = "summary"
    ))),
    error = function(e) e
  )

  expect_match(err$message, "Test failures")
})

test_that("teardown error", {
  skip_on_cran()
  skip_on_covr()
  withr::local_envvar(TESTTHAT_PARALLEL = "TRUE")
  err <- tryCatch(
    capture.output(suppressMessages(testthat::test_local(
      test_path("test-parallel", "teardown"),
      reporter = "summary"
    ))),
    error = function(e) e
  )
  expect_s3_class(err$parent, "callr_error")
  expect_match(
    err$message,
    "At least one parallel worker failed to run teardown"
  )
  expect_match(
    err$parent$parent$parent$message,
    "Error in teardown",
    fixed = TRUE
  )
})

test_that("error in parallel setup code", {
  skip_on_covr()
  withr::local_envvar(TESTTHAT_PARALLEL = "TRUE")
  err <- tryCatch(
    capture.output(suppressMessages(testthat::test_local(
      test_path("test-parallel", "syntax-error"),
      reporter = "summary"
    ))),
    error = function(e) e
  )
  expect_s3_class(err, "testthat_process_error")
  # contains test file's name
  expect_match(conditionMessage(err), "test-error-1.R")
  # contains original error message
  expect_match(conditionMessage(err), "unexpected symbol")
  # contains the text of the syntax error
  expect_match(conditionMessage(err), "but this")
})

# regression test: test_file() used to crash with a NULL reporter
test_that("ListReporter with test_file and NULL reporter", {
  test_file_path <- 'test-list-reporter/test-exercise-list-reporter.R'
  expect_error(test_file(test_path(test_file_path), reporter = NULL), NA)
})

# regression: check that an exception is reported if it is raised in the test file outside
# of a test (test_that() call).
# N.B: the exception here happens between two tests: "before" and "after"
test_that("ListReporter - exception outside of test_that()", {
  test_file_path <- 'test-list-reporter/test-exception-outside-tests.R'
  res <- test_file(test_path(test_file_path), reporter = NULL)

  expect_true(is.list(res))
  # 2 results: first test "before" + the exception. N.B: the 2nd test "after" is not reported
  expect_length(res, 2)

  df <- as.data.frame(res)

  # the first result should be the results of test "before", that was successful
  expect_identical(df$test[1], 'before')
  expect_equal(df$passed[1], 1)
  expect_false(df$error[1])

  # the 2nd result should be the exception
  expect_true(is.na(df$test[2])) # no test name
  expect_true(df$error[2])       # it was an error
  expect_match(res[[2]]$results[[1]]$message, "dying outside of tests")
})


test_that("captures error if only thing in file", {
  test_file_path <- 'test-list-reporter/test-only-error.R'
  res <- test_file(test_path(test_file_path), reporter = NULL)

  expect_length(res, 1)
  expect_s3_class(res[[1]]$results[[1]], "expectation_error")
})

# ListReporter on a "standard" test file: 2 contexts, passing, failing and crashing tests
test_that("exercise ListReporter", {
  test_file_path <- 'test-list-reporter/test-exercise-list-reporter.R'
  res <- test_file(test_path(test_file_path), reporter = NULL)
  expect_s3_class(res, "testthat_results")

  # we convert the results to data frame for convenience
  df <- as.data.frame(res)
  expect_equal(nrow(df), 5)
  expect_equal(df$test, c("test1", "test2", "test-pass", "test-fail", "test-error"))

  # test "A failing test" is the only failing test
  expect_equal(df$failed, c(0, 0, 0, 1, 0))
  expect_identical(expectation_type(res[[4]]$results[[1]]), "failure")

  # test "A crashing test" is the only crashing test
  expect_equal(df$error, c(FALSE, FALSE, FALSE, FALSE, TRUE))
  expect_identical(expectation_type(res[[5]]$results[[1]]), "error")
})

# bare expectations are ignored
test_that("ListReporter and bare expectations", {
  test_file_path <- 'test-list-reporter/test-bare-expectations.R'
  res <- test_file(test_path(test_file_path), reporter = NULL)

  df <- as.data.frame(res)

  # 2 tests, "before" and "after". no result for the bare expectation
  expect_identical(df$test, c("before", "after"))
})

test_that("error in parallel setup code", {
  skip_on_covr()
  withr::local_envvar(TESTTHAT_PARALLEL = "TRUE")
  err <- tryCatch(
    capture.output(suppressMessages(testthat::test_local(
      test_path("test-parallel", "setup"),
      reporter = "summary"
    ))),
    error = function(e) e
  )
  expect_s3_class(err, "testthat_process_error")
  expect_match(conditionMessage(err), "Error in setup", fixed = TRUE)
})

test_that("useful failure if empty", {
  zero <- character(0)
  expect_snapshot_failure(expect_match(zero, 'asdf'))
})

test_that("useful failure messages for scalars", {
  local_reproducible_output(unicode = TRUE)

  one <- "bcde"
  expect_snapshot_failure(expect_match(one, 'asdf'))
  expect_snapshot_failure(expect_match(one, 'asdf', fixed = TRUE))
})

test_that("useful failure messages for vectors", {
  local_reproducible_output(unicode = TRUE)

  many <- c("a", "a", "b")
  expect_snapshot_failure(expect_match(many, "a"))
  expect_snapshot_failure(expect_match(many, "c", all = FALSE))

  paragraph <- c("This is a multiline\nparagraph.", "Second element.")
  expect_snapshot_failure(expect_match(paragraph, "paragraph"))

  na <- c("NA", NA)
  expect_snapshot_failure(expect_match(na, "NA"))
})

test_that("expect_match validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_match(1)
    expect_match("x", 1)
    expect_match("x", "x", fixed = 1)
    expect_match("x", "x", perl = 1)
    expect_match("x", "x", all = 1)
  })
})

test_that("expect_no_match validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_no_match(1, "x")
    expect_no_match("x", 1)
    expect_no_match("x", "x", fixed = 1)
    expect_no_match("x", "x", perl = 1)
    expect_no_match("x", "x", all = 1)
  })
})

test_that("extra arguments passed onto grepl", {
  expect_failure(expect_match("\\s", "\\s"))
  expect_success(expect_match("\\s", "\\s", fixed = TRUE))

  expect_failure(expect_match("test", "TEST"))
  expect_success(expect_match("test", "TEST", ignore.case = TRUE))
})

test_that("expect_no_match works", {
  expect_success(expect_no_match("[a]", "[b]"))
  expect_success(expect_no_match("[a]", "[b]", fixed = TRUE))

  x <- "te*st"
  expect_snapshot_failure(expect_no_match(x, "e*", fixed = TRUE))
  x <- "test"
  expect_snapshot_failure(expect_no_match(x, "TEST", ignore.case = TRUE))
})

test_that("empty string is never a match", {
  expect_success(expect_no_match(character(), "x"))
})

# show_text() ------------------------------------------------------------------

test_that("show_text() shows success and failure", {
  local_reproducible_output(unicode = TRUE)
  expect_snapshot({
    base::writeLines(show_text(c("a", "b"), c(TRUE, FALSE)))
  })
})

test_that("show_text() truncates values and lines", {
  local_reproducible_output(unicode = TRUE)
  lines <- map_chr(
    split(letters, (seq_along(letters) - 1) %/% 3),
    paste,
    collapse = "\n"
  )

  expect_snapshot({
    base::writeLines(show_text(lines, max_lines = 3))
    base::writeLines(show_text(lines, max_items = 3))
    base::writeLines(show_text(lines, max_items = 2, max_lines = 4))
  })
})

test_that("stdout/stderr in parallel code", {
  skip_on_covr()
  withr::local_envvar(TESTTHAT_PARALLEL = "TRUE")

  assemble_msgs <- function(txt, test_name) {
    prefix <- paste0("> ", test_name, ": ")
    parts <- sub(
      prefix,
      "",
      grep(prefix, out, fixed = TRUE, value = TRUE),
      fixed = TRUE
    )
    paste(parts, collapse = "")
  }

  for (reporter in c("summary", "progress")) {
    out <- capture.output(suppressMessages(testthat::test_local(
      test_path("test-parallel", "stdout"),
      reporter = reporter
    )))
    msg2 <- assemble_msgs(out, "test-stdout-2.R")
    expect_match(msg2, "This is a message!", fixed = TRUE)
    msg3 <- assemble_msgs(out, "test-stdout-3.R")
    expect_match(msg3, "[1]  1  2  3", fixed = TRUE)
  }
})

#' Report results for `R CMD check`
#'
#' `R CMD check` displays only the last 13 lines of the result, so this
#' report is designed to ensure that you see something useful there.
#'
#' @export
#' @family reporters
CheckReporter <- R6::R6Class(
  "CheckReporter",
  inherit = Reporter,
  public = list(
    problems = NULL,
    skips = NULL,
    warnings = NULL,
    n_ok = 0L,
    old_in_check_reporter = NULL,

    initialize = function(...) {
      self$capabilities$parallel_support <- TRUE
      self$problems <- Stack$new()
      self$warnings <- Stack$new()
      self$skips <- Stack$new()

      super$initialize(...)
    },

    add_result = function(context, test, result) {
      if (expectation_broken(result)) {
        self$problems$push(result)
        try(save_test(result$srcref, "_problems"), silent = TRUE)
      } else if (expectation_warning(result)) {
        self$warnings$push(result)
      } else if (expectation_skip(result)) {
        self$skips$push(result)
      } else {
        self$n_ok <- self$n_ok + 1L
      }
    },

    start_reporter = function() {
      self$old_in_check_reporter <- in_check_reporter()
      the$in_check_reporter <- TRUE
    },

    end_reporter = function() {
      the$in_check_reporter <- self$old_in_check_reporter

      if (self$skips$size() || self$warnings$size() || self$problems$size()) {
        self$cat_line(summary_line(
          n_fail = self$problems$size(),
          n_warn = self$warnings$size(),
          n_skip = self$skips$size(),
          n_pass = self$n_ok
        ))
        self$cat_line()
      }

      skip_report(self, line = 2)

      # Don't show warnings in revdep checks in order to focus on failures
      if (self$warnings$size() > 0 && !on_cran()) {
        warnings <- self$warnings$as_list()

        self$rule("Warnings", line = 2)
        self$cat_line(map_chr(warnings, issue_summary, rule = TRUE))
        self$cat_line()
      }

      if (self$problems$size() > 0) {
        problems <- self$problems$as_list()
        saveRDS(problems, "testthat-problems.rds", version = 2)

        self$rule("Failed tests", line = 2)
        self$cat_line(map_chr(problems, issue_summary, rule = TRUE))
        self$cat_line()

        if (some(problems, \(x) isTRUE(attr(x, "snapshot")))) {
          self$rule("Snapshots", line = 1)
          self$cat_line(snapshot_check_hint())
        }
      } else {
        # clean up
        unlink("testthat-problems.rds")
      }

      self$cat_line(summary_line(
        n_fail = self$problems$size(),
        n_warn = self$warnings$size(),
        n_skip = self$skips$size(),
        n_pass = self$n_ok
      ))
    }
  )
)

summary_line <- function(n_fail, n_warn, n_skip, n_pass) {
  colourise_if <- function(text, colour, cond) {
    if (cond) colourise(text, colour) else text
  }

  # Ordered from most important to least important
  paste0(
    "[ ",
    colourise_if("FAIL", "failure", n_fail > 0),
    " ",
    n_fail,
    " | ",
    colourise_if("WARN", "warn", n_warn > 0),
    " ",
    n_warn,
    " | ",
    colourise_if("SKIP", "skip", n_skip > 0),
    " ",
    n_skip,
    " | ",
    colourise_if("PASS", "success", n_fail == 0),
    " ",
    n_pass,
    " ]"
  )
}

snapshot_check_hint <- function() {
  intro <- "To review and process snapshots locally:"

  if (on_gh() && Sys.getenv("GITHUB_JOB") == "R-CMD-check") {
    repository <- Sys.getenv("GITHUB_REPOSITORY")
    run_id <- Sys.getenv("GITHUB_RUN_ID")

    call <- sprintf(
      "testthat::snapshot_download_gh(\"%s\", \"%s\")",
      repository,
      run_id
    )
    copy <- sprintf("* Run `%s` to download snapshots.", call)
  } else {
    copy <- c(
      if (on_ci()) {
        "* Download and unzip artifact."
      } else {
        "* Locate check directory."
      },
      "* Copy 'tests/testthat/_snaps' to local package."
    )
  }

  action <- c(
    "* Run `testthat::snapshot_accept()` to accept all changes.",
    "* Run `testthat::snapshot_review()` to review all changes."
  )
  c(intro, copy, action)
}

run <- function(x) {
  cli::format_inline(paste0("{.run testthat::", x, "}"))
}

#' Skip a test
#'
#' @description
#' `skip_if()` and `skip_if_not()` allow you to skip tests, immediately
#' concluding a [test_that()] block without executing any further expectations.
#' This allows you to skip a test without failure, if for some reason it
#' can't be run (e.g. it depends on the feature of a specific operating system,
#' or it requires a specific version of a package).
#'
#' See `vignette("skipping")` for more details.
#'
#' @section Helpers:
#'
#' * `skip_if_not_installed("pkg")` skips tests if package "pkg" is not
#'   installed or cannot be loaded (using `requireNamespace()`). Generally,
#'   you can assume that suggested packages are installed, and you do not
#'   need to check for them specifically, unless they are particularly
#'   difficult to install.
#'
#' * `skip_if_offline()` skips if an internet connection is not available
#'   (using [curl::nslookup()]) or if the test is run on CRAN. Requires
#'   \{curl\} to be installed and included in the dependencies of your package.
#'
#' * `skip_if_translated("msg")` skips tests if the "msg" is translated.
#'
#' * `skip_on_bioc()` skips on Bioconductor (using the `IS_BIOC_BUILD_MACHINE`
#'    env var).
#'
#' * `skip_on_cran()` skips on CRAN (using the `NOT_CRAN` env var set by
#'    devtools and friends).
#'
#' * `skip_on_covr()` skips when covr is running (using the `R_COVR` env var).
#'
#' * `skip_on_ci()` skips on continuous integration systems like GitHub Actions,
#'    travis, and appveyor (using the `CI` env var).
#'
#' * `skip_on_os()` skips on the specified operating system(s) ("windows",
#'   "mac", "linux", or "solaris").
#'
#' @param message A message describing why the test was skipped.
#' @param host A string with a hostname to lookup
#' @export
#' @examples
#' if (FALSE) skip("Some Important Requirement is not available")
#'
#' test_that("skip example", {
#'   expect_equal(1, 1L)    # this expectation runs
#'   skip('skip')
#'   expect_equal(1, 2)     # this one skipped
#'   expect_equal(1, 3)     # this one is also skipped
#' })
skip <- function(message = "Skipping") {
  message <- paste0(message, collapse = "\n")
  cond <- structure(
    list(message = paste0("Reason: ", message)),
    class = c("skip", "condition")
  )
  stop(cond)
}

# Called automatically if the test contains no expectations
skip_empty <- function() {
  cond <- structure(
    list(message = "Reason: empty test"),
    class = c("skip_empty", "skip", "condition")
  )
  stop(cond)
}

#' @export
#' @rdname skip
#' @param condition Boolean condition to check. `skip_if_not()` will skip if
#'   `FALSE`, `skip_if()` will skip if `TRUE`.
skip_if_not <- function(condition, message = NULL) {
  if (is.null(message)) {
    message <- paste0(deparse1(substitute(condition)), " is not TRUE")
  }
  if (!isTRUE(condition)) {
    skip(message)
  } else {
    invisible()
  }
}

#' @export
#' @rdname skip
skip_if <- function(condition, message = NULL) {
  if (is.null(message)) {
    message <- paste0(deparse1(substitute(condition)), " is TRUE")
  }
  if (isTRUE(condition)) {
    skip(message)
  } else {
    invisible()
  }
}

#' @export
#' @param pkg Name of package to check for
#' @param minimum_version Minimum required version for the package
#' @rdname skip
skip_if_not_installed <- function(pkg, minimum_version = NULL) {
  # most common case: it's not installed
  tryCatch(
    find.package(pkg),
    error = function(e) skip(paste0("{", pkg, "} is not installed"))
  )
  # rarer: it's installed, but fails to load
  if (!requireNamespace(pkg, quietly = TRUE)) {
    skip(paste0("{", pkg, "} cannot be loaded"))
  }

  if (!is.null(minimum_version)) {
    installed_version <- package_version(pkg)
    if (installed_version < minimum_version) {
      skip(paste0(
        "Installed ", pkg, " is version ", installed_version, "; ",
        "but ", minimum_version, " is required"
      ))
    }
  }

  invisible()
}
package_version <- function(x) {
  utils::packageVersion(x)
}

#' @export
#' @rdname skip
skip_if_offline <- function(host = "captive.apple.com") {
  skip_on_cran()
  check_installed("curl")

  skip_if_not(has_internet(host), "offline")
}
has_internet <- function(host) {
  !is.null(curl::nslookup(host, error = FALSE))
}

#' @export
#' @rdname skip
skip_on_cran <- function() {
  skip_if(on_cran(), "On CRAN")
}

#' @export
#' @param os Character vector of one or more operating systems to skip on.
#'   Supported values are `"windows"`, `"mac"`, `"linux"`, and `"solaris"`.
#' @param arch Character vector of one or more architectures to skip on.
#'   Common values include `"i386"` (32 bit), `"x86_64"` (64 bit), and
#'   `"aarch64"` (M1 mac). Supplying `arch` makes the test stricter; i.e. both
#'   `os` and `arch` must match in order for the test to be skipped.
#' @rdname skip
skip_on_os <- function(os, arch = NULL) {
  os <- match.arg(
    os,
    choices = c("windows", "mac", "linux", "solaris"),
    several.ok = TRUE
  )

  msg <- switch(system_os(),
    windows = if ("windows" %in% os) "On Windows",
    darwin =  if ("mac" %in% os) "On Mac",
    linux =   if ("linux" %in% os) "On Linux",
    sunos =   if ("solaris" %in% os) "On Solaris"
  )

  if (!is.null(arch) && !is.null(msg)) {
    if (!is.character(arch)) {
      abort("`arch` must be a character vector")
    }

    if (system_arch() %in% arch) {
      msg <- paste(msg, system_arch())
    } else {
      msg <- NULL
    }
  }

  if (is.null(msg)) {
    invisible(TRUE)
  } else {
    skip(msg)
  }
}

system_os <- function() tolower(Sys.info()[["sysname"]])
system_arch <- function() R.version$arch

#' @export
#' @rdname skip
skip_on_ci <- function() {
  skip_if(on_ci(), "On CI")
}

#' @export
#' @rdname skip
skip_on_covr <- function() {
  skip_if(in_covr(), "On covr")
}

#' @export
#' @rdname skip
skip_on_bioc <- function() {
  skip_if(on_bioc(), "On Bioconductor")
}

#' @export
#' @param msgid R message identifier used to check for translation: the default
#'   uses a message included in most translation packs. See the complete list in
#'   [`R-base.pot`](https://github.com/wch/r-source/blob/master/src/library/base/po/R-base.pot).
#' @rdname skip
skip_if_translated <- function(msgid = "'%s' not found") {
  skip_if(
    gettext(msgid) != msgid,
    paste0("\"", msgid, "\" is translated")
  )
}

gettext <- function(msgid, domain = "R") {
  base::gettext(msgid, domain = domain)
}


#' Superseded skip functions
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' * `skip_on_travis()` and `skip_on_appveyor()` have been superseded by
#'   [skip_on_ci()].
#'
#' @export
#' @keywords internal
skip_on_travis <- function() {
  skip_if(env_var_is_true("TRAVIS"), "On Travis")
}

#' @export
#' @rdname skip_on_travis
skip_on_appveyor <- function() {
  skip_if(env_var_is_true("APPVEYOR"), "On Appveyor")
}

# helpers -----------------------------------------------------------------

on_ci <- function() {
 env_var_is_true("CI")
}
in_covr <- function() {
  env_var_is_true("R_COVR")
}
on_bioc <- function() {
  env_var_is_true("IS_BIOC_BUILD_MACHINE")
}
on_cran <- function() {
  !interactive() && !env_var_is_true("NOT_CRAN")
}

env_var_is_true <- function(x) {
  isTRUE(as.logical(Sys.getenv(x, "false")))
}

MainprocessSnapshotReporter <- R6::R6Class("MainprocessSnapshotReporter",
  inherit = SnapshotReporter,
  public = list(
    end_file = function() {
      # No thing to do, this is done in the subprocess
    }
  )
)

SubprocessSnapshotReporter <- R6::R6Class("SubprocessSnapshotReporter",
  inherit = SnapshotReporter,
  public = list(
    start_file = function(path, test = NULL) {
      private$filename <- path
      private$test <- test
      super$start_file(path, test)
    },

    end_file = function() {
      private$filename <- NULL
      super$end_file()
    },

    end_context = function(context) {
      private$context <- NULL
      super$end_context()
    },

    end_test = function(context, test) {
      private$context <- NULL
      private$test <- NULL
      super$end_test(context, test)
    },

    start_test = function(context, test) {
      private$context <- context
      private$test <- test
      super$start_test(context, test)
    },

    announce_file_snapshot = function(name) {
      private$event("announce_file_snapshot", name)
      super$announce_file_snapshot(name)
    },

    end_reporter = function() {
      # do not delete unused snapshots, that's up to the main process
    }
  ),
  private = list(
    filename = NULL,
    context = NULL,
    test = NULL,
    event = function(cmd, ...) {
      msg <- list(
        code = PROCESS_MSG,
        type = "snapshotter",
        cmd = cmd,
        filename = private$filename,
        context = private$context,
        test = private$test,
        time = NULL,
        args = list(...)
      )
      class(msg) <- c("testthat_message", "callr_message", "condition")
      signalCondition(msg)
    }
  )
)

# We can't use expect_snapshot_reporter() here because it uses test_one_file()
# which wraps code in `test_code()` which turns the error into a test failure
# It also only captures the output, but we also want to see the error

test_that("produces useful output", {
  run_tests <- \() source(test_path("reporters/tests.R"))
  expect_snapshot(with_reporter("stop", run_tests()), error = TRUE)
})

test_that("can suppress praise", {
  run_tests <- \() source(test_path("reporters/successes.R"))
  expect_silent(with_reporter(StopReporter$new(praise = FALSE), run_tests()))
})

test_that("works nicely with nested tests", {
  run_tests <- \() source(test_path("reporters/nested.R"))
  expect_snapshot(with_reporter("stop", run_tests()), error = TRUE)
})

test_that("errors when needed", {
  r <- StopReporter$new()
  r$start_test()
  expect_no_error(r$end_test())

  r$start_test()
  r$n_fail <- 1
  r$n_success <- 0
  expect_snapshot(error = TRUE, r$end_test())
})

#' Do you expect this code to return the same value as last time?
#'
#' Captures the result of function, flexibly serializing it into a text
#' representation that's stored in a snapshot file. See [expect_snapshot()]
#' for more details on snapshot testing.
#'
#' @param style Serialization style to use:
#'   * `json` uses [jsonlite::fromJSON()] and [jsonlite::toJSON()]. This
#'      produces the simplest output but only works for relatively simple
#'      objects.
#'   * `json2` uses [jsonlite::serializeJSON()] and [jsonlite::unserializeJSON()]
#'     which are more verbose but work for a wider range of type.
#'   * `deparse` uses [deparse()], which generates a depiction of the object
#'     using R code.
#'   * `serialize()` produces a binary serialization of the object using
#'     [serialize()]. This is all but guaranteed to work for any R object,
#'     but produces a completely opaque serialization.
#' @param ... Passed on to [waldo::compare()] so you can control the details of
#'   the comparison.
#' @inheritParams expect_snapshot
#' @inheritParams compare
#' @export
expect_snapshot_value <- function(
  x,
  style = c("json", "json2", "deparse", "serialize"),
  cran = FALSE,
  tolerance = testthat_tolerance(),
  ...,
  variant = NULL
) {
  edition_require(3, "expect_snapshot_value()")

  style <- arg_match(style)
  check_bool(cran)
  check_number_decimal(tolerance, min = 0)

  variant <- check_variant(variant)
  lab <- quo_label(enquo(x))

  save <- switch(
    style,
    json = function(x) jsonlite::toJSON(x, auto_unbox = TRUE, pretty = TRUE),
    json2 = function(x) jsonlite::serializeJSON(x, pretty = TRUE),
    deparse = function(x) paste0(deparse(x), collapse = "\n"),
    serialize = function(x) {
      jsonlite::base64_enc(serialize(x, NULL, version = 2))
    }
  )
  load <- switch(
    style,
    json = function(x) jsonlite::fromJSON(x, simplifyVector = FALSE),
    json2 = function(x) jsonlite::unserializeJSON(x),
    deparse = function(x) reparse(x),
    serialize = function(x) unserialize(jsonlite::base64_dec(x))
  )

  with_is_snapshotting(force(x))
  check_roundtrip(
    x,
    load(save(x)),
    label = lab,
    style = style,
    ...,
    tolerance = tolerance
  )

  expect_snapshot_helper(
    lab,
    x,
    save = save,
    load = load,
    cran = cran,
    ...,
    tolerance = tolerance,
    variant = variant,
    trace_env = caller_env()
  )
}

# Safe environment for evaluating deparsed objects, based on inspection of
# https://github.com/wch/r-source/blob/5234fe7b40aad8d3929d240c83203fa97d8c79fc/src/main/deparse.c#L845
reparse <- function(x) {
  env <- env(emptyenv())
  env_bind(
    env,
    !!!env_get_list(
      base_env(),
      c(
        c("c", "structure", ":", "-"),
        c("list", "numeric", "integer", "logical", "character"),
        "function",
        c("quote", "alist", "pairlist", "as.pairlist", "expression")
      )
    )
  )
  env_bind(env, !!!env_get_list(ns_env("methods"), c("new", "getClass")))

  eval(parse(text = x), env)
}

check_roundtrip <- function(
  x,
  y,
  label,
  style,
  ...,
  tolerance = testthat_tolerance(),
  error_call = caller_env()
) {
  check <- waldo_compare(
    x,
    y,
    x_arg = "original",
    y_arg = "new",
    ...,
    tolerance = tolerance
  )
  if (length(check) > 0) {
    cli::cli_abort(
      c(
        "{.code {label}} could not be safely serialized with {.arg style} = {.str {style}}.",
        " " = "Serializing then deserializing the object returned something new:\n\n{no_wrap(check)}\n",
        i = "You may need to try a different {.arg style}."
      ),
      call = error_call
    )
  }
}

#' Quasi-labelling
#'
#' The first argument to every `expect_` function can use unquoting to
#' construct better labels. This makes it easy to create informative labels when
#' expectations are used inside a function or a for loop. `quasi_label()` wraps
#' up the details, returning the expression and label.
#'
#' @section Limitations:
#' Because all `expect_` function use unquoting to generate more informative
#' labels, you can not use unquoting for other purposes. Instead, you'll need
#' to perform all other unquoting outside of the expectation and only test
#' the results.
#'
#' @param quo A quosure created by `rlang::enquo()`.
#' @param label An optional label to override the default. This is
#'   only provided for internal usage. Modern expectations should not
#'   include a `label` parameter.
#' @param arg Argument name shown in error message if `quo` is missing.
#' @keywords internal
#' @return A list containing two elements:
#' \item{val}{The evaluate value of `quo`}
#' \item{lab}{The quasiquoted label generated from `quo`}
#' @export
#' @examples
#' f <- function(i) if (i > 3) i * 9 else i * 10
#' i <- 10
#'
#' # This sort of expression commonly occurs inside a for loop or function
#' # And the failure isn't helpful because you can't see the value of i
#' # that caused the problem:
#' show_failure(expect_equal(f(i), i * 10))
#'
#' # To overcome this issue, testthat allows you to unquote expressions using
#' # !!. This causes the failure message to show the value rather than the
#' # variable name
#' show_failure(expect_equal(f(!!i), !!(i * 10)))
quasi_label <- function(quo, label = NULL, arg = "quo") {
  force(quo)
  if (quo_is_missing(quo)) {
    stop("argument `", arg, "` is missing, with no default.", call. = FALSE)
  }

  expr <- quo_get_expr(quo)

  list(
    val = eval_bare(expr, quo_get_env(quo)),
    lab = label %||% expr_label(expr)
  )
}

quasi_capture <- function(.quo, .label, .capture, ...) {
  act <- list()
  act$lab <- .label %||% quo_label(.quo)
  act$cap <- .capture(act$val <- eval_bare(quo_get_expr(.quo), quo_get_env(.quo)), ...)

  act
}

expr_label <- function(x) {
  if (is_syntactic_literal(x)) {
    deparse1(x)
  } else if (is.name(x)) {
    paste0("`", as.character(x), "`")
  } else if (is_call(x)) {
    chr <- deparse(x)
    if (length(chr) > 1) {
      if (is_call(x, "function")) {
        x[[3]] <- quote(...)
      } else if (is_call_infix(x)) {
        left <- deparse(x[[2]], width.cutoff = 29)
        right <- deparse(x[[3]], width.cutoff = 28)

        if (length(left) > 1) {
          x[[2]] <- quote(expr = ...)
        }
        if (length(right) > 1) {
          x[[3]] <- quote(expr = ...)
        }
      } else {
        x <- call2(x[[1]], quote(expr = ...))
      }
    }
    deparse1(x)
  } else {
    # Any other object that's been inlined in
    x <- deparse(x)
    if (length(x) > 1) {
      x <- paste0(x[[1]], "...)")
    }
    x
  }
}

is_call_infix <- function(x) {
  if (!is_call(x, n = 2)) {
    return(FALSE)
  }

  fn <- x[[1]]
  if (!is_symbol(fn)) {
    return(FALSE)
  }

  name <- as_string(fn)
  base <- c(
    ":", "::", ":::", "$", "@", "^", "*", "/", "+", "-", ">", ">=",
    "<", "<=", "==", "!=", "!", "&", "&&", "|", "||", "~", "<-", "<<-"
  )
  name %in% base || grepl("^%.*%$", name)
}

#' Verify output
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' This function is superseded in favour of `expect_snapshot()` and friends.
#'
#' This is a regression test that records interwoven code and output into a
#' file, in a similar way to knitting an `.Rmd` file (but see caveats below).
#'
#' `verify_output()` is designed particularly for testing print methods and error
#' messages, where the primary goal is to ensure that the output is helpful to
#' a human. Obviously, you can't test that with code, so the best you can do is
#' make the results explicit by saving them to a text file. This makes the output
#' easy to verify in code reviews, and ensures that you don't change the output
#' by accident.
#'
#' `verify_output()` is designed to be used with git: to see what has changed
#' from the previous run, you'll need to use `git diff` or similar.
#'
#' @section Syntax:
#' `verify_output()` can only capture the abstract syntax tree, losing all
#' whitespace and comments. To mildly offset this limitation:
#'
#' - Strings are converted to R comments in the output.
#' - Strings starting with `# ` are converted to headers in the output.
#'
#' @section CRAN:
#' On CRAN, `verify_output()` will never fail, even if the output changes.
#' This avoids false positives because tests of print methods and error
#' messages are often fragile due to implicit dependencies on other packages,
#' and failure does not imply incorrect computation, just a change in
#' presentation.
#'
#' @param path Path to record results.
#'
#'   This should usually be a call to [test_path()] in order to ensure that
#'   the same path is used when run interactively (when the working directory
#'   is typically the project root), and when run as an automated test (when
#'   the working directory will be `tests/testthat`).
#' @param code Code to execute. This will usually be a multiline expression
#'   contained within `{}` (similarly to `test_that()` calls).
#' @param width Width of console output
#' @param crayon Enable cli/crayon package colouring?
#' @param unicode Enable cli package UTF-8 symbols? If you set this to
#'   `TRUE`, call `skip_if(!cli::is_utf8_output())` to disable the
#'   test on your CI platforms that don't support UTF-8 (e.g. Windows).
#' @param env The environment to evaluate `code` in.
#' @export
#' @keywords internal
verify_output <- function(path, code, width = 80, crayon = FALSE,
                          unicode = FALSE, env = caller_env()) {

  local_reproducible_output(width = width, crayon = crayon, unicode = unicode)

  expr <- substitute(code)
  output <- verify_exec(expr, env = env)

  if (!interactive() && on_cran()) {
    skip("On CRAN")
  }
  compare_file(path, output, update = TRUE)
  invisible()
}

verify_exec <- function(expr, env = caller_env(), replay = output_replay) {

  if (is_call(expr, "{")) {
    exprs <- as.list(expr[-1])
  } else {
    exprs <- list(expr)
  }

  device_path <- withr::local_tempfile(pattern = "verify_exec_")
  withr::local_pdf(device_path)
  grDevices::dev.control(displaylist = "enable")

  exprs <- lapply(exprs, function(x) if (is.character(x)) paste0("# ", x) else expr_deparse(x))
  source <- unlist(exprs, recursive = FALSE)

  handler <- evaluate::new_output_handler(value = testthat_print)
  results <- evaluate::evaluate(source, envir = env,
    new_device = FALSE,
    output_handler = handler
  )
  output <- unlist(lapply(results, replay))
  output <- gsub("\r", "", output, fixed = TRUE)
  output
}

output_replay <- function(x) {
  UseMethod("output_replay", x)
}

#' @export
output_replay.character <- function(x) {
  c(split_lines(x), "")
}

#' @export
output_replay.source <- function(x) {
  lines <- split_lines(x$src)

  # Remove header of lines so they don't get prefixed
  first <- lines[[1]]
  if (grepl("^# # ", first)) {
    header <- gsub("^# # ", "", first)
    lines <- lines[-1]
  } else {
    header <- NULL
  }

  n <- length(lines)
  if (n > 0) {
    lines[1] <- paste0("> ", lines[1])
    if (n > 1) {
      lines[2:n] <- paste0("+ ", lines[2:n])
    }
  }

  if (!is.null(header)) {
    underline <- strrep("=", nchar(header))
    lines <- c("", header, underline, "", lines)
  }

  lines
}

#' @export
output_replay.error <- function(x) {
  msg <- cnd_message(x)
  if (is.null(x$call)) {
    msg <- paste0("Error: ", msg)
  } else {
    call <- deparse(x$call)
    msg <- paste0("Error in ", call, ": ", msg)
  }
  c(split_lines(msg), "")
}

#' @export
output_replay.warning <- function(x) {
  msg <- cnd_message(x)
  if (is.null(x$call)) {
    msg <- paste0("Warning: ", msg)
  } else {
    call <- deparse(x$call)
    msg <- paste0("Warning in ", call, ": ", msg)
  }
  c(split_lines(msg), "")
}

#' @export
output_replay.message <- function(x) {
  # Messages are the only conditions where a new line is appended automatically
  msg <- paste0("Message: ", sub("\n$", "", cnd_message(x)))
  c(split_lines(msg), "")
}

#' @export
output_replay.recordedplot <- function(x) {
  abort("Plots are not supported")
}


# Helpers ------------------------------------------------------------

split_lines <- function(x) {
  strsplit(x, "\n")[[1]]
}

test_that("can snapshot values", {
  x <- list("a", 1.5, 1L, TRUE)
  expect_snapshot_value(x, style = "json")
  expect_snapshot_value(x, style = "json2")
  expect_snapshot_value(x, style = "deparse")
  expect_snapshot_value(x, style = "serialize")
})

test_that("can control snapshot value details", {
  expect_snapshot_value(1.2, tolerance = 0.1)

  # including through ...
  f <- ~1
  expect_snapshot_value(f, style = "serialize", ignore_formula_env = TRUE)
})

test_that("tolerance passed to check_roundtrip", {
  expect_snapshot_value(0.900000000000001, style = "json")
})

test_that("reparse handles common cases", {
  roundtrip <- function(x) reparse(deparse(x))
  expect_equal(roundtrip(-1), -1)
  expect_equal(roundtrip(c(1, 2, 3)), c(1, 2, 3))
  expect_equal(roundtrip(list(1, 2, 3)), list(1, 2, 3))
  expect_equal(roundtrip(mtcars), mtcars)
  expect_equal(roundtrip(1:10), 1:10)
  expect_equal(roundtrip(numeric()), numeric())

  f <- function(x) x + 1
  expect_equal(roundtrip(f), f, ignore_function_env = TRUE)
})

test_that("errors if can't roundtrip", {
  snapper <- local_test_snapshotter()
  snapper$start_file("snapshot-4", "test")

  expect_error(expect_snapshot_value(NULL), "safely serialized")
})

test_that("check_roundtrip() gives nice error", {
  # disable crayon usage
  local_bindings(crayon = FALSE, .env = get_reporter())

  wrapper <- function(...) check_roundtrip(...)
  expect_snapshot(
    wrapper(NULL, list(), label = "foo", style = "json"),
    error = TRUE
  )
})

test_that("expect_snapshot_value validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_snapshot_value(123, cran = "yes")
    expect_snapshot_value(123, tolerance = "high")
  })
})

#' Does code return a vector with the specified length?
#'
#' @seealso [expect_vector()] to make assertions about the "size" of a vector
#' @inheritParams expect_that
#' @param n Expected length.
#' @family expectations
#' @export
#' @examples
#' expect_length(1, 1)
#' expect_length(1:10, 10)
#'
#' \dontrun{
#' expect_length(1:10, 1)
#' }
expect_length <- function(object, n) {
  stopifnot(is.numeric(n), length(n) == 1)

  act <- quasi_label(enquo(object), arg = "object")
  act$n <- length(act$val)

  expect(
    act$n == n,
    sprintf("%s has length %i, not length %i.", act$lab, act$n, n)
  )

  invisible(act$val)
}

test_that("atomic scalars deparsed to single values", {
  expect_equal(expr_label(NULL), "NULL")
  expect_equal(expr_label(TRUE), "TRUE")
  expect_equal(expr_label(1L), "1L")
  expect_equal(expr_label(1), "1")
  expect_equal(expr_label("a"), '"a"')
})

test_that("symbols are quoted", {
  expect_equal(expr_label(quote(a)), "`a`")
})

test_that("missing arguments are propagated", {
  x <- list(missing = missing_arg())
  expect_snapshot_failure(expect_null(x$missing))
})

test_that("is_call_infix() handles complex calls (#1472)", {
  expect_false(is_call_infix(quote(
    base::any(
      c(
        veryyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy_long_name = TRUE
      ),
      na.rm = TRUE
    )
  )))

  withr::local_envvar(
    "_R_CHECK_LENGTH_1_LOGIC2_" = "TRUE",
  )
  expect_true(
    base::any(
      c(
        veryyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy_long_name = TRUE
      ),
      na.rm = TRUE
    )
  )
})

test_that("long vectors get ...", {
  long <- "123456789_123456789_123456789_123456789_123456789_123456789_"
  expect_equal(
    expr_label(c(long, long)),
    paste0('c("', long, '", ...)')
  )
})

test_that("produces useful summaries for long calls", {
  expect_snapshot({
    expr_label(quote(foo(
      a = "this is a long argument",
      b = "this is a long argument",
      c = "this is a long argument"
    )))

    expr_label(quote(
      arg + arg + arg + arg + arg + arg + arg + arg + arg + arg + arg + arg
    ))
    expr_label(quote(
      arg + (arg + arg + arg + arg + arg + arg + arg + arg + arg + arg + arg)
    ))

    expr_label(quote(function(a, b, c) {
      a + b + c
    }))
  })
})

test_that("other inlined other objects are deparsed", {
  expect_equal(expr_label(c(1, 2, 3)), "c(1, 2, 3)")
  expect_equal(expr_label(list(1, 2, 3)), "list(1, 2, 3)")
  expect_equal(
    expr_label(1:100 + 0),
    "c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, ...)"
  )
})

test_that("labelling compound {} expression gives single string", {
  out <- expr_label(quote({
    1 + 2
  }))

  expect_length(out, 1)
  expect_type(out, "character")
})

test_that("can label multiline functions", {
  expect_equal(
    expr_label(quote(function(x, y) {})),
    "`function(x, y) ...`"
  )
})

test_that("informative error for missing arg", {
  expect_snapshot(error = TRUE, expect_equal())
})

test_that("labelling compound {} expression gives single string", {
  out <- expr_label(quote({
    1 + 2
  }))

  expect_length(out, 1)
  expect_type(out, "character")
})

test_that("can label multiline functions", {
  expect_equal(
    expr_label(quote(function(x, y) {})),
    "function(x, y) ..."
  )
})

describe("describe", {
  it("can contain nested describe blocks", {
    describe("addition", {
      it("should be able to add two numbers", {
        expect_equal(2, 1 + 1)
      })
      describe("sub feature", {
        it("should also work", {
          expect_equal(2, 1 + 1)
        })
      })
    })
  })
})

test_that("can write snaphot tests", {
  local_description_set()

  describe("snapshot tests in describe", {
    expect_snapshot(1 + 1)
    it("and in it", {
      expect_snapshot(2 + 2)
    })
  })
})

test_that("unimplemented specs generate skips", {
  expectations <- capture_expectations({
    it("can have not yet implemented specs", {
      describe("Millennium Prize Problems", {
        it("can be shown that P != NP")
      })
    })
  })
  expect_length(expectations, 1)
  expect_s3_class(expectations[[1]], "expectation_skip")
})

someExternalVariable <- 1
describe("variable scoping", {
  someInternalVariable <- 1
  it("should be possible to use variables from outer environments", {
    expect_equal(1, someExternalVariable)
    expect_equal(1, someInternalVariable)

    someInternalVariable <- 2
    someExternalVariable <- 3
  })

  # prefix is needed to test this use case
  testthat::it("even when using it() directly", {
    expect_equal(1, someExternalVariable)
    expect_equal(1, someInternalVariable)
  })

  it("shouldn't affect other tests", {
    expect_equal(1, someExternalVariable)
    expect_equal(1, someInternalVariable)
  })
})

test_that("has to have a valid description for the block", {
  expect_snapshot(error = TRUE, {
    describe()
    describe(c("a", "b"))

    it()
    it(c("a", "b"))
  })
})

test_that("names are concatenated", {
  expectations <- capture_expectations({
    describe("a", {
      describe("b", {
        it("c", {
          it("d", {
            expect_true(TRUE)
          })
        })
      })
    })
  })
  expect_equal(expectations[[1]]$test, "a / b / c / d")
})

test_that("skips are scoped to describe/it", {
  expectations <- capture_expectations({
    describe("", skip())
    describe("", expect_true())
    it("", skip())
    it("", expect_true(TRUE))
  })
  expect_length(expectations, 4)
})

test_that("otel instrumentation works", {
  skip_if_not_installed("otelsdk")

  record <- with_otel_record({
    test_that("testing is traced", {
      expect_equal(1, 1)
      expect_error(stop("expected error"))
    })
    test_that("all expectations are recorded", {
      expect_equal(1, 1)
      expect_true(TRUE)
      expect_length(1:3, 3)
      expect_warning(warning("expected warning"))
      expect_error(stop("expected error"))
    })
  })

  traces <- record$traces
  expect_length(traces, 2L)
  span <- traces[[1L]]
  expect_equal(
    span$name,
    "test that otel instrumentation works / testing is traced"
  )
  expect_equal(span$instrumentation_scope$name, "org.r-lib.testthat")
  span <- traces[[2L]]
  expect_equal(span$attributes[["test.status"]], "pass")
  expect_equal(span$attributes[["test.expectations.total"]], 5)
  expect_equal(span$attributes[["test.expectations.passed"]], 5)
  expect_equal(span$attributes[["test.expectations.failed"]], 0)
  expect_equal(span$attributes[["test.expectations.error"]], 0)
  expect_equal(span$attributes[["test.expectations.skipped"]], 0)
  expect_equal(span$attributes[["test.expectations.warning"]], 0)
  expect_equal(span$status, "ok")
})

test_that("otel instrumentation works with describe/it", {
  skip_if_not_installed("otelsdk")

  record <- with_otel_record({
    with_reporter("silent", {
      describe("a feature", {
        it("passes", {
          expect_true(TRUE)
        })
        it("fails", {
          expect_equal(1, 1)
          expect_true(FALSE)
        })
      })
    })
  })

  traces <- record$traces
  expect_length(traces, 3L)
  expect_equal(traces[[1L]]$name, "test that a feature / passes")
  expect_equal(traces[[1L]]$attributes[["test.expectations.total"]], 1)
  expect_equal(traces[[1L]]$attributes[["test.status"]], "pass")
  expect_equal(traces[[1L]]$status, "ok")
  expect_equal(traces[[2L]]$name, "test that a feature / fails")
  expect_equal(traces[[2L]]$attributes[["test.expectations.total"]], 2)
  expect_equal(traces[[2L]]$attributes[["test.expectations.passed"]], 1)
  expect_equal(traces[[2L]]$attributes[["test.expectations.failed"]], 1)
  expect_equal(traces[[2L]]$attributes[["test.status"]], "fail")
  expect_equal(traces[[2L]]$status, "error")
  expect_equal(traces[[3L]]$name, "test that a feature")
  expect_equal(traces[[3L]]$attributes[["test.expectations.total"]], 0)
})

test_that("otel instrumentation works with nested test_that", {
  skip_if_not_installed("otelsdk")

  record <- with_otel_record({
    with_reporter("silent", {
      test_that("outer test", {
        expect_true(TRUE)
        test_that("inner test fails", {
          expect_equal(1, 2)
        })
      })
    })
  })

  traces <- record$traces
  expect_length(traces, 2L)
  expect_equal(traces[[1L]]$name, "test that outer test / inner test fails")
  expect_equal(traces[[1L]]$attributes[["test.expectations.total"]], 1)
  expect_equal(traces[[1L]]$attributes[["test.expectations.failed"]], 1)
  expect_equal(traces[[1L]]$attributes[["test.status"]], "fail")
  expect_equal(traces[[1L]]$status, "error")
  expect_equal(traces[[2L]]$name, "test that outer test")
  expect_equal(traces[[2L]]$attributes[["test.expectations.total"]], 1)
  expect_equal(traces[[2L]]$attributes[["test.status"]], "pass")
  expect_equal(traces[[2L]]$status, "ok")
})

#' Do you expect a value bigger or smaller than this?
#'
#' These functions compare values of comparable data types, such as numbers,
#' dates, and times.
#'
#' @inheritParams expect_equal
#' @param object,expected A value to compare and its expected bound.
#' @family expectations
#' @examples
#' a <- 9
#' expect_lt(a, 10)
#'
#' \dontrun{
#' expect_lt(11, 10)
#' }
#'
#' a <- 11
#' expect_gt(a, 10)
#' \dontrun{
#' expect_gt(9, 10)
#' }
#' @name comparison-expectations
NULL

expect_compare_ <- function(
  operator = c("<", "<=", ">", ">="),
  act,
  exp,
  trace_env = caller_env()
) {
  operator <- match.arg(operator)
  op <- match.fun(operator)

  cmp <- op(act$val, exp$val)
  if (length(cmp) != 1 || !is.logical(cmp)) {
    cli::cli_abort(
      "Result of comparison must be `TRUE`, `FALSE`, or `NA`",
      call = trace_env
    )
  } else if (!isTRUE(cmp)) {
    msg <- failure_compare(act, exp, operator)
    fail(msg, trace_env = trace_env)
  } else {
    pass()
  }
}

failure_compare <- function(act, exp, operator) {
  actual_op <- switch(operator, "<" = ">=", "<=" = ">", ">" = "<=", ">=" = "<")

  msg_exp <- sprintf("Expected %s %s %s.", act$lab, operator, exp$lab)

  if (is.numeric(act$val)) {
    digits <- max(
      digits(act$val),
      digits(exp$val),
      min_digits(act$val, exp$val)
    )

    msg_act <- sprintf(
      "Actual comparison: %s %s %s",
      num_exact(act$val, digits),
      actual_op,
      num_exact(exp$val, digits)
    )

    diff <- act$val - exp$val
    if (is.na(diff)) {
      msg_diff <- NULL
    } else {
      msg_diff <- sprintf(
        "Difference: %s %s 0",
        num_exact(diff, digits),
        actual_op
      )
    }
  } else {
    msg_act <- sprintf(
      "Actual comparison: \"%s\" %s \"%s\"",
      act$val,
      actual_op,
      exp$val
    )

    if (inherits(act$val, c("Date", "POSIXt"))) {
      diff <- act$val - exp$val
      if (is.na(diff)) {
        msg_diff <- NULL
      } else {
        msg_diff <- sprintf(
          "Difference: %s %s 0 %s",
          dt_diff(diff),
          actual_op,
          attr(diff, "unit")
        )
      }
    } else {
      msg_diff <- NULL
    }
  }

  c(msg_exp, msg_act, msg_diff)
}

#' @export
#' @rdname comparison-expectations
expect_lt <- function(object, expected, label = NULL, expected.label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  expect_compare_("<", act, exp)
  invisible(act$val)
}

#' @export
#' @rdname comparison-expectations
expect_lte <- function(object, expected, label = NULL, expected.label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  expect_compare_("<=", act, exp)
  invisible(act$val)
}

#' @export
#' @rdname comparison-expectations
expect_gt <- function(object, expected, label = NULL, expected.label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  expect_compare_(">", act, exp)
  invisible(act$val)
}

#' @export
#' @rdname comparison-expectations
expect_gte <- function(object, expected, label = NULL, expected.label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  expect_compare_(">=", act, exp)
  invisible(act$val)
}


# Wordy names -------------------------------------------------------------

#' Deprecated numeric comparison functions
#'
#' These functions have been deprecated in favour of the more concise
#' [expect_gt()] and [expect_lt()].
#'
#' @export
#' @param ... All arguments passed on to `expect_lt()`/`expect_gt()`.
#' @keywords internal
expect_less_than <- function(...) {
  cli::cli_warn("Deprecated: please use {.fn expect_lt} instead.")
  expect_lt(...)
}

#' @rdname expect_less_than
#' @export
expect_more_than <- function(...) {
  cli::cli_warn("Deprecated: please use {.fn expect_gt} instead.")
  expect_gt(...)
}


# Helpers -----------------------------------------------------------------

num_exact <- function(x, digits = 6) {
  sprintf(paste0("%0.", digits, "f"), x)
}

min_digits <- function(x, y, tolerance = testthat_tolerance()) {
  if (is.integer(x) && is.integer(y)) {
    return(0L)
  }

  attributes(x) <- NULL
  attributes(y) <- NULL

  n <- digits(abs(x - y))
  if (!is.null(tolerance)) {
    n <- min(n, digits(tolerance))
  }

  as.integer(n) + 1L
}

digits <- function(x) {
  x <- x[!is.na(x) & x != 0]
  if (length(x) == 0) {
    return(0)
  }
  scale <- -log10(min(abs(x)))
  if (scale <= 0) {
    0L
  } else {
    ceiling(round(scale, digits = 2))
  }
}

dt_diff <- function(x) {
  val <- unclass(x)
  digits <- digits(abs(val)) + 1
  paste(num_exact(val, digits), attr(x, "unit"))
}

#' Expect that a condition holds.
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' An old style of testing that's no longer encouraged.
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' This style of testing is formally deprecated as of the 3rd edition.
#' Use a more specific `expect_` function instead.
#'
#' @param object Object to test.
#'
#'   Supports limited unquoting to make it easier to generate readable failures
#'   within a function or for loop. See [quasi_label] for more details.
#' @param condition, a function that returns whether or not the condition
#'   is met, and if not, an error message to display.
#' @param label Used to customise failure messages. For expert use only.
#' @param info Extra information to be included in the message. This argument
#'   is soft-deprecated and should not be used in new code. Instead see
#'   alternatives in [quasi_label].
#' @return the (internal) expectation result as an invisible list
#' @keywords internal
#' @export
#' @seealso [fail()] for an expectation that always fails.
#' @examples
#' expect_that(5 * 2, equals(10))
#' expect_that(sqrt(2) ^ 2, equals(2))
#' \dontrun{
#' expect_that(sqrt(2) ^ 2, is_identical_to(2))
#' }
expect_that <- function(object, condition, info = NULL, label = NULL) {
  edition_deprecate(3, "expect_that()")
  condition(object)
}

#' Old-style expectations.
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' Initial testthat used a style of testing that looked like
#' `expect_that(a, equals(b)))` this allowed expectations to read like
#' English sentences, but was verbose and a bit too cutesy. This style
#' will continue to work but has been soft-deprecated - it is no longer
#' documented, and new expectations will only use the new style
#' `expect_equal(a, b)`.
#'
#' @name oldskool
#' @keywords internal
NULL

#' @export
#' @rdname oldskool
is_a <- function(class) {
  function(x) expect_is(x, class)
}

#' @export
#' @rdname oldskool
has_names <- function(expected, ignore.order = FALSE, ignore.case = FALSE) {
  function(x) {
    expect_named(
      x,
      expected = expected,
      ignore.order = ignore.order,
      ignore.case = ignore.case
    )
  }
}

#' @export
#' @rdname oldskool
is_less_than <- function(expected, label = NULL, ...) {
  function(x) expect_lt(x, expected)
}

#' @export
#' @rdname oldskool
is_more_than <- function(expected, label = NULL, ...) {
  function(x) expect_gt(x, expected)
}

#' @export
#' @rdname oldskool
equals <- function(expected, label = NULL, ...) {
  function(x) expect_equal(x, expected, ..., expected.label = label)
}

#' @export
#' @rdname oldskool
is_equivalent_to <- function(expected, label = NULL) {
  function(x) expect_equivalent(x, expected, expected.label = label)
}

#' @export
#' @rdname oldskool
is_identical_to <- function(expected, label = NULL) {
  function(x) expect_identical(x, expected, expected.label = label)
}

#' @export
#' @rdname oldskool
equals_reference <- function(file, label = NULL, ...) {
  function(x) expect_known_value(x, file, expected.label = label, ...)
}

#' @export
#' @rdname oldskool
shows_message <- function(regexp = NULL, all = FALSE, ...) {
  function(x) expect_message(x, regexp = regexp, all = all, ...)
}

#' @export
#' @rdname oldskool
gives_warning <- function(regexp = NULL, all = FALSE, ...) {
  function(x) expect_warning(x, regexp = regexp, all = all, ...)
}

#' @export
#' @rdname oldskool
prints_text <- function(regexp = NULL, ...) {
  function(x) expect_output(x, regexp, ...)
}

#' @export
#' @rdname oldskool
throws_error <- function(regexp = NULL, ...) {
  function(x) expect_error(x, regexp, ...)
}

#' Does code take less than the expected amount of time to run?
#'
#' This is useful for performance regression testing.
#'
#' @keywords internal
#' @export
#' @param amount maximum duration in seconds
takes_less_than <- function(amount) {
  cli::cli_warn(
    "{.fn takes_less_than} is deprecated because it is stochastic and unreliable."
  )

  function(expr) {
    duration <- system.time(force(expr))["elapsed"]

    if (duration >= amount) {
      fail(paste0("took ", duration, " seconds, which is more than ", amount))
    } else {
      pass()
    }
  }
}

#' Negate an expectation
#'
#' This negates an expectation, making it possible to express that you
#' want the opposite of a standard expectation. This function is deprecated
#' and will be removed in a future version.
#'
#' @param f an existing expectation function
#' @keywords internal
#' @export
not <- function(f) {
  cli::cli_warn("{.fn not} is deprecated.")
  stopifnot(is.function(f))

  negate <- function(expt) {
    if (expectation_success(expt)) {
      msg <- paste0("NOT(", expt$message, ")")
      fail(msg, srcref = expt$srcref)
    } else {
      pass()
    }
  }

  function(...) {
    negate(capture_expectation(f(...)))
  }
}

test_that("basic comparisons work", {
  x <- 10
  expect_success(expect_lt(x, 11))
  expect_snapshot_failure(expect_lt(x, 10))
  expect_success(expect_lte(x, 10))

  expect_success(expect_gt(11, 10))
  expect_snapshot_failure(expect_gt(x, 10))
  expect_success(expect_gte(x, 10))
})

test_that("useful output when numbers are very small", {
  x <- 1e-5
  expect_snapshot_failure(expect_lte(1.1 * x, x))
  expect_snapshot_failure(expect_gt(x, 1.1 * x))
})

test_that("useful output when difference is zero", {
  x <- 100
  expect_snapshot_failure(expect_lt(x, 100))
})

test_that("useful output when differnce is large", {
  x <- 100
  expect_snapshot_failure(expect_lt(x, 0.001))
})

test_that("comparison result object invisibly", {
  out <- expect_invisible(expect_lt(1, 10))
  expect_equal(out, 1)
})

test_that("comparisons with Inf work", {
  expect_success(expect_lt(10, Inf))
  expect_failure(expect_lt(Inf, Inf))
  expect_success(expect_lte(Inf, Inf))

  expect_success(expect_gt(Inf, 10))
  expect_failure(expect_gt(Inf, Inf))
  expect_success(expect_gte(Inf, Inf))

  x <- Inf
  expect_snapshot_failure(expect_lt(x, Inf))
})

test_that("comparisons with NA work", {
  expect_failure(expect_lt(10, NA_real_))
  expect_failure(expect_lt(NA_real_, 10))
  expect_failure(expect_lt(NA_real_, NA_real_))
  expect_failure(expect_lte(NA_real_, NA_real_))

  expect_failure(expect_gt(10, NA_real_))
  expect_failure(expect_gt(NA_real_, 10))
  expect_failure(expect_gt(NA_real_, NA_real_))
  expect_failure(expect_gte(NA_real_, NA_real_))

  x <- NA_real_
  expect_snapshot_failure(expect_lt(x, 10))
})


test_that("comparisons with negative numbers work", {
  expect_success(expect_lt(-5, -2))
  expect_snapshot_failure(expect_gt(-5, -2))
})

test_that("comparisons with POSIXct objects work", {
  time <- as.POSIXct("2020-01-01 01:00:00")
  time2 <- time + 1.5
  expect_success(expect_lt(time, time2))

  # set digits.secs = 1 to ensure consistent output with older R versions
  withr::with_options(c(digits.secs = 1), {
    expect_snapshot_failure(expect_lt(time2, time))
  })
})

test_that("comparisons with Date objects work", {
  date <- as.Date("2020-01-01")
  date2 <- date + 1
  expect_success(expect_gt(date2, date))
  expect_success(expect_gte(date2, date))

  expect_snapshot_failure(expect_gt(date, date2))
})

test_that("comparisons of date/time with NA work", {
  time <- as.POSIXct("2020-01-01 01:00:00")
  date <- as.Date("2020-01-01")

  expect_failure(expect_lt(time, NA))
  expect_failure(expect_gt(date, NA))
})

test_that("comparisons with character objects work", {
  expect_success(expect_lte("a", "b"))

  expect_snapshot_failure(expect_lte("b", "a"))
})

test_that("comparison must yield a single logical", {
  expect_snapshot(error = TRUE, expect_lt(1:10, 5))
})

test_that("wordy versions are deprecated", {
  expect_warning(expect_less_than(1, 2), "Deprecated")
  expect_warning(expect_more_than(2, 1), "Deprecated")
})

#' Error if any test fails
#'
#' @description
#' The default reporter used when [expect_that()] is run interactively.
#' It responds by displaying a summary of the number of successes and failures
#' and [stop()]ping on if there are any failures.
#'
#' @export
#' @family reporters
StopReporter <- R6::R6Class(
  "StopReporter",
  inherit = Reporter,
  public = list(
    # All expectations that need to be reported (error, failure, warning, skip)
    issues = NULL,
    # Expectations that should cause the test to fail (error, failure)
    n_fail = 0L,
    # Successful expectations
    n_success = 0L,
    praise = TRUE,
    depth = 0,

    initialize = function(praise = TRUE) {
      super$initialize()
      self$issues <- Stack$new()
      self$praise <- praise
    },

    start_test = function(context, test) {
      if (self$depth == 0) {
        self$n_fail <- 0L
        self$n_success <- 0L
        self$issues <- Stack$new()
      }
      self$depth <- self$depth + 1
    },

    add_result = function(context, test, result) {
      if (expectation_success(result)) {
        self$n_success <- self$n_success + 1
        return()
      }

      if (expectation_broken(result)) {
        self$n_fail <- self$n_fail + 1
      }
      self$issues$push(result)
    },

    end_test = function(context, test) {
      self$depth <- self$depth - 1
      if (self$depth > 0) {
        return()
      }

      self$local_user_output()

      for (issue in self$issues$as_list()) {
        self$cat_line(issue_summary(issue, rule = TRUE, location = FALSE))
      }

      if (self$praise && self$n_fail == 0 && self$n_success > 0) {
        emoji <- praise_emoji()
        self$cat_line(cli::format_inline(
          "{.strong Test passed with {self$n_success} success{?es}{emoji}}."
        ))
      }

      if (self$n_fail > 0) {
        cli::cli_abort(
          "Test failed with {self$n_fail} failure{?s} and {self$n_success} success{?es}.",
          call = NULL
        )
      }
    }
  )
)

#' Run a test
#'
#' @description
#' A test encapsulates a series of expectations about a small, self-contained
#' unit of functionality. Each test contains one or more expectations, such as
#' [expect_equal()] or [expect_error()], and lives in a `test/testhat/test*`
#' file, often together with other tests that relate to the same function or set
#' of functions.
#'
#' Each test has its own execution environment, so an object created in a test
#' also dies with the test. Note that this cleanup does not happen automatically
#' for other aspects of global state, such as session options or filesystem
#' changes. Avoid changing global state, when possible, and reverse any changes
#' that you do make.
#'
#' @param desc Test name. Names should be brief, but evocative. It's common to
#'   write the description so that it reads like a natural sentence, e.g.
#'   `test_that("multiplication works", { ... })`.
#' @param code Test code containing expectations. Braces (`{}`) should always
#'   be used in order to get accurate location data for test failures.
#' @return When run interactively, returns `invisible(TRUE)` if all tests
#'   pass, otherwise throws an error.
#' @export
#' @examples
#' test_that("trigonometric functions match identities", {
#'   expect_equal(sin(pi / 4), 1 / sqrt(2))
#'   expect_equal(cos(pi / 4), 1 / sqrt(2))
#'   expect_equal(tan(pi / 4), 1)
#' })
#'
#' \dontrun{
#' test_that("trigonometric functions match identities", {
#'   expect_equal(sin(pi / 4), 1)
#' })
#' }
test_that <- function(desc, code) {
  local_description_push(desc)

  code <- substitute(code)
  test_code(code, parent.frame())
}

# Access error fields with `[[` rather than `$` because the
# `$.Throwable` from the rJava package throws with unknown fields
test_code <- function(code, env, reporter = NULL, skip_on_empty = TRUE) {
  # Must initialise interactive reporter before local_test_context()
  reporter <- get_reporter() %||% local_interactive_reporter()
  local_test_context()

  frame <- caller_env()

  otel_n_success <- 0L
  otel_n_failure <- 0L
  otel_n_error <- 0L
  otel_n_skip <- 0L
  otel_n_warning <- 0L

  test <- test_description()
  if (!is.null(test)) {
    span <- otel_local_test_span(test, scope = frame)
    reporter$start_test(context = reporter$.context, test = test)
    withr::defer({
      otel_update_span(
        span,
        otel_n_success,
        otel_n_failure,
        otel_n_error,
        otel_n_skip,
        otel_n_warning
      )
      reporter$end_test(context = reporter$.context, test = test)
    })
  }

  if (the$top_level_test) {
    # Not strictly necessary but nice to reset the count
    the$test_expectations <- 0
    the$top_level_test <- FALSE
    withr::defer(the$top_level_test <- TRUE)
  }
  # Used to skip if the test _and_ its subtests have no expectations
  starting_expectations <- the$test_expectations

  ok <- TRUE
  snapshot_skipped <- FALSE

  # @param debug_end How many frames should be skipped to find the
  #   last relevant frame call. Only useful for the DebugReporter.
  register_expectation <- function(e, debug_end) {
    srcref <- e[["srcref"]] %||% find_expectation_srcref(frame)
    e <- as.expectation(e, srcref = srcref)

    # Data for the DebugReporter
    if (debug_end >= 0) {
      start <- eval_bare(quote(base::sys.nframe()), test_env) + 1L
      e$start_frame <- start
      e$end_frame <- sys.nframe() - debug_end - 1L
    }

    e$test <- test %||% "(code run outside of `test_that()`)"

    # record keeping for otel
    switch(
      expectation_type(e),
      success = otel_n_success <<- otel_n_success + 1L,
      failure = otel_n_failure <<- otel_n_failure + 1L,
      error = otel_n_error <<- otel_n_error + 1L,
      skip = otel_n_skip <<- otel_n_skip + 1L,
      warning = otel_n_warning <<- otel_n_warning + 1L,
      NULL
    )

    ok <<- ok && expectation_ok(e)
    reporter$add_result(context = reporter$.context, test = test, result = e)
  }

  expressions_opt <- getOption("expressions")
  expressions_opt_new <- min(expressions_opt + 500L, 500000L)

  handle_error <- function(e) {
    the$test_expectations <- the$test_expectations + 1L

    # Increase option(expressions) to handle errors here if possible, even in
    # case of a stack overflow. This is important for the DebugReporter.
    local_options(expressions = expressions_opt_new)

    # Add structured backtrace to the expectation
    if (can_entrace(e)) {
      e <- cnd_entrace(e)
    }

    register_expectation(e, 2)
    invokeRestart("end_test")
  }
  handle_fatal <- function(e) {
    the$test_expectations <- the$test_expectations + 1L
    register_expectation(e, 0)
  }
  handle_expectation <- function(e) {
    the$test_expectations <- the$test_expectations + 1L
    register_expectation(e, 11)
    invokeRestart("muffle_expectation")
  }
  handle_warning <- function(e) {
    # When options(warn) < 0, warnings are expected to be ignored.
    if (getOption("warn") < 0) {
      return()
    }

    # When options(warn) >= 2, warnings are converted to errors.
    # So, do not handle it here so that it will be handled by handle_error.
    if (getOption("warn") >= 2) {
      return()
    }

    if (!inherits(e, "testthat_warn")) {
      e <- cnd_entrace(e)
    }

    register_expectation(e, 5)
    tryInvokeRestart("muffleWarning")
  }
  handle_message <- function(e) {
    if (edition_get() < 3) {
      tryInvokeRestart("muffleMessage")
    }
  }
  handle_skip <- function(e) {
    the$test_expectations <- the$test_expectations + 1L

    debug_end <- if (inherits(e, "skip_empty")) -1 else 2
    register_expectation(e, debug_end)
    invokeRestart("end_test")
  }
  handle_interrupt <- function(e) {
    if (!is.null(test)) {
      cat("\n")
      cli::cli_inform(c("!" = "Interrupting test: {test}"))
    }
  }

  test_env <- new.env(parent = env)
  old <- options(rlang_trace_top_env = test_env)[[1]]
  withr::defer(options(rlang_trace_top_env = old))

  withr::local_options(testthat_topenv = test_env)

  before <- inspect_state()
  withRestarts(
    tryCatch(
      withCallingHandlers(
        {
          eval(code, test_env)
          new_expectations <- the$test_expectations > starting_expectations
          if (snapshot_skipped) {
            skip("On CRAN")
          } else if (!new_expectations && skip_on_empty) {
            skip_empty()
          }
        },
        expectation = handle_expectation,
        packageNotFoundError = function(e) {
          if (on_cran()) {
            skip(paste0("{", e$package, "} is not installed."))
          }
        },
        snapshot_on_cran = function(cnd) {
          snapshot_skipped <<- TRUE
          invokeRestart("muffle_cran_snapshot")
        },
        skip = handle_skip,
        warning = handle_warning,
        message = handle_message,
        error = handle_error,
        interrupt = handle_interrupt
      ),
      # some errors may need handling here, e.g., stack overflow
      error = handle_fatal
    ),
    end_test = function() {}
  )
  after <- inspect_state()

  if (!is.null(test)) {
    cnd <- testthat_state_condition(before, after, call = sys.call(-1))
    if (!is.null(cnd)) {
      register_expectation(cnd, 0)
    }
  }

  invisible(ok)
}


# Maintain a stack of descriptions
local_description_push <- function(description, frame = caller_env()) {
  check_string(description, call = frame)
  local_description_set(c(the$description, description), frame = frame)
}
local_description_set <- function(
  description = character(),
  frame = caller_env()
) {
  check_character(description, call = frame)

  old <- the$description
  the$description <- description
  withr::defer(the$description <- old, frame)

  invisible(old)
}

test_description <- function(desc = the$description) {
  if (length(desc) == 0) {
    NULL
  } else {
    paste(desc, collapse = " / ")
  }
}

#' @keywords internal
#' @rdname run_cpp_tests
#' @export
expect_cpp_tests_pass <- function(package) {
  run_testthat_tests <- get_routine(package, "run_testthat_tests")

  output <- ""
  tests_passed <- TRUE

  tryCatch(
    output <- capture_output_lines(tests_passed <- .Call(run_testthat_tests, FALSE)),
    error = function(e) {
      warning(sprintf("failed to call test entrypoint '%s'", run_testthat_tests))
    }
  )

  # Drop first line of output (it's jut a '####' delimiter)
  info <- paste(output[-1], collapse = "\n")

  expect(tests_passed, paste("C++ unit tests:", info, sep = "\n"))
}

#' Do C++ tests past?
#'
#' Test compiled code in the package `package`. A call to this function will
#' automatically be generated for you in `tests/testthat/test-cpp.R` after
#' calling [use_catch()]; you should not need to manually call this expectation
#' yourself.
#'
#' @param package The name of the package to test.
#' @keywords internal
#' @export
run_cpp_tests <- function(package) {
  skip_on_os("solaris")
  check_installed("xml2", "to run run_cpp_tests()")

  run_testthat_tests <- get_routine(package, "run_testthat_tests")

  output <- ""
  tests_passed <- TRUE

  catch_error <- FALSE
  tryCatch({
    output <- capture_output_lines(tests_passed <- .Call(run_testthat_tests, TRUE))
  },
    error = function(e) {
      catch_error <- TRUE
      reporter <- get_reporter()

      context_start("Catch")
      reporter$start_test(context = "Catch", test = "Catch")
      reporter$add_result(context = "Catch", test = "Catch", result = expectation("failure", e$message))
      reporter$end_test(context = "Catch", test = "Catch")
    }
  )


  if (catch_error) {
    return()
  }

  report <- xml2::read_xml(paste(output, collapse = "\n"))

  contexts <- xml2::xml_find_all(report, "//TestCase")

  for (context in contexts) {
    context_name <- sub(" [|][^|]+$", "", xml2::xml_attr(context, "name"))

    context_start(context_name)

    tests <- xml2::xml_find_all(context, "./Section")
    for (test in tests) {
      test_name <- xml2::xml_attr(test, "name")

      result <- xml2::xml_find_first(test, "./OverallResults")
      successes <- as.integer(xml2::xml_attr(result, "successes"))

      get_reporter()$start_test(context = context_name, test = test_name)

      for (i in seq_len(successes)) {
        exp <- expectation("success", "")
        exp$test <- test_name
        get_reporter()$add_result(context = context_name, test = test_name, result = exp)
      }

      failures <- xml2::xml_find_all(test, "./Expression")
      for (failure in failures) {
        org <- xml2::xml_find_first(failure, "Original")
        org_text <- xml2::xml_text(org, trim = TRUE)

        filename <- xml2::xml_attr(failure, "filename")
        type <- xml2::xml_attr(failure, "type")

        type_msg <- switch(type,
          "CATCH_CHECK_FALSE" = "isn't false.",
          "CATCH_CHECK_THROWS" = "did not throw an exception.",
          "CATCH_CHECK_THROWS_AS" = "threw an exception with unexpected type.",
          "isn't true."
        )

        org_text <- paste(org_text, type_msg)

        line <- xml2::xml_attr(failure, "line")
        failure_srcref <- srcref(srcfile(file.path("src", filename)), c(line, line, 1, 1))

        exp <- expectation("failure", org_text, srcref = failure_srcref)
        exp$test <- test_name

        get_reporter()$add_result(context = context_name, test = test_name, result = exp)
      }

      exceptions <- xml2::xml_find_all(test, "./Exception")
      for (exception in exceptions) {
        exception_text <- xml2::xml_text(exception, trim = TRUE)
        filename <- xml2::xml_attr(exception, "filename")
        line <- xml2::xml_attr(exception, "line")

        exception_srcref <- srcref(srcfile(file.path("src", filename)), c(line, line, 1, 1))

        exp <- expectation("error", exception_text, srcref = exception_srcref)
        exp$test <- test_name

        get_reporter()$add_result(context = context_name, test = test_name, result = exp)
      }

      get_reporter()$end_test(context = context_name, test = test_name)
    }
  }
}

#' Use Catch for C++ Unit Testing
#'
#' Add the necessary infrastructure to enable C++ unit testing
#' in \R packages with [Catch](https://github.com/catchorg/Catch2) and
#' `testthat`.
#'
#' Calling `use_catch()` will:
#'
#' 1. Create a file `src/test-runner.cpp`, which ensures that the
#'    `testthat` package will understand how to run your package's
#'    unit tests,
#'
#' 2. Create an example test file `src/test-example.cpp`, which
#'    showcases how you might use Catch to write a unit test,
#'
#' 3. Add a test file `tests/testthat/test-cpp.R`, which ensures that
#'    `testthat` will run your compiled tests during invocations of
#'    `devtools::test()` or `R CMD check`, and
#'
#' 4. Create a file `R/catch-routine-registration.R`, which ensures that
#'    \R will automatically register this routine when
#'    `tools::package_native_routine_registration_skeleton()` is invoked.
#'
#' You will also need to:
#'
#' * Add xml2 to Suggests, with e.g. `usethis::use_package("xml2", "Suggests")`
#' * Add testthat to LinkingTo, with e.g.
#'   `usethis::use_package("testthat", "LinkingTo")`
#'
#' C++ unit tests can be added to C++ source files within the
#' `src` directory of your package, with a format similar
#' to \R code tested with `testthat`. Here's a simple example
#' of a unit test written with `testthat` + Catch:
#'
#' \preformatted{
#' context("C++ Unit Test") {
#'   test_that("two plus two is four") {
#'     int result = 2 + 2;
#'     expect_true(result == 4);
#'   }
#' }
#' }
#'
#' When your package is compiled, unit tests alongside a harness
#' for running these tests will be compiled into your \R package,
#' with the C entry point `run_testthat_tests()`. `testthat`
#' will use that entry point to run your unit tests when detected.
#'
#' @section Functions:
#'
#' All of the functions provided by Catch are
#' available with the `CATCH_` prefix -- see
#' [here](https://github.com/catchorg/Catch2/blob/master/docs/assertions.md)
#' for a full list. `testthat` provides the
#' following wrappers, to conform with `testthat`'s
#' \R interface:
#'
#' \tabular{lll}{
#' \strong{Function} \tab \strong{Catch} \tab \strong{Description} \cr
#' `context` \tab `CATCH_TEST_CASE` \tab The context of a set of tests. \cr
#' `test_that` \tab `CATCH_SECTION` \tab A test section. \cr
#' `expect_true` \tab `CATCH_CHECK` \tab Test that an expression evaluates to `true`. \cr
#' `expect_false` \tab `CATCH_CHECK_FALSE` \tab Test that an expression evalutes to `false`. \cr
#' `expect_error` \tab `CATCH_CHECK_THROWS` \tab Test that evaluation of an expression throws an exception. \cr
#' `expect_error_as` \tab `CATCH_CHECK_THROWS_AS` \tab Test that evaluation of an expression throws an exception of a specific class. \cr
#' }
#'
#' In general, you should prefer using the `testthat`
#' wrappers, as `testthat` also does some work to
#' ensure that any unit tests within will not be compiled or
#' run when using the Solaris Studio compilers (as these are
#' currently unsupported by Catch). This should make it
#' easier to submit packages to CRAN that use Catch.
#'
#' @section Symbol Registration:
#'
#' If you've opted to disable dynamic symbol lookup in your
#' package, then you'll need to explicitly export a symbol
#' in your package that `testthat` can use to run your unit
#' tests. `testthat` will look for a routine with one of the names:
#'
#' \preformatted{
#'     C_run_testthat_tests
#'     c_run_testthat_tests
#'     run_testthat_tests
#' }
#'
#' See [Controlling Visibility](https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Controlling-visibility)
#' and [Registering Symbols](https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Registering-symbols)
#' in the **Writing R Extensions** manual for more information.
#'
#' @section Advanced Usage:
#'
#' If you'd like to write your own Catch test runner, you can
#' instead use the `testthat::catchSession()` object in a file
#' with the form:
#'
#' \preformatted{
#' #define TESTTHAT_TEST_RUNNER
#' #include <testthat.h>
#'
#' void run()
#' {
#'     Catch::Session& session = testthat::catchSession();
#'     // interact with the session object as desired
#' }
#' }
#'
#' This can be useful if you'd like to run your unit tests
#' with custom arguments passed to the Catch session.
#'
#' @param dir The directory containing an \R package.
#'
#' @section Standalone Usage:
#'
#' If you'd like to use the C++ unit testing facilities provided
#' by Catch, but would prefer not to use the regular `testthat`
#' \R testing infrastructure, you can manually run the unit tests
#' by inserting a call to:
#'
#' \preformatted{
#' .Call("run_testthat_tests", PACKAGE = <pkgName>)
#' }
#'
#' as necessary within your unit test suite.
#'
#' @export
#' @seealso [Catch](https://github.com/catchorg/Catch2/blob/master/docs/assertions.md),
#'   the library used to enable C++ unit testing.
use_catch <- function(dir = getwd()) {
  desc_path <- file.path(dir, "DESCRIPTION")
  if (!file.exists(desc_path)) {
    stop("no DESCRIPTION file at path '", desc_path, "'", call. = FALSE)
  }

  desc <- read.dcf(desc_path, all = TRUE)
  pkg <- desc$Package
  if (!nzchar(pkg)) {
    stop("no 'Package' field in DESCRIPTION file '", desc_path, "'", call. = FALSE)
  }

  src_dir <- file.path(dir, "src")
  if (!file.exists(src_dir) && !dir.create(src_dir)) {
    stop("failed to create 'src/' directory '", src_dir, "'", call. = FALSE)
  }

  test_runner_path <- file.path(src_dir, "test-runner.cpp")

  # Copy the test runner.
  success <- file.copy(
    system.file(package = "testthat", "resources", "test-runner.cpp"),
    test_runner_path,
    overwrite = TRUE
  )

  if (!success) {
    stop("failed to copy 'test-runner.cpp' to '", src_dir, "'", call. = FALSE)
  }

  # Copy the test example.
  success <- file.copy(
    system.file(package = "testthat", "resources", "test-example.cpp"),
    file.path(src_dir, "test-example.cpp"),
    overwrite = TRUE
  )

  if (!success) {
    stop("failed to copy 'test-example.cpp' to '", src_dir, "'", call. = FALSE)
  }

  # Copy the 'test-cpp.R' file.
  test_dir <- file.path(dir, "tests", "testthat")
  if (!file.exists(test_dir) && !dir.create(test_dir, recursive = TRUE)) {
    stop("failed to create 'tests/testthat/' directory '", test_dir, "'", call. = FALSE)
  }

  template_file <- system.file(package = "testthat", "resources", "test-cpp.R")
  contents <- readChar(template_file, file.info(template_file)$size, TRUE)
  transformed <- sprintf(contents, pkg)
  output_path <- file.path(test_dir, "test-cpp.R")
  cat(transformed, file = output_path)

  # Copy the 'test-runner.R file.
  template_file <- system.file(package = "testthat", "resources", "catch-routine-registration.R")
  contents <- readChar(template_file, file.info(template_file)$size, TRUE)
  transformed <- sprintf(contents, pkg)
  output_path <- file.path(dir, "R", "catch-routine-registration.R")
  cat(transformed, file = output_path)

  message("> Added C++ unit testing infrastructure.")
  message("> Please ensure you have 'LinkingTo: testthat' in your DESCRIPTION.")
  message("> Please ensure you have 'Suggests: xml2' in your DESCRIPTION.")
  message("> Please ensure you have 'useDynLib(", pkg, ", .registration = TRUE)' in your NAMESPACE.")
}

get_routine <- function(package, routine) {

  # check to see if the package has explicitly exported
  # the associated routine (check common prefixes as we
  # don't necessarily have access to the NAMESPACE and
  # know what the prefix is)
  namespace <- asNamespace(package)
  prefixes <- c("C_", "c_", "C", "c", "_", "")
  for (prefix in prefixes) {
    name <- paste(prefix, routine, sep = "")
    if (exists(name, envir = namespace)) {
      symbol <- get(name, envir = namespace)
      if (inherits(symbol, "NativeSymbolInfo")) {
        return(symbol)
      }
    }
  }

  # otherwise, try to resolve the symbol dynamically
  for (prefix in prefixes) {
    name <- paste(prefix, routine, sep = "")
    resolved <- tryCatch(
      getNativeSymbolInfo(routine, PACKAGE = package),
      error = function(e) NULL
    )
    if (inherits(resolved, "NativeSymbolInfo")) {
      return(resolved)
    }
  }

  # if we got here, we failed to find the symbol -- throw an error
  fmt <- "failed to locate routine '%s' in package '%s'"
  stop(sprintf(fmt, routine, package), call. = FALSE)
}

(function() {
  .Call(run_testthat_tests, TRUE)
})

#' Locate a file in the testing directory
#'
#' Many tests require some external file (e.g. a `.csv` if you're testing a
#' data import function) but the working directory varies depending on the way
#' that you're running the test (e.g. interactively, with `devtools::test()`,
#' or with `R CMD check`). `test_path()` understands these variations and
#' automatically generates a path relative to `tests/testthat`, regardless of
#' where that directory might reside relative to the current working directory.
#'
#' @param ... Character vectors giving path components.
#' @return A character vector giving the path.
#' @export
#' @examples
#' \dontrun{
#' test_path("foo.csv")
#' test_path("data", "foo.csv")
#' }
test_path <- function(...) {
  if (is_testing() && !isTRUE(getOption("testthat_interactive"))) {
    base <- NULL
  } else if (pkgload::is_loading()) {
    # Probably called from a helper file
    base <- NULL
  } else {
    base <- "tests/testthat"

    if (!dir.exists(base)) {
      cli::cli_abort("Can't find {.path {base}}.")
    }
  }

  file_path(base, ...)
}

file_path <- function(...) {
  paths <- compact(list2(...))
  if (length(paths) == 0) {
    "."
  } else {
    do.call(file.path, paths)
  }
}

test_that("crash", {
  skip_on_cran()
  skip_on_covr()
  skip_if_not(getRversion() >= "4.4.0")

  withr::local_envvar(TESTTHAT_PARALLEL = "TRUE")

  pkg <- test_path("test-parallel", "crash")
  err <- callr::r(
    function() {
      tryCatch(
        testthat::test_local(
          ".",
          reporter = "summary",
          stop_on_failure = FALSE
        ),
        error = function(e) e
      )
    },
    wd = pkg
  )
  expect_s3_class(err, "testthat_process_error")
  expect_equal(err$test_file, "test-crash-3.R")
})

test_that("source_file always uses UTF-8 encoding", {
  has_locale <- function(l) {
    has <- TRUE
    tryCatch(
      withr::with_locale(c(LC_CTYPE = l), "foobar"),
      warning = function(w) has <<- FALSE,
      error = function(e) has <<- FALSE
    )
    has
  }

  ## Some text in UTF-8
  tmp <- tempfile()
  withr::defer(unlink(tmp))
  utf8 <- as.raw(c(
    0xc3,
    0xa1,
    0x72,
    0x76,
    0xc3,
    0xad,
    0x7a,
    0x74,
    0xc5,
    0xb1,
    0x72,
    0xc5,
    0x91,
    0x20,
    0x74,
    0xc3,
    0xbc,
    0x6b,
    0xc3,
    0xb6,
    0x72,
    0x66,
    0xc3,
    0xba,
    0x72,
    0xc3,
    0xb3,
    0x67,
    0xc3,
    0xa9,
    0x70
  ))
  writeBin(c(charToRaw("x <- \""), utf8, charToRaw("\"\n")), tmp)

  run_test <- function(locale) {
    if (has_locale(locale)) {
      env <- new.env()
      withr::with_locale(
        c(LC_CTYPE = locale),
        source_file(tmp, env = env, wrap = FALSE)
      )
      expect_equal(Encoding(env$x), "UTF-8")
      expect_equal(charToRaw(env$x), utf8)
    }
  }

  ## Try to read it in latin1 and UTF-8 locales
  ## They have different names on Unix and Windows
  run_test("en_US.ISO8859-1")
  run_test("en_US.UTF-8")
  run_test("English_United States.1252")
  run_test("German_Germany.1252")
  run_test(Sys.getlocale("LC_CTYPE"))
})

test_that("source_file wraps error", {
  expect_snapshot(error = TRUE, {
    source_file(test_path("reporters/error-setup.R"), wrap = FALSE)
  })
})

test_that("checks its inputs", {
  expect_snapshot(error = TRUE, {
    source_file(1)
    source_file("x")
    source_file(".", "x")
  })
})

# filter_desc -------------------------------------------------------------

test_that("works with all subtest types", {
  code <- exprs(
    test_that("foo", {}),
    describe("bar", {}),
    it("baz", {})
  )
  expect_equal(filter_desc(code, "foo"), code[1])
  expect_equal(filter_desc(code, "bar"), code[2])
  expect_equal(filter_desc(code, "baz"), code[3])
})

test_that("only returns non-subtest code before subtest", {
  code <- exprs(
    f(),
    test_that("bar", {}),
    describe("foo", {}),
    g(),
    h()
  )
  expect_equal(filter_desc(code, "foo"), code[c(1, 3)])
})

test_that("can select recursively", {
  code <- exprs(
    x <- 1,
    describe("a", {
      y <- 1
      describe("b", {
        z <- 1
      })
      y <- 2
    }),
    x <- 2
  )

  expect_equal(
    filter_desc(code, c("a", "b")),
    exprs(
      x <- 1,
      describe("a", {
        y <- 1
        describe("b", {
          z <- 1
        })
      })
    )
  )
})

test_that("works on code like the describe() example", {
  code <- exprs(
    describe("math library", {
      x1 <- 1
      x2 <- 1
      describe("addition()", {
        it("can add two numbers", {
          expect_equal(x1 + x2, addition(x1, x2))
        })
      })
      describe("division()", {
        x1 <- 10
        x2 <- 2
        it("can divide two numbers", {
          expect_equal(x1 / x2, division(x1, x2))
        })
        it("can handle division by 0") #not yet implemented
      })
    })
  )

  expect_equal(
    filter_desc(
      code,
      c("math library", "division()", "can divide two numbers")
    ),
    exprs(
      describe("math library", {
        x1 <- 1
        x2 <- 1
        describe("division()", {
          x1 <- 10
          x2 <- 2
          it("can divide two numbers", {
            expect_equal(x1 / x2, division(x1, x2))
          })
        })
      })
    )
  )

  # what happens for an unimplemented specification?
  expect_snapshot(
    error = TRUE,
    filter_desc(
      code,
      c("math library", "division()", "can handle division by 0")
    )
  )
})

test_that("preserve srcrefs", {
  code <- parse(
    keep.source = TRUE,
    text = '
    test_that("foo", {
      # this is a comment
    })
  '
  )
  expect_snapshot(filter_desc(code, "foo"))
})

test_that("errors if zero or duplicate labels", {
  code <- exprs(
    f(),
    test_that("baz", {}),
    test_that("baz", {}),
    g()
  )

  expect_snapshot(error = TRUE, {
    filter_desc(code, "baz")
    filter_desc(code, "missing")
  })
})

test_that("source_dir()", {
  res <- source_dir("test_dir", pattern = "hello", chdir = TRUE, wrap = FALSE)
  expect_equal(res[[1]](), "Hello World")

  res <- source_dir(
    normalizePath("test_dir"),
    pattern = "hello",
    chdir = TRUE,
    wrap = FALSE
  )
  expect_equal(res[[1]](), "Hello World")

  res <- source_dir("test_dir", pattern = "hello", chdir = FALSE, wrap = FALSE)
  expect_equal(res[[1]](), "Hello World")

  res <- source_dir(
    normalizePath("test_dir"),
    pattern = "hello",
    chdir = FALSE,
    wrap = FALSE
  )
  expect_equal(res[[1]](), "Hello World")
})

test_that("basic report works", {
  withr::defer(unlink(test_path("testthat-problems.rds")))
  expect_snapshot_reporter(CheckReporter$new())

  rds <- test_path("testthat-problems.rds")
  expect_true(file.exists(rds))
})

test_that("doesn't truncate long lines", {
  on.exit(unlink(test_path("testthat-problems.rds")))

  expect_snapshot_reporter(CheckReporter$new(), test_path("reporters/long-test.R"))
})

test_that("always shows summary", {
  file.create(test_path("testthat-problems.rds"))

  expect_snapshot_reporter(CheckReporter$new(), test_path("reporters/successes.R"))
  # and cleans up testthat-problems
  expect_false(file.exists(test_path("testthat-problems.rds")))
})

test_that("shows warnings when not on CRAN", {
  on.exit(unlink(test_path("testthat-problems.rds")))

  withr::local_options("NOT_CRAN" = "true")
  expect_snapshot_reporter(CheckReporter$new(), test_path("reporters/tests.R"))
})

test_that("set_state_inspector() verifies its inputs", {
  expect_snapshot(set_state_inspector(function(x) 123), error = TRUE)
})

test_that("can detect state changes", {
  local_options(x = NULL)
  set_state_inspector(function() list(x = getOption("x")))
  withr::defer(set_state_inspector(NULL))

  expect_snapshot_reporter(CheckReporter$new(), test_path("reporters/state-change.R"))
})

test_that("detect number of cpus to use", {
  withr::local_options(Ncpus = 100L)
  withr::local_envvar(TESTTHAT_CPUS = NA)
  expect_equal(default_num_cpus(), 100L)

  withr::local_options(Ncpus = 100L)
  withr::local_envvar(TESTTHAT_CPUS = 10)
  expect_equal(default_num_cpus(), 100L)

  withr::local_options(list(Ncpus = NULL))
  withr::local_envvar(TESTTHAT_CPUS = NA)
  expect_equal(default_num_cpus(), 2L)

  withr::local_options(list(Ncpus = NULL))
  withr::local_envvar(TESTTHAT_CPUS = NA)
  expect_equal(default_num_cpus(), 2L)

  withr::local_options(list(Ncpus = NULL))
  withr::local_envvar(TESTTHAT_CPUS = 13)
  expect_equal(default_num_cpus(), 13L)
})

test_that("good error if bad option", {
  withr::local_options(Ncpus = "bad")
  expect_snapshot(default_num_cpus(), error = TRUE)
})

test_that("ok", {
  skip_on_covr()
  withr::local_envvar(c(TESTTHAT_PARALLEL = "TRUE"))
  # we cannot run these with the silent reporter, because it is not
  # parallel compatible, and they'll not run in parallel
  capture.output(suppressMessages(
    ret <- test_local(
      test_path("test-parallel", "ok"),
      reporter = "summary",
      stop_on_failure = FALSE
    )
  ))
  tdf <- as.data.frame(ret)
  tdf <- tdf[order(tdf$file), ]
  expect_equal(tdf$failed, c(0, 1, 0))
  expect_equal(tdf$skipped, c(FALSE, FALSE, TRUE))
})

test_that("fail", {
  skip_on_covr()
  withr::local_envvar(c(TESTTHAT_PARALLEL = "TRUE"))
  # we cannot run these with the silent reporter, because it is not
  # parallel compatible, and they'll not run in parallel
  capture.output(suppressMessages(
    ret <- test_local(
      test_path("test-parallel", "fail"),
      reporter = "summary",
      stop_on_failure = FALSE
    )
  ))
  tdf <- as.data.frame(ret)
  tdf <- tdf[order(tdf$file), ]
  expect_equal(tdf$failed, c(1))
})

test_that("snapshots", {
  skip_on_covr()
  skip_on_cran()
  withr::local_envvar(c(TESTTHAT_PARALLEL = "TRUE"))

  tmp <- withr::local_tempdir("testthat-snap-")
  file.copy(test_path("test-parallel", "snap"), tmp, recursive = TRUE)
  # we cannot run these with the silent reporter, because it is not
  # parallel compatible, and they'll not run in parallel
  capture.output(suppressMessages(
    ret <- test_local(
      file.path(tmp, "snap"),
      reporter = "summary",
      stop_on_failure = FALSE
    )
  ))
  tdf <- as.data.frame(ret)
  tdf <- tdf[order(tdf$file), ]
  expect_equal(tdf$failed, c(0, 0, 1))
  snaps <- file.path(tmp, "snap", "tests", "testthat", "_snaps")
  expect_true(file.exists(file.path(snaps, "snap-1.md")))
  expect_true(file.exists(file.path(snaps, "snap-2.md")))
  expect_true(file.exists(file.path(snaps, "snap-3.md")))
})

test_that("new snapshots are added", {
  skip_on_covr()
  skip_on_cran()
  withr::local_envvar(c(TESTTHAT_PARALLEL = "TRUE", CI = "false"))

  tmp <- withr::local_tempdir("testthat-snap-")
  file.copy(test_path("test-parallel", "snap"), tmp, recursive = TRUE)
  unlink(file.path(tmp, "snap", "tests", "testthat", "_snaps", "snap-2.md"))

  # we cannot run these with the silent reporter, because it is not
  # parallel compatible, and they'll not run in parallel
  capture.output(suppressMessages(
    ret <- test_local(
      file.path(tmp, "snap"),
      reporter = "summary",
      stop_on_failure = FALSE
    )
  ))
  tdf <- as.data.frame(ret)
  tdf <- tdf[order(tdf$file), ]
  expect_equal(tdf$failed, c(0, 0, 1))
  snaps <- file.path(tmp, "snap", "tests", "testthat", "_snaps")
  expect_true(file.exists(file.path(snaps, "snap-1.md")))
  expect_true(file.exists(file.path(snaps, "snap-2.md")))
  expect_true(file.exists(file.path(snaps, "snap-3.md")))
})

test_that("snapshots are removed if test file has no snapshots", {
  skip_on_covr()
  skip_on_cran()
  withr::local_envvar(c(TESTTHAT_PARALLEL = "TRUE"))

  tmp <- withr::local_tempdir("testthat-snap-")
  file.copy(test_path("test-parallel", "snap"), tmp, recursive = TRUE)
  writeLines(
    "test_that(\"2\", { expect_true(TRUE) })",
    file.path(tmp, "snap", "tests", "testthat", "test-snap-2.R")
  )

  # we cannot run these with the silent reporter, because it is not
  # parallel compatible, and they'll not run in parallel
  capture.output(suppressMessages(
    ret <- test_local(
      file.path(tmp, "snap"),
      reporter = "summary",
      stop_on_failure = FALSE
    )
  ))
  tdf <- as.data.frame(ret)
  tdf <- tdf[order(tdf$file), ]
  expect_equal(tdf$failed, c(0, 0, 1))
  snaps <- file.path(tmp, "snap", "tests", "testthat", "_snaps")
  expect_true(file.exists(file.path(snaps, "snap-1.md")))
  expect_false(file.exists(file.path(snaps, "snap-2.md")))
  expect_true(file.exists(file.path(snaps, "snap-3.md")))
})

test_that("snapshots are removed if test file is removed", {
  skip_on_covr()
  skip_on_cran()

  withr::local_envvar(c(TESTTHAT_PARALLEL = "TRUE"))
  withr::defer(unlink(tmp, recursive = TRUE))
  dir.create(tmp <- tempfile("testthat-snap-"))
  file.copy(test_path("test-parallel", "snap"), tmp, recursive = TRUE)
  unlink(file.path(tmp, "snap", "tests", "testthat", "test-snap-2.R"))
  withr::local_envvar(CI = NA_character_)
  # we cannot run these with the silent reporter, because it is not
  # parallel compatible, and they'll not run in parallel
  capture.output(suppressMessages(
    ret <- test_local(
      file.path(tmp, "snap"),
      reporter = "summary",
      stop_on_failure = FALSE
    )
  ))
  tdf <- as.data.frame(ret)
  tdf <- tdf[order(tdf$file), ]
  expect_equal(tdf$failed, c(0, 1))
  snaps <- file.path(tmp, "snap", "tests", "testthat", "_snaps")
  expect_true(file.exists(file.path(snaps, "snap-1.md")))
  expect_false(file.exists(file.path(snaps, "snap-2.md")))
  expect_true(file.exists(file.path(snaps, "snap-3.md")))
})

test_that("can control output with file arg/option", {
  # powered through Reporter base class so we only test one reporter
  path <- withr::local_tempfile()

  with_reporter(
    MinimalReporter$new(file = path),
    test_one_file(test_path("reporters/tests.R"))
  )
  expect_snapshot_output(readLines(path))

  withr::local_options(testthat.output_file = path)
  with_reporter(
    MinimalReporter$new(),
    test_one_file(test_path("reporters/tests.R"))
  )
  expect_snapshot_output(readLines(path))
})

test_that("should not automatically skip in non-utf-8 locales", {
  withr::local_locale(LC_CTYPE = "C")
  expect_true(TRUE)
})

test_that("reports issues immediately but not successes", {
  expect_snapshot_reporter(
    LlmReporter$new(),
    test_path("reporters/tests.R")
  )
})

test_that("reports only summary for all successes", {
  expect_snapshot_reporter(
    LlmReporter$new(),
    test_path("reporters/successes.R")
  )
})

test_that("fails after max_fail tests", {
  withr::local_options(testthat.progress.max_fails = 3)
  expect_snapshot_reporter(
    LlmReporter$new(),
    test_path(c("reporters/fail-many.R", "reporters/fail.R"))
  )
})

test_that("is_llm() detects known LLM agent environment variables", {
  withr::local_envvar(
    AGENT = NA,
    CLAUDECODE = NA,
    GEMINI_CLI = NA,
    CURSOR_AGENT = NA
  )
  expect_false(is_llm())

  local({
    withr::local_envvar(AGENT = "1")
    expect_true(is_llm())
  })
  local({
    withr::local_envvar(CLAUDECODE = "1")
    expect_true(is_llm())
  })
})

#' Do you expect a vector with this size and/or prototype?
#'
#' `expect_vector()` is a thin wrapper around [vctrs::vec_assert()], converting
#' the results of that function in to the expectations used by testthat. This
#' means that it used the vctrs of `ptype` (prototype) and `size`. See
#' details in <https://vctrs.r-lib.org/articles/type-size.html>
#'
#' @inheritParams expect_that
#' @param ptype (Optional) Vector prototype to test against. Should be a
#'   size-0 (empty) generalised vector.
#' @param size (Optional) Size to check for.
#' @export
#' @examplesIf requireNamespace("vctrs")
#' expect_vector(1:10, ptype = integer(), size = 10)
#' show_failure(expect_vector(1:10, ptype = integer(), size = 5))
#' show_failure(expect_vector(1:10, ptype = character(), size = 5))
expect_vector <- function(object, ptype = NULL, size = NULL) {
  check_installed("vctrs")
  check_number_whole(size, min = 0, allow_null = TRUE)
  act <- quasi_label(enquo(object))
  # vec_assert() automatically adds backticks so we hack out the ones
  # added by as_label()
  act$lab <- gsub("^`|`$", "", act$lab)

  failed <- FALSE
  withCallingHandlers(
    vctrs::vec_assert(act$val, ptype = ptype, size = size, arg = act$lab),
    vctrs_error_scalar_type = function(e) {
      failed <<- TRUE
      fail(e$message)
    },
    vctrs_error_assert = function(e) {
      failed <<- TRUE
      fail(e$message)
    }
  )

  if (!failed) {
    pass()
  }
  invisible(act$val)
}

test_that("basic properties upheld", {
  skip_if_not_installed("vctrs", "0.1.0.9002")

  expect_success(expect_vector(1:10, size = 10))

  x <- 1:10
  expect_snapshot(error = TRUE, {
    expect_vector(x, size = 5)
  })

  y <- NULL
  expect_snapshot(error = TRUE, {
    expect_vector(y)
  })
})

test_that("expect_vector validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_vector(1:5, size = "large")
  })
})

# Manage a test files worth of snapshots - if the test file uses variants, this
# will correspond to multiple output files.
FileSnaps <- R6::R6Class(
  "FileSnaps",
  public = list(
    snap_path = NULL,
    file = NULL,
    type = NULL,
    snaps = NULL,

    initialize = function(snap_path, file, type = c("old", "cur", "new")) {
      self$snap_path <- snap_path
      self$file <- file
      self$type <- arg_match(type)

      if (self$type == "old") {
        # Find variants
        variants <- c("_default", dirs(self$snap_path))
        paths <- set_names(self$path(variants), variants)
        paths <- paths[file.exists(paths)]

        self$snaps <- lapply(paths, read_snaps)
      } else {
        self$snaps <- list(`_default` = list())
      }
    },

    get = function(test, variant, i) {
      test_snaps <- self$snaps[[variant]][[test]]
      if (i > length(test_snaps)) {
        NULL
      } else {
        test_snaps[[i]]
      }
    },

    set = function(test, variant, i, data) {
      self$snaps[[variant]][[test]][[i]] <- data
    },

    append = function(test, variant, data) {
      self$snaps[[variant]][[test]] <- c(self$snaps[[variant]][[test]], data)
      length(self$snaps[[variant]][[test]])
    },

    reset = function(test, old) {
      for (variant in names(self$snaps)) {
        cur_test <- self$snaps[[variant]][[test]]
        old_test <- old$snaps[[variant]][[test]]

        if (length(cur_test) == 0) {
          self$snaps[[variant]][[test]] <- old_test
        } else if (length(old_test) > length(cur_test)) {
          self$snaps[[variant]][[test]] <- c(
            cur_test,
            old_test[-seq_along(cur_test)]
          )
        }
      }
      invisible()
    },

    write = function(variants = names(self$snaps)) {
      for (variant in variants) {
        default <- variant == "_default"
        if (!default) {
          dir.create(file.path(self$snap_path, variant), showWarnings = FALSE)
        }

        write_snaps(
          self$snaps[[variant]],
          self$path(variant),
          delete = default
        )
      }
      invisible()
    },

    delete = function(variant = "_default") {
      unlink(self$path(variant))
      invisible()
    },

    variants = function() {
      names(self$snaps)
    },

    filename = function() {
      paste0(self$file, if (self$type == "new") ".new", ".md")
    },

    path = function(variant = "_default") {
      ifelse(
        variant == "_default",
        file.path(self$snap_path, self$filename()),
        file.path(self$snap_path, variant, self$filename())
      )
    }
  )
)

dirs <- function(path) {
  list.dirs(path, recursive = FALSE, full.names = FALSE)
}

test_that("append manages current snapshot index", {
  path <- withr::local_tempdir()
  snaps <- FileSnaps$new(path, "file")

  i <- snaps$append("test1", "_default", 1)
  expect_equal(i, 1)
  i <- snaps$append("test1", "_default", 2)
  expect_equal(i, 2)
  i <- snaps$append("test1", "windows", 3)
  expect_equal(i, 1)
  i <- snaps$append("test1", "windows", 3)
  expect_equal(i, 2)

  i <- snaps$append("test2", "_default", 3)
  expect_equal(i, 1)
})

test_that("can retrieve appended snaps", {
  snaps <- FileSnaps$new(withr::local_tempdir(), "file")

  snaps$append("test1", "_default", "1")
  snaps$append("test1", "_default", "2")
  snaps$append("test2", "_default", "3")
  snaps$append("test2", "windows", "4")

  expect_equal(snaps$get("test1", "_default", 2), "2")
  expect_equal(snaps$get("test2", "_default", 1), "3")
  expect_equal(snaps$get("test2", "windows", 1), "4")

  # Returns NULL if don't exist
  expect_equal(snaps$get("MISSING", "_default", 1), NULL)
  expect_equal(snaps$get("test1", "MISSING", 1), NULL)
  expect_equal(snaps$get("test1", "_default", 100), NULL)
})

test_that("can reset snapshots", {
  snaps1 <- FileSnaps$new(withr::local_tempdir(), "file")
  snaps2 <- FileSnaps$new(withr::local_tempdir(), "file")

  snaps1$append("test1", "_default", "1")
  snaps1$append("test1", "_default", "2")
  snaps1$append("test2", "_default", "3")

  snaps2$append("test1", "_default", "4")

  snaps2$reset("test1", snaps1)
  expect_equal(snaps2$snaps$`_default`$test1, c("4", "2"))

  # And can copy complete snapshot
  snaps2$reset("test2", snaps1)
  expect_equal(snaps2$snaps$`_default`$test2, "3")

  # And shouldn't change if we reset again
  snaps2$reset("test1", snaps1)
  expect_equal(snaps2$snaps$`_default`$test1, c("4", "2"))
})

test_that("can round trip cur to old snaps", {
  path <- withr::local_tempdir()
  cur <- FileSnaps$new(path, "file", "cur")
  cur$append("test1", "_default", "1")
  cur$append("test2", "_default", "2")
  cur$append("test2", "windows", "3")
  cur$write()

  old <- FileSnaps$new(path, "file", "old")
  expect_equal(cur$snaps, old$snaps)
})

test_that("snaps delete default variant if no snaps", {
  snaps <- FileSnaps$new(withr::local_tempdir(), "file", "cur")
  brio::write_lines("x", snaps$path())

  snaps$write()
  expect_false(file.exists(snaps$path()))
})

#' Mock a sequence of output from a function
#'
#' Specify multiple return values for mocking
#'
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> Values to return in sequence.
#' @param recycle whether to recycle. If `TRUE`, once all values have been returned,
#' they will be returned again in sequence.
#'
#' @return A function that you can use within `local_mocked_bindings()` and
#' `with_mocked_bindings()`
#' @export
#'
#' @examples
#' # inside local_mocked_bindings()
#' \dontrun{
#' local_mocked_bindings(readline = mock_output_sequence("3", "This is a note", "n"))
#' }
#' # for understanding
#' mocked_sequence <- mock_output_sequence("3", "This is a note", "n")
#' mocked_sequence()
#' mocked_sequence()
#' mocked_sequence()
#' try(mocked_sequence())
#' recycled_mocked_sequence <- mock_output_sequence(
#'   "3", "This is a note", "n",
#'   recycle = TRUE
#' )
#' recycled_mocked_sequence()
#' recycled_mocked_sequence()
#' recycled_mocked_sequence()
#' recycled_mocked_sequence()
#' @family mocking
mock_output_sequence <- function(..., recycle = FALSE) {
  values <- rlang::list2(...)
  i <- 1
  function(...) {
    if (i > length(values) && !recycle) {
      cli::cli_abort(c(
        "Can't find value for {i}th iteration.",
        i = "{.arg ...} has only {length(values)} values.",
        i = "You can set {.arg recycle} to {.code TRUE}."
      ))
    }
    index <- (i - 1) %% length(values) + 1
    value <- rep_len(values, length.out = index)[[index]]
    i <<- i + 1
    value
  }
}

test_that("mock_output_sequence() works", {
  mocked_sequence <- mock_output_sequence("3", "This is a note", "n")
  expect_equal(mocked_sequence(), "3")
  expect_equal(mocked_sequence(), "This is a note")
  expect_equal(mocked_sequence(), "n")
  expect_snapshot(mocked_sequence(), error = TRUE)
})

test_that("mock_output_sequence() works -- list", {
  x <- list("3", "This is a note", "n")
  mocked_sequence <- mock_output_sequence(!!!x)
  expect_equal(mocked_sequence(), "3")
  expect_equal(mocked_sequence(), "This is a note")
  expect_equal(mocked_sequence(), "n")
})

test_that("mock_output_sequence()'s recycling works", {
  mocked_sequence <- mock_output_sequence(
    "3",
    "This is a note",
    "n",
    recycle = TRUE
  )
  expect_equal(mocked_sequence(), "3")
  expect_equal(mocked_sequence(), "This is a note")
  expect_equal(mocked_sequence(), "n")
  expect_equal(mocked_sequence(), "3")
  expect_equal(mocked_sequence(), "This is a note")
  expect_equal(mocked_sequence(), "n")
})

#' Evaluate a promise, capturing all types of output.
#'
#' @param code Code to evaluate.
#' @keywords internal
#' @export
#' @return A list containing
#'  \item{result}{The result of the function}
#'  \item{output}{A string containing all the output from the function}
#'  \item{warnings}{A character vector containing the text from each warning}
#'  \item{messages}{A character vector containing the text from each message}
#' @examples
#' evaluate_promise({
#'   print("1")
#'   message("2")
#'   warning("3")
#'   4
#' })
evaluate_promise <- function(code, print = FALSE) {
  warnings <- Stack$new()
  handle_warning <- function(condition) {
    if (!is_deprecation(condition)) {
      warnings$push(condition)
      maybe_restart("muffleWarning")
    }
  }

  messages <- Stack$new()
  handle_message <- function(condition) {
    messages$push(condition)
    maybe_restart("muffleMessage")
  }

  path <- withr::local_tempfile()
  result <- withr::with_output_sink(
    path,
    withCallingHandlers(
      withVisible(code),
      warning = handle_warning,
      message = handle_message
    )
  )

  if (result$visible && print) {
    withr::with_output_sink(path, print(result$value), append = TRUE)
  }

  output <- paste0(brio::read_lines(path), collapse = "\n")

  list(
    result = result$value,
    output = output,
    warnings = get_messages(warnings$as_list()),
    messages = get_messages(messages$as_list())
  )
}

file_teardown_env <- new.env(parent = emptyenv())
file_teardown_env$queue <- list()

#' Run code before/after tests
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' We no longer recommend using `setup()` and `teardown()`; instead
#' we think it's better practice to use a **test fixture** as described in
#' `vignette("test-fixtures")`.
#'
#' Code in a `setup()` block is run immediately in a clean environment.
#' Code in a `teardown()` block is run upon completion of a test file,
#' even if it exits with an error. Multiple calls to `teardown()` will be
#' executed in the order they were created.
#'
#' @param code Code to evaluate
#' @param env Environment in which code will be evaluated. For expert
#'   use only.
#' @export
#' @keywords internal
#' @examples
#' \dontrun{
#' # Old approach
#' tmp <- tempfile()
#' setup(writeLines("some test data", tmp))
#' teardown(unlink(tmp))
#' }
#'
#' # Now recommended:
#' local_test_data <- function(env = parent.frame()) {
#'   tmp <- tempfile()
#'   writeLines("some test data", tmp)
#'   withr::defer(unlink(tmp), env)
#'
#'   tmp
#' }
#' # Then call local_test_data() in your tests
teardown <- function(code, env = parent.frame()) {
  edition_deprecate(
    3,
    "teardown()",
    "Please use test fixtures instead see vignette('test-fixtures') for details"
  )

  fun <- new_function(list(), enexpr(code), env = env)
  file_teardown_env$queue <- append(file_teardown_env$queue, fun)

  invisible()
}

#' @export
#' @rdname teardown
setup <- function(code, env = parent.frame()) {
  edition_deprecate(
    3,
    "setup()",
    "Please use test fixtures instead see vignette('test-fixtures') for details"
  )

  out <- eval_tidy(enquo(code), env = env)
  invisible(out)
}

teardown_reset <- function() {
  file_teardown_env$queue <- list()
}

teardown_run <- function(path = ".") {
  if (length(file_teardown_env$queue) == 0) {
    return()
  }

  old_dir <- setwd(path)
  withr::defer(setwd(old_dir))

  lapply(file_teardown_env$queue, function(f) try(f()))
  teardown_reset()
  gc()
}

test_that("teardown adds to queue", {
  local_edition(2)
  withr::defer({
    teardown_reset()
  })

  expect_length(file_teardown_env$queue, 0)

  teardown({})
  expect_length(file_teardown_env$queue, 1)

  teardown({})
  expect_length(file_teardown_env$queue, 2)
})

test_that("teardowns runs in order", {
  local_edition(2)
  withr::defer(teardown_reset())

  a <- 1
  teardown(a <<- 2)
  teardown(a <<- 3)
  expect_length(file_teardown_env$queue, 2)

  teardown_run()
  expect_equal(a, 3)
  expect_length(file_teardown_env$queue, 0)
})

test_that("teardown run after tests complete", {
  test_file(test_path("test-teardown/test-teardown.R"), "silent")
  expect_false(file.exists(test_path("test-teardown/teardown.txt")))
})

#' Test reporter: summary of errors.
#'
#' This is a reporter designed for interactive usage: it lets you know which
#' tests have run successfully and as well as fully reporting information about
#' failures and errors.
#'
#' You can use the `max_reports` field to control the maximum number
#' of detailed reports produced by this reporter. This is useful when running
#' with [auto_test()]
#'
#' As an additional benefit, this reporter will praise you from time-to-time
#' if all your tests pass.
#'
#' @export
#' @family reporters
SummaryReporter <- R6::R6Class("SummaryReporter",
  inherit = Reporter,
  public = list(
    failures = NULL,
    skips = NULL,
    warnings = NULL,
    max_reports = NULL,
    show_praise = TRUE,
    omit_dots = FALSE,

    initialize = function(show_praise = TRUE,
                          omit_dots = getOption("testthat.summary.omit_dots"),
                          max_reports = getOption("testthat.summary.max_reports", 10L),
                          ...) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
      self$failures <- Stack$new()
      self$skips <- Stack$new()
      self$warnings <- Stack$new()
      self$max_reports <- max_reports
      self$show_praise <- show_praise
      self$omit_dots <- omit_dots
    },

    is_full = function() {
      self$failures$size() >= self$max_reports
    },

    start_file = function(file) {
      context_start_file(file)
    },

    start_context = function(context) {
      self$cat_tight(context, ": ")
    },

    end_context = function(context) {
      self$cat_line()
    },

    add_result = function(context, test, result) {
      if (expectation_broken(result)) {
        self$failures$push(result)
      } else if (expectation_skip(result)) {
        self$skips$push(result)
      } else if (expectation_warning(result)) {
        self$warnings$push(result)
      } else {
        if (isTRUE(self$omit_dots)) {
          return()
        }
      }

      self$cat_tight(private$get_summary(result))
    },

    end_reporter = function() {
      skips <- self$skips$as_list()
      failures <- self$failures$as_list()
      warnings <- self$warnings$as_list()

      self$cat_line()
      private$cat_reports("Skipped", skips, Inf, skip_summary)
      private$cat_reports("Warnings", warnings, Inf, skip_summary)
      private$cat_reports("Failed", failures, self$max_reports, failure_summary)

      if (self$failures$size() >= self$max_reports) {
        self$cat_line(
          "Maximum number of ", self$max_reports, " failures reached, ",
          "some test results may be missing."
        )
        self$cat_line()
      }

      self$rule("DONE", line = 2)
      if (self$show_praise) {
        if (length(failures) == 0 && stats::runif(1) < 0.1) {
          self$cat_line(colourise(praise(), "success"))
        }
        if (length(failures) > 0 && stats::runif(1) < 0.25) {
          self$cat_line(colourise(encourage(), "error"))
        }
      }
    }
  ),

  private = list(
    get_summary = function(result) {
      if (expectation_broken(result)) {
        if (self$failures$size() <= length(labels)) {
          return(colourise(labels[self$failures$size()], "error"))
        }
      }

      single_letter_summary(result)
    },

    cat_reports = function(header, expectations, max_n, summary_fun,
                           collapse = "\n\n") {
      n <- length(expectations)
      if (n == 0L) {
        return()
      }

      self$rule(header, line = 2)

      if (n > max_n) {
        expectations <- expectations[seq_len(max_n)]
      }

      labels <- seq_along(expectations)
      exp_summary <- function(i) {
        summary_fun(expectations[[i]], labels[i])
      }
      report_summary <- vapply(seq_along(expectations), exp_summary, character(1))

      self$cat_tight(paste(report_summary, collapse = collapse))
      if (n > max_n) {
        self$cat_line()
        self$cat_line("  ... and ", n - max_n, " more")
      }

      self$cat_line()
      self$cat_line()
    }
  )
)

labels <- c(1:9, letters, LETTERS)

skip_summary <- function(x, label) {
  header <- paste0(label, ". ", x$test)

  paste0(
    colourise(header, "skip"), expectation_location(x, " (", ")"), " - ", x$message
  )
}

failure_summary <- function(x, label, width = cli::console_width()) {
  header <- paste0(label, ". ", issue_header(x))

  paste0(
    cli::rule(header, col = testthat_style("error")), "\n",
    format(x)
  )
}

#' Test reporter: minimal.
#'
#' The minimal test reporter provides the absolutely minimum amount of
#' information: whether each expectation has succeeded, failed or experienced
#' an error.  If you want to find out what the failures and errors actually
#' were, you'll need to run a more informative test reporter.
#'
#' @export
#' @family reporters
MinimalReporter <- R6::R6Class("MinimalReporter",
  inherit = Reporter,
  public = list(
    initialize = function(...) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
    },

    add_result = function(context, test, result) {
      self$cat_tight(single_letter_summary(result))
    },

    end_reporter = function() {
      self$cat_line()
    }
  )
)

#' Test reporter: location
#'
#' This reporter simply prints the location of every expectation and error.
#' This is useful if you're trying to figure out the source of a segfault,
#' or you want to figure out which code triggers a C/C++ breakpoint
#'
#' @export
#' @family reporters
LocationReporter <- R6::R6Class(
  "LocationReporter",
  inherit = Reporter,
  public = list(
    start_test = function(context, test) {
      self$cat_line("Start test: ", test)
    },

    add_result = function(context, test, result) {
      status <- expectation_type(result)
      self$cat_line("  ", expectation_location(result), " [", status, "]")
    },

    end_test = function(context, test) {
      self$cat_line("End test: ", test)
      self$cat_line()
    }
  )
)

#' Watch a directory for changes (additions, deletions & modifications).
#'
#' This is used to power the [auto_test()] and
#' [auto_test_package()] functions which are used to rerun tests
#' whenever source code changes.
#'
#' Use Ctrl + break (windows), Esc (mac gui) or Ctrl + C (command line) to
#' stop the watcher.
#'
#' @param path character vector of paths to watch.  Omit trailing backslash.
#' @param pattern file pattern passed to [dir()]
#' @param callback function called everytime a change occurs.  It should
#'   have three parameters: added, deleted, modified, and should return
#'   TRUE to keep watching, or FALSE to stop.
#' @param hash hashes are more accurate at detecting changes, but are slower
#'   for large files.  When FALSE, uses modification time stamps
#' @export
#' @keywords internal
watch <- function(path, callback, pattern = NULL, hash = TRUE) {
  prev <- dir_state(path, pattern, hash = hash)

  repeat {
    Sys.sleep(1)

    curr <- dir_state(path, pattern, hash = hash)
    changes <- compare_state(prev, curr)

    if (changes$n > 0) {
      # cat("C")
      keep_going <- TRUE
      try(keep_going <- callback(changes$added, changes$deleted, changes$modified))

      if (!isTRUE(keep_going)) return(invisible())
    } else {
      # cat(".")
    }

    prev <- curr
  }
}

safe_digest <- function(path) {
  if (!file.exists(path)) return(NA_character_)
  if (is_directory(path)) return(NA_character_)
  if (!is_readable(path)) return(NA_character_)

  digest::digest(path, file = TRUE)
}

#' Capture the state of a directory.
#'
#' @param path path to directory
#' @param pattern regular expression with which to filter files
#' @param hash use hash (slow but accurate) or time stamp (fast but less
#'   accurate)
#' @keywords internal
dir_state <- function(path, pattern = NULL, hash = TRUE) {
  files <- dir(path, pattern, full.names = TRUE)

  # It's possible for any of the files to be deleted between the dir()
  # call above and the calls below; `file.info` handles this
  # gracefully, but digest::digest doesn't -- so we wrap it. Both
  # cases will return NA for files that have gone missing.
  if (hash) {
    file_states <- vapply(files, safe_digest, character(1))
  } else {
    file_states <- stats::setNames(file.info(files)$mtime, files)
  }
  file_states[!is.na(file_states)]
}

#' Compare two directory states.
#'
#' @param old previous state
#' @param new current state
#' @return list containing number of changes and files which have been
#'   `added`, `deleted` and `modified`
#' @keywords internal
compare_state <- function(old, new) {
  added <- setdiff(names(new), names(old))
  deleted <- setdiff(names(old), names(new))

  same <- intersect(names(old), names(new))
  modified <- names(new[same])[new[same] != old[same]]

  n <- length(added) + length(deleted) + length(modified)

  list(n = n, added = added, deleted = deleted, modified = modified)
}

# Helpers -----------------------------------------------------------------

is_directory <- function(x) file.info(x)$isdir
is_readable <- function(x) file.access(x, 4) == 0

test_that("compare state works correctly", {
  loc <- withr::local_tempfile(pattern = "watcher")
  dir.create(loc)

  empty <- dir_state(loc)
  expect_equal(length(empty), 0)

  file.create(file.path(loc, "test-1.txt"))
  one <- dir_state(loc)
  expect_equal(length(one), 1)
  expect_equal(basename(names(one)), "test-1.txt")

  diff <- compare_state(empty, one)
  expect_equal(diff$n, 1)
  expect_equal(basename(diff$added), "test-1.txt")

  write.table(mtcars, file.path(loc, "test-1.txt"))
  diff <- compare_state(one, dir_state(loc))
  expect_equal(diff$n, 1)
  expect_equal(basename(diff$modified), "test-1.txt")

  file.rename(file.path(loc, "test-1.txt"), file.path(loc, "test-2.txt"))
  diff <- compare_state(one, dir_state(loc))
  expect_equal(diff$n, 2)
  expect_equal(basename(diff$deleted), "test-1.txt")
  expect_equal(basename(diff$added), "test-2.txt")

  diff <- compare_state(
    c(file1 = "62da2", file2 = "e14a6", file3 = "6e6dd"),
    c(file1 = "62da2", file2 = "e14a6", file21 = "532fa", file3 = "3f4sa")
  )
  expect_equal(diff$n, 2)
  expect_equal(basename(diff$added), "file21")
  expect_equal(basename(diff$modified), "file3")
})

test_that("watcher works correctly", {
  skip_on_ci()
  skip_on_os("windows")
  skip_on_cran()
  if (Sys.which("bash") == "") {
    skip("bash not available")
  }
  if (system("bash -c 'which touch'", ignore.stdout = TRUE) != 0L) {
    skip("touch (or which) not available")
  }

  loc <- withr::local_tempfile(pattern = "watcher")
  dir.create(loc)

  code_path <- file.path(loc, "R")
  test_path <- file.path(loc, "tests")

  dir.create(code_path)
  dir.create(test_path)

  delayed.bash.cmd <- function(command) {
    system(paste0("bash -c 'sleep 1;", command, "'"), wait = FALSE)
  }

  add.code.file <- function(file.name) {
    delayed.bash.cmd(paste0("touch ", file.path(code_path, file.name)))
  }

  remove.code.file <- function(file.name) {
    delayed.bash.cmd(paste0("rm ", file.path(code_path, file.name)))
  }

  test.added <- function(added, deleted, modified) {
    expect_equal(length(added), 1)
    expect_equal(grepl("test1.R", added), TRUE)
    expect_equal(length(deleted), 0)
    expect_equal(length(modified), 0)
    FALSE
  }

  test.removed <- function(added, deleted, modified) {
    expect_equal(length(added), 0)
    expect_equal(length(deleted), 1)
    expect_equal(grepl("test1.R", deleted), TRUE)
    expect_equal(length(modified), 0)
    FALSE
  }

  add.code.file("test1.R")
  watch(c(code_path, test_path), test.added)

  remove.code.file("test1.R")
  watch(c(code_path, test_path), test.removed)
})

find_parallel <- function(path, load_package = "source", package = NULL) {
  # If env var is set, then use that
  parenv <- Sys.getenv("TESTTHAT_PARALLEL", NA_character_)
  if (!is.na(parenv)) {
    if (toupper(parenv) == "TRUE") {
      return(TRUE)
    }
    if (toupper(parenv) == "FALSE") {
      return(FALSE)
    }
    cli::cli_abort(
      "{.envvar TESTTHAT_PARALLEL} must be {.code TRUE} or {.code FALSE}."
    )
  }

  # Make sure we get the local package package if not "installed"
  if (load_package != "installed") {
    package <- NULL
  }
  desc <- find_description(path, package)
  if (is.null(desc)) {
    return(FALSE)
  }

  par <- identical(
    toupper(desc$get_field("Config/testthat/parallel", default = "FALSE")),
    "TRUE"
  )

  if (par) {
    ed <- as.integer(desc$get_field("Config/testthat/edition", default = 2L))
    if (ed < 3) {
      cli::cli_inform("Running tests in parallel requires the 3rd edition.")
      par <- FALSE
    }
  }

  par
}

methods::setOldClass("proc_time")

#' Capture test results and metadata
#'
#' This reporter gathers all results, adding additional information such as
#' test elapsed time, and test filename if available. Very useful for reporting.
#'
#' @export
#' @family reporters
ListReporter <- R6::R6Class(
  "ListReporter",
  inherit = Reporter,
  public = list(
    running = NULL,
    current_file = "", # so we can still subset with this
    results = NULL,

    initialize = function() {
      super$initialize()
      self$capabilities$parallel_support <- TRUE
      self$capabilities$parallel_updates <- TRUE
      self$results <- Stack$new()
      self$running <- new.env(parent = emptyenv())
    },

    start_test = function(context, test) {
      # is this a new test block?
      if (
        !identical(self$running[[self$current_file]]$context, context) ||
          !identical(self$running[[self$current_file]]$test, test)
      ) {
        self$running[[self$current_file]]$context <- context
        self$running[[self$current_file]]$test <- test
        self$running[[self$current_file]]$expectations <- Stack$new()
        self$running[[self$current_file]]$start_time <- proc.time()
      }
    },

    add_result = function(context, test, result) {
      if (is.null(self$running[[self$current_file]]$expectations)) {
        # we received a result outside of a test:
        # could be a bare expectation or an exception/error
        if (!inherits(result, 'error')) {
          return()
        }
        self$running[[self$current_file]]$expectations <- Stack$new()
      }

      self$running[[self$current_file]]$expectations$push(result)
    },

    end_test = function(context, test) {
      elapsed <- as.double(
        proc.time() - self$running[[self$current_file]]$start_time
      )

      results <- list()
      if (!is.null(self$running[[self$current_file]]$expectations)) {
        results <- self$running[[self$current_file]]$expectations$as_list()
      }

      self$results$push(list(
        file = self$current_file %||% NA_character_,
        context = context,
        test = test,
        user = elapsed[1],
        system = elapsed[2],
        real = elapsed[3],
        results = results
      ))

      self$running[[self$current_file]]$expectations <- NULL
    },

    start_file = function(name) {
      if (!name %in% names(self$running)) {
        newfile <- list(
          start_time = NA,
          expectations = NULL,
          context = NULL,
          test = NULL
        )
        assign(name, newfile, envir = self$running)
      }
      self$current_file <- name
    },

    end_file = function() {
      # fallback in case we have errors but no expectations
      self$end_context(self$current_file)
      rm(list = self$current_file, envir = self$running)
    },

    end_context = function(context) {
      results <- self$running[[self$current_file]]$expectations
      if (is.null(results)) {
        return()
      }

      self$running[[self$current_file]]$expectations <- NULL

      # look for exceptions raised outside of tests
      # they happened just before end_context since they interrupt the test_
      # file execution
      results <- results$as_list()
      if (length(results) == 0) {
        return()
      }

      self$results$push(list(
        file = self$current_file %||% NA_character_,
        context = context,
        test = NA_character_,
        user = NA_real_,
        system = NA_real_,
        real = NA_real_,
        results = results
      ))
    },

    get_results = function() {
      testthat_results(self$results$as_list())
    }
  )
)

#' Create a `testthat_results` object from the test results
#' as stored in the ListReporter results field.
#'
#' @param results a list as stored in ListReporter
#' @return its list argument as a `testthat_results` object
#' @seealso ListReporter
#' @keywords internal
testthat_results <- function(results) {
  stopifnot(is.list(results))
  structure(results, class = "testthat_results")
}

# return if all tests are successful w/o error
all_passed <- function(res) {
  if (length(res) == 0) {
    return(TRUE)
  }

  df <- as.data.frame.testthat_results(res)
  sum(df$failed) == 0 && !any(df$error)
}

any_warnings <- function(res) {
  if (length(res) == 0) {
    return(FALSE)
  }

  df <- as.data.frame.testthat_results(res)
  any(df$warning > 0)
}


#' @export
as.data.frame.testthat_results <- function(x, ...) {
  if (length(x) == 0) {
    return(
      data.frame(
        file = character(0),
        context = character(0),
        test = character(0),
        nb = integer(0),
        failed = integer(0),
        skipped = logical(0),
        error = logical(0),
        warning = integer(0),
        user = numeric(0),
        system = numeric(0),
        real = numeric(0),
        passed = integer(0),
        result = list(),
        stringsAsFactors = FALSE
      )
    )
  }

  rows <- lapply(x, summarize_one_test_results)
  do.call(rbind, rows)
}

summarize_one_test_results <- function(test) {
  test_results <- test$results
  nb_tests <- length(test_results)

  nb_failed <- nb_skipped <- nb_warning <- nb_passed <- 0L
  error <- FALSE

  if (nb_tests > 0) {
    # error reports should be handled differently.
    # They may not correspond to an expect_that() test so remove them
    last_test <- test_results[[nb_tests]]
    error <- expectation_error(last_test)
    if (error) {
      test_results <- test_results[-nb_tests]
      nb_tests <- length(test_results)
    }

    nb_passed <- sum(map_lgl(test_results, expectation_success))
    nb_skipped <- sum(map_lgl(test_results, expectation_skip))
    nb_failed <- sum(map_lgl(test_results, expectation_failure))
    nb_warning <- sum(map_lgl(test_results, expectation_warning))
  }

  context <- if (length(test$context) > 0) test$context else ""

  res <- data.frame(
    file = test$file,
    context = context,
    test = test$test,
    nb = nb_tests,
    failed = nb_failed,
    skipped = as.logical(nb_skipped),
    error = error,
    warning = nb_warning,
    user = test$user,
    system = test$system,
    real = test$real,
    stringsAsFactors = FALSE
  )

  # Added at end for backward compatibility
  res$passed <- nb_passed

  # Cannot easily add list columns in data.frame()
  res$result <- list(test_results)
  res
}


#' @export
print.testthat_results <- function(x, ...) {
  print(as.data.frame(x))
}

test_that("MultiReporter", {
  reports <- lapply(seq_len(3), function(x) ListReporter$new())
  reporter <- MultiReporter$new(reporters = reports)

  with_reporter(reporter, test_one_file("context.R"))

  dfs <- lapply(reports, function(x) as.data.frame(x$get_results()))

  expect_equal(dfs[[2]][1:7], dfs[[1]][1:7])
  expect_equal(dfs[[3]][1:7], dfs[[1]][1:7])
})

#' Manage test reporting
#'
#' The job of a reporter is to aggregate the results from files, tests, and
#' expectations and display them in an informative way. Every testthat function
#' that runs multiple tests provides a `reporter` argument which you can
#' use to override the default (which is selected by [default_reporter()]).
#'
#' You only need to use this `Reporter` object directly if you are creating
#' a new reporter. Currently, creating new Reporters is undocumented,
#' so if you want to create your own, you'll need to make sure that you're
#' familiar with [R6](https://adv-r.hadley.nz/R6.html) and then need read the
#' source code for a few.
#'
#' @keywords internal
#' @export
#' @export Reporter
#' @aliases Reporter
#' @importFrom R6 R6Class
#' @family reporters
#' @examples
#' path <- testthat_example("success")
#'
#' test_file(path)
#' # Override the default by supplying the name of a reporter
#' test_file(path, reporter = "minimal")
Reporter <- R6::R6Class("Reporter",
  public = list(
    capabilities = list(parallel_support = FALSE, parallel_updates = FALSE),
    start_reporter = function() {},
    start_context =  function(context) {},
    start_test =     function(context, test) {},
    start_file =     function(filename) {},
    add_result =     function(context, test, result) {},
    end_test =       function(context, test) {},
    end_context =    function(context) {},
    end_reporter =   function() {},
    end_file =       function() {},
    is_full =        function() FALSE,
    update =         function() {},

    width = 80,
    unicode = TRUE,
    crayon = TRUE,
    rstudio = TRUE,
    hyperlinks = TRUE,

    out = NULL,

    initialize = function(file = getOption("testthat.output_file", stdout())) {
      if (is.character(file)) {
        file <- normalizePath(file, mustWork = FALSE)
      }
      self$out <- file
      if (is.character(self$out) && file.exists(self$out)) {
        # If writing to a file, overwrite it if it exists
        file.remove(self$out)
      }

      # Capture at init so not affected by test settings
      self$width <- cli::console_width()
      self$unicode <- cli::is_utf8_output()
      self$crayon <- cli::num_ansi_colors() > 1
      self$rstudio <- Sys.getenv("RSTUDIO") == "1"
      self$hyperlinks <- cli::ansi_hyperlink_types()[["run"]]
    },

    # To be used when the reporter needs to produce output inside of an active
    # test, which is almost always from $add_result()
    local_user_output = function(.env = parent.frame()) {
      local_reproducible_output(
        width = self$width,
        crayon = self$crayon,
        rstudio = self$rstudio,
        hyperlinks = self$hyperlinks,
        .env = .env
      )
      # Can't set unicode with local_reproducible_output() because it can
      # generate a skip if you're temporarily using a non-UTF-8 locale
      withr::local_options(cli.unicode = self$unicode, .local_envir = .env)
    },

    cat_tight = function(...) {
      cat(..., sep = "", file = self$out, append = TRUE)
    },

    cat_line = function(...) {
      cli::cat_line(..., file = self$out)
    },

    rule = function(...) {
      cli::cat_rule(..., file = self$out)
    },

    # The hierarchy of contexts are implied - a context starts with a
    # call to context(), and ends either with the end of the file, or
    # with the next call to context() in the same file. These private
    # methods paper over the details so that context appear to work
    # in the same way as tests and expectations.
    .context = NULL,
    .start_context = function(context) {
      if (!is.null(self$.context)) {
        self$end_context(self$.context)
      }
      self$.context <- context
      self$start_context(context)

      invisible()
    },
    end_context_if_started = function(context) {
      if (!is.null(self$.context)) {
        self$end_context(self$.context)
        self$.context <- NULL
      }
      invisible()
    }
  )
)

#' Retrieve the default reporter
#'
#' The defaults are:
#' * [ProgressReporter] for interactive, non-parallel; override with
#'   `testthat.default_reporter`
#' * [ParallelProgressReporter] for interactive, parallel packages;
#'   override with `testthat.default_parallel_reporter`
#' * [CompactProgressReporter] for single-file interactive; override with
#'   `testthat.default_compact_reporter`
#' * [CheckReporter] for R CMD check; override with `testthat.default_check_reporter`
#'
#' @export
#' @keywords internal
default_reporter <- function() {
  getOption("testthat.default_reporter", "Progress")
}

#' @export
#' @rdname default_reporter
default_parallel_reporter <- function() {
  getOption("testthat.default_parallel_reporter", "ParallelProgress")
}

#' @export
#' @rdname default_reporter
default_compact_reporter <- function() {
  getOption("testthat.default_compact_reporter", "CompactProgress")
}

#' @export
#' @rdname default_reporter
check_reporter <- function() {
  getOption("testthat.default_check_reporter", "Check")
}

# /-----------------------------\     /-------------------------------\
# | Main R process              |     | Subprocess 1                  |
# | +------------------------+  |     | +---------------------------+ |
# | | test_dir_parallel()    |  |     | | test_file()               | |
# | | +-------------------+  |  |     | | +-----------------------+ | |
# | | | Event loop        |< ------+  | | | SubprocessReporter    | | |
# | | +-------------------+  |  |  |  | | | +-------------------+ | | |
# | |    |                   |  |  |  | | | | test_that()       | | | |
# | |    v                   |  |  |  | | | +-------------------+ | | |
# | | +-------------------+  |  |  |  | | |    |                  | | |
# | | | Progress2Reporter |  |  |  |  | | |    v                  | | |
# | | +-------------------+  |  |  |  | | | +-------------------+ | | |
# | +------------------------+  |  |--------| signalCondition() | | | |
# \-----------------------------/  |  | | | +-------------------+ | | |
#                                  |  | | +-----------------------+ | |
#                                  |  | +---------------------------+ |
#                                  |  \-------------------------------/
#                                  |  /-------------------------------\
#                                  |--| Subprocess 2                  |
#                                  |  \-------------------------------/
#                                  |  /-------------------------------\
#                                  \--| Subprocess 3                  |
#                                     \-------------------------------/
#
# ## Notes
#
# * Subprocesses run `callr::r_session` R sessions. They are re-used,
#   one R session can be used for several test_file() calls.
# * Helper and setup files are loaded in the subprocesses after this.
# * The main process puts all test files in the task queue, and then
#   runs an event loop.

test_files_parallel <- function(
  test_dir,
  test_package,
  test_paths,
  load_helpers = TRUE,
  reporter = default_reporter(TRUE),
  env = NULL,
  stop_on_failure = FALSE,
  stop_on_warning = FALSE,
  wrap = TRUE, # unused, to match test_files signature
  load_package = c("none", "installed", "source"),
  shuffle = FALSE
) {
  # TODO: support timeouts. 20-30s for each file by default?

  num_workers <- min(default_num_cpus(), length(test_paths))
  cli::cli_inform("Starting {num_workers} test process{?es}.")

  # Set up work queue ------------------------------------------
  queue <- NULL
  withr::defer(queue_teardown(queue))

  # Start workers in parallel and add test tasks to queue.
  queue <- queue_setup(
    test_paths = test_paths,
    test_package = test_package,
    test_dir = test_dir,
    load_helpers = load_helpers,
    num_workers = num_workers,
    load_package = load_package,
    shuffle = shuffle
  )

  withr::with_dir(test_dir, {
    reporters <- test_files_reporter(reporter, "parallel")
    with_reporter(reporters$multi, {
      parallel_updates <- reporter$capabilities$parallel_updates
      if (parallel_updates) {
        parallel_event_loop_smooth(queue, reporters, ".")
      } else {
        parallel_event_loop_chunky(queue, reporters, ".")
      }
    })

    test_files_check(
      reporters$list$get_results(),
      stop_on_failure = stop_on_failure,
      stop_on_warning = stop_on_warning
    )
  })
}


default_num_cpus <- function() {
  # Use common option, if set
  ncpus <- getOption("Ncpus", NULL)
  if (!is.null(ncpus)) {
    ncpus <- suppressWarnings(as.integer(ncpus))
    if (is.na(ncpus)) {
      cli::cli_abort(
        "{.code getOption('Ncpus')} must be an integer.",
        call = NULL
      )
    }
    return(ncpus)
  }

  # Otherwise use env var if set
  ncpus <- Sys.getenv("TESTTHAT_CPUS", "")
  if (ncpus != "") {
    ncpus <- suppressWarnings(as.integer(ncpus))
    if (is.na(ncpus)) {
      cli::cli_abort("{.envvar TESTTHAT_CPUS} must be an integer.")
    }
    return(ncpus)
  }

  # Otherwise 2
  2L
}

parallel_event_loop_smooth <- function(queue, reporters, test_dir) {
  update_interval <- 0.1
  next_update <- proc.time()[[3]] + update_interval

  while (!queue$is_idle()) {
    # How much time do we have to poll before the next UI update?
    now <- proc.time()[[3]]
    poll_time <- max(next_update - now, 0)
    next_update <- now + update_interval

    msgs <- queue$poll(poll_time)

    updated <- FALSE
    for (x in msgs) {
      if (x$code == PROCESS_OUTPUT) {
        lns <- paste0("> ", x$path, ": ", x$message)
        cat("\n", file = stdout())
        base::writeLines(lns, stdout())
        next
      }
      if (x$code != PROCESS_MSG) {
        next
      }

      m <- x$message
      if (!inherits(m, "testthat_message")) {
        cli::cli_inform(as.character(m))
        next
      }

      if (m$cmd != "DONE") {
        reporters$multi$start_file(m$filename)
        reporters$multi$start_test(m$context, m$test)
        if (m$type == "snapshotter") {
          snapshotter <- getOption("testthat.snapshotter")
          do.call(snapshotter[[m$cmd]], m$args)
        } else {
          do.call(reporters$multi[[m$cmd]], m$args)
          updated <- TRUE
        }
      }
    }

    # We need to spin, even if there were no events
    if (!updated) {
      reporters$multi$update()
    }
  }
}

parallel_event_loop_chunky <- function(queue, reporters, test_dir) {
  files <- list()
  while (!queue$is_idle()) {
    msgs <- queue$poll(Inf)
    for (x in msgs) {
      if (x$code == PROCESS_OUTPUT) {
        lns <- paste0("> ", x$path, ": ", x$message)
        base::writeLines(lns, stdout())
        next
      }
      if (x$code != PROCESS_MSG) {
        next
      }

      m <- x$message
      if (!inherits(m, "testthat_message")) {
        cli::cli_inform(as.character(m))
        next
      }

      # Record all events until we get end of file, then we replay them all
      # with the local reporters. This prevents out of order reporting.
      if (m$cmd != "DONE") {
        files[[m$filename]] <- append(files[[m$filename]], list(m))
      } else {
        replay_events(reporters$multi, files[[m$filename]])
        reporters$multi$end_context_if_started()
        files[[m$filename]] <- NULL
      }
    }
  }
}

replay_events <- function(reporter, events) {
  snapshotter <- getOption("testthat.snapshotter")
  for (m in events) {
    if (m$type == "snapshotter") {
      do.call(snapshotter[[m$cmd]], m$args)
    } else {
      do.call(reporter[[m$cmd]], m$args)
    }
  }
}

queue_setup <- function(
  test_paths,
  test_package,
  test_dir,
  num_workers,
  load_helpers,
  load_package,
  shuffle = FALSE
) {
  # TODO: observe `load_package`, but the "none" default is not
  # OK for the subprocess, because it'll not have the tested package
  if (load_package == "none") {
    load_package <- "source"
  }

  # TODO: similarly, load_helpers = FALSE, coming from devtools,
  # is not appropriate in the subprocess
  load_helpers <- TRUE

  test_package <- test_package %||% Sys.getenv("TESTTHAT_PKG")

  hyperlinks <- cli::ansi_has_hyperlink_support()

  # First we load the package "manually", in case it is testthat itself
  load_hook <- expr({
    switch(
      !!load_package,
      installed = library(!!test_package, character.only = TRUE),
      source = pkgload::load_all(!!test_dir, helpers = FALSE, quiet = TRUE)
    )

    # Ensure snapshot can generate hyperlinks for snapshot_accept()
    options(
      cli.hyperlink = !!hyperlinks,
      cli.hyperlink_run = !!hyperlinks
    )

    asNamespace("testthat")$queue_process_setup(
      test_package = !!test_package,
      test_dir = !!test_dir,
      load_helpers = !!load_helpers,
      load_package = "none"
    )
  })
  queue <- task_q$new(concurrency = num_workers, load_hook = load_hook)

  fun <- transport_fun(function(path, shuffle) {
    asNamespace("testthat")$queue_task(path, shuffle)
  })
  for (path in test_paths) {
    queue$push(fun, list(path, shuffle))
  }

  queue
}

queue_process_setup <- function(
  test_package,
  test_dir,
  load_helpers,
  load_package
) {
  env <- asNamespace("testthat")$test_files_setup_env(
    test_package,
    test_dir,
    load_package
  )

  asNamespace("testthat")$test_files_setup_state(
    test_dir = test_dir,
    test_package = test_package,
    load_helpers = load_helpers,
    env = env,
    frame = .GlobalEnv
  )

  # record testing env for mocks & queue_task
  # manual implementation of local_testing_env()
  the$testing_env <- env
}

queue_task <- function(path, shuffle = FALSE) {
  withr::local_envvar("TESTTHAT_IS_PARALLEL" = "true")
  snapshotter <- SubprocessSnapshotReporter$new(snap_dir = "_snaps")
  withr::local_options(testthat.snapshotter = snapshotter)
  reporters <- list(
    SubprocessReporter$new(),
    snapshotter
  )
  multi <- MultiReporter$new(reporters = reporters)
  with_reporter(
    multi,
    test_one_file(path, env = the$testing_env, shuffle = shuffle)
  )
  NULL
}

# Clean up subprocesses: we call teardown methods, but we only give them a
# second, before killing the whole process tree using ps's env var marker
# method.
queue_teardown <- function(queue) {
  if (is.null(queue)) {
    return()
  }

  tasks <- queue$list_tasks()
  num <- nrow(tasks)

  # calling quit() here creates a race condition, and the output of
  # the deferred_run() might be lost. Instead we close the input
  # connection in a separate task.
  clean_fn <- function() {
    withr::deferred_run(.GlobalEnv)
  }

  topoll <- integer()
  for (i in seq_len(num)) {
    if (
      !is.null(tasks$worker[[i]]) && tasks$worker[[i]]$get_state() == "idle"
    ) {
      # The worker might have crashed or exited, so this might fail.
      # If it does then we'll just ignore that worker
      tryCatch(
        {
          tasks$worker[[i]]$call(clean_fn)
          topoll <- c(topoll, i)
        },
        error = function(e) NULL
      )
    }
  }

  # Give covr a bit more time
  if (in_covr()) {
    grace <- 30L
  } else {
    grace <- 1L
  }
  first_error <- NULL
  limit <- Sys.time() + grace
  while (length(topoll) > 0 && (timeout <- limit - Sys.time()) > 0) {
    timeout <- as.double(timeout, units = "secs") * 1000
    conns <- lapply(tasks$worker[topoll], function(x) x$get_poll_connection())
    pr <- unlist(processx::poll(conns, as.integer(timeout)))
    for (i in which(pr == "ready")) {
      msg <- tasks$worker[[topoll[i]]]$read()
      first_error <- first_error %||% msg$error
    }
    topoll <- topoll[pr != "ready"]
  }

  topoll <- integer()
  for (i in seq_len(num)) {
    if (
      !is.null(tasks$worker[[i]]) && tasks$worker[[i]]$get_state() == "idle"
    ) {
      tryCatch(
        {
          close(tasks$worker[[i]]$get_input_connection())
          topoll <- c(topoll, i)
        },
        error = function(e) NULL
      )
    }
  }

  limit <- Sys.time() + grace
  while (length(topoll) > 0 && (timeout <- limit - Sys.time()) > 0) {
    timeout <- as.double(timeout, units = "secs") * 1000
    conns <- lapply(tasks$worker[topoll], function(x) x$get_poll_connection())
    pr <- unlist(processx::poll(conns, as.integer(timeout)))
    topoll <- topoll[pr != "ready"]
  }

  for (i in seq_len(num)) {
    if (!is.null(tasks$worker[[i]])) {
      if (ps::ps_is_supported()) {
        tryCatch(tasks$worker[[i]]$kill_tree(), error = function(e) NULL)
      } else {
        tryCatch(tasks$worker[[i]]$kill(), error = function(e) NULL)
      }
    }
  }

  if (!is.null(first_error)) {
    cli::cli_abort(
      "At least one parallel worker failed to run teardown",
      parent = first_error
    )
  }
}

# Reporter that just forwards events in the subprocess back to the main process
#
# Ideally, these messages would be throttled, i.e. if the test code
# emits many expectation conditions fast, SubprocessReporter should
# collect several of them and only emit a condition a couple of times
# a second. End-of-test and end-of-file events would be transmitted
# immediately.
SubprocessReporter <- R6::R6Class(
  "SubprocessReporter",
  inherit = Reporter,
  public = list(
    start_file = function(filename) {
      private$filename <- filename
      private$event("start_file", filename)
    },
    start_test = function(context, test) {
      private$context <- context
      private$test <- test
      private$event("start_test", context, test)
    },
    start_context = function(context) {
      private$context <- context
      private$event("start_context", context)
    },
    add_result = function(context, test, result) {
      if (inherits(result, "expectation_success")) {
        # Strip bulky components to reduce data transfer cost
        result[["srcref"]] <- NULL
        result[["trace"]] <- NULL
      }
      private$event("add_result", context, test, result)
    },
    end_test = function(context, test) {
      private$event("end_test", context, test)
    },
    end_context = function(context) {
      private$event("end_context", context)
    },
    end_file = function() {
      private$event("end_file")
    },
    end_reporter = function() {
      private$event("DONE")
    }
  ),

  private = list(
    filename = NULL,
    context = NULL,
    test = NULL,
    event = function(cmd, ...) {
      msg <- list(
        code = PROCESS_MSG,
        type = "reporter",
        cmd = cmd,
        filename = private$filename,
        context = private$context,
        test = private$test,
        time = proc.time()[[3]],
        args = list(...)
      )
      class(msg) <- c("testthat_message", "callr_message", "condition")
      signalCondition(msg)
    }
  )
)

test_that("reporter doesn't change without warning", {
  skip_if_not_installed("xml2")
  expect_snapshot_reporter(JunitReporterMock$new())
})

test_that("permit Java-style class names", {
  class <- "package_name_or_domain.ClassName"
  expect_equal(classnameOK(class), class)
})

test_that("ANSI escapes are stripped from all user text in XML", {
  skip_if_not_installed("xml2")

  tmp <- withr::local_tempfile(fileext = ".xml")
  reporter <- JunitReporterMock$new(file = tmp)
  reporter$start_reporter()

  text_with_ansi <- "\033[33mFirst line\033[0m\nSecond line"
  reporter$start_context("c")
  reporter$start_test("c", "t")
  reporter$add_result("c", "t", new_expectation("error", text_with_ansi))
  reporter$add_result("c", "t", new_expectation("failure", text_with_ansi))
  reporter$add_result("c", "t", new_expectation("skip", text_with_ansi))
  reporter$end_test()
  reporter$end_context()
  reporter$end_reporter()

  expect_no_error(xml2::read_xml(tmp))
})

test_that("warnings outside context don't cause xml_add_child errors", {
  skip_if_not_installed("xml2")

  tmp <- withr::local_tempfile(fileext = ".xml")
  reporter <- JunitReporterMock$new(file = tmp)
  reporter$start_reporter()

  # This would previously fail with "no applicable method for 'xml_add_child'"
  expect_no_error({
    reporter$add_result(NULL, "test", new_expectation("warning", "test"))
  })
  reporter$end_reporter()
})

#' Run all tests in a directory
#'
#' @description
#' This function is the low-level workhorse that powers [test_local()] and
#' [test_package()]. Generally, you should not call this function directly.
#' In particular, you are responsible for ensuring that the functions to test
#' are available in the test `env` (e.g. via `load_package`).
#'
#' See `vignette("special-files")` to learn more about the conventions for test,
#' helper, and setup files that testthat uses, and what you might use each for.
#'
#' @section Environments:
#' Each test is run in a clean environment to keep tests as isolated as
#' possible. For package tests, that environment inherits from the package's
#' namespace environment, so that tests can access internal functions
#' and objects.
#'
#' @param path Path to directory containing tests.
#' @param package If these tests belong to a package, the name of the package.
#' @param filter If not `NULL`, only tests with file names matching this
#'   regular expression will be executed. Matching is performed on the file
#'   name after it's stripped of `"test-"` and `".R"`.
#' @param env Environment in which to execute the tests. Expert use only.
#' @param ... Additional arguments passed to [grepl()] to control filtering.
#' @param load_helpers Source helper files before running the tests?
#' @param stop_on_failure If `TRUE`, throw an error if any tests fail.
#' @param stop_on_warning If `TRUE`, throw an error if any tests generate
#'   warnings.
#' @param load_package Strategy to use for load package code:
#'   * "none", the default, doesn't load the package.
#'   * "installed", uses [library()] to load an installed package.
#'   * "source", uses [pkgload::load_all()] to a source package.
#'     To configure the arguments passed to `load_all()`, add this
#'     field in your DESCRIPTION file:
#'
#'     ```
#'     Config/testthat/load-all: list(export_all = FALSE, helpers = FALSE)
#'     ```
#' @param wrap DEPRECATED
#' @return A list (invisibly) containing data about the test results.
#' @inheritParams with_reporter
#' @inheritParams source_file
#' @export
test_dir <- function(path,
                     filter = NULL,
                     reporter = NULL,
                     env = NULL,
                     ...,
                     load_helpers = TRUE,
                     stop_on_failure = TRUE,
                     stop_on_warning = FALSE,
                     wrap = lifecycle::deprecated(),
                     package = NULL,
                     load_package = c("none", "installed", "source")
                     ) {

  load_package <- arg_match(load_package)

  start_first <- find_test_start_first(path, load_package, package)
  test_paths <- find_test_scripts(
    path,
    filter = filter,
    ...,
    full.names = FALSE,
    start_first = start_first
  )
  if (length(test_paths) == 0) {
    abort("No test files found")
  }

  if (!is_missing(wrap)) {
    lifecycle::deprecate_stop("3.0.0", "test_dir(wrap = )")
  }

  want_parallel <- find_parallel(path, load_package, package)

  if (is.null(reporter)) {
    if (want_parallel) {
      reporter <- default_parallel_reporter()
    } else {
      reporter <- default_reporter()
    }
  }
  reporter <- find_reporter(reporter)
  parallel <- want_parallel && reporter$capabilities$parallel_support

  test_files(
    test_dir = path,
    test_paths = test_paths,
    test_package = package,
    reporter = reporter,
    load_helpers = load_helpers,
    env = env,
    stop_on_failure = stop_on_failure,
    stop_on_warning = stop_on_warning,
    load_package = load_package,
    parallel = parallel
  )
}

#' Run tests in a single file
#'
#' Helper, setup, and teardown files located in the same directory as the
#' test will also be run. See `vignette("special-files")` for details.
#'
#' @inherit test_dir return params
#' @inheritSection test_dir Environments
#' @param path Path to file.
#' @param ... Additional parameters passed on to `test_dir()`
#' @param desc Optionally, supply a string here to run only a single
#'   test (`test_that()` or `describe()`) with this `desc`ription.
#' @export
#' @examples
#' path <- testthat_example("success")
#' test_file(path)
#' test_file(path, desc = "some tests have warnings")
#' test_file(path, reporter = "minimal")
test_file <- function(path,
                      reporter = default_compact_reporter(),
                      desc = NULL,
                      package = NULL,
                      ...) {
  if (!file.exists(path)) {
    stop("`path` does not exist", call. = FALSE)
  }

  test_files(
    test_dir = dirname(path),
    test_package = package,
    test_paths = basename(path),
    reporter = reporter,
    desc = desc,
    ...
  )
}

test_files <- function(test_dir,
                       test_package,
                       test_paths,
                       load_helpers = TRUE,
                       reporter = default_reporter(),
                       env = NULL,
                       stop_on_failure = FALSE,
                       stop_on_warning = FALSE,
                       desc = NULL,
                       wrap = TRUE,
                       load_package = c("none", "installed", "source"),
                       parallel = FALSE,
                       error_call = caller_env()) {

  if (!isTRUE(wrap)) {
    lifecycle::deprecate_stop("3.0.0", "test_dir(wrap = )")
  }

  # Must keep these two blocks in sync
  if (parallel) {
    test_files_parallel(
      test_dir = test_dir,
      test_package = test_package,
      test_paths = test_paths,
      load_helpers = load_helpers,
      reporter = reporter,
      env = env,
      stop_on_failure = stop_on_failure,
      stop_on_warning = stop_on_warning,
      load_package = load_package
    )
  } else {
    test_files_serial(
      test_dir = test_dir,
      test_package = test_package,
      test_paths = test_paths,
      load_helpers = load_helpers,
      reporter = reporter,
      env = env,
      stop_on_failure = stop_on_failure,
      stop_on_warning = stop_on_warning,
      desc = desc,
      load_package = load_package,
      error_call = error_call
    )
  }

}

test_files_serial <- function(test_dir,
                       test_package,
                       test_paths,
                       load_helpers = TRUE,
                       reporter = default_reporter(),
                       env = NULL,
                       stop_on_failure = FALSE,
                       stop_on_warning = FALSE,
                       desc = NULL,
                       wrap = TRUE,
                       load_package = c("none", "installed", "source"),
                       error_call = caller_env()) {

  # Because load_all() called by test_files_setup_env() will have already
  # loaded them. We don't want to rely on testthat's loading since that
  # only affects the test environment and we want to keep the helpers
  # loaded in the user's session.
  load_package <- arg_match(load_package)
  if (load_package == "source") {
    load_helpers <- FALSE
  }

  env <- test_files_setup_env(test_package, test_dir, load_package, env)
  # record testing env for mocks
  local_testing_env(env)

  test_files_setup_state(test_dir, test_package, load_helpers, env)
  reporters <- test_files_reporter(reporter)

  with_reporter(reporters$multi,
    lapply(
      test_paths,
      test_one_file,
      env = env,
      desc = desc,
      error_call = error_call
    )
  )

  test_files_check(reporters$list$get_results(),
    stop_on_failure = stop_on_failure,
    stop_on_warning = stop_on_warning
  )
}

test_files_setup_env <- function(test_package,
                                 test_dir,
                                 load_package = c("none", "installed", "source"),
                                 env = NULL) {
  library(testthat)

  load_package <- arg_match(load_package)
  if (load_package == "installed") {
    library(test_package, character.only = TRUE)
  } else if (load_package == "source") {
    # Allow configuring what we export to the search path (#1636)
    args <- find_load_all_args(test_dir)
    pkgload::load_all(
      test_dir,
      export_all = args[["export_all"]],
      helpers = args[["helpers"]],
      quiet = TRUE
    )
  }

  env %||% test_env(test_package)
}

find_load_all_args <- function(path) {
  default <- list(export_all = TRUE, helpers = TRUE)

  desc <- find_description(path)
  if (is.null(desc)) {
    return(default)
  }

  args <- desc$get_field("Config/testthat/load-all", default = NULL)
  if (is.null(args)) {
    return(default)
  }

  args <- parse_expr(args)
  if (!is_call(args, "list")) {
    abort("`Config/testthat/load-all` must be a list.", call = NULL)
  }

  args <- as.list(args[-1])
  list(
    export_all = args[["export_all"]] %||% default[["export_all"]],
    helpers = args[["helpers"]] %||% default[["helpers"]]
  )
}

test_files_setup_state <- function(
    test_dir,
    test_package,
    load_helpers,
    env,
    frame = parent.frame()
) {
  # Define testing environment
  local_test_directory(test_dir, test_package, .env = frame)
  withr::local_options(
    topLevelEnvironment = env_parent(env),
    .local_envir = frame
  )

  # Load helpers, setup, and teardown (on exit)
  local_teardown_env(frame)
  if (load_helpers) {
    source_test_helpers(".", env)
  }
  source_test_setup(".", env)
  withr::defer(source_test_teardown(".", env), frame)      # old school
}

test_files_reporter <- function(reporter, .env = parent.frame()) {
  lister <- ListReporter$new()
  reporters <- list(
    find_reporter(reporter),
    lister, # track data
    local_snapshotter("_snaps", fail_on_new = FALSE, .env = .env)
  )
  list(
    multi = MultiReporter$new(reporters = compact(reporters)),
    list = lister
  )
}

test_files_check <- function(results, stop_on_failure = TRUE, stop_on_warning = FALSE) {
  if (stop_on_failure && !all_passed(results)) {
    stop("Test failures", call. = FALSE)
  }
  if (stop_on_warning && any_warnings(results)) {
    stop("Tests generated warnings", call. = FALSE)
  }

  invisible(results)
}

test_one_file <- function(path,
                          env = test_env(),
                          desc = NULL,
                          error_call = caller_env()) {
  reporter <- get_reporter()
  on.exit(teardown_run(), add = TRUE)

  reporter$start_file(path)
  source_file(
    path,
    env = env(env),
    desc = desc,
    error_call = error_call
  )
  reporter$end_context_if_started()
  reporter$end_file()
}

# Helpers -----------------------------------------------------------------

#' Run code after all test files
#'
#' This environment has no purpose other than as a handle for [withr::defer()]:
#' use it when you want to run code after all tests have been run.
#' Typically, you'll use `withr::defer(cleanup(), teardown_env())`
#' immediately after you've made a mess in a `setup-*.R` file.
#'
#' @export
teardown_env <- function() {
  if (is.null(the$teardown_env)) {
    abort("`teardown_env()` has not been initialized", .internal = TRUE)
  }

  the$teardown_env
}

local_teardown_env <- function(frame = parent.frame()) {
  local_bindings(teardown_env = frame, .env = the, .frame = frame)
}

#' Find test files
#'
#' @param path path to tests
#' @param invert If `TRUE` return files which **don't** match.
#' @param ... Additional arguments passed to [grepl()] to control filtering.
#' @param start_first A character vector of file patterns (globs, see
#'   [utils::glob2rx()]). The patterns are for the file names (base names),
#'   not for the whole paths. testthat starts the files matching the
#'   first pattern first,  then the ones matching the second, etc. and then
#'   the rest of the files, alphabetically. Parallel tests tend to finish
#'   quicker if you start the slowest files first. `NULL` means alphabetical
#'   order.
#' @inheritParams test_dir
#' @return A character vector of paths
#' @keywords internal
#' @export
find_test_scripts <- function(path, filter = NULL, invert = FALSE, ..., full.names = TRUE, start_first = NULL) {
  files <- dir(path, "^test.*\\.[rR]$", full.names = full.names)
  files <- filter_test_scripts(files, filter, invert, ...)
  order_test_scripts(files, start_first)
}

filter_test_scripts <- function(files, filter = NULL, invert = FALSE, ...) {
  if (is.null(filter)) {
    return(files)
  }

  which_files <- grepl(filter, context_name(files), ...)
  if (isTRUE(invert)) {
    which_files <- !which_files
  }
  files[which_files]
}

find_test_start_first <- function(path, load_package, package) {
  # Make sure we get the local package package if not "installed"
  if (load_package != "installed") package <- NULL
  desc <- find_description(path, package)
  if (is.null(desc)) {
    return(NULL)
  }

  conf <- desc$get_field("Config/testthat/start-first", NULL)
  if (is.null(conf)) {
    return(NULL)
  }

  trimws(strsplit(conf, ",")[[1]])
}

order_test_scripts <- function(paths, start_first) {
  if (is.null(start_first)) return(paths)
  filemap <- data.frame(
    stringsAsFactors = FALSE,
    base = sub("\\.[rR]$", "", sub("^test[-_\\.]?", "", basename(paths))),
    orig = paths
  )
  rxs <- utils::glob2rx(start_first)
  mch <- lapply(rxs, function(rx) filemap$orig[grep(rx, filemap$base)])
  unique(c(unlist(mch), paths))
}

#' Test package examples
#'
#' These helper functions make it easier to test the examples in a package.
#' Each example counts as one test, and it succeeds if the code runs without
#' an error. Generally, this is redundant with R CMD check, and is not
#' recommended in routine practice.
#'
#' @keywords internal
#' @param path For `test_examples()`, path to directory containing Rd files.
#'   For `test_example()`, path to a single Rd file. Remember the working
#'   directory for tests is `tests/testthat`.
#' @param title Test title to use
#' @param rd A parsed Rd object, obtained from [tools::Rd_db()] or otherwise.
#' @export
test_examples <- function(path = "../..") {
  res <- test_examples_source(path) %||% test_examples_installed()
  if (is.null(res)) {
    cli::cli_abort("Could not find examples.")
  }
  invisible(res)
}

test_examples_source <- function(path = "../..") {
  if (!dir.exists(file.path(path, "man"))) {
    return()
  }
  Rd <- tools::Rd_db(dir = path)
  if (length(Rd) == 0) {
    return()
  }
  lapply(Rd, test_rd)
}

test_examples_installed <- function(package = testing_package()) {
  if (identical(package, "") || is.null(package)) {
    return()
  }

  Rd <- tools::Rd_db(package = package)
  if (length(Rd) == 0) {
    return()
  }
  lapply(Rd, test_rd)
}


#' @export
#' @rdname test_examples
test_rd <- function(rd, title = attr(rd, "Rdfile")) {
  test_example(rd, title %||% "example")
}

#' @export
#' @rdname test_examples
test_example <- function(path, title = path) {
  local_description_push(title)

  ex_path <- withr::local_tempfile(pattern = "test_example-", fileext = ".R")
  tools::Rd2ex(path, ex_path)
  if (!file.exists(ex_path)) {
    return(invisible(FALSE))
  }

  ok <- test_code(
    code = parse(ex_path, encoding = "UTF-8"),
    env = globalenv(),
    reporter = get_reporter() %||% StopReporter$new(),
    skip_on_empty = FALSE
  )
  if (ok) {
    succeed(path)
  }

  invisible(ok)
}

test_that("can test documentation from path or Rd object", {
  rd_path <- test_path("../../man/expect_length.Rd")
  skip_if_not(file.exists(rd_path))

  capture_output({
    test_example(rd_path)
    test_rd(tools::parse_Rd(rd_path))
  })
})
test_that("returns false if no examples", {
  rd_path <- test_path("../../man/test_examples.Rd")
  skip_if_not(file.exists(rd_path))

  expect_false(test_example(rd_path))
})

#' Test reporter: TAP format.
#'
#' This reporter will output results in the Test Anything Protocol (TAP),
#' a simple text-based interface between testing modules in a test harness.
#' For more information about TAP, see http://testanything.org
#'
#' @export
#' @family reporters
TapReporter <- R6::R6Class("TapReporter",
  inherit = Reporter,
  public = list(
    results = list(),
    n = 0L,
    has_tests = FALSE,
    contexts = NA_character_,

    initialize = function(...) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
    },

    start_context = function(context) {
      self$contexts[self$n + 1] <- context
    },

    add_result = function(context, test, result) {
      self$has_tests <- TRUE
      self$n <- self$n + 1L
      self$results[[self$n]] <- result
    },

    end_reporter = function() {
      if (!self$has_tests) {
        return()
      }

      self$cat_line("1..", self$n)
      for (i in 1:self$n) {
        if (!is.na(self$contexts[i])) {
          self$cat_line("# Context ", self$contexts[i])
        }
        result <- self$results[[i]]
        if (expectation_success(result)) {
          self$cat_line("ok ", i, " ", result$test)
        } else if (expectation_broken(result)) {
          self$cat_line("not ok ", i, " ", result$test)
          msg <- gsub("(^|\n)", "\\1  ", format(result))
          self$cat_line(msg)
        } else {
          self$cat_line(
            "ok ", i, " # ", toupper(expectation_type(result)), " ",
            format(result)
          )
        }
      }
    }
  )
)

#' Do you expect an object with this length or shape?
#'
#' `expect_length()` inspects the [length()] of an object; `expect_shape()`
#' inspects the "shape" (i.e. [nrow()], [ncol()], or [dim()]) of
#' higher-dimensional objects like data.frames, matrices, and arrays.
#'
#' @seealso [expect_vector()] to make assertions about the "size" of a vector.
#' @inheritParams expect_that
#' @param n Expected length.
#' @family expectations
#' @export
#' @examples
#' expect_length(1, 1)
#' expect_length(1:10, 10)
#' show_failure(expect_length(1:10, 1))
#'
#' x <- matrix(1:9, nrow = 3)
#' expect_shape(x, nrow = 3)
#' show_failure(expect_shape(x, nrow = 4))
#' expect_shape(x, ncol = 3)
#' show_failure(expect_shape(x, ncol = 4))
#' expect_shape(x, dim = c(3, 3))
#' show_failure(expect_shape(x, dim = c(3, 4, 5)))
expect_length <- function(object, n) {
  check_number_whole(n, min = 0)

  act <- quasi_label(enquo(object))
  act$n <- length(act$val)

  if (act$n != n) {
    fail(c(
      sprintf("Expected %s to have length %i.", act$lab, n),
      sprintf("Actual length: %i.", act$n)
    ))
  } else {
    pass()
  }
  invisible(act$val)
}

#' @param nrow,ncol Expected [nrow()]/[ncol()] of `object`.
#' @param dim Expected [dim()] of `object`.
#' @rdname expect_length
#' @param ... Not used; used to force naming of other arguments.
#' @export
expect_shape = function(object, ..., nrow, ncol, dim) {
  check_dots_empty()
  check_exclusive(nrow, ncol, dim)
  act <- quasi_label(enquo(object))

  dim_object <- base::dim(object)
  if (is.null(dim_object)) {
    fail(sprintf("Expected %s to have dimensions.", act$lab))
  } else if (!missing(nrow)) {
    check_number_whole(nrow, allow_na = TRUE)
    act$nrow <- dim_object[1L]

    if (!identical(as.integer(act$nrow), as.integer(nrow))) {
      fail(c(
        sprintf("Expected %s to have %i rows.", act$lab, nrow),
        sprintf("Actual rows: %i.", act$nrow)
      ))
    } else {
      pass()
    }
  } else if (!missing(ncol)) {
    check_number_whole(ncol, allow_na = TRUE)

    if (length(dim_object) == 1L) {
      fail(sprintf("Expected %s to have two or more dimensions.", act$lab))
    } else {
      act$ncol <- dim_object[2L]

      if (!identical(as.integer(act$ncol), as.integer(ncol))) {
        fail(c(
          sprintf("Expected %s to have %i columns.", act$lab, ncol),
          sprintf("Actual columns: %i.", act$ncol)
        ))
      } else {
        pass()
      }
    }
  } else {
    # !missing(dim)
    if (!is.numeric(dim) && !is.integer(dim)) {
      stop_input_type(dim, "a numeric vector")
    }
    act$dim <- dim_object

    if (length(act$dim) != length(dim)) {
      fail(c(
        sprintf("Expected %s to have %i dimensions.", act$lab, length(dim)),
        sprintf("Actual dimensions: %i.", length(act$dim))
      ))
    } else if (!identical(as.integer(act$dim), as.integer(dim))) {
      fail(c(
        sprintf("Expected %s to have dim (%s).", act$lab, toString(dim)),
        sprintf("Actual dim: (%s).", toString(act$dim))
      ))
    } else {
      pass()
    }
  }

  invisible(act$val)
}

test_that("length computed correctly", {
  expect_success(expect_length(1, 1))
  expect_success(expect_length(1:10, 10))
  expect_success(expect_length(letters[1:5], 5))
})

test_that("generates actionable failure message", {
  x <- 1:10
  expect_snapshot_failure(expect_length(x, 2))
})

test_that("uses S4 length method", {
  A <- setClass("ExpectLengthA", slots = c(x = "numeric", y = "numeric"))
  setMethod("length", "ExpectLengthA", function(x) 5L)
  expect_success(expect_length(A(x = 1:9, y = 3), 5))
})

test_that("returns input", {
  x <- list(1:10, letters)
  out <- expect_length(x, 2)
  expect_identical(out, x)
})

test_that("expect_length validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_length(1:5, "a")
  })
})


test_that("dim compared correctly", {
  expect_success(expect_shape(matrix(nrow = 5, ncol = 4), dim = c(5L, 4L)))
  expect_snapshot_failure(expect_shape(
    matrix(nrow = 6, ncol = 3),
    dim = c(6L, 2L)
  ))
  expect_snapshot_failure(expect_shape(
    matrix(nrow = 6, ncol = 3),
    dim = c(7L, 3L)
  ))
  expect_success(expect_shape(data.frame(1:10, 11:20), dim = c(10, 2)))
  expect_success(expect_shape(array(dim = 1:3), dim = 1:3))
  expect_snapshot_failure(expect_shape(array(dim = 1:3), dim = 1:2))
  expect_snapshot_failure(expect_shape(array(dim = 1:3), dim = 1:4))
  expect_success(expect_shape(array(integer()), dim = 0L))
  dd <- c(0L, 0L, 0L, 5L, 0L, 0L, 0L)
  expect_success(expect_shape(array(dim = dd), dim = dd))

  x <- cbind(1:2, 3:4)
  out <- expect_shape(x, dim = c(2L, 2L))
  expect_identical(out, x)
})

test_that("nrow compared correctly", {
  expect_success(expect_shape(matrix(nrow = 5, ncol = 4), nrow = 5L))
  expect_snapshot_failure(expect_shape(matrix(nrow = 5, ncol = 5), nrow = 6L))
  expect_success(expect_shape(data.frame(1:10, 11:20), nrow = 10L))
  expect_snapshot_failure(expect_shape(1, nrow = 1))
  expect_success(expect_shape(array(integer()), nrow = 0L))
  dd <- c(0L, 0L, 0L, 5L, 0L, 0L, 0L)
  expect_success(expect_shape(array(dim = dd), nrow = 0L))

  x <- cbind(1:2, 3:4)
  out <- expect_shape(x, dim = c(2L, 2L))
  expect_identical(out, x)
})

test_that("ncol compared correctly", {
  expect_success(expect_shape(matrix(nrow = 5, ncol = 4), ncol = 4L))
  expect_snapshot_failure(expect_shape(matrix(nrow = 5, ncol = 5), ncol = 7L))
  expect_success(expect_shape(data.frame(1:10, 11:20), ncol = 2L))
  expect_snapshot_failure(expect_shape(array(1), ncol = 1))
  expect_snapshot_failure(expect_shape(array(integer()), ncol = 0L))
  dd <- c(0L, 0L, 0L, 5L, 0L, 0L, 0L)
  expect_success(expect_shape(array(dim = dd), ncol = 0L))

  x <- cbind(1:2, 3:4)
  out <- expect_shape(x, dim = c(2L, 2L))
  expect_identical(out, x)
})

test_that("uses S3 dim method", {
  local_bindings(
    dim.testthat_expect_shape = function(x) 1:2,
    .env = globalenv()
  )
  x <- structure(integer(), class = "testthat_expect_shape")
  expect_success(expect_shape(x, dim = 1:2))
})

test_that("NA handling (e.g. dbplyr)", {
  local_bindings(
    dim.testthat_expect_shape_missing = function(x) c(NA_integer_, 10L),
    .env = globalenv()
  )
  x <- structure(integer(), class = "testthat_expect_shape_missing")

  expect_success(expect_shape(x, nrow = NA_integer_))
  expect_success(expect_shape(x, ncol = 10L))
  expect_success(expect_shape(x, dim = c(NA_integer_, 10L)))

  expect_snapshot_failure(expect_shape(x, nrow = 10L))
  expect_snapshot_failure(expect_shape(x, ncol = NA_integer_))
  expect_snapshot_failure(expect_shape(x, dim = c(10L, NA_integer_)))
})

test_that("uses S4 dim method", {
  A <- setClass("ExpectShapeA", slots = c(x = "numeric", y = "numeric"))
  setMethod("dim", "ExpectShapeA", function(x) 8:10)
  expect_success(expect_shape(A(x = 1:9, y = 3), dim = 8:10))
})

test_that("checks inputs arguments, ", {
  expect_snapshot(error = TRUE, {
    expect_shape(1:10)
    expect_shape(1:10, nrow = 1L, ncol = 2L)
    expect_shape(1:10, 2)
    expect_shape(array(1), nrow = "x")
    expect_shape(array(1), ncol = "x")
    expect_shape(array(1), dim = "x")
  })
})

#' Do you expect the results/output to equal a known value?
#'
#' For complex printed output and objects, it is often challenging to describe
#' exactly what you expect to see. `expect_known_value()` and
#' `expect_known_output()` provide a slightly weaker guarantee, simply
#' asserting that the values have not changed since the last time that you ran
#' them.
#'
#' These expectations should be used in conjunction with git, as otherwise
#' there is no way to revert to previous values. Git is particularly useful
#' in conjunction with `expect_known_output()` as the diffs will show you
#' exactly what has changed.
#'
#' Note that known values updates will only be updated when running tests
#' interactively. `R CMD check` clones the package source so any changes to
#' the reference files will occur in a temporary directory, and will not be
#' synchronised back to the source package.
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' `expect_known_output()` and friends are deprecated in the 3rd edition;
#' please use [expect_snapshot_output()] and friends instead.
#'
#' @export
#' @param file File path where known value/output will be stored.
#' @param update Should the file be updated? Defaults to `TRUE`, with
#'   the expectation that you'll notice changes because of the first failure,
#'   and then see the modified files in git.
#' @param version The serialization format version to use. The default, 2, was
#'   the default format from R 1.4.0 to 3.5.3. Version 3 became the default from
#'   R 3.6.0 and can only be read by R versions 3.5.0 and higher.
#' @param ... Passed on to [waldo::compare()].
#' @keywords internal
#' @inheritParams expect_equal
#' @inheritParams capture_output_lines
#' @examples
#' tmp <- tempfile()
#'
#' # The first run always succeeds
#' expect_known_output(mtcars[1:10, ], tmp, print = TRUE)
#'
#' # Subsequent runs will succeed only if the file is unchanged
#' # This will succeed:
#' expect_known_output(mtcars[1:10, ], tmp, print = TRUE)
#'
#' \dontrun{
#' # This will fail
#' expect_known_output(mtcars[1:9, ], tmp, print = TRUE)
#' }
expect_known_output <- function(
  object,
  file,
  update = TRUE,
  ...,
  info = NULL,
  label = NULL,
  print = FALSE,
  width = 80
) {
  check_string(file)
  check_bool(update)
  check_bool(print)
  check_number_whole(width, min = 1)
  edition_deprecate(
    3,
    "expect_known_output()",
    "Please use `expect_snapshot_output()` instead"
  )

  act <- list()
  act$quo <- enquo(object)
  act$lab <- label %||% quo_label(act$quo)
  act <- append(act, eval_with_output(object, print = print, width = width))

  expect_file_unchanged_(file, act$out, update = update, info = info, ...)
  invisible(act$val)
}

expect_file_unchanged_ <- function(
  path,
  lines,
  ...,
  update = TRUE,
  info = NULL,
  trace_env = caller_env()
) {
  if (!file.exists(path)) {
    cli::cli_warn("Creating reference output.")
    brio::write_lines(lines, path)
    pass()
    return()
  }

  old_lines <- brio::read_lines(path)
  if (update) {
    brio::write_lines(lines, path)
    if (!all_utf8(lines)) {
      cli::cli_warn("New reference output is not UTF-8 encoded.")
    }
  }
  if (!all_utf8(old_lines)) {
    cli::cli_warn("Reference output is not UTF-8 encoded.")
  }

  comp <- waldo_compare(
    x = old_lines,
    x_arg = "old",
    y = lines,
    y_arg = "new",
    ...
  )
  if (length(comp) != 0) {
    msg <- sprintf(
      "Results have changed from known value recorded in %s.\n\n%s",
      encodeString(path, quote = "'"),
      paste0(comp, collapse = "\n\n")
    )
    fail(msg, info = info, trace_env = trace_env)
  } else {
    pass()
  }
}

#' Do you expect the output/result to equal a known good value?
#'
#' `expect_output_file()` behaves identically to [expect_known_output()].
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' `expect_output_file()` is deprecated in the 3rd edition;
#' please use [expect_snapshot_output()] and friends instead.
#'
#' @export
#' @keywords internal
expect_output_file <- function(
  object,
  file,
  update = TRUE,
  ...,
  info = NULL,
  label = NULL,
  print = FALSE,
  width = 80
) {
  check_string(file)
  check_bool(update)
  check_bool(print)
  check_number_whole(width, min = 1)
  # Code is a copy of expect_known_output()
  edition_deprecate(
    3,
    "expect_output_file()",
    "Please use `expect_snapshot_output()` instead"
  )

  act <- list()
  act$quo <- enquo(object)
  act$lab <- label %||% quo_label(act$quo)
  act <- append(act, eval_with_output(object, print = print, width = width))

  expect_file_unchanged_(file, act$out, update = update, info = info, ...)
  invisible(act$val)
}

#' @export
#' @rdname expect_known_output
expect_known_value <- function(
  object,
  file,
  update = TRUE,
  ...,
  info = NULL,
  label = NULL,
  version = 2
) {
  check_string(file)
  check_bool(update)
  check_number_whole(version, min = 1)
  edition_deprecate(
    3,
    "expect_known_value()",
    "Please use `expect_snapshot_value()` instead"
  )

  act <- quasi_label(enquo(object), label)

  if (!file.exists(file)) {
    cli::cli_warn("Creating reference value.")
    saveRDS(object, file, version = version)
    pass()
  } else {
    ref_val <- readRDS(file)
    comp <- compare(act$val, ref_val, ...)
    if (update && !comp$equal) {
      saveRDS(act$val, file, version = version)
    }

    if (!comp$equal) {
      msg <- sprintf(
        "%s has changed from known value recorded in %s.\n%s",
        act$lab,
        encodeString(file, quote = "'"),
        comp$message
      )
      fail(msg, info = info)
    } else {
      pass()
    }
  }

  invisible(act$val)
}

#' @export
#' @rdname expect_known_output
#' @usage NULL
expect_equal_to_reference <- function(..., update = FALSE) {
  edition_deprecate(
    3,
    "expect_equal_to_reference()",
    "Please use `expect_snapshot_value()` instead"
  )
  expect_known_value(..., update = update)
}

#' @export
#' @rdname expect_known_output
#' @param hash Known hash value. Leave empty and you'll be informed what
#'   to use in the test output.
expect_known_hash <- function(object, hash = NULL) {
  check_installed("digest")
  edition_deprecate(
    3,
    "expect_known_hash()",
    "Please use `expect_snapshot_value()` instead"
  )

  act <- quasi_label(enquo(object))
  act_hash <- digest::digest(act$val)
  if (!is.null(hash)) {
    act_hash <- substr(act_hash, 1, nchar(hash))
  }

  if (is.null(hash)) {
    cli::cli_warn("No recorded hash: use {substr(act_hash, 1, 10)}.")
    pass()
  } else {
    if (hash != act_hash) {
      fail(sprintf(
        "Expected value to hash to %s.\nActual hash: %s",
        hash,
        act_hash
      ))
    } else {
      pass()
    }
  }

  invisible(act$val)
}

all_utf8 <- function(x) {
  !anyNA(iconv(x, "UTF-8", "UTF-8"))
}

local_edition(2)

# expect_known_output -----------------------------------------------------

test_that("uses specified width", {
  old <- options(width = 20)
  withr::defer(options(old))

  x <- 1:100
  expect_known_output(print(x), "width-80.txt")
})

test_that("creates file on first run", {
  file <- withr::local_tempfile()
  expect_known_output(cat("ok!\n"), file) |>
    expect_success() |>
    expect_warning("Creating reference")

  expect_true(file.exists(file))
})

test_that("ignores incomplete last line", {
  file <- withr::local_tempfile()
  writeLines("Hi!", file)
  expect_success(expect_known_output(cat("Hi!"), file))
  expect_success(expect_known_output(cat("Hi!\n"), file))
  expect_failure(expect_known_output(cat("Hi!\n\n"), file))
  expect_failure(expect_known_output(cat("oops"), file))
})

test_that("updates by default", {
  file <- withr::local_tempfile()
  writeLines("Hi!", file)
  expect_failure(expect_known_output(cat("oops"), file, update = FALSE))

  expect_equal(readLines(file), "Hi!")
  expect_failure(expect_known_output(cat("oops"), file, update = TRUE))
  expect_success(expect_known_output(cat("oops"), file))
})

test_that("works with utf-8 output", {
  skip_on_cran()
  skip_on_os("windows")

  text <- c("\u00fc", "\u2a5d", "\u6211", "\u0438")
  expect_known_output(cat(text, sep = "\n"), "test-expect-known.txt")
})

test_that("Warning for non-UTF-8 reference files", {
  x <- "\xe9\xe1\xed\xf6\xfc"
  Encoding(x) <- "latin1"

  tmp <- tempfile()
  withr::defer(unlink(tmp))
  writeBin(x, tmp)

  suppressWarnings(
    expect_failure(
      expect_known_output("foobar", tmp, update = FALSE)
    )
  )
})


# expect_known_value ------------------------------------------------------

test_that("correctly matches to a file", {
  x <- 1
  expect_success(expect_known_value(x, "one.rds"))

  x <- 2
  expect_failure(expect_known_value(x, "one.rds", update = FALSE))
})

test_that("first run is successful", {
  expect_known_value(2, "two.rds") |>
    expect_success() |>
    expect_warning("Creating reference")
  unlink("two.rds")
})

test_that("equal_to_ref does not overwrite existing", {
  tmp_rds <- tempfile(fileext = ".rds")
  withr::defer(unlink(tmp_rds))
  ref_obj1 <- 1:3
  ref_obj2 <- 2:4
  saveRDS(ref_obj1, tmp_rds)

  expect_success(expect_equal_to_reference(ref_obj1, tmp_rds))

  # Failure does not update object
  expect_failure(expect_equal_to_reference(ref_obj2, tmp_rds))
  expect_equal(readRDS(tmp_rds), ref_obj1)

  # Now failure does update object
  expect_failure(expect_equal_to_reference(ref_obj2, tmp_rds, update = TRUE))
  expect_success(expect_equal_to_reference(ref_obj2, tmp_rds))
})

test_that("serializes to version 2 by default", {
  tmp_rds <- tempfile(fileext = ".rds")
  withr::defer(unlink(tmp_rds))

  expect_warning(
    expect_known_value("a", tmp_rds),
    "Creating reference"
  )

  expect_identical(tools:::get_serialization_version(tmp_rds)[[1]], 2L)
})

test_that("version 3 is possible", {
  tmp_rds <- tempfile(fileext = ".rds")
  withr::defer(unlink(tmp_rds))

  expect_warning(
    expect_known_value("a", tmp_rds, version = 3),
    "Creating reference"
  )

  expect_identical(tools:::get_serialization_version(tmp_rds)[[1]], 3L)
})

# expect_known_hash -------------------------------------------------------

test_that("empty hash succeeds with warning", {
  expect_known_hash(1:10) |>
    expect_success() |>
    expect_warning("No recorded hash")
})

test_that("only succeeds if hash is correct", {
  expect_success(expect_known_hash(1:10, "c08951d2c2"))
  expect_failure(expect_known_hash(1:10, "c08951d2c3"))
})

## ----setup, include = FALSE---------------------------------------------------
library(testthat)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## -----------------------------------------------------------------------------
expect_length <- function(object, n) {
  # 1. Capture object and label
  act <- quasi_label(rlang::enquo(object), arg = "object")

  # 2. Call expect()
  act$n <- length(act$val)
  expect(
    act$n == n,
    sprintf("%s has length %i, not length %i.", act$lab, act$n, n)
  )

  # 3. Invisibly return the value
  invisible(act$val)
}

## -----------------------------------------------------------------------------
mtcars %>%
  expect_type("list") %>%
  expect_s3_class("data.frame") %>% 
  expect_length(11)

## -----------------------------------------------------------------------------
expect_length <- function(object, n) {
  act <- quasi_label(rlang::enquo(object), arg = "object")

  act$n <- length(act$val)
  if (act$n == n) {
    succeed()
    return(invisible(act$val))
  }

  message <- sprintf("%s has length %i, not length %i.", act$lab, act$n, n)
  fail(message)
}

# nocov start
praise <- function() {
  plain <- c(
    "You rock!",
    "You are a coding rockstar!",
    "Keep up the good work.",
    "Woot!",
    "Way to go!",
    "Nice code.",
    praise::praise("Your tests are ${adjective}!"),
    praise::praise("${EXCLAMATION} - ${adjective} code.")
  )
  utf8 <- c(
    "\U0001f600", # smile
    "\U0001f973", # party face
    "\U0001f638", # cat grin
    paste0(strrep("\U0001f389\U0001f38a", 5), "\U0001f389"),
    "\U0001f485 Your tests are beautiful \U0001f485",
    "\U0001f947 Your tests deserve a gold medal \U0001f947",
    "\U0001f308 Your tests are over the rainbow \U0001f308",
    "\U0001f9ff Your tests look perfect \U0001f9ff",
    "\U0001f3af Your tests hit the mark \U0001f3af",
    "\U0001f41d Your tests are the bee's knees \U0001f41d",
    "\U0001f3b8 Your tests rock \U0001f3b8",
    "\U0001f44f Your tests get an ovation \U0001f44f"
  )

  x <- if (cli::is_utf8_output()) c(plain, utf8) else plain
  sample(x, 1)
}

praise_emoji <- function() {
  if (!cli::is_utf8_output()) {
    return("")
  }

  emoji <- c(
    "\U0001f600", # smile
    "\U0001f973", # party face
    "\U0001f638", # cat grin
    "\U0001f308", # rainbow
    "\U0001f947", # gold medal
    "\U0001f389", # party popper
    "\U0001f38a" # confetti ball
  )
  paste0(" ", sample(emoji, 1))
}

encourage <- function() {
  x <- c(
    "Keep trying!",
    "Don't worry, you'll get it.",
    "No one is perfect!",
    "No one gets it right on their first try",
    "Frustration is a natural part of programming :)",
    "I believe in you!"
  )

  sample(x, 1)
}
# nocov end

#' Accept or reject modified snapshots
#'
#' * `snapshot_accept()` accepts all modified snapshots.
#' * `snapshot_reject()` rejects all modified snapshots by deleting the `.new` variants.
#' * `snapshot_review()` opens a Shiny app that shows a visual diff of each
#'    modified snapshot. This is particularly useful for whole file snapshots
#'    created by `expect_snapshot_file()`.
#'
#' @param files Optionally, filter effects to snapshots from specified files.
#'   This can be a snapshot name (e.g. `foo` or `foo.md`), a snapshot file name
#'   (e.g. `testfile/foo.txt`), or a snapshot file directory (e.g. `testfile/`).
#'
#' @param path Path to tests.
#' @export
snapshot_accept <- function(files = NULL, path = "tests/testthat") {
  changed <- snapshot_meta(files, path)
  if (nrow(changed) == 0) {
    cli::cli_inform("No snapshots to update.")
    return(invisible())
  }

  cli::cli_inform("Updating snapshots: {.path {changed$name}}.")
  unlink(changed$cur)
  file.rename(changed$new, changed$cur)

  rstudio_tickle()
  invisible()
}

#' @rdname snapshot_accept
#' @export
snapshot_reject <- function(files = NULL, path = "tests/testthat") {
  changed <- snapshot_meta(files, path)
  if (nrow(changed) == 0) {
    inform("No snapshots to reject")
    return(invisible())
  }

  inform(c("Rejecting snapshots:", changed$name))
  unlink(changed$new)

  rstudio_tickle()
  invisible()
}

#' @rdname snapshot_accept
#' @param ... Additional arguments passed on to [shiny::runApp()].
#' @export
snapshot_review <- function(files = NULL, path = "tests/testthat", ...) {
  check_installed(c("shiny", "diffviewer"), "to use snapshot_review()")

  changed <- snapshot_meta(files, path)
  if (nrow(changed) == 0) {
    cli::cli_inform("No snapshots to update.")
    return(invisible())
  }

  review_app(changed$name, changed$cur, changed$new, ...)
  rstudio_tickle()
  invisible()
}

review_app <- function(name, old_path, new_path, ...) {
  stopifnot(
    length(name) == length(old_path),
    length(old_path) == length(new_path)
  )

  n <- length(name)
  case_index <- stats::setNames(seq_along(name), name)
  handled <- rep(FALSE, n)

  ui <- shiny::fluidPage(
    style = "margin: 0.5em",
    shiny::fluidRow(
      style = "display: flex; margin-bottom: 0.5em",
      shiny::div(
        style = "flex: 1 1",
        if (n > 1) shiny::selectInput("cases", NULL, case_index, width = "100%")
      ),
      shiny::div(
        class = "btn-group",
        style = "margin-left: 1em; flex: 0 0 auto",
        shiny::actionButton("reject", "Reject", class = "btn-danger"),
        if (n > 1) shiny::actionButton("skip", "Skip"),
        shiny::actionButton("accept", "Accept", class = "btn-success"),
      )
    ),
    shiny::fluidRow(
      diffviewer::visual_diff_output("diff")
    )
  )
  server <- function(input, output, session) {
    i <- shiny::reactive(if (n == 1) 1L else as.numeric(input$cases))
    output$diff <- diffviewer::visual_diff_render({
      diffviewer::visual_diff(old_path[[i()]], new_path[[i()]])
    })

    # Can't skip if there's only one file to review
    shiny::updateActionButton(session, "skip", disabled = (n <= 1))

    # Handle buttons - after clicking update move input$cases to next case,
    # and remove current case (for accept/reject). If no cases left, close app
    shiny::observeEvent(input$reject, {
      cli::cli_inform("Rejecting snapshot: {.path {new_path[[i()]]}}.")
      unlink(new_path[[i()]])
      update_cases()
    })
    shiny::observeEvent(input$accept, {
      cli::cli_inform("Accepting snapshot: {.path {old_path[[i()]]}}.")
      file.rename(new_path[[i()]], old_path[[i()]])
      update_cases()
    })
    shiny::observeEvent(input$skip, {
      i <- next_case()
      shiny::updateSelectInput(session, "cases", selected = i)
    })

    update_cases <- function() {
      handled[[i()]] <<- TRUE
      i <- next_case()

      shiny::updateSelectInput(
        session,
        "cases",
        choices = case_index[!handled],
        selected = i
      )

      n_left <- sum(!handled)
      # Disable skip button if only one case remains
      shiny::updateActionButton(session, "skip", disabled = (n_left <= 1))
    }
    next_case <- function() {
      if (all(handled)) {
        cli::cli_inform("Review complete.")
        shiny::stopApp()
        return()
      }

      # Find next case;
      remaining <- case_index[!handled]
      next_cases <- which(remaining > i())
      if (length(next_cases) == 0) {
        remaining[[1]]
      } else {
        remaining[[next_cases[[1]]]]
      }
    }
  }

  cli::cli_inform(c(
    "Starting Shiny app for snapshot review.",
    i = "Use {.kbd Ctrl + C} to quit."
  ))
  shiny::runApp(
    shiny::shinyApp(ui, server),
    quiet = TRUE,
    launch.browser = shiny::paneViewer(),
    ...
  )
  invisible()
}

# helpers -----------------------------------------------------------------

snapshot_meta <- function(files = NULL, path = "tests/testthat") {
  all <- dir(file.path(path, "_snaps"), recursive = TRUE, full.names = TRUE)
  cur <- all[!grepl("\\.new\\.", all)]

  snap_file <- basename(dirname(cur)) != "_snaps"
  snap_test <- ifelse(
    snap_file,
    basename(dirname(cur)),
    gsub("\\.md$", "", basename(cur))
  )

  if (length(cur) == 0) {
    new <- character()
  } else {
    new <- paste0(tools::file_path_sans_ext(cur), ".new.", tools::file_ext(cur))
    new[!file.exists(new)] <- NA
  }

  snap_name <- ifelse(
    snap_file,
    file.path(snap_test, basename(cur)),
    basename(cur)
  )
  out <- data.frame(
    test = snap_test,
    name = snap_name,
    cur = cur,
    new = new,
    stringsAsFactors = FALSE
  )
  out <- out[!is.na(out$new), , drop = FALSE]
  out <- out[order(out$test, out$cur), , drop = FALSE]
  rownames(out) <- NULL

  if (!is.null(files)) {
    is_dir <- substr(files, nchar(files), nchar(files)) == "/"
    dirs <- files[is_dir]
    files <- files[!is_dir]

    dirs <- substr(dirs, 1, nchar(dirs) - 1)
    # Match regardless of whether user include .md or not
    files <- c(files, paste0(files, ".md"))

    out <- out[out$name %in% files | out$test %in% dirs, , drop = FALSE]
  }

  out
}

#' Snapshot testing
#'
#' @description
#' Snapshot tests (aka golden tests) are similar to unit tests except that the
#' expected result is stored in a separate file that is managed by testthat.
#' Snapshot tests are useful for when the expected value is large, or when
#' the intent of the code is something that can only be verified by a human
#' (e.g. this is a useful error message). Learn more in
#' `vignette("snapshotting")`.
#'
#' `expect_snapshot()` runs code as if you had executed it at the console, and
#' records the results, including output, messages, warnings, and errors.
#' If you just want to compare the result, try [expect_snapshot_value()].
#'
#' @section Workflow:
#' The first time that you run a snapshot expectation it will run `x`,
#' capture the results, and record them in `tests/testthat/_snaps/{test}.md`.
#' Each test file gets its own snapshot file, e.g. `test-foo.R` will get
#' `_snaps/foo.md`.
#'
#' It's important to review the Markdown files and commit them to git. They are
#' designed to be human readable, and you should always review new additions
#' to ensure that the salient information has been captured. They should also
#' be carefully reviewed in pull requests, to make sure that snapshots have
#' updated in the expected way.
#'
#' On subsequent runs, the result of `x` will be compared to the value stored
#' on disk. If it's different, the expectation will fail, and a new file
#' `_snaps/{test}.new.md` will be created. If the change was deliberate,
#' you can approve the change with [snapshot_accept()] and then the tests will
#' pass the next time you run them.
#'
#' Note that snapshotting can only work when executing a complete test file
#' (with [test_file()], [test_dir()], or friends) because there's otherwise
#' no way to figure out the snapshot path. If you run snapshot tests
#' interactively, they'll just display the current value.
#'
#' @param x Code to evaluate.
#' @param cran Should these expectations be verified on CRAN? By default,
#'   they are not, because snapshot tests tend to be fragile because they
#'   often rely on minor details of dependencies.
#' @param error Do you expect the code to throw an error? The expectation
#'   will fail (even on CRAN) if an unexpected error is thrown or the
#'   expected error is not thrown.
#' @param variant If non-`NULL`, results will be saved in
#'   `_snaps/{variant}/{test.md}`, so `variant` must be a single string
#'   suitable for use as a directory name.
#'
#'   You can use variants to deal with cases where the snapshot output varies
#'   and you want to capture and test the variations. Common use cases include
#'   variations for operating system, R version, or version of key dependency.
#'   Variants are an advanced feature. When you use them, you'll need to
#'   carefully think about your testing strategy to ensure that all important
#'   variants are covered by automated tests, and ensure that you have a way
#'   to get snapshot changes out of your CI system and back into the repo.
#' @param transform Optionally, a function to scrub sensitive or stochastic
#'   text from the output. Should take a character vector of lines as input
#'   and return a modified character vector as output.
#' @param cnd_class Whether to include the class of messages,
#'   warnings, and errors in the snapshot. Only the most specific
#'   class is included, i.e. the first element of `class(cnd)`.
#' @export
expect_snapshot <- function(x,
                            cran = FALSE,
                            error = FALSE,
                            transform = NULL,
                            variant = NULL,
                            cnd_class = FALSE) {
  edition_require(3, "expect_snapshot()")
  variant <- check_variant(variant)
  if (!is.null(transform)) {
    transform <- as_function(transform)
  }

  x <- enquo0(x)

  # Execute code, capturing last error
  state <- new_environment(list(error = NULL))
  replay <- function(x) {
    snapshot_replay(
      x,
      state,
      transform = transform,
      cnd_class = cnd_class
    )
  }
  with_is_snapshotting(
    out <- verify_exec(quo_get_expr(x), quo_get_env(x), replay)
  )

  # Use expect_error() machinery to confirm that error is as expected
  msg <- compare_condition_3e("error", NULL, state$error, quo_label(x), error)
  if (!is.null(msg)) {
    if (error) {
      expect(FALSE, msg, trace = state$error[["trace"]])
    } else {
      cnd_signal(state$error)
    }
    return()
  }

  expect_snapshot_helper("code", out,
    cran = cran,
    save = function(x) paste0(x, collapse = "\n"),
    load = function(x) split_by_line(x)[[1]],
    variant = variant,
    trace_env = caller_env()
  )
}

snapshot_replay <- function(x, state, ..., transform = NULL) {
  UseMethod("snapshot_replay", x)
}
#' @export
snapshot_replay.character <- function(x, state, ..., transform = NULL) {
  c(snap_header(state, "Output"), snapshot_lines(x, transform))
}
#' @export
snapshot_replay.source <- function(x, state, ..., transform = NULL) {
  c(snap_header(state, "Code"), snapshot_lines(x$src))
}
#' @export
snapshot_replay.condition <- function(x,
                                      state,
                                      ...,
                                      transform = NULL,
                                      cnd_class = FALSE) {

  cnd_message <- env_get(ns_env("rlang"), "cnd_message")

  if (inherits(x, "message")) {
    msg <- cnd_message(x)
    type <- "Message"
  } else {
    if (inherits(x, "error")) {
      state$error <- x
    }
    msg <- cnd_message(x, prefix = TRUE)
    type <- "Condition"
  }

  if (cnd_class) {
    type <- paste0(type, " <", class(x)[[1]], ">")
  }

  c(snap_header(state, type), snapshot_lines(msg, transform))
}

snapshot_lines <- function(x, transform = NULL) {
  x <- split_lines(x)
  if (!is.null(transform)) {
    x <- transform(x)
  }
  x <- indent(x)
  x
}

add_implicit_nl <- function(x) {
  if (substr(x, nchar(x), nchar(x)) == "\n") {
    x
  } else {
    paste0(x, "\n")
  }
}

snap_header <- function(state, header) {
  if (!identical(state$header, header)) {
    state$header <- header
    header
  }
}

#' Snapshot helpers
#'
#' @description
#' `r lifecycle::badge("questioning")`
#'
#' These snapshotting functions are questioning because they were developed
#' before [expect_snapshot()] and we're not sure that they still have a
#' role to play.
#'
#' * `expect_snapshot_output()` captures just output printed to the console.
#' * `expect_snapshot_error()` captures an error message and
#'   optionally checks its class.
#' * `expect_snapshot_warning()` captures a warning message and
#'   optionally checks its class.
#'
#' @inheritParams expect_snapshot
#' @keywords internal
#' @export
expect_snapshot_output <- function(x, cran = FALSE, variant = NULL) {
  edition_require(3, "expect_snapshot_output()")
  variant <- check_variant(variant)

  lab <- quo_label(enquo(x))
  with_is_snapshotting(
    val <- capture_output_lines(x, print = TRUE, width = NULL)
  )

  expect_snapshot_helper(lab, val,
    cran = cran,
    save = function(x) paste0(x, collapse = "\n"),
    load = function(x) split_by_line(x)[[1]],
    variant = variant,
    trace_env = caller_env()
  )
}

#' @param class Class of expected error or warning. The expectation will
#'   always fail (even on CRAN) if an error of this class isn't seen
#'   when executing `x`.
#' @export
#' @rdname expect_snapshot_output
expect_snapshot_error <- function(x, class = "error", cran = FALSE, variant = NULL) {
  edition_require(3, "expect_snapshot_error()")
  expect_snapshot_condition(
    "error", {{x}},
    class = class,
    cran = cran,
    variant = variant
  )
}

#' @export
#' @rdname expect_snapshot_output
expect_snapshot_warning <- function(x, class = "warning", cran = FALSE, variant = NULL) {
  edition_require(3, "expect_snapshot_warning()")
  expect_snapshot_condition(
    "warning", {{x}},
    class = class,
    cran = cran,
    variant = variant
  )
}

expect_snapshot_condition <- function(base_class, x, class, cran = FALSE, variant = NULL) {
  variant <- check_variant(variant)

  lab <- quo_label(enquo(x))
  with_is_snapshotting(
    val <- capture_matching_condition(x, cnd_matcher(class))
  )
  if (is.null(val)) {
    if (base_class == class) {
      fail(sprintf("%s did not generate %s", lab, base_class))
    } else {
      fail(sprintf("%s did not generate %s with class '%s'", lab, base_class, class))
    }
  }

  expect_snapshot_helper(
    lab,
    conditionMessage(val),
    cran = cran,
    variant = variant,
    trace_env = caller_env()
  )
}

expect_snapshot_helper <- function(lab, val,
                                   cran = FALSE,
                                   save = identity,
                                   load = identity,
                                   ...,
                                   tolerance = testthat_tolerance(),
                                   variant = NULL,
                                   trace_env = caller_env()
                                   ) {
  if (!cran && !interactive() && on_cran()) {
    skip("On CRAN")
  }

  snapshotter <- get_snapshotter()
  if (is.null(snapshotter)) {
    snapshot_not_available(save(val))
    return(invisible())
  }

  comp <- snapshotter$take_snapshot(val,
    save = save,
    load = load,
    ...,
    tolerance = tolerance,
    variant = variant,
    trace_env = trace_env
  )

  if (!identical(variant, "_default")) {
    variant_lab <- paste0(" (variant '", variant, "')")
  } else {
    variant_lab <- ""
  }
  hint <- snapshot_accept_hint(variant, snapshotter$file)

  expect(
    length(comp) == 0,
    sprintf(
      "Snapshot of %s has changed%s:\n%s\n\n%s",
      lab,
      variant_lab,
      paste0(comp, collapse = "\n\n"),
      hint
    ),
    trace_env = trace_env
  )
}

snapshot_accept_hint <- function(variant, file, reset_output = TRUE) {
  if (reset_output) {
    local_reporter_output()
  }

  if (is.null(variant) || variant == "_default") {
    name <- file
  } else {
    name <- file.path(variant, file)
  }

  paste0(
    cli::format_inline("* Run {.run testthat::snapshot_accept('{name}')} to accept the change."), "\n",
    cli::format_inline("* Run {.run testthat::snapshot_review('{name}')} to interactively review the change.")
  )
}

snapshot_not_available <- function(message) {
  local_reporter_output()

  cat(cli::rule("Snapshot"), "\n", sep = "")
  cli::cli_inform(c(
    i = "Can't save or compare to reference when testing interactively."
  ))
  cat(message, "\n", sep = "")
  cat(cli::rule(), "\n", sep = "")
}

local_snapshot_dir <- function(snap_names, .env = parent.frame()) {
  path <- withr::local_tempdir(.local_envir = .env)
  dir.create(file.path(path, "_snaps"), recursive = TRUE)

  dirs <- setdiff(unique(dirname(snap_names)), ".")
  for (dir in dirs) {
    dir.create(file.path(path, "_snaps", dir), recursive = TRUE, showWarnings = FALSE)
  }

  snap_paths <- file.path(path, "_snaps", snap_names)
  lapply(snap_paths, brio::write_lines, text = "")

  path
}

# if transform() wiped out the full message, don't indent, #1487
indent <- function(x) if (length(x)) paste0("  ", x) else x

check_variant <- function(x) {
  if (is.null(x)) {
    "_default"
  } else if (is_string(x)) {
    x
  } else {
    abort("If supplied, `variant` must be a string")
  }
}

with_is_snapshotting <- function(code) {
  withr::local_envvar(TESTTHAT_IS_SNAPSHOT = "true")
  code
}

test_that("informs about files being accepted", {
  path <- local_snapshot_dir(c(
    "a.md",
    "a.new.md",
    "test/b.txt",
    "test/b.new.txt"
  ))

  expect_snapshot(snapshot_accept(path = path))
  expect_equal(
    dir(file.path(path, "_snaps"), recursive = TRUE),
    c("a.md", "test/b.txt")
  )
})

test_that("useful mesasge if no files to accept", {
  path <- local_snapshot_dir(character())
  expect_snapshot(snapshot_accept(path = path))
})

test_that("can accept files created by expect_snapshot()", {
  # without extension
  path <- local_snapshot_dir(c("a.md", "a.new.md", "b.md", "b.new.md"))
  suppressMessages(snapshot_accept("a", path = path))
  expect_equal(dir(file.path(path, "_snaps")), c("a.md", "b.md", "b.new.md"))

  # with extension
  path <- local_snapshot_dir(c("a.md", "a.new.md", "b.md", "b.new.md"))
  suppressMessages(snapshot_accept("a.md", path = path))
  expect_equal(dir(file.path(path, "_snaps")), c("a.md", "b.md", "b.new.md"))

  # or whole directory
  path <- local_snapshot_dir(c("a.md", "a.new.md", "b.md", "b.new.md"))
  suppressMessages(snapshot_accept(path = path))
  expect_equal(dir(file.path(path, "_snaps")), c("a.md", "b.md"))
})

test_that("can accept files created by dotted tests in name", {
  # e.g. test-data.frame.R will create _snaps/data.frame.md

  # without extension
  path <- local_snapshot_dir(c("data.frame.md", "data.frame.new.md"))
  suppressMessages(snapshot_accept("data.frame", path = path))
  expect_equal(dir(file.path(path, "_snaps")), "data.frame.md")

  # with extension
  path <- local_snapshot_dir(c("data.frame.md", "data.frame.new.md"))
  suppressMessages(snapshot_accept("data.frame.md", path = path))
  expect_equal(dir(file.path(path, "_snaps")), "data.frame.md")
})

test_that("can accept files created by expect_snapshot_file()", {
  path <- local_snapshot_dir(c("test/a.txt", "test/a.new.txt"))
  suppressMessages(snapshot_accept("test/a.txt", path = path))
  expect_equal(dir(file.path(path, "_snaps"), recursive = TRUE), "test/a.txt")

  # including markdown files
  path <- local_snapshot_dir(c("test/a.md", "test/a.new.md"))
  suppressMessages(snapshot_accept("test/", path = path))
  expect_equal(dir(file.path(path, "_snaps"), recursive = TRUE), "test/a.md")

  # or the whole directory
  path <- local_snapshot_dir(c(
    "test/a.md",
    "test/a.new.md",
    "test/b.txt",
    "test/b.new.txt"
  ))
  suppressMessages(snapshot_accept("test/", path = path))
  expect_equal(
    dir(file.path(path, "_snaps"), recursive = TRUE),
    c("test/a.md", "test/b.txt")
  )
})

test_that("can work with variants", {
  # Can accept all
  path <- local_snapshot_dir(c("foo/a.md", "foo/a.new.md"))
  expect_snapshot(snapshot_accept(path = path))
  expect_equal(dir(file.path(path, "_snaps", "foo")), "a.md")

  # Can accept specified
  path <- local_snapshot_dir(c("foo/a.md", "foo/a.new.md"))
  expect_snapshot(snapshot_accept("foo/a", path = path))
  expect_equal(dir(file.path(path, "_snaps", "foo")), "a.md")
})

test_that("snapshot_reject deletes .new files", {
  path <- local_snapshot_dir(c("a.md", "a.new.md", "b.md", "b.new.md"))

  expect_snapshot(snapshot_reject(path = path))
  expect_equal(dir(file.path(path, "_snaps")), c("a.md", "b.md"))
})


# snapshot_meta -----------------------------------------------------------

test_that("returns empty data frame for empty directory", {
  path <- tempfile()
  dir.create(path)

  expect_equal(nrow(snapshot_meta(path = path)), 0)
})

#' Do you expect a vector with (these) names?
#'
#' You can either check for the presence of names (leaving `expected`
#' blank), specific names (by supplying a vector of names), or absence of
#' names (with `NULL`).
#'
#' @inheritParams expect_that
#' @param expected Character vector of expected names. Leave missing to
#'   match any names. Use `NULL` to check for absence of names.
#' @param ignore.order If `TRUE`, sorts names before comparing to
#'   ignore the effect of order.
#' @param ignore.case If `TRUE`, lowercases all names to ignore the
#'   effect of case.
#' @family expectations
#' @export
#' @examples
#' x <- c(a = 1, b = 2, c = 3)
#' expect_named(x)
#' expect_named(x, c("a", "b", "c"))
#'
#' # Use options to control sensitivity
#' expect_named(x, c("B", "C", "A"), ignore.order = TRUE, ignore.case = TRUE)
#'
#' # Can also check for the absence of names with NULL
#' z <- 1:4
#' expect_named(z, NULL)
expect_named <- function(
  object,
  expected,
  ignore.order = FALSE,
  ignore.case = FALSE,
  info = NULL,
  label = NULL
) {
  check_bool(ignore.order)
  check_bool(ignore.case)

  act <- quasi_label(enquo(object), label)

  if (missing(expected)) {
    act_names <- names(act$val)
    if (is.null(act_names)) {
      msg <- sprintf("Expected %s to have names.", act$lab)
      fail(msg)
    } else {
      pass()
    }
  } else {
    exp <- quasi_label(enquo(expected), arg = "expected")

    exp$val <- normalise_names(exp$val, ignore.order, ignore.case)
    act_names <- labelled_value(
      normalise_names(names(act$val), ignore.order, ignore.case),
      act$lab
    )

    msg <- "Expected %s to have names %s."
    if (ignore.order) {
      expect_setequal_(msg, act_names, exp)
    } else {
      expect_waldo_equal_(msg, act_names, exp)
    }
  }

  invisible(act$val)
}

normalise_names <- function(x, ignore.order = FALSE, ignore.case = FALSE) {
  if (is.null(x)) {
    return()
  }

  if (ignore.order) {
    x <- sort(x)
  }
  if (ignore.case) {
    x <- tolower(x)
  }

  x
}

#' @importFrom magrittr %>%
#' @export
magrittr::`%>%`

can_entrace <- function(cnd) {
  !inherits(cnd, "Throwable")
}

# Need to strip environment and source references to make lightweight
# function suitable to send to another process
transport_fun <- function(f) {
  environment(f) <- .GlobalEnv
  f <- zap_srcref(f)
  f
}

# Handled specially in test_code so no backtrace
testthat_warn <- function(message, ...) {
  warn(message, class = "testthat_warn", ...)
}

split_by_line <- function(x) {
  trailing_nl <- grepl("\n$", x)

  x <- strsplit(x, "\n")
  x[trailing_nl] <- lapply(x[trailing_nl], c, "")
  x
}

rstudio_tickle <- function() {
  if (!is_installed("rstudioapi")) {
    return()
  }

  if (!rstudioapi::hasFun("executeCommand")) {
    return()
  }

  rstudioapi::executeCommand("vcsRefresh")
  if (!is_positron()) {
    rstudioapi::executeCommand("refreshFiles")
  }
}

is_positron <- function() {
  nzchar(Sys.getenv("POSITRON", ""))
}

first_upper <- function(x) {
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

in_check_reporter <- function() {
  isTRUE(the$in_check_reporter)
}

r_version <- function() paste0("R", getRversion()[, 1:2])

# Supress cli wrapping
no_wrap <- function(x) {
  x <- gsub(" ", "\u00a0", x, fixed = TRUE)
  x <- gsub("\n", "\f", x, fixed = TRUE)
  x
}

paste_c <- function(...) {
  paste0(c(...), collapse = "")
}

# from rematch2
re_match <- function(text, pattern, perl = TRUE, ...) {
  stopifnot(is.character(pattern), length(pattern) == 1, !is.na(pattern))
  text <- as.character(text)
  match <- regexpr(pattern, text, perl = perl, ...)
  start <- as.vector(match)
  length <- attr(match, "match.length")
  end <- start + length - 1L
  matchstr <- substring(text, start, end)
  matchstr[start == -1] <- NA_character_
  res <- data.frame(stringsAsFactors = FALSE, .text = text, .match = matchstr)
  if (!is.null(attr(match, "capture.start"))) {
    gstart <- attr(match, "capture.start")
    glength <- attr(match, "capture.length")
    gend <- gstart + glength - 1L
    groupstr <- substring(text, gstart, gend)
    groupstr[gstart == -1] <- NA_character_
    dim(groupstr) <- dim(gstart)
    res <- cbind(groupstr, res, stringsAsFactors = FALSE)
  }
  names(res) <- c(attr(match, "capture.names"), ".text", ".match")
  class(res) <- c("tbl_df", "tbl", class(res))
  res
}

readLines <- function(...) stop("Use read_lines!")
writeLines <- function(...) stop("Use write_lines!")

read_lines <- function(path, n = -1L, encoding = "UTF-8") {
  base::readLines(path, n = n, encoding = encoding, warn = FALSE)
}

write_lines <- function(text, path) {
  base::writeLines(enc2utf8(text), path, useBytes = TRUE)
}

all_utf8 <- function(x) {
  ! any(is.na(iconv(x, "UTF-8", "UTF-8")))
}

## -----------------------------------------------------------------------------
library(testthat)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

# Pretend we're snapshotting
snapper <- local_snapshotter(fail_on_new = FALSE)
snapper$start_file("snapshotting.Rmd", "test")

# Pretend we're testing testthat so we can use mocking
Sys.setenv(TESTTHAT_PKG = "testthat")

## -----------------------------------------------------------------------------
dice <- function() {
  sample(6, 1)
}

test_that("dice returns different numbers", {
  withr::local_seed(1234)

  expect_equal(dice(), 4)
  expect_equal(dice(), 2)
  expect_equal(dice(), 6)
})

## -----------------------------------------------------------------------------
roll_three <- function() {
  sum(dice(), dice(), dice())
}

test_that("three dice adds values of individual calls", {
  local_mocked_bindings(dice = mock_output_sequence(1, 2, 3))
  expect_equal(roll_three(), 6)
})

## -----------------------------------------------------------------------------
continue <- function(prompt) {
  cat(prompt, "\n", sep = "")

  repeat {
    val <- readline("Do you want to continue? (y/n) ")
    if (val %in% c("y", "n")) {
      return(val == "y")
    }
    cat("! You must enter y or n\n")
  }  
}

readline <- NULL

## -----------------------------------------------------------------------------
test_that("user must respond y or n", {
  mock_readline <- local({
    i <- 0
    function(prompt) {
      i <<- i + 1
      cat(prompt)
      val <- if (i == 1) "x" else "y"
      cat(val, "\n", sep = "")
      val
    }
  })

  local_mocked_bindings(readline = mock_readline)
  expect_snapshot(val <- continue("This is dangerous"))
  expect_true(val)
})

## -----------------------------------------------------------------------------
f <- mock_output_sequence(1, 12, 123)
f()
f()
f()

## -----------------------------------------------------------------------------
test_that("user must respond y or n", {
  local_mocked_bindings(readline = mock_output_sequence("x", "y"))
  expect_true(continue("This is dangerous"))
})

## -----------------------------------------------------------------------------
save_file <- function(path, data) {
  if (file.exists(path)) {
    if (!continue("`path` already exists")) {
      stop("Failed to continue")
    }
  }
  writeLines(data, path)
}

test_that("save_file() requires confirmation to overwrite file", {
  path <- withr::local_tempfile(lines = letters)

  local_mocked_bindings(continue = function(...) TRUE)
  save_file(path, "a")
  expect_equal(readLines(path), "a")

  local_mocked_bindings(continue = function(...) FALSE)
  expect_snapshot(save_file(path, "a"), error = TRUE)
})

test_that("get_routine() finds own 'run_testthat_tests'", {
  routine <- get_routine("testthat", "run_testthat_tests")
  expect_s3_class(routine, "NativeSymbolInfo")
})

test_that("get_routine() fails when no routine exists", {
  expect_snapshot(error = TRUE, get_routine("utils", "no_such_routine"))
})

test_that("validates inputs", {
  expect_snapshot(error = TRUE, {
    expect_cpp_tests_pass(123)
    run_cpp_tests(123)
  })
})

test_that("useful messaging", {
  path <- withr::local_tempdir()
  writeLines("Package: foo", file.path(path, "DESCRIPTION"))
  dir.create(file.path(path, "R"))

  expect_snapshot(use_catch(path))
})

skip_if_not_installed("xml2")
run_cpp_tests("testthat")

#' Fail if any tests fail
#'
#' This reporter will simply throw an error if any of the tests failed. It is
#' best combined with another reporter, such as the
#' [SummaryReporter].
#'
#' @export
#' @family reporters
FailReporter <- R6::R6Class(
  "FailReporter",
  inherit = Reporter,
  public = list(
    failed = FALSE,

    initialize = function(...) {
      self$capabilities$parallel_support <- TRUE
      super$initialize(...)
    },

    add_result = function(context, test, result) {
      self$failed <- self$failed || expectation_broken(result)
    },

    end_reporter = function() {
      if (self$failed) {
        cli::cli_abort("Failures detected.")
      }
    }
  )
)

#' Extract a reprex from a failed expectation
#'
#' @description
#' `extract_test()` creates a minimal reprex for a failed expectation.
#' It extracts all non-test code before the failed expectation as well as
#' all code inside the test up to and including the failed expectation.
#'
#' This is particularly useful when you're debugging test failures in
#' someone else's package.
#'
#' @param location A string giving the location in the form
#'   `FILE:LINE[:COLUMN]`.
#' @param path Path to write the reprex to. Defaults to `stdout()`.
#' @param package If supplied, will be used to construct a test environment
#'   for the extracted code.
#' @return This function is called for its side effect of rendering a
#'   reprex to `path`. This function will never error: if extraction
#'   fails, the error message will be written to `path`.
#' @export
#' @examples
#' # If you see a test failure like this:
#' # -- Failure (test-extract.R:46:3): errors if can't find test -------------
#' # Expected FALSE to be TRUE.
#' # Differences:
#' # `actual`:   FALSE
#' # `expected`: TRUE
#'
#' # You can run this:
#' \dontrun{extract_test("test-extract.R:46:3")}
#' # to see just the code needed to reproduce the failure
extract_test <- function(
  location,
  path = stdout(),
  package = Sys.getenv("TESTTHAT_PKG")
) {
  check_string(location)
  check_string(package)

  pieces <- strsplit(location, ":")[[1]]
  if (!length(pieces) %in% c(2, 3)) {
    cli::cli_abort(c(
      "Expected {.arg location} to be of the form FILE:LINE[:COLUMN]",
      i = "Got {.arg location}: {.val {location}}"
    ))
  }

  test_path <- test_path(pieces[[1]])
  line <- as.integer(pieces[2])
  lines <- extract_test_(test_path, line, package)

  base::writeLines(lines, con = path)
}

#' Simulate a test environment
#'
#' This function is designed to allow you to simulate testthat's testing
#' environment in an interactive session. To undo it's affect, you
#' will need to restart your R session.
#'
#' @keywords internal
#' @param package Name of installed package.
#' @param path Path to `tests/testthat`.
#' @export
#' @rdname topic-name
simulate_test_env <- function(package, path) {
  check_string(package)
  check_string(path)

  env <- test_env(package)
  source_test_helpers(path, env = env)
  source_test_setup(path, env = env)

  invisible(env)
}

extract_test_ <- function(
  test_path,
  line,
  package = Sys.getenv("TESTTHAT_PKG")
) {
  source <- paste0("# Extracted from ", test_path, ":", line)
  exprs <- parse_file(test_path)

  lines <- tryCatch(
    extract_test_lines(exprs, line, package),
    error = function(cnd) {
      lines <- strsplit(conditionMessage(cnd), "\n")[[1]]
      lines <- c("", "Failed to extract test: ", lines)
      paste0("# ", lines)
    }
  )
  lines <- c(source, "", lines)
  lines
}

save_test <- function(srcref, dir, package = Sys.getenv("TESTTHAT_PKG")) {
  if (env_var_is_false("TESTTHAT_PROBLEMS")) {
    return()
  }

  test_path <- utils::getSrcFilename(srcref, full.names = TRUE)
  if (is.null(test_path) || !file.exists(test_path)) {
    return()
  }
  line <- srcref[[3]]
  extracted <- extract_test_(test_path, line, package)

  test_name <- tools::file_path_sans_ext(basename(test_path))
  dir_create(dir)
  problems_path <- file.path(dir, paste0(test_name, "-", line, ".R"))
  cat("Saving ", problems_path, "\n", sep = "")
  writeLines(extracted, problems_path)

  invisible(problems_path)
}

extract_test_lines <- function(
  exprs,
  line,
  package = "",
  error_call = caller_env()
) {
  check_number_whole(line, min = 1, call = error_call)

  srcrefs <- attr(exprs, "srcref")
  is_subtest <- map_lgl(exprs, is_subtest)

  # First we find the test
  is_test <- is_subtest &
    start_line(srcrefs) <= line &
    end_line(srcrefs) >= line
  if (!any(is_test)) {
    cli::cli_abort("Failed to find test at line {line}.", call = error_call)
  }
  call <- exprs[[which(is_test)[[1]]]]
  test_contents <- attr(call[[3]], "srcref")[-1] # drop `{`
  keep <- start_line(test_contents) <= line
  lines <- c(header("test"), srcref_to_character(test_contents[keep]))

  # We first find the prequel, all non-test code before the test
  is_prequel <- !is_subtest & start_line(srcrefs) < line
  if (any(is_prequel)) {
    lines <- c(
      header("prequel"),
      srcref_to_character(srcrefs[is_prequel]),
      "",
      lines
    )
  }

  if (package != "") {
    lines <- c(
      header("setup"),
      "library(testthat)",
      sprintf(
        'test_env <- simulate_test_env(package = "%s", path = "..")',
        package
      ),
      "attach(test_env, warn.conflicts = FALSE)",
      "",
      lines
    )
  }
  lines
}

# Helpers ---------------------------------------------------------------------

parse_file <- function(path, error_call = caller_env()) {
  check_string(path, call = error_call)
  if (!file.exists(path)) {
    cli::cli_abort(
      "{.arg path} ({.path {path}}) does not exist.",
      call = error_call
    )
  }
  parse(path, keep.source = TRUE)
}

parse_text <- function(text) {
  text <- sub("^\n", "", text)
  indent <- regmatches(text, regexpr("^ *", text))
  text <- gsub(paste0("(?m)^", indent), "", text, perl = TRUE)

  parse(text = text, keep.source = TRUE)
}

srcref_to_character <- function(x) {
  unlist(map(x, as.character))
}
start_line <- function(srcrefs) {
  map_int(srcrefs, \(x) x[[1]])
}
end_line <- function(srcrefs) {
  map_int(srcrefs, \(x) x[[3]])
}

header <- function(x) {
  paste0("# ", x, " ", strrep("-", 80 - nchar(x) - 3))
}

test_that("can extract test from file", {
  exprs <- parse_file(test_path("extract/simple.R"))

  dir <- withr::local_tempdir()
  out_path <- save_test(attr(exprs, "srcref")[[1]], dir = dir)

  expect_snapshot(base::writeLines(readLines(out_path)))
})

# extract_test_lines -----------------------------------------------------------

test_that("can include test env setup", {
  # fmt: skip
  exprs <- parse_text("
    test_that('foo', {
      expect_true(TRUE)
    })
  ")
  expect_snapshot(base::writeLines(extract_test_lines(exprs, 2, "test")))
})

test_that("can extract prequel", {
  # fmt: skip
  exprs <- parse_text("
    x <- 1
    y <- 2
    test_that('foo', {
      expect_true(TRUE)
    })
  ")
  expect_snapshot(base::writeLines(extract_test_lines(exprs, 4)))
})

test_that("preserves code format but not comments", {
  # fmt: skip
  exprs <- parse_text("
    1 +   1 # 2
    test_that('foo', {
      2 +   2 # 4
    })
  ")
  expect_snapshot(base::writeLines(extract_test_lines(exprs, 3)))
})

test_that("can extract selected expectation", {
  # fmt: skip
  exprs <- parse_text("
    test_that('foo', {
      expect_true(TRUE)
      expect_false(FALSE)
    })
  ")
  expect_snapshot(base::writeLines(extract_test_lines(exprs, 2)))
})

test_that("errors if can't find test", {
  # fmt: skip
  exprs <- parse_text("
    # line 1
    test_that('foo', {
      expect_true(TRUE)
    })
    # line 5
  ")
  expect_error(extract_test_lines(exprs, 1), "Failed to find test")
  expect_error(extract_test_lines(exprs, 5), "Failed to find test")
})

#' Retrieve paths to built-in example test files
#'
#' `testthat_examples()` retrieves path to directory of test files,
#' `testthat_example()` retrieves path to a single test file.
#'
#' @keywords internal
#' @param filename Name of test file
#' @export
#' @examples
#' dir(testthat_examples())
#' testthat_example("success")
testthat_examples <- function() {
  system.file("examples", package = "testthat")
}

#' @export
#' @rdname testthat_examples
testthat_example <- function(filename) {
  system.file(
    "examples", paste0("test-", filename, ".R"),
    package = "testthat",
    mustWork = TRUE
  )
}

# data is list of character vectors
snap_to_md <- function(data) {
  h2 <- paste0("# ", names(data), "\n\n")
  code_block <- function(x) paste0(indent_add(x), collapse = "\n\n---\n\n")
  data <- vapply(data, code_block, character(1))

  paste0(h2, data, "\n\n", collapse = "")
}

snap_from_md <- function(lines) {
  lines <- gsub("\r", "", lines, fixed = TRUE)

  h2 <- grepl("^# ", lines)
  tests_group <- cumsum(h2)
  tests <- split(lines[!h2], tests_group[!h2])
  names(tests) <- gsub("^# ", "", lines[h2])

  split_tests <- function(lines) {
    sep <- grepl("^-{3,}", lines)
    case_group <- cumsum(sep)

    # Remove first line and last line, separator, line above and line below
    sep_loc <- which(sep)
    drop <- c(1, sep_loc, sep_loc + 1, sep_loc - 1, length(lines))

    cases <- unname(split(lines[-drop], case_group[-drop]))
    code_unblock <- function(x) paste0(indent_del(x), collapse = "\n")
    vapply(cases, code_unblock, character(1))
  }

  lapply(tests, split_tests)
}

read_snaps <- function(path) {
  if (file.exists(path)) {
    lines <- brio::read_lines(path)
    snap_from_md(lines)
  } else {
    list()
  }
}

write_snaps <- function(snaps, path, delete = FALSE) {
  snaps <- compact(snaps)
  if (length(snaps) == 0) {
    if (delete) {
      unlink(path)
    }
    return()
  }

  out <- snap_to_md(snaps)
  brio::write_file(out, path)
}

# Helpers -----------------------------------------------------------------

indent_add <- function(x, prefix = "    ") {
  paste0(prefix, gsub("\n", paste0("\n", prefix), x, fixed = TRUE))
}
indent_del <- function(x, prefix = "    ") {
  x <- gsub(paste0("^", prefix), "", x)
  x <- gsub(paste0("\n", prefix), "\n", x)
  x
}

test_that("single test case can roundtrip", {
  x <- list(test = '[1] "x"')
  x_snap <- snap_to_md(x)
  x_lines <- strsplit(x_snap, "\n")[[1]]

  y <- snap_from_md(x_lines)
  expect_equal(x, y)
})

test_that("multiple tests can roundtrip", {
  x <- list(foo = c("a", "b"), bar = "d", baz = letters[1:3])
  x_snap <- snap_to_md(x)
  x_lines <- strsplit(x_snap, "\n")[[1]]

  y <- snap_from_md(x_lines)
  expect_equal(x, y)
})

test_that("snapshots always use \n", {
  path <- withr::local_tempfile()
  x <- list(foo = c("a", "b"), bar = "d", baz = letters[1:3])
  write_snaps(x, path)

  snap <- brio::read_file(path)
  has_cr <- grepl("\r", snap, fixed = TRUE)
  expect_equal(has_cr, FALSE)
})

test_that("snap_from_md handles missing final newlines", {
  one_newline <- withr::local_tempfile(
    fileext = ".md",
    lines = c(
      "# test_case",
      "",
      "result1",
      "",
      "---",
      "",
      "result2"
    )
  )
  expect_equal(
    read_snaps(one_newline),
    list(test_case = c("result1", "result2"))
  )
})

#' Provide human-readable comparison of two objects
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `compare` is similar to [base::all.equal()], but somewhat buggy in its
#' use of `tolerance`. Please use [waldo](https://waldo.r-lib.org/) instead.
#'
#' @export
#' @param x,y Objects to compare
#' @param ... Additional arguments used to control specifics of comparison
#' @keywords internal
#' @order 1
compare <- function(x, y, ...) {
  UseMethod("compare", x)
}

comparison <- function(equal = TRUE, message = "Equal") {
  check_bool(equal)
  check_character(message)

  structure(
    list(
      equal = equal,
      message = paste(message, collapse = "\n")
    ),
    class = "comparison"
  )
}
difference <- function(..., fmt = "%s") {
  comparison(FALSE, sprintf(fmt, ...))
}
no_difference <- function() {
  comparison()
}

#' @export
print.comparison <- function(x, ...) {
  if (x$equal) {
    cat("Equal\n")
    return()
  }

  cat(x$message)
}

#' @export
#' @rdname compare
#' @order 2
compare.default <- function(x, y, ..., max_diffs = 9) {
  same <- all.equal(x, y, ...)
  if (length(same) > max_diffs) {
    same <- c(same[1:max_diffs], "...")
  }

  comparison(identical(same, TRUE), as.character(same))
}

print_out <- function(x, ...) {
  lines <- capture_output_lines(x, ..., print = TRUE)
  paste0(lines, collapse = "\n")
}

# Common helpers ---------------------------------------------------------------

same_length <- function(x, y) length(x) == length(y)
diff_length <- function(x, y) {
  difference(fmt = "Lengths differ: %i is not %i", length(x), length(y))
}

same_type <- function(x, y) identical(typeof(x), typeof(y))
diff_type <- function(x, y) {
  difference(fmt = "Types not compatible: %s is not %s", typeof(x), typeof(y))
}

same_class <- function(x, y) {
  if (!is.object(x) && !is.object(y)) {
    return(TRUE)
  }
  identical(class(x), class(y))
}
diff_class <- function(x, y) {
  difference(
    fmt = "Classes differ: %s is not %s",
    format_class(class(x)),
    format_class(class(y))
  )
}

same_attr <- function(x, y) {
  is.null(attr.all.equal(x, y))
}
diff_attr <- function(x, y) {
  out <- attr.all.equal(x, y)
  difference(out)
}

vector_equal <- function(x, y) {
  (is.na(x) & is.na(y)) | (!is.na(x) & !is.na(y) & x == y)
}

vector_equal_tol <- function(x, y, tolerance = .Machine$double.eps^0.5) {
  (is.na(x) & is.na(y)) |
    (!is.na(x) & !is.na(y)) & (x == y | abs(x - y) < tolerance)
}


# character ---------------------------------------------------------------

#' @param max_diffs Maximum number of differences to show
#' @param max_lines Maximum number of lines to show from each difference
#' @param check.attributes If `TRUE`, also checks values of attributes.
#' @param width Width of output device
#' @rdname compare
#' @export
#' @examples
#' # Character -----------------------------------------------------------------
#' x <- c("abc", "def", "jih")
#' compare(x, x)
#'
#' y <- paste0(x, "y")
#' compare(x, y)
#'
#' compare(letters, paste0(letters, "-"))
#'
#' x <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Duis cursus
#'  tincidunt auctor. Vestibulum ac metus bibendum, facilisis nisi non, pulvinar
#'  dolor. Donec pretium iaculis nulla, ut interdum sapien ultricies a. "
#' y <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Duis cursus
#'  tincidunt auctor. Vestibulum ac metus1 bibendum, facilisis nisi non, pulvinar
#'  dolor. Donec pretium iaculis nulla, ut interdum sapien ultricies a. "
#' compare(x, y)
#' compare(c(x, x), c(y, y))
#'
compare.character <- function(
  x,
  y,
  check.attributes = TRUE,
  ...,
  max_diffs = 5,
  max_lines = 5,
  width = cli::console_width()
) {
  if (identical(x, y)) {
    return(no_difference())
  }

  if (!same_type(x, y)) {
    return(diff_type(x, y))
  }
  if (!same_class(x, y)) {
    return(diff_class(x, y))
  }
  if (!same_length(x, y)) {
    return(diff_length(x, y))
  }
  if (check.attributes && !same_attr(x, y)) {
    return(diff_attr(x, y))
  }

  diff <- !vector_equal(x, y)

  if (!any(diff)) {
    no_difference()
  } else {
    mismatches <- mismatch_character(x, y, diff)
    difference(format(
      mismatches,
      max_diffs = max_diffs,
      max_lines = max_lines,
      width = width
    ))
  }
}

mismatch_character <- function(x, y, diff = !vector_equal(x, y)) {
  structure(
    list(
      i = which(diff),
      x = x[diff],
      y = y[diff],
      n = length(diff),
      n_diff = sum(diff)
    ),
    class = "mismatch_character"
  )
}

#' @export
format.mismatch_character <- function(
  x,
  ...,
  max_diffs = 5,
  max_lines = 5,
  width = cli::console_width()
) {
  width <- width - 6 # allocate space for labels
  n_show <- seq_len(min(x$n_diff, max_diffs))

  encode <- function(x) encodeString(x, quote = '"')
  show_x <- str_trunc(encode(x$x[n_show]), width * max_lines)
  show_y <- str_trunc(encode(x$y[n_show]), width * max_lines)
  show_i <- x$i[n_show]

  sidebyside <- Map(
    function(x, y, pos) {
      x <- paste0("x[", pos, "]: ", str_chunk(x, width))
      y <- paste0("y[", pos, "]: ", str_chunk(y, width))
      paste(c(x, y), collapse = "\n")
    },
    show_x,
    show_y,
    show_i
  )

  summary <- paste0(x$n_diff, "/", x$n, " mismatches")
  paste0(summary, "\n", paste0(sidebyside, collapse = "\n\n"))
}

#' @export
print.mismatch_character <- function(x, ...) {
  cat(format(x, ...), "\n", sep = "")
}

str_trunc <- function(x, length) {
  too_long <- nchar(x) > length

  x[too_long] <- paste0(substr(x[too_long], 1, length - 3), "...")
  x
}
str_chunk <- function(x, length) {
  lines <- ceiling(nchar(x) / length)
  start <- (seq_len(lines) - 1) * length + 1

  substring(x, start, start + length - 1)
}

# compare.numeric ---------------------------------------------------------

#' @export
#' @rdname compare
#' @param tolerance Numerical tolerance: any differences (in the sense of
#'   [base::all.equal()]) smaller than this value will be ignored.
#'
#'   The default tolerance is `sqrt(.Machine$double.eps)`, unless long doubles
#'   are not available, in which case the test is skipped.
#' @examples
#' # Numeric -------------------------------------------------------------------
#'
#' x <- y <- runif(100)
#' y[sample(100, 10)] <- 5
#' compare(x, y)
#'
#' x <- y <- 1:10
#' x[5] <- NA
#' x[6] <- 6.5
#' compare(x, y)
#'
#' # Compare ignores minor numeric differences in the same way
#' # as all.equal.
#' compare(x, x + 1e-9)
compare.numeric <- function(
  x,
  y,
  tolerance = testthat_tolerance(),
  check.attributes = TRUE,
  ...,
  max_diffs = 9
) {
  all_equal <- all.equal(
    x,
    y,
    tolerance = tolerance,
    check.attributes = check.attributes,
    ...
  )
  if (isTRUE(all_equal)) {
    return(no_difference())
  }

  if (!typeof(y) %in% c("integer", "double")) {
    return(diff_type(x, y))
  }
  if (!same_class(x, y)) {
    return(diff_class(x, y))
  }
  if (!same_length(x, y)) {
    return(diff_length(x, y))
  }
  if (check.attributes && !same_attr(x, y)) {
    return(diff_attr(x, y))
  }

  diff <- !vector_equal_tol(x, y, tolerance = tolerance)

  if (!any(diff)) {
    no_difference()
  } else {
    mismatches <- mismatch_numeric(x, y, diff)
    difference(format(mismatches, max_diffs = max_diffs))
  }
}

#' Default numeric tolerance
#'
#' testthat's default numeric tolerance is `r testthat_tolerance()`.
#'
#' @keywords internal
#' @export
testthat_tolerance <- function() {
  if (identical(capabilities("long.double"), FALSE)) {
    skip("Long doubles not available and `tolerance` not supplied")
  }

  .Machine$double.eps^0.5
}

mismatch_numeric <- function(x, y, diff = !vector_equal(x, y)) {
  structure(
    list(
      i = which(diff),
      x = x[diff],
      y = y[diff],
      n = length(diff),
      n_diff = sum(diff),
      mu_diff = mean(abs(x[diff] - y[diff]), na.rm = TRUE)
    ),
    class = "mismatch_numeric"
  )
}

#' @export
format.mismatch_numeric <- function(x, ..., max_diffs = 9, digits = 3) {
  summary <- paste0(x$n_diff, "/", x$n, " mismatches")
  if (x$n_diff > 1) {
    mu <- format(x$mu_diff, digits = digits, trim = TRUE)
    summary <- paste0(summary, " (average diff: ", mu, ")")
  }

  n_show <- seq_len(min(x$n_diff, max_diffs))

  diffs <- paste0(
    format(paste0("[", x$i[n_show], "]")),
    " ",
    format(x$x[n_show], digits = digits),
    " - ",
    format(x$y[n_show], digits = digits),
    " == ",
    format(x$x[n_show] - x$y[n_show], digits = digits)
  )

  if (x$n_diff > length(n_show)) {
    diffs <- c(diffs, "...")
  }

  paste0(summary, "\n", paste(diffs, collapse = "\n"))
}

#' @export
print.mismatch_numeric <- function(x, ...) {
  cat(format(x, ...), "\n", sep = "")
}

# compare.time ------------------------------------------------------------

#' @rdname compare
#' @export
compare.POSIXt <- function(x, y, tolerance = 0.001, ..., max_diffs = 9) {
  if (!inherits(y, "POSIXt")) {
    return(diff_class(x, y))
  }
  if (!same_length(x, y)) {
    return(diff_length(x, y))
  }

  x <- standardise_tzone(as.POSIXct(x))
  y <- standardise_tzone(as.POSIXct(y))

  if (!same_attr(x, y)) {
    return(diff_attr(x, y))
  }

  diff <- !vector_equal_tol(x, y, tolerance = tolerance)

  if (!any(diff)) {
    no_difference()
  } else {
    mismatches <- mismatch_numeric(x, y, diff)
    difference(format(mismatches, max_diffs = max_diffs))
  }
}

standardise_tzone <- function(x) {
  if (
    is.null(attr(x, "tzone")) || identical(attr(x, "tzone"), Sys.timezone())
  ) {
    attr(x, "tzone") <- ""
  }

  x
}

test_that("deprecated in 3rd edition", {
  expect_warning(local_mock(), "deprecated")
  expect_warning(with_mock(is_testing = function() FALSE), "deprecated")
})

test_that("can change value of internal function", {
  local_edition(2)

  with_mock(
    test_mock2 = function() 5,
    expect_equal(test_mock1(), 5)
  )

  # and value is restored on error
  expect_error(
    with_mock(
      test_mock2 = function() 5,
      stop("!")
    )
  )
  expect_equal(test_mock1(), 10)
})


test_that("mocks can access local variables", {
  local_edition(2)
  x <- 5

  with_mock(
    test_mock2 = function() x,
    expect_equal(test_mock1(), 5)
  )
})

test_that("non-empty mock with return value", {
  local_edition(2)
  expect_true(with_mock(
    compare = function(x, y, ...) list(equal = TRUE, message = "TRUE"),
    TRUE
  ))
})

test_that("nested mock", {
  local_edition(2)
  with_mock(
    all.equal = function(x, y, ...) TRUE,
    {
      with_mock(
        expect_warning = expect_error,
        {
          expect_warning(stopifnot(!compare(3, "a")$equal))
        }
      )
    },
    .env = asNamespace("base")
  )
  expect_false(isTRUE(all.equal(3, 5)))
  expect_warning(warning("test"))
})

test_that("can't mock non-existing", {
  local_edition(2)
  expect_error(with_mock(..bogus.. = identity, TRUE), "Function [.][.]bogus[.][.] not found in environment testthat")
})

test_that("can't mock non-function", {
  local_edition(2)
  expect_error(with_mock(pkg_and_name_rx = FALSE, TRUE), "Function pkg_and_name_rx not found in environment testthat")
})

test_that("empty or no-op mock", {
  local_edition(2)
  expect_warning(
    expect_null(with_mock()),
    "Not mocking anything. Please use named parameters to specify the functions you want to mock.",
    fixed = TRUE
  )
  expect_warning(
    expect_true(with_mock(TRUE)),
    "Not mocking anything. Please use named parameters to specify the functions you want to mock.",
    fixed = TRUE
  )
})

test_that("visibility", {
  local_edition(2)
  expect_warning(expect_false(withVisible(with_mock())$visible))
  expect_true(withVisible(with_mock(compare = function() {}, TRUE))$visible)
  expect_false(withVisible(with_mock(compare = function() {}, invisible(5)))$visible)
})

test_that("multiple return values", {
  local_edition(2)
  expect_true(with_mock(FALSE, TRUE, compare = function() {}))
  expect_equal(with_mock(3, compare = function() {}, 5), 5)
})

test_that("can access variables defined in function", {
  local_edition(2)
  x <- 5
  expect_equal(with_mock(x, compare = function() {}), 5)
})

test_that("can mock if package is not loaded", {
  local_edition(2)
  if ("package:curl" %in% search()) {
    skip("curl is loaded")
  }
  skip_if_not_installed("curl")
  with_mock(`curl::curl` = identity, expect_identical(curl::curl, identity))
})

test_that("changes to variables are preserved between calls and visible outside", {
  local_edition(2)
  x <- 1
  with_mock(
    show_menu = function() {},
    x <- 3,
    expect_equal(x, 3)
  )
  expect_equal(x, 3)
})

test_that("mock extraction", {
  local_edition(2)
  expect_identical(
    extract_mocks(list(compare = compare), .env = asNamespace("testthat"))$compare$name,
    as.name("compare")
  )
  expect_error(
    extract_mocks(list(..bogus.. = identity), "testthat"),
    "Function [.][.]bogus[.][.] not found in environment testthat"
  )
  expect_equal(
    length(extract_mocks(list(not = identity, show_menu = identity), "testthat")),
    2
  )
})
# local_mock --------------------------------------------------------------

test_that("local_mock operates locally", {
  local_edition(2)
  f <- function() {
    local_mock(compare = function(x, y) FALSE)
    compare(1, 1)
  }

  expect_false(f())
  expect_equal(compare(1, 1), no_difference())
})

test_that("list comparison truncates to max_diffs", {
  x <- as.list(as.character(1:1e3))
  y <- lapply(x, paste0, ".")

  lines1 <- strsplit(compare(x, y)$message, "\n")[[1]]
  expect_length(lines1, 10)

  lines2 <- strsplit(compare(x, y, max_diffs = 99)$message, "\n")[[1]]
  expect_length(lines2, 100)
})

test_that("no diff", {
  expect_equal(compare(1, 1), no_difference())
})

test_that("vector_equal_tol handles infinity", {
  expect_true(vector_equal_tol(Inf, Inf))
  expect_true(vector_equal_tol(-Inf, -Inf))
  expect_false(vector_equal_tol(Inf, -Inf))
  expect_false(vector_equal_tol(Inf, 0))
})

test_that("vector_equal_tol handles na", {
  expect_true(vector_equal_tol(NA, NA))
  expect_false(vector_equal_tol(NA, 0))
})


# character ---------------------------------------------------------------

test_that("types must be the same", {
  expect_match(compare("a", 1L)$message, "character is not integer")
})

test_that("base lengths must be identical", {
  expect_match(compare("a", letters)$message, "1 is not 26")
})

test_that("classes must be identical", {
  c1 <- "a"
  c2 <- structure("a", class = "mycharacter")

  expect_match(compare(c1, c2)$message, "\"character\" is not \"mycharacter\"")
})

test_that("attributes must be identical", {
  x1 <- "a"
  x2 <- c(a = "a")
  x3 <- c(b = "a")
  x4 <- structure("a", a = 1)
  x5 <- structure("a", b = 1)

  expect_match(compare(x1, x2)$message, "names for current")
  expect_match(compare(x2, x3)$message, "Names: 1 string mismatch")

  expect_match(compare(x1, x4)$message, "target is NULL")
  expect_match(compare(x4, x5)$message, "Names: 1 string mismatch")
})

test_that("two identical vectors are the same", {
  expect_true(compare(letters, letters)$equal)
})

test_that("equal if both missing or both the same (multiple values)", {
  expect_true(compare(c("ABC", NA), c("ABC", NA))$equal)

  expect_false(compare(c(NA, NA), c("ABC", NA))$equal)
  expect_false(compare(c("AB", NA), c("ABC", NA))$equal)
  expect_false(compare(c("AB", "AB"), c("ABC", "AB"))$equal)
})

test_that("computes correct number of mismatches", {
  x <- mismatch_character(c("a", "b", "c"), c("c", "d", "e"))
  expect_equal(x$n, 3)
})

test_that("only differences are shown", {
  x <- mismatch_character(letters, c(letters[-26], "a"))

  lines <- strsplit(format(x), "\n")[[1]]
  expect_equal(lines[1], "1/26 mismatches")
  expect_equal(lines[2], 'x[26]: "z"')
})

test_that("not all lines are shown", {
  a <- "1234567890"
  b <- paste(rep(a, 10), collapse = "")

  x <- mismatch_character(a, b)
  lines <- strsplit(format(x, width = 16), "\n")[[1]]

  expect_equal(lines[1], "1/1 mismatches")
  expect_equal(length(lines), 8)
})

test_that("vectors longer than `max_diffs` (#513)", {
  comp <- compare(letters[1:2], LETTERS[1:2], max_diffs = 1)
  expect_s3_class(comp, "comparison")
  expect_false(comp$equal)
  expect_equal(comp$message, "2/2 mismatches\nx[1]: \"a\"\ny[1]: \"A\"")
})

# numeric ------------------------------------------------------------------

test_that("numeric types are compatible", {
  expect_true(compare(1, 1L)$equal)
  expect_true(compare(1L, 1)$equal)
})

test_that("non-numeric types are not compatible", {
  expect_match(compare(1, "a")$message, "double is not character")
})

test_that("base lengths must be identical", {
  expect_match(compare(1, c(1, 2))$message, "1 is not 2")
})

test_that("classes must be identical", {
  f1 <- factor("a")
  f2 <- factor("a", ordered = TRUE)

  expect_match(compare(1L, f1)$message, "\"integer\" is not \"factor\"")
  expect_match(
    compare(1L, f2)$message,
    "\"integer\" is not \"ordered\"/\"factor\""
  )
})

test_that("attributes must be identical", {
  x1 <- 1L
  x2 <- c(a = 1L)
  x3 <- c(b = 1L)
  x4 <- structure(1L, a = 1)
  x5 <- structure(1L, b = 1)

  expect_match(compare(x1, x2)$message, "names for current")
  expect_match(compare(x2, x3)$message, "Names: 1 string mismatch")

  expect_match(compare(x1, x4)$message, "target is NULL")
  expect_match(compare(x4, x5)$message, "Names: 1 string mismatch")
})

test_that("unless check.attributes is FALSE", {
  x1 <- 1L
  x2 <- c(a = 1L)
  x3 <- structure(1L, a = 1)

  expect_equal(compare(x1, x2, check.attributes = FALSE)$message, "Equal")
  expect_equal(compare(x1, x3, check.attributes = FALSE)$message, "Equal")
  expect_equal(compare(x2, x3, check.attributes = FALSE)$message, "Equal")
})

test_that("two identical vectors are the same", {
  expect_true(compare(1:10, 1:10)$equal)
})

test_that("named arguments to all.equal passed through", {
  expect_equal(415, 416, tolerance = 0.01)
})

test_that("tolerance used for individual comparisons", {
  x1 <- 1:3
  x2 <- x1 + c(0, 0, 0.1)

  expect_false(compare(x1, x2)$equal)
  expect_true(compare(x1, x2, tolerance = 0.1)$equal)
})

test_that("mismatch_numeric truncates diffs", {
  x <- mismatch_numeric(1:11, 11:1)
  expect_equal(x$n, 11)
  expect_equal(x$n_diff, 10)

  lines <- strsplit(format(x, max_diffs = 5), "\n")[[1]]
  expect_equal(length(lines), 5 + 2)
})

# time --------------------------------------------------------------------

test_that("both POSIXt classes are compatible", {
  x1 <- Sys.time()
  x2 <- as.POSIXlt(x1)
  expect_true(compare(x1, x2)$equal)
  expect_true(compare(x2, x1)$equal)
})

test_that("other classes are not", {
  expect_match(
    compare(Sys.time(), 1)$message,
    "\"POSIXct\"/\"POSIXt\" is not \"numeric\""
  )
})

test_that("base lengths must be identical", {
  x1 <- Sys.time()
  x2 <- c(x1, x1 - 3600)

  expect_match(compare(x1, x2)$message, "1 is not 2")
})

test_that("tzones must be identical", {
  # skip on minimal setups
  tryCatch(
    {
      t1 <- ISOdatetime(2016, 2, 29, 12, 13, 14, "EST")
      t2 <- ISOdatetime(2016, 2, 29, 12, 13, 14, "US/Eastern")
    },
    warning = function(w) skip(conditionMessage(w))
  )

  expect_match(compare(t1, t2)$message, '"tzone": 1 string mismatch')
})

test_that("two identical vectors are the same", {
  x <- Sys.time()
  expect_true(compare(x, x)$equal)
})

test_that("two different values are not the same", {
  x1 <- Sys.time()
  x2 <- x1 + 3600
  expect_false(compare(x1, x2)$equal)
})

test_that("uses all.equal tolerance", {
  x1 <- structure(1457284588.83749, class = c("POSIXct", "POSIXt"))
  x2 <- structure(1457284588.837, class = c("POSIXct", "POSIXt"))
  expect_true(compare(x1, x2)$equal)
})

#' Do you expect every value in a vector to have this value?
#'
#' These expectations are similar to `expect_true(all(x == "x"))`,
#' `expect_true(all(x))` and `expect_true(all(!x))` but give more informative
#' failure messages if the expectations are not met.
#'
#' @inheritParams expect_equal
#' @export
#' @examples
#' x1 <- c(1, 1, 1, 1, 1, 1)
#' expect_all_equal(x1, 1)
#'
#' x2 <- c(1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2)
#' show_failure(expect_all_equal(x2, 1))
#'
#' # expect_all_true() and expect_all_false() are helpers for common cases
#' set.seed(1016)
#' show_failure(expect_all_true(rpois(100, 10) < 20))
#' show_failure(expect_all_false(rpois(100, 10) > 20))
expect_all_equal <- function(object, expected) {
  act <- quasi_label(enquo(object))
  exp <- quasi_label(enquo(expected))

  expect_all_equal_(act, exp)
  invisible(act$val)
}

#' @export
#' @rdname expect_all_equal
expect_all_true <- function(object) {
  act <- quasi_label(enquo(object))
  exp <- labelled_value(TRUE, "TRUE")

  expect_all_equal_(act, exp)
  invisible(act$val)
}

#' @export
#' @rdname expect_all_equal
expect_all_false <- function(object) {
  act <- quasi_label(enquo(object))
  exp <- labelled_value(FALSE, "FALSE")

  expect_all_equal_(act, exp)
  invisible(act$val)
}


expect_all_equal_ <- function(act, exp, trace_env = caller_env()) {
  check_vector(act$val, error_call = trace_env, error_arg = "object")
  if (length(act$val) == 0) {
    cli::cli_abort("{.arg object} must not be empty.", call = trace_env)
  }

  check_vector(exp$val, error_call = trace_env, error_arg = "expected")
  if (length(exp$val) != 1) {
    cli::cli_abort("{.arg expected} must be length 1.", call = trace_env)
  }

  exp$val <- rep(exp$val, length(act$val))
  names(exp$val) <- names(act$val)
  expect_waldo_equal_(
    "Expected every element of %s to equal %s.",
    act,
    exp,
    tolerance = testthat_tolerance(),
    trace_env = trace_env
  )
}

rlang_version <- function() {
  if (use_rlang_1_0()) "rlang-1.0" else "rlang-pre-1.0"
}

local_use_rlang_1_0 <- function(frame = caller_env()) {
  local_options("testthat:::rlang_dep" = TRUE, .frame = frame)
}

#' Mocking tools
#'
#' @description
#' `with_mocked_bindings()` and `local_mocked_bindings()` provide tools for
#' "mocking", temporarily redefining a function so that it behaves differently
#' during tests. This is helpful for testing functions that depend on external
#' state (i.e. reading a value from a file or a website, or pretending a package
#' is or isn't installed).
#'
#' These functions represent a second attempt at bringing mocking to testthat,
#' incorporating what we've learned from the mockr, mockery, and mockthat
#' packages.
#'
#' # Use
#'
#' There are four places that the function you are trying to mock might
#' come from:
#'
#' * Internal to your package.
#' * Imported from an external package via the `NAMESPACE`.
#' * The base environment.
#' * Called from an external package with `::`.
#'
#' They are described in turn below.
#'
#' ## Internal & imported functions
#'
#' You mock internal and imported functions the same way. For example, take
#' this code:
#'
#' ```R
#' some_function <- function() {
#'   another_function()
#' }
#' ```
#'
#' It doesn't matter whether `another_function()` is defined by your package
#' or you've imported it from a dependency with `@import` or `@importFrom`,
#' you mock it the same way:
#'
#' ```R
#' local_mocked_bindings(
#'   another_function = function(...) "new_value"
#' )
#' ```
#'
#' ## Base functions
#'
#' To mock a function in the base package, you need to make sure that you
#' have a binding for this function in your package. It's easiest to do this
#' by binding the value to `NULL`. For example, if you wanted to mock
#' `interactive()` in your package, you'd need to include this code somewhere
#' in your package:
#'
#' ```R
#' interactive <- NULL
#' ```
#'
#' Why is this necessary? `with_mocked_bindings()` and `local_mocked_bindings()`
#' work by temporarily modifying the bindings within your package's namespace.
#' When these tests are running inside of `R CMD check` the namespace is locked
#' which means it's not possible to create new bindings so you need to make sure
#' that the binding exists already.
#'
#' ## Namespaced calls
#'
#' It's trickier to mock functions in other packages that you call with `::`.
#' For example, take this minor variation:
#'
#' ```R
#' some_function <- function() {
#'   anotherpackage::another_function()
#' }
#' ```
#'
#' To mock this function, you'd need to modify `another_function()` inside the
#' `anotherpackage` package. You _can_ do this by supplying the `.package`
#' argument to `local_mocked_bindings()` but we don't recommend it because
#' it will affect all calls to `anotherpackage::another_function()`, not just
#' the calls originating in your package. Instead, it's safer to either import
#' the function into your package, or make a wrapper that you can mock:
#'
#' ```R
#' some_function <- function() {
#'   my_wrapper()
#' }
#' my_wrapper <- function(...) {
#'   anotherpackage::another_function(...)
#' }
#'
#' local_mocked_bindings(
#'   my_wrapper = function(...) "new_value"
#' )
#' ```
#' @export
#' @param ... Name-value pairs providing new values (typically functions) to
#'   temporarily replace the named bindings.
#' @param code Code to execute with specified bindings.
#' @param .env Environment that defines effect scope. For expert use only.
#' @param .package The name of the package where mocked functions should be
#'   inserted. Generally, you should not supply this as it will be automatically
#'   detected when whole package tests are run or when there's one package
#'   under active development (i.e. loaded with [pkgload::load_all()]).
#'   We don't recommend using this to mock functions in other packages,
#'   as you should not modify namespaces that you don't own.
local_mocked_bindings <- function(..., .package = NULL, .env = caller_env()) {
  bindings <- list2(...)
  check_bindings(bindings)

  .package <- .package %||% dev_package()
  ns_env <- ns_env(.package)

  # Rebind in namespace, imports, and the global environment
  envs <- list(ns_env, env_parent(ns_env), globalenv())
  bindings_found <- rep_named(names(bindings), FALSE)
  for (env in envs) {
    local_bindings_rebind(!!!bindings, .env = env, .frame = .env)
    bindings_found <- bindings_found | env_has(env, names(bindings))
  }

  # And mock S3 methods
  methods_env <- ns_env[[".__S3MethodsTable__."]]
  local_bindings_rebind(!!!bindings, .env = methods_env, .frame = .env)

  # If needed, also mock in the package environment so we can call directly
  if (is_attached(paste0("package:", .package))) {
    local_bindings_rebind(!!!bindings, .env = pkg_env(.package), .frame = .env)
  }
  # And in the current testing environment
  test_env <- the$testing_env
  if (!is.null(test_env)) {
    local_bindings_rebind(!!!bindings, .env = test_env, .frame = .env)
  }

  if (any(!bindings_found)) {
    missing <- names(bindings)[!bindings_found]
    cli::cli_abort("Can't find binding for {.arg {missing}}")
  }

  invisible()
}

#' @rdname local_mocked_bindings
#' @export
with_mocked_bindings <- function(code, ..., .package = NULL) {
  local_mocked_bindings(..., .package = .package)
  code
}

local_testing_env <- function(env, frame = caller_env()) {
  local_bindings(testing_env = env, .env = the, .frame = frame)
}

# helpers -----------------------------------------------------------------

# Wrapper around local_bindings() that only rebinds existing values,
# automatically unlocking as needed. We can only rebind because most of
# these environments are locked, meaning we can't add new bindings.
local_bindings_rebind <- function(..., .env = .frame, .frame = caller_env()) {
  bindings <- list2(...)
  bindings <- bindings[env_has(.env, names(bindings))]

  if (length(bindings) == 0) {
    return()
  }

  nms <- names(bindings)
  locked <- env_binding_unlock(.env, nms)
  withr::defer(env_binding_lock(.env, nms[locked]), envir = .frame)

  local_bindings(!!!bindings, .env = .env, .frame = .frame)

  invisible()
}

dev_package <- function() {
  if (is_testing() && testing_package() != "") {
    testing_package()
  } else {
    loaded <- loadedNamespaces()
    is_dev <- map_lgl(loaded, function(x) !is.null(pkgload::dev_meta(x)))
    if (sum(is_dev) == 0) {
      cli::cli_abort("No packages loaded with pkgload")
    } else if (sum(is_dev) == 1) {
      loaded[is_dev]
    } else {
      cli::cli_abort("Multiple packages loaded with pkgload")
    }
  }
}

check_bindings <- function(x, error_call = caller_env()) {
  if (!is_named(x)) {
    cli::cli_abort(
      "All elements of {.arg ...} must be named.",
      call = error_call
    )
  }
}

# For testing -------------------------------------------------------------

test_mock_direct <- function() {
  "y"
}

test_mock_internal <- function() {
  test_mock_internal2()
}
test_mock_internal2 <- function() "y"

test_mock_imports <- function() {
  as.character(sym("y"))
}

test_mock_namespaced <- function() {
  as.character(rlang::sym("y"))
}

test_mock_method <- function(x) {
  UseMethod("test_mock_method")
}
#' @export
test_mock_method.integer <- function(x) {
  "y"
}

test_mock_base <- function() {
  interactive()
}
interactive <- NULL

show_bindings <- function(name, env = caller_env()) {
  envs <- env_parents(env)
  has_binding <- Filter(function(env) env_has(env, name), envs)
  lapply(has_binding, env_desc)
  invisible()
}

test_mock_value <- 10

env_desc <- function(env) {
  cat(obj_address(env), ": ", env_name(env), "\n", sep = "")
}

test_that("produces consistent output", {
  withr::local_options(testthat.edition_ignore = TRUE)
  local_edition(2)
  local_mocked_bindings(
    show_menu = function(choices, title = NULL) {
      cat(paste0(format(seq_along(choices)), ": ", choices, sep = "\n"), "\n", sep = "")
      0L
    },
    sink_number = function() 0L
  )
  withr::local_options(testthat_format_srcrefs = FALSE)
  expect_snapshot_reporter(DebugReporter$new())
})

get_vars_from_debug_reporter <- function(choice, fun, envir = parent.frame()) {
  frame <- get_frame_from_debug_reporter(choice, fun, envir)
  ls(frame)
}

get_frame_from_debug_reporter <- function(choice, fun, envir = parent.frame()) {
  local_edition(2)
  force(choice)
  test_debug_reporter_parent_frame <- NULL

  with_mocked_bindings(
    show_menu = function(choices, title = NULL) {
      # if (choice > 0) print(choices)
      my_choice <- choice
      choice <<- 0L
      my_choice
    },
    browse_frame = function(frame, skip) {
      test_debug_reporter_parent_frame <<- frame
    },
    sink_number = function() 0L,
    with_reporter(
      "debug",
      test_that("debug_reporter_test", { fun() })
    )
  )

  test_debug_reporter_parent_frame
}

success_fun <- function() {
  aa <- 1
  expect_true(TRUE)
}

test_that("debug reporter is not called for successes", {
  expect_null(get_frame_from_debug_reporter(2, success_fun))
})

test_that("browser() is called for the correct frame for failures", {
  fun_1 <- function() {
    aa <- 1
    expect_true(FALSE)
  }

  fun_2 <- function() {
    f <- function() expect_true(FALSE)
    f()
  }

  fun_3 <- function() {
    f <- function() {
      g <- function() expect_true(FALSE)
      g()
    }
    f()
  }

  expect_equal(get_vars_from_debug_reporter(1, fun_1), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_1), "aa")

  expect_equal(get_vars_from_debug_reporter(1, fun_2), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_2), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_2), character())

  expect_equal(get_vars_from_debug_reporter(1, fun_3), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_3), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_3), "g")
  expect_equal(get_vars_from_debug_reporter(4, fun_3), character())
})

test_that("browser() is called for the correct frame for warnings", {
  fun_1 <- function() {
    aa <- 1
    warning("warn")
  }

  fun_2 <- function() {
    f <- function() warning("warn")
    f()
  }

  fun_3 <- function() {
    f <- function() {
      g <- function() warning("warn")
      g()
    }
    f()
  }

  expect_equal(get_vars_from_debug_reporter(1, fun_1), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_1), "aa")

  expect_equal(get_vars_from_debug_reporter(1, fun_2), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_2), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_2), character())

  expect_equal(get_vars_from_debug_reporter(1, fun_3), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_3), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_3), "g")
  expect_equal(get_vars_from_debug_reporter(4, fun_3), character())
})

test_that("browser() is called for the correct frame for errors", {
  fun_1 <- function() {
    aa <- 1
    stop("error")
  }

  fun_2 <- function() {
    f <- function() stop("error")
    f()
  }

  fun_3 <- function() {
    f <- function() {
      g <- function() stop("error")
      g()
    }
    f()
  }

  expect_equal(get_vars_from_debug_reporter(1, fun_1), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_1), "aa")

  expect_equal(get_vars_from_debug_reporter(1, fun_2), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_2), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_2), character())

  expect_equal(get_vars_from_debug_reporter(1, fun_3), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_3), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_3), "g")
  expect_equal(get_vars_from_debug_reporter(4, fun_3), character())
})

test_that("browser() is called for the correct frame for skips", {
  fun_1 <- function() {
    aa <- 1
    skip("skip")
  }

  fun_2 <- function() {
    f <- function() skip("skip")
    f()
  }

  fun_3 <- function() {
    f <- function() {
      g <- function() skip("skip")
      g()
    }
    f()
  }

  expect_equal(get_vars_from_debug_reporter(1, fun_1), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_1), "aa")

  expect_equal(get_vars_from_debug_reporter(1, fun_2), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_2), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_2), character())

  expect_equal(get_vars_from_debug_reporter(1, fun_3), character())
  expect_equal(get_vars_from_debug_reporter(2, fun_3), "f")
  expect_equal(get_vars_from_debug_reporter(3, fun_3), "g")
  expect_equal(get_vars_from_debug_reporter(4, fun_3), character())
})

test_that("with_mocked_bindings affects local bindings", {
  out <- with_mocked_bindings(
    test_mock_internal(),
    test_mock_internal2 = function() "x"
  )
  expect_equal(out, "x")
  expect_equal(test_mock_internal(), "y")
})

test_that("local_mocked_bindings affects local bindings", {
  local({
    local_mocked_bindings(test_mock_internal = function() "x")
    expect_equal(test_mock_internal(), "x")
  })

  expect_equal(test_mock_internal(), "y")
})

test_that("unlocks and relocks binding if needed", {
  ns_env <- ns_env("testthat")
  expect_true(env_binding_are_locked(ns_env, "test_mock_direct"))

  local({
    local_mocked_bindings(test_mock_direct = function(...) "x")
    expect_false(env_binding_are_locked(ns_env, "test_mock_direct"))
  })

  expect_true(env_binding_are_locked(ns_env, "test_mock_direct"))
})

test_that("can make wrapper", {
  local_mock_x <- function(env = caller_env()) {
    local_mocked_bindings(test_mock_internal2 = function() "x", .env = env)
  }

  local({
    local_mock_x()
    expect_equal(test_mock_internal(), "x")
  })

  expect_equal(test_mock_internal(), "y")
})

test_that("with_mocked_bindings() validates its inputs", {
  expect_snapshot(error = TRUE, {
    with_mocked_bindings(1 + 1, function() 2)
  })
})

# -------------------------------------------------------------------------

test_that("can mock directly", {
  local_mocked_bindings(test_mock_direct = function(...) "x")
  expect_equal(test_mock_direct(), "x")
})

test_that("can mock bindings from imports", {
  local_mocked_bindings(sym = function(...) "x")
  expect_equal(test_mock_imports(), "x")
})

test_that("can mock bindings in another package", {
  local_mocked_bindings(sym = function(...) "x", .package = "rlang")
  expect_equal(test_mock_namespaced(), "x")
})

test_that("can mock S3 methods", {
  local({
    local_mocked_bindings(test_mock_method.integer = function(...) "x")
    expect_equal(test_mock_method(1L), "x")
  })
  expect_equal(test_mock_method(1L), "y")
})

test_that("can't mock bindings that don't exist", {
  expect_snapshot(local_mocked_bindings(f = function() "x"), error = TRUE)
})

test_that("can mock base functions with in-package bindings", {
  local_mocked_bindings(interactive = function() TRUE)
  expect_equal(test_mock_base(), TRUE)
})

test_that("can mock values", {
  local_mocked_bindings(test_mock_value = 100)
  expect_equal(test_mock_value, 100)
})

#' Get and set active reporter.
#'
#' `get_reporter()` and `set_reporter()` access and modify the current "active"
#' reporter. Generally, these functions should not be called directly; instead
#' use `with_reporter()` to temporarily change, then reset, the active reporter.
#'
#'
#' @param reporter Reporter to use to summarise output. Can be supplied
#'   as a string (e.g. "summary") or as an R6 object
#'   (e.g. `SummaryReporter$new()`).
#'
#'   See [Reporter] for more details and a list of built-in reporters.
#' @param code Code to execute.
#' @return `with_reporter()` invisible returns the reporter active when `code`
#'   was evaluated.
#' @param start_end_reporter Should the reporters `start_reporter()` and
#'   `end_reporter()` methods be called? For expert use only.
#' @keywords internal
#' @name reporter-accessors
NULL

#' @rdname reporter-accessors
#' @export
set_reporter <- function(reporter) {
  old <- the$reporter
  the$reporter <- reporter
  invisible(old)
}

#' @rdname reporter-accessors
#' @export
get_reporter <- function() {
  the$reporter
}

#' @rdname reporter-accessors
#' @export
with_reporter <- function(reporter, code, start_end_reporter = TRUE) {
  # Ensure we don't propagate the local description to the new reporter
  local_description_set()
  reporter <- find_reporter(reporter)

  old <- set_reporter(reporter)
  withr::defer(set_reporter(old))

  if (start_end_reporter) {
    reporter$start_reporter()
  }

  tryCatch(code, testthat_abort_reporter = function(cnd) {
    cat(conditionMessage(cnd), "\n")
    NULL
  })

  if (start_end_reporter) {
    reporter$end_reporter()
  }

  invisible(reporter)
}

stop_reporter <- function(message) {
  cli::cli_abort(
    message,
    class = "testthat_abort_reporter",
    call = NULL
  )
}

#' Find reporter object given name or object.
#'
#' If not found, will return informative error message.
#' Pass a character vector to create a [MultiReporter] composed
#' of individual reporters.
#' Will return null if given NULL.
#'
#' @param reporter name of reporter(s), or reporter object(s)
#' @keywords internal
find_reporter <- function(reporter) {
  if (is.null(reporter)) {
    return(NULL)
  }

  if (inherits(reporter, "R6ClassGenerator")) {
    reporter$new()
  } else if (inherits(reporter, "Reporter")) {
    reporter
  } else if (is_string(reporter)) {
    find_reporter_one(reporter)
  } else if (is.character(reporter)) {
    reporters <- lapply(reporter, find_reporter_one, call = current_env())
    MultiReporter$new(reporters)
  } else {
    stop_input_type(
      reporter,
      c(
        "a string",
        "a character vector",
        "a reporter object",
        "a reporter class"
      )
    )
  }
}

find_reporter_one <- function(reporter, ..., call = caller_env()) {
  name <- reporter
  substr(name, 1, 1) <- toupper(substr(name, 1, 1))
  name <- paste0(name, "Reporter")

  if (!exists(name)) {
    cli::cli_abort("Cannot find test reporter {.arg {reporter}}.", call = call)
  }

  get(name)$new(...)
}

test_that("can locate reporter from name", {
  expect_equal(find_reporter("minimal"), MinimalReporter$new())
  expect_equal(find_reporter("summary"), SummaryReporter$new())
})

test_that("useful error message if can't find reporter", {
  expect_error(
    find_reporter(c("summary", "blah")),
    "Can not find test reporter blah"
  )
})

test_that("character vector yields multi reporter", {
  expect_equal(
    find_reporter(c("summary", "stop")),
    MultiReporter$new(
      reporters = list(
        SummaryReporter$new(),
        StopReporter$new()
      )
    )
  )
})

test_that("multiplication works", {
  n_tests <- function(n) {
    for (i in 1:n) {
      test_that(paste0("run ", i), {
        suceed()
      })
    }
  }
  expect_snapshot(
    {
      show_all <- SlowReporter$new(min_time = 0)
      with_reporter(show_all, n_tests(10))
    },
    transform = \(x) gsub("\\d.\\d\\ds", "-.--s", x),
  )
})

srcref_line <- function(code) {
  srcref <- attr(substitute(code), "srcref")
  if (!is.list(srcref)) {
    stop("code doesn't have srcref", call. = FALSE)
  }

  results <- with_reporter("silent", code)$expectations()
  unlist(lapply(results, function(x) x$srcref[1])) - srcref[[1]][1]
}

test_that("line numbers captured for expectations and warnings", {
  f <- function() warning("Uh oh")
  lines <- srcref_line({
    test_that("simple", {        # line 1
      expect_true(FALSE)         # line 2
      f()                        # line 3
    })
  })
  expect_equal(lines, c(2, 3))
})

test_that("line numbers captured when called indirectly", {
  lines <- srcref_line({
    test_that("simple", {                # line 1
      f <- function() g()                # line 2
      g <- function() h()                # line 3
      h <- function() expect_true(FALSE) # line 4
                                         # line 5
      h()                                # line 6
    })
  })
  expect_equal(lines, 4)

  lines <- srcref_line({
    f <- function() g()                  # line 1
    g <- function() h()                  # line 2
    h <- function() expect_true(FALSE)   # line 3
    test_that("simple", {                # line 4
      h()                                # line 5
    })
  })
  expect_equal(lines, 5)
})

test_that("line numbers captured inside a loop", {
  lines <- srcref_line({
    test_that("simple", {               # line 1
      for (i in 1:4) expect_true(TRUE)  # line 2
    })
  })
  expect_equal(lines, rep(2, 4))
})

test_that("line numbers captured for skip()s and stops()", {
  lines <- srcref_line({
    test_that("simple", {             # line 1
      skip("Not this time")           # line 2
    })                                # line 3
  })
  expect_equal(lines, 2)

  lines <- srcref_line({
    test_that("simple", {             # line 1
      stop("Not this time")           # line 2
    })                                # line 3
  })
  expect_equal(lines, 2)
})

test_that("line numbers captured for on.exit()", {
  lines <- srcref_line({
    test_that("simple", {             # line 1
      on.exit({stop("Error")})        # line 2
    })                                # line 3
  })
  expect_equal(lines, 2)

  # Falls back to test if no srcrf
  lines <- srcref_line({
    test_that("simple", {             # line 1
      on.exit(stop("Error"))          # line 2
    })                                # line 3
  })
  expect_equal(lines, 1)
})

#' Test reporter: gather all errors silently.
#'
#' This reporter quietly runs all tests, simply gathering all expectations.
#' This is helpful for programmatically inspecting errors after a test run.
#' You can retrieve the results with the `expectations()`
#' method.
#'
#' @export
#' @family reporters
SilentReporter <- R6::R6Class("SilentReporter",
  inherit = Reporter,
  public = list(
    .expectations = NULL,

    initialize = function(...) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
      self$.expectations <- Stack$new()
    },

    add_result = function(context, test, result) {
      self$.expectations$push(result)
    },

    expectations = function() {
      self$.expectations$as_list()
    }
  )
)

#' Test reporter: RStudio
#'
#' This reporter is designed for output to RStudio. It produces results in
#' any easily parsed form.
#'
#' @export
#' @family reporters
RStudioReporter <- R6::R6Class("RStudioReporter",
  inherit = Reporter,
  public = list(
    initialize = function(...) {
      self$capabilities$parallel_support <- TRUE
      super$initialize(...)
    },

    add_result = function(context, test, result) {
      if (expectation_success(result)) {
        return()
      }

      loc <- expectation_location(result)
      status <- expectation_type(result)
      first_line <- strsplit(result$message, "\n")[[1]][1]

      self$cat_line(loc, " [", status, "] ", test, ". ", first_line)
    }
  )
)

otel_tracer_name <- "org.r-lib.testthat"

# generic otel helpers ---------------------------------------------------------

otel_cache_tracer <- NULL
otel_local_test_span <- NULL
otel_update_span <- NULL

local({
  otel_is_tracing <- FALSE
  otel_tracer <- NULL

  otel_cache_tracer <<- function() {
    requireNamespace("otel", quietly = TRUE) || return()
    otel_tracer <<- otel::get_tracer(otel_tracer_name)
    otel_is_tracing <<- tracer_enabled(otel_tracer)
  }

  otel_local_test_span <<- function(name, scope = parent.frame()) {
    otel_is_tracing || return()
    otel::start_local_active_span(
      sprintf("test that %s", name),
      tracer = otel_tracer,
      activation_scope = scope
    )
  }

  otel_update_span <<- function(
    span,
    n_success,
    n_failure,
    n_error,
    n_skip,
    n_warning
  ) {
    otel_is_tracing || return()

    total <- n_success + n_failure + n_error + n_skip + n_warning
    test_status <- if (n_error > 0) {
      "error"
    } else if (n_failure > 0) {
      "fail"
    } else if (total == 0 || n_skip == total) {
      "skip"
    } else {
      "pass"
    }
    span$set_attribute("test.expectations.total", total)
    span$set_attribute("test.expectations.passed", n_success)
    span$set_attribute("test.expectations.failed", n_failure)
    span$set_attribute("test.expectations.error", n_error)
    span$set_attribute("test.expectations.skipped", n_skip)
    span$set_attribute("test.expectations.warning", n_warning)
    span$set_attribute("test.status", test_status)

    if (test_status %in% c("pass", "skip")) {
      span$set_status("ok")
    } else {
      span$set_status("error", paste("Test", test_status))
    }
  }
})

tracer_enabled <- function(tracer) {
  .subset2(tracer, "is_enabled")()
}

with_otel_record <- function(expr) {
  on.exit(otel_cache_tracer())
  otelsdk::with_otel_record({
    otel_cache_tracer()
    expr
  })
}

#' Does a string match a regular expression?
#'
#' @details
#' `expect_match()` is a wrapper around [grepl()]. See its documentation for
#' more detail about the individual arguments. `expect_no_match()` provides
#' the complementary case, checking that a string *does not* match a regular
#' expression.
#'
#' @inheritParams expect_that
#' @inheritParams base::grepl
#' @param regexp Regular expression to test against.
#' @param all Should all elements of actual value match `regexp` (TRUE),
#'   or does only one need to match (FALSE).
#' @param fixed If `TRUE`, treats `regexp` as a string to be matched exactly
#'   (not a regular expressions). Overrides `perl`.
#' @inheritDotParams base::grepl -pattern -x -perl -fixed
#' @family expectations
#' @export
#' @examples
#' expect_match("Testing is fun", "fun")
#' expect_match("Testing is fun", "f.n")
#' expect_no_match("Testing is fun", "horrible")
#'
#' \dontrun{
#' expect_match("Testing is fun", "horrible")
#'
#' # Zero-length inputs always fail
#' expect_match(character(), ".")
#' }
expect_match <- function(object, regexp, perl = FALSE, fixed = FALSE, ..., all = TRUE,
                         info = NULL, label = NULL) {
  # Capture here to avoid environment-related messiness
  act <- quasi_label(enquo(object), label, arg = "object")
  stopifnot(is.character(regexp), length(regexp) == 1)

  stopifnot(is.character(act$val))
  if (length(object) == 0) {
    fail(sprintf("%s is empty.", act$lab), info = info)
  }

  expect_match_(
    act = act,
    regexp = regexp,
    perl = perl,
    fixed = fixed,
    ...,
    all = all,
    info = info,
    label = label,
    negate = FALSE
  )
}

#' @describeIn expect_match Check that a string doesn't match a regular
#'   expression.
#' @export
expect_no_match <- function(object, regexp, perl = FALSE, fixed = FALSE, ..., all = TRUE,
                            info = NULL, label = NULL) {
  # Capture here to avoid environment-related messiness
  act <- quasi_label(enquo(object), label, arg = "object")
  stopifnot(is.character(regexp), length(regexp) == 1)

  stopifnot(is.character(act$val))
  if (length(object) == 0) {
    fail(sprintf("%s is empty.", act$lab), info = info)
  }

  expect_match_(
    act = act,
    regexp = regexp,
    perl = perl,
    fixed = fixed,
    ...,
    all = all,
    info = info,
    label = label,
    negate = TRUE
  )
}

expect_match_ <- function(act, regexp, perl = FALSE, fixed = FALSE, ..., all = TRUE,
                         info = NULL, label = NULL, negate = FALSE) {

  matches <- grepl(regexp, act$val, perl = perl, fixed = fixed, ...)
  condition <- if (negate) !matches else matches
  escape <- if (fixed) identity else escape_regex

  if (length(act$val) == 1) {
    values <- paste0("Actual value: \"", escape(encodeString(act$val)), "\"")
  } else {
    values <- paste0(
      "Actual values:\n",
      paste0("* ", escape(encodeString(act$val)), collapse = "\n")
    )
  }
  expect(
    if (all) all(condition) else any(condition),
    sprintf(
      if (negate) "%s does match %s.\n%s" else "%s does not match %s.\n%s",
      escape(act$lab),
      encodeString(regexp, quote = '"'),
      values
    ),
    info = info
  )
  invisible(act$val)
}

#' Multi reporter: combine several reporters in one.
#'
#' This reporter is useful to use several reporters at the same time, e.g.
#' adding a custom reporter without removing the current one.
#'
#' @export
#' @family reporters
MultiReporter <- R6::R6Class("MultiReporter",
  inherit = Reporter,
  public = list(
    reporters = list(),

    initialize = function(reporters = list()) {
      super$initialize()
      self$capabilities$parallel_support <- TRUE
      self$reporters <- reporters
    },

    start_reporter = function() {
      o_apply(self$reporters, "start_reporter")
    },
    start_file = function(filename) {
      o_apply(self$reporters, "start_file", filename)
    },
    start_context = function(context) {
      o_apply(self$reporters, "start_context", context)
    },
    start_test = function(context, test) {
      o_apply(self$reporters, "start_test", context, test)
    },
    add_result = function(context, test, result) {
      o_apply(self$reporters, "add_result", context = context, test = test, result = result)
    },
    end_test = function(context, test) {
      o_apply(self$reporters, "end_test", context, test)
    },
    end_context = function(context) {
      o_apply(self$reporters, "end_context", context)
    },
    end_reporter = function() {
      o_apply(self$reporters, "end_reporter")
    },
    end_file = function() {
      o_apply(self$reporters, "end_file")
    },
    update = function() {
      o_apply(self$reporters, "update")
    }
  )
)

o_apply <- function(objects, method, ...) {
  x <- NULL # silence check note
  f <- new_function(exprs(x = ), expr(
    `$`(x, !!method)(...)
  ))

  lapply(objects, f)
}

#' Download snapshots from GitHub
#'
#' @description
#' If your snapshots fail on GitHub, it can be a pain to figure out exactly
#' why, or to incorporate them into your local package. This function makes it
#' easy, only requiring you to interactively select which job you want to
#' take the artifacts from.
#'
#' Note that you should not generally need to use this function manually;
#' instead copy and paste from the hint emitted on GitHub. This hint is only
#' emitted when running in a job named "R-CMD-check", since that's where the
#' testthat artifact is typically uploaded.
#'
#' @param repository Repository owner/name, e.g. `"r-lib/testthat"`.
#' @param run_id Run ID, e.g. `"47905180716"`. You can find this in the action url.
#' @param dest_dir Package root directory. Defaults to the current directory.
#' @export
snapshot_download_gh <- function(repository, run_id, dest_dir = ".") {
  check_string(repository)
  check_string(run_id)
  check_string(dest_dir)

  check_installed("gh")

  dest_snaps <- file.path(dest_dir, "tests", "testthat", "_snaps")
  if (!dir.exists(dest_snaps)) {
    cli::cli_abort("No snapshot directory found in {.file {dest_dir}}.")
  }

  job_id <- gh_find_job(repository, run_id)
  artifact_id <- gh_find_artifact(repository, job_id)

  path <- withr::local_tempfile(pattern = "gh-snaps-")
  gh_download_artifact(repository, artifact_id, path)

  files <- dir(path, full.names = TRUE)
  if (length(files) != 1) {
    cli::cli_abort("Unexpected artifact format.")
  }
  inner_dir <- files[[1]]
  src_snaps <- file.path(inner_dir, "tests", "testthat", "_snaps")
  dir_copy(src_snaps, dest_snaps)
}

gh_find_job <- function(repository, run_id) {
  jobs_json <- gh::gh(
    "/repos/{repository}/actions/runs/{run_id}/jobs",
    repository = repository,
    run_id = run_id
  )
  jobs <- data.frame(
    id = map_dbl(jobs_json$jobs, \(x) x$id),
    name = map_chr(jobs_json$jobs, \(x) x$name)
  )
  jobs <- jobs[order(jobs$name), ]

  idx <- utils::menu(jobs$name, title = "Which job?")
  if (idx == 0) {
    cli::cli_abort("Selection cancelled.")
  }
  jobs$id[[idx]]
}

gh_find_artifact <- function(repository, job_id) {
  job_logs <- gh::gh(
    "GET /repos/{repository}/actions/jobs/{job_id}/logs",
    repository = repository,
    job_id = job_id,
    .send_headers = c("Accept" = "application/vnd.github.v3+json")
  )

  log_lines <- strsplit(job_logs$message, "\r?\n")[[1]]
  matches <- re_match(log_lines, "Artifact download URL: (?<artifact_url>.*)")
  matches <- matches[!is.na(matches$artifact_url), ]
  if (nrow(matches) == 0) {
    cli::cli_abort("Failed to find artifact.")
  }

  # Take last artifact URL; if the job has failed the previous artifact will
  # be the R CMD check logs
  artifact_url <- matches$artifact_url[nrow(matches)]
  basename(artifact_url)
}

gh_download_artifact <- function(repository, artifact_id, path) {
  zip_path <- withr::local_tempfile(pattern = "gh-zip-")
  gh::gh(
    "/repos/{repository}/actions/artifacts/{artifact_id}/{archive_format}",
    repository = repository,
    artifact_id = artifact_id,
    archive_format = "zip",
    .destfile = zip_path
  )
  utils::unzip(zip_path, exdir = path)
  invisible(path)
}

# Directory helpers ------------------------------------------------------------

dir_create <- function(paths) {
  for (path in paths) {
    dir.create(path, recursive = TRUE, showWarnings = FALSE)
  }
  invisible(paths)
}

dir_copy <- function(src_dir, dst_dir) {
  # First create directories
  dirs <- list.dirs(src_dir, recursive = TRUE, full.names = FALSE)
  dir_create(file.path(dst_dir, dirs))

  # Then copy files
  files <- dir(src_dir, recursive = TRUE)
  src_files <- file.path(src_dir, files)
  dst_files <- file.path(dst_dir, files)
  same <- map_lgl(seq_along(files), \(i) {
    same_file(src_files[[i]], dst_files[[i]])
  })

  n_new <- sum(!same)
  if (n_new == 0) {
    cli::cli_inform(c(i = "No new snapshots."))
  } else {
    cli::cli_inform(c(
      v = "Copying {n_new} new snapshots: {.file {files[!same]}}."
    ))
  }

  file.copy(src_files[!same], dst_files[!same], overwrite = TRUE)
  invisible()
}

same_file <- function(x, y) {
  file.exists(x) && file.exists(y) && hash_file(x) == hash_file(y)
}

on_gh <- function() {
  Sys.getenv("GITHUB_ACTIONS") == "true"
}

colourise <- function(text, as = c("success", "skip", "warning", "failure", "error")) {
  if (has_colour()) {
    unclass(cli::make_ansi_style(testthat_style(as))(text))
  } else {
    text
  }
}

has_colour <- function() {
  isTRUE(getOption("testthat.use_colours", TRUE)) &&
    cli::num_ansi_colors() > 1
}

testthat_style <- function(type = c("success", "skip", "warning", "failure", "error")) {
  type <- match.arg(type)

  c(
    success = "green",
    skip = "blue",
    warning = "magenta",
    failure = "orange",
    error = "orange"
  )[[type]]
}

test_that("can supress colours", {
  op <- options(
    crayon.enabled = TRUE,
    testthat.use_colours = TRUE
  )
  check <- crayon::has_style(colourise("X"))
  # Must restore original options before expectation is triggered
  options(op)

  expect_true(check)
})

test_that("We don't have colours if we don't want to", {
  op <- options(
    crayon.enabled = TRUE,
    testthat.use_colours = FALSE
  )
  check <- crayon::has_style(colourise("X"))
  options(op)

  expect_false(check)
})

test_that("can suppress colours", {
  op <- options(
    crayon.enabled = TRUE,
    testthat.use_colours = TRUE,
    cli.num_colors = 8L
  )
  check <- cli::ansi_has_any(colourise("X"))
  # Must restore original options before expectation is triggered
  options(op)

  expect_true(check)
})

test_that("We don't have colours if we don't want to", {
  op <- options(
    crayon.enabled = TRUE,
    testthat.use_colours = FALSE
  )
  check <- cli::ansi_has_any(colourise("X"))
  options(op)

  expect_false(check)
})

# See https://www.tidyverse.org/blog/2019/09/callr-task-q/
# for a detailed explanation on how the task queue works.
#
# Changes in this version, compared to the blog post:
# * We use data frames instead of tibbles. This requires some caution
#   and the df_add_row() function below.
# * We do not collect the results in a result column, because we
#   just return them immediately, as we get them.
# * We do not need a pop() method, because poll() will just return
#   every message.

PROCESS_DONE <- 200L
PROCESS_STARTED <- 201L
PROCESS_MSG <- 301L
PROCESS_OUTPUT <- 302L
PROCESS_EXITED <- 500L
PROCESS_CRASHED <- 501L
PROCESS_CLOSED <- 502L
PROCESS_FAILURES <- c(PROCESS_EXITED, PROCESS_CRASHED, PROCESS_CLOSED)

task_q <- R6::R6Class(
  "task_q",
  public = list(
    initialize = function(concurrency = 4L, ...) {
      private$start_workers(concurrency, ...)
      invisible(self)
    },
    list_tasks = function() private$tasks,
    get_num_waiting = function() {
      sum(!private$tasks$idle & private$tasks$state == "waiting")
    },
    get_num_running = function() {
      sum(!private$tasks$idle & private$tasks$state == "running")
    },
    get_num_done = function() sum(private$tasks$state == "done"),
    is_idle = function() sum(!private$tasks$idle) == 0,

    push = function(fun, args = list(), id = NULL) {
      if (is.null(id)) {
        id <- private$get_next_id()
      }
      if (id %in% private$tasks$id) {
        cli::cli_abort("Duplicate task id.")
      }
      before <- which(private$tasks$idle)[1]
      private$tasks <- df_add_row(
        private$tasks,
        .before = before,
        id = id,
        idle = FALSE,
        state = "waiting",
        fun = I(list(fun)),
        args = I(list(args)),
        worker = I(list(NULL)),
        path = args[[1]],
        startup = I(list(NULL))
      )
      private$schedule()
      invisible(id)
    },

    poll = function(timeout = 0) {
      limit <- Sys.time() + timeout
      as_ms <- function(x) {
        if (x == Inf) -1 else as.integer(as.double(x, "secs") * 1000)
      }
      repeat {
        pr <- vector(mode = "list", nrow(private$tasks))
        topoll <- which(private$tasks$state == "running")
        pr[topoll] <- processx::poll(
          private$tasks$worker[topoll],
          as_ms(timeout)
        )
        results <- lapply(seq_along(pr), function(i) {
          # nothing from this worker?
          if (is.null(pr[[i]]) || all(pr[[i]] != "ready")) {
            return()
          }

          # there is a testthat message?
          worker <- private$tasks$worker[[i]]
          msg <- if (pr[[i]][["process"]] == "ready") {
            worker$read()
          }

          # there is an output message?
          has_output <- pr[[i]][["output"]] == "ready" ||
            pr[[i]][["error"]] == "ready"
          outmsg <- NULL
          if (has_output) {
            lns <- c(
              safely(worker$read_output_lines(), character()),
              safely(worker$read_error_lines(), character())
            )
            inc <- paste0(
              safely(worker$read_output(), ""),
              safely(worker$read_error(), "")
            )
            if (nchar(inc)) {
              lns <- c(lns, strsplit(inc, "\n", fixed = TRUE)[[1]])
            }
            # startup message?
            if (is.na(private$tasks$path[i])) {
              private$tasks$startup[[i]] <- c(private$tasks$startup[[i]], lns)
            } else {
              outmsg <- structure(
                list(
                  code = PROCESS_OUTPUT,
                  message = lns,
                  path = private$tasks$path[i]
                ),
                class = "testthat_message"
              )
            }
          }

          ## TODO: why can this be NULL?
          if (is.null(msg) || msg$code == PROCESS_MSG) {
            private$tasks$state[[i]] <- "running"
          } else if (msg$code == PROCESS_STARTED) {
            private$tasks$state[[i]] <- "ready"
            msg <- NULL
          } else if (msg$code == PROCESS_DONE) {
            if (!is.null(msg$error)) {
              private$handle_error(msg, i)
            }
            private$tasks$state[[i]] <- "ready"
          } else if (msg$code %in% PROCESS_FAILURES) {
            private$handle_error(msg, i)
          } else {
            file <- private$tasks$args[[i]][[1]]
            cli::cli_abort(
              c(
                "Unknown message from testthat subprocess: {msg$code}.",
                "i" = "In file {.file {file}}."
              ),
              test_file = file,
              class = c("testthat_process_error", "testthat_error")
            )
          }
          compact(list(msg, outmsg))
        })
        # single list for all workers
        results <- compact(unlist(results, recursive = FALSE))

        private$schedule()
        if (is.finite(timeout)) {
          timeout <- limit - Sys.time()
        }
        if (length(results) || timeout < 0) break
      }
      results
    }
  ),

  private = list(
    tasks = NULL,
    next_id = 1L,
    get_next_id = function() {
      id <- private$next_id
      private$next_id <- id + 1L
      paste0(".", id)
    },

    start_workers = function(concurrency, ...) {
      nl <- I(replicate(concurrency, NULL))
      private$tasks <- data.frame(
        stringsAsFactors = FALSE,
        id = paste0(".idle-", seq_len(concurrency)),
        idle = TRUE,
        state = "running",
        fun = nl,
        args = nl,
        worker = nl,
        path = NA_character_,
        startup = nl
      )
      rsopts <- callr::r_session_options(stdout = "|", stderr = "|", ...)
      for (i in seq_len(concurrency)) {
        rs <- callr::r_session$new(rsopts, wait = FALSE)
        private$tasks$worker[[i]] <- rs
      }
    },

    schedule = function() {
      ready <- which(private$tasks$state == "ready")
      if (!length(ready)) {
        return()
      }
      rss <- private$tasks$worker[ready]

      private$tasks$worker[ready] <- replicate(length(ready), NULL)
      private$tasks$state[ready] <-
        ifelse(private$tasks$idle[ready], "waiting", "done")
      done <- which(private$tasks$state == "done")
      if (any(done)) {
        private$tasks <- private$tasks[-done, ]
      }

      waiting <- which(private$tasks$state == "waiting")[1:length(ready)]
      private$tasks$worker[waiting] <- rss
      private$tasks$state[waiting] <-
        ifelse(private$tasks$idle[waiting], "ready", "running")
      lapply(waiting, function(i) {
        if (!private$tasks$idle[i]) {
          private$tasks$worker[[i]]$call(
            private$tasks$fun[[i]],
            private$tasks$args[[i]]
          )
        }
      })
    },

    handle_error = function(msg, task_no) {
      cat("\n") # get out of the progress bar, if any
      fun <- private$tasks$fun[[task_no]]
      file <- private$tasks$args[[task_no]][[1]]
      if (is.null(fun)) {
        msg$error$stdout <- msg$stdout
        msg$error$stderr <- paste(
          c(private$tasks$startup[[task_no]], msg$stderr),
          collapse = "\n"
        )
        cli::cli_abort(
          c(
            "testthat subprocess failed to start.",
            " " = "{no_wrap(msg$error$stderr)}"
          ),
          test_file = NULL,
          class = c("testthat_process_error", "testthat_error"),
          call = NULL
        )
      } else {
        cli::cli_abort(
          "testthat subprocess exited in file {.file {file}}.",
          test_file = file,
          parent = msg$error,
          class = c("testthat_process_error", "testthat_error"),
          call = NULL
        )
      }
    }
  )
)

df_add_row <- function(df, ..., .before = NULL) {
  before <- .before %||% (nrow(df) + 1L)
  row <- data.frame(stringsAsFactors = FALSE, ...)
  if (before > nrow(df)) {
    rbind(df, row)
  } else if (before <= 1L) {
    rbind(row, df)
  } else {
    rbind(df[1:(before - 1), ], row, df[before:nrow(df), ])
  }
}

safely <- function(expr, default = NULL) {
  tryCatch(
    expr,
    error = function(e) default
  )
}

silence_r_cmd_check <- function() callr::r_session

#' Report test progress for LLMs
#'
#' @description
#' `LlmReporter` is designed for use with Large Language Models (LLMs).
#' It reports problems (warnings, skips, errors, and failures) as they
#' occur and the total number of successes at the end.
#'
#' `LlmReporter` is used by default when tests are run by a coding agent.
#' Currently we detect Claude Code, Cursor, and Gemini CLI.
#' If using another tool, configure it to set env var `AGENT=1`.
#'
#' @export
#' @family reporters
LlmReporter <- R6::R6Class(
  "LlmReporter",
  inherit = ProgressReporter,
  public = list(
    initialize = function(max_failures = testthat_max_fails(), ...) {
      super$initialize(
        show_praise = FALSE,
        max_failures = max_failures,
        ...
      )
      self$width <- 20
      self$rstudio <- FALSE
      self$hyperlinks <- FALSE
    },

    # No header
    show_header = function() {},

    # No status updates
    show_status = function(complete = FALSE, time = 0, pad = FALSE) {},

    add_result = function(context, test, result) {
      if (self$is_full()) {
        return()
      }

      if (expectation_broken(result)) {
        self$n_fail <- self$n_fail + 1
        self$report_issue(result)
      } else if (expectation_skip(result)) {
        self$n_skip <- self$n_skip + 1
        self$report_issue(result)
      } else if (expectation_warning(result)) {
        self$n_warn <- self$n_warn + 1
        self$report_issue(result)
      } else {
        self$n_ok <- self$n_ok + 1
        # Do nothing for passing tests
      }
    },

    report_issue = function(result) {
      self$local_user_output()

      type <- toupper(expectation_type(result))
      header <- paste0(type, ": ", expectation_location(result))
      rule <- strrep("-", max(50 - nchar(header), 10))

      self$cat_line(header, " ", rule)
      self$cat_line(format(result))
      self$cat_line()
    },

    end_context = function(context) {
      if (self$is_full()) {
        self$report_full()
      }
    },

    end_reporter = function() {
      if (self$is_full()) {
        return()
      }

      self$cat_line(paste_c(
        "[ ",
        c("FAIL ", self$n_fail, " | "),
        c("WARN ", self$n_warn, " | "),
        c("SKIP ", self$n_skip, " | "),
        c("PASS ", self$n_ok),
        " ]"
      ))
    }
  )
)

is_llm <- function() {
  nzchar(Sys.getenv("AGENT")) ||
    nzchar(Sys.getenv("CLAUDECODE")) ||
    nzchar(Sys.getenv("GEMINI_CLI")) ||
    nzchar(Sys.getenv("CURSOR_AGENT"))
}

new_capture <- function(class) {
  exiting_handlers <- rep_named(class, list(identity))

  calling_handlers <- rep_named(
    class,
    alist(function(cnd) {
      if (can_entrace(cnd)) {
        cnd <- cnd_entrace(cnd)
      }
      return_from(env, cnd)
    })
  )

  formals <- pairlist2(code = , entrace = FALSE)

  # R CMD check global variable NOTE
  code <- entrace <- NULL

  body <- expr({
    if (!entrace) {
      return(tryCatch(
        {
          code
          NULL
        },
        !!!exiting_handlers
      ))
    }

    env <- environment()
    withCallingHandlers(
      {
        code
        NULL
      },
      !!!calling_handlers
    )
  })

  new_function(formals, body, ns_env("testthat"))
}

#' Capture conditions, including messages, warnings, expectations, and errors.
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' These functions allow you to capture the side-effects of a function call
#' including printed output, messages and warnings. We no longer recommend
#' that you use these functions, instead relying on the [expect_message()]
#' and friends to bubble up unmatched conditions. If you just want to silence
#' unimportant warnings, use [suppressWarnings()].
#'
#' @param code Code to evaluate
#' @param entrace Whether to add a [backtrace][rlang::trace_back] to
#'   the captured condition.
#' @return Singular functions (`capture_condition`, `capture_expectation` etc)
#'   return a condition object. `capture_messages()` and `capture_warnings`
#'   return a character vector of message text.
#' @keywords internal
#' @export
#' @examples
#' f <- function() {
#'   message("First")
#'   warning("Second")
#'   message("Third")
#' }
#'
#' capture_message(f())
#' capture_messages(f())
#'
#' capture_warning(f())
#' capture_warnings(f())
#'
#' # Condition will capture anything
#' capture_condition(f())
capture_condition <- new_capture("condition")

#' @export
#' @rdname capture_condition
capture_error <- new_capture("error")

#' @export
#' @rdname capture_condition
capture_expectation <- new_capture("expectation")

#' @export
#' @rdname capture_condition
capture_message <- new_capture("condition")

#' @export
#' @rdname capture_condition
capture_warning <- new_capture("warning")

#' @export
#' @rdname capture_condition
capture_messages <- function(code) {
  out <- Stack$new()

  withCallingHandlers(
    code,
    message = function(condition) {
      out$push(condition)
      tryInvokeRestart("muffleMessage")
    }
  )

  get_messages(out$as_list())
}

#' @export
#' @rdname capture_condition
capture_warnings <- function(code, ignore_deprecation = FALSE) {
  out <- Stack$new()

  withCallingHandlers(
    code,
    warning = function(condition) {
      if (ignore_deprecation && is_deprecation(condition)) {
        return()
      }

      out$push(condition)
      tryInvokeRestart("muffleWarning")
    }
  )

  get_messages(out$as_list())
}

get_messages <- function(x) {
  map_chr(x, cnd_message)
}

#' Is an error informative?
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `is_informative_error()` is a generic predicate that indicates
#' whether testthat users should explicitly test for an error
#' class. Since we no longer recommend you do that, this generic
#' has been deprecated.
#'
#' @param x An error object.
#' @inheritParams rlang::args_dots_empty
#'
#' @details
#' A few classes are hard-coded as uninformative:
#' - `simpleError`
#' - `rlang_error` unless a subclass is detected
#' - `Rcpp::eval_error`
#' - `Rcpp::exception`
#'
#' @keywords internal
#' @export
is_informative_error <- function(x, ...) {
  lifecycle::deprecate_stop("3.0.0", "is_informative_error()")
}

test_that("checks for any type of output", {
  expect_failure(expect_silent(warning("!")))
  expect_failure(expect_silent(message("!")))
  expect_failure(expect_silent(print("!")))

  expect_success(expect_silent(""))
})

test_that("returns first argument", {
  expect_equal(expect_silent(1), 1)
})

test_that("deprecations are ignored", {
  foo <- function() {
    lifecycle::deprecate_warn("1.0.0", "foo()")
  }
  expect_success(expect_silent(foo))
})

test_that("errors thrown at block level are entraced", {
  f <- function() g()
  g <- function() stop("foo")
  f()
})

test_that("errors thrown from a quasi-labelled argument are entraced", {
  foo <- function() stop("foo")
  expect_s3_class(foo(), "foo")
})

test_that("errors thrown from a quasi-labelled argument are entraced (deep case)", {
  foo <- function() stop("foo")
  f <- function() g()
  g <- function() expect_s3_class(foo(), "foo")
  expect_s3_class(f(), "foo")
})

test_that("errors thrown from a quasi-labelled argument are entraced (deep deep case)", {
  foo <- function() bar()
  bar <- function() stop("foobar")

  f <- function() g()
  g <- function() expect_s3_class(foo(), "foo")

  f()
})

test_that("failed expect_error() prints a backtrace", {
  f <- function() signaller()

  signaller <- function() stop("bar")
  expect_error(f(), "foo")
})

test_that("Errors are inspected with `conditionMessage()`", {
  rlang::local_bindings(
    .env = globalenv(),
    conditionMessage.foobar = function(...) "dispatched"
  )
  rlang::abort("Wrong message", "foobar")
})

test_that("also get backtraces for warnings", {
  foo <- function() bar()
  bar <- function() warning("foobar", call. = FALSE)

  foo()
  expect_true(TRUE)
})

test_that("deep stacks are shown", {
  f <- function(x) {
    if (x > 0) f(x - 1) else stop("This is deep")
  }
  f(25)
})

# Expectations ----------------------------------------------------------------
f <- function() g()
g <- function() h()
h <- function() expect_true(FALSE)

f()

test_that("nested expectations get backtraces", {
  f()
})

test_that("variants save different values", {
  expect_snapshot(r_version(), variant = r_version())
})

test_that("can snapshot output", {
  foo <- function() cat("y")
  expect_snapshot_output(foo())
  expect_snapshot_output(foo())
  expect_snapshot_output(foo())
  expect_snapshot_output(foo())
})

test_that("can snapshot everything", {
  f <- function() {
    print("1")
    message("2")
    warning("3")
    stop("4")
  }
  expect_snapshot(f(), error = TRUE)
})

test_that("empty lines are preserved", {
  f <- function() {
    cat("1\n\n")
    message("2\n")
    warning("3\n")
    stop("4\n\n")
  }
  expect_snapshot(f(), error = TRUE)
})

test_that("line-endings fixed before comparison", {
  x <- "a\n\rb"
  expect_snapshot(cat(x))
})

test_that("multiple outputs of same type are collapsed", {
  expect_snapshot({
    x <- 1
    y <- 1
    {message("a"); message("b")}
    {warning("a"); warning("b")}
  })
})

test_that("can scrub output/messages/warnings/errors", {
  secret <- function() {
    print("secret")
    message("secret")
    warning("secret")
    stop("secret")
  }
  redact <- function(x) gsub("secret", "<redacted>", x)
  expect_snapshot(secret(), transform = redact, error = TRUE)

  # Or with an inline fun
  expect_snapshot(print("secret"), transform = ~ gsub("secret", "****", .x))
})

test_that("always checks error status", {
  expect_error(expect_snapshot(stop("!"), error = FALSE))
  expect_failure(expect_snapshot(print("!"), error = TRUE))
})

test_that("can capture error/warning messages", {
  expect_snapshot_error(stop("This is an error"))
  expect_snapshot_warning(warning("This is a warning"))
})

test_that("snapshot captures deprecations", {
  foo <- function() {
    lifecycle::deprecate_warn("1.0.0", "foo()")
  }

  expect_snapshot(foo())
  expect_snapshot_warning(foo())
  expect_snapshot_warning(foo(), class = "lifecycle_warning_deprecated")
})

test_that("can check error/warning classes", {
  expect_snapshot(expect_snapshot_error(1), error = TRUE)
  expect_snapshot(expect_snapshot_error(1, class = "myerror"), error = TRUE)
  expect_snapshot(expect_snapshot_warning(1), error = TRUE)
  expect_snapshot(expect_snapshot_warning(1, class = "mywarning"), error = TRUE)
})

test_that("snapshot handles multi-line input", {
  expect_snapshot({
    1 + 2
    3 + 4
    "this is a comment"
  })
})

test_that("snapshot captures output if visible", {
  f_visible <- function() "x"
  f_invisible <- function() invisible("x")

  expect_snapshot(f_visible())
  expect_snapshot(f_invisible())
})

test_that("captures custom classes", {
  f <- function() {
    inform("Hello", class = "testthat_greeting")
    warn("Goodbye", class = "testthat_farewell")
    abort("Eeek!", class = "testthat_scream")
  }
  expect_snapshot(f(), error = TRUE)
})

test_that("even with multiple lines", {
  expect_snapshot_output(cat("a\nb\nc"))
  expect_snapshot_output(cat("a\nb\nc\n"))
})

test_that("`expect_snapshot()` does not inject", {
  expect_snapshot({
    x <- quote(!!foo)
    expect_equal(x, call("!", call("!", quote(foo))))
  })
})

test_that("full condition message is printed with rlang", {
  expect_snapshot(error = TRUE, {
    foo <- error_cnd("foo", message = "Title parent.")
    abort("Title.", parent = foo)
  })
})

test_that("can print with and without condition classes", {
  f <- function() {
    message("foo")
    warning("bar")
    stop("baz")
  }
  expect_snapshot(error = TRUE, cnd_class = TRUE, f())
  expect_snapshot(error = TRUE, cnd_class = FALSE, f())
})

test_that("errors and warnings are folded", {
  f <- function() {
    warning("foo")
    stop("bar")
  }
  expect_snapshot(error = TRUE, f())
})

# I don't know how to test this automatically; wrapping it in another
# snapshot doesn't capture the behaviour I expected, presumably due to the
# way that errors bubble up
# test_that("errors in snapshots behave like regular errors", {
#   f <- function() g()
#   g <- function() h()
#   h <- function() abort("!")
#
#   expect_snapshot(f())
#   expect_snapshot(1 + 1)
# })

test_that("hint is informative", {
  local_reproducible_output(crayon = TRUE, hyperlinks = TRUE, rstudio = TRUE)

  expect_snapshot({
    cat(snapshot_accept_hint("_default", "bar.R", reset_output = FALSE))
    cat(snapshot_accept_hint("foo", "bar.R", reset_output = FALSE))
  })
})

test_that("expect_snapshot requires a non-empty test label", {

  test_that("", {
    expect_error(expect_snapshot(1 + 1))
  })

})

test_that("can record all types of output", {
  verify_output(test_path("test-verify-output.txt"), {
    "Output"
    1 + 2
    invisible(1:10)
    12345678 + 12345678 + 12345678 + 12345678 + 12345678 + 12345678 +
      12345678 + 12345678 + 12345678 + 12345678 + 12345678

    "# Header"
    "Other output"
    letters
  })
})

test_that("can record all types of output", {
  local_bindings(
    .env = global_env(),
    conditionMessage.foobar = function(cnd) {
      paste("Dispatched!", cnd$message)
    }
  )

  verify_output(test_path("test-verify-conditions.txt"), {
    message("Message")

    "With calls"
    warning("Warning")
    stop("Error")

    "Without calls"
    warning("Warning", call. = FALSE)
    stop("Error", call. = FALSE)

    "With `conditionMessage()` method"
    cnd_signal(message_cnd("foobar", message = "Message"))
    cnd_signal(warning_cnd("foobar", message = "Warning"))
    cnd_signal(error_cnd("foobar", message = "Error"))
  })
})

test_that("can't record plots", {
  skip_if(interactive())
  expect_error(verify_output(tempfile(), plot(1:10)), "Plots")
})

test_that("verify_output() splits condition messages on newlines", {
  verify_output(test_path("test-verify-conditions-lines.txt"), {
    message("First.\nSecond.")
    warning("First.\nSecond.")
    stop("First.\nSecond.")
  })
})

test_that("can use constructed calls in verify_output() (#945)", {
  verify_output(test_path("test-verify-constructed-calls.txt"), {
    expr(foo(!!c("bar", "baz")))

    # Can unquote local objects
    binding <- quote(foo)
    expr(foo(!!binding))
  })
})

test_that("verify_output() doesn't use cli unicode by default", {
  verify_output(
    test_path("test-verify-unicode-false.txt"),
    {
      cat(cli::symbol$info, cli::symbol$cross, "\n")
    }
  )

  local_reproducible_output(unicode = TRUE)
  verify_output(
    test_path("test-verify-unicode-true.txt"),
    unicode = TRUE,
    {
      cat(cli::symbol$info, cli::symbol$cross, "\n")
    })
})

test_that("verify_output() handles carriage return", {
  verify_output(test_path("test-verify-conditions-cr.txt"), {
    cat("\r\n")
  })
})

test_that("verify_exec() doesn't leave tempfiles around", {
  before <- dir(tempdir())
  verify_exec(quote(1 + 1))
  after <- dir(tempdir())

  expect_equal(before, after)
})

test_that("verify_exec() strips CR", {
  act <- verify_exec(quote(cat("\r\n")))
  exp <- verify_exec(quote(cat("\n")))
  expect_equal(act[-1], exp[-1])
})

test_that("captures warnings, messages and output", {
  out <- evaluate_promise({
    message("m", appendLF = FALSE)
    warning("w")
    cat("out")
  })

  expect_equal(out$output, "out")
  expect_equal(out$messages, "m")
  expect_equal(out$warnings, "w")
})

test_that("capture_warnings captures warnings", {
  out <- capture_warnings({
    warning("a")
    warning("b")
  })
  expect_equal(out, c("a", "b"))
})

test_that("capture_messages captures messages", {
  out <- capture_messages({
    message("a")
    message("b")
  })
  expect_equal(out, c("a\n", "b\n")) # message adds LF by default
})

test_that("is_informative_error is defunct", {
  expect_snapshot(is_informative_error(TRUE), error = TRUE)
})

test_that("capture_warnings can ignore deprecation warnings", {
  foo <- function() {
    lifecycle::deprecate_warn("1.0.0", "foo()")
  }
  expect_warning(
    expect_equal(capture_warnings(foo(), TRUE), character()),
    class = "lifecycle_warning_deprecated"
  )
})

#' Does code throw an error, warning, message, or other condition?
#'
#' @description
#' `expect_error()`, `expect_warning()`, `expect_message()`, and
#' `expect_condition()` check that code throws an error, warning, message,
#' or condition with a message that matches `regexp`, or a class that inherits
#' from `class`. See below for more details.
#'
#' In the 3rd edition, these functions match (at most) a single condition. All
#' additional and non-matching (if `regexp` or `class` are used) conditions
#' will bubble up outside the expectation. If these additional conditions
#' are important you'll need to catch them with additional
#' `expect_message()`/`expect_warning()` calls; if they're unimportant you
#' can ignore with [suppressMessages()]/[suppressWarnings()].
#'
#' It can be tricky to test for a combination of different conditions,
#' such as a message followed by an error. [expect_snapshot()] is
#' often an easier alternative for these more complex cases.
#'
#' @section Testing `message` vs `class`:
#' When checking that code generates an error, it's important to check that the
#' error is the one you expect. There are two ways to do this. The first
#' way is the simplest: you just provide a `regexp` that match some fragment
#' of the error message. This is easy, but fragile, because the test will
#' fail if the error message changes (even if its the same error).
#'
#' A more robust way is to test for the class of the error, if it has one.
#' You can learn more about custom conditions at
#' <https://adv-r.hadley.nz/conditions.html#custom-conditions>, but in
#' short, errors are S3 classes and you can generate a custom class and check
#' for it using `class` instead of `regexp`.
#'
#' If you are using `expect_error()` to check that an error message is
#' formatted in such a way that it makes sense to a human, we recommend
#' using [expect_snapshot()] instead.
#'
#' @export
#' @family expectations
#' @inheritParams expect_that
#' @param regexp Regular expression to test against.
#'   * A character vector giving a regular expression that must match the
#'     error message.
#'   * If `NULL`, the default, asserts that there should be an error,
#'     but doesn't test for a specific value.
#'   * If `NA`, asserts that there should be no errors, but we now recommend
#'     using [expect_no_error()] and friends instead.
#'
#'   Note that you should only use `message` with errors/warnings/messages
#'   that you generate. Avoid tests that rely on the specific text generated by
#'   another package since this can easily change. If you do need to test text
#'   generated by another package, either protect the test with `skip_on_cran()`
#'   or use `expect_snapshot()`.
#' @inheritDotParams expect_match -object -regexp -info -label -all
#' @param class Instead of supplying a regular expression, you can also supply
#'   a class name. This is useful for "classed" conditions.
#' @param inherit Whether to match `regexp` and `class` across the
#'   ancestry of chained errors.
#' @param all *DEPRECATED* If you need to test multiple warnings/messages
#'   you now need to use multiple calls to `expect_message()`/
#'   `expect_warning()`
#' @seealso [expect_no_error()], `expect_no_warning()`,
#'   `expect_no_message()`, and `expect_no_condition()` to assert
#'   that code runs without errors/warnings/messages/conditions.
#' @return If `regexp = NA`, the value of the first argument; otherwise
#'   the captured condition.
#' @examples
#' # Errors ------------------------------------------------------------------
#' f <- function() stop("My error!")
#' expect_error(f())
#' expect_error(f(), "My error!")
#'
#' # You can use the arguments of grepl to control the matching
#' expect_error(f(), "my error!", ignore.case = TRUE)
#'
#' # Note that `expect_error()` returns the error object so you can test
#' # its components if needed
#' err <- expect_error(rlang::abort("a", n = 10))
#' expect_equal(err$n, 10)
#'
#' # Warnings ------------------------------------------------------------------
#' f <- function(x) {
#'   if (x < 0) {
#'     warning("*x* is already negative")
#'     return(x)
#'   }
#'   -x
#' }
#' expect_warning(f(-1))
#' expect_warning(f(-1), "already negative")
#' expect_warning(f(1), NA)
#'
#' # To test message and output, store results to a variable
#' expect_warning(out <- f(-1), "already negative")
#' expect_equal(out, -1)
#'
#' # Messages ------------------------------------------------------------------
#' f <- function(x) {
#'   if (x < 0) {
#'     message("*x* is already negative")
#'     return(x)
#'   }
#'
#'   -x
#' }
#' expect_message(f(-1))
#' expect_message(f(-1), "already negative")
#' expect_message(f(1), NA)
expect_error <- function(object,
                         regexp = NULL,
                         class = NULL,
                         ...,
                         inherit = TRUE,
                         info = NULL,
                         label = NULL) {

  if (edition_get() >= 3) {
    expect_condition_matching("error", {{ object }},
      regexp = regexp,
      class = class,
      ...,
      inherit = inherit,
      info = info,
      label = label
    )
  } else {
    act <- quasi_capture(enquo(object), label, capture_error, entrace = TRUE)
    msg <- compare_condition_2e(
      act$cap,
      act$lab,
      regexp = regexp,
      class = class,
      ...,
      inherit = inherit
    )

    # Access error fields with `[[` rather than `$` because the
    # `$.Throwable` from the rJava package throws with unknown fields
    expect(is.null(msg), msg, info = info, trace = act$cap[["trace"]])
    invisible(act$val %||% act$cap)
  }
}


#' @export
#' @rdname expect_error
expect_warning <- function(object,
                           regexp = NULL,
                           class = NULL,
                           ...,
                           inherit = TRUE,
                           all = FALSE,
                           info = NULL,
                           label = NULL) {

  if (edition_get() >= 3) {
    if (!missing(all)) {
      warn("The `all` argument is deprecated")
    }

    expect_condition_matching("warning", {{ object }},
      regexp = regexp,
      class = class,
      ...,
      inherit = inherit,
      info = info,
      label = label
    )
  } else {
    act <- quasi_capture(enquo(object), label, capture_warnings, ignore_deprecation = identical(regexp, NA))
    msg <- compare_messages(
      act$cap, act$lab, regexp = regexp, all = all, ...,
      cond_type = "warnings"
    )
    expect(is.null(msg), msg, info = info)

    invisible(act$val)
  }
}

#' @export
#' @rdname expect_error
expect_message <- function(object,
                           regexp = NULL,
                           class = NULL,
                           ...,
                           inherit = TRUE,
                           all = FALSE,
                           info = NULL,
                           label = NULL) {

  if (edition_get() >= 3) {
    expect_condition_matching("message", {{ object }},
      regexp = regexp,
      class = class,
      ...,
      inherit = inherit,
      info = info,
      label = label
    )
  } else {
    act <- quasi_capture(enquo(object), label, capture_messages)
    msg <- compare_messages(act$cap, act$lab, regexp = regexp, all = all, ...)
    expect(is.null(msg), msg, info = info)

    invisible(act$val)
  }
}

#' @export
#' @rdname expect_error
expect_condition <- function(object,
                             regexp = NULL,
                             class = NULL,
                             ...,
                             inherit = TRUE,
                             info = NULL,
                             label = NULL) {

  if (edition_get() >= 3) {
    expect_condition_matching("condition", {{ object }},
      regexp = regexp,
      class = class,
      ...,
      inherit = inherit,
      info = info,
      label = label
    )
  } else {

    act <- quasi_capture(enquo(object), label, capture_condition, entrace = TRUE)
    msg <- compare_condition_2e(
      act$cap,
      act$lab,
      regexp = regexp,
      class = class,
      ...,
      inherit = inherit,
      cond_type = "condition"
    )
    expect(is.null(msg), msg, info = info, trace = act$cap[["trace"]])

    invisible(act$val %||% act$cap)
  }
}

expect_condition_matching <- function(base_class,
                                      object,
                                      regexp = NULL,
                                      class = NULL,
                                      ...,
                                      inherit = TRUE,
                                      info = NULL,
                                      label = NULL,
                                      trace_env = caller_env(),
                                      error_call = caller_env()) {
  matcher <- cnd_matcher(
    base_class,
    class,
    regexp,
    ...,
    inherit = inherit,
    ignore_deprecation = base_class == "warning" && identical(regexp, NA),
    error_call = error_call
  )

  act <- quasi_capture(
    enquo(object),
    label,
    capture_matching_condition,
    matches = matcher
  )

  expected <- !identical(regexp, NA)
  msg <- compare_condition_3e(base_class, class, act$cap, act$lab, expected)

  # Access error fields with `[[` rather than `$` because the
  # `$.Throwable` from the rJava package throws with unknown fields
  expect(is.null(msg), msg, info = info, trace = act$cap[["trace"]], trace_env = trace_env)

  # If a condition was expected, return it. Otherwise return the value
  # of the expression.
  invisible(if (expected) act$cap else act$val)
}

# -------------------------------------------------------------------------

cnd_matcher <- function(base_class,
                        class = NULL,
                        pattern = NULL,
                        ...,
                        inherit = TRUE,
                        ignore_deprecation = FALSE,
                        error_call = caller_env()) {
  check_string(class, allow_null = TRUE, call = error_call)
  check_string(pattern, allow_null = TRUE, allow_na = TRUE, call = error_call)

  if (is.null(pattern) && dots_n(...) > 0) {
    cli::cli_abort(
      "Can't specify {.arg ...} without {.arg pattern}.",
      call = error_call
    )
  }

  function(cnd) {
    if (!inherit) {
      cnd$parent <- NULL
    }

    if (ignore_deprecation && is_deprecation(cnd)) {
      return(FALSE)
    }

    matcher <- function(x) {
      if (!inherits(x, base_class)) {
        return(FALSE)
      }
      if (!is.null(class) && !inherits(x, class)) {
        return(FALSE)
      }
      if (!is.null(pattern) && !isNA(pattern)) {
        withCallingHandlers(
          grepl(pattern, conditionMessage(x), ...),
          error = function(e) {
            cli::cli_abort(
              "Failed to compare {base_class} to {.arg pattern}.",
              parent = e,
              call = error_call
            )
          }
        )

      } else {
        TRUE
      }
    }
    cnd_some(cnd, matcher)
  }
}

has_classes <- function(x, classes) {
  all(classes %in% class(x))
}

is_deprecation <- function(x) {
  inherits(x, "lifecycle_warning_deprecated")
}

cnd_some <- function(.cnd, .p, ...) {
  .p <- as_function(.p)

  while (is_condition(.cnd)) {
    if (.p(.cnd, ...)) {
      return(TRUE)
    }

    .cnd <- .cnd$parent
  }

  FALSE
}

capture_matching_condition <- function(expr, matches) {
  matched <- NULL
  tl <- current_env()

  withCallingHandlers(expr, condition = function(cnd) {
    if (!is.null(matched) || !matches(cnd)) {
      return()
    }

    if (can_entrace(cnd)) {
      cnd <- cnd_entrace(cnd)
    }
    matched <<- cnd

    if (inherits(cnd, "message") || inherits(cnd, "warning")) {
      cnd_muffle(cnd)
    } else if (inherits(cnd, "error") || inherits(cnd, "skip")) {
      return_from(tl, cnd)
    }
  })

  matched
}

# Helpers -----------------------------------------------------------------

compare_condition_3e <- function(cond_type, cond_class, cond, lab, expected) {
  if (expected) {
    if (is.null(cond)) {
      if (is.null(cond_class)) {
        sprintf("%s did not throw the expected %s.", lab, cond_type)
      } else {
        sprintf("%s did not throw a %s with class <%s>.", lab, cond_type, cond_class)
      }
    } else {
      NULL
    }
  } else {
    if (!is.null(cond)) {
      sprintf(
        "%s threw an unexpected %s.\nMessage: %s\nClass:   %s",
        lab,
        cond_type,
        cnd_message(cond),
        paste(class(cond), collapse = "/")
      )
    } else {
      NULL
    }
  }
}

compare_condition_2e <- function(cond,
                                 lab,
                                 regexp = NULL,
                                 class = NULL,
                                 ...,
                                 inherit = TRUE,
                                 cond_type = "error") {

  # Expecting no condition
  if (identical(regexp, NA)) {
    if (!is.null(cond)) {
      return(sprintf(
        "%s threw an %s.\nMessage: %s\nClass:   %s",
        lab,
        cond_type,
        cnd_message(cond),
        paste(class(cond), collapse = "/")
      ))
    } else {
      return()
    }
  }

  # Otherwise we're definitely expecting a condition
  if (is.null(cond)) {
    return(sprintf("%s did not throw an %s.", lab, cond_type))
  }

  matches <- cnd_matches_2e(cond, class, regexp, inherit, ...)
  ok_class <- matches[["class"]]
  ok_msg <- matches[["msg"]]

  # All good
  if (ok_msg && ok_class) {
    return()
  }

  problems <- c(if (!ok_class) "class", if (!ok_msg) "message")
  message <- cnd_message(cond)

  details <- c(
    if (!ok_class) {
      sprintf(
        "Expected class: %s\nActual class:   %s\nMessage:        %s",
        paste0(class, collapse = "/"),
        paste0(class(cond), collapse = "/"),
        message
      )
    },
    if (!ok_msg) {
      sprintf(
        "Expected match: %s\nActual message: %s",
        encodeString(regexp, quote = '"'),
        encodeString(message, quote = '"')
      )
    }
  )

  sprintf(
    "%s threw an %s with unexpected %s.\n%s",
    lab,
    cond_type,
    paste(problems, collapse = " and "),
    paste(details, collapse = "\n")
  )
}

cnd_matches_2e <- function(cnd, class, regexp, inherit, ...) {
  if (!inherit) {
    cnd$parent <- NULL
  }

  ok_class <- is.null(class) || cnd_inherits(cnd, class)
  ok_msg <- is.null(regexp) || cnd_some(cnd, function(x) {
    any(grepl(regexp, cnd_message(x), ...))
  })

  c(class = ok_class, msg = ok_msg)
}


compare_messages <- function(messages,
                             lab,
                             regexp = NA, ...,
                             all = FALSE,
                             cond_type = "messages") {
  bullets <- paste0("* ", messages, collapse = "\n")
  # Expecting no messages
  if (identical(regexp, NA)) {
    if (length(messages) > 0) {
      return(sprintf("%s generated %s:\n%s", lab, cond_type, bullets))
    } else {
      return()
    }
  }

  # Otherwise we're definitely expecting messages
  if (length(messages) == 0) {
    return(sprintf("%s did not produce any %s.", lab, cond_type))
  }

  if (is.null(regexp)) {
    return()
  }

  matched <- grepl(regexp, messages, ...)

  # all/any ok
  if ((all && all(matched)) || (!all && any(matched))) {
    return()
  }

  sprintf(
    "%s produced unexpected %s.\n%s\n%s",
    lab,
    cond_type,
    paste0("Expected match: ", encodeString(regexp)),
    paste0("Actual values:\n", bullets)
  )
}

# Disable rlang backtrace reminders so they don't interfere with
# expected error messages
cnd_message <- function(x) {
  withr::local_options(rlang_backtrace_on_error = "none")
  conditionMessage(x)
}

test_that("old school types still work", {
  local_edition(2L)

  expect_success(expect_that(1L, is_a("integer")))
})

test_that("old school names still work", {
  local_edition(2L)

  expect_success(expect_that("a", has_names(NULL)))
})

test_that("old school comparisons still work", {
  local_edition(2L)

  expect_success(expect_that(10, is_less_than(11)))
  expect_failure(expect_that(10, is_more_than(11)))
})

test_that("old school equality tests still work", {
  local_edition(2L)

  expect_success(expect_that(10, equals(10)))
  expect_success(expect_that(10, is_identical_to(10)))
  expect_success(expect_that(10, is_equivalent_to(10)))
})

test_that("old school output tests still work", {
  local_edition(2L)

  expect_success(expect_that(stop("!"), throws_error()))
  expect_success(expect_that(warning("!"), gives_warning()))
  expect_success(expect_that(message("!"), shows_message()))
  expect_success(expect_that(print("!"), prints_text()))
})

#' Report results in Teamcity format
#'
#' This reporter will output results in the Teamcity message format.
#' For more information about Teamcity messages, see
#' http://confluence.jetbrains.com/display/TCD7/Build+Script+Interaction+with+TeamCity
#'
#' @export
#' @family reporters
TeamcityReporter <- R6::R6Class(
  "TeamcityReporter",
  inherit = Reporter,
  public = list(
    i = NA_integer_,

    initialize = function(...) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
    },

    start_context = function(context) {
      private$report_event("testSuiteStarted", context)
    },

    end_context = function(context) {
      private$report_event("testSuiteFinished", context)
      self$cat_line()
      self$cat_line()
    },

    start_test = function(context, test) {
      private$report_event("testSuiteStarted", test)
      self$i <- 1L
    },

    end_test = function(context, test) {
      private$report_event("testSuiteFinished", test)
      self$cat_line()
    },

    add_result = function(context, test, result) {
      testName <- paste0("expectation ", self$i)
      self$i <- self$i + 1L

      if (expectation_skip(result)) {
        private$report_event("testIgnored", testName, message = format(result))
        return()
      }

      private$report_event("testStarted", testName)

      if (!expectation_ok(result)) {
        lines <- strsplit(format(result), "\n")[[1]]

        private$report_event(
          "testFailed",
          testName,
          message = lines[1],
          details = paste(lines[-1], collapse = "\n")
        )
      }
      private$report_event("testFinished", testName)
    }
  ),

  private = list(
    report_event = function(event, name, ...) {
      values <- list(name = name, ...)

      values <- map_chr(values, teamcity_escape)
      if (length(values) == 0) {
        value_string <- ""
      } else {
        value_string <- paste0(names(values), "='", values, "'", collapse = " ")
      }

      self$cat_line("##teamcity[", event, " ", value_string, "]")
    }
  )
)

# teamcity escape character is |
teamcity_escape <- function(s) {
  s <- gsub("(['|]|\\[|\\])", "|\\1", s)
  gsub("\n", "|n", s)
}

# Source: https://github.com/rstudio/shiny/blob/master/R/stack.R
# License: GPL-3
# Relicensed a MIT with permission.

# A Stack object backed by a list. The backing list will grow or shrink as
# the stack changes in size.
Stack <- R6Class(
  "Stack",
  class = FALSE,
  public = list(
    initialize = function(init = 20L) {
      # init is the initial size of the list. It is also used as the minimum
      # size of the list as it shrinks.
      private$stack <- vector("list", init)
      private$init <- init
      private$count <- 0L
    },

    push = function(..., .list = NULL) {
      args <- c(list(...), .list)
      new_size <- private$count + length(args)

      # Grow if needed; double in size
      while (new_size > length(private$stack)) {
        private$stack[length(private$stack) * 2L] <- list(NULL)
      }
      private$stack[private$count + seq_along(args)] <- args
      private$count <- new_size

      invisible(self)
    },

    size = function() {
      private$count
    },

    # Return the entire stack as a list, where the first item in the list is the
    # oldest item in the stack, and the last item is the most recently added.
    as_list = function() {
      private$stack[seq_len(private$count)]
    }
  ),

  private = list(
    stack = NULL, # A list that holds the items
    count = 0L, # Current number of items in the stack
    init = 20L # Initial and minimum size of the stack
  )
)

#' Do you expect printed output to match this pattern?
#'
#' Test for output produced by `print()` or `cat()`. This is best used for
#' very simple output; for more complex cases use [expect_snapshot()].
#'
#' @export
#' @family expectations
#' @inheritParams expect_that
#' @param regexp Regular expression to test against.
#'   * A character vector giving a regular expression that must match the output.
#'   * If `NULL`, the default, asserts that there should output,
#'     but doesn't check for a specific value.
#'   * If `NA`, asserts that there should be no output.
#' @inheritDotParams expect_match -object -regexp -info -label
#' @inheritParams capture_output
#' @return The first argument, invisibly.
#' @examples
#' str(mtcars)
#' expect_output(str(mtcars), "32 obs")
#' expect_output(str(mtcars), "11 variables")
#'
#' # You can use the arguments of grepl to control the matching
#' expect_output(str(mtcars), "11 VARIABLES", ignore.case = TRUE)
#' expect_output(str(mtcars), "$ mpg", fixed = TRUE)
expect_output <- function(
  object,
  regexp = NULL,
  ...,
  info = NULL,
  label = NULL,
  width = 80
) {
  check_number_whole(width, min = 1)

  act <- quasi_capture(enquo(object), label, capture_output, width = width)

  if (identical(regexp, NA)) {
    if (!identical(act$cap, "")) {
      msg <- c(
        sprintf("Expected %s to produce no output.", act$lab),
        sprintf("Actual output:\n%s", encodeString(act$cap))
      )
      fail(msg, info = info)
    } else {
      pass()
    }
  } else if (is.null(regexp) || identical(act$cap, "")) {
    if (identical(act$cap, "")) {
      msg <- sprintf("Expected %s to produce output.", act$lab)
      fail(msg, info = info)
    } else {
      pass()
    }
  } else {
    act_out <- labelled_value(act$cap, paste0("output from ", act$lab))
    expect_match_(act_out, enc2native(regexp), ..., title = "output")
  }

  invisible(act$val)
}

capture_success_failure <- function(expr) {
  cnd <- NULL

  n_success <- 0
  n_failure <- 0

  last_failure <- NULL

  withCallingHandlers(
    expr,
    expectation_failure = function(cnd) {
      last_failure <<- cnd
      n_failure <<- n_failure + 1
      invokeRestart("muffle_expectation")
    },
    expectation_success = function(cnd) {
      n_success <<- n_success + 1
      invokeRestart("muffle_expectation")
    }
  )

  list(
    n_success = n_success,
    n_failure = n_failure,
    last_failure = last_failure
  )
}

format_success_failure <- function(status, exp_n_success, exp_n_failure) {
  pluralise <- function(n, singular, plural) {
    paste(n, ngettext(n, singular, plural))
  }

  tick <- cli::col_green(cli::symbol$tick)
  cross <- cli::col_red(cli::symbol$cross)

  success_ok <- status$n_success == exp_n_success
  failure_ok <- status$n_failure == exp_n_failure

  c(
    sprintf(
      "Expected %s and %s.",
      pluralise(exp_n_success, "success", "successes"),
      pluralise(exp_n_failure, "failure", "failures")
    ),
    sprintf(
      "%s Observed %s.",
      if (success_ok) tick else cross,
      pluralise(status$n_success, "success", "successes")
    ),
    sprintf(
      "%s Observed %s.",
      if (failure_ok) tick else cross,
      pluralise(status$n_failure, "failure", "failures")
    )
  )
}

#' Test your custom expectations
#'
#' @description
#' `expect_success()` checks that there's exactly one success and no failures;
#' `expect_failure()` checks that there's exactly one failure and no successes.
#' `expect_snapshot_failure()` records the failure message so that you can
#' manually check that it is informative.
#'
#' Use `show_failure()` in examples to print the failure message without
#' throwing an error.
#'
#' @param expr Code to evaluate
#' @param message Check that the failure message matches this regexp.
#' @param ... Other arguments passed on to [expect_match()].
#' @export
expect_success <- function(expr) {
  status <- capture_success_failure(expr)

  if (status$n_success == 1 && status$n_failure == 0) {
    pass()
    return(invisible())
  }
  fail(format_success_failure(status, exp_n_success = 1, exp_n_failure = 0))
  invisible()
}

#' @export
#' @rdname expect_success
expect_failure <- function(expr, message = NULL, ...) {
  status <- capture_success_failure(expr)

  if (status$n_failure == 1 && status$n_success == 0) {
    if (is.null(message)) {
      pass()
    } else {
      act <- labelled_value(status$last_failure$message, "failure message")
      expect_match_(act, message, ..., title = "message")
    }
    return(invisible())
  }

  fail(format_success_failure(status, exp_n_success = 0, exp_n_failure = 1))
  invisible()
}

#' @export
#' @rdname expect_success
expect_snapshot_failure <- function(expr) {
  expr <- enquo0(expr)
  expect_snapshot_(expr, error = TRUE, error_class = "expectation_failure")
}

#' Test for absence of success or failure
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' These functions are deprecated because [expect_success()] and
#' [expect_failure()] now test for exactly one success or no failures, and
#' exactly one failure and no successes.
#'
#' @keywords internal
#' @export
expect_no_success <- function(expr) {
  lifecycle::deprecate_warn("3.3.0", "expect_no_success()", "expect_failure()")
  status <- capture_success_failure(expr)

  if (status$n_success > 0) {
    fail("Expectation succeeded")
  } else {
    pass()
  }
  invisible()
}

#' @export
#' @rdname expect_no_success
expect_no_failure <- function(expr) {
  lifecycle::deprecate_warn("3.3.0", "expect_no_failure()", "expect_success()")
  status <- capture_success_failure(expr)

  if (status$n_failure > 0) {
    fail("Expectation failed")
  } else {
    pass()
  }
  invisible()
}

expect_snapshot_skip <- function(x, cran = FALSE) {
  expect_snapshot_condition_("skip", x)
}
expect_skip <- function(code) {
  expect_condition_matching_("skip", code)
}
expect_no_skip <- function(code) {
  expect_no_("skip", code)
}


#' @export
#' @rdname expect_success
show_failure <- function(expr) {
  exp <- capture_expectation(expr)

  if (!is.null(exp) && expectation_failure(exp)) {
    cat(cli::style_bold("Failed expectation:\n"))
    cat(exp$message, "\n", sep = "")
  }

  invisible()
}

expect_snapshot_reporter <- function(
  reporter,
  paths = test_path("reporters/tests.R")
) {
  local_options(rlang_trace_format_srcrefs = FALSE)
  withr::local_seed(1014)

  expect_snapshot_output(
    with_reporter(reporter, {
      for (path in paths) {
        test_one_file(path)
      }
    })
  )
}

# Use specifically for testthat tests in order to override the
# defaults found when starting the reporter
local_output_override <- function(
  width = 80,
  crayon = TRUE,
  unicode = TRUE,
  .env = parent.frame()
) {
  reporter <- get_reporter()
  if (is.null(reporter)) {
    return()
  }

  old_width <- reporter$width
  old_crayon <- reporter$crayon
  old_unicode <- reporter$unicode

  reporter$width <- width
  reporter$crayon <- crayon
  reporter$unicode <- unicode

  withr::defer(
    {
      reporter$width <- old_width
      reporter$crayon <- old_crayon
      reporter$unicode <- old_unicode
    },
    .env
  )
}

## -----------------------------------------------------------------------------
library(testthat)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

# Pretend we're snapshotting
snapper <- local_snapshotter(fail_on_new = FALSE)
snapper$start_file("snapshotting.Rmd", "test")

# Pretend we're testing testthat so we can use mocking
Sys.setenv(TESTTHAT_PKG = "testthat")

## -----------------------------------------------------------------------------
check_installed <- function(pkg, min_version = NULL) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(sprintf("{%s} is not installed.", pkg))
  }
  if (!is.null(min_version)) {
    pkg_version <- packageVersion(pkg)
    if (pkg_version < min_version) {
      stop(sprintf(
        "{%s} version %s is installed, but %s is required.", 
        pkg, 
        pkg_version, 
        min_version
      ))
    }
  }

  invisible()
}

## -----------------------------------------------------------------------------
test_that("check_installed() checks package is installed", {
  expect_no_error(check_installed("testthat"))
  expect_snapshot(check_installed("doesntexist"), error = TRUE)
})

## -----------------------------------------------------------------------------
test_that("check_installed() checks minimum version", {
  expect_no_error(check_installed("testthat", "1.0.0"))
  expect_snapshot(check_installed("testthat", "99.99.999"), error = TRUE)
})

## -----------------------------------------------------------------------------
test_that("check_installed() checks minimum version", {
  expect_no_error(check_installed("testthat", "1.0.0"))
  expect_snapshot(
    check_installed("testthat", "99.99.999"), 
    error = TRUE, 
    transform = function(lines) gsub(packageVersion("testthat"), "<version>", lines)
  )
})

## -----------------------------------------------------------------------------
requireNamespace <- NULL
packageVersion <- NULL

## -----------------------------------------------------------------------------
test_that("check_installed() checks package is installed", {
  local_mocked_bindings(requireNamespace = function(...) TRUE)
  expect_no_error(check_installed("package-name"))

  local_mocked_bindings(requireNamespace = function(...) FALSE)
  expect_snapshot(check_installed("package-name"), error = TRUE)
})

## -----------------------------------------------------------------------------
test_that("check_installed() checks minimum version", {
  local_mocked_bindings(
    requireNamespace = function(...) TRUE,
    packageVersion = function(...) numeric_version("2.0.0")
  )
  
  expect_no_error(check_installed("package-name", "1.0.0"))
  expect_snapshot(check_installed("package-name", "3.4.5"), error = TRUE)
})

## -----------------------------------------------------------------------------
system_os <- NULL

## -----------------------------------------------------------------------------
# test_that("can skip on multiple oses", {
#   local_mocked_bindings(system_os = function() "windows")
# 
#   expect_skip(skip_on_os("windows"))
#   expect_skip(skip_on_os(c("windows", "linux")))
#   expect_no_skip(skip_on_os("linux"))
# })

## -----------------------------------------------------------------------------
# local_mocked_bindings(
#   get_revdeps = function() character(),
#   gh_milestone_number = function(...) NA
# )

## -----------------------------------------------------------------------------
unix_time <- function() unclass(Sys.time())
unix_time()

## -----------------------------------------------------------------------------
elapsed <- function() {
  start <- unix_time()
  function() {
    unix_time() - start
  }
}

timer <- elapsed()
Sys.sleep(0.5)
timer()

## -----------------------------------------------------------------------------
test_that("elapsed() measures elapsed time", {
  time <- 1
  local_mocked_bindings(unix_time = function() time)

  timer <- elapsed()
  expect_equal(timer(), 0)

  time <- 2
  expect_equal(timer(), 1)
})

## -----------------------------------------------------------------------------
# old <- getFromNamespace("my_function", "mypackage")
# assignInNamespace("my_function", new, "mypackage")
# 
# # run the test...
# 
# # restore the previous value
# assignInNamespace("my_function", old, "mypackage")

## ----setup--------------------------------------------------------------------
library(testthat)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

# Pretend we're snapshotting
snapper <- local_snapshotter(fail_on_new = FALSE)
snapper$start_file("snapshotting.Rmd", "test")

## -----------------------------------------------------------------------------
expect_df <- function(tbl) {
  expect_s3_class(tbl, "data.frame")
}

## -----------------------------------------------------------------------------
# from tidytext
expect_nrow <- function(tbl, n) {
  expect_s3_class(tbl, "data.frame")
  expect_equal(nrow(tbl), n)
}

## -----------------------------------------------------------------------------
try({
test_that("success", {
  expect_nrow(mtcars, 32)
})

test_that("failure 1", {
  expect_nrow(mtcars, 30)
})

test_that("failure 2", {
  expect_nrow(matrix(1:5), 2)
})
})

## -----------------------------------------------------------------------------
expect_length <- function(object, n) {  
  # 1. Capture object and label
  act <- quasi_label(rlang::enquo(object))
  
  act_n <- length(act$val)
  if (act_n != n) {
    # 2. Fail if expectations are violated
    fail(c(
      sprintf("Expected %s to have length %i.", act$lab, n),
      sprintf("Actual length: %i.", act_n)
    ))
  } else {
    # 3. Pass if expectations are met
    pass()
  }
  
  # 4. Invisibly return the input value
  invisible(act$val)
}

## -----------------------------------------------------------------------------
test_that("mtcars is a 13 row data frame", {
  mtcars |>
    expect_type("list") |>
    expect_s3_class("data.frame") |> 
    expect_length(11)
})

## -----------------------------------------------------------------------------
test_that("expect_length works as expected", {
  x <- 1:10
  expect_success(expect_length(x, 10))
  expect_failure(expect_length(x, 11))
})

test_that("expect_length gives useful feedback", {
  x <- 1:10
  expect_snapshot_failure(expect_length(x, 11))
})

## -----------------------------------------------------------------------------
expect_length(mean, 1)

## -----------------------------------------------------------------------------
expect_vector_length <- function(object, n) {  
  act <- quasi_label(rlang::enquo(object))

  # It's non-trivial to check if an object is a vector in base R so we
  # use an rlang helper
  if (!rlang::is_vector(act$val)) {
    fail(c(
      sprintf("Expected %s to be a vector", act$lab),
      sprintf("Actual type: %s", typeof(act$val))
    ))
  } else {
    act_n <- length(act$val)
    if (act_n != n) {
      fail(c(
        sprintf("Expected %s to have length %i.", act$lab, n),
        sprintf("Actual length: %i.", act_n)
      ))
    } else {
      pass()
    }
  }

  invisible(act$val)
}

## -----------------------------------------------------------------------------
try({
expect_vector_length(mean, 1)
expect_vector_length(mtcars, 15)
})

## -----------------------------------------------------------------------------
expect_s3_class <- function(object, class) {
  if (!rlang::is_string(class)) {
    rlang::abort("`class` must be a string.")
  }

  act <- quasi_label(rlang::enquo(object))

  if (!is.object(act$val)) {
    fail(sprintf("Expected %s to be an object.", act$lab))
  } else if (isS4(act$val)) {
    fail(c(
      sprintf("Expected %s to be an S3 object.", act$lab),
      "Actual OO type: S4"
    ))
  } else if (!inherits(act$val, class)) {
    fail(c(
      sprintf("Expected %s to inherit from %s.", act$lab, class),
      sprintf("Actual class: %s", class(act$val))
    ))
  } else {
    pass()
  }

  invisible(act$val)
}

## -----------------------------------------------------------------------------
try({
x1 <- 1:10
TestClass <- methods::setClass("Test", contains = "integer")
x2 <- TestClass()
x3 <- factor()

expect_s3_class(x1, "integer")
expect_s3_class(x2, "integer")
expect_s3_class(x3, "integer")
expect_s3_class(x3, "factor")
})

## -----------------------------------------------------------------------------
try({
expect_s3_class(x1, 1)
})

## -----------------------------------------------------------------------------
expect_s3_object <- function(object, class = NULL) {
  if (!rlang::is_string(class) && is.null(class)) {
    rlang::abort("`class` must be a string or NULL.")
  }

  act <- quasi_label(rlang::enquo(object))

  if (!is.object(act$val)) {
    fail(sprintf("Expected %s to be an object.", act$lab))
  } else if (isS4(act$val)) {
    fail(c(
      sprintf("Expected %s to be an S3 object.", act$lab),
      "Actual OO type: S4"
    ))
  } else if (!is.null(class) && !inherits(act$val, class)) {
    fail(c(
      sprintf("Expected %s to inherit from %s.", act$lab, class),
      sprintf("Actual class: %s", class(act$val))
    ))
  } else {
    pass()
  }

  invisible(act$val)
}

## -----------------------------------------------------------------------------
expect_length_ <- function(act, n, trace_env = caller_env()) {
  act_n <- length(act$val)
  if (act_n != n) {
    fail(
      sprintf("%s has length %i, not length %i.", act$lab, act_n, n), 
      trace_env = trace_env
    )
  } else {
    pass()
  }
}

expect_length <- function(object, n) {  
  act <- quasi_label(rlang::enquo(object))

  expect_length_(act, n)
  invisible(act$val)
}

test_that("basical principles of equality hold", {
  local_edition(2)
  expect_success(expect_equal(1, 1))
  expect_failure(expect_equal(1, 2))
  expect_success(expect_identical(1, 1))
  expect_failure(expect_identical(1, 2))

  local_edition(3)
  expect_success(expect_equal(1, 1))
  expect_failure(expect_equal(1, 2))
  expect_success(expect_identical(1, 1))
  expect_failure(expect_identical(1, 2))
})

test_that("expect_equal() ignores numeric type; expect_identical() does not", {
  local_edition(2)
  expect_success(expect_equal(1, 1L))
  expect_failure(expect_identical(1, 1L))

  local_edition(3)
  expect_success(expect_equal(1, 1L))
  expect_failure(expect_identical(1, 1L))
})

test_that("returns value", {
  one <- 1

  local_edition(3)
  expect_equal(expect_equal(one, one), 1)
  expect_equal(expect_identical(one, one), 1)

  local_edition(2)
  expect_equal(expect_equal(one, one), 1)
  expect_equal(expect_identical(one, one), 1)
})

test_that("can control numeric tolerance", {
  x1 <- 1
  x2 <- x1 + 1e-6

  local_edition(2)
  expect_failure(expect_equal(x1, x2))
  expect_success(expect_equal(x1, x2, tolerance = 1e-5))
  expect_success(expect_equivalent(x1, x2, tolerance = 1e-5))

  # with partial matching
  withr::local_options(warnPartialMatchArgs = FALSE)
  expect_success(expect_equal(x1, x2, tol = 1e-5))

  local_edition(3)
  expect_failure(expect_equal(x1, x2))
  expect_success(expect_equal(x1, x2, tolerance = 1e-5))
})

test_that("second edition only optionally sets tolerance", {
  local_edition(2)

  # all.equal.POSIXct sets default tolerance to 0.001
  x <- .POSIXct(1)
  y <- .POSIXct(1 + 1e-4)
  expect_success(expect_equal(x, y))
})

test_that("provide useful feedback on failure (3e)", {
  x <- 1
  expect_snapshot_failure(expect_identical(x, "a"))
  expect_snapshot_failure(expect_equal(x, "a"))

  local_edition(2)
  withr::local_options(testthat.edition_ignore = TRUE)
  expect_snapshot_failure(expect_identical(x, "a"))
  expect_snapshot_failure(expect_equal(x, "a"))
})

test_that("correctly spaces lines", {
  expect_snapshot_failure(expect_equal(list(a = 1), list(a = "b", b = 10)))
})

test_that("provide useful feedback on failure (2e)", {
  local_edition(2)
  withr::local_options(testthat.edition_ignore = TRUE)

  x <- 1
  expect_snapshot_failure(expect_identical(x, "a"))
  expect_snapshot_failure(expect_equal(x, "a"))
})

test_that("default labels use unquoting", {
  x <- 1
  y <- 2
  expect_snapshot_failure(expect_equal(x, !!y))
})

test_that("% is not treated as sprintf format specifier (#445)", {
  expect_failure(expect_equal("+", "%"))
  expect_failure(expect_equal("%", "+"))
  expect_success(expect_equal("%", "%"))
})

# 2nd edition ---------------------------------------------------

test_that("useful message if objects equal but not identical", {
  local_edition(2)

  f <- function() x
  environment(f) <- new_environment()
  g <- function() x
  environment(g) <- new_environment()

  expect_snapshot_failure(expect_identical(f, g))
})

test_that("attributes for object (#452)", {
  local_edition(2)

  oops <- structure(0, oops = "oops")
  expect_equal(oops, oops)
  expect_snapshot_failure(expect_equal(oops, 0))
  expect_equal(as.numeric(oops), 0)
})

test_that("expect_equivalent ignores attributes and numeric differences", {
  local_edition(2)

  x <- y <- 1
  attr(y, "y") <- y
  expect_success(expect_equivalent(x, y))
  expect_success(expect_equivalent(x, 1L))
})

test_that("expect_equivalent returns value", {
  local_edition(2)
  one <- 1

  expect_equal(expect_equivalent(one, one), 1)
})

test_that("expect_equal validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_equal(1, 2, tolerance = "high")
    expect_equal(1, 2, tolerance = -1)
  })
})

test_that("basic skips work as expected", {
  expect_snapshot_skip(skip())

  expect_no_skip(skip_if(FALSE))
  expect_snapshot_skip(skip_if(TRUE))

  expect_no_skip(skip_if_not(TRUE))
  expect_snapshot_skip(skip_if_not(FALSE))

  expect_snapshot_skip(skip_empty())
})

test_that("autogenerated message is always single line", {
  a_very_long_argument_name <- FALSE

  cnd <- capture_condition(skip_if_not(
    a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name ||
      a_very_long_argument_name
  ))

  expect_length(cnd$message, 1)
  # ensure the message is not repeated, #1290
  expect_snapshot_output(cat(cnd$message))
})

# skip helpers ------------------------------------------------------------

test_that("skip_if_not_installed() works as expected, absence and version", {
  local_mocked_bindings(package_version = function(x) "3.0.0")

  expect_snapshot_skip(skip_if_not_installed("doesntexist"))
  expect_snapshot_skip(skip_if_not_installed("testthat", "9999.9999.999"))
  expect_no_skip(skip_if_not_installed("testthat", "1.0.0"))
})

test_that("skip_if_not_installed() works as expected, offline", {
  skip_if_not_installed("curl")
  skip_on_cran()

  expect_no_skip(skip_if_offline())

  local_mocked_bindings(has_internet = function(host) FALSE)
  expect_snapshot_skip(skip_if_offline())
})

test_that("skip_on_cran() works as expected", {
  local({
    local_on_cran(FALSE)
    expect_no_skip(skip_on_cran())
  })

  local({
    local_on_cran(TRUE)
    expect_skip(skip_on_cran())
  })

  withr::local_envvar(NOT_CRAN = NA)
  local_mocked_bindings(interactive = function() TRUE)
  expect_no_skip(skip_on_cran())

  local_mocked_bindings(interactive = function() FALSE)
  expect_skip(skip_on_cran())
})

test_that("local_on_cran sets NOT_CRAN", {
  local({
    local_on_cran(TRUE)
    expect_equal(on_cran(), TRUE)
    expect_equal(Sys.getenv("NOT_CRAN"), "false")
  })

  local({
    local_on_cran(FALSE)
    expect_equal(on_cran(), FALSE)
    expect_equal(Sys.getenv("NOT_CRAN"), "true")
  })
})

test_that("local_assume_not_on_cran() sets NOT_CRAN if not already set", {
  withr::local_envvar(NOT_CRAN = NA)
  local({
    local_assume_not_on_cran()
    expect_equal(Sys.getenv("NOT_CRAN"), "true")
  })

  withr::local_envvar(NOT_CRAN = "false")
  local({
    local_assume_not_on_cran()
    expect_equal(Sys.getenv("NOT_CRAN"), "false")
  })
})

test_that("skip_on_ci() works as expected", {
  withr::local_envvar(CI = "false")
  expect_no_skip(skip_on_ci())

  withr::local_envvar(CI = "true")
  expect_snapshot_skip(skip_on_ci())
})

test_that("skip_on_covr() works as expected", {
  withr::local_envvar(R_COVR = "false")
  expect_no_skip(skip_on_covr())

  withr::local_envvar(R_COVR = "true")
  expect_snapshot_skip(skip_on_covr())
})

test_that("skip_on_bioc() works as expected", {
  expect_no_skip(skip_on_bioc())

  withr::local_envvar(IS_BIOC_BUILD_MACHINE = "true")
  expect_snapshot_skip(skip_on_bioc())
})

test_that("superseded CI skips still work", {
  expect_no_skip(skip_on_travis())
  expect_no_skip(skip_on_appveyor())

  withr::local_envvar(TRAVIS = "true", APPVEYOR = "true")
  expect_snapshot_skip(skip_on_travis())
  expect_snapshot_skip(skip_on_appveyor())
})

test_that("skip_if_translated() works as expected", {
  local_mocked_bindings(gettext = function(msg, ...) msg)
  expect_no_skip(skip_if_translated())

  local_mocked_bindings(gettext = function(msg, ...) toupper(msg))
  expect_snapshot_skip(skip_if_translated())
})

# skip_on_os() ------------------------------------------------------------

test_that("skip on os checks os names", {
  expect_snapshot(skip_on_os("amiga"), error = TRUE)
})

test_that("can skip on multiple oses", {
  local_mocked_bindings(system_os = function() "windows")

  expect_snapshot_skip(skip_on_os("windows"))
  expect_snapshot_skip(skip_on_os(c("windows", "linux")))
  expect_no_skip(skip_on_os("linux"))
  expect_no_skip(skip_on_os("mac"))
  expect_no_skip(skip_on_os("emscripten"))
})

test_that("can refine os with arch", {
  local_mocked_bindings(
    system_os = function() "windows",
    system_arch = function() "i386"
  )

  expect_snapshot_skip(skip_on_os("windows"))
  expect_snapshot_skip(skip_on_os("windows", "i386"))
  expect_no_skip(skip_on_os("windows", "x86_64"))
  expect_no_skip(skip_on_os("linux", "i386"))
})

test_that("skip_unless_r works as expected", {
  expect_no_skip(skip_unless_r(">= 0.0.0"))
  expect_no_skip(skip_unless_r(paste("==", getRversion())))
  expect_no_skip(skip_unless_r("<= 999.999.999"))

  expect_skip(skip_unless_r(">= 999.999.999"))
  expect_skip(skip_unless_r("== 0.0.0"))
  expect_skip(skip_unless_r("<= 0.0.0"))

  expect_snapshot(error = TRUE, skip_unless_r("idfjdij"))
})

test_that("skip_unless_r gives the expected output", {
  local_mocked_bindings(getRversion = \() numeric_version("4.5.0"))

  expect_snapshot_skip(skip_unless_r(">= 999.999.999"))
  expect_snapshot_skip(skip_unless_r("== 0.0.0"))
})

test_that("returns condition or value", {
  expect_equal(expect_error(1, NA), 1)
  expect_s3_class(expect_error(stop("!")), "simpleError")
})

test_that("regexp = NULL checks for presence of error", {
  expect_success(expect_error(stop()))

  f <- function() {}
  expect_snapshot_failure(expect_error(f()))
})

test_that("regexp = NA checks for absence of error", {
  expect_success(expect_error({}, NA))

  f <- function() stop("Yes")
  expect_snapshot_failure(expect_error(f(), NA))
})

test_that("regexp = string matches for error message", {
  expect_success(expect_error(stop("Yes"), "Yes"))
  expect_error(expect_error(stop("Yes"), "No"))

  f <- function() {}
  expect_snapshot_failure(expect_error(f(), "No"))
})

test_that("class = string matches class of error", {
  blah <- function() {
    abort("hi", class = c("blah", "error", "condition"))
  }

  expect_success(expect_error(blah(), class = "blah"))
  # otherwise bubbles up
  expect_error(expect_error(blah(), class = "blech"), class = "blah")
})

test_that("base_class must match when class is set", {
  foo <- function() warn("foo", class = "bar")
  expect_warning(expect_failure(expect_error(foo(), class = "bar")))
  expect_success(expect_warning(foo(), class = "bar"))
})

test_that("expect_error validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_error(stop("!"), regexp = 1)
    expect_error(stop("!"), class = 1)
    expect_error(stop("!"), inherit = "yes")
  })
})

test_that("... passed on to grepl", {
  expect_success(expect_error(stop("X"), "x", ignore.case = TRUE))
})

test_that("message method is called when expecting error", {
  local_bindings(
    conditionMessage.foobar = function(err) "dispatched!",
    .env = globalenv()
  )
  fb <- function() abort("foobar", "foobar")

  expect_error(fb(), "dispatched!", class = "foobar")
  expect_snapshot_failure(expect_error(fb(), NA))
})

test_that("rlang backtrace reminders are not included in error message", {
  f <- function() g()
  g <- function() h()
  h <- function() abort("foo")
  expect_error(f(), "foo$")
})

test_that("can capture Throwable conditions from rJava", {
  local_bindings(
    conditionMessage.Throwable = function(c, ...) unclass(c)$message,
    conditionCall.Throwable = function(c, ...) unclass(c)$call,
    `$.Throwable` = function(...) stop("forbidden"),
    `$<-.Throwable` = function(...) stop("forbidden"),
    .env = globalenv()
  )

  throw <- function(msg) stop(error_cnd("Throwable", message = msg))
  expect_error(throw("foo"), "foo", class = "Throwable")
})

test_that("capture correct trace_env (#1994)", {
  # This should fail, not error
  status <- capture_success_failure({
    stop("oops") |> expect_error() |> expect_warning()
  })
  expect_equal(status$n_success, 1) # from expect_error()
  expect_equal(status$n_failure, 1) # from expect_warning()

  status <- capture_success_failure({
    stop("oops") |> expect_error() |> expect_warning()
  })
  expect_equal(status$n_success, 1) # from expect_error()
  expect_equal(status$n_failure, 1) # from expect_warning()
})

# expect_warning() ----------------------------------------------------------

test_that("warnings are converted to errors when options('warn') >= 2", {
  withr::with_options(c(warn = 2), {
    expect_warning(warning("foo"))
    expect_error(warning("foo"))
  })
})

test_that("can silence warnings", {
  expect_warning(suppressWarnings(warning("foo")), NA)

  # Can't test with `expect_warning()` because the warning is still
  # signalled, it's just not printed
  # https://github.com/wch/r-source/blob/886ab4a0/src/main/errors.c#L388-L484
  withr::with_options(c(warn = -1), warning("foo"))
})

test_that("when checking for no warnings, exclude deprecation warnings", {
  foo <- function() {
    lifecycle::deprecate_warn("1.0.0", "foo()")
  }
  expect_warning(
    expect_warning(foo(), NA),
    class = "lifecycle_warning_deprecated"
  )
})

test_that("when checking for no warnings, exclude deprecation warnings (2e)", {
  local_edition(2)

  foo <- function() {
    options(lifecycle_verbosity = "warning")
    lifecycle::deprecate_warn("1.0.0", "foo()")
  }

  expect_warning(
    expect_warning(foo(), NA),
    class = "lifecycle_warning_deprecated"
  )
})

test_that("expect_warning validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_warning(warning("!"), regexp = 1)
    expect_warning(warning("!"), class = 1)
    expect_warning(warning("!"), inherit = "yes")
    expect_warning(warning("!"), all = "yes")
  })
})

# expect_message ----------------------------------------------------------

test_that("regexp = NA checks for absence of message", {
  expect_success(expect_message({}, NA))

  f <- \() message("!")
  expect_snapshot_failure(expect_message(f(), NA))
})

test_that("expect_message validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_message(message("!"), regexp = 1)
    expect_message(message("!"), class = 1)
    expect_message(message("!"), inherit = "yes")
    expect_message(message("!"), all = "yes")
  })
})


# expect_condition --------------------------------------------------------

test_that("continues evaluation", {
  expect_condition({
    message("Hi")
    new_variable <- 1
  })
  expect_equal(exists("new_variable"), TRUE)
})

test_that("but not after error", {
  expect_condition({
    stop("Hi")
    new_variable <- 1
  })
  expect_equal(exists("new_variable"), FALSE)
})

test_that("captured condition is muffled", {
  expect_message(expect_condition(message("Hi")), NA)
  expect_warning(expect_condition(warning("Hi")), NA)
  expect_error(expect_condition(stop("Hi")), NA)
})

test_that("condition class is included in failure", {
  f1 <- function() signal(class = "foo")
  expect_snapshot_failure(expect_condition(f1(), class = "bar"))
})

test_that("only matching condition is captured, others bubble up", {
  f1 <- function() {
    message("Hi")
    message("Bye")
  }
  expect_condition(expect_condition(f1(), "Hi"), "Bye")
  expect_condition(expect_condition(f1(), "Bye"), "Hi")

  f2 <- function() {
    message("Hi")
    stop("Bye")
  }
  expect_error(expect_condition(f2(), "Hi"), "Bye")
})

test_that("cnd expectations consistently return condition (#1371)", {
  f <- function(out, action) {
    action
    out
  }

  expect_s3_class(expect_message(f(NULL, message(""))), "simpleMessage")
  expect_s3_class(expect_warning(f(NULL, warning(""))), "simpleWarning")
  expect_s3_class(expect_error(f(NULL, stop(""))), "simpleError")

  # Used to behave differently with non-`NULL` values
  expect_s3_class(
    expect_message(f("return value", message(""))),
    "simpleMessage"
  )
  expect_s3_class(
    expect_warning(f("return value", warning(""))),
    "simpleWarning"
  )
  expect_s3_class(expect_error(f("return value", stop(""))), "simpleError")

  # If there is no condition expected we return the value
  expect_equal(
    expect_message(f("return value", NULL), regexp = NA),
    "return value"
  )
  expect_equal(
    expect_warning(f("return value", NULL), regexp = NA),
    "return value"
  )
  expect_equal(
    expect_error(f("return value", NULL), regexp = NA),
    "return value"
  )
})

test_that("cli width wrapping doesn't affect text matching", {
  skip_if_not_installed("cli", "3.0.2")
  skip_if_not_installed("rlang", "1.0.0")

  local_use_cli()

  expect_error(
    abort(
      "foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz"
    ),
    "foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz foobarbaz"
  )
})

test_that("can match parent conditions (#1493)", {
  parent <- error_cnd("foo", message = "Parent message.")
  f <- function() abort("Tilt.", parent = parent)

  expect_error(f(), class = "foo")
  expect_error(f(), "^Parent message.$")

  # Pattern and class must match the same condition
  expect_error(expect_error(f(), "Tilt.", class = "foo"))

  # Can disable parent matching
  expect_error(expect_error(f(), class = "foo", inherit = FALSE))
  expect_error(expect_error(f(), "Parent message.", inherit = FALSE))
})

test_that("expect_condition validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_condition(stop("!"), regexp = 1)
    expect_condition(stop("!"), class = 1)
    expect_condition(stop("!"), inherit = "yes")
  })
})

test_that("unused arguments generate an error", {
  expect_snapshot(error = TRUE, {
    expect_condition(stop("Hi!"), foo = "bar")
    expect_condition(stop("Hi!"), , , "bar")
    expect_condition(stop("Hi!"), , , "bar", fixed = TRUE)
    expect_condition(stop("Hi!"), "x", foo = "bar")
    expect_condition(stop("Hi!"), pattern = "bar", fixed = TRUE)
  })
})


# second edition ----------------------------------------------------------

test_that("other conditions are swallowed", {
  f <- function(...) {
    conds <- c(...)
    for (cond in conds) {
      switch(
        cond,
        message = message("message"),
        warning = warning("warning"),
        error = stop("error"),
        condition = signal("signal", class = "signal")
      )
    }
  }

  local_edition(2)
  # if condition text doesn't match, expectation fails (not errors)
  expect_failure(expect_error(f("error"), "not a match"))
  expect_failure(expect_warning(f("warning"), "not a match"))
  expect_failure(expect_message(f("message"), "not a match"))
  expect_failure(expect_condition(f("condition"), "not a match"))

  # if error/condition class doesn't match, expectation fails
  expect_failure(expect_error(f("error"), class = "not a match"))
  expect_failure(expect_condition(f("message"), class = "not a match"))

  # expect_message() and expect_warning() swallow all messages/warnings
  expect_message(expect_message(f("message", "message")), NA)
  expect_warning(expect_warning(f("warning", "warning")), NA)
})

test_that("can match parent conditions (edition 2, #1493)", {
  local_edition(2)

  parent <- error_cnd("foo", message = "Parent message.")
  f <- function() abort("Tilt.", parent = parent)

  expect_error(f(), class = "foo")
  expect_error(f(), "^Parent message.$")

  # Can disable parent matching
  expect_error(expect_error(f(), class = "foo", inherit = FALSE))
  expect_error(expect_error(f(), "Parent message.", inherit = FALSE))
})

f <- function() NULL
g <- function() cat("!")

test_that("expect = NA checks for no output", {
  expect_success(expect_output(f(), NA))
  expect_failure(expect_output(g(), NA), "produced output")
})

test_that("expect = NULL checks for some output", {
  expect_failure(expect_output(f(), NULL), "produced no output")
  expect_success(expect_output(g(), NULL))
})

test_that("expect = string checks for match", {
  expect_success(expect_output(g(), "!"))
  expect_failure(expect_output(g(), "x"), 'does not match "x"')
  expect_failure(expect_output("a", "x"), "produced no output")
})

test_that("multiline outputs captures and matches", {
  expect_success(expect_output(cat("1\n2"), "1\n2"))
})

test_that("expect_output sets width", {
  x <- expect_output(getOption("width"), NA)
  expect_equal(x, 80)

  x <- expect_output(getOption("width"), NA, width = 20)
  expect_equal(x, 20)
})

test_that("... passed on to grepl", {
  expect_success(expect_output(print("X"), "x", ignore.case = TRUE))
})

test_that("returns first argument", {
  expect_equal(expect_output(1, NA), 1)
})

test_that("uses unicode characters in output where available", {
  skip_if_not(l10n_info()$`UTF-8`)

  bar <- "\u2551"
  expect_success(expect_output(cat(bar), "\u2551"))
})

# setequal ----------------------------------------------------------------

test_that("ignores order and duplicates", {
  expect_success(expect_setequal(letters, rev(letters)))
  expect_success(expect_setequal(c("a", "a", "b"), c("b", "b", "a")))
})

test_that("checks both directions of containment", {
  expect_failure(expect_setequal(letters, letters[-1]))
  expect_failure(expect_setequal(letters[-1], letters))
})

test_that("truncates long differences", {
  cnd <- catch_cnd(expect_setequal("a", letters))
  expect_match(cnd$message, "...")
})

test_that("can compare data frames", {
  # this isn't really a legit use case but one package does it
  df <- data.frame(x = 1:10, y = 10:1)
  expect_success(expect_setequal(unname(df), unname(df)))
})

test_that("warns if both inputs are named", {
  expect_snapshot(expect_setequal(c(a = 1), c(b = 1)))
})

test_that("checks inputs", {
  fun <- sum
  expect_snapshot(error = TRUE, {
    expect_setequal(sum, 1)
    expect_setequal(1, sum)
    expect_setequal(!!fun, 1)
    expect_setequal(1, !!fun)
  })
})

test_that("useful message on failure", {
  expect_snapshot_failure(expect_setequal("actual", "expected"))

  x <- 1:2
  y <- 2
  expect_snapshot_failure(expect_setequal(x, y))

  x <- 2
  y <- 2:3
  expect_snapshot_failure(expect_setequal(x, y))

  x <- 1:2
  y <- 2:3
  expect_snapshot_failure(expect_setequal(x, y))

  # doesn't repeat values
  x <- c("a", "a")
  y <- c("b", "b", "b")
  expect_snapshot_failure(expect_setequal(x, y))

  # still looks good when expected is inlined
  x <- c("a", "b", "c")
  expect_snapshot_failure(expect_setequal(x, c("a", "b", "c", "d")))
})

test_that("truncates long vectors", {
  x <- 1:2
  y <- 1:50
  expect_snapshot_failure(expect_setequal(x, y))
})

# mapequal ----------------------------------------------------------------

test_that("ignores order", {
  expect_success(expect_mapequal(list(a = 1, b = 2), list(b = 2, a = 1)))
  expect_success(expect_mapequal(c(a = 1, b = 2), c(b = 2, a = 1)))
})

test_that("fails if names don't match", {
  x <- list(a = 1, b = 2)
  y <- list(a = 1)
  expect_snapshot_failure(expect_mapequal(x, y))
  expect_snapshot_failure(expect_mapequal(y, x))
})

test_that("fails if values don't match", {
  x <- list(a = 1, b = 2)
  y <- list(a = 1, b = 3)
  expect_snapshot_failure(expect_mapequal(x, y))
})

test_that("NULLs are not dropped", {
  expect_success(expect_mapequal(list(a = 1, b = NULL), list(b = NULL, a = 1)))
})

test_that("warns if empty vector", {
  expect_snapshot(expect_success(expect_mapequal(list(), list())))
})

test_that("ignores integer/numeric differences", {
  expect_success(expect_mapequal(list(a = 1L), list(a = 1)))
})

test_that("uses equality behaviour of current edition", {
  local_edition(2)
  expect_success(expect_mapequal(c(a = 1), c(a = 1L)))
})

test_that("validates its inputs", {
  unnamed <- list(1)
  named <- list(a = 1)
  duplicated <- list(x = 1, x = 2)

  expect_snapshot(error = TRUE, {
    expect_mapequal(sum, named)
    expect_mapequal(named, sum)

    expect_mapequal(unnamed, named)
    expect_mapequal(named, unnamed)

    expect_mapequal(named, duplicated)
    expect_mapequal(duplicated, named)
  })
})

# contains ----------------------------------------------------------------

test_that("expect_contains() succeeds when appropriate", {
  expect_success(expect_contains(letters, "a"))
  expect_success(expect_contains(letters, letters))
  expect_success(expect_contains(letters, character()))
})

test_that("expect_contains() gives useful message on failure", {
  x1 <- c("a", "b", "c")
  x2 <- c("c", "d")
  x3 <- c("d", "e")

  expect_snapshot_failure(expect_contains(x1, x2))
  expect_snapshot_failure(expect_contains(x1, x3))
})


# in ----------------------------------------------------------------

test_that("expect_in() succeeds when appropriate", {
  expect_success(expect_in("a", letters))
  expect_success(expect_in(letters, letters))
  expect_success(expect_in(character(), letters))
})

test_that("expect_in() gives useful message on failure", {
  x1 <- c("a", "b")
  x2 <- c("b", "c")
  x3 <- c("d", "e")

  expect_snapshot_failure(expect_in(x1, x2))
  expect_snapshot_failure(expect_in(x1, x3))
})

# disjoint ----------------------------------------------------------------

test_that("expect_disjoint() succeeds when appropriate", {
  expect_success(expect_disjoint(1, letters))
  expect_success(expect_disjoint(LETTERS, letters))
  expect_success(expect_disjoint(character(), letters))
})

test_that("expect_disjoint() gives useful message on failure", {
  x1 <- c("a", "b", "c")
  x2 <- c("c", "d")
  x3 <- c("b", "c", "d")

  expect_snapshot_failure(expect_disjoint(x1, x2))
  expect_snapshot_failure(expect_disjoint(x1, x3))
  expect_snapshot_failure(expect_disjoint(NA, c("a", NA)))
})

test_that("validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_all_equal(mean, 1)
    expect_all_equal(logical(), 1)
    expect_all_equal(1:10, mean)
    expect_all_equal(1:10, 1:2)
  })
})

test_that("can compare atomic vectors", {
  x <- rep(TRUE, 10)
  expect_success(expect_all_equal(x, TRUE))

  x[5] <- FALSE
  expect_snapshot_failure(expect_all_equal(x, TRUE))
})

test_that("can compare named lists", {
  x <- list(a = 1, b = 1, c = 2, d = 1, e = 1)
  expect_snapshot_failure(expect_all_equal(x, list(1)))
})

test_that("has tolerance enabled", {
  expect_success(expect_all_equal(1, 1L))
})

test_that("truncates very long differences", {
  x <- rep(TRUE, 10)
  expect_snapshot_failure(expect_all_equal(x, FALSE))
})

test_that("has TRUE and FALSE helpers", {
  x1 <- rep(TRUE, 10)
  x2 <- rep(FALSE, 10)
  expect_success(expect_all_true(x1))
  expect_success(expect_all_false(x2))
})

test_that("succeeds only when same object", {
  local_edition(2)
  x <- y <- 1
  expect_success(expect_reference(x, y))
  expect_failure(expect_reference(x, 1))
})

test_that("returns value", {
  local_edition(2)
  one <- 1
  expect_equal(expect_reference(one, one), 1)
})

test_that("length computed correctly", {
  expect_success(expect_length(1, 1))
  expect_failure(expect_length(1, 2), "has length 1, not length 2.")
  expect_success(expect_length(1:10, 10))
  expect_success(expect_length(letters[1:5], 5))
})

test_that("uses S4 length method", {
  A <- setClass("ExpectLengthA", slots = c(x = "numeric", y = "numeric"))
  setMethod("length", "ExpectLengthA", function(x) 5L)
  expect_success(expect_length(A(x = 1:9, y = 3), 5))
})

test_that("returns input", {
  x <- list(1:10, letters)
  out <- expect_length(x, 2)
  expect_identical(out, x)
})

test_that("can control appearance of dots", {
  expect_snapshot_reporter(SummaryReporter$new(
    show_praise = FALSE,
    omit_dots = FALSE
  ))
  expect_snapshot_reporter(SummaryReporter$new(
    show_praise = FALSE,
    omit_dots = TRUE
  ))
})

test_that("can control maximum reports", {
  expect_snapshot_reporter(SummaryReporter$new(
    show_praise = FALSE,
    max_reports = 2
  ))
})

test_that("basic principles of visibility hold", {
  expect_success(expect_invisible(x <- 10))
  expect_failure(expect_invisible(x))

  expect_success(expect_visible(x))
  expect_failure(expect_visible(x <- 1))
})

test_that("generates useful failure messages", {

  expect_snapshot_failure(expect_visible(invisible(1)))
  expect_snapshot_failure(expect_invisible(1))

})

test_that("invisibly returns evaluated value", {
  out <- expect_invisible(expect_invisible(x <- 2 + 2))
  expect_equal(out, 4)

  out <- expect_invisible(expect_visible(2 + 2))
  expect_equal(out, 4)
})

test_that("expect_failure() requires 1 failure and zero successes", {
  expect_success(expect_failure(fail()))

  expect_failure(expect_failure({}))
  expect_failure(expect_failure(pass()))
  expect_failure(expect_failure({
    pass()
    fail()
  }))

  expect_failure(expect_failure({
    fail()
    # Following succeed/fail are never reached
    pass()
    fail()
  }))
})

test_that("expect_failure() can optionally match message", {
  expect_success(expect_failure(fail("apple"), "apple"))
  expect_failure(expect_failure(fail("apple"), "banana"))
})

test_that("expect_failure() generates a useful error messages", {
  expect_no_failure <- function() {}
  expect_many_failures <- function() {
    fail()
    fail()
  }
  expect_has_success <- function() {
    fail()
    pass()
  }
  expect_both_wrong <- function() {
    pass()
  }
  expect_failure_foo <- function() fail("foo")

  expect_snapshot_failure({
    expect_failure(expect_no_failure())
    expect_failure(expect_many_failures())
    expect_failure(expect_has_success())
    expect_failure(expect_both_wrong())
    expect_failure(expect_failure_foo(), "bar")
  })
})

test_that("expect_success() requires 1 success and zero failures", {
  expect_success(expect_success(pass()))

  expect_failure(expect_success({}))
  expect_failure(expect_success(fail()))
  expect_failure(expect_success({
    pass()
    fail()
  }))
  expect_failure(expect_success({
    pass()
    pass()
  }))
})

test_that("errors in expect_success bubble up", {
  expect_snapshot(expect_success(abort("error")), error = TRUE)
})

test_that("show_failure", {
  expect_null(show_failure(NULL))
  expect_snapshot(show_failure(expect_true(FALSE)))
})


test_that("expect_success() generates a useful error messages", {
  expect_no_success <- function() {}
  expect_many_successes <- function() {
    pass()
    pass()
  }
  expect_has_failure <- function() {
    fail()
    pass()
  }
  expect_both_wrong <- function() {
    fail()
  }

  expect_snapshot_failure({
    expect_success(expect_no_success())
    expect_success(expect_many_successes())
    expect_success(expect_has_failure())
    expect_success(expect_both_wrong())
  })
})

test_that("can count successes and failures", {
  status <- capture_success_failure({})
  expect_equal(status$n_success, 0)
  expect_equal(status$n_failure, 0)

  status <- capture_success_failure({
    pass()
    pass()
    fail()
  })
  expect_equal(status$n_success, 2)
  expect_equal(status$n_failure, 1)

  # No code run after first fail
  status <- capture_success_failure({
    pass()
    fail()
    pass()
    fail()
  })
  expect_equal(status$n_success, 2)
  expect_equal(status$n_failure, 2)
})

test_that("expect_no are deprecated", {
  expect_snapshot({
    expect_no_failure(pass())
    expect_no_success(fail())
  })
})

test_that("expect_no still work", {
  withr::local_options(lifecycle_verbosity = "quiet")
  expect_success(expect_no_failure(pass()))
  expect_failure(expect_no_failure(fail()))
  expect_success(expect_no_success(fail()))
  expect_failure(expect_no_success(pass()))
})

test_that("expect_no_* conditions behave as expected", {
  # base R
  expect_snapshot_failure(expect_no_error(stop("error")))
  expect_snapshot_failure(expect_no_warning(warning("warning")))
  expect_snapshot_failure(expect_no_message(message("message")))

  # rlang equivalents
  expect_snapshot_failure(expect_no_error(abort("error")))
  expect_snapshot_failure(expect_no_warning(warn("warning")))
  expect_snapshot_failure(expect_no_message(inform("message")))
})

test_that("expect_no_* pass with pure code", {
  expect_success(out <- expect_no_error(1))
  expect_equal(out, 1)

  expect_success(expect_no_warning(1))
  expect_success(expect_no_message(1))
  expect_success(expect_no_condition(1))
})

test_that("expect_no_ continues execution", {
  b <- 1
  expect_failure(expect_no_warning({
    warning("x")
    b <- 2
  }))
  expect_equal(b, 2)
})

test_that("expect_no_* don't emit success when they fail", {
  expect_snapshot_failure(expect_no_error(stop("!")))
})

test_that("capture correct trace_env (#1994)", {
  status <- capture_success_failure(
    expect_warning(expect_error(stop("oops")))
  )
  expect_equal(status$n_success, 1) # from expect_error()
  expect_equal(status$n_failure, 1) # from expect_warning()
})

test_that("unmatched conditions bubble up", {
  expect_error(expect_no_error(abort("foo"), message = "bar"), "foo")
  expect_warning(expect_no_warning(warn("foo"), message = "bar"), "foo")
  expect_message(expect_no_message(inform("foo"), message = "bar"), "foo")
  expect_condition(
    expect_no_condition(signal("foo", "x"), message = "bar"),
    "foo"
  )
})

test_that("only matches conditions of specified type", {
  foo <- function() {
    warn("This is a problem!", class = "test")
  }
  expect_warning(expect_no_error(foo(), class = "test"), class = "test")
})

test_that("matched conditions give informative message", {
  foo <- function() {
    warn("This is a problem!", class = "test")
  }

  expect_snapshot(error = TRUE, {
    expect_no_warning(foo())
    expect_no_warning(foo(), message = "problem")
    expect_no_warning(foo(), class = "test")
    expect_no_warning(foo(), message = "problem", class = "test")
  })
})

test_that("deprecations always bubble up", {
  foo <- function() {
    lifecycle::deprecate_warn("1.0.0", "foo()")
  }
  expect_warning(
    expect_no_warning(foo()),
    class = "lifecycle_warning_deprecated"
  )
})

test_that("captures error before first test", {
  local_output_override()

  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path("reporters/error-setup.R")
  )
})

test_that("gracefully handles multiple contexts", {
  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path("reporters/context.R")
  )
})

test_that("can control max fails with env var or option", {
  withr::local_envvar(TESTTHAT_MAX_FAILS = 11)
  expect_equal(testthat_max_fails(), 11)

  withr::local_options(testthat.progress.max_fails = 12)
  expect_equal(testthat_max_fails(), 12)
})

test_that("fails after max_fail tests", {
  withr::local_options(testthat.progress.max_fails = 10)
  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path(c("reporters/fail-many.R", "reporters/fail.R"))
  )
})

test_that("can fully suppress incremental updates", {
  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path("reporters/successes.R")
  )

  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = Inf, min_time = Inf),
    test_path("reporters/successes.R")
  )
})

test_that("reports backtraces", {
  withr::local_envvar(TESTTHAT_MAX_FAILS = Inf)
  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path("reporters/backtraces.R")
  )
})

test_that("records skips", {
  expect_snapshot_reporter(
    ProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path("reporters/skips.R")
  )
})

# compact display ---------------------------------------------------------

test_that("compact display is informative", {
  expect_snapshot_reporter(
    CompactProgressReporter$new(),
    test_path("reporters/tests.R")
  )
})

test_that("display of successes only is compact", {
  expect_snapshot_reporter(
    CompactProgressReporter$new(),
    test_path("reporters/successes.R")
  )

  expect_snapshot_reporter(
    CompactProgressReporter$new(),
    test_path("reporters/skips.R")
  )

  # And even more compact if in RStudio pane
  local_reproducible_output(rstudio = TRUE)
  expect_snapshot_reporter(
    CompactProgressReporter$new(),
    test_path("reporters/successes.R")
  )
})

# parallel progress reporter ----------------------------------------------

test_that("ParallelProgressReporter fails after max_fail tests", {
  withr::local_options(testthat.progress.max_fails = 10)
  expect_snapshot_reporter(
    ParallelProgressReporter$new(update_interval = 0, min_time = Inf),
    test_path(c("reporters/fail-many.R", "reporters/fail.R"))
  )
})

test_that("logical tests act as expected", {
  df <- data.frame(1:10)

  expect_success(expect_true(TRUE))
  expect_snapshot_failure(expect_true(df))

  expect_success(expect_false(FALSE))
  expect_snapshot_failure(expect_false(df))
})

test_that("logical tests ignore attributes", {
  expect_success(expect_true(c(a = TRUE)))
  expect_success(expect_false(c(a = FALSE)))
})

test_that("additional info returned in message", {
  expect_failure(expect_true(FALSE, "NOPE"), "NOPE")
  expect_failure(expect_false(TRUE, "YUP"), "YUP")
})

test_that("expect_null works", {
  x <- NULL
  df <- data.frame(1:10)

  expect_success(expect_null(x))
  expect_snapshot_failure(expect_null(df))
})

test_that("returns the input value", {
  res <- expect_true(TRUE)
  expect_equal(res, TRUE)
})

test_that("expected_named verifies presence of names", {
  expect_success(expect_named(c(a = 1)))

  x <- 1:10
  expect_snapshot_failure(expect_named(x))
})

test_that("expected_named verifies actual of names", {
  expect_success(expect_named(c(a = 1), "a"))

  x <- c(a = 1)
  expect_snapshot_failure(expect_named(x, "b"))
})

test_that("always returns inputs", {
  x <- c(a = 1)
  expect_equal(expect_named(x), x)
  expect_equal(expect_named(x, "a"), x)
  expect_equal(expect_named(x, "a", ignore.order = TRUE), x)
})

test_that("expected_named optionally ignores order and case", {
  x <- c(a = 1, b = 2)
  expect_success(expect_named(x, c("A", "B"), ignore.case = TRUE))
  expect_success(expect_named(x, c("b", "a"), ignore.order = TRUE))
})

test_that("provide useful feedback on failure", {
  x1 <- c(a = 1)
  x2 <- c(a = 1, b = 2)

  expect_snapshot_failure(expect_named(x1, c("a", "b"), ignore.order = TRUE))
  expect_snapshot_failure(expect_named(x2, "a", ignore.order = TRUE))
  expect_snapshot_failure(expect_named(x1, "b", ignore.order = TRUE))

  expect_snapshot_failure(expect_named(x1, c("a", "b"), ignore.order = FALSE))
  expect_snapshot_failure(expect_named(x2, "a", ignore.order = FALSE))
  expect_snapshot_failure(expect_named(x1, "b", ignore.order = FALSE))
})

test_that("expect_named validates its inputs", {
  expect_snapshot(error = TRUE, {
    expect_named(c(a = 1), "a", ignore.order = "yes")
    expect_named(c(a = 1), "a", ignore.case = "yes")
  })
})

test_that("expect_named accepts glue for 'expected'", {
  n <- structure(
    c("v1", "v2", "v3", "v4", "v5"),
    class = c("glue", "character")
  )
  v <- set_names(1:5, n)

  expect_named(v, n)
})

#' Find slow tests
#'
#' @description
#' `SlowReporter` is designed to identify slow tests. It reports the
#' execution time for each test and can optionally filter out tests that
#' run faster than a specified threshold (default: 1 second). This reporter
#' is useful for performance optimization and identifying tests that may
#' benefit from optimization or parallelization.
#'
#' @export
#' @family reporters
SlowReporter <- R6::R6Class(
  "SlowReporter",
  inherit = Reporter,
  public = list(
    min_time = 0.5,
    test_timings = NULL,
    current_test_start = NULL,
    current_file = NULL,

    initialize = function(min_time = 0.5, ...) {
      super$initialize(...)
      self$min_time <- min_time
      self$test_timings <- list()
    },

    start_reporter = function(context) {
      self$cat_line(
        cli::style_bold("Slow tests"),
        " (showing tests >= ",
        self$min_time,
        "s)"
      )
      self$cat_line()
    },

    start_file = function(file) {
      self$current_file <- file
    },

    start_test = function(context, test) {
      self$current_test_start <- proc.time()[[3]]
    },

    end_test = function(context, test) {
      if (is.null(self$current_test_start)) {
        return()
      }

      time_taken <- proc.time()[[3]] - self$current_test_start

      # Store timing information
      timing <- list(
        file = self$current_file,
        test = test,
        time = time_taken
      )
      self$test_timings[[length(self$test_timings) + 1]] <- timing

      if (time_taken >= self$min_time) {
        self$show_timing(timing)
      }

      self$current_test_start <- NULL
    },

    end_reporter = function() {
      if (length(self$test_timings) == 0) {
        return()
      }

      all_times <- map_dbl(self$test_timings, \(x) x$time)
      is_slow <- all_times >= self$min_time

      self$cat_line()
      self$rule(cli::style_bold("Summary"), line = 2)
      self$cat_line("All tests:  ", sprintf("%.2fs", sum(all_times)))
      self$cat_line("Slow tests: ", sprintf("%.2fs", sum(all_times[is_slow])))

      if (sum(is_slow) <= 10) {
        return()
      }

      # Sort by time descending for summary
      slowest <- self$test_timings[order(all_times, decreasing = TRUE)]

      self$cat_line()
      self$rule(cli::style_bold("Slowest tests:"), line = 1)

      # Show top 10 slowest tests
      for (i in 1:10) {
        self$show_timing(slowest[[i]])
      }

      if (length(slowest) > 10) {
        self$cat_line("... and ", length(slowest) - 10, " more slow tests")
      }

      self$cat_line()
    },
    show_timing = function(timing) {
      time <- sprintf("%.2fs", timing$time)
      self$cat_line("[", time, "] ", time, " ", timing$file, ": ", timing$test)
    }
  )
)

#' Check for global state changes
#'
#' @description
#' One of the most pernicious challenges to debug is when a test runs fine
#' in your test suite, but fails when you run it interactively (or similarly,
#' it fails randomly when running your tests in parallel). One of the most
#' common causes of this problem is accidentally changing global state in a
#' previous test (e.g. changing an option, an environment variable, or the
#' working directory). This is hard to debug, because it's very hard to figure
#' out which test made the change.
#'
#' Luckily testthat provides a tool to figure out if tests are changing global
#' state. You can register a state inspector with `set_state_inspector()` and
#' testthat will run it before and after each test, store the results, then
#' report if there are any differences. For example, if you wanted to see if
#' any of your tests were changing options or environment variables, you could
#' put this code in `tests/testthat/helper-state.R`:
#'
#' ```R
#' set_state_inspector(function() {
#'   list(
#'     options = options(),
#'     envvars = Sys.getenv()
#'   )
#' })
#' ```
#'
#' (You might discover other packages outside your control are changing
#' the global state, in which case you might want to modify this function
#' to ignore those values.)
#'
#' Other problems that can be troublesome to resolve are CRAN check notes that
#' report things like connections being left open. You can easily debug
#' that problem with:
#'
#' ```R
#' set_state_inspector(function() {
#'   getAllConnections()
#' })
#' ```
#'
#' @export
#' @param callback Either a zero-argument function that returns an object
#'   capturing global state that you're interested in, or `NULL`.
#' @inheritParams waldo::compare
set_state_inspector <- function(callback, tolerance = testthat_tolerance()) {
  if (
    !is.null(callback) &&
      !(is.function(callback) && length(formals(callback)) == 0)
  ) {
    cli::cli_abort("{.arg callback} must be a zero-arg function, or NULL")
  }

  the$state_inspector <- callback
  the$state_inspector_tolerance <- tolerance
  invisible()
}

testthat_state_condition <- function(before, after, call) {
  diffs <- waldo_compare(
    before,
    after,
    x_arg = "before",
    y_arg = "after",
    tolerance = the$state_inspector_tolerance
  )

  if (length(diffs) == 0) {
    return(NULL)
  }

  srcref <- attr(call, "srcref")
  warning_cnd(
    message = c("Global state has changed:", set_names(format(diffs), "")),
    srcref = srcref
  )
}

inspect_state <- function() {
  if (is.null(the$state_inspector)) {
    NULL
  } else {
    the$state_inspector()
  }
}

#' Run all tests in a package
#'
#' @description
#' * `test_local()` tests a local source package.
#' * `test_package()` tests an installed package.
#' * `test_check()` checks a package during `R CMD check`.
#'
#' See `vignette("special-files")` to learn about the various files that
#' testthat works with.
#'
#' @section `R CMD check`:
#' To run testthat automatically from `R CMD check`, make sure you have
#' a `tests/testthat.R` that contains:
#'
#' ```
#' library(testthat)
#' library(yourpackage)
#'
#' test_check("yourpackage")
#' ```
#'
#' @inherit test_dir return params
#' @inheritSection test_dir Environments
#' @param ... Additional arguments passed to [test_dir()]
#' @export
#' @rdname test_package
test_package <- function(package, reporter = check_reporter(), ...) {
  test_path <- system.file("tests", "testthat", package = package)
  if (test_path == "") {
    cli::cli_inform("No installed testthat tests found for {.pkg {package}}.")
    return(invisible())
  }

  test_dir(
    test_path,
    package = package,
    reporter = reporter,
    ...,
    load_package = "installed"
  )
}

#' @export
#' @rdname test_package
test_check <- function(package, reporter = check_reporter(), ...) {
  require(package, character.only = TRUE)
  options(cli.hyperlink = FALSE)

  withr::local_envvar(TESTTHAT_IS_CHECKING = "true")

  test_dir(
    "testthat",
    package = package,
    reporter = reporter,
    ...,
    load_package = "installed"
  )
}

#' @export
#' @rdname test_package
test_local <- function(
  path = ".",
  reporter = NULL,
  ...,
  load_package = "source",
  shuffle = FALSE
) {
  package <- pkgload::pkg_name(path)
  test_path <- file.path(pkgload::pkg_path(path), "tests", "testthat")

  local_assume_not_on_cran()
  test_dir(
    test_path,
    package = package,
    reporter = reporter,
    ...,
    load_package = load_package,
    shuffle = shuffle
  )
}

#' Report progress interactively
#'
#' @description
#' `ProgressReporter` is designed for interactive use. Its goal is to
#' give you actionable insights to help you understand the status of your
#' code. This reporter also praises you from time-to-time if all your tests
#' pass. It's the default reporter for [test_dir()].
#'
#' `ParallelProgressReporter` is very similar to `ProgressReporter`, but
#' works better for packages that want parallel tests.
#'
#' `CompactProgressReporter` is a minimal version of `ProgressReporter`
#' designed for use with single files. It's the default reporter for
#' [test_file()].
#'
#' @export
#' @family reporters
ProgressReporter <- R6::R6Class(
  "ProgressReporter",
  inherit = Reporter,
  public = list(
    show_praise = TRUE,
    min_time = 1,
    start_time = NULL,
    last_update = NULL,
    update_interval = NULL,

    skips = NULL,
    problems = NULL,

    max_fail = NULL,
    n_ok = 0,
    n_skip = 0,
    n_warn = 0,
    n_fail = 0,

    frames = NULL,
    dynamic = FALSE,

    ctxt_start_time = NULL,
    ctxt_issues = NULL,
    ctxt_n = 0,
    ctxt_n_ok = 0,
    ctxt_n_skip = 0,
    ctxt_n_warn = 0,
    ctxt_n_fail = 0,
    ctxt_name = "",
    file_name = "",

    initialize = function(
      show_praise = TRUE,
      max_failures = testthat_max_fails(),
      min_time = 1,
      update_interval = 0.1,
      ...
    ) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
      self$show_praise <- show_praise
      self$max_fail <- max_failures
      self$min_time <- min_time
      self$update_interval <- update_interval

      self$skips <- Stack$new()
      self$problems <- Stack$new()
      self$ctxt_issues <- Stack$new()

      # Capture at init so not affected by test settings
      self$frames <- cli::get_spinner()$frames
      self$dynamic <- cli::is_dynamic_tty()
    },

    is_full = function() {
      self$n_fail >= self$max_fail
    },

    start_reporter = function(context) {
      self$start_time <- proc.time()
      self$show_header()
    },

    start_file = function(file) {
      self$file_name <- file
      self$ctxt_issues <- Stack$new()
      self$ctxt_start_time <- proc.time()

      context_start_file(self$file_name)
    },

    start_context = function(context) {
      self$ctxt_name <- context
      self$ctxt_issues <- Stack$new()

      self$ctxt_n <- 0L
      self$ctxt_n_ok <- 0L
      self$ctxt_n_fail <- 0L
      self$ctxt_n_warn <- 0L
      self$ctxt_n_skip <- 0L

      self$ctxt_start_time <- proc.time()
      self$show_status()
    },

    show_header = function() {
      self$cat_line(
        colourise(cli::symbol$tick, "success"),
        " | ",
        colourise("F", "failure"),
        " ",
        colourise("W", "warning"),
        " ",
        colourise(" S", "skip"),
        " ",
        colourise(" OK", "success"),
        " | ",
        "Context"
      )
    },

    status_data = function() {
      list(
        n = self$ctxt_n,
        n_ok = self$ctxt_n_ok,
        n_fail = self$ctxt_n_fail,
        n_warn = self$ctxt_n_warn,
        n_skip = self$ctxt_n_skip,
        name = self$ctxt_name
      )
    },

    show_status = function(complete = FALSE, time = 0, pad = FALSE) {
      data <- self$status_data()
      if (complete) {
        if (data$n_fail > 0) {
          status <- cli::col_red(cli::symbol$cross)
        } else {
          status <- cli::col_green(cli::symbol$tick)
        }
      } else {
        # Do not print if not enough time has passed since we last printed.
        if (!self$should_update()) {
          return()
        }
        status <- spinner(self$frames, data$n)
        if (data$n_fail > 0) {
          status <- colourise(status, "failure")
        } else if (data$n_warn > 0) {
          status <- colourise(status, "warning")
        }
      }

      col_format <- function(n, type) {
        if (n == 0) {
          if (type == "skip") {
            "  "
          } else {
            " "
          }
        } else {
          if (type == "skip") {
            colourise(sprintf("%2d", n), type)
          } else {
            colourise(n, type)
          }
        }
      }

      message <- paste0(
        status,
        " | ",
        col_format(data$n_fail, "fail"),
        " ",
        col_format(data$n_warn, "warn"),
        " ",
        col_format(data$n_skip, "skip"),
        " ",
        sprintf("%3d", data$n_ok),
        " | ",
        data$name
      )

      if (complete && time > self$min_time) {
        message <- paste0(
          message,
          cli::col_grey(sprintf(" [%.1fs]", time))
        )
      }

      if (pad) {
        message <- strpad(message, self$width)
        message <- cli::ansi_substr(message, 1, self$width)
      }

      if (!complete) {
        message <- strpad(message, self$width)
        self$cat_tight(self$cr(), message)
      } else {
        self$cat_line(self$cr(), message)
      }
    },

    cr = function() {
      if (self$dynamic) {
        "\r"
      } else {
        "\n"
      }
    },

    end_context = function(context) {
      time <- proc.time() - self$ctxt_start_time
      self$last_update <- NULL

      # context with no expectation = automatic file context in file
      # that also has manual contexts
      if (self$ctxt_n == 0) {
        return()
      }

      self$show_status(complete = TRUE, time = time[[3]])
      self$report_issues(self$ctxt_issues)

      if (self$is_full()) {
        self$report_full()
      }
    },

    report_full = function() {
      snapshotter <- get_snapshotter()
      if (!is.null(snapshotter)) {
        snapshotter$end_file()
      }

      # Separate from the progress bar
      self$cat_line()
      stop_reporter(c(
        "Maximum number of failures exceeded; quitting.",
        i = "Increase this number with (e.g.) {.run testthat::set_max_fails(Inf)}"
      ))
    },

    add_result = function(context, test, result) {
      self$ctxt_n <- self$ctxt_n + 1L

      if (expectation_broken(result)) {
        self$n_fail <- self$n_fail + 1
        self$ctxt_n_fail <- self$ctxt_n_fail + 1
        self$ctxt_issues$push(result)
        self$problems$push(result)
      } else if (expectation_skip(result)) {
        self$n_skip <- self$n_skip + 1
        self$ctxt_n_skip <- self$ctxt_n_skip + 1
        self$skips$push(result)
      } else if (expectation_warning(result)) {
        self$n_warn <- self$n_warn + 1
        self$ctxt_n_warn <- self$ctxt_n_warn + 1
        self$ctxt_issues$push(result)
      } else {
        self$n_ok <- self$n_ok + 1
        self$ctxt_n_ok <- self$ctxt_n_ok + 1
      }

      self$show_status()
    },

    end_reporter = function() {
      if (self$is_full()) {
        return()
      }

      self$cat_line()

      colour_if <- function(n, type) {
        colourise(n, if (n == 0) "success" else type)
      }

      self$rule(cli::style_bold("Results"), line = 2)
      time <- proc.time() - self$start_time
      if (time[[3]] > self$min_time) {
        self$cat_line("Duration: ", sprintf("%.1f s", time[[3]]), col = "cyan")
        self$cat_line()
      }

      skip_report(self)

      if (self$problems$size() > 0) {
        problems <- self$problems$as_list()
        self$rule("Failed tests", line = 1)
        for (problem in problems) {
          self$cat_line(issue_summary(problem))
          self$cat_line()
        }
      }

      status <- summary_line(self$n_fail, self$n_warn, self$n_skip, self$n_ok)
      self$cat_line(status)

      if (self$is_full()) {
        self$rule("Terminated early", line = 2)
      }

      if (!self$show_praise || stats::runif(1) > 0.1) {
        return()
      }

      self$cat_line()
      if (self$n_fail == 0) {
        self$cat_line(colourise(praise(), "success"))
      } else {
        self$cat_line(colourise(encourage(), "error"))
      }
    },

    report_issues = function(issues) {
      if (issues$size() > 0) {
        self$rule()

        issues <- issues$as_list()
        summary <- map_chr(issues, issue_summary)
        self$cat_tight(paste(summary, collapse = "\n\n"))

        self$cat_line()
        self$rule()
      }
    },

    should_update = function() {
      if (self$update_interval == 0) {
        return(TRUE)
      }
      if (identical(self$update_interval, Inf)) {
        return(FALSE)
      }

      time <- proc.time()[[3]]
      if (
        !is.null(self$last_update) &&
          (time - self$last_update) < self$update_interval
      ) {
        return(FALSE)
      }
      self$last_update <- time
      TRUE
    }
  )
)

testthat_max_fails <- function() {
  val <- getOption("testthat.progress.max_fails")

  if (is.null(val)) {
    env <- Sys.getenv("TESTTHAT_MAX_FAILS")
    val <- if (!identical(env, "")) as.numeric(env) else 10
  }
  val
}

#' @export
#' @rdname ProgressReporter
CompactProgressReporter <- R6::R6Class(
  "CompactProgressReporter",
  inherit = ProgressReporter,
  public = list(
    initialize = function(min_time = Inf, ...) {
      super$initialize(min_time = min_time, ...)
    },

    start_file = function(name) {
      if (!self$rstudio) {
        self$cat_line()
        self$rule(cli::style_bold(paste0("Testing ", name)), line = 2)
      }
      super$start_file(name)
    },

    start_reporter = function(context) {},

    end_context = function(context) {
      if (self$ctxt_issues$size() == 0) {
        return()
      }

      self$cat_line()
      self$cat_line()

      issues <- self$ctxt_issues$as_list()
      summary <- map_chr(issues, issue_summary, rule = TRUE)
      self$cat_tight(paste(summary, collapse = "\n\n"))

      self$cat_line()
    },

    end_reporter = function() {
      had_feedback <- self$n_fail > 0 || self$n_warn > 0

      if (self$n_skip > 0) {
        if (!had_feedback) {
          self$cat_line()
        }
        self$cat_line()
        skip_report(self)
      }

      if (had_feedback) {
        self$show_status()
        self$cat_line()
      } else if (self$is_full()) {
        self$cat_line(" Terminated early")
      } else if (self$n_skip == 0 && !self$rstudio) {
        self$cat_line(cli::style_bold(" Done!"))
      }
    },

    show_status = function(complete = NULL) {
      self$local_user_output()
      status <- summary_line(self$n_fail, self$n_warn, self$n_skip, self$n_ok)
      self$cat_tight(self$cr(), status)
    }
  )
)

# parallel progres reporter -----------------------------------------------

#' @export
#' @rdname ProgressReporter

ParallelProgressReporter <- R6::R6Class(
  "ParallelProgressReporter",
  inherit = ProgressReporter,
  public = list(
    files = list(),
    spin_frame = 0L,
    is_rstudio = FALSE,

    initialize = function(...) {
      super$initialize(...)
      self$capabilities$parallel_support <- TRUE
      self$capabilities$parallel_updates <- TRUE
      self$update_interval <- 0.05
      self$is_rstudio <- Sys.getenv("RSTUDIO", "") == "1"
    },

    start_file = function(file) {
      if (!file %in% names(self$files)) {
        self$files[[file]] <- list(
          issues = Stack$new(),
          n_fail = 0L,
          n_skip = 0L,
          n_warn = 0L,
          n_ok = 0L,
          name = context_name(file),
          start_time = proc.time()
        )
      }
      self$file_name <- file
    },

    start_context = function(context) {
      # we'll just silently ignore this
    },

    end_context = function(context) {
      # we'll just silently ignore this
    },

    end_file = function() {
      if (!self$is_full()) {
        fsts <- self$files[[self$file_name]]
        time <- proc.time() - fsts$start_time

        # Workaround for https://github.com/rstudio/rstudio/issues/7649
        if (self$is_rstudio) {
          self$cat_tight(strpad(self$cr(), self$width + 1)) # +1 for \r
        }
        self$show_status(complete = TRUE, time = time[[3]], pad = TRUE)
        self$report_issues(fsts$issues)

        self$files[[self$file_name]] <- NULL
        if (length(self$files)) {
          self$update(force = TRUE)
        }
      } else {
        self$update(force = TRUE)
        self$cat_line()
        self$rule()

        issues <- unlist(
          map(self$files, \(x) x$issues$as_list()),
          recursive = FALSE
        )
        summary <- map_chr(issues, issue_summary)
        self$cat_tight(paste(summary, collapse = "\n\n"))

        self$cat_line()
        self$report_full()
      }
    },

    end_reporter = function() {
      self$cat_tight(self$cr(), strpad("", self$width))
      super$end_reporter()
    },

    show_header = function() {
      super$show_header()
      self$update(force = TRUE)
    },

    status_data = function() {
      self$files[[self$file_name]]
    },

    add_result = function(context, test, result) {
      self$ctxt_n <- self$ctxt_n + 1L
      file <- self$file_name
      if (expectation_broken(result)) {
        self$n_fail <- self$n_fail + 1
        self$files[[file]]$n_fail <- self$files[[file]]$n_fail + 1L
        self$files[[file]]$issues$push(result)
        self$problems$push(result)
      } else if (expectation_skip(result)) {
        self$n_skip <- self$n_skip + 1
        self$files[[file]]$n_skip <- self$files[[file]]$n_skip + 1L
        self$skips$push(result)
      } else if (expectation_warning(result)) {
        self$n_warn <- self$n_warn + 1
        self$files[[file]]$n_warn <- self$files[[file]]$n_warn + 1L
        self$files[[file]]$issues$push(result)
      } else {
        self$n_ok <- self$n_ok + 1
        self$files[[file]]$n_ok <- self$files[[file]]$n_ok + 1
      }
    },

    update = function(force = FALSE) {
      if (!force && !self$should_update()) {
        return()
      }
      self$spin_frame <- self$spin_frame + 1L
      status <- spinner(self$frames, self$spin_frame)

      message <- paste(
        status,
        summary_line(self$n_fail, self$n_warn, self$n_skip, self$n_ok),
        if (length(self$files) > 0) "@" else "Starting up...",
        paste(context_name(names(self$files)), collapse = ", ")
      )
      message <- strpad(message, self$width)
      message <- cli::ansi_substr(message, 1, self$width)
      self$cat_tight(self$cr(), message)
    }
  )
)

# helpers -----------------------------------------------------------------

spinner <- function(frames, i) {
  frames[((i - 1) %% length(frames)) + 1]
}

issue_header <- function(x, pad = FALSE, location = TRUE) {
  type <- expectation_type(x)
  if (has_colour()) {
    type <- colourise(first_upper(type), type)
  } else {
    type <- first_upper(type)
  }
  if (pad) {
    type <- strpad(type, 7)
  }

  paste0(type, if (location) expectation_location(x, " (", ")"), ": ", x$test)
}

issue_summary <- function(x, rule = FALSE, location = TRUE) {
  header <- cli::style_bold(issue_header(x, location = location))
  if (rule) {
    # Don't truncate long test names
    width <- max(cli::ansi_nchar(header) + 6, getOption("width"))
    header <- cli::rule(header, width = width)
  }

  paste0(header, "\n", format(x))
}

strpad <- function(x, width = cli::console_width()) {
  n <- pmax(0, width - cli::ansi_nchar(x))
  paste0(x, strrep(" ", n))
}

skip_report <- function(reporter, line = 1) {
  n <- reporter$skips$size()
  if (n == 0) {
    return()
  }

  reporter$rule(paste0("Skipped tests (", n, ")"), line = line)
  reporter$cat_line(skip_bullets(reporter$skips$as_list()))
  reporter$cat_line()
}


skip_bullets <- function(skips) {
  message <- map_chr(skips, "[[", "message")
  message <- gsub("Reason: ", "", message)
  message <- gsub(":?\n(\n|.)+", "", message) # only show first line

  locs <- map_chr(skips, expectation_location)
  locs_by_skip <- split(locs, message)
  n <- lengths(locs_by_skip)
  skip_summary <- map_chr(locs_by_skip, paste, collapse = ", ")

  bullets <- paste0(
    cli::symbol$bullet,
    " ",
    names(locs_by_skip),
    " (",
    n,
    "): ",
    skip_summary
  )
  cli::ansi_strwrap(bullets, exdent = 2)
}


#' Set maximum number of test failures allowed before aborting the run
#'
#' This sets the `TESTTHAT_MAX_FAILS` env var which will affect both the
#' current R process and any processes launched from it.
#'
#' @param n Maximum number of failures allowed.
#' @export
#' @keywords internal
set_max_fails <- function(n) {
  Sys.setenv('TESTTHAT_MAX_FAILS' = n)
}

writeLines(
  "If you see me, something has gone wrong with old-school teardown",
  "DELETE-ME"
)
writeLines(
  "If you see me, something has gone wrong with new-school teardown",
  "DELETE-ME-2"
)
withr::defer(unlink("DELETE-ME-2"), teardown_env())

#' Declare that an expectation either passes or fails
#'
#' @description
#' These are the primitives that you can use to implement your own expectations.
#' Every path through an expectation should either call `pass()`, `fail()`,
#' or throw an error (e.g. if the arguments are invalid). Expectations should
#' always return `invisible(act$val)`.
#'
#' Learn more about creating your own expectations in
#' `vignette("custom-expectation")`.
#'
#' @param message A character vector describing the failure. The
#'   first element should describe the expected value, and the second (and
#'   optionally subsequence) elements should describe what was actually seen.
#' @param info Character vector continuing additional information. Included
#'   for backward compatibility only and new expectations should not use it.
#' @param srcref Location of the failure. Should only needed to be explicitly
#'   supplied when you need to forward a srcref captured elsewhere.
#' @param trace_env If `trace` is not specified, this is used to generate an
#'   informative traceback for failures. You should only need to set this if
#'   you're calling `fail()` from a helper function; see
#'   `vignette("custom-expectation")` for details.
#' @param trace An optional backtrace created by [rlang::trace_back()].
#'   When supplied, the expectation is displayed with the backtrace.
#'   Expert use only.
#' @export
#' @examples
#' expect_length <- function(object, n) {
#'   act <- quasi_label(rlang::enquo(object), arg = "object")
#'
#'   act_n <- length(act$val)
#'   if (act_n != n) {
#'     fail(sprintf("%s has length %i, not length %i.", act$lab, act_n, n))
#'   } else {
#'     pass()
#'   }
#'
#'   invisible(act$val)
#' }
fail <- function(
  message = "Failure has been forced",
  info = NULL,
  srcref = NULL,
  trace_env = caller_env(),
  trace = NULL
) {
  check_character(message)
  check_character(info, allow_null = TRUE)

  trace <- trace %||% capture_trace(trace_env)
  message <- paste(c(message, info), collapse = "\n")
  expectation("failure", message, srcref = srcref, trace = trace)
  invisible()
}

snapshot_fail <- function(message, trace_env = caller_env()) {
  trace <- capture_trace(trace_env)
  message <- paste(message, collapse = "\n")
  expectation("failure", message, trace = trace, snapshot = TRUE)
  invisible()
}

capture_trace <- function(trace_env) {
  trace <- trace_back(top = getOption("testthat_topenv"), bottom = trace_env)
  # Only include trace if there's at least one function apart from the expectation
  if (!is.null(trace) && trace_length(trace) <= 1) {
    trace <- NULL
  }
  trace
}

#' @rdname fail
#' @export
pass <- function() {
  expectation("success", "success")
  invisible()
}

#' Mark a test as successful
#'
#' This is an older version of [pass()] that exists for backwards compatibility.
#' You should now use `pass()` instead.
#'
#' @export
#' @inheritParams fail
#' @keywords internal
succeed <- function(message = "Success has been forced", info = NULL) {
  message <- paste(c(message, info), collapse = "\n")

  expectation("success", message)
}

#' Do you expect `TRUE` or `FALSE`?
#'
#' @description
#' These are fall-back expectations that you can use when none of the other
#' more specific expectations apply. The disadvantage is that you may get
#' a less informative error message.
#'
#' Attributes are ignored.
#'
#' @inheritParams expect_that
#' @family expectations
#' @examples
#' expect_true(2 == 2)
#' # Failed expectations will throw an error
#' show_failure(expect_true(2 != 2))
#'
#' # where possible, use more specific expectations, to get more informative
#' # error messages
#' a <- 1:4
#' show_failure(expect_true(length(a) == 3))
#' show_failure(expect_equal(length(a), 3))
#'
#' x <- c(TRUE, TRUE, FALSE, TRUE)
#' show_failure(expect_true(all(x)))
#' show_failure(expect_all_true(x))
#' @name logical-expectations
NULL

#' @export
#' @rdname logical-expectations
expect_true <- function(object, info = NULL, label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- labelled_value(TRUE, "TRUE")

  expect_waldo_constant_(act, exp, info = info, ignore_attr = TRUE)
  invisible(act$val)
}

#' @export
#' @rdname logical-expectations
expect_false <- function(object, info = NULL, label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- labelled_value(FALSE, "FALSE")

  expect_waldo_constant_(act, exp, info = info, ignore_attr = TRUE)
  invisible(act$val)
}

#' Do you expect `NULL`?
#'
#' This is a special case because `NULL` is a singleton so it's possible
#' check for it either with `expect_equal(x, NULL)` or `expect_type(x, "NULL")`.
#'
#' @inheritParams expect_that
#' @export
#' @family expectations
#' @examples
#' x <- NULL
#' y <- 10
#'
#' expect_null(x)
#' show_failure(expect_null(y))
expect_null <- function(object, info = NULL, label = NULL) {
  act <- quasi_label(enquo(object), label)
  exp <- labelled_value(NULL, "NULL")

  expect_waldo_constant_(act, exp, info = info)
  invisible(act$val)
}

expect_waldo_constant_ <- function(
  act,
  exp,
  ...,
  info = NULL,
  trace_env = caller_env()
) {
  comp <- waldo_compare(
    act$val,
    exp$val,
    ...,
    x_arg = "actual",
    y_arg = "expected"
  )
  if (length(comp) != 0) {
    msg <- c(
      sprintf("Expected %s to be %s.", act$lab, exp$lab),
      "Differences:",
      paste0(comp, "\n")
    )
    fail(msg, info = info, trace_env = trace_env)
  } else {
    pass()
  }
}

#' Does code run without error, warning, message, or other condition?
#'
#' @description
#' These expectations are the opposite of [expect_error()],
#' `expect_warning()`, `expect_message()`, and `expect_condition()`. They
#' assert the absence of an error, warning, or message, respectively.
#'
#' @inheritParams expect_error
#' @param message,class The default, `message = NULL, class = NULL`,
#'   will fail if there is any error/warning/message/condition.
#'
#'   In many cases, particularly when testing warnings and messages, you will
#'   want to be more specific about the condition you are hoping **not** to see,
#'   i.e. the condition that motivated you to write the test.  Similar to
#'   `expect_error()` and friends, you can specify the `message` (a regular
#'   expression that the message of the condition must match) and/or the
#'   `class` (a class the condition must inherit from). This ensures that
#'   the message/warnings you don't want never recur, while allowing new
#'   messages/warnings to bubble up for you to deal with.
#'
#'   Note that you should only use `message` with errors/warnings/messages
#'   that you generate, or that base R generates (which tend to be stable).
#'   Avoid tests that rely on the specific text generated by another package
#'   since this can easily change. If you do need to test text generated by
#'   another package, either protect the test with `skip_on_cran()` or
#'   use `expect_snapshot()`.
#' @inheritParams rlang::args_dots_empty
#' @export
#' @examples
#' expect_no_warning(1 + 1)
#'
#' foo <- function(x) {
#'   warning("This is a problem!")
#' }
#'
#' # warning doesn't match so bubbles up:
#' expect_no_warning(foo(), message = "bananas")
#'
#' # warning does match so causes a failure:
#' try(expect_no_warning(foo(), message = "problem"))
expect_no_error <- function(object,
                            ...,
                            message = NULL,
                            class = NULL) {
  check_dots_empty()
  expect_no_("error", {{ object }}, regexp = message, class = class)
}


#' @export
#' @rdname expect_no_error
expect_no_warning <- function(object,
                              ...,
                              message = NULL,
                              class = NULL
                              ) {
  check_dots_empty()
  expect_no_("warning", {{ object }}, regexp = message, class = class)
}

#' @export
#' @rdname expect_no_error
expect_no_message <- function(object,
                              ...,
                              message = NULL,
                              class = NULL
                              ) {
  check_dots_empty()
  expect_no_("message", {{ object }}, regexp = message, class = class)
}

#' @export
#' @rdname expect_no_error
expect_no_condition <- function(object,
                                ...,
                                message = NULL,
                                class = NULL
                                ) {
  check_dots_empty()
  expect_no_("condition", {{ object }}, regexp = message, class = class)
}


expect_no_ <- function(base_class,
                       object,
                       regexp = NULL,
                       class = NULL,
                       trace_env = caller_env()) {

  matcher <- cnd_matcher(
    base_class,
    class,
    pattern = regexp,
    ignore_deprecation = base_class == "warning" && is.null(regexp) && is.null(class)
  )

  capture <- function(code) {
    try_fetch(
      {
        code
        # We can't call succeed() here because that generates a condition
        # that causes `expect_no_condition()` to always fail
        NULL
      },
      !!base_class := function(cnd) {
        if (!matcher(cnd)) {
          return(zap())
        }

        expected <- paste0(
          "Expected ", quo_label(enquo(object)), " to run without any ", base_class, "s",
          if (!is.null(class)) paste0(" of class '", class, "'"),
          if (!is.null(regexp)) paste0(" matching pattern '", regexp, "'"),
          "."
        )
        actual <- paste0(
          "Actually got a <", class(cnd)[[1]], "> with text:\n",
          indent_lines(rlang::cnd_message(cnd))
        )
        message <- format_error_bullets(c(expected, i = actual))
        fail(message, trace_env = trace_env)
      }
    )
  }

  act <- quasi_capture(enquo(object), NULL, capture)
  if (is.null(act$cap)) {
    succeed()
  }
  invisible(act$val)
}

indent_lines <- function(x) {
  paste0("  ", gsub("\n", "\n  ", x))
}

#' Do you expect a reference to this object?
#'
#' `expect_reference()` compares the underlying memory addresses of
#' two symbols. It is for expert use only.
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' `expect_reference()` is deprecated in the third edition. If you know what
#' you're doing, and you really need this behaviour, just use `is_reference()`
#' directly: `expect_true(rlang::is_reference(x, y))`.
#'
#' @inheritParams expect_equal
#' @family expectations
#' @keywords internal
#' @export
expect_reference <- function(
  object,
  expected,
  info = NULL,
  label = NULL,
  expected.label = NULL
) {
  edition_deprecate(3, "expect_reference()")

  act <- quasi_label(enquo(object), label)
  exp <- quasi_label(enquo(expected), expected.label)

  if (!is_reference(act$val, exp$val)) {
    msg <- sprintf("Expected %s to be a reference to %s.", act$lab, exp$lab)
    fail(msg, info = info)
  } else {
    pass()
  }
  invisible(act$val)
}

# expect_reference() needs dev version of rlang
utils::globalVariables("is_reference")

#' Do you expect a vector containing these values?
#'
#' * `expect_setequal(x, y)` tests that every element of `x` occurs in `y`,
#'    and that every element of `y` occurs in `x`.
#' * `expect_contains(x, y)` tests that `x` contains every element of `y`
#'   (i.e. `y` is a subset of `x`).
#' * `expect_in(x, y)` tests that every element of `x` is in `y`
#'   (i.e. `x` is a subset of `y`).
#' * `expect_disjoint(x, y)` tests that no element of `x` is in `y`
#'   (i.e. `x` is disjoint from `y`).
#' * `expect_mapequal(x, y)` treats lists as if they are mappings between names
#'   and values. Concretely, checks that `x` and `y` have the same names, then
#'   checks that `x[names(y)]` equals `y`.
#'
#' Note that `expect_setequal()` ignores names, and you will be warned if both
#' `object` and `expected` have them.
#'
#' @inheritParams expect_equal
#' @export
#' @examples
#' expect_setequal(letters, rev(letters))
#' show_failure(expect_setequal(letters[-1], rev(letters)))
#'
#' x <- list(b = 2, a = 1)
#' expect_mapequal(x, list(a = 1, b = 2))
#' show_failure(expect_mapequal(x, list(a = 1)))
#' show_failure(expect_mapequal(x, list(a = 1, b = "x")))
#' show_failure(expect_mapequal(x, list(a = 1, b = 2, c = 3)))
expect_setequal <- function(object, expected) {
  act <- quasi_label(enquo(object))
  exp <- quasi_label(enquo(expected))

  check_vector(act$val, error_arg = "object")
  check_vector(exp$val, error_arg = "expected")
  if (!is.null(names(act$val)) && !is.null(names(exp$val))) {
    testthat_warn("expect_setequal() ignores names")
  }

  expect_setequal_("Expected %s to have the same values as %s.", act, exp)
}

expect_setequal_ <- function(msg, act, exp, trace_env = caller_env()) {
  act_miss <- unique(act$val[!act$val %in% exp$val])
  exp_miss <- unique(exp$val[!exp$val %in% act$val])

  if (length(exp_miss) == 0 && length(act_miss) == 0) {
    pass()
  } else {
    msg_exp <- sprintf(msg, act$lab, exp$lab)
    msg_act <- c(
      sprintf("Actual: %s", values(act$val)),
      sprintf("Expected: %s", values(exp$val)),
      if (length(act_miss)) sprintf("Needs: %s", values(act_miss)),
      if (length(exp_miss)) sprintf("Absent: %s", values(exp_miss))
    )
    fail(c(msg_exp, msg_act), trace_env = trace_env)
  }

  invisible(act$val)
}

values <- function(x) {
  has_extra <- length(x) > 10
  if (has_extra) {
    x <- x[1:9]
  }
  if (is.character(x)) {
    x <- encodeString(x, quote = '"')
  }

  out <- paste0(x, collapse = ", ")
  if (has_extra) {
    out <- paste0(out, ", ...")
  }
  out
}

is_vector <- function(x) is.list(x) || (is.atomic(x) && !is.null(x))

#' @export
#' @rdname expect_setequal
expect_mapequal <- function(object, expected) {
  act <- quasi_label(enquo(object))
  exp <- quasi_label(enquo(expected))

  check_vector(act$val, error_arg = "object")
  check_map_names(act$val, error_arg = "object")
  check_vector(exp$val, error_arg = "expected")
  check_map_names(exp$val, error_arg = "expected")

  act_nms <- names(act$val)
  exp_nms <- names(exp$val)

  # Length-0 vectors are OK whether named or unnamed.
  if (length(act$val) == 0 && length(exp$val) == 0) {
    testthat_warn("`object` and `expected` are empty vectors.")
    pass()
  } else {
    if (!setequal(act_nms, exp_nms)) {
      msg <- "Expected %s to have the same names as %s."
      act_names <- labelled_value(names(act$val), act$lab)
      exp_names <- labelled_value(names(exp$val), exp$lab)
      expect_setequal_(msg, act_names, exp_names)
    } else {
      if (edition_get() >= 3) {
        act <- labelled_value(act$val[exp_nms], act$lab)

        msg <- "Expected %s to contain the same values as %s."
        expect_waldo_equal_(msg, act, exp, tolerance = testthat_tolerance())
      } else {
        # Packages depend on 2e behaviour, but the expectation isn't written
        # to be reused, and we don't want to bother
        expect_equal(act$val[exp_nms], exp$val)
      }
    }
  }

  invisible(act$val)
}

#' @export
#' @rdname expect_setequal
expect_contains <- function(object, expected) {
  act <- quasi_label(enquo(object))
  exp <- quasi_label(enquo(expected))

  check_vector(act$val, error_arg = "object")
  check_vector(exp$val, error_arg = "expected")

  exp_miss <- !exp$val %in% act$val
  if (any(exp_miss)) {
    msg_exp <- sprintf(
      "Expected %s to contain all values in %s.",
      act$lab,
      exp$lab
    )
    msg_act <- c(
      sprintf("Actual: %s", values(act$val)),
      sprintf("Expected: %s", values(exp$val)),
      sprintf("Missing: %s", values(exp$val[exp_miss]))
    )
    fail(c(msg_exp, msg_act))
  } else {
    pass()
  }

  invisible(act$val)
}


#' @export
#' @rdname expect_setequal
expect_in <- function(object, expected) {
  act <- quasi_label(enquo(object))
  exp <- quasi_label(enquo(expected))

  check_vector(act$val, error_arg = "object")
  check_vector(exp$val, error_arg = "expected")

  act_miss <- !act$val %in% exp$val
  if (any(act_miss)) {
    msg_exp <- sprintf(
      "Expected %s to only contain values from %s.",
      act$lab,
      exp$lab
    )
    msg_act <- c(
      sprintf("Actual: %s", values(act$val)),
      sprintf("Expected: %s", values(exp$val)),
      sprintf("Invalid: %s", values(act$val[act_miss]))
    )
    fail(c(msg_exp, msg_act))
  } else {
    pass()
  }

  invisible(act$val)
}

#' @export
#' @rdname expect_setequal
expect_disjoint <- function(object, expected) {
  act <- quasi_label(enquo(object))
  exp <- quasi_label(enquo(expected))

  check_vector(act$val, error_arg = "object")
  check_vector(exp$val, error_arg = "expected")

  act_common <- act$val %in% exp$val
  if (any(act_common)) {
    fail(c(
      sprintf("Expected %s to be disjoint from %s.", act$lab, exp$lab),
      sprintf("Actual: %s", values(act$val)),
      sprintf("Expected: None of %s", values(exp$val)),
      sprintf("Invalid: %s", values(act$val[act_common]))
    ))
  } else {
    pass()
  }

  invisible(act$val)
}

# Helpers ----------------------------------------------------------------------

check_vector <- function(x, error_arg, error_call = caller_env()) {
  if (!is_vector(x)) {
    stop_input_type(x, "a vector", arg = error_arg, call = error_call)
  }
}

check_map_names <- function(x, error_arg, error_call = caller_env()) {
  nms <- names2(x)
  if (anyDuplicated(nms)) {
    dups <- unique(nms[duplicated(nms)])
    cli::cli_abort(
      c(
        "All elements in {.arg {error_arg}} must have unique names.",
        x = "Duplicate names: {.str {dups}}"
      ),
      call = error_call
    )
  }
  if (any(nms == "")) {
    empty <- which(nms == "")
    cli::cli_abort(
      c(
        "All elements in {.arg {error_arg}} must have names.",
        x = "Empty names at position{?s}: {empty}"
      ),
      call = error_call
    )
  }
}

test_that("can't access variables from other tests (1)", {
  a <- 10
  expect_true(TRUE)
})
test_that("can't access variables from other tests (2)", {
  expect_false(exists("a"))
})

test_that("messages are suppressed", {
  local_edition(2)
  message("YOU SHOULDN'T SEE ME")
  pass()
})

test_that("errors are captured", {
  f <- function() g()
  g <- function() stop("I made a mistake", call. = FALSE)

  reporter <- with_reporter("silent", {
    test_that("", {
      f()
    })
  })
  expect_equal(length(reporter$expectations()), 1)
})

test_that("errors captured even when looking for messages", {
  reporter <- with_reporter("silent", {
    test_that("", {
      expect_message(stop("a"))
    })
  })
  expect_equal(length(reporter$expectations()), 1)
  expect_true(expectation_error(reporter$expectations()[[1L]]))
})

test_that("errors captured even when looking for warnings", {
  reporter <- with_reporter("silent", {
    test_that("", {
      expect_warning(stop())
    })
  })
  expect_equal(length(reporter$expectations()), 1)
  expect_true(expectation_error(reporter$expectations()[[1L]]))
})

test_that("failures are errors", {
  f <- function() {
    expect_true(FALSE)
    expect_false(TRUE)
  }

  expect_error(f(), class = "expectation_failure")
})

test_that("infinite recursion is captured", {
  skip_on_covr()
  f <- function() f()

  reporter <- with_reporter("silent", {
    withr::with_options(
      list(expressions = sys.nframe() + 100),
      test_that("", {
        f()
      })
    )
  })
  expect_equal(length(reporter$expectations()), 1)
})

test_that("return value from test_that", {
  with_reporter(
    "",
    success <- test_that("success", {
      pass()
    })
  )
  expect_true(success)
  with_reporter(
    "",
    success <- test_that("success", {
      succeed("Yes!")
    })
  )
  expect_true(success)

  with_reporter(
    "",
    error <- test_that("error", {
      barf
    })
  )
  expect_false(error)

  with_reporter(
    "",
    failure <- test_that("failure", {
      expect_true(FALSE)
    })
  )
  expect_false(failure)
  with_reporter(
    "",
    failure <- test_that("failure", {
      fail()
    })
  )
  expect_false(failure)
  with_reporter(
    "",
    success <- test_that("failure", {
      fail()
    })
  )
  expect_false(success)

  with_reporter(
    "",
    skip <- test_that("skip", {
      skip("skipping")
    })
  )
  expect_false(skip)
})

test_that("empty test skips automatically", {
  expectations <- capture_expectations(skip <- test_that("success", {}))
  expect_false(skip)
  expect_s3_class(expectations[[1]], "expectation_skip")
})

test_that("nested tests skipped correctly", {
  expectations <- capture_expectations({
    describe("outer", {
      it("1")
      it("2", expect_true(TRUE))
    })
  })
  expect_length(expectations, 2)
  expect_s3_class(expectations[[1]], "expectation_skip")
  expect_s3_class(expectations[[2]], "expectation_success")
})

test_that("can signal warnings and messages without restart", {
  expect_null(signalCondition(message_cnd("foo")))

  return("Skipping following test because it verbosely registers the warning")
  expect_null(signalCondition(warning_cnd("foo")))
})


test_that("no braces required in testthat 2e", {
  local_edition(2)
  expect_warning(
    test_that("", expect_true(TRUE)),
    NA
  )
})

test_that("missing packages cause a skip on CRAN", {
  local_on_cran(TRUE)

  expectations <- capture_expectations(test_that("", {
    library(notinstalled)
  }))
  expect_length(expectations, 1)
  expect_s3_class(expectations[[1]], "expectation_skip")

  local_on_cran(FALSE)
  expectations <- capture_expectations(test_that("", {
    library(notinstalled)
  }))
  expect_length(expectations, 1)
  expect_s3_class(expectations[[1]], "expectation_error")
})

#' Do you expect the result to be (in)visible?
#'
#' Use this to test whether a function returns a visible or invisible
#' output. Typically you'll use this to check that functions called primarily
#' for their side-effects return their data argument invisibly.
#'
#' @param call A function call.
#' @inheritParams expect_that
#' @return The evaluated `call`, invisibly.
#' @export
#' @examples
#' expect_invisible(x <- 10)
#' expect_visible(x)
#'
#' # Typically you'll assign the result of the expectation so you can
#' # also check that the value is as you expect.
#' greet <- function(name) {
#'   message("Hi ", name)
#'   invisible(name)
#' }
#' out <- expect_invisible(greet("Hadley"))
#' expect_equal(out, "Hadley")
expect_invisible <- function(call, label = NULL) {
  lab <- label %||% expr_label(enexpr(call))
  vis <- withVisible(call)

  if (!identical(vis$visible, FALSE)) {
    fail(c(
      sprintf("Expected %s to return invisibly.", lab),
      "Actual visibility: visible."
    ))
  } else {
    pass()
  }
  invisible(vis$value)
}

#' @export
#' @rdname expect_invisible
expect_visible <- function(call, label = NULL) {
  lab <- label %||% expr_label(enexpr(call))
  vis <- withVisible(call)

  if (!identical(vis$visible, TRUE)) {
    fail(c(
      sprintf("Expected %s to return visibly.", lab),
      "Actual visibility: invisible."
    ))
  } else {
    pass()
  }
  invisible(vis$value)
}

snapshot_cleanup <- function(
  path,
  test_files_seen = character(),
  snap_files_seen = character()
) {
  outdated <- snapshot_outdated(path, test_files_seen, snap_files_seen)

  if (length(outdated) > 0) {
    cli::cli_inform("Deleting unused snapshots: {.path {outdated}}")
    unlink(file.path(path, outdated), recursive = TRUE)
  }

  # Delete empty directories:
  # nest dir() inside list.dirs() to avoid picking up `.` directories
  dirs <- list.dirs(dir(path, full.names = TRUE))
  empty <- dirs[map_lgl(dirs, is_dir_empty)]
  unlink(empty, recursive = TRUE)

  # Delete snapshot folder
  if (is_dir_empty(path)) {
    unlink(path, recursive = TRUE)
  }

  rstudio_tickle()

  invisible(outdated)
}

is_dir_empty <- function(x) {
  length(dir(x, recursive = TRUE)) == 0
}

snapshot_outdated <- function(
  path,
  test_files_seen = character(),
  snap_files_seen = character()
) {
  all_files <- dir(path, recursive = TRUE)
  expected <- snapshot_expected(path, test_files_seen, snap_files_seen)
  setdiff(all_files, expected)
}

snapshot_expected <- function(
  snap_dir,
  test_files_seen = character(),
  snap_files_seen = character()
) {
  if (length(test_files_seen) > 0) {
    snaps <- c(
      paste0(test_files_seen, ".md"),
      paste0(test_files_seen, ".new.md")
    )
  } else {
    snaps <- character()
  }

  # Empirically determine variants
  snap_dirs <- list.dirs(snap_dir, recursive = FALSE)
  is_variant <- dir_contains(snap_dirs, c(snaps, snap_files_seen))
  variants <- basename(snap_dirs[is_variant])

  snap_files_seen_new <- paste0(
    tools::file_path_sans_ext(snap_files_seen),
    ".new.",
    tools::file_ext(snap_files_seen)
  )

  sort(c(
    snaps,
    outer(variants, snaps, file.path),
    snap_files_seen,
    outer(variants, snap_files_seen, file.path),
    snap_files_seen_new,
    outer(variants, snap_files_seen_new, file.path)
  ))
}

dir_contains <- function(paths, expected_files) {
  map_lgl(paths, \(path) any(file.exists(file.path(path, expected_files))))
}

test_that("snapshot cleanup makes nice message if needed", {
  dir <- local_snap_dir(c("a.md", "b.md"))
  expect_snapshot({
    snapshot_cleanup(dir)
    snapshot_cleanup(dir, c("a", "b"))
  })
})

test_that("deletes empty dirs", {
  dir <- local_snap_dir(character())
  dir.create(file.path(dir, "a", "b", "c"), recursive = TRUE)
  dir.create(file.path(dir, "b"), recursive = TRUE)
  dir.create(file.path(dir, "c"), recursive = TRUE)

  snapshot_cleanup(dir)
  expect_equal(dir(dir), character())
})

test_that("detects outdated snapshots", {
  dir <- local_snap_dir(c("a.md", "b.md", "b.new.md"))
  expect_equal(snapshot_outdated(dir, c("a", "b")), character())
  expect_equal(snapshot_outdated(dir, "a"), c("b.md", "b.new.md"))
  expect_equal(snapshot_outdated(dir, "b"), "a.md")
  expect_equal(snapshot_outdated(dir), c("a.md", "b.md", "b.new.md"))
})

test_that("preserves variants", {
  dir <- local_snap_dir(c("a.md", "windows/a.md", "windows/b.md"))
  expect_equal(snapshot_outdated(dir, "a"), "windows/b.md")

  # Doesn't delete new files in variants
  dir <- local_snap_dir(c("a.md", "windows/a.md", "windows/a.new.md"))
  expect_equal(snapshot_outdated(dir, "a"), character())
})

test_that("detects outdated snapshot files", {
  dir <- local_snap_dir(c("a/foo.txt", "b/foo.txt", "b/foo.new.txt"))

  expect_equal(
    snapshot_outdated(dir, character(), character()),
    c("a/foo.txt", "b/foo.new.txt", "b/foo.txt")
  )
  expect_equal(
    snapshot_outdated(dir, character(), "a/foo.txt"),
    c("b/foo.new.txt", "b/foo.txt")
  )
  expect_equal(
    snapshot_outdated(dir, character(), "b/foo.txt"),
    "a/foo.txt"
  )
  expect_equal(
    snapshot_outdated(dir, character(), c("a/foo.txt", "b/foo.txt")),
    character()
  )
})

test_that("detects individual snapshots files to remove", {
  dir <- local_snap_dir(c("a/a1", "a/a2", "b/b1"))
  expect_equal(
    snapshot_outdated(dir, c("a", "b"), "a/a1"),
    c("a/a2", "b/b1")
  )
})

#' The previous building block of all `expect_` functions
#'
#' Previously, we recommended using `expect()` when writing your own
#' expectations. Now we instead recommend [pass()] and [fail()]. See
#' `vignette("custom-expectation")` for details.
#'
#' @param ok `TRUE` or `FALSE` indicating if the expectation was successful.
#' @param failure_message A character vector describing the failure. The
#'   first element should describe the expected value, and the second (and
#'   optionally subsequence) elements should describe what was actually seen.
#' @inheritParams fail
#' @return An expectation object from either `succeed()` or `fail()`.
#'   with a `muffle_expectation` restart.
#' @seealso [exp_signal()]
#' @keywords internal
#' @export
expect <- function(
  ok,
  failure_message,
  info = NULL,
  srcref = NULL,
  trace = NULL,
  trace_env = caller_env()
) {
  check_bool(ok)
  check_character(failure_message)

  if (!ok) {
    fail(
      failure_message,
      info,
      srcref = srcref,
      trace = trace,
      trace_env = trace_env
    )
  } else {
    # For backwards compatibility
    succeed(failure_message)
  }
}

#' Expectation conditions
#'
#' @description
#' `new_expectation()` creates an expectation condition object and
#' `exp_signal()` signals it. `expectation()` does both. `is.expectation()`
#' tests if a captured condition is a testthat expectation.
#'
#' These functions are primarily for internal use. If you are creating your
#' own expectation, you do not need these functions are instead should use
#' [pass()] or [fail()]. See `vignette("custom-expectation")` for more
#' details.
#'
#' @param type Expectation type. Must be one of "success", "failure", "error",
#'   "skip", "warning".
#' @param message Message describing test failure
#' @param srcref Optional `srcref` giving location of test.
#' @keywords internal
#' @inheritParams expect
#' @export
expectation <- function(type, message, ..., srcref = NULL, trace = NULL) {
  exp <- new_expectation(type, message, ..., srcref = srcref, trace = trace)
  exp_signal(exp)
}
#' @rdname expectation
#' @param ... Additional attributes for the expectation object.
#' @param .subclass An optional subclass for the expectation object.
#' @export
new_expectation <- function(
  type,
  message,
  ...,
  srcref = NULL,
  trace = NULL,
  .subclass = NULL
) {
  type <- match.arg(type, c("success", "failure", "error", "skip", "warning"))

  structure(
    list(
      message = message,
      srcref = srcref,
      trace = trace
    ),
    class = c(
      .subclass,
      paste0("expectation_", type),
      "expectation",
      # Make broken expectations catchable by try()
      if (type %in% c("failure", "error")) "error",
      "condition"
    ),
    ...
  )
}
#' @rdname expectation
#' @param exp An expectation object, as created by
#'   [new_expectation()].
#' @export
exp_signal <- function(exp) {
  withRestarts(
    if (expectation_broken(exp)) {
      stop(exp)
    } else {
      signalCondition(exp)
    },
    muffle_expectation = function(e) NULL,
    # Legacy support for shinytest2
    # https://github.com/r-lib/testthat/pull/2271#discussion_r2528722708
    continue_test = function(e) NULL
  )

  invisible(exp)
}


#' @export
#' @rdname expectation
#' @param x object to test for class membership
is.expectation <- function(x) inherits(x, "expectation")

#' @export
print.expectation <- function(x, ...) {
  cat(
    cli::style_bold("<", paste0(class(x), collapse = "/"), ">"),
    "\n",
    sep = ""
  )
  cat(format(x), "\n", sep = "")
  invisible(x)
}

#' @export
format.expectation_success <- function(x, ...) {
  "As expected"
}

#' @export
format.expectation <- function(x, ...) {
  # Access error fields with `[[` rather than `$` because the
  # `$.Throwable` from the rJava package throws with unknown fields
  if (is.null(x[["trace"]]) || trace_length(x[["trace"]]) == 0L) {
    return(x$message)
  }

  trace_lines <- format(x$trace, ...)
  lines <- c(x$message, cli::style_bold("Backtrace:"), trace_lines)
  paste(lines, collapse = "\n")
}

# as.expectation ----------------------------------------------------------

as.expectation <- function(x, srcref = NULL) {
  UseMethod("as.expectation", x)
}

#' @export
as.expectation.expectation <- function(x, srcref = NULL) {
  x$srcref <- x$srcref %||% srcref
  x
}

#' @export
as.expectation.error <- function(x, srcref = NULL) {
  if (is.null(x$call)) {
    header <- paste0("Error: ")
  } else {
    header <- paste0("Error in `", deparse1(x$call), "`: ")
  }

  msg <- paste0(
    if (!is_simple_error(x)) {
      paste0("<", paste(class(x), collapse = "/"), ">\n")
    },
    header,
    cnd_message(x)
  )

  new_expectation("error", msg, srcref = srcref, trace = x[["trace"]])
}


is_simple_error <- function(x) {
  class(x)[[1]] %in% c("simpleError", "rlang_error")
}

#' @export
as.expectation.warning <- function(x, srcref = NULL) {
  new_expectation(
    "warning",
    cnd_message(x),
    srcref = srcref,
    trace = x[["trace"]]
  )
}

#' @export
as.expectation.skip <- function(x, ..., srcref = NULL) {
  new_expectation("skip", cnd_message(x), srcref = srcref, trace = x[["trace"]])
}

#' @export
as.expectation.default <- function(x, srcref = NULL) {
  cli::cli_abort(
    "Don't know how to convert {.cls {class(x)}} to expectation.",
    call = NULL
  )
}

# expectation_type --------------------------------------------------------

expectation_type <- function(exp) {
  stopifnot(is.expectation(exp))
  gsub("^expectation_", "", class(exp)[[1]])
}

expectation_success <- function(exp) expectation_type(exp) == "success"
expectation_failure <- function(exp) expectation_type(exp) == "failure"
expectation_error <- function(exp) expectation_type(exp) == "error"
expectation_skip <- function(exp) expectation_type(exp) == "skip"
expectation_warning <- function(exp) expectation_type(exp) == "warning"
expectation_broken <- function(exp) {
  expectation_failure(exp) || expectation_error(exp)
}
expectation_ok <- function(exp) {
  expectation_type(exp) %in% c("success", "warning")
}

single_letter_summary <- function(x) {
  switch(
    expectation_type(x),
    skip = colourise("S", "skip"),
    success = colourise(".", "success"),
    error = colourise("E", "error"),
    failure = colourise("F", "failure"),
    warning = colourise("W", "warning"),
    "?"
  )
}

expectation_location <- function(x, prefix = "", suffix = "") {
  srcref <- x$srcref
  if (!inherits(srcref, "srcref")) {
    return("")
  }

  filename <- attr(srcref, "srcfile")$filename
  cli::format_inline(
    "{prefix}{.file {filename}:{srcref[1]}:{srcref[2]}}{suffix}"
  )
}

test_that("expectation contains failure message even when successful", {
  e <- expect(TRUE, "I failed")
  expect_equal(e$message, "I failed")
})

test_that("info only evaluated on failure", {
  expect_no_error(expect(TRUE, "fail", info = stop("!")))
})

test_that("validates key inputs", {
  expect_snapshot(error = TRUE, {
    expect(1)
    expect(TRUE, 1)
  })
})

test_that("can subclass expectation", {
  exp <- new_expectation(
    "failure",
    "didn't work",
    .subclass = "foo",
    bar = "baz"
  )
  expect_true(inherits_all(
    exp,
    c("foo", "expectation_failure", "expectation", "error", "condition")
  ))
  expect_identical(attr(exp, "bar"), "baz")
})

test_that("`expect()` and `exp_signal()` signal expectations", {
  expect_success(expect(TRUE, ""))
  expect_failure(expect(FALSE, ""))

  expect_success(exp_signal(new_expectation("success", "")))
  expect_failure(exp_signal(new_expectation("failure", "")))
})

test_that("conditionMessage() is called during conversion", {
  local_bindings(
    conditionMessage.foobar = function(...) "dispatched",
    .env = global_env()
  )

  wrn <- warning_cnd("foobar", message = "wrong")
  expect_identical(as.expectation(wrn)$message, "dispatched")

  err <- error_cnd("foobar", message = "wrong")
  expect_match(as.expectation(err)$message, "Error: dispatched")

  err <- cnd(c("foobar", "skip"), message = "wrong")
  expect_identical(as.expectation(err)$message, "dispatched")
})

test_that("error message includes call", {
  f <- function() stop("Error!")
  cnd <- catch_cnd(f())
  expect_equal(format(as.expectation(cnd)), "Error in `f()`: Error!")
})

# To allow the Java-style class name format that Jenkins prefers,
# "package_name_or_domain.ClassName", allow "."s in the class name.
classnameOK <- function(text) {
  gsub("[^._A-Za-z0-9]+", "_", text)
}


#' Report results in jUnit XML format
#'
#' @description
#' This reporter includes detailed results about each test and summaries,
#' written to a file (or stdout) in jUnit XML format. This can be read by
#' the Jenkins Continuous Integration System to report on a dashboard etc.
#' Requires the _xml2_ package.
#'
#' To fit into the jUnit structure, `context()` becomes the `<testsuite>`
#' name as well as the base of the `<testcase> classname`. The
#' `test_that()` name becomes the rest of the `<testcase> classname`.
#' The deparsed `expect_that()` call becomes the `<testcase>` name.
#' On failure, the message goes into the `<failure>` node message
#' argument (first line only) and into its text content (full message).
#' Execution time and some other details are also recorded.
#'
#' References for the jUnit XML format:
#' <https://github.com/testmoapp/junitxml>
#'
#' @export
#' @family reporters
JunitReporter <- R6::R6Class(
  "JunitReporter",
  inherit = Reporter,
  public = list(
    results = NULL,
    timer = NULL,
    doc = NULL,
    errors = NULL,
    failures = NULL,
    skipped = NULL,
    tests = NULL,
    root = NULL,
    suite = NULL,
    suite_time = NULL,
    file_name = NULL,

    elapsed_time = function() {
      time <- (private$proctime() - self$timer)[["elapsed"]]
      self$timer <- private$proctime()
      time
    },

    reset_suite = function() {
      self$errors <- 0
      self$failures <- 0
      self$skipped <- 0
      self$tests <- 0
      self$suite_time <- 0
    },

    start_reporter = function() {
      check_installed("xml2", "to use JunitReporter")

      self$timer <- private$proctime()
      self$doc <- xml2::xml_new_document()
      self$root <- xml2::xml_add_child(self$doc, "testsuites")
      self$reset_suite()
    },

    start_file = function(file) {
      self$file_name <- file
    },

    start_test = function(context, test) {
      if (is.null(context)) {
        context_start_file(self$file_name)
      }
    },

    start_context = function(context) {
      self$suite <- xml2::xml_add_child(
        self$root,
        "testsuite",
        name = context,
        timestamp = private$timestamp(),
        hostname = private$hostname()
      )
    },

    end_context = function(context) {
      # Always uses . as decimal place in output regardless of options set in test
      withr::local_options(list(OutDec = "."))
      xml2::xml_attr(self$suite, "tests") <- as.character(self$tests)
      xml2::xml_attr(self$suite, "skipped") <- as.character(self$skipped)
      xml2::xml_attr(self$suite, "failures") <- as.character(self$failures)
      xml2::xml_attr(self$suite, "errors") <- as.character(self$errors)
      #jenkins junit plugin requires time has at most 3 digits
      xml2::xml_attr(self$suite, "time") <- as.character(round(
        self$suite_time,
        3
      ))

      self$reset_suite()
    },

    add_result = function(context, test, result) {
      withr::local_options(list(OutDec = "."))
      self$tests <- self$tests + 1

      time <- self$elapsed_time()
      self$suite_time <- self$suite_time + time

      # If no context was started (e.g., warnings outside tests), create a default one
      if (is.null(self$suite)) {
        self$start_context(context %||% "(unknown)")
      }

      # XML node for test case
      name <- test %||% "(unnamed)"
      testcase <- xml2::xml_add_child(
        self$suite,
        "testcase",
        time = toString(time),
        classname = classnameOK(context),
        name = classnameOK(name)
      )

      first_line <- function(x) {
        loc <- expectation_location(x, " (", ")")
        paste0(strsplit(cli::ansi_strip(x$message), split = "\n")[[1]][1], loc)
      }

      # add an extra XML child node if not a success
      if (expectation_error(result)) {
        # "type" in Java is the exception class
        error <- xml2::xml_add_child(
          testcase,
          "error",
          type = "error",
          message = first_line(result)
        )
        xml2::xml_text(error) <- cli::ansi_strip(format(result))
        self$errors <- self$errors + 1
      } else if (expectation_failure(result)) {
        # "type" in Java is the type of assertion that failed
        failure <- xml2::xml_add_child(
          testcase,
          "failure",
          type = "failure",
          message = first_line(result)
        )
        xml2::xml_text(failure) <- cli::ansi_strip(format(result))
        self$failures <- self$failures + 1
      } else if (expectation_skip(result)) {
        xml2::xml_add_child(testcase, "skipped", message = first_line(result))
        self$skipped <- self$skipped + 1
      } else if (expectation_warning(result)) {
        warning_node <- xml2::xml_add_child(testcase, "system-out")
        xml2::xml_text(warning_node) <- cli::ansi_strip(format(result))
      }
    },

    end_reporter = function() {
      if (is.character(self$out)) {
        xml2::write_xml(self$doc, self$out, format = TRUE)
      } else if (inherits(self$out, "connection")) {
        file <- withr::local_tempfile()
        xml2::write_xml(self$doc, file, format = TRUE)
        cat(brio::read_file(file), file = self$out)
      } else {
        cli::cli_abort("Unsupported output type: {toString(self$out)}.")
      }
    } # end_reporter
  ), # public

  private = list(
    proctime = function() {
      proc.time()
    },
    timestamp = function() {
      strftime(Sys.time(), "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
    },
    hostname = function() {
      Sys.info()[["nodename"]]
    }
  ) # private
)

# Fix components of JunitReporter that otherwise vary from run-to-run
#
# The following functions need to be mocked out to run a unit test
# against static contents of reporters/junit.txt:
#   - proctime - originally wrapper for proc.time()
#   - timestamp - originally wrapper for toString(Sys.time())
#   - hostname  - originally wrapper for Sys.info()[["nodename"]]
#
JunitReporterMock <- R6::R6Class(
  "JunitReporterMock",
  inherit = JunitReporter,
  public = list(),
  private = list(
    proctime = function() {
      c(user = 0, system = 0, elapsed = 0)
    },
    timestamp = function() {
      "1999:12:31 23:59:59"
    },
    hostname = function() {
      "nodename"
    }
  )
)

#' Mock S3 and S4 methods
#'
#' @description
#' These functions temporarily override S3 or S4 methods. They can mock
#' methods that don't already exist, or temporarily remove a method by setting
#' `definition = NULL`.
#'
#' Learn more about mocking in `vignette("mocking")`.
#'
#' @param generic A string giving the name of the generic.
#' @param signature A character vector giving the signature of the method.
#' @param definition A function providing the method definition, or `NULL` to
#'   temporarily remove the method.
#' @param frame Calling frame which determines the scope of the mock.
#'   Only needed when wrapping in another local helper.
#' @export
#' @examples
#' x <- as.POSIXlt(Sys.time())
#' local({
#'   local_mocked_s3_method("length", "POSIXlt", function(x) 42)
#'   length(x)
#' })
#'
#' length(x)
local_mocked_s3_method <- function(
  generic,
  signature,
  definition,
  frame = caller_env()
) {
  check_string(generic)
  check_string(signature)
  check_function(definition, allow_null = TRUE)

  old <- utils::getS3method(generic, signature, optional = TRUE)

  # Set the new method, or a pass-through stub if removing
  definition <- definition %||% function(...) NextMethod()
  registerS3method(generic, signature, definition, envir = frame)

  # On cleanup, restore old method or remove the one we added
  if (is.null(old)) {
    withr::defer(remove_s3_method(generic, signature, envir = frame), frame)
  } else {
    withr::defer(
      registerS3method(generic, signature, old, envir = frame),
      frame
    )
  }

  invisible()
}

remove_s3_method <- function(generic, class, envir) {
  # Extracted from registerS3method()
  group_generics <- c("Math", "Ops", "matrixOps", "Summary", "Complex")
  if (generic %in% group_generics) {
    s3_envir <- .BaseNamespaceEnv
  } else {
    genfun <- get(generic, envir = envir)
    s3_envir <- environment(genfun) %||% .BaseNamespaceEnv
  }

  if (env_has(s3_envir, ".__S3MethodsTable__.")) {
    table <- env_get(s3_envir, ".__S3MethodsTable__.")
    env_unbind(table, paste0(generic, ".", class))
  }
}

#' @rdname local_mocked_s3_method
#' @export
local_mocked_s4_method <- function(
  generic,
  signature,
  definition,
  frame = caller_env()
) {
  check_string(generic)
  check_character(signature)
  check_function(definition, allow_null = TRUE)

  generic_def <- methods::getGeneric(generic)
  if (is.null(generic_def)) {
    cli::cli_abort("Can't find generic {.fn {generic}}.")
  }

  set_method <- function(generic, signature, def) {
    env <- topenv(frame)
    old <- methods::getMethod(generic, signature, optional = TRUE)
    if (is.null(def)) {
      methods::removeMethod(generic, signature, env)
    } else {
      suppressMessages(methods::setMethod(generic, signature, def, env))
    }
    old
  }

  old <- set_method(generic_def, signature, definition)
  withr::defer(set_method(generic_def, signature, old), frame)

  invisible()
}


#' Mock an R6 class
#'
#' @description
#' This function allows you to temporarily override an R6 class definition.
#' It works by creating a subclass then using [local_mocked_bindings()] to
#' temporarily replace the original definition. This means that it will not
#' affect subclasses of the original class; please file an issue if you need
#' this.
#'
#' Learn more about mocking in `vignette("mocking")`.
#'
#' @export
#' @param class An R6 class definition.
#' @param public,private A named list of public and private methods/data.
#' @inheritParams local_mocked_s3_method
local_mocked_r6_class <- function(
  class,
  public = list(),
  private = list(),
  frame = caller_env()
) {
  if (!inherits(class, "R6ClassGenerator")) {
    stop_input_type(class, "an R6 class definition")
  }
  if (!is.list(public)) {
    stop_input_type(public, "a list")
  }
  if (!is.list(private)) {
    stop_input_type(private, "a list")
  }

  mocked_class <- mock_r6_class(class, public, private)
  local_mocked_bindings("{class$classname}" := mocked_class, .env = frame)
}

mock_r6_class <- function(class, public = list(), private = list()) {
  R6::R6Class(
    paste0("Mocked", class$classname),
    inherit = class,
    private = private,
    public = public
  )
}

# For testing ------------------------------------------------------------------

TestMockClass <- R6::R6Class(
  "TestMockClass",
  public = list(
    sum = function() {
      self$public_fun() +
        self$public_val +
        private$private_fun() +
        private$private_val
    },
    public_fun = function() 1,
    public_val = 20
  ),
  private = list(
    private_fun = function() 300,
    private_val = 4000
  )
)

# Silence R CMD check NOTE: I think it's because TestMockPerson is a claas
# constructor
#' @importFrom methods new
TestMockPerson <- methods::setClass(
  "TestMockPerson",
  slots = c(name = "character", age = "numeric")
)
methods::setGeneric("mock_age", function(x) standardGeneric("mock_age"))
methods::setMethod("mock_age", "TestMockPerson", function(x) x@age)

# S3 --------------------------------------------------------------------------

test_that("can mock S3 methods", {
  x <- as.POSIXlt(Sys.time())

  local({
    local_mocked_s3_method("length", "POSIXlt", function(x) 42)
    expect_length(x, 42)
  })

  expect_length(x, 1)
})

test_that("validates its inputs", {
  expect_snapshot(error = TRUE, {
    local_mocked_s3_method(1)
    local_mocked_s3_method("mean", 1)
    local_mocked_s3_method("mean", "bar", 1)
    local_mocked_s3_method("notAGeneric", "bar", function() {})
  })
})

test_that("can mock S3 method that doesn't exist yet", {
  x <- structure(list(), class = "test_mock_class")

  local({
    local_mocked_s3_method("length", "test_mock_class", function(x) 42)
    expect_length(x, 42)
  })

  # Method should be removed after scope ends
  expect_length(x, 0)
})

test_that("can temporarily remove S3 method with NULL", {
  x <- structure(list(), class = "test_mock_class2")
  local_mocked_s3_method("length", "test_mock_class2", function(x) 42)

  local({
    # Now remove it
    local_mocked_s3_method("length", "test_mock_class2", NULL)
    expect_length(x, 0)
  })

  # Method should be restored after scope ends
  expect_length(x, 42)
})

# S4 --------------------------------------------------------------------------

test_that("can mock S4 methods", {
  jim <- TestMockPerson(name = "Jim", age = 32)

  local({
    local_mocked_s4_method("mock_age", "TestMockPerson", function(x) 42)
    expect_equal(mock_age(jim), 42)
  })

  expect_equal(mock_age(jim), 32)
})

test_that("validates its inputs", {
  expect_snapshot(error = TRUE, {
    local_mocked_s4_method(1)
    local_mocked_s4_method("mean", 1)
    local_mocked_s4_method("mean", "bar", 1)
    local_mocked_s4_method("notAGeneric", "bar", function() {})
  })
})

test_that("can mock S4 method that doesn't exist yet", {
  jim <- TestMockPerson(name = "Jim", age = 32)

  local({
    local_mocked_s4_method("show", "TestMockPerson", function(object) {
      cat("Person:", object@name, "\n")
    })
    expect_output(show(jim), "Person: Jim")
  })

  expect_null(methods::getMethod("show", "TestMockPerson", optional = TRUE))
})

test_that("can temporarily remove S4 method with NULL", {
  jim <- TestMockPerson(name = "Jim", age = 32)
  expect_equal(mock_age(jim), 32)

  local({
    local_mocked_s4_method("mock_age", "TestMockPerson", NULL)
    # Method is removed, so this should error
    expect_error(mock_age(jim), "unable to find")
  })

  expect_equal(mock_age(jim), 32)
})

# R6 --------------------------------------------------------------------------

test_that("can mock R6 methods", {
  local({
    local_mocked_r6_class(TestMockClass, public = list(sum = function() 2))
    obj <- TestMockClass$new()
    expect_equal(obj$sum(), 2)
  })

  obj <- TestMockClass$new()
  expect_equal(obj$sum(), 4321)
})

test_that("can mock all R6 components", {
  local_mocked_r6_class(
    TestMockClass,
    public = list(public_fun = function() 0, public_val = 0),
    private = list(private_fun = function() 0, private_val = 0)
  )
  obj <- TestMockClass$new()
  expect_equal(obj$sum(), 0)
})

test_that("validates its inputs", {
  expect_snapshot(error = TRUE, {
    local_mocked_r6_class(mean)
    local_mocked_r6_class(TestMockClass, public = 1)
    local_mocked_r6_class(TestMockClass, private = 1)
  })
})

SnapshotReporter <- R6::R6Class(
  "SnapshotReporter",
  inherit = Reporter,
  public = list(
    snap_dir = character(),
    file = NULL,
    test = NULL,
    test_file_seen = character(),
    snap_file_seen = character(),
    snap_file_saved = character(),
    variants_changed = FALSE,
    fail_on_new = NULL,
    desc = NULL,

    old_snaps = NULL,
    cur_snaps = NULL,
    new_snaps = NULL,

    initialize = function(
      snap_dir = "_snaps",
      fail_on_new = NULL,
      desc = NULL
    ) {
      self$snap_dir <- normalizePath(snap_dir, mustWork = FALSE)
      self$fail_on_new <- fail_on_new
      self$desc <- desc
    },

    start_file = function(path, test = NULL) {
      self$file <- context_name(path)
      self$test_file_seen <- c(self$test_file_seen, self$file)
      self$snap_file_saved <- character()

      self$variants_changed <- character()

      self$old_snaps <- FileSnaps$new(self$snap_dir, self$file, type = "old")
      self$cur_snaps <- FileSnaps$new(self$snap_dir, self$file, type = "cur")
      self$new_snaps <- FileSnaps$new(self$snap_dir, self$file, type = "new")

      if (!is.null(self$desc)) {
        # When filtering tests, we need to copy over all of the old snapshots,
        # apart from the one that matches the test
        snaps <- self$old_snaps$snaps
        test_name <- test_description(self$desc)
        for (variant in names(snaps)) {
          # In the case of subtests, snaps are named a / b / c1, a / b / c2 etc.
          # So if we run a / b, we want to remove a / b, a / b / c, a / b / c2
          # Subtests that use / in their names are not currently supported.
          matches <- startsWith(names(snaps[[variant]]), test_name)
          # Can't just remove because we want to preserve order
          snaps[[variant]][matches] <- rep(list(NULL), sum(matches))
        }
        self$cur_snaps$snaps <- snaps
      }

      if (!is.null(test)) {
        self$start_test(NULL, test)
      }
    },

    start_test = function(context, test) {
      if (is.null(test)) {
        return()
      }

      self$test <- gsub("\n", "", test)
    },

    # Called by expectation
    take_snapshot = function(
      value,
      save = identity,
      load = identity,
      ...,
      tolerance = testthat_tolerance(),
      variant = NULL,
      trace_env = caller_env()
    ) {
      check_string(self$test, allow_empty = FALSE)
      i <- self$new_snaps$append(self$test, variant, save(value))

      old_raw <- self$old_snaps$get(self$test, variant, i)
      if (!is.null(old_raw)) {
        self$cur_snaps$append(self$test, variant, old_raw)
        old <- load(old_raw)

        comp <- waldo_compare(
          x = old,
          x_arg = "old",
          y = value,
          y_arg = "new",
          ...,
          tolerance = tolerance,
          quote_strings = FALSE
        )

        if (length(comp) > 0L) {
          self$variants_changed <- union(self$variants_changed, variant)
        } else {
          # Use the old value for the new snapshot so the snapshot remains
          # unchanged if the values compare as equal
          self$new_snaps$set(self$test, variant, i, old_raw)
        }
        comp
      } else {
        value_enc <- save(value)

        self$cur_snaps$append(self$test, variant, value_enc)
        fail_on_new <- self$fail_on_new %||% on_ci()

        message <- paste0(
          "Adding new snapshot",
          if (variant != "_default") paste0(" for variant '", variant, "'"),
          ":\n",
          value_enc
        )
        if (fail_on_new) {
          fail(message, trace_env = trace_env)
          return()
        }
        testthat_warn(message)
        character()
      }
    },

    take_file_snapshot = function(
      name,
      path,
      file_equal,
      variant = NULL,
      trace_env = caller_env()
    ) {
      self$announce_file_snapshot(name)

      save_path <- paste0(c(self$file, variant, name), collapse = "/")
      if (save_path %in% self$snap_file_saved) {
        cli::cli_abort(
          "Snapshot file names must be unique. {.val {name}} has already been used.",
          call = trace_env
        )
      }
      self$snap_file_saved <- c(self$snap_file_saved, save_path)

      snapshot_file_equal(
        snap_dir = self$snap_dir,
        snap_test = self$file,
        snap_name = name,
        snap_variant = variant,
        path = path,
        file_equal = file_equal,
        fail_on_new = self$fail_on_new,
        trace_env = trace_env
      )
    },

    # Also called from announce_snapshot_file()
    announce_file_snapshot = function(name) {
      self$snap_file_seen <- c(self$snap_file_seen, file.path(self$file, name))
    },

    add_result = function(context, test, result) {
      if (is.null(self$test)) {
        return()
      }

      # If expectation errors or skips, need to copy snapshots from old to cur
      # TODO: the logic is not correct here for subtests, probably because
      # the code was not written under the assumption that start_test()
      # generates a stack of tests. You can see the problem by running
      # local_on_cran() then testing describe.R.
      if (expectation_error(result) || expectation_skip(result)) {
        self$cur_snaps$reset(self$test, self$old_snaps)
      }
    },

    end_file = function() {
      dir.create(self$snap_dir, showWarnings = FALSE)

      self$cur_snaps$write()

      for (variant in self$new_snaps$variants()) {
        if (variant %in% self$variants_changed) {
          self$new_snaps$write(variant)
        } else {
          self$new_snaps$delete(variant)
        }
      }
    },
    end_reporter = function() {
      # clean up if we've seen all files
      tests <- context_name(find_test_scripts(".", full.names = FALSE))
      if (!on_ci() && all(tests %in% self$test_file_seen)) {
        snapshot_cleanup(
          self$snap_dir,
          test_files_seen = self$test_file_seen,
          snap_files_seen = self$snap_file_seen
        )
      }
    },

    is_active = function() {
      !is.null(self$file) && !is.null(self$test)
    },

    snap_files = function() {
      dir(self$snap_dir, recursive = TRUE)
    }
  )
)

# set/get active snapshot reporter ----------------------------------------

get_snapshotter <- function() {
  x <- getOption("testthat.snapshotter")
  if (is.null(x)) {
    return()
  }

  if (!x$is_active()) {
    return()
  }

  x
}

#' Instantiate local snapshotting context
#'
#' Needed if you want to run snapshot tests outside of the usual testthat
#' framework For expert use only.
#'
#' @export
#' @keywords internal
local_snapshotter <- function(
  snap_dir = "_snaps",
  reporter = SnapshotReporter,
  cleanup = FALSE,
  desc = NULL,
  fail_on_new = NULL,
  frame = caller_env()
) {
  reporter <- reporter$new(
    snap_dir = snap_dir,
    fail_on_new = fail_on_new,
    desc = desc
  )
  withr::local_options("testthat.snapshotter" = reporter, .local_envir = frame)

  reporter
}

local_test_snapshotter <- function(
  snap_dir = NULL,
  desc = NULL,
  frame = caller_env()
) {
  snap_dir <- snap_dir %||% withr::local_tempdir(.local_envir = frame)
  local_snapshotter(
    snap_dir = snap_dir,
    desc = desc,
    fail_on_new = FALSE,
    frame = frame
  )
}

test_that("can establish local snapshotter for testing", {
  snapper <- local_test_snapshotter()

  snapper$start_file("snapshot-1", "test")
  expect_true(snapper$is_active())
  expect_equal(snapper$file, "snapshot-1")
  expect_equal(snapper$test, "test")
})

test_that("basic workflow", {
  local_on_cran(FALSE)

  path <- withr::local_tempdir()
  snapper <- local_test_snapshotter(snap_dir = path)
  snapper$start_file("snapshot-2")
  # output if not active (because test not set here)
  expect_snapshot_output("x") |>
    expect_message("Can't save") |>
    expect_output("[1] \"x\"", fixed = TRUE)

  # warns on first creation
  snapper$start_file("snapshot-2", "test")
  expect_warning(expect_snapshot_output("x"), "Adding new")
  snapper$end_file()
  expect_true(file.exists(file.path(path, "snapshot-2.md")))
  expect_false(file.exists(file.path(path, "snapshot-2.new.md")))

  # succeeds if unchanged
  snapper$start_file("snapshot-2", "test")
  expect_success(expect_snapshot_output("x"))
  snapper$end_file()
  expect_true(file.exists(file.path(path, "snapshot-2.md")))
  expect_false(file.exists(file.path(path, "snapshot-2.new.md")))

  # fails if changed
  snapper$start_file("snapshot-2", "test")
  expect_failure(expect_snapshot_output("y"))
  snapper$end_file()
  expect_true(file.exists(file.path(path, "snapshot-2.md")))
  expect_true(file.exists(file.path(path, "snapshot-2.new.md")))
})

test_that("defaults to failing on CI", {
  local_on_cran(FALSE)
  withr::local_envvar(CI = "true")

  path <- withr::local_tempdir()
  snapper <- local_snapshotter(snap_dir = path)
  snapper$start_file("snapshot-2")
  # warns on first creation
  snapper$start_file("snapshot-2", "test")
  expect_error(expect_snapshot_output("x"), "Adding new")
})

test_that("only create new files for changed variants", {
  local_on_cran(FALSE)

  snapper <- local_test_snapshotter()
  snapper$start_file("variants", "test")
  expect_warning(expect_snapshot_output("x"), "Adding new")
  expect_warning(expect_snapshot_output("x", variant = "a"), "Adding new")
  expect_warning(expect_snapshot_output("x", variant = "b"), "Adding new")
  snapper$end_file()
  expect_setequal(
    snapper$snap_files(),
    c("variants.md", "a/variants.md", "b/variants.md")
  )

  # failure in default
  snapper$start_file("variants", "test")
  expect_failure(expect_snapshot_output("y"))
  expect_success(expect_snapshot_output("x", variant = "a"))
  expect_success(expect_snapshot_output("x", variant = "b"))
  snapper$end_file()
  expect_setequal(
    snapper$snap_files(),
    c("variants.md", "variants.new.md", "a/variants.md", "b/variants.md")
  )
  unlink(file.path(snapper$snap_dir, "variants.new.md"))

  # failure in variant
  snapper$start_file("variants", "test")
  expect_success(expect_snapshot_output("x"))
  expect_success(expect_snapshot_output("x", variant = "a"))
  expect_failure(expect_snapshot_output("y", variant = "b"))
  snapper$end_file()
  expect_setequal(
    snapper$snap_files(),
    c("variants.md", "a/variants.md", "b/variants.md", "b/variants.new.md")
  )
})

test_that("only reverting change in variant deletes .new", {
  local_on_cran(FALSE)

  snapper <- local_test_snapshotter()
  snapper$start_file("v", "test")
  expect_warning(expect_snapshot_output("x", variant = "a"), "Adding new")
  expect_warning(expect_snapshot_output("x", variant = "b"), "Adding new")
  snapper$end_file()
  expect_setequal(snapper$snap_files(), c("a/v.md", "b/v.md"))

  # failure
  snapper$start_file("v", "test")
  expect_failure(expect_snapshot_output("y", variant = "a"))
  snapper$end_file()
  expect_setequal(snapper$snap_files(), c("a/v.md", "b/v.md", "a/v.new.md"))

  # success
  snapper$start_file("v", "test")
  expect_success(expect_snapshot_output("x", variant = "a"))
  snapper$end_file()
  expect_setequal(snapper$snap_files(), c("a/v.md", "b/v.md"))
})


test_that("removing tests removes snap file", {
  local_on_cran(FALSE)

  path <- withr::local_tempdir()
  snapper <- local_test_snapshotter(snap_dir = path)
  snapper$start_file("snapshot-3", "test")
  expect_warning(expect_snapshot_output("x"), "Adding new")
  snapper$end_file()
  expect_true(file.exists(file.path(path, "snapshot-3.md")))

  snapper$start_file("snapshot-3", "test")
  snapper$end_file()
  expect_false(file.exists(file.path(path, "snapshot-3.md")))
})

test_that("errors in test doesn't change snapshot", {
  local_on_cran(FALSE)
  snapper <- local_test_snapshotter()

  # First run
  snapper$start_file("snapshot-5", "test")
  expect_warning(expect_snapshot_output("x"), "Adding new")
  snapper$end_file()

  # Second run has error
  snapper$start_file("snapshot-5", "test")
  snapper$add_result(NULL, NULL, as.expectation(simpleError("error")))
  snapper$end_file()

  # Third run
  snapper$start_file("snapshot-5", "test")
  expect_warning(expect_snapshot_output("x"), NA)
  snapper$end_file()

  # No warning if snapshot already happened
  snapper$start_file("snapshot-5", "test")
  expect_snapshot_output("x")
  expect_warning(
    snapper$add_result(NULL, NULL, as.expectation(simpleError("error"))),
    NA
  )
  snapper$end_file()
})

test_that("skips and unexpected errors reset snapshots", {
  regenerate <- FALSE

  if (regenerate) {
    withr::local_envvar(c(TESTTHAT_REGENERATE_SNAPS = "true"))
  }

  catch_cnd(
    test_file(
      test_path("test-snapshot", "test-snapshot.R"),
      reporter = NULL
    )
  )

  path <- "test-snapshot/_snaps/snapshot.md"
  snaps <- snap_from_md(brio::read_lines(path))
  titles <- c("errors reset snapshots", "skips reset snapshots")
  expect_true(all(titles %in% names(snaps)))
})

test_that("`expect_error()` can fail inside `expect_snapshot()`", {
  out <- test_file(
    test_path("test-snapshot", "test-expect-condition.R"),
    reporter = NULL
  )
  err <- out[[1]]$results[[1]]
  expect_snapshot(err$message)
})


test_that("can filter with desc", {
  path <- withr::local_tempdir()

  # First record some results
  suppressWarnings({
    snapper <- local_test_snapshotter(snap_dir = path)
    snapper$start_file("snapshot")
    snapper$start_test(test = "x")
    expect_snapshot_output(cat("x"), cran = TRUE)
    snapper$end_test()
    snapper$start_test(test = "y")
    expect_snapshot_output(cat("y"), cran = TRUE)
    snapper$end_test()
    snapper$end_file()
  })
  snaps_all <- readLines(file.path(path, "snapshot.md"))

  # Now pretend we just ran one
  snapper <- local_test_snapshotter(snap_dir = path, desc = "x")
  snapper$start_file("snapshot")
  snapper$start_test(test = "x")
  expect_snapshot_output(cat("x"), cran = TRUE)
  snapper$end_test()
  snapper$end_file()
  snaps_filtered <- readLines(file.path(path, "snapshot.md"))

  expect_equal(snaps_all, snaps_filtered)
})

test_that("expect_snapshot_file works", {
  path <- write_tmp_lines(letters)
  expect_snapshot_file(path, "foo.r", compare = compare_file_text)

  path <- withr::local_tempfile()
  png(path, width = 300, height = 300, type = "cairo")
  plot(1:10, xlab = "", ylab = "", pch = 20, cex = 5, axes = FALSE)
  dev.off()
  expect_snapshot_file(path, "foo.png")

  path <- withr::local_tempfile()
  mtcars2 <- mtcars[1:5, 1:4]
  write.csv(mtcars2, path)
  expect_snapshot_file(path, "foo.csv", compare = compare_file_text)

  # Deprecated `binary` argument still works
  withr::local_options(lifecycle_verbosity = "quiet")
  expect_snapshot_file(path, "foo-not-binary.csv", binary = FALSE)
})


test_that("expect_snapshot_file works in a different directory", {
  path <- withr::local_tempdir()
  withr::local_dir(path)

  brio::write_lines("a", "a.txt", eol = "\r\n")

  # expect no warning
  expect_no_warning(expect_snapshot_file("a.txt"))
})

test_that("expect_snapshot_file works with variant", {
  local_on_cran(FALSE)

  expect_snapshot_file(
    write_tmp_lines(r_version()),
    "version.txt",
    compare = compare_file_text,
    variant = r_version()
  )
})

test_that("expect_snapshot_file finds duplicate snapshot files", {
  local_on_cran(FALSE)

  expect_snapshot(
    expect_snapshot_file(
      write_tmp_lines(r_version()),
      "version.txt",
      variant = r_version()
    ),
    error = TRUE
  )
})

test_that("basic workflow", {
  local_on_cran(FALSE)
  snapper <- local_test_snapshotter()

  path <- write_tmp_lines(letters)
  # warns on first run
  snapper$start_file("snapshot-6", "test")
  expect_warning(expect_snapshot_file(path, "letters.txt"), "Adding new")
  snapper$end_file()

  # succeeds if unchanged
  snapper$start_file("snapshot-6", "test")
  expect_success(expect_snapshot_file(path, "letters.txt"))
  snapper$end_file()

  # fails if changed
  snapper$start_file("snapshot-6", "test")
  path2 <- write_tmp_lines(letters[-1])
  expect_failure(expect_snapshot_file(path2, "letters.txt"), "has changed")
  snapper$end_file()
})

test_that("can announce snapshot file", {
  snapper <- local_test_snapshotter()
  snapper$start_file("snapshot-announce", "test")
  announce_snapshot_file(name = "bar.svg")
  expect_equal(snapper$snap_file_seen, "snapshot-announce/bar.svg")
})

test_that("can transform snapshot contents", {
  path <- withr::local_tempfile(lines = c("secret", "ssh secret squirrel"))

  redact <- function(x) gsub("secret", "<redacted>", x)
  expect_snapshot_file(path, "secret.txt", transform = redact)
})

# snapshot_file_equal -----------------------------------------------------

test_that("warns on first creation", {
  path <- write_tmp_lines("a")
  snap_dir <- withr::local_tempdir()

  snapshot_file_equal_ <- function(path) {
    snapshot_file_equal(
      snap_dir = snap_dir,
      snap_test = "my-test",
      snap_name = "test.txt",
      snap_variant = NULL,
      path = path,
      fail_on_new = FALSE
    )
  }

  # Warns on first run
  expect_snapshot(out <- snapshot_file_equal_(path))
  expect_true(out)

  # Errors on non-existing file
  expect_snapshot(snapshot_file_equal_("doesnt-exist.txt"), error = TRUE)

  # Unchanged returns TRUE
  expect_true(snapshot_file_equal_(path))
  expect_true(file.exists(file.path(snap_dir, "my-test/test.txt")))
  expect_false(file.exists(file.path(snap_dir, "my-test/test.new.txt")))

  # Changed returns FALSE
  path2 <- write_tmp_lines("b")
  expect_false(snapshot_file_equal_(path2))
  expect_true(file.exists(file.path(snap_dir, "my-test/test.txt")))
  expect_true(file.exists(file.path(snap_dir, "my-test/test.new.txt")))

  # Changing again overwrites
  path2 <- write_tmp_lines("c")
  expect_false(snapshot_file_equal_(path2))
  expect_equal(
    brio::read_lines(file.path(snap_dir, "my-test/test.new.txt")),
    "c"
  )

  # Unchanged cleans up
  expect_true(snapshot_file_equal_(path))
  expect_true(file.exists(file.path(snap_dir, "my-test/test.txt")))
  expect_false(file.exists(file.path(snap_dir, "my-test/test.new.txt")))
})

# helpers -----------------------------------------------------------------

test_that("text comparison ignores CR", {
  path1 <- write_tmp_lines(c("a", "b"))
  path2 <- write_tmp_lines(c("a", "b"), eol = "\r\n")

  expect_false(compare_file_binary(path1, path2))
  expect_true(compare_file_text(path1, path2))
})

test_that("split_path handles edge cases", {
  expect_equal(split_path(""), list(dir = "", name = "", ext = ""))

  expect_equal(split_path("a"), list(dir = "", name = "a", ext = ""))
  expect_equal(split_path("a"), list(dir = "", name = "a", ext = ""))
  expect_equal(split_path(".b"), list(dir = "", name = "", ext = "b"))
  expect_equal(split_path(".b.c"), list(dir = "", name = "", ext = "b.c"))

  expect_equal(split_path("x/a"), list(dir = "x", name = "a", ext = ""))
  expect_equal(split_path("x/a"), list(dir = "x", name = "a", ext = ""))
  expect_equal(split_path("x/.b"), list(dir = "x", name = "", ext = "b"))
  expect_equal(split_path("x/.b.c"), list(dir = "x", name = "", ext = "b.c"))
})

test_that("generates informative hint", {
  local_mocked_bindings(in_check_reporter = function() FALSE)
  withr::local_envvar(GITHUB_ACTIONS = "false", TESTTHAT_WD = NA)

  expect_snapshot(snapshot_hint("lala", reset_output = FALSE))
})

test_that("expect_snapshot_file validates its inputs", {
  path <- withr::local_tempfile(lines = "x")

  expect_snapshot(error = TRUE, {
    expect_snapshot_file(123)
    expect_snapshot_file("doesnt-exist.txt")
    expect_snapshot_file(path, 123)
    expect_snapshot_file(path, "test.txt", cran = "yes")
  })
})

#' Do you expect this code to create the same file as last time?
#'
#' @description
#' Whole file snapshot testing is designed for testing objects that don't have
#' a convenient textual representation, with initial support for images
#' (`.png`, `.jpg`, `.svg`), data frames (`.csv`), and text files
#' (`.R`, `.txt`, `.json`, ...).
#'
#' The first time `expect_snapshot_file()` is run, it will create
#' `_snaps/{test}/{name}.{ext}` containing reference output. Future runs will
#' be compared to this reference: if different, the test will fail and the new
#' results will be saved in `_snaps/{test}/{name}.new.{ext}`. To review
#' failures, call [snapshot_review()].
#'
#' We generally expect this function to be used via a wrapper that takes care
#' of ensuring that output is as reproducible as possible, e.g. automatically
#' skipping tests where it's known that images can't be reproduced exactly.
#'
#' @param path Path to file to snapshot. Optional for
#'   `announce_snapshot_file()` if `name` is supplied.
#' @param name Snapshot name, taken from `path` by default.
#' @param binary `r lifecycle::badge("deprecated")` Please use the
#'   `compare` argument instead.
#' @param compare A function used to compare the snapshot files. It should take
#'   two inputs, the paths to the `old` and `new` snapshot, and return either
#'   `TRUE` or `FALSE`. This defaults to `compare_file_text` if `name` has
#'   extension `.r`, `.R`, `.Rmd`, `.md`, or `.txt`, and otherwise uses
#'   `compare_file_binary`.
#'
#'   `compare_file_binary()` compares byte-by-byte and
#'   `compare_file_text()` compares lines-by-line, ignoring
#'   the difference between Windows and Mac/Linux line endings.
#' @param variant If not-`NULL`, results will be saved in
#'   `_snaps/{variant}/{test}/{name}`. This allows you to create
#'   different snapshots for different scenarios, like different operating
#'   systems or different R versions.
#'
#'   Note that there's no way to declare all possible variants up front which
#'   means that as soon as you start using variants, you are responsible for
#'   deleting snapshot variants that are no longer used. (testthat will still
#'   delete all variants if you delete the test.)
#' @inheritParams expect_snapshot
#'
#' @section Announcing snapshots:
#' testthat automatically detects dangling snapshots that have been
#' written to the `_snaps` directory but which no longer have
#' corresponding R code to generate them. These dangling files are
#' automatically deleted so they don't clutter the snapshot
#' directory.
#'
#' This can cause problems if your test is conditionally executed, either
#' because of an `if` statement or a [skip()]. To avoid files being deleted in
#' this case, you can call `announce_snapshot_file()` before the conditional
#' code.
#'
#' ```R
#' test_that("can save a file", {
#'   if (!can_save()) {
#'     announce_snapshot_file(name = "data.txt")
#'     skip("Can't save file")
#'   }
#'   path <- withr::local_tempfile()
#'   expect_snapshot_file(save_file(path, mydata()), "data.txt")
#' })
#' ```
#'
#' @export
#' @examples
#'
#' # To use expect_snapshot_file() you'll typically need to start by writing
#' # a helper function that creates a file from your code, returning a path
#' save_png <- function(code, width = 400, height = 400) {
#'   path <- tempfile(fileext = ".png")
#'   png(path, width = width, height = height)
#'   on.exit(dev.off())
#'   code
#'
#'   path
#' }
#' path <- save_png(plot(1:5))
#' path
#'
#' \dontrun{
#' expect_snapshot_file(save_png(hist(mtcars$mpg)), "plot.png")
#' }
#'
#' # You'd then also provide a helper that skips tests where you can't
#' # be sure of producing exactly the same output.
#' expect_snapshot_plot <- function(name, code) {
#'   # Announce the file before touching skips or running `code`. This way,
#'   # if the skips are active, testthat will not auto-delete the corresponding
#'   # snapshot file.
#'   name <- paste0(name, ".png")
#'   announce_snapshot_file(name = name)
#'
#'   # Other packages might affect results
#'   skip_if_not_installed("ggplot2", "2.0.0")
#'   # Or maybe the output is different on some operating systems
#'   skip_on_os("windows")
#'   # You'll need to carefully think about and experiment with these skips
#'
#'   path <- save_png(code)
#'   expect_snapshot_file(path, name)
#' }
expect_snapshot_file <- function(
  path,
  name = basename(path),
  binary = deprecated(),
  cran = FALSE,
  compare = NULL,
  transform = NULL,
  variant = NULL
) {
  lab <- quo_label(enquo(path))

  check_string(path)
  if (!file.exists(path)) {
    cli::cli_abort("{.path {path}} doesn't exist.")
  }
  check_string(name)
  check_bool(cran)
  check_variant(variant)

  edition_require(3, "expect_snapshot_file()")

  announce_snapshot_file(name = name)
  if (!cran && on_cran()) {
    signal_snapshot_on_cran()
    return(invisible())
  }

  snapshotter <- get_snapshotter()
  if (is.null(snapshotter)) {
    snapshot_not_available(path)
    return(invisible())
  }

  is_text <- is_text_file(name)
  if (!is_missing(binary)) {
    lifecycle::deprecate_warn(
      "3.0.3",
      "expect_snapshot_file(binary = )",
      "expect_snapshot_file(compare = )"
    )
    compare <- if (binary) compare_file_binary else compare_file_text
  }
  if (is.null(compare)) {
    compare <- if (is_text) compare_file_text else compare_file_binary
  }

  if (!is.null(transform)) {
    lines <- brio::read_lines(path)
    lines <- transform(lines)
    brio::write_lines(lines, path)
  }

  equal <- snapshotter$take_file_snapshot(
    name,
    path,
    file_equal = compare,
    variant = variant,
  )
  if (inherits(equal, "expectation_failure")) {
    return(equal)
  }

  file <- snapshotter$file

  if (!equal) {
    if (is_text) {
      base <- paste0(c(snapshotter$snap_dir, variant, file), collapse = "/")
      old_path <- paste0(c(base, name), collapse = "/")
      new_path <- paste0(c(base, new_name(name)), collapse = "/")

      comp <- waldo_compare(
        x = brio::read_lines(old_path),
        x_arg = "old",
        y = brio::read_lines(new_path),
        y_arg = "new",
        quote_strings = FALSE
      )
      comp <- c("Differences:", comp)
    } else {
      comp <- NULL
    }

    hint <- snapshot_hint(paste0(file, "/"), show_accept = is_text)

    msg <- c(
      sprintf("Snapshot of %s has changed.", lab),
      comp,
      hint
    )
    snapshot_fail(msg)
  } else {
    pass()
  }
}

is_text_file <- function(path) {
  ext <- tools::file_ext(path)
  ext %in% c("r", "R", "txt", "md", "Rmd", "qmd", "json")
}

#' @rdname expect_snapshot_file
#' @export
announce_snapshot_file <- function(path, name = basename(path)) {
  edition_require(3, "announce_snapshot_file()")

  snapshotter <- get_snapshotter()
  if (!is.null(snapshotter)) {
    snapshotter$announce_file_snapshot(name)
  }
}

snapshot_file_equal <- function(
  snap_dir, # _snaps/
  snap_test, # test file name
  snap_name, # snapshot file name
  snap_variant, # variant (optional)
  path, # path to new file
  file_equal = compare_file_binary,
  fail_on_new = NULL,
  trace_env = caller_env()
) {
  if (!file.exists(path)) {
    cli::cli_abort("{.path {path}} not found.", call = trace_env)
  }

  if (is.null(snap_variant)) {
    snap_test_dir <- file.path(snap_dir, snap_test)
  } else {
    snap_test_dir <- file.path(snap_dir, snap_variant, snap_test)
  }
  fail_on_new <- fail_on_new %||% on_ci()

  cur_path <- file.path(snap_test_dir, snap_name)
  new_path <- file.path(snap_test_dir, new_name(snap_name))

  if (file.exists(cur_path)) {
    eq <- file_equal(cur_path, path)
    if (!eq) {
      file.copy(path, new_path, overwrite = TRUE)
    } else {
      # in case it exists from a previous run
      unlink(new_path)
    }
    eq
  } else {
    dir.create(snap_test_dir, showWarnings = FALSE, recursive = TRUE)
    file.copy(path, cur_path)

    message <- paste_c(
      "Adding new file snapshot: 'tests/testthat/_snaps/",
      c(snap_variant, if (!is.null(snap_variant)) "/"),
      c(snap_test, "/"),
      snap_name,
      "'"
    )

    # We want to fail on CI since this suggests that the user has failed
    # to record the value locally
    if (fail_on_new) {
      snapshot_fail(message, trace_env = trace_env)
    } else {
      testthat_warn(message)
    }

    TRUE
  }
}

# Helpers -----------------------------------------------------------------

new_name <- function(x) {
  pieces <- split_path(x)
  paste0(
    pieces$dir,
    ifelse(pieces$dir == "", "", "/"),
    pieces$name,
    ".new.",
    pieces$ext
  )
}

split_path <- function(path) {
  dir <- dirname(path)
  dir[dir == "."] <- ""
  name <- basename(path)

  ext_loc <- regexpr(".", name, fixed = TRUE)
  no_ext <- ext_loc == -1L

  name_sans_ext <- ifelse(no_ext, name, substr(name, 1, ext_loc - 1))
  ext <- ifelse(no_ext, "", substr(name, ext_loc + 1, nchar(name)))

  list(
    dir = dir,
    name = name_sans_ext,
    ext = ext
  )
}

write_tmp_lines <- function(
  lines,
  ext = ".txt",
  eol = "\n",
  envir = caller_env()
) {
  path <- withr::local_tempfile(fileext = ext, .local_envir = envir)
  brio::write_lines(lines, path, eol = eol)
  path
}

local_snap_dir <- function(paths, .env = parent.frame()) {
  dir <- withr::local_tempfile(.local_envir = .env)
  withr::defer(unlink(paths), envir = .env)

  dirs <- file.path(dir, unique(dirname(paths)))
  for (d in dirs) {
    dir.create(d, showWarnings = FALSE, recursive = TRUE)
  }
  file.create(file.path(dir, paths))

  dir
}

#' @rdname expect_snapshot_file
#' @param old,new Paths to old and new snapshot files.
#' @export
compare_file_binary <- function(old, new) {
  old <- brio::read_file_raw(old)
  new <- brio::read_file_raw(new)
  identical(old, new)
}
#' @rdname expect_snapshot_file
#' @export
compare_file_text <- function(old, new) {
  old <- brio::read_lines(old)
  new <- brio::read_lines(new)
  identical(old, new)
}

#' Do you expect an S3/S4/R6/S7 object that inherits from this class?
#'
#' @description
#' See <https://adv-r.hadley.nz/oo.html> for an overview of R's OO systems, and
#' the vocabulary used here.
#'
#' * `expect_type(x, type)` checks that `typeof(x)` is `type`.
#' * `expect_s3_class(x, class)` checks that `x` is an S3 object that
#'   [inherits()] from `class`
#' * `expect_s3_class(x, NA)` checks that `x` isn't an S3 object.
#' * `expect_s4_class(x, class)` checks that `x` is an S4 object that
#'   [is()] `class`.
#' * `expect_s4_class(x, NA)` checks that `x` isn't an S4 object.
#' * `expect_r6_class(x, class)` checks that `x` an R6 object that
#'   inherits from `class`.
#' * `expect_s7_class(x, Class)` checks that `x` is an S7 object that
#'   [S7::S7_inherits()] from `Class`
#'
#' See [expect_vector()] for testing properties of objects created by vctrs.
#'
#' @param type String giving base type (as returned by [typeof()]).
#' @param class The required type varies depending on the function:
#'   * `expect_type()`: a string.
#'   * `expect_s3_class()`: a string or character vector. The behaviour of
#'     multiple values (i.e. a character vector) is controlled by the
#'     `exact` argument.
#'   * `expect_s4_class()`: a string.
#'   * `expect_r6_class()`: a string.
#'   * `expect_s7_class()`: an [S7::S7_class()] object.
#'
#'   For historical reasons, `expect_s3_class()` and `expect_s4_class()` also
#'   take `NA` to assert that the `object` is not an S3 or S4 object.
#' @inheritParams expect_that
#' @family expectations
#' @examples
#' x <- data.frame(x = 1:10, y = "x", stringsAsFactors = TRUE)
#' # A data frame is an S3 object with class data.frame
#' expect_s3_class(x, "data.frame")
#' show_failure(expect_s4_class(x, "data.frame"))
#' # A data frame is built from a list:
#' expect_type(x, "list")
#'
#' f <- factor(c("a", "b", "c"))
#' o <- ordered(f)
#'
#' # Using multiple class names tests if the object inherits from any of them
#' expect_s3_class(f, c("ordered", "factor"))
#' # Use exact = TRUE to test for exact match
#' show_failure(expect_s3_class(f, c("ordered", "factor"), exact = TRUE))
#' expect_s3_class(o, c("ordered", "factor"), exact = TRUE)
#'
#' # An integer vector is an atomic vector of type "integer"
#' expect_type(x$x, "integer")
#' # It is not an S3 object
#' show_failure(expect_s3_class(x$x, "integer"))
#'
#' # Above, we requested data.frame() converts strings to factors:
#' show_failure(expect_type(x$y, "character"))
#' expect_s3_class(x$y, "factor")
#' expect_type(x$y, "integer")
#' @name inheritance-expectations
NULL

#' @export
#' @rdname inheritance-expectations
expect_type <- function(object, type) {
  check_string(type)

  act <- quasi_label(enquo(object))
  act_type <- typeof(act$val)

  if (!identical(act_type, type)) {
    fail(c(
      sprintf("Expected %s to have type %s.", act$lab, format_class(type)),
      sprintf("Actual type: %s", format_class(act_type))
    ))
  } else {
    pass()
  }
  invisible(act$val)
}

#' @export
#' @rdname inheritance-expectations
#' @param exact If `FALSE`, the default, checks that `object` inherits
#'   from any element of `class`. If `TRUE`, checks that object has a class
#'   that exactly matches `class`.
expect_s3_class <- function(object, class, exact = FALSE) {
  check_bool(exact)

  act <- quasi_label(enquo(object))
  act$class <- format_class(class(act$val))
  exp_lab <- format_class(class)

  if (identical(class, NA)) {
    if (isS3(object)) {
      fail(sprintf("Expected %s not to be an S3 object.", act$lab))
    } else {
      pass()
    }
  } else if (is.character(class)) {
    if (!isS3(act$val)) {
      fail(c(
        sprintf("Expected %s to be an S3 object.", act$lab),
        sprintf("Actual OO type: %s.", oo_type(act$val))
      ))
    } else if (exact && !identical(class(act$val), class)) {
      fail(c(
        sprintf("Expected %s to have class %s.", act$lab, exp_lab),
        sprintf("Actual class: %s.", act$class)
      ))
    } else if (!inherits(act$val, class)) {
      fail(c(
        sprintf("Expected %s to inherit from %s.", act$lab, exp_lab),
        sprintf("Actual class: %s.", act$class)
      ))
    } else {
      pass()
    }
  } else {
    stop_input_type(class, c("a character vector", "NA"))
  }

  invisible(act$val)
}

#' @export
#' @rdname inheritance-expectations
expect_s4_class <- function(object, class) {
  act <- quasi_label(enquo(object))
  act$class <- format_class(methods::is(act$val))
  exp_lab <- format_class(class)

  if (identical(class, NA)) {
    if (isS4(object)) {
      fail(sprintf("Expected %s not to be an S4 object.", act$lab))
    } else {
      pass()
    }
  } else if (is.character(class)) {
    if (!isS4(act$val)) {
      fail(c(
        sprintf("Expected %s to be an S4 object.", act$lab),
        sprintf("Actual OO type: %s.", oo_type(act$val))
      ))
    } else if (!methods::is(act$val, class)) {
      fail(c(
        sprintf("Expected %s to inherit from %s.", act$lab, exp_lab),
        sprintf("Actual class: %s.", act$class)
      ))
    } else {
      pass()
    }
  } else {
    stop_input_type(class, c("a character vector", "NA"))
  }

  invisible(act$val)
}

#' @export
#' @rdname inheritance-expectations
expect_r6_class <- function(object, class) {
  act <- quasi_label(enquo(object))
  check_string(class)

  if (!inherits(act$val, "R6")) {
    fail(c(
      sprintf("Expected %s to be an R6 object.", act$lab),
      sprintf("Actual OO type: %s.", oo_type(act$val))
    ))
  } else if (!inherits(act$val, class)) {
    act_class <- format_class(class(act$val))
    exp_class <- format_class(class)
    fail(c(
      sprintf("Expected %s to inherit from %s.", act$lab, exp_class),
      sprintf("Actual class: %s.", act_class)
    ))
  } else {
    pass()
  }

  invisible(act$val)
}

#' @export
#' @rdname inheritance-expectations
expect_s7_class <- function(object, class) {
  check_installed("S7")
  if (!inherits(class, "S7_class")) {
    stop_input_type(class, "an S7 class object")
  }

  act <- quasi_label(enquo(object))

  if (!S7::S7_inherits(object)) {
    fail(c(
      sprintf("Expected %s to be an S7 object.", act$lab),
      sprintf("Actual OO type: %s.", oo_type(act$val))
    ))
  } else if (!S7::S7_inherits(object, class)) {
    exp_class <- attr(class, "name", TRUE)
    act_class <- setdiff(base::class(object), "S7_object")
    act_class_desc <- paste0("<", act_class, ">", collapse = "/")

    fail(c(
      sprintf("Expected %s to inherit from <%s>.", act$lab, exp_class),
      sprintf("Actual class: %s.", act_class_desc)
    ))
  } else {
    pass()
  }

  invisible(act$val)
}

#' Do you expect to inherit from this class?
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `expect_is()` is an older form that uses [inherits()] without checking
#' whether `x` is S3, S4, or neither. Instead, I'd recommend using
#' [expect_type()], [expect_s3_class()], or [expect_s4_class()] to more clearly
#' convey your intent.
#'
#' @section 3rd edition:
#' `r lifecycle::badge("deprecated")`
#'
#' `expect_is()` is formally deprecated in the 3rd edition.
#'
#' @keywords internal
#' @param class Class name passed to `inherits()`.
#' @inheritParams expect_type
#' @export
expect_is <- function(object, class, info = NULL, label = NULL) {
  check_character(class)
  edition_deprecate(
    3,
    "expect_is()",
    "Use `expect_type()`, `expect_s3_class()`, or `expect_s4_class()` instead"
  )

  act <- quasi_label(enquo(object), label)
  act$class <- format_class(class(act$val))
  exp_lab <- format_class(class(class))

  if (!inherits(act$val, class)) {
    msg <- sprintf(
      "Expected %s to inherit from %s.\nActual inheritance: %s",
      act$lab,
      exp_lab,
      act$class
    )
    fail(msg, info = info)
  } else {
    pass()
  }
  invisible(act$val)
}

# Helpers ----------------------------------------------------------------------

isS3 <- function(x) is.object(x) && !isS4(x)

format_class <- function(x) {
  paste0(encodeString(x, quote = '"'), collapse = "/")
}

oo_type <- function(x) {
  if (!is.object(x)) {
    return("none")
  }
  if (isS4(x)) {
    "S4"
  } else {
    if (inherits(x, "R6")) {
      "R6"
    } else if (inherits(x, "S7")) {
      "S7"
    } else {
      "S3"
    }
  }
}

test_that("get_routine() finds own 'run_testthat_tests'", {
  routine <- get_routine("testthat", "run_testthat_tests")
  expect_s3_class(routine, "NativeSymbolInfo")
})

test_that("get_routine() fails when no routine exists", {
  expect_error(get_routine("utils", "no_such_routine"))
})

run_cpp_tests("testthat")

test_that("expect_type checks typeof", {
  expect_success(expect_type(factor("a"), "integer"))
  expect_failure(expect_type(factor("a"), "double"))
})

test_that("expect_is checks class", {
  local_edition(2)

  expect_success(expect_is(factor("a"), "factor"))
  expect_failure(expect_is(factor("a"), "integer"))
})

test_that("expect_s3/s4_class fails if appropriate type", {
  A <- methods::setClass("A", contains = "list")

  expect_failure(expect_s3_class(1, "double"), "not an S3 object")
  expect_failure(expect_s3_class(A(), "double"), "not an S3 object")
  expect_failure(expect_s4_class(factor(), "double"), "not an S4 object")
})

test_that("expect_s[34]_class can check not S3/S4", {
  expect_success(expect_s3_class(1, NA))
  expect_snapshot_failure(expect_s3_class(factor(), NA))

  A <- methods::setClass("A", contains = "list")
  expect_success(expect_s4_class(1, NA))
  expect_snapshot_failure(expect_s4_class(A(), NA))
})

test_that("test_s4_class respects class hierarchy", {
  A <- methods::setClass("A", contains = "list")
  B <- methods::setClass("B", contains = "list")
  C <- methods::setClass("C", contains = c("A", "B"))
  on.exit({
    methods::removeClass("A")
    methods::removeClass("B")
    methods::removeClass("C")
  })

  expect_success(expect_s4_class(C(), "A"))
  expect_success(expect_s4_class(C(), "B"))
  expect_snapshot_failure(expect_s4_class(C(), "D"))
})

test_that("test_s3_class respects class hierarchy", {
  x <- structure(list(), class = c("a", "b"))
  expect_success(expect_s3_class(x, "a"))
  expect_success(expect_s3_class(x, "b"))
  expect_snapshot_failure(expect_s3_class(x, "c"))
  expect_snapshot_failure(expect_s3_class(x, c("c", "d")))
})

test_that("test_s3_class can request exact match", {
  x <- structure(list(), class = c("a", "b"))
  expect_failure(expect_s3_class(x, "a", exact = TRUE))
  expect_success(expect_s3_class(x, c("a", "b"), exact = TRUE))
})

test_that("expect_s3_class allows unquoting of first argument", {
  f <- factor("a")
  expect_success(expect_s3_class(!! rlang::quo(f), "factor"))
})


# expect_s7_class --------------------------------------------------------

test_that("checks its inputs", {
  expect_snapshot(expect_s7_class(1, 1), error = TRUE)
})

test_that("can check with actual class", {
  Foo <- S7::new_class("Foo", package = NULL)
  Bar <- S7::new_class("Bar", package = NULL)
  expect_success(expect_s7_class(Foo(), class = Foo))
  expect_snapshot_failure(expect_s7_class(Foo(), class = Bar))

  Baz <- S7::new_class("Baz", parent = Foo, package = NULL)
  expect_snapshot_failure(expect_s7_class(Baz(), class = Bar))
})

#' Do you expect code to execute silently?
#'
#' Checks that the code produces no output, messages, or warnings.
#'
#' @inheritParams expect_error
#' @return The first argument, invisibly.
#' @family expectations
#' @export
#' @examples
#' expect_silent("123")
#'
#' f <- function() {
#'   message("Hi!")
#'   warning("Hey!!")
#'   print("OY!!!")
#' }
#' \dontrun{
#' expect_silent(f())
#' }
expect_silent <- function(object) {
  act <- quasi_capture(enquo(object), NULL, evaluate_promise)

  outputs <- c(
    if (!identical(act$cap$output, "")) "output",
    if (length(act$cap$warnings) > 0) "warnings",
    if (length(act$cap$messages) > 0) "messages"
  )

  if (length(outputs) != 0) {
    fail(c(
      sprintf("Expected %s to run silently.", act$lab),
      sprintf("Actual noise: %s.", paste(outputs, collapse = ", "))
    ))
  } else {
    pass()
  }
  invisible(act$cap$result)
}

#' Mock functions in a package.
#'
#' @description
#' `r lifecycle::badge("defunct")`
#'
#' `with_mock()` and `local_mock()` are now defunct, and can be replaced by
#' [with_mocked_bindings()] and [local_mocked_bindings()]. These functions only
#' worked by abusing of R's internals.
#'
#' @param ... named parameters redefine mocked functions, unnamed parameters
#'   will be evaluated after mocking the functions
#' @param .env the environment in which to patch the functions,
#'   defaults to the top-level environment.  A character is interpreted as
#'   package name.
#' @param .local_envir Environment in which to add exit handler.
#'   For expert use only.
#' @keywords internal
#' @export
with_mock <- function(..., .env = topenv()) {
  lifecycle::deprecate_stop("3.2.0", "with_mock()", "with_mocked_bindings()")
}

#' @export
#' @rdname with_mock
local_mock <- function(..., .env = topenv(), .local_envir = parent.frame()) {
  lifecycle::deprecate_stop("3.2.0", "local_mock()", "local_mocked_bindings()")
}

library(testthat)

local({
  if (!requireNamespace("usethis", quietly = TRUE)) {
    return()
  }

  quietly <- function(expr) {
    suppressMessages(capture_output(result <- expr))
    result
  }

  perform_test <- function(pkgName, catchEnabled) {
    owd <- setwd(tempdir())
    withr::defer(setwd(owd))

    pkgPath <- file.path(tempdir(), pkgName)
    libPath <- file.path(tempdir(), "rlib")
    if (!utils::file_test("-d", libPath)) {
      dir.create(libPath)
    }
    .libPaths(c(libPath, .libPaths()))

    withr::defer({
      unlink(pkgPath, recursive = TRUE)
      unlink(libPath, recursive = TRUE)
    })

    quietly(usethis::create_package(pkgPath, open = FALSE))
    quietly(testthat::use_catch(pkgPath))

    cat(
      "LinkingTo: testthat",
      file = file.path(pkgPath, "DESCRIPTION"),
      append = TRUE,
      sep = "\n"
    )

    cat(
      sprintf("useDynLib(%s, .registration=TRUE)", pkgName),
      file = file.path(pkgPath, "NAMESPACE"),
      append = TRUE,
      sep = "\n"
    )

    if (!catchEnabled) {
      isWindows <- Sys.info()[["sysname"]] == "Windows"
      makevarsPath <- file.path(
        pkgPath,
        "src",
        if (isWindows) "Makevars.win" else "Makevars"
      )

      cat(
        "PKG_CPPFLAGS = -DTESTTHAT_DISABLED",
        file = makevarsPath,
        sep = "\n"
      )
    }

    install.packages(pkgs = pkgPath, repos = NULL, type = "source")
    library(pkgName, character.only = TRUE)
    stopifnot(.Call("run_testthat_tests", FALSE, PACKAGE = pkgName))
    pkgload::unload(pkgName)
  }

  withr::local_envvar(R_TESTS = '')
  perform_test("testthatclient1", TRUE)
  perform_test("testthatclient2", FALSE)
})

test_that("one plus one is two", {
  expect_equal(1 + 1, 2)
})

test_that("you can skip tests if needed", {
  skip("This test hasn't been written yet")
})

test_that("some tests have warnings", {
  expect_equal(log(-1), NaN)
})

test_that("some more successes just to pad things out", {
  expect_true(TRUE)
  expect_false(FALSE)
})

test_that("errors reset snapshots", {
  if (nzchar(Sys.getenv("TESTTHAT_REGENERATE_SNAPS"))) {
    expect_snapshot(print(1))
  } else {
    expect_snapshot(stop("failing"))
  }
})

test_that("skips reset snapshots", {
  if (nzchar(Sys.getenv("TESTTHAT_REGENERATE_SNAPS"))) {
    expect_snapshot(print(1))
  } else {
    expect_snapshot(skip("skipping"))
  }
})

test_that("helpers are not on the path", {
  expect_false(
    "my_helper" %in% names(rlang::pkg_env("testConfigLoadAll"))
  )
})

test_that("internal functions are not on the path", {
  fns <- names(rlang::pkg_env("testConfigLoadAll"))
  expect_false("internal" %in% fns)
  expect_true("exported" %in% fns)
})

# the objective is to test what happens if some code fails outside of tests
# i.e. not inside a test_that() call.

test_that("before", expect_true(TRUE))

stop('dying outside of tests')

test_that("after", expect_true(TRUE))

#' An R package to make testing fun!
#'
#' Try the example below. Have a look at the references and learn more
#' from function documentation such as [test_that()].
#'
#' @section Options:
#' - `testthat.use_colours`: Should the output be coloured? (Default: `TRUE`).
#' - `testthat.summary.max_reports`: The maximum number of detailed test
#'    reports printed for the summary reporter (default: 10).
#' - `testthat.summary.omit_dots`: Omit progress dots in the summary reporter
#'    (default: `FALSE`).
#'
#' @keywords internal
"_PACKAGE"

#' @import rlang
#' @importFrom brio writeLines readLines
#' @useDynLib testthat, .registration = TRUE
NULL

the <- new.env(parent = emptyenv())


# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

test_that("logical tests act as expected", {
  expect_true(TRUE)
  expect_false(FALSE)
})

test_that("logical tests ignore attributes", {
  expect_true(c(a = TRUE))
  expect_false(c(a = FALSE))
})

test_that("equality holds", {
  expect_equal(5, 5)
  expect_identical(10, 10)
})

test_that("can't access variables from other tests 2", {
  a <- 10
})

test_that("can't access variables from other tests 1", {
  expect_false(exists("a"))
})

#' Make an equality test.
#'
#' This a convenience function to make a expectation that checks that
#' input stays the same.
#'
#' @param x a vector of values
#' @param expectation the type of equality you want to test for
#'   (`"equals"`, `"is_equivalent_to"`, `"is_identical_to"`)
#' @export
#' @keywords internal
#' @examples
#' x <- 1:10
#' make_expectation(x)
#'
#' make_expectation(mtcars$mpg)
#'
#' df <- data.frame(x = 2)
#' make_expectation(df)
make_expectation <- function(x, expectation = "equals") {
  obj <- substitute(x)
  expectation <- match.arg(
    expectation,
    c("equals", "is_equivalent_to", "is_identical_to")
  )

  dput(substitute(
    expect_equal(obj, values),
    list(obj = obj, expectation = as.name(expectation), values = x)
  ))
}

succeed_after <- function(i) {
  function() {
    i <<- i - 1
    expect_equal(i, 0)
  }
}

test_that("tries multiple times", {
  third_try <- succeed_after(3)
  expect_snapshot(result <- try_again(3, third_try()))
  expect_equal(result, 0)

  third_try <- succeed_after(3)
  expect_snapshot(try_again(1, third_try()), error = TRUE)
})

test_that("handles errors", {
  fails_twice <- local({
    i <- 0
    function() {
      i <<- i + 1
      if (i <= 2) stop("fail") else 1
    }
  })

  expect_snapshot(result <- try_again(3, expect_equal(fails_twice(), 1)))
  expect_equal(result, 1)
})

test_that("source_dir()", {
  res <- source_dir("test_dir", pattern = "hello", chdir = TRUE, wrap = FALSE)
  expect_equal(res[[1]](), "Hello World")

  res <- source_dir(normalizePath("test_dir"), pattern = "hello", chdir = TRUE, wrap = FALSE)
  expect_equal(res[[1]](), "Hello World")

  res <- source_dir("test_dir", pattern = "hello", chdir = FALSE, wrap = FALSE)
  expect_equal(res[[1]](), "Hello World")

  res <- source_dir(normalizePath("test_dir"), pattern = "hello", chdir = FALSE, wrap = FALSE)
  expect_equal(res[[1]](), "Hello World")
})

test_that("simple", {
  stop("argh")
})

test_that("after one success", {
  expect_true(TRUE)
  stop("argh")
  expect_true(TRUE)
})

test_that("after one failure", {
  expect_true(FALSE)
  stop("argh")
})

test_that("in the test", {
  expect_true(stop("Argh"))
})

test_that("in expect_error", {
  expect_error(stop("Argh"))
})

test_that("SummaryReport gives up if too many errors", {
  expect_equal(Inf, 1)
  expect_equal(Inf, 2)
  expect_equal(Inf, 3)
  expect_equal(Inf, 4)
  expect_equal(Inf, 5)
  expect_equal(Inf, 6)
  expect_equal(Inf, 7)
  expect_equal(Inf, 8)
  expect_equal(Inf, 9)
  expect_equal(Inf, 10)
})

find_expectation_srcref <- function(test_code_frame = NULL, top = caller_env()) {
  # It's not possible to give useful srcrefs interactively so don't even try
  path <- getOption("testthat_path")
  if (is.null(path)) {
    return(NULL)
  }

  # Scope our search to the current file loaded with source_file()
  file_srcref <- srcref(srcfile(path), c(1, 1, 1e5, 1e5))

  # Now attempt to narrow the scope to a call that leads to test_code(). That's
  # usually test_that() but might be describe(), it(), or another wrapper.
  testthat_srcref <- find_srcref(
    top = test_code_frame,
    container = file_srcref
  )

  # Now we can find the bottom-most call with a srcref that's inside that scope
  call_srcref <- find_srcref(
    top = top,
    bottom = test_code_frame,
    container = testthat_srcref %||% file_srcref
  )

  # If we can't find that we fall back to the test
  call_srcref %||% testthat_srcref
}

find_srcref <- function(bottom = NULL,
                        top = caller_env(),
                        container = NULL) {

  idx <- sys_index(bottom, top)
  calls <- sys.calls()[rev(idx)]

  for (call in calls) {
    srcref <- attr(call, "srcref")

    if (!is.null(srcref)) {
      if (is.null(container) || srcref_inside(srcref, container)) {
        return(srcref)
      }
    }
  }

  NULL
}

srcref_inside <- function(needle, haystack) {
  stopifnot(inherits(needle, "srcref"), inherits(haystack, "srcref"))

  needle_file <- attr(needle, "srcfile")$filename
  haystack_file <- attr(haystack, "srcfile")$filename

  if (!identical(needle_file, haystack_file)) {
    return(FALSE)
  }

  sign_pair <- function(x, y) {
    diff <- y - x
    if (diff[1] == 0) sign(diff[2]) else sign(diff[1])
  }

  sign_pair(needle[1:2], haystack[1:2]) <= 0 &&
    sign_pair(needle[3:4], haystack[3:4]) >= 0
}

sys_index <- function(bottom = NULL, top = caller_env()) {
  frames <- sys.frames()
  if (is.null(bottom)) {
    bottom_idx <- 1
  } else {
    bottom_idx <- Position(function(env) identical(bottom, env), frames)
    if (is.na(bottom_idx)) {
      abort("Can't find `bottom` on stack")
    }
  }

  top_idx <- Position(function(env) identical(top, env), frames)
  if (is.na(top_idx)) {
    abort("Can't find `top` on stack")
  }

  seq2(bottom_idx, top_idx)
}

test_that("test1", expect_true(TRUE))
test_that("test2", expect_true(TRUE))

test_that("test-pass", expect_true(TRUE))

test_that("test-fail", expect_true(FALSE))

test_that("test-error", {
  stop('argh')
  expect_true(TRUE)
})

#' Evaluate an expectation multiple times until it succeeds
#'
#' If you have a flaky test, you can use `try_again()` to run it a few times
#' until it succeeds. In most cases, you are better fixing the underlying
#' cause of the flakeyness, but sometimes that's not possible.
#'
#' @param times Number of times to retry.
#' @param code Code to evaluate.
#' @export
#' @examples
#' usually_return_1 <- function(i) {
#'   if (runif(1) < 0.1) 0 else 1
#' }
#'
#' \dontrun{
#' # 10% chance of failure:
#' expect_equal(usually_return_1(), 1)
#'
#' # 1% chance of failure:
#' try_again(1, expect_equal(usually_return_1(), 1))
#'
#' # 0.1% chance of failure:
#' try_again(2, expect_equal(usually_return_1(), 1))
#' }
try_again <- function(times, code) {
  check_number_whole(times, min = 1)

  code <- enquo(code)

  i <- 1
  while (i <= times) {
    tryCatch(
      return(eval(get_expr(code), get_env(code))),
      expectation_failure = function(cnd) {
        cli::cli_inform(c(i = "Expectation failed; trying again ({i})..."))
        NULL
      },
      error = function(cnd) {
        cli::cli_inform(c(i = "Expectation errored; trying again ({i})..."))
        NULL
      }
    )

    i <- i + 1
  }

  eval(get_expr(code), get_env(code))
}

expect_equal(2, 2)
expect_true(TRUE)
expect_error(stop("!"))

stopifnot(
  tryCatch(
    expect_true(TRUE),
    expectation_failure = function(e) FALSE
  )
)
stopifnot(
  tryCatch(
    expect_true(FALSE),
    expectation_failure = function(e) TRUE
  )
)

describe("succeeded", {
  it("succeeded-1", expect_true(TRUE))
  it("succeeded-2", expect_true(TRUE))
})

describe("failed then succeeded", {
  it("failed-1", expect_true(FALSE))
  it("failed-2", expect_true(FALSE))
  it("succeeded", expect_true(TRUE))
})

test_that("just one failure", {
  expect_true(FALSE)
})

test_that("one failure on two", {
  expect_false(FALSE)
  expect_true(FALSE)
})

test_that("no failure", {
  expect_false(FALSE)
  expect_true(TRUE)
})

#' describe: a BDD testing language
#'
#' A simple [behavior-driven development
#' (BDD)](https://en.wikipedia.org/wiki/Behavior-driven_development)
#' [domain-specific language](https://en.wikipedia.org/wiki/Domain-specific_language)
#' for writing tests. The language is similar to [RSpec](https://rspec.info/)
#' for Ruby or [Mocha](https://mochajs.org/) for JavaScript. BDD tests read
#' like sentences and it should thus be easier to understand what the
#' specification of a function/component is.
#'
#' Tests using the `describe` syntax not only verify the tested code, but
#' also document its intended behaviour. Each `describe` block specifies a
#' larger component or function and contains a set of specifications. A
#' specification is defined by an `it` block. Each `it` block
#' functions as a test and is evaluated in its own environment. You
#' can also have nested `describe` blocks.
#'
#' This test syntax helps to test the intended behaviour of your code. For
#' example: you want to write a new function for your package. Try to describe
#' the specification first using `describe`, before your write any code.
#' After that, you start to implement the tests for each specification (i.e.
#' the `it` block).
#'
#' Use `describe` to verify that you implement the right things and use
#' [test_that()] to ensure you do the things right.
#'
#' @param description description of the feature
#' @param code test code containing the specs
#' @keywords internal
#' @export
#' @examples
#' describe("matrix()", {
#'   it("can be multiplied by a scalar", {
#'     m1 <- matrix(1:4, 2, 2)
#'     m2 <- m1 * 2
#'     expect_equal(matrix(1:4 * 2, 2, 2), m2)
#'   })
#'   it("can have not yet tested specs")
#' })
#'
#' # Nested specs:
#' ## code
#' addition <- function(a, b) a + b
#' division <- function(a, b) a / b
#'
#' ## specs
#' describe("math library", {
#'   describe("addition()", {
#'     it("can add two numbers", {
#'       expect_equal(1 + 1, addition(1, 1))
#'     })
#'   })
#'   describe("division()", {
#'     it("can divide two numbers", {
#'       expect_equal(10 / 2, division(10, 2))
#'     })
#'     it("can handle division by 0") #not yet implemented
#'   })
#' })
describe <- function(description, code) {
  local_description_push(description)

  code <- substitute(code)
  test_code(code, parent.frame())
}

#' @export
#' @rdname describe
it <- function(description, code = NULL) {
  local_description_push(description)

  code <- substitute(code)
  test_code(code, parent.frame())
}

---
title: "testthat 3e"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{testthat 3e}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

testthat 3.0.0 introduces the idea of an "edition" of testthat.
An edition is a bundle of behaviours that you have to explicitly choose to use, allowing us to make otherwise backward incompatible changes.
This is particularly important for testthat since it has a very large number of packages that use it (almost 5,000 at last count).
Choosing to use the 3rd edition allows you to use our latest recommendations for ongoing and new work, while historical packages continue to use the old behaviour.

(We don't anticipate creating new editions very often, and they'll always be matched with major version, i.e. if there's another edition, it'll be the fourth edition and will come with testthat 4.0.0.)

This vignette shows you how to activate the 3rd edition, introduces the main features, and discusses common challenges when upgrading a package.
If you have a problem that this vignette doesn't cover, please let me know, as it's likely that the problem also affects others.

```{r, message = FALSE}
library(testthat)
local_edition(3)
```

## Activating

The usual way to activate the 3rd edition is to add a line to your `DESCRIPTION`:

    Config/testthat/edition: 3

This will activate the 3rd edition for every test in your package.

You can also control the edition used for individual tests with `testthat::local_edition()`:

```{r}
test_that("I can use the 3rd edition", {
  local_edition(3)
  expect_true(TRUE)
})
```

This is also useful if you've switched to the 3rd edition and have a couple of tests that fail.
You can use `local_edition(2)` to revert back to the old behaviour, giving you some breathing room to figure out the underlying issue.

```{r}
test_that("I want to use the 2nd edition", {
  local_edition(2)
  expect_true(TRUE)
})
```

## Changes

There are three major changes in the 3rd edition:

-   A number of outdated functions are now **deprecated**, so you'll be warned about them every time you run your tests (but they won't cause `R CMD check` to fail).

-   testthat no longer silently swallows **messages**; you now need to deliberately handle them.

-   `expect_equal()` and `expect_identical()` now use the [**waldo**](https://waldo.r-lib.org/) package instead of `identical()` and `all.equal()`.
    This makes them more consistent and provides an enhanced display of differences when a test fails.

### Deprecations

A number of outdated functions have been deprecated.
Most of these functions have not been recommended for a number of years, but before the introduction of the edition idea, I didn't have a good way of preventing people from using them without breaking a lot of code on CRAN.

-   `context()` is formally deprecated.
    testthat has been moving away from `context()` in favour of file names for quite some time, and now you'll be strongly encouraged remove these calls from your tests.

-   `expect_is()` is deprecated in favour of the more specific `expect_type()`, `expect_s3_class()`, and `expect_s4_class()`.
    This ensures that you check the expected class along with the expected OO system.

-   The very old `expect_that()` syntax is now deprecated.
    This was an overly clever API that I regretted even before the release of testthat 1.0.0.

-   `expect_equivalent()` has been deprecated since it is now equivalent (HA HA) to `expect_equal(ignore_attr = TRUE)`.

-   `setup()` and `teardown()` are deprecated in favour of test fixtures.
    See `vignette("test-fixtures")` for details.

-   `expect_known_output()`, `expect_known_value()`, `expect_known_hash()`, and `expect_equal_to_reference()` are all deprecated in favour of `expect_snapshot_output()` and `expect_snapshot_value()`.

-   `with_mock()` and `local_mock()` are deprecated; please use 
`with_mocked_bindings()` or `local_mocked_bindings()` instead.

Fixing these deprecation warnings should be straightforward.

### Warnings

In the second edition, `expect_warning()` swallows all warnings regardless of whether or not they match the `regexp` or `class`:

```{r}
f <- function() {
  warning("First warning")
  warning("Second warning")
  warning("Third warning")
}

local_edition(2)
expect_warning(f(), "First")
```

In the third edition, `expect_warning()` captures at most one warning so the others will bubble up:

```{r}
local_edition(3)
expect_warning(f(), "First")
```

You can either add additional expectations to catch these warnings, or silence them all with `suppressWarnings()`:

```{r}
f() |> 
  expect_warning("First") |> 
  expect_warning("Second") |> 
  expect_warning("Third")

f() |> 
  expect_warning("First") |> 
  suppressWarnings()
```

Alternatively, you might want to capture them all in a snapshot test:

```{r}
test_that("f() produces expected outputs/messages/warnings", {
  expect_snapshot(f())  
})
```

The same principle also applies to `expect_message()`, but message handling has changed in a more radical way, as described next.

### Messages

For reasons that I can no longer remember, testthat silently ignores all messages.
This is inconsistent with other types of output, so as of the 3rd edition, they now bubble up to your test results.
You'll have to explicit ignore them with `suppressMessages()`, or if they're important, test for their presence with `expect_message()`.

### waldo

Probably the biggest day-to-day difference (and the biggest reason to upgrade!) is the use of [`waldo::compare()`](https://waldo.r-lib.org/reference/compare.html) inside of `expect_equal()` and `expect_identical()`.
The goal of waldo is to find and concisely describe the difference between a pair of R objects, and it's designed specifically to help you figure out what's gone wrong in your unit tests.

```{r, error = TRUE}
f1 <- factor(letters[1:3])
f2 <- ordered(letters[1:3], levels = letters[1:4])

local_edition(2)
expect_equal(f1, f2)

local_edition(3)
expect_equal(f1, f2)
```

waldo looks even better in your console because it carefully uses colours to help highlight the differences.

The use of waldo also makes precise the difference between `expect_equal()` and `expect_identical()`: `expect_equal()` sets `tolerance` so that waldo will ignore small numerical differences arising from floating point computation.
Otherwise the functions are identical (HA HA).

This change is likely to result in the most work during an upgrade, because waldo can give slightly different results to both `identical()` and `all.equal()` in moderately common situations.
I believe on the whole the differences are meaningful and useful, so you'll need to handle them by tweaking your tests.
The following changes are most likely to affect you:

-   `expect_equal()` previously ignored the environments of formulas and functions.
    This is most like to arise if you are testing models.
    It's worth thinking about what the correct values should be, but if that is to annoying you can opt out of the comparison with `ignore_function_env` or `ignore_formula_env`.

-   `expect_equal()` used a combination of `all.equal()` and a home-grown `testthat::compare()` which unfortunately used a slightly different definition of tolerance.
    Now `expect_equal()` always uses the same definition of tolerance everywhere, which may require tweaks to your existing tolerance values.

-   `expect_equal()` previously ignored timezone differences when one object had the current timezone set implicitly (with `""`) and the other had it set explicitly:

    ```{r, error = TRUE}
    dt1 <- dt2 <- ISOdatetime(2020, 1, 2, 3, 4, 0)
    attr(dt1, "tzone") <- ""
    attr(dt2, "tzone") <- Sys.timezone()

    local_edition(2)
    expect_equal(dt1, dt2)

    local_edition(3)
    expect_equal(dt1, dt2)
    ```

### Reproducible output

In the third edition, `test_that()` automatically calls `local_reproducible_output()` which automatically sets a number of options and environment variables to ensure output is as reproducible across systems.
This includes setting:

-   `options(crayon.enabled = FALSE)` and `options(cli.unicode = FALSE)` so that the crayon and cli packages produce raw ASCII output.

-   `Sys.setLocale("LC_COLLATE" = "C")` so that sorting a character vector returns the same order regardless of the system language.

-   `options(width = 80)` so print methods always generate the same output regardless of your actual console width.

See the documentation for more details.

## Upgrading

The changes lend themselves to the following workflow for upgrading from the 2nd to the 3rd edition:

1. Activate edition 3. You can let [`usethis::use_testthat(3)`](https://usethis.r-lib.org/reference/use_testthat.html) do this for you.
2. Remove or replace deprecated functions, going over the list of above.
3. If your output got noisy, quiet things down by either capturing or suppressing warnings and messages.
4. Inspect test outputs if objects are not "all equal" anymore.

## Alternatives

You might wonder why we came up with the idea of an "edition", rather than creating a new package like testthat3.
We decided against making a new package because the 2nd and 3rd edition share a very large amount of code, so making a new package would have substantially increased the maintenance burden: the majority of bugs would've needed to be fixed in two places.

If you're a programmer in other languages, you might wonder why we can't rely on [semantic versioning](https://semver.org).
The main reason is that CRAN checks all packages that use testthat with the latest version of testthat, so simply incrementing the major version number doesn't actually help with reducing R CMD check failures on CRAN.

---
title: "Running tests in parallel"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Running tests in parallel}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options:
  markdown:
    wrap: sentence
---

```{r setup, include = FALSE}
library(testthat)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

## Setup

To enable parallel testing, you must first be using the 3rd edition[^1].
Then add the following line to the `DESCRIPTION`:

[^1]: See `vignette("third-edition")` for details.

```         
Config/testthat/parallel: true
```

If needed (for example, for debugging) you can temporarily suppress parallel testing with `Sys.setenv(TESTTHAT_PARALLEL = "false")`.

By default, testthat will use `getOption("Ncpus", 2)` cores.
To increase that value for your development machine we recommend setting `TESTTHAT_CPUS` in your `.Renviron`.
The easiest way to do that is call `usethis::edit_r_environ()` and then add something like the following:

```         
TESTTHAT_CPUS=4
```

Tests are run in alphabetical order by default, but you can often improve performance by starting the slowest tests first.
Specify these tests by supplying a comma separated list of glob patterns[^2] to the `Config/testthat/start-first` field in your `DESCRIPTION`, e.g.:

[^2]: See `?utils::glob2rx` for details

```         
Config/testthat/start-first: watcher, parallel*
```

## Basic operation

Each worker begins by loading testthat and the package being tested.
It then runs any setup files (so if you have existing setup files you'll need to make sure they work when executed in parallel).

testthat runs test *files* in parallel.
Once the worker pool is initialized, testthat then starts sending test files to workers, by default in alphabetical order: as soon as a subprocess has finished, it receives another file, until all files are done.
This means that state is persisted across test files: options are *not* reset, loaded packages are *not* unloaded, the global environment is *not* cleared, etc.
You are responsible for making sure each file leaves the world as it finds it.

## Common problems

-   If tests fail stochastically (i.e. they sometimes work and sometimes fail) you may have accidentally introduced a dependency between your test files.
    This sort of dependency is hard to track down due to the random nature, and you'll need to check all tests to make sure that they're not accidentally changing global state.
    `set_state_inspector()` will make this easier.

-   If you use [packaged scope test fixtures](https://testthat.r-lib.org/articles/test-fixtures.html#package), you'll need to review them to make sure that they work in parallel.
    For example, if you were previously creating a temporary database in the test directory, you'd need to instead create it in the session temporary directory so that each process gets its own independent version.

## Performance

There is some overhead associated with running tests in parallel:

-   Startup cost is linear in the number of subprocesses, because we need to create them in a loop.
    This is about 50ms on my laptop.
    Each subprocess needs to load testthat and the tested package, this happens in parallel, and we cannot do too much about it.

-   Clean up time is again linear in the number of subprocesses, and it about 80ms per subprocess on my laptop.

-   It seems that sending a message (i.e. a passing or failing expectation) is about 2ms currently.
    This is the total cost that includes sending the message, receiving it, and replying it to a non-parallel reporter.

This overhead generally means that if you have many test files that take a short amount of time, you're unlikely to see a huge benefit by using parallel tests.
For example, testthat itself takes about 10s to run tests in serial, and 8s to run the tests in parallel.

## Reporters

### Default reporters

See `default_reporter()` for how testthat selects the default reporter for `devtools::test()` and `testthat::test_local()`.
In short, testthat selects `ProgressReporter` for non-parallel and `ParallelProgressReporter` for parallel tests by default.
(Other testthat test functions, like `test_check()`, `test_file()` , etc. select different reporters by default.)

### Parallel support

Most reporters support parallel tests.
If a reporter is passed to `devtools::test()`, `testthat::test_dir()`, etc. directly, and it does not support parallel tests, then testthat runs the test files sequentially.

Currently the following reporters *don't* support parallel tests:

-   `DebugReporter`, because it is not currently possible to debug subprocesses.

-   `JunitReporter`, because this reporter records timing information for each test block, and this is currently only available for reporters that support multiple active test files.
    (See "Writing parallel reporters" below.)

-   `LocationReporter` because testthat currently does not include location information for successful tests when running in parallel, to minimize messaging between the processes.

-   `StopReporter`, as this is a reporter that testthat uses for interactive `expect_that()` calls.

The other built-in reporters all support parallel tests, with some subtle differences:

-   Reporters that stop after a certain number of failures can only stop at the end of a test file.

-   Reporters report all information about a file at once, unless they support *parallel updates*.
    E.g.
    `ProgressReporter` does not update its display until a test file is complete.

-   The standard output and standard error, i.e. `print()`, `cat()`, `message()`, etc. output from the test files are lost currently.
    If you want to use `cat()` or `message()` for print-debugging test cases, then the best is to temporarily run tests sequentially, by changing the `Config` entry in `DESCRIPTION` or setting `Sys.setenv(TESTTHAT_PARALLEL = "false")`.

### Writing parallel reporters

To support parallel tests, a reporter must be able to function when the test files run in a subprocess.
For example `DebugReporter` does not support parallel tests, because it requires direct interaction with the frames in the subprocess.
When running in parallel, testthat does not provide location information (source references) for test successes.

To support parallel tests, a reporter must set `self$capabilities$parallel_support` to `TRUE` in its `initialize()` method:

``` r
...
initialize = function(...) {
  super$initialize(...)
  self$capabilities$parallel_support <- TRUE
  ...
}
...
```

When running in parallel, testthat runs the reporter in the main process, and relays information between the reporter and the test code transparently.
(Currently the reporter does not even know that the tests are running in parallel.)

If a reporter does not support parallel updates (see below), then testthat internally caches all calls to the reporter methods from subprocesses, until a test file is complete.
This is because these reporters are not prepared for running multiple test files concurrently.
Once a test file is complete, testthat calls the reporter's `$start_file()` method, relays all `$start_test()` , `$end_test()`, `$add_result()`, etc. calls in the order they came in from the subprocess, and calls `$end_file()` .

### Parallel updates

The `ParallelProgressReporter` supports parallel updates.
This means that once a message from a subprocess comes in, the reporter is updated immediately.
For this to work, a reporter must be able to handle multiple test files concurrently.

A reporter declares parallel update support by setting `self$capabilities$parallel_updates` to `TRUE`:

``` r
...
initialize = function(...) {
  super$initialize(...)
  self$capabilities$parallel_support <- TRUE
  self$capabilities$parallel_updates <- TRUE
  ...
}
...
```

For these reporters, testthat does not cache the messages from the subprocesses.
Instead, when a message comes in:

-   It calls the `$start_file()` method, letting the reporter know which file the following calls apply to.
    This means that the reporter can receive multiple `$start_file()` calls for the same file.

-   Then relays the message from the subprocess, calling the appropriate `$start_test()` , `$add_result()`, etc. method.

testthat also calls the new `$update()` method of the reporter regularly, even if it does not receive any messages from the subprocess.
(Currently aims to do this every 100ms, but there are no guarantees.) The `$update()` method may implement a spinner to let the user know that the tests are running.

---
title: "Special files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Special files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
#| include: false
library(testthat)
```

This vignette describes the various special files that testthat understands: test, helper, setup/teardown, snapshot, and everything else.

## Test files

These are the bread and butter of testthat.
Test files live in `tests/testthat/`, start with either `test-` or `test_`, and end with `.r` or `.R`.
We recommend organising your test files so that there's a one-to-one correspondence between the files in `R/` and the files in `tests/testthat/` so that (e.g.) `R/myfile.R` has a matching `tests/testthat/test-myfile.R`.
This correspondence is maintained by functions like `usethis::use_r()` and `usethis::use_test()` and is taken advantage of by functions like `devtools::test_active_file()` and `devtools::test_coverage_active_file()`.

Test files are executed in alphabetical order, but you should strive to avoid dependencies between test files.
In principle, you should be able to be run your test files in any order or even at the same time.

## Helper files

Helper files live in `tests/testthat/`, start with `helper`, and end with `.r` or `.R`.
They are sourced by `devtools::load_all()` (so they're available interactively when developing your packages) and by `test_check()` and friends (so that they're available no matter how your tests are executed).

Helper files are a useful place for functions that you've extracted from repeated code in your tests, whether that be test fixtures (`vignette("test-fixtures")`), custom expectations (`vignette("custom-expectation")`), or skip helpers (`vignette("skipping")`).

## Setup files

Setup files live in `tests/testthat/`, start with `setup`, and end with `.r` or `.R`.
Typically there is only one setup file which, by convention, is `tests/testthat/setup.R`.
Setup files are sourced by `test_check()` and friends (so that they're available no matter how your tests are executed), but they are *not* sourced by `devtools::load_all()`.

Setup files are good place to put truly global test setup that would be impractical to build into every single test and that might be tailored for test execution in non-interactive or remote environments.
Examples:

-   Turning off behaviour aimed at an interactive user, such as messaging or writing to the clipboard.
-   Setting up a cache folder.

If any of your setup should be reversed after test execution (i.e. it needs to be torn down), we recommend maintaining that teardown code alongside the setup code, in `setup.R`, because this makes it easier to ensure they stay in sync.
The artificial environment `teardown_env()` exists as a magical handle to use in `withr::defer()` and `withr::local_*()`.
A legacy approach (which still works, but is no longer recommended) is to put teardown code in `tests/testthat/teardown.R`.

Here's a `setup.R` example from the reprex package, where we turn off clipboard and HTML preview functionality during testing:

```{r eval = FALSE}
op <- options(reprex.clipboard = FALSE, reprex.html_preview = FALSE)

withr::defer(options(op), teardown_env())
```

Since we are just modifying options here, we can be even more concise and use the pre-built function `withr::local_options()` and pass `teardown_env()` as the `.local_envir`:

```{r eval = FALSE}
withr::local_options(
  list(reprex.clipboard = FALSE, reprex.html_preview = FALSE),
  .local_envir = teardown_env()
)
```

### Teardown files

Teardown files live in `tests/testhat/`, start with `teardown` and end with `.r` or `.R`.
They are executed after the tests are run, but we no longer recommend using them as it's easier to check that you clean up every mess that you make if you interleave setup and tear down code as described above.

## Snapshot files

Snapshot files live in `tests/testthat/_snaps/`.
Snapshot file are named automatically based on the name of the test file so that `tests/testthat/test-one.R` will generated `tests/testthat/_snaps/one.md`.
Learn more about snapshot tests in `vignette("snapshotting")`.

## Other files and folders

Other files and folders in `tests/testthat/` are ignored by testthat, making them a good place to store persistent test data.
Since the precise location of the `test/testthat/` directory varies slightly depending on how you're running the test, we recommend creating paths to these files and directories using `test_path()`.

---
title: "Skipping tests"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Skipping tests}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Some times you have tests that you don't want to run in certain circumstances.
This vignette describes how to **skip** tests to avoid execution in undesired environments.
Skipping is a relatively advanced topic because in most cases you want all your tests to run everywhere.
The most common exceptions are:

-   You're testing a web service that occasionally fails, and you don't want to run the tests on CRAN.
    Or maybe the API requires authentication, and you can only run the tests when you've [securely distributed](https://gargle.r-lib.org/articles/articles/managing-tokens-securely.html) some secrets.

-   You're relying on features that not all operating systems possess, and want to make sure your code doesn't run on a platform where it doesn't work.
    This platform tends to be Windows, since amongst other things, it lacks full utf8 support.

-   You're writing your tests for multiple versions of R or multiple versions of a dependency and you want to skip when a feature isn't available.
    You generally don't need to skip tests if a suggested package is not installed.
    This is only needed in exceptional circumstances, e.g. when a package is not available on some operating system.

```{r setup}
library(testthat)
```

## Basics

testthat comes with a variety of helpers for the most common situations:

-   `skip_on_cran()` skips tests on CRAN.
    This is useful for slow tests and tests that occasionally fail for reasons outside of your control.

-   `skip_on_os()` allows you to skip tests on a specific operating system.
    Generally, you should strive to avoid this as much as possible (so your code works the same on all platforms), but sometimes it's just not possible.

-   `skip_on_ci()` skips tests on most continuous integration platforms (e.g. GitHub Actions, Travis, Appveyor).

You can also easily implement your own using either `skip_if()` or `skip_if_not()`, which both take an expression that should yield a single `TRUE` or `FALSE`.

All reporters show which tests as skipped.
As of testthat 3.0.0, ProgressReporter (used interactively) and CheckReporter (used inside of `R CMD check`) also display a summary of skips across all tests.
It looks something like this:

    ── Skipped tests  ───────────────────────────────────────────────────────
    ● No token (3)
    ● On CRAN (1)

You should keep an on eye this when developing interactively to make sure that you're not accidentally skipping the wrong things.

## Helpers

If you find yourself using the same `skip_if()`/`skip_if_not()` expression across multiple tests, it's a good idea to create a helper function.
This function should start with `skip_` and live somewhere in your `R/` directory.

```{r}
skip_if_Tuesday <- function() {
  if (as.POSIXlt(Sys.Date())$wday != 2) {
    return(invisible(TRUE))
  }
  
  skip("Not run on Tuesday")
}
```

It's important to test your skip helpers because it's easy to miss if you're skipping more often than desired, and the test code is never run.
This is unlikely to happen locally (since you'll see the skipped tests in the summary), but is quite possible in continuous integration.

For that reason, it's a good idea to add a test that you skip is activated when you expect.
skips are a special type of condition, so you can test for their presence/absence with `expect_condition()`.
For example, imagine that you've defined a custom skipper that skips tests whenever an environment variable `DANGER` is set:

```{r}
skip_if_dangerous <- function() {
  if (identical(Sys.getenv("DANGER"), "")) {
    return(invisible(TRUE))
  }
  
  skip("Not run in dangerous enviromnents")
}
```

Then you can use `expect_condition()` to test that it skips tests when it should, and doesn't skip when it shouldn't:

```{r}
test_that("skip_if_dangerous work", {
  # Test that a skip happens
  withr::local_envvar(DANGER = "yes")
  expect_condition(skip_if_dangerous(), class = "skip") 

  # Test that a skip doesn't happen
  withr::local_envvar(DANGER = "")
  expect_condition(skip_if_dangerous(), NA, class = "skip")
})
```

Testing `skip_if_Tuesday()` is harder because there's no way to control the skipping from the outside.
That means you'd need to "mock" its behaviour in a test, using the [mockery](https://github.com/r-lib/mockery) or [mockr](https://krlmlr.github.io/mockr/) packages.

---
title: "Skipping tests"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Skipping tests}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Sometimes you have tests that you can't or don't want to run in certain circumstances.
This vignette describes how to **skip** tests to avoid execution in undesired environments.
The most common scenarios are:

-   You're testing a web service that occasionally fails, and you don't want to run the tests on CRAN.
    Or the API requires authentication, and you can only run the tests when you've [securely distributed](https://gargle.r-lib.org/articles/articles/managing-tokens-securely.html) secrets.

-   You're relying on features that not all operating systems possess, and want to make sure your code doesn't run on platforms where it doesn't work.
    The most common platform with limitations is Windows, which among other things lacks full UTF-8 support.

-   You're writing tests for multiple versions of R or multiple versions of a dependency, and you want to skip when a feature isn't available.
    You generally don't need to skip tests if a suggested package is not installed.
    This is only needed in exceptional circumstances, e.g., when a package is not available on some operating systems.

```{r setup}
library(testthat)
```

## Basics

testthat comes with a variety of helpers for the most common situations:

-   `skip_if_not_installed()` skips if a required package is not installed. You can optionally supply a minimal version too.

-   `skip_on_cran()` skips tests on CRAN. `skip_on_bioc()` skips tests on Bioconductor.
    This is useful for slow tests and tests that occasionally fail for reasons outside of your control.

-   `skip_on_os()` allows you to skip tests on a specific operating system.
    Generally, you should strive to avoid this as much as possible (so your code works the same on all platforms), but sometimes it's just not possible.

-   `skip_on_ci()` skips tests on most CI platforms (e.g., GitHub Actions).

-   `skip_on_covr()` skips tests during code coverage.

-   `skip_unless_r(">= 4.2")` only runs tests for newer R versions. 
    `skip_unless_r("< 4.2")` only runs tests for older R versions.

You can implement your own using skips `skip_if()` or `skip_if_not()`:

```{r}
#| eval: false

# Only run test if a token file is available
skip_if_not(file.exists("secure-token.json"))

# Only run test if R has memory profiling capabilities
skip_if_not(capabilities("profmem"))

# Only run if we've opted-in to slow tests with an env var
skip_if(Sys.getenv("RUN_SLOW_TESTS") == "true")
```

All reporters show which tests are skipped.
As of testthat 3.0.0, ProgressReporter (used interactively) and CheckReporter (used inside `R CMD check`) also display a summary of skips across all tests.
It looks something like this:

```         
── Skipped tests  ───────────────────────────────────────────────────────
● No token (3)
● On CRAN (1)
```

This display is really important, and you should keep an eye on it when working on your test suite. If you accidentally skip too many tests, you can trick yourself into believing your code is working correctly, when actually you're just not testing it.

## Helpers

If you find yourself using the same `skip_if()` or `skip_if_not()` expression across multiple tests, it's a good idea to create a helper function.
This function should start with `skip_` and live in a `tests/testthat/helper-{something}.R` file:

```{r}
skip_if_dangerous <- function() {
  if (!identical(Sys.getenv("DANGER"), "")) {
    skip("Not run in dangerous environments.")
  } else {
    invisible()
  }
}
```

## Embedding `skip()` in package functions

Another useful technique is to embed a `skip()` directly into a package function.
For example, take a look at [`pkgdown:::convert_markdown_to_html()`](https://github.com/r-lib/pkgdown/blob/v2.0.7/R/markdown.R#L95-L106), which absolutely cannot work if the Pandoc tool is unavailable:

```{r eval = FALSE}
convert_markdown_to_html <- function(in_path, out_path, ...) {
  if (rmarkdown::pandoc_available("2.0")) {
    from <- "markdown+gfm_auto_identifiers-citations+emoji+autolink_bare_uris"
  } else if (rmarkdown::pandoc_available("1.12.3")) {
    from <- "markdown_github-hard_line_breaks+tex_math_dollars+tex_math_single_backslash+header_attributes"
  } else {
    if (is_testing()) {
      testthat::skip("Pandoc not available")
    } else {
      abort("Pandoc not available")
    }
  }
  
  ...
}
```

If Pandoc is not available when `convert_markdown_to_html()` executes, it throws an error *unless* it appears to be part of a test run, in which case the test is skipped.
This is an alternative to implementing a custom skipper, e.g., `skip_if_no_pandoc()`, and inserting it into many of pkgdown's tests.

We don't want pkgdown to have a runtime dependency on testthat, so pkgdown includes a copy of `testthat::is_testing()`:

```{r}
is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}
```

It might look like the code still has a runtime dependency on testthat, because of the call to `testthat::skip()`.
But `testthat::skip()` is only executed during a test run, which means that testthat is installed.

We have mixed feelings about this approach.
On the one hand, it feels elegant and concise, and it absolutely guarantees that you'll never miss a needed skip in one of your tests.
On the other hand, it mixes code and tests in an unusual way, and when you're focused on the tests, it's easy to miss the fact that a package function contains a `skip()`.

---
title: "Snapshot tests"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Snapshot tests}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(1014)
```

The goal of a unit test is to record the expected output of a function using code.
This is a powerful technique because not only does it ensure that code doesn't change unexpectedly, it also expresses the desired behaviour in a way that a human can understand.

However, it's not always convenient to record the expected behaviour with code.
Some challenges include:

-   Text output that includes many characters like quotes and newlines that require special handling in a string.

-   Output that is large, making it painful to define the reference output, and bloating the size of the test file and making it hard to navigate.

-   Binary formats like plots or images, which are very difficult to describe in code: i.e. the plot looks right, the error message is useful to a human, the print method uses colour effectively.

For these situations, testthat provides an alternative mechanism: snapshot tests.
Instead of using code to describe expected output, snapshot tests (also known as [golden tests](https://ro-che.info/articles/2017-12-04-golden-tests)) record results in a separate human readable file.
Snapshot tests in testthat are inspired primarily by [Jest](https://jestjs.io/docs/en/snapshot-testing), thanks to a number of very useful discussions with Joe Cheng.

```{r setup}
library(testthat)
```

```{r include = FALSE}
snapper <- testthat:::SnapshotReporter$new()
options(testthat.snapshotter = snapper)
snapper$start_file("pizza", "test")
```

## Basic workflow

We'll illustrate the basic workflow with a simple function that generates an HTML heading.
It can optionally include an `id` attribute, which allows you to construct a link directly to that heading.

```{r}
bullets <- function(text, id = NULL) {
  paste0(
    "<ul", if (!is.null(id)) paste0(" id=\"", id, "\""), ">\n", 
    paste0("  <li>", text, "</li>\n", collapse = ""),
    "</ul>\n"
  )
}
cat(bullets("a", id = "x"))
```

Testing this simple function is relatively painful.
To write the test you have to carefully escape the newlines and quotes.
And then when you re-read the test in the future, all that escaping makes it hard to tell exactly what it's supposed to return.

```{r}
test_that("bullets", {
  expect_equal(bullets("a"), "<ul>\n  <li>a</li>\n</ul>\n")
  expect_equal(bullets("a", id = "x"), "<ul id=\"x\">\n  <li>a</li>\n</ul>\n")
})
```

This is a great place to use snapshot testing.
To do this we make two changes to our code: we use `expect_snapshot_output()` instead of `expect_equal()`, and wrap the call in `cat()` (to avoid `[1]` in the output, like I my first interactive example).

```{r}
test_that("bullets", {
  expect_snapshot_output(cat(bullets("a")))
  expect_snapshot_output(cat(bullets("a", "b")))
})
```

```{r, include = FALSE}
# Reset snapshot test
snapper$end_file()
snapper$start_file("pizza", "test")
```

We we run the test for the first time, it automatically generates reference output, and prints it, so that you can visually confirm that it's correct.
The output is automatically saved in `_snaps/{name}.R`.
The name of the snapshot matches your test file name --- e.g. if your test is `test-pizza.R` then your snapshot will be saved in `test/testthat/_snaps/pizza.md`.
As the file name suggests, this is a markdown file, which I'll explain shortly.

If you run the test again, it'll succeed:

```{r}
test_that("bullets", {
  expect_snapshot_output(cat(bullets("a")))
  expect_snapshot_output(cat(bullets("a", "b")))
})
```

```{r, include = FALSE}
# Reset snapshot test
snapper$end_file()
snapper$start_file("pizza", "test")
```

But if you change the underlying code, say to tweak the indenting, the test will fail:

```{r}
bullets <- function(text, id = NULL) {
  paste0(
    "<ul", if (!is.null(id)) paste0(" id=\"", id, "\""), ">\n", 
    paste0("<li>", text, "</li>\n", collapse = ""),
    "</ul>\n"
  )
}
test_that("bullets", {
  expect_snapshot_output(cat(bullets("a")))
  expect_snapshot_output(cat(bullets("a", "b")))
})
```

If this is a deliberate change, you can follow the advice in the message and update the snapshots for that file by running `snapshot_accept("pizza")`; otherwise you can fix the bug and your tests will pass once more.
(You can also accept snapshot for all files with `snapshot_accept()`).

### Snapshot format

Snapshots are recorded using a subset of markdown.
You might wonder why we use markdown?
It's important that snapshots be readable by humans, because humans have to look at it during code reviews.
Reviewers often don't run your code but still want to understand the changes.

Here's the snapshot file generated by the test above:

``` {.md}
# bullets

    <ul>
      <li>a</li>
    </ul>
  
---

    <ul id="x">
      <li>a</li>
    </ul>
```

Each test starts with `# {test name}`, a level 1 heading.
Within a test, each snapshot expectation is indented by four spaces, i.e. as code, and are separated by `---`, a horizontal rule.

### Interactive usage

Because the snapshot output uses the name of the current test file and the current test, snapshot expectations don't really work when run interactively at the console.
Since they can't automatically find the reference output, they instead just print the current value for manual inspection.

## Other types of snapshot

`expect_snapshot_output()` is probably the most commonly used snapshot function but there are a few other variants that are useful in other case.

-   `expect_snapshot_error()` works like `expect_snapshot_output()` but captures an error message, rather than console output.

-   `expect_snapshot_value()` checks that the value of any R object stays the same.
    It has a number of serialisation approaches built-in that offer a tradeoff between accuracy and human readability.

## Whole file snapshotting

`expect_snapshot()`, `expect_snapshot_output()`, `expect_snapshot_error()`, and `expect_snapshot_value()` all store their snapshots in a single file per test.
But that doesn't work for all file types --- for example, what happens if you want to snapshot an image?
`expect_snapshot_file()` provides an alternative workflow that generates one snapshot per expectation, rather than one file per test.
Assuming you're in `test-burger.R` then the snapshot created by `expect_snapshot_file(code_that_returns_path_to_file(), "toppings.png")` would be saved in `tests/testthat/_snaps/burger/toppings.png`.
If a future change in the code creates a different file it will be saved in `tests/testthat/_snaps/burger/toppings.new.png`.

Unlike `expect_snapshot()` and friends, `expect_snapshot_file()` can't provide an automatic diff when the test fails.
Instead you'll need to call `snapshot_review()`.
This launches a Shiny app that allows you to visually review each change and approve it if it's deliberate:

![](review-image.png) ![](review-text.png)

The display varies based on the file type (currently text files, common image files, and csv files are supported).

Sometimes the failure occurs in a non-interactive environment where you can't run `snapshot_review()`, e.g. in `R CMD check`.
In this case, the easiest fix is to retrieve the `.new` file, copy it into the appropriate directory, then run `snapshot_review()` locally.
If your code was run on a CI platform, you'll need to start by downloading the run "artifact", which contains the check folder.

In most cases, we don't expect you to use `expect_snapshot_file()` directly.
Instead, you'll use it via a wrapper that does its best to gracefully skip tests when differences in platform or package versions make it unlikely to generate perfectly reproducible output.

## Previous work

This is not the first time that testthat has attempted to provide snapshot testing (although it's the first time I knew what other languages called them).
This section describes some of the previous attempts and why we believe the new approach is better.

-   `verify_output()` has three main drawbacks:

    -   You have to supply a path where the output will be saved.
        This seems like a small issue, but thinking of a good name, and managing the difference between interactive and test-time paths introduces a suprising amount of friction.

    -   It always overwrites the previous result; automatically assuming that the changes are correct.
        That means you have to use it with git and it's easy to accidentally accept unwanted changes.

    -   It's relatively coarse grained, which means tests that use it tend to keep growing and growing.

-   `expect_known_output()` is finer grained version of `verify_output()` that captures output from a single function.
    The requirement to produce a path for each individual expectation makes it even more painful to use.

-   `expect_known_value()` and `expect_known_hash()` have all the disadvantages of `expect_known_output()`, but also produce binary output meaning that you can't easily review test differences in pull requests.

```{r, include = FALSE}
snapper$snaps_cleanup()
```

---
title: "Custom expectations"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Custom expectations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
library(testthat)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

This vignette shows you how to create custom expectations that work identically to the built-in `expect_` functions.  Since these functions will need to be loaded when your package is loaded for testing, it is recommended that `expect_` functions be defined in `test-helpers.R` in your packages `R/` directory.

## Creating an expectation

There are three main parts to writing an expectation, as illustrated by `expect_length()`:

```{r}
expect_length <- function(object, n) {
  # 1. Capture object and label
  act <- quasi_label(rlang::enquo(object), arg = "object")

  # 2. Call expect()
  act$n <- length(act$val)
  expect(
    act$n == n,
    sprintf("%s has length %i, not length %i.", act$lab, act$n, n)
  )

  # 3. Invisibly return the value
  invisible(act$val)
}
```

## Quasi-labelling

The first step in any expectation is to capture the actual object, and generate a label for it to use if a failure occur. All testthat expectations support quasiquotation so that you can unquote variables. This makes it easier to generate good labels when the expectation is called from a function or within a for loop.

By convention, the first argument to every `expect_` function is called `object`, and you capture it's value (`val`) and label (`lab`) with `act <- quasi_label(enquo(object))`, where `act` is short for actual.

### Verify the expectation

Next, you should verify the expectation. This often involves a little computation (here just figuring out the `length`), and you should typically store the results back into the `act` object.

Next you call `expect()`. This has two arguments:

1.  `ok`: was the expectation successful? This is usually easy to write

2.  `failure_message`: What informative error message should be reported to
    the user so that they can diagnose the problem. This is often hard to 
    write!
    
    For historical reasons, most built-in expectations generate these with
    `sprintf()`, but today I'd recommend using the 
    [glue](https://glue.tidyverse.org) package

### Invisibly return the input

Expectation functions are called primarily for their side-effects (triggering a failure), so should invisibly return their input, `act$val`. This allows expectations to be chained:

```{r}
mtcars %>%
  expect_type("list") %>%
  expect_s3_class("data.frame") %>% 
  expect_length(11)
```

## `succeed()` and `fail()`

For expectations with more complex logic governing when success or failure occurs, you can use `succeed()` and `fail()`. These are simple wrappers around `expect()` that allow you to write code that looks like this:

```{r}
expect_length <- function(object, n) {
  act <- quasi_label(rlang::enquo(object), arg = "object")

  act$n <- length(act$val)
  if (act$n == n) {
    succeed()
    return(invisible(act$val))
  }

  message <- sprintf("%s has length %i, not length %i.", act$lab, act$n, n)
  fail(message)
}
```

## Testing your expectations

Use the expectations `expect_success()` and `expect_failure()` to test your expectation.

```{r}
test_that("length computed correctly", {
  expect_success(expect_length(1, 1))
  expect_failure(expect_length(1, 2), "has length 1, not length 2.")
  expect_success(expect_length(1:10, 10))
  expect_success(expect_length(letters[1:5], 5))
})
```

# full condition message is printed with rlang

    Code
      foo <- error_cnd("foo", message = "Title parent.")
      abort("Title.", parent = foo)
    Error <rlang_error>
      Title.

# can print with and without condition classes

    Code
      f()
    Message <simpleMessage>
      foo
    Warning <simpleWarning>
      bar
    Error <simpleError>
      baz

---

    Code
      f()
    Message <simpleMessage>
      foo
    Warning <simpleWarning>
      bar
    Error <simpleError>
      baz

# errors and warnings are folded

    Code
      f()
    Warning <simpleWarning>
      foo
    Error <simpleError>
      bar

# set_state_inspector() verifies its inputs

    Code
      set_state_inspector(function(x) 123)
    Condition
      Error in `set_state_inspector()`:
      ! `callback` must be a zero-arg function, or NULL

# can detect state changes

    [ FAIL 0 | WARN 1 | SKIP 0 | PASS 1 ]
    
    == Warnings ====================================================================
    -- Warning ('reporters/state-change.R:1:1'): options ---------------------------
    Global state has changed:
    `before$x` is NULL
    `after$x` is a double vector (1)
    
    [ FAIL 0 | WARN 1 | SKIP 0 | PASS 1 ]

# runs all tests and records output

                  file context                                      test nb failed skipped error warning passed
    1     test-basic.R                     logical tests act as expected  2      0   FALSE FALSE       0      2
    2     test-basic.R                   logical tests ignore attributes  2      0   FALSE FALSE       0      2
    3     test-basic.R                                    equality holds  2      0   FALSE FALSE       0      2
    4     test-basic.R         can't access variables from other tests 2  1      0    TRUE FALSE       0      0
    5     test-basic.R         can't access variables from other tests 1  1      0   FALSE FALSE       0      1
    6     test-empty.R                                        empty test  1      0    TRUE FALSE       0      0
    7     test-empty.R                             empty test with error  0      0   FALSE  TRUE       0      0
    8    test-errors.R                                            simple  0      0   FALSE  TRUE       0      0
    9    test-errors.R                                 after one success  1      0   FALSE  TRUE       0      1
    10   test-errors.R                                 after one failure  1      1   FALSE  TRUE       0      0
    11   test-errors.R                                       in the test  0      0   FALSE  TRUE       0      0
    12   test-errors.R                                   in expect_error  1      0   FALSE FALSE       0      1
    13 test-failures.R                                  just one failure  1      1   FALSE FALSE       0      0
    14 test-failures.R                                one failure on two  2      1   FALSE FALSE       0      1
    15 test-failures.R                                        no failure  2      0   FALSE FALSE       0      2
    16   test-helper.R                                       helper test  1      0   FALSE FALSE       0      1
    17     test-skip.R                                        Skips skip  1      0    TRUE FALSE       0      0

# can force cli to display RStudio style hyperlinks

    Code
      str(cli::ansi_hyperlink_types())
    Output
      List of 4
       $ href    : logi FALSE
       $ run     : logi FALSE
       $ help    : logi FALSE
       $ vignette: logi FALSE

---

    Code
      str(cli::ansi_hyperlink_types())
    Output
      List of 4
       $ href    : logi TRUE
       $ run     : logi TRUE
        ..- attr(*, "type")= chr "rstudio"
       $ help    : logi TRUE
        ..- attr(*, "type")= chr "rstudio"
       $ vignette: logi TRUE
        ..- attr(*, "type")= chr "rstudio"

# informs about files being accepted

    Code
      snapshot_accept(path = path)
    Message
      Updating snapshots: 'a.md' and 'test/b.txt'.

# useful mesasge if no files to accept

    Code
      snapshot_accept(path = path)
    Message
      No snapshots to update.

# can work with variants

    Code
      snapshot_accept(path = path)
    Message
      Updating snapshots: 'foo/a.md'.

---

    Code
      snapshot_accept("foo/a", path = path)
    Message
      Updating snapshots: 'foo/a.md'.

# snapshot_reject deletes .new files

    Code
      snapshot_reject(path = path)
    Message
      Rejecting snapshots:
      * a.md
      * b.md

# validates its inputs

    Code
      local_mocked_s3_method(1)
    Condition
      Error in `local_mocked_s3_method()`:
      ! `generic` must be a single string, not the number 1.
    Code
      local_mocked_s3_method("mean", 1)
    Condition
      Error in `local_mocked_s3_method()`:
      ! `signature` must be a single string, not the number 1.
    Code
      local_mocked_s3_method("mean", "bar", 1)
    Condition
      Error in `local_mocked_s3_method()`:
      ! `definition` must be a function or `NULL`, not the number 1.
    Code
      local_mocked_s3_method("notAGeneric", "bar", function() { })
    Condition
      Error in `get()`:
      ! object 'notAGeneric' not found

---

    Code
      local_mocked_s4_method(1)
    Condition
      Error in `local_mocked_s4_method()`:
      ! `generic` must be a single string, not the number 1.
    Code
      local_mocked_s4_method("mean", 1)
    Condition
      Error in `local_mocked_s4_method()`:
      ! `signature` must be a character vector, not the number 1.
    Code
      local_mocked_s4_method("mean", "bar", 1)
    Condition
      Error in `local_mocked_s4_method()`:
      ! `definition` must be a function or `NULL`, not the number 1.
    Code
      local_mocked_s4_method("notAGeneric", "bar", function() { })
    Condition
      Error in `local_mocked_s4_method()`:
      ! Can't find generic `notAGeneric()`.

---

    Code
      local_mocked_r6_class(mean)
    Condition
      Error in `local_mocked_r6_class()`:
      ! `class` must be an R6 class definition, not a function.
    Code
      local_mocked_r6_class(TestMockClass, public = 1)
    Condition
      Error in `local_mocked_r6_class()`:
      ! `public` must be a list, not the number 1.
    Code
      local_mocked_r6_class(TestMockClass, private = 1)
    Condition
      Error in `local_mocked_r6_class()`:
      ! `private` must be a list, not the number 1.

# full condition message is printed with rlang

    Code
      foo <- error_cnd("foo", message = "Title parent.")
      abort("Title.", parent = foo)
    Condition
      Error:
      ! Title.
      Caused by error:
      ! Title parent.

# can print with and without condition classes

    Code
      f()
    Message <simpleMessage>
      foo
    Condition <simpleWarning>
      Warning in `f()`:
      bar
    Condition <simpleError>
      Error in `f()`:
      ! baz

---

    Code
      f()
    Message
      foo
    Condition
      Warning in `f()`:
      bar
      Error in `f()`:
      ! baz

# errors and warnings are folded

    Code
      f()
    Condition
      Warning in `f()`:
      foo
      Error in `f()`:
      ! bar

# reports issues immediately but not successes

    FAILURE: 'reporters/tests.R:13:3' -----------------
    Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    FAILURE: 'reporters/tests.R:17:8' -----------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
    
    ERROR: 'reporters/tests.R:24:3' -------------------
    Error in `eval(code, test_env)`: stop
    
    ERROR: 'reporters/tests.R:30:8' -------------------
    Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
    
    SKIP: 'reporters/tests.R:38:3' --------------------
    Reason: skip
    
    SKIP: 'reporters/tests.R:41:1' --------------------
    Reason: empty test
    
    WARNING: 'reporters/tests.R:47:5' -----------------
    def
    Backtrace:
        x
     1. \-f()
    
    SKIP: 'reporters/tests.R:45:1' --------------------
    Reason: empty test
    
    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]

# reports only summary for all successes

    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 7 ]

# fails after max_fail tests

    FAILURE: 'reporters/fail-many.R:3:5' --------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    FAILURE: 'reporters/fail-many.R:3:5' --------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    FAILURE: 'reporters/fail-many.R:3:5' --------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    
    Maximum number of failures exceeded; quitting.
    i Increase this number with (e.g.) `testthat::set_max_fails(Inf)` 

# missing arguments are propagated

    Code
      expect_null(x$missing)
    Condition
      Error:
      ! Expected `x$missing` to be NULL.
      Differences:
      `actual` is absent
      `expected` is NULL

# produces useful summaries for long calls

    Code
      expr_label(quote(foo(a = "this is a long argument", b = "this is a long argument",
        c = "this is a long argument")))
    Output
      [1] "`foo(...)`"
    Code
      expr_label(quote(arg + arg + arg + arg + arg + arg + arg + arg + arg + arg +
      arg + arg))
    Output
      [1] "`... + arg`"
    Code
      expr_label(quote(arg + (arg + arg + arg + arg + arg + arg + arg + arg + arg +
        arg + arg)))
    Output
      [1] "`arg + ...`"
    Code
      expr_label(quote(function(a, b, c) {
        a + b + c
      }))
    Output
      [1] "`function(a, b, c) ...`"

# informative error for missing arg

    Code
      expect_equal()
    Condition
      Error in `expect_equal()`:
      ! argument `object` is missing, with no default.

# can snapshot output

    y

---

    y

---

    y

---

    y

# can snapshot everything

    Code
      f()
    Output
      [1] "1"
    Message
      2
    Condition
      Warning in `f()`:
      3
      Error in `f()`:
      ! 4

# empty lines are preserved

    Code
      f()
    Output
      1
      
    Message
      2
      
    Condition
      Warning in `f()`:
      3
      Error in `f()`:
      ! 4

# line-endings fixed before comparison

    Code
      cat(x)
    Output
      a
      b

# multiple outputs of same type are collapsed

    Code
      x <- 1
      y <- 1
      {
        message("a")
        message("b")
      }
    Message
      a
      b
    Code
      {
        warning("a")
        warning("b")
      }
    Condition
      Warning:
      a
      Warning:
      b

# can scrub output/messages/warnings/errors

    Code
      secret()
    Output
      [1] "<redacted>"
    Message
      <redacted>
    Condition
      Warning in `<redacted>()`:
      <redacted>
      Error in `<redacted>()`:
      ! <redacted>

---

    Code
      print("secret")
    Output
      [1] "****"

# always checks error status

    Code
      expect_snapshot(print("!"), error = TRUE)
    Condition
      Error:
      ! Expected `print("!")` to throw a error.

# snapshots of failures fail

    Code
      expect_snapshot(fail())
    Condition
      Error:
      ! Failure has been forced

# can capture error/warning messages

    This is an error

---

    This is a warning

# snapshot captures deprecations

    Code
      foo()
    Condition
      Warning:
      `foo()` was deprecated in testthat 1.0.0.

---

    `foo()` was deprecated in testthat 1.0.0.

---

    `foo()` was deprecated in testthat 1.0.0.

# can check error/warning classes

    Code
      expect_snapshot_error(1)
    Condition
      Error:
      ! 1 did not generate error

---

    Code
      expect_snapshot_error(1, class = "myerror")
    Condition
      Error:
      ! 1 did not generate error with class 'myerror'

---

    Code
      expect_snapshot_warning(1)
    Condition
      Error:
      ! 1 did not generate warning

---

    Code
      expect_snapshot_warning(1, class = "mywarning")
    Condition
      Error:
      ! 1 did not generate warning with class 'mywarning'

# snapshot handles multi-line input

    Code
      1 + 2
    Output
      [1] 3
    Code
      3 + 4
    Output
      [1] 7
    Code
      # this is a comment

# snapshot captures output if visible

    Code
      f_visible()
    Output
      [1] "x"

---

    Code
      f_invisible()

# captures custom classes

    Code
      f()
    Message
      Hello
    Condition
      Warning:
      Goodbye
      Error in `f()`:
      ! Eeek!

# even with multiple lines

    a
    b
    c

---

    a
    b
    c

# `expect_snapshot()` does not inject

    Code
      x <- quote(!!foo)
      expect_equal(x, call("!", call("!", quote(foo))))

# full condition message is printed with rlang

    Code
      foo <- error_cnd("foo", message = "Title parent.")
      abort("Title.", parent = foo)
    Condition
      Error:
      ! Title.
      Caused by error:
      ! Title parent.

# can print with and without condition classes

    Code
      f()
    Message <simpleMessage>
      foo
    Condition <simpleWarning>
      Warning in `f()`:
      bar
    Condition <simpleError>
      Error in `f()`:
      ! baz

---

    Code
      f()
    Message
      foo
    Condition
      Warning in `f()`:
      bar
      Error in `f()`:
      ! baz

# errors and warnings are folded

    Code
      f()
    Condition
      Warning in `f()`:
      foo
      Error in `f()`:
      ! bar

# hint is informative

    Code
      snapshot_hint("bar.R", reset_output = FALSE)
    Output
      * Run `testthat::snapshot_accept("bar.R")` to accept the change.
      * Run `testthat::snapshot_review("bar.R")` to review the change.

# expect_snapshot validates its inputs

    Code
      expect_snapshot(1 + 1, cran = "yes")
    Condition
      Error in `expect_snapshot()`:
      ! `cran` must be `TRUE` or `FALSE`, not the string "yes".
    Code
      expect_snapshot(1 + 1, error = "yes")
    Condition
      Error in `expect_snapshot()`:
      ! `error` must be `TRUE` or `FALSE`, not the string "yes".
    Code
      expect_snapshot(1 + 1, cnd_class = "yes")
    Condition
      Error in `expect_snapshot()`:
      ! `cnd_class` must be `TRUE` or `FALSE`, not the string "yes".

# expect_snapshot_output validates its inputs

    Code
      expect_snapshot_output(cat("test"), cran = "yes")
    Condition
      Error in `expect_snapshot_output()`:
      ! `cran` must be `TRUE` or `FALSE`, not the string "yes".

# expect_snapshot_error validates its inputs

    Code
      expect_snapshot_error(stop("!"), class = 123)
    Condition
      Error in `expect_snapshot_error()`:
      ! `class` must be a single string, not the number 123.
    Code
      expect_snapshot_error(stop("!"), cran = "yes")
    Condition
      Error in `expect_snapshot_error()`:
      ! `cran` must be `TRUE` or `FALSE`, not the string "yes".

# expect_snapshot_warning validates its inputs

    Code
      expect_snapshot_warning(warning("!"), class = 123)
    Condition
      Error in `expect_snapshot_warning()`:
      ! `class` must be a single string, not the number 123.
    Code
      expect_snapshot_warning(warning("!"), cran = "yes")
    Condition
      Error in `expect_snapshot_warning()`:
      ! `cran` must be `TRUE` or `FALSE`, not the string "yes".

# reporter doesn't change without warning

    <?xml version="1.0" encoding="UTF-8"?>
    <testsuites>
      <testsuite name="Successes" timestamp="1999:12:31 23:59:59" hostname="nodename" tests="1" skipped="0" failures="0" errors="0" time="0">
        <testcase time="0" classname="Successes" name="Success"/>
      </testsuite>
      <testsuite name="Failures" timestamp="1999:12:31 23:59:59" hostname="nodename" tests="2" skipped="0" failures="2" errors="0" time="0">
        <testcase time="0" classname="Failures" name="Failure_1">
          <failure type="failure" message="Expected `x` to be TRUE. ('reporters/tests.R:13:3')">Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    </failure>
        </testcase>
        <testcase time="0" classname="Failures" name="Failure_2a">
          <failure type="failure" message="Expected FALSE to be TRUE. ('reporters/tests.R:17:8')">Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)</failure>
        </testcase>
      </testsuite>
      <testsuite name="Errors" timestamp="1999:12:31 23:59:59" hostname="nodename" tests="2" skipped="0" failures="0" errors="2" time="0">
        <testcase time="0" classname="Errors" name="Error_1">
          <error type="error" message="Error in `eval(code, test_env)`: stop ('reporters/tests.R:24:3')">Error in `eval(code, test_env)`: stop</error>
        </testcase>
        <testcase time="0" classname="Errors" name="errors_get_tracebacks">
          <error type="error" message="Error in `h()`: ! ('reporters/tests.R:30:8')">Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()</error>
        </testcase>
      </testsuite>
      <testsuite name="Skips" timestamp="1999:12:31 23:59:59" hostname="nodename" tests="2" skipped="2" failures="0" errors="0" time="0">
        <testcase time="0" classname="Skips" name="explicit_skips_are_reported">
          <skipped message="Reason: skip ('reporters/tests.R:38:3')"/>
        </testcase>
        <testcase time="0" classname="Skips" name="empty_tests_are_implicitly_skipped">
          <skipped message="Reason: empty test ('reporters/tests.R:41:1')"/>
        </testcase>
      </testsuite>
      <testsuite name="Warnings" timestamp="1999:12:31 23:59:59" hostname="nodename" tests="2" skipped="1" failures="0" errors="0" time="0">
        <testcase time="0" classname="Warnings" name="warnings_get_backtraces">
          <system-out>def
    Backtrace:
        x
     1. \-f()</system-out>
        </testcase>
        <testcase time="0" classname="Warnings" name="warnings_get_backtraces">
          <skipped message="Reason: empty test ('reporters/tests.R:45:1')"/>
        </testcase>
      </testsuite>
    </testsuites>

# can snapshot values

    [
      "a",
      1.5,
      1,
      true
    ]

---

    {
      "type": "list",
      "attributes": {},
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["a"]
        },
        {
          "type": "double",
          "attributes": {},
          "value": [1.5]
        },
        {
          "type": "integer",
          "attributes": {},
          "value": [1]
        },
        {
          "type": "logical",
          "attributes": {},
          "value": [true]
        }
      ]
    }

---

    list("a", 1.5, 1L, TRUE)

---

    WAoAAAACAAQCAwACAwAAAAATAAAABAAAABAAAAABAAQACQAAAAFhAAAADgAAAAE/+AAAAAAA
    AAAAAA0AAAABAAAAAQAAAAoAAAABAAAAAQ==

# can control snapshot value details

    1.2

---

    WAoAAAACAAQCAwACAwAAAAMGAAAEAgAAAAEABAAJAAAABWNsYXNzAAAAEAAAAAEABAAJAAAA
    B2Zvcm11bGEAAAQCAAAAAQAEAAkAAAAMLkVudmlyb25tZW50AAAABAAAAAAAAAAEAAAAAAAA
    AAQAAAAAAAAA+QAAAAAAAAACAAQACQAAAAh0ZXN0dGhhdAAEAAkAAAAFMy4yLjAAAAD+AAAA
    EwAAAB0AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4A
    AAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA
    /gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAAEwAAAB0AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4A
    AAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA
    /gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAAEwAAAB0A
    AAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA
    /gAAAP4AAAQCAAAAAQAEAAkAAAABZgAAAwYAAAQCAAAB/wAAABAAAAABAAQACQAAAAdmb3Jt
    dWxhAAAEAgAAAv8AAAP/AAAA/gAAAAEABAAJAAAAAX4AAAACAAAADgAAAAE/8AAAAAAAAAAA
    AP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+AAAA/gAAAP4AAAD+
    AAAA/gAAAP4AAAD+AAAI/wAAAAIAAAAOAAAAAT/wAAAAAAAAAAAA/g==

# tolerance passed to check_roundtrip

    0.9

# check_roundtrip() gives nice error

    Code
      wrapper(NULL, list(), label = "foo", style = "json")
    Condition
      Error in `wrapper()`:
      ! `foo` could not be safely serialized with `style = "json"`.
        Serializing then deserializing the object returned something new:
      
      `original` is NULL
      `new` is a list
      
      i You may need to try a different `style`.

# snapshot tests in describe

    Code
      1 + 1
    Output
      [1] 2

# snapshot tests in describe / and in it

    Code
      2 + 2
    Output
      [1] 4

# has to have a valid description for the block

    Code
      describe()
    Condition
      Error in `describe()`:
      ! `description` must be a single string, not absent.
    Code
      describe(c("a", "b"))
    Condition
      Error in `describe()`:
      ! `description` must be a single string, not a character vector.
    Code
      it()
    Condition
      Error in `it()`:
      ! `description` must be a single string, not absent.
    Code
      it(c("a", "b"))
    Condition
      Error in `it()`:
      ! `description` must be a single string, not a character vector.

# reporter basics work

    ##teamcity[testSuiteStarted name='Successes']
    ##teamcity[testSuiteStarted name='Success']
    ##teamcity[testStarted name='expectation 1']
    ##teamcity[testFinished name='expectation 1']
    ##teamcity[testSuiteFinished name='Success']
    
    ##teamcity[testSuiteFinished name='Successes']
    
    
    ##teamcity[testSuiteStarted name='Failures']
    ##teamcity[testSuiteStarted name='Failure:1']
    ##teamcity[testStarted name='expectation 1']
    ##teamcity[testFailed name='expectation 1' message='FALSE is not TRUE' details='|n`actual`:   FALSE|n`expected`: TRUE ']
    ##teamcity[testFinished name='expectation 1']
    ##teamcity[testSuiteFinished name='Failure:1']
    
    ##teamcity[testSuiteStarted name='Failure:2a']
    ##teamcity[testStarted name='expectation 1']
    ##teamcity[testFailed name='expectation 1' message='FALSE is not TRUE' details='|n`actual`:   FALSE|n`expected`: TRUE |nBacktrace:|n    x|n 1. \-f()|n 2.   \-testthat::expect_true(FALSE)']
    ##teamcity[testFinished name='expectation 1']
    ##teamcity[testSuiteFinished name='Failure:2a']
    
    ##teamcity[testSuiteFinished name='Failures']
    
    
    ##teamcity[testSuiteStarted name='Errors']
    ##teamcity[testSuiteStarted name='Error:1']
    ##teamcity[testStarted name='expectation 1']
    ##teamcity[testFailed name='expectation 1' message='Error in `eval(code, test_env)`: stop' details='']
    ##teamcity[testFinished name='expectation 1']
    ##teamcity[testSuiteFinished name='Error:1']
    
    ##teamcity[testSuiteStarted name='errors get tracebacks']
    ##teamcity[testStarted name='expectation 1']
    ##teamcity[testFailed name='expectation 1' message='Error in `h()`: !' details='Backtrace:|n    x|n 1. \-f()|n 2.   \-g()|n 3.     \-h()']
    ##teamcity[testFinished name='expectation 1']
    ##teamcity[testSuiteFinished name='errors get tracebacks']
    
    ##teamcity[testSuiteFinished name='Errors']
    
    
    ##teamcity[testSuiteStarted name='Skips']
    ##teamcity[testSuiteStarted name='explicit skips are reported']
    ##teamcity[testIgnored name='expectation 1' message='Reason: skip']
    ##teamcity[testSuiteFinished name='explicit skips are reported']
    
    ##teamcity[testSuiteStarted name='empty tests are implicitly skipped']
    ##teamcity[testIgnored name='expectation 1' message='Reason: empty test']
    ##teamcity[testSuiteFinished name='empty tests are implicitly skipped']
    
    ##teamcity[testSuiteFinished name='Skips']
    
    
    ##teamcity[testSuiteStarted name='Warnings']
    ##teamcity[testSuiteStarted name='warnings get backtraces']
    ##teamcity[testStarted name='expectation 1']
    ##teamcity[testFinished name='expectation 1']
    ##teamcity[testIgnored name='expectation 2' message='Reason: empty test']
    ##teamcity[testSuiteFinished name='warnings get backtraces']
    
    ##teamcity[testSuiteFinished name='Warnings']
    
    

# reporter basics works

    'reporters/tests.R:13:3' [failure] Failure:1. Expected `x` to be TRUE.
    'reporters/tests.R:17:8' [failure] Failure:2a. Expected FALSE to be TRUE.
    'reporters/tests.R:24:3' [error] Error:1. Error in `eval(code, test_env)`: stop
    'reporters/tests.R:30:8' [error] errors get tracebacks. Error in `h()`: !
    'reporters/tests.R:38:3' [skip] explicit skips are reported. Reason: skip
    'reporters/tests.R:41:1' [skip] empty tests are implicitly skipped. Reason: empty test
    'reporters/tests.R:47:5' [warning] warnings get backtraces. def
    'reporters/tests.R:45:1' [skip] warnings get backtraces. Reason: empty test

# reporter works

    1..9
    # Context Successes
    ok 1 Success
    # Context Failures
    not ok 2 Failure:1
      Expected `x` to be TRUE.
      Differences:
      `actual`:   FALSE
      `expected`: TRUE 
      
    not ok 3 Failure:2a
      Expected FALSE to be TRUE.
      Differences:
      `actual`:   FALSE
      `expected`: TRUE 
      
      Backtrace:
          x
       1. \-f()
       2.   \-testthat::expect_true(FALSE)
    # Context Errors
    not ok 4 Error:1
      Error in `eval(code, test_env)`: stop
    not ok 5 errors get tracebacks
      Error in `h()`: !
      Backtrace:
          x
       1. \-f()
       2.   \-g()
       3.     \-h()
    # Context Skips
    ok 6 # SKIP Reason: skip
    ok 7 # SKIP Reason: empty test
    # Context Warnings
    ok 8 # WARNING def
    Backtrace:
        x
     1. \-f()
    ok 9 # SKIP Reason: empty test

# regexp = NULL checks for presence of error

    Code
      expect_error(f())
    Condition
      Error:
      ! Expected `f()` to throw a error.

# regexp = NA checks for absence of error

    Code
      expect_error(f(), NA)
    Condition
      Error:
      ! Expected `f()` not to throw any errors.
      Actually got a <simpleError> with message:
        Yes

# regexp = string matches for error message

    Code
      expect_error(f(), "No")
    Condition
      Error:
      ! Expected `f()` to throw a error.

# expect_error validates its inputs

    Code
      expect_error(stop("!"), regexp = 1)
    Condition
      Error in `expect_error()`:
      ! `regexp` must be a single string, `NA`, or `NULL`, not the number 1.
    Code
      expect_error(stop("!"), class = 1)
    Condition
      Error in `expect_error()`:
      ! `class` must be a single string or `NULL`, not the number 1.
    Code
      expect_error(stop("!"), inherit = "yes")
    Condition
      Error in `expect_error()`:
      ! `inherit` must be `TRUE` or `FALSE`, not the string "yes".

# message method is called when expecting error

    Code
      expect_error(fb(), NA)
    Condition
      Error:
      ! Expected `fb()` not to throw any errors.
      Actually got a <foobar> with message:
        dispatched!

# expect_warning validates its inputs

    Code
      expect_warning(warning("!"), regexp = 1)
    Condition
      Error in `expect_warning()`:
      ! `regexp` must be a single string, `NA`, or `NULL`, not the number 1.
    Code
      expect_warning(warning("!"), class = 1)
    Condition
      Error in `expect_warning()`:
      ! `class` must be a single string or `NULL`, not the number 1.
    Code
      expect_warning(warning("!"), inherit = "yes")
    Condition
      Error in `expect_warning()`:
      ! `inherit` must be `TRUE` or `FALSE`, not the string "yes".
    Code
      expect_warning(warning("!"), all = "yes")
    Condition
      Error in `expect_warning()`:
      ! `all` must be `TRUE` or `FALSE`, not the string "yes".

# regexp = NA checks for absence of message

    Code
      expect_message(f(), NA)
    Condition
      Error:
      ! Expected `f()` not to throw any messages.
      Actually got a <simpleMessage> with message:
        !
        

# expect_message validates its inputs

    Code
      expect_message(message("!"), regexp = 1)
    Condition
      Error in `expect_message()`:
      ! `regexp` must be a single string, `NA`, or `NULL`, not the number 1.
    Code
      expect_message(message("!"), class = 1)
    Condition
      Error in `expect_message()`:
      ! `class` must be a single string or `NULL`, not the number 1.
    Code
      expect_message(message("!"), inherit = "yes")
    Condition
      Error in `expect_message()`:
      ! `inherit` must be `TRUE` or `FALSE`, not the string "yes".
    Code
      expect_message(message("!"), all = "yes")
    Condition
      Error in `expect_message()`:
      ! `all` must be `TRUE` or `FALSE`, not the string "yes".

# condition class is included in failure

    Code
      expect_condition(f1(), class = "bar")
    Condition
      Error:
      ! Expected `f1()` to throw a condition with class <bar>.

# expect_condition validates its inputs

    Code
      expect_condition(stop("!"), regexp = 1)
    Condition
      Error in `expect_condition()`:
      ! `regexp` must be a single string, `NA`, or `NULL`, not the number 1.
    Code
      expect_condition(stop("!"), class = 1)
    Condition
      Error in `expect_condition()`:
      ! `class` must be a single string or `NULL`, not the number 1.
    Code
      expect_condition(stop("!"), inherit = "yes")
    Condition
      Error in `expect_condition()`:
      ! `inherit` must be `TRUE` or `FALSE`, not the string "yes".

# unused arguments generate an error

    Code
      expect_condition(stop("Hi!"), foo = "bar")
    Condition
      Error in `expect_condition()`:
      ! Can't supply `...` unless `regexp` is set.
      * Unused arguments: `foo`.
      i Did you mean to use `regexp` so `...` is passed to `grepl()`?
    Code
      expect_condition(stop("Hi!"), , , "bar")
    Condition
      Error in `expect_condition()`:
      ! Can't supply `...` unless `regexp` is set.
      * Unused arguments: `..1`.
      i Did you mean to use `regexp` so `...` is passed to `grepl()`?
    Code
      expect_condition(stop("Hi!"), , , "bar", fixed = TRUE)
    Condition
      Error in `expect_condition()`:
      ! Can't supply `...` unless `regexp` is set.
      * Unused arguments: `..1` and `fixed`.
      i Did you mean to use `regexp` so `...` is passed to `grepl()`?
    Code
      expect_condition(stop("Hi!"), "x", foo = "bar")
    Condition
      Error in `expect_condition()`:
      ! Failed to compare condition to `regexp`.
      Caused by error in `grepl()`:
      ! unused argument (foo = "bar")
    Code
      expect_condition(stop("Hi!"), pattern = "bar", fixed = TRUE)
    Condition
      Error in `expect_condition()`:
      ! Can't supply `...` unless `regexp` is set.
      * Unused arguments: `pattern` and `fixed`.
      i Did you mean to use `regexp` so `...` is passed to `grepl()`?

# expect = NA checks for no output

    Code
      expect_output(g(), NA)
    Condition
      Error:
      ! Expected `g()` to produce no output.
      Actual output:
      !

# expect = NULL checks for some output

    Code
      expect_output(f(), NULL)
    Condition
      Error:
      ! Expected `f()` to produce output.

# expect = string checks for match

    Code
      expect_output(g(), "x")
    Condition
      Error:
      ! Expected output from `g()` to match regexp "x".
      Actual output:
      x | !

---

    Code
      expect_output("a", "x")
    Condition
      Error:
      ! Expected "a" to produce output.

# expect_output validates its inputs

    Code
      expect_output(cat("hello"), "hello", width = "wide")
    Condition
      Error in `expect_output()`:
      ! `width` must be a whole number, not the string "wide".

# works in parallel

    Code
      results[, c(1:8, 12:13)]
    Output
        file context test nb failed skipped error warning passed         result
      1   f1          t11  2      0   FALSE FALSE       0      2 msg111, msg112
      2   f2          t21  2      0   FALSE FALSE       0      2 msg211, msg212
      3   f2          t22  1      0    TRUE FALSE       0      0        skip221

# basic skips work as expected

    Reason: Skipping

---

    Reason: TRUE is TRUE

---

    Reason: FALSE is not TRUE

---

    Reason: empty test

# autogenerated message is always single line

    Reason: a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name || a_very_long_argument_name is not TRUE

# skip_if_not_installed() works as expected, absence and version

    Reason: {doesntexist} is not installed

---

    Reason: Installed testthat is version 3.0.0; but 9999.9999.999 is required

# skip_if_not_installed() works as expected, offline

    Reason: offline

# skip_on_ci() works as expected

    Reason: On CI

# skip_on_covr() works as expected

    Reason: On covr

# skip_on_bioc() works as expected

    Reason: On Bioconductor

# superseded CI skips still work

    Reason: On Travis

---

    Reason: On Appveyor

# skip_if_translated() works as expected

    Reason: "'%s' not found" is translated

# skip on os checks os names

    Code
      skip_on_os("amiga")
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of "windows", "mac", "linux", "solaris", "emscripten"

# can skip on multiple oses

    Reason: On Windows

---

    Reason: On Windows

# can refine os with arch

    Reason: On Windows

---

    Reason: On Windows i386

# skip_unless_r works as expected

    Code
      skip_unless_r("idfjdij")
    Condition
      Error in `skip_unless_r()`:
      ! `spec` must be an valid version specification, like ">= 4.0.0", not "idfjdij".

# skip_unless_r gives the expected output

    Reason: Current R version (4.5.0) does not satisfy requirement (>= 999.999.999)

---

    Reason: Current R version (4.5.0) does not satisfy requirement (== 0.0.0)

# captures error before first test

    v | F W  S  OK | Context
    
    / |          0 | reporters/error-setup                                          
    - | 1        0 | reporters/error-setup                                          
    x | 1        0 | reporters/error-setup
    --------------------------------------------------------------------------------
    Error ('reporters/error-setup.R:3:6'): (code run outside of `test_that()`)
    Error in `h()`: !
    Backtrace:
        x
     1. +-testthat::setup(f())
     2. | \-rlang::eval_tidy(enquo(code), env = env)
     3. \-f()
     4.   \-g()
     5.     \-h()
    --------------------------------------------------------------------------------
    
    == Results =====================================================================
    -- Failed tests ----------------------------------------------------------------
    Error ('reporters/error-setup.R:3:6'): (code run outside of `test_that()`)
    Error in `h()`: !
    Backtrace:
        x
     1. +-testthat::setup(f())
     2. | \-rlang::eval_tidy(enquo(code), env = env)
     3. \-f()
     4.   \-g()
     5.     \-h()
    
    [ FAIL 1 | WARN 0 | SKIP 0 | PASS 0 ]
    
    I believe in you!

# gracefully handles multiple contexts

    v | F W  S  OK | Context
    
    / |          0 | reporters/context                                              
    / |          0 | my context                                                     
    - |          1 | my context                                                     
    v |          1 | my context
    
    == Results =====================================================================
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 1 ]
    
    You rock!

# fails after max_fail tests

    v | F W  S  OK | Context
    
    / |          0 | reporters/fail-many                                            
    - | 1        0 | reporters/fail-many                                            
    \ | 2        0 | reporters/fail-many                                            
    | | 3        0 | reporters/fail-many                                            
    / | 4        0 | reporters/fail-many                                            
    - | 5        0 | reporters/fail-many                                            
    \ | 6        0 | reporters/fail-many                                            
    | | 7        0 | reporters/fail-many                                            
    / | 8        0 | reporters/fail-many                                            
    - | 9        0 | reporters/fail-many                                            
    \ | 10        0 | reporters/fail-many                                           
    | | 11        0 | reporters/fail-many                                           
    x | 11        0 | reporters/fail-many
    --------------------------------------------------------------------------------
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    --------------------------------------------------------------------------------
    Maximum number of failures exceeded; quitting at end of file.
    i Increase this number with (e.g.) `testthat::set_max_fails(Inf)` 
    
    == Results =====================================================================
    -- Failed tests ----------------------------------------------------------------
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    Failure ('reporters/fail-many.R:3:5'): Example
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    [ FAIL 11 | WARN 0 | SKIP 0 | PASS 0 ]
    == Terminated early ============================================================
    
    I believe in you!

# can fully suppress incremental updates

    v | F W  S  OK | Context
    
    / |          0 | reporters/successes                                            
    - |          1 | reporters/successes                                            
    \ |          2 | reporters/successes                                            
    | |          3 | reporters/successes                                            
    / |          4 | reporters/successes                                            
    - |          5 | reporters/successes                                            
    \ |          6 | reporters/successes                                            
    | |          7 | reporters/successes                                            
    v |          7 | reporters/successes
    
    == Results =====================================================================
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 7 ]
    
    You rock!

---

    v | F W  S  OK | Context
    
    v |          7 | reporters/successes
    
    == Results =====================================================================
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 7 ]
    
    You rock!

# reports backtraces

    v | F W  S  OK | Context
    
    / |          0 | reporters/backtraces                                           
    - | 1        0 | reporters/backtraces                                           
    \ | 2        0 | reporters/backtraces                                           
    | | 3        0 | reporters/backtraces                                           
    / | 4        0 | reporters/backtraces                                           
    - | 5        0 | reporters/backtraces                                           
    \ | 6        0 | reporters/backtraces                                           
    | | 6 1      0 | reporters/backtraces                                           
    / | 6 1      1 | reporters/backtraces                                           
    - | 7 1      1 | reporters/backtraces                                           
    \ | 8 1      1 | reporters/backtraces                                           
    | | 9 1      1 | reporters/backtraces                                           
    x | 9 1      1 | reporters/backtraces
    --------------------------------------------------------------------------------
    Error ('reporters/backtraces.R:5:8'): errors thrown at block level are entraced
    Error in `g()`: foo
    Backtrace:
        x
     1. \-f()
     2.   \-g()
    
    Error ('reporters/backtraces.R:10:10'): errors thrown from a quasi-labelled argument are entraced
    Error in `foo()`: foo
    Backtrace:
        x
     1. +-testthat::expect_s3_class(foo(), "foo")
     2. | \-testthat::quasi_label(enquo(object), arg = "object")
     3. |   \-rlang::eval_bare(expr, quo_get_env(quo))
     4. \-foo()
    
    Error ('reporters/backtraces.R:15:10'): errors thrown from a quasi-labelled argument are entraced (deep case)
    Error in `foo()`: foo
    Backtrace:
        x
     1. +-testthat::expect_s3_class(f(), "foo")
     2. | \-testthat::quasi_label(enquo(object), arg = "object")
     3. |   \-rlang::eval_bare(expr, quo_get_env(quo))
     4. \-f()
     5.   \-g()
     6.     +-testthat::expect_s3_class(foo(), "foo")
     7.     | \-testthat::quasi_label(enquo(object), arg = "object")
     8.     |   \-rlang::eval_bare(expr, quo_get_env(quo))
     9.     \-foo()
    
    Error ('reporters/backtraces.R:23:10'): errors thrown from a quasi-labelled argument are entraced (deep deep case)
    Error in `bar()`: foobar
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     +-testthat::expect_s3_class(foo(), "foo")
     4.     | \-testthat::quasi_label(enquo(object), arg = "object")
     5.     |   \-rlang::eval_bare(expr, quo_get_env(quo))
     6.     \-foo()
     7.       \-bar()
    
    Error ('reporters/backtraces.R:34:16'): failed expect_error() prints a backtrace
    Error in `signaller()`: bar
    Backtrace:
        x
     1. +-testthat::expect_error(f(), "foo")
     2. | \-testthat:::expect_condition_matching(...)
     3. |   \-testthat:::quasi_capture(...)
     4. |     +-testthat (local) .capture(...)
     5. |     | \-base::withCallingHandlers(...)
     6. |     \-rlang::eval_bare(quo_get_expr(.quo), quo_get_env(.quo))
     7. \-f()
     8.   \-signaller()
    
    Error ('reporters/backtraces.R:43:3'): Errors are inspected with `conditionMessage()`
    <foobar/rlang_error/error/condition>
    Error in `eval(code, test_env)`: dispatched
    Backtrace:
        x
     1. \-rlang::abort("Wrong message", "foobar")
    
    Warning ('reporters/backtraces.R:48:10'): also get backtraces for warnings
    foobar
    Backtrace:
        x
     1. \-foo()
     2.   \-bar()
    
    Error ('reporters/backtraces.R:56:5'): deep stacks are shown
    Error in `f(x - 1)`: This is deep
    Backtrace:
         x
      1. \-f(25)
      2.   \-f(x - 1)
      3.     \-f(x - 1)
      4.       \-f(x - 1)
      5.         \-f(x - 1)
      6.           \-f(x - 1)
      7.             \-f(x - 1)
      8.               \-f(x - 1)
      9.                 \-f(x - 1)
     10.                   \-f(x - 1)
     11.                     \-f(x - 1)
     12.                       \-f(x - 1)
     13.                         \-f(x - 1)
     14.                           \-f(x - 1)
     15.                             \-f(x - 1)
     16.                               \-f(x - 1)
     17.                                 \-f(x - 1)
     18.                                   \-f(x - 1)
     19.                                     \-f(x - 1)
     20.                                       \-f(x - 1)
     21.                                         \-f(x - 1)
     22.                                           \-f(x - 1)
     23.                                             \-f(x - 1)
     24.                                               \-f(x - 1)
     25.                                                 \-f(x - 1)
     26.                                                   \-f(x - 1)
    
    Failure ('reporters/backtraces.R:64:6'): (code run outside of `test_that()`)
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
     4.       \-testthat::expect_true(FALSE)
    
    Failure ('reporters/backtraces.R:69:3'): nested expectations get backtraces
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
     4.       \-testthat::expect_true(FALSE)
    --------------------------------------------------------------------------------
    
    == Results =====================================================================
    -- Failed tests ----------------------------------------------------------------
    Error ('reporters/backtraces.R:5:8'): errors thrown at block level are entraced
    Error in `g()`: foo
    Backtrace:
        x
     1. \-f()
     2.   \-g()
    
    Error ('reporters/backtraces.R:10:10'): errors thrown from a quasi-labelled argument are entraced
    Error in `foo()`: foo
    Backtrace:
        x
     1. +-testthat::expect_s3_class(foo(), "foo")
     2. | \-testthat::quasi_label(enquo(object), arg = "object")
     3. |   \-rlang::eval_bare(expr, quo_get_env(quo))
     4. \-foo()
    
    Error ('reporters/backtraces.R:15:10'): errors thrown from a quasi-labelled argument are entraced (deep case)
    Error in `foo()`: foo
    Backtrace:
        x
     1. +-testthat::expect_s3_class(f(), "foo")
     2. | \-testthat::quasi_label(enquo(object), arg = "object")
     3. |   \-rlang::eval_bare(expr, quo_get_env(quo))
     4. \-f()
     5.   \-g()
     6.     +-testthat::expect_s3_class(foo(), "foo")
     7.     | \-testthat::quasi_label(enquo(object), arg = "object")
     8.     |   \-rlang::eval_bare(expr, quo_get_env(quo))
     9.     \-foo()
    
    Error ('reporters/backtraces.R:23:10'): errors thrown from a quasi-labelled argument are entraced (deep deep case)
    Error in `bar()`: foobar
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     +-testthat::expect_s3_class(foo(), "foo")
     4.     | \-testthat::quasi_label(enquo(object), arg = "object")
     5.     |   \-rlang::eval_bare(expr, quo_get_env(quo))
     6.     \-foo()
     7.       \-bar()
    
    Error ('reporters/backtraces.R:34:16'): failed expect_error() prints a backtrace
    Error in `signaller()`: bar
    Backtrace:
        x
     1. +-testthat::expect_error(f(), "foo")
     2. | \-testthat:::expect_condition_matching(...)
     3. |   \-testthat:::quasi_capture(...)
     4. |     +-testthat (local) .capture(...)
     5. |     | \-base::withCallingHandlers(...)
     6. |     \-rlang::eval_bare(quo_get_expr(.quo), quo_get_env(.quo))
     7. \-f()
     8.   \-signaller()
    
    Error ('reporters/backtraces.R:43:3'): Errors are inspected with `conditionMessage()`
    <foobar/rlang_error/error/condition>
    Error in `eval(code, test_env)`: dispatched
    Backtrace:
        x
     1. \-rlang::abort("Wrong message", "foobar")
    
    Error ('reporters/backtraces.R:56:5'): deep stacks are shown
    Error in `f(x - 1)`: This is deep
    Backtrace:
         x
      1. \-f(25)
      2.   \-f(x - 1)
      3.     \-f(x - 1)
      4.       \-f(x - 1)
      5.         \-f(x - 1)
      6.           \-f(x - 1)
      7.             \-f(x - 1)
      8.               \-f(x - 1)
      9.                 \-f(x - 1)
     10.                   \-f(x - 1)
     11.                     \-f(x - 1)
     12.                       \-f(x - 1)
     13.                         \-f(x - 1)
     14.                           \-f(x - 1)
     15.                             \-f(x - 1)
     16.                               \-f(x - 1)
     17.                                 \-f(x - 1)
     18.                                   \-f(x - 1)
     19.                                     \-f(x - 1)
     20.                                       \-f(x - 1)
     21.                                         \-f(x - 1)
     22.                                           \-f(x - 1)
     23.                                             \-f(x - 1)
     24.                                               \-f(x - 1)
     25.                                                 \-f(x - 1)
     26.                                                   \-f(x - 1)
    
    Failure ('reporters/backtraces.R:64:6'): (code run outside of `test_that()`)
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
     4.       \-testthat::expect_true(FALSE)
    
    Failure ('reporters/backtraces.R:69:3'): nested expectations get backtraces
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
     4.       \-testthat::expect_true(FALSE)
    
    [ FAIL 9 | WARN 1 | SKIP 0 | PASS 1 ]
    
    I believe in you!

# records skips

    v | F W  S  OK | Context
    
    / |          0 | reporters/skips                                                
    - |      1   0 | reporters/skips                                                
    \ |      2   0 | reporters/skips                                                
    v |      2   0 | reporters/skips
    
    == Results =====================================================================
    -- Skipped tests (2) -----------------------------------------------------------
    * longer skip (1): 'reporters/skips.R:6:3'
    * regular skip (1): 'reporters/skips.R:2:3'
    
    [ FAIL 0 | WARN 0 | SKIP 2 | PASS 0 ]
    
    You rock!

# compact display is informative

    
    == Testing reporters/tests.R ===================================================
    
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 0 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 0 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 1 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 2 | WARN 0 | SKIP 0 | PASS 1 ]
    
    -- Failure ('reporters/tests.R:12:3'): Failure:1 -------------------------------
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    
    -- Failure ('reporters/tests.R:16:8'): Failure:2a ------------------------------
    FALSE is not TRUE
    
    `actual`:   FALSE
    `expected`: TRUE 
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
    
    [ FAIL 2 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 3 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 4 | WARN 0 | SKIP 0 | PASS 1 ]
    
    -- Error ('reporters/tests.R:23:3'): Error:1 -----------------------------------
    Error in `eval(code, test_env)`: stop
    
    -- Error ('reporters/tests.R:29:8'): errors get tracebacks ---------------------
    Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
    
    [ FAIL 4 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 4 | WARN 0 | SKIP 1 | PASS 1 ]
    [ FAIL 4 | WARN 0 | SKIP 2 | PASS 1 ]
    [ FAIL 4 | WARN 0 | SKIP 2 | PASS 1 ]
    [ FAIL 4 | WARN 1 | SKIP 2 | PASS 1 ]
    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]
    
    -- Warning ('reporters/tests.R:47:5'): warnings get backtraces -----------------
    def
    Backtrace:
        x
     1. \-f()
    
    -- Skipped tests (3) -----------------------------------------------------------
    * empty test (2): 'reporters/tests.R:40:1', 'reporters/tests.R:45:1'
    * skip (1): 'reporters/tests.R:37:3'
    
    
    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]

# display of successes only is compact

    
    == Testing reporters/successes.R ===============================================
    
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 0 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 2 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 3 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 4 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 5 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 6 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 7 ] Done!

---

    
    == Testing reporters/skips.R ===================================================
    
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 0 ]
    [ FAIL 0 | WARN 0 | SKIP 1 | PASS 0 ]
    [ FAIL 0 | WARN 0 | SKIP 2 | PASS 0 ]
    
    -- Skipped tests (2) -----------------------------------------------------------
    * longer skip (1): 'reporters/skips.R:6:3'
    * regular skip (1): 'reporters/skips.R:2:3'
    

---

    
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 0 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 1 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 2 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 3 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 4 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 5 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 6 ]
    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 7 ]

# expected_named verifies presence of names

    Code
      expect_named(x)
    Condition
      Error:
      ! Expected `x` to have names.

# expected_named verifies actual of names

    Code
      expect_named(x, "b")
    Condition
      Error:
      ! Expected `x` to have names "b".
      Differences:
      `actual`:   "a"
      `expected`: "b"

# provide useful feedback on failure

    Code
      expect_named(x1, c("a", "b"), ignore.order = TRUE)
    Condition
      Error:
      ! Expected `x1` to have names `c("a", "b")`.
      Actual: "a"
      Expected: "a", "b"
      Absent: "b"

---

    Code
      expect_named(x2, "a", ignore.order = TRUE)
    Condition
      Error:
      ! Expected `x2` to have names "a".
      Actual: "a", "b"
      Expected: "a"
      Needs: "b"

---

    Code
      expect_named(x1, "b", ignore.order = TRUE)
    Condition
      Error:
      ! Expected `x1` to have names "b".
      Actual: "a"
      Expected: "b"
      Needs: "a"
      Absent: "b"

---

    Code
      expect_named(x1, c("a", "b"), ignore.order = FALSE)
    Condition
      Error:
      ! Expected `x1` to have names `c("a", "b")`.
      Differences:
      `actual`:   "a"    
      `expected`: "a" "b"

---

    Code
      expect_named(x2, "a", ignore.order = FALSE)
    Condition
      Error:
      ! Expected `x2` to have names "a".
      Differences:
      `actual`:   "a" "b"
      `expected`: "a"    

---

    Code
      expect_named(x1, "b", ignore.order = FALSE)
    Condition
      Error:
      ! Expected `x1` to have names "b".
      Differences:
      `actual`:   "a"
      `expected`: "b"

# expect_named validates its inputs

    Code
      expect_named(c(a = 1), "a", ignore.order = "yes")
    Condition
      Error in `expect_named()`:
      ! `ignore.order` must be `TRUE` or `FALSE`, not the string "yes".
    Code
      expect_named(c(a = 1), "a", ignore.case = "yes")
    Condition
      Error in `expect_named()`:
      ! `ignore.case` must be `TRUE` or `FALSE`, not the string "yes".

# get_routine() fails when no routine exists

    Code
      get_routine("utils", "no_such_routine")
    Condition
      Error in `get_routine()`:
      ! Failed to locate routine `no_such_routine` in package utils.

# validates inputs

    Code
      expect_cpp_tests_pass(123)
    Condition
      Error in `expect_cpp_tests_pass()`:
      ! `package` must be a single string, not the number 123.
    Code
      run_cpp_tests(123)
    Condition
      Error in `run_cpp_tests()`:
      ! `package` must be a single string, not the number 123.

# useful messaging

    Code
      use_catch(path)
    Message
      v Added C++ unit testing infrastructure.
      i Please ensure you have LinkingTo: testthat in your DESCRIPTION.
      i Please ensure you have Suggests: xml2 in your DESCRIPTION.

# with_mocked_bindings() validates its inputs

    Code
      with_mocked_bindings(1 + 1, function() 2)
    Condition
      Error in `local_mocked_bindings()`:
      ! All elements of `...` must be named.

# can't mock bindings that don't exist

    Code
      local_mocked_bindings(f = function() "x")
    Condition
      Error in `local_mocked_bindings()`:
      ! Can't find binding for `f`

# produces consistent output

    1: expect_true(x)
    2: expect_waldo_constant_(act, exp, info = info, ignore_attr = TRUE)
    3: fail(msg, info = info, trace_env = trace_env)
    
    1: f()
    2: expect_true(FALSE)
    3: expect_waldo_constant_(act, exp, info = info, ignore_attr = TRUE)
    4: fail(msg, info = info, trace_env = trace_env)
    
    1: stop("stop")
    
    1: f()
    2: g()
    3: h()
    4: stop("!")
    
    1: skip("skip")
    
    1: f()
    2: warning("def")
    

# produces useful output

    Code
      with_reporter("stop", run_tests())
    Output
      Test passed with 1 success.
      -- Failure: Failure:1 ----------------------------------------------------------
      Expected `x` to be TRUE.
      Differences:
      `actual`:   FALSE
      `expected`: TRUE 
      
    Condition
      Error:
      ! Test failed with 1 failure and 0 successes.

# works nicely with nested tests

    Code
      with_reporter("stop", run_tests())
    Output
      Test passed with 2 successes.
      -- Failure: failed then succeeded / failed-1 -----------------------------------
      Expected FALSE to be TRUE.
      Differences:
      `actual`:   FALSE
      `expected`: TRUE 
      
      -- Failure: failed then succeeded / failed-2 -----------------------------------
      Expected FALSE to be TRUE.
      Differences:
      `actual`:   FALSE
      `expected`: TRUE 
      
    Condition
      Error:
      ! Test failed with 2 failures and 1 success.

# errors when needed

    Code
      r$end_test()
    Condition
      Error:
      ! Test failed with 1 failure and 0 successes.

# snapshot cleanup makes nice message if needed

    Code
      snapshot_cleanup(dir)
    Message
      Deleting unused snapshots: 'a.md' and 'b.md'
    Code
      snapshot_cleanup(dir, c("a", "b"))

# mock_output_sequence() works

    Code
      mocked_sequence()
    Condition
      Error in `mocked_sequence()`:
      ! Can't find value for 4th iteration.
      i `...` has only 3 values.
      i You can set `recycle` to `TRUE`.

# expect_failure() generates a useful error messages

    Code
      expect_failure(expect_no_failure())
    Condition
      Error:
      ! Expected 0 successes and 1 failure.
      v Observed 0 successes.
      x Observed 0 failures.
    Code
      expect_failure(expect_many_failures())
    Condition
      Error:
      ! Expected 0 successes and 1 failure.
      v Observed 0 successes.
      x Observed 2 failures.
    Code
      expect_failure(expect_has_success())
    Condition
      Error:
      ! Expected 0 successes and 1 failure.
      x Observed 1 success.
      v Observed 1 failure.
    Code
      expect_failure(expect_both_wrong())
    Condition
      Error:
      ! Expected 0 successes and 1 failure.
      x Observed 1 success.
      x Observed 0 failures.
    Code
      expect_failure(expect_failure_foo(), "bar")
    Condition
      Error:
      ! Expected failure message to match regexp "bar".
      Actual message:
      x | foo

# errors in expect_success bubble up

    Code
      expect_success(abort("error"))
    Condition
      Error:
      ! error

# show_failure

    Code
      show_failure(expect_true(FALSE))
    Output
      Failed expectation:
      Expected FALSE to be TRUE.
      Differences:
      `actual`:   FALSE
      `expected`: TRUE 
      

# expect_success() generates a useful error messages

    Code
      expect_success(expect_no_success())
    Condition
      Error:
      ! Expected 1 success and 0 failures.
      x Observed 0 successes.
      v Observed 0 failures.
    Code
      expect_success(expect_many_successes())
    Condition
      Error:
      ! Expected 1 success and 0 failures.
      x Observed 2 successes.
      v Observed 0 failures.
    Code
      expect_success(expect_has_failure())
    Condition
      Error:
      ! Expected 1 success and 0 failures.
      v Observed 1 success.
      x Observed 1 failure.
    Code
      expect_success(expect_both_wrong())
    Condition
      Error:
      ! Expected 1 success and 0 failures.
      x Observed 0 successes.
      x Observed 1 failure.

# expect_no are deprecated

    Code
      expect_no_failure(pass())
    Condition
      Warning:
      `expect_no_failure()` was deprecated in testthat 3.3.0.
      i Please use `expect_success()` instead.
    Code
      expect_no_success(fail())
    Condition
      Warning:
      `expect_no_success()` was deprecated in testthat 3.3.0.
      i Please use `expect_failure()` instead.

# reporter as expected

    Start test: Success
      'reporters/tests.R:6:3' [success]
    End test: Success
    
    Start test: Failure:1
      'reporters/tests.R:13:3' [failure]
    End test: Failure:1
    
    Start test: Failure:2a
      'reporters/tests.R:17:8' [failure]
    End test: Failure:2a
    
    Start test: Error:1
      'reporters/tests.R:24:3' [error]
    End test: Error:1
    
    Start test: errors get tracebacks
      'reporters/tests.R:30:8' [error]
    End test: errors get tracebacks
    
    Start test: explicit skips are reported
      'reporters/tests.R:38:3' [skip]
    End test: explicit skips are reported
    
    Start test: empty tests are implicitly skipped
      'reporters/tests.R:41:1' [skip]
    End test: empty tests are implicitly skipped
    
    Start test: warnings get backtraces
      'reporters/tests.R:47:5' [warning]
      'reporters/tests.R:45:1' [skip]
    End test: warnings get backtraces
    

# expect_no_* conditions behave as expected

    Code
      expect_no_error(stop("error"))
    Condition
      Error:
      ! Expected `stop("error")` not to throw any errors.
      Actually got a <simpleError> with message:
        error

---

    Code
      expect_no_warning(warning("warning"))
    Condition
      Error:
      ! Expected `warning("warning")` not to throw any warnings.
      Actually got a <simpleWarning> with message:
        warning

---

    Code
      expect_no_message(message("message"))
    Condition
      Error:
      ! Expected `message("message")` not to throw any messages.
      Actually got a <simpleMessage> with message:
        message
        

---

    Code
      expect_no_error(abort("error"))
    Condition
      Error:
      ! Expected `abort("error")` not to throw any errors.
      Actually got a <rlang_error> with message:
        error

---

    Code
      expect_no_warning(warn("warning"))
    Condition
      Error:
      ! Expected `warn("warning")` not to throw any warnings.
      Actually got a <rlang_warning> with message:
        warning

---

    Code
      expect_no_message(inform("message"))
    Condition
      Error:
      ! Expected `inform("message")` not to throw any messages.
      Actually got a <rlang_message> with message:
        message

# expect_no_* don't emit success when they fail

    Code
      expect_no_error(stop("!"))
    Condition
      Error:
      ! Expected `stop("!")` not to throw any errors.
      Actually got a <simpleError> with message:
        !

# matched conditions give informative message

    Code
      expect_no_warning(foo())
    Condition
      Error:
      ! Expected `foo()` not to throw any warnings.
      Actually got a <test> with message:
        This is a problem!
    Code
      expect_no_warning(foo(), message = "problem")
    Condition
      Error:
      ! Expected `foo()` not to throw any warnings matching pattern 'problem'.
      Actually got a <test> with message:
        This is a problem!
    Code
      expect_no_warning(foo(), class = "test")
    Condition
      Error:
      ! Expected `foo()` not to throw any warnings of class 'test'.
      Actually got a <test> with message:
        This is a problem!
    Code
      expect_no_warning(foo(), message = "problem", class = "test")
    Condition
      Error:
      ! Expected `foo()` not to throw any warnings of class 'test' matching pattern 'problem'.
      Actually got a <test> with message:
        This is a problem!

# logical tests act as expected

    Code
      expect_true(df)
    Condition
      Error:
      ! Expected `df` to be TRUE.
      Differences:
      `actual` is an S3 object of class <data.frame>, a list
      `expected` is a logical vector (TRUE)

---

    Code
      expect_false(df)
    Condition
      Error:
      ! Expected `df` to be FALSE.
      Differences:
      `actual` is an S3 object of class <data.frame>, a list
      `expected` is a logical vector (FALSE)

# expect_null works

    Code
      expect_null(df)
    Condition
      Error:
      ! Expected `df` to be NULL.
      Differences:
      `actual` is an S3 object of class <data.frame>, a list
      `expected` is NULL

# useful error message if can't find reporter

    Code
      find_reporter(1)
    Condition
      Error in `find_reporter()`:
      ! `reporter` must be a string, a character vector, a reporter object, or a reporter class, not the number 1.
    Code
      find_reporter("blah")
    Condition
      Error in `find_reporter()`:
      ! Cannot find test reporter `blah`.
    Code
      find_reporter(c("summary", "blah"))
    Condition
      Error in `find_reporter()`:
      ! Cannot find test reporter `blah`.

# basic report works

    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]
    
    == Skipped tests (3) ===========================================================
    * empty test (2): 'reporters/tests.R:41:1', 'reporters/tests.R:45:1'
    * skip (1): 'reporters/tests.R:38:3'
    
    == Warnings ====================================================================
    -- Warning ('reporters/tests.R:47:5'): warnings get backtraces -----------------
    def
    Backtrace:
        x
     1. \-f()
    
    == Failed tests ================================================================
    -- Failure ('reporters/tests.R:13:3'): Failure:1 -------------------------------
    Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    -- Failure ('reporters/tests.R:17:8'): Failure:2a ------------------------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
    -- Error ('reporters/tests.R:24:3'): Error:1 -----------------------------------
    Error in `eval(code, test_env)`: stop
    -- Error ('reporters/tests.R:30:8'): errors get tracebacks ---------------------
    Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
    
    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]

# doesn't truncate long lines

    [ FAIL 1 | WARN 0 | SKIP 0 | PASS 0 ]
    
    == Failed tests ================================================================
    -- Failure ('reporters/long-test.R:2:3'): That very long test messages are not truncated because they contain useful information that you probably want to read --
    Failure has been forced
    
    [ FAIL 1 | WARN 0 | SKIP 0 | PASS 0 ]

# always shows summary

    [ FAIL 0 | WARN 0 | SKIP 0 | PASS 7 ]

# shows warnings when not on CRAN

    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]
    
    == Skipped tests (3) ===========================================================
    * empty test (2): 'reporters/tests.R:41:1', 'reporters/tests.R:45:1'
    * skip (1): 'reporters/tests.R:38:3'
    
    == Warnings ====================================================================
    -- Warning ('reporters/tests.R:47:5'): warnings get backtraces -----------------
    def
    Backtrace:
        x
     1. \-f()
    
    == Failed tests ================================================================
    -- Failure ('reporters/tests.R:13:3'): Failure:1 -------------------------------
    Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    -- Failure ('reporters/tests.R:17:8'): Failure:2a ------------------------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
    -- Error ('reporters/tests.R:24:3'): Error:1 -----------------------------------
    Error in `eval(code, test_env)`: stop
    -- Error ('reporters/tests.R:30:8'): errors get tracebacks ---------------------
    Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
    
    [ FAIL 4 | WARN 1 | SKIP 3 | PASS 1 ]

# generates informative snapshot hints

    Code
      base::writeLines(snapshot_check_hint())
    Output
      To review and process snapshots locally:
      * Locate check directory.
      * Copy 'tests/testthat/_snaps' to local package.
      * Run `testthat::snapshot_accept()` to accept all changes.
      * Run `testthat::snapshot_review()` to review all changes.

---

    Code
      base::writeLines(snapshot_check_hint())
    Output
      To review and process snapshots locally:
      * Download and unzip artifact.
      * Copy 'tests/testthat/_snaps' to local package.
      * Run `testthat::snapshot_accept()` to accept all changes.
      * Run `testthat::snapshot_review()` to review all changes.

---

    Code
      base::writeLines(snapshot_check_hint())
    Output
      To review and process snapshots locally:
      * Run `testthat::snapshot_download_gh("r-lib/testthat", "123")` to download snapshots.
      * Run `testthat::snapshot_accept()` to accept all changes.
      * Run `testthat::snapshot_review()` to review all changes.

# basic principles of visibility hold

    Code
      expect_invisible(x)
    Condition
      Error:
      ! Expected `x` to return invisibly.
      Actual visibility: visible.

---

    Code
      expect_visible(x <- 1)
    Condition
      Error:
      ! Expected `x <- 1` to return visibly.
      Actual visibility: invisible.

# generates useful failure messages

    Code
      expect_visible(invisible(1))
    Condition
      Error:
      ! Expected `invisible(1)` to return visibly.
      Actual visibility: invisible.

---

    Code
      expect_invisible(1)
    Condition
      Error:
      ! Expected 1 to return invisibly.
      Actual visibility: visible.

# expect_snapshot_file finds duplicate snapshot files

    Code
      expect_snapshot_file(write_tmp_lines(r_version()), "version.txt", variant = r_version())
    Condition
      Error in `expect_snapshot_file()`:
      ! Snapshot file names must be unique. "version.txt" has already been used.

# warns on first creation

    Code
      out <- snapshot_file_equal_(path)
    Condition
      Warning:
      Adding new file snapshot: 'tests/testthat/_snaps/my-test/test.txt'

---

    Code
      snapshot_file_equal_("doesnt-exist.txt")
    Condition
      Error in `snapshot_file_equal_()`:
      ! 'doesnt-exist.txt' not found.

# generates informative hint

    Code
      snapshot_hint("lala", reset_output = FALSE)
    Output
      * Run `testthat::snapshot_accept("lala")` to accept the change.
      * Run `testthat::snapshot_review("lala")` to review the change.

# expect_snapshot_file validates its inputs

    Code
      expect_snapshot_file(123)
    Condition
      Error in `expect_snapshot_file()`:
      ! `path` must be a single string, not the number 123.
    Code
      expect_snapshot_file("doesnt-exist.txt")
    Condition
      Error in `expect_snapshot_file()`:
      ! 'doesnt-exist.txt' doesn't exist.
    Code
      expect_snapshot_file(path, 123)
    Condition
      Error in `expect_snapshot_file()`:
      ! `name` must be a single string, not the number 123.
    Code
      expect_snapshot_file(path, "test.txt", cran = "yes")
    Condition
      Error in `expect_snapshot_file()`:
      ! `cran` must be `TRUE` or `FALSE`, not the string "yes".

# useful failure if empty

    Code
      expect_match(zero, "asdf")
    Condition
      Error:
      ! Expected `zero` to have at least one element.

# useful failure messages for scalars

    Code
      expect_match(one, "asdf")
    Condition
      Error:
      ! Expected `one` to match regexp "asdf".
      Actual text:
      ✖ │ bcde

---

    Code
      expect_match(one, "asdf", fixed = TRUE)
    Condition
      Error:
      ! Expected `one` to match string "asdf".
      Actual text:
      ✖ │ bcde

# useful failure messages for vectors

    Code
      expect_match(many, "a")
    Condition
      Error:
      ! Expected every element of `many` to match regexp "a".
      Actual text:
      ✔ │ a
      ✔ │ a
      ✖ │ b

---

    Code
      expect_match(many, "c", all = FALSE)
    Condition
      Error:
      ! Expected some element of `many` to match regexp "c".
      Actual text:
      ✖ │ a
      ✖ │ a
      ✖ │ b

---

    Code
      expect_match(paragraph, "paragraph")
    Condition
      Error:
      ! Expected every element of `paragraph` to match regexp "paragraph".
      Actual text:
      ✔ │ This is a multiline
        │ paragraph.
      ✖ │ Second element.

---

    Code
      expect_match(na, "NA")
    Condition
      Error:
      ! Expected every element of `na` to match regexp "NA".
      Actual text:
      ✔ │ NA
      ✖ │ <NA>

# expect_match validates its inputs

    Code
      expect_match(1)
    Condition
      Error in `expect_match()`:
      ! `object` must be a character vector, not the number 1.
    Code
      expect_match("x", 1)
    Condition
      Error in `expect_match()`:
      ! `regexp` must be a single string, not the number 1.
    Code
      expect_match("x", "x", fixed = 1)
    Condition
      Error in `expect_match()`:
      ! `fixed` must be `TRUE` or `FALSE`, not the number 1.
    Code
      expect_match("x", "x", perl = 1)
    Condition
      Error in `expect_match()`:
      ! `perl` must be `TRUE` or `FALSE`, not the number 1.
    Code
      expect_match("x", "x", all = 1)
    Condition
      Error in `expect_match()`:
      ! `all` must be `TRUE` or `FALSE`, not the number 1.

# expect_no_match validates its inputs

    Code
      expect_no_match(1, "x")
    Condition
      Error in `expect_no_match()`:
      ! `object` must be a character vector, not the number 1.
    Code
      expect_no_match("x", 1)
    Condition
      Error in `expect_no_match()`:
      ! `regexp` must be a single string, not the number 1.
    Code
      expect_no_match("x", "x", fixed = 1)
    Condition
      Error in `expect_no_match()`:
      ! `fixed` must be `TRUE` or `FALSE`, not the number 1.
    Code
      expect_no_match("x", "x", perl = 1)
    Condition
      Error in `expect_no_match()`:
      ! `perl` must be `TRUE` or `FALSE`, not the number 1.
    Code
      expect_no_match("x", "x", all = 1)
    Condition
      Error in `expect_no_match()`:
      ! `all` must be `TRUE` or `FALSE`, not the number 1.

# expect_no_match works

    Code
      expect_no_match(x, "e*", fixed = TRUE)
    Condition
      Error:
      ! Expected `x` not to match string "e*".
      Actual text:
      x | te*st

---

    Code
      expect_no_match(x, "TEST", ignore.case = TRUE)
    Condition
      Error:
      ! Expected `x` not to match regexp "TEST".
      Actual text:
      x | test

# show_text() shows success and failure

    Code
      base::writeLines(show_text(c("a", "b"), c(TRUE, FALSE)))
    Output
      ✔ │ a
      ✖ │ b

# show_text() truncates values and lines

    Code
      base::writeLines(show_text(lines, max_lines = 3))
    Output
      ✔ │ a
        │ b
        │ ...
      ... and 8 more.
      
    Code
      base::writeLines(show_text(lines, max_items = 3))
    Output
      ✔ │ a
        │ b
        │ c
      ✔ │ d
        │ e
        │ f
      ✔ │ g
        │ h
        │ i
      ... and 6 more.
      
    Code
      base::writeLines(show_text(lines, max_items = 2, max_lines = 4))
    Output
      ✔ │ a
        │ b
        │ c
      ✔ │ d
        │ ...
      ... and 8 more.
      

# checks its inputs

    Code
      local_edition("x")
    Condition
      Error in `local_edition()`:
      ! `x` must be a whole number, not the string "x".
    Code
      local_edition(5)
    Condition
      Error in `local_edition()`:
      ! `x` must be a whole number between 2 and 3, not the number 5.

# deprecation only fired for newer edition

    Code
      edition_deprecate(3, "old stuff")
    Condition
      Warning:
      `old stuff` was deprecated in the 3rd edition.

# now defunct

    Code
      local_mock()
    Condition
      Error:
      ! `local_mock()` was deprecated in testthat 3.2.0 and is now defunct.
      i Please use `local_mocked_bindings()` instead.
    Code
      with_mock(is_testing = function() FALSE)
    Condition
      Error:
      ! `with_mock()` was deprecated in testthat 3.2.0 and is now defunct.
      i Please use `with_mocked_bindings()` instead.

# provide useful feedback on failure (3e)

    Code
      expect_identical(x, "a")
    Condition
      Error:
      ! Expected `x` to be identical to "a".
      Differences:
      `actual` is a double vector (1)
      `expected` is a character vector ('a')

---

    Code
      expect_equal(x, "a")
    Condition
      Error:
      ! Expected `x` to equal "a".
      Differences:
      `actual` is a double vector (1)
      `expected` is a character vector ('a')

---

    Code
      expect_identical(x, "a")
    Condition
      Error:
      ! Expected `x` to be identical to "a".
      Differences:
      Types not compatible: double is not character

---

    Code
      expect_equal(x, "a")
    Condition
      Error:
      ! Expected `x` to equal "a".
      Differences:
      Types not compatible: double is not character

# correctly spaces lines

    Code
      expect_equal(list(a = 1), list(a = "b", b = 10))
    Condition
      Error:
      ! Expected `list(a = 1)` to equal `list(a = "b", b = 10)`.
      Differences:
      `actual` is length 1
      `expected` is length 2
      
      `names(actual)`:   "a"    
      `names(expected)`: "a" "b"
      
      `actual$a` is a double vector (1)
      `expected$a` is a character vector ('b')
      
      `actual$b` is absent
      `expected$b` is a double vector (10)

# provide useful feedback on failure (2e)

    Code
      expect_identical(x, "a")
    Condition
      Error:
      ! Expected `x` to be identical to "a".
      Differences:
      Types not compatible: double is not character

---

    Code
      expect_equal(x, "a")
    Condition
      Error:
      ! Expected `x` to equal "a".
      Differences:
      Types not compatible: double is not character

# default labels use unquoting

    Code
      expect_equal(x, !!y)
    Condition
      Error:
      ! Expected `x` to equal 2.
      Differences:
        `actual`: 1.0
      `expected`: 2.0

# useful message if objects equal but not identical

    Code
      expect_identical(f, g)
    Condition
      Error:
      ! Expected `f` to be identical to `g`.
      Differences:
      Objects equal but not identical

# attributes for object (#452)

    Code
      expect_equal(oops, 0)
    Condition
      Error:
      ! Expected `oops` to equal 0.
      Differences:
      Attributes: < Modes: list, NULL >
      Attributes: < Lengths: 1, 0 >
      Attributes: < names for target but not for current >
      Attributes: < current is not list-like >

# expect_equal validates its inputs

    Code
      expect_equal(1, 2, tolerance = "high")
    Condition
      Error in `expect_equal()`:
      ! `tolerance` must be a number or `NULL`, not the string "high".
    Code
      expect_equal(1, 2, tolerance = -1)
    Condition
      Error in `expect_equal()`:
      ! `tolerance` must be a number larger than or equal to 0 or `NULL`, not the number -1.

# can control appearance of dots

    reporters/tests: 
    Successes: .
    Failures: 12
    Errors: 34
    Skips: SS
    Warnings: WS
    
    == Skipped =====================================================================
    1. explicit skips are reported ('reporters/tests.R:38:3') - Reason: skip
    
    2. empty tests are implicitly skipped ('reporters/tests.R:41:1') - Reason: empty test
    
    3. warnings get backtraces ('reporters/tests.R:45:1') - Reason: empty test
    
    == Warnings ====================================================================
    1. warnings get backtraces ('reporters/tests.R:47:5') - def
    
    == Failed ======================================================================
    -- 1. Failure ('reporters/tests.R:13:3'): Failure:1 ----------------------------
    Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    -- 2. Failure ('reporters/tests.R:17:8'): Failure:2a ---------------------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
    
    -- 3. Error ('reporters/tests.R:24:3'): Error:1 --------------------------------
    Error in `eval(code, test_env)`: stop
    
    -- 4. Error ('reporters/tests.R:30:8'): errors get tracebacks ------------------
    Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
    
    == DONE ========================================================================

---

    reporters/tests: 
    Successes: 
    Failures: 12
    Errors: 34
    Skips: SS
    Warnings: WS
    
    == Skipped =====================================================================
    1. explicit skips are reported ('reporters/tests.R:38:3') - Reason: skip
    
    2. empty tests are implicitly skipped ('reporters/tests.R:41:1') - Reason: empty test
    
    3. warnings get backtraces ('reporters/tests.R:45:1') - Reason: empty test
    
    == Warnings ====================================================================
    1. warnings get backtraces ('reporters/tests.R:47:5') - def
    
    == Failed ======================================================================
    -- 1. Failure ('reporters/tests.R:13:3'): Failure:1 ----------------------------
    Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    -- 2. Failure ('reporters/tests.R:17:8'): Failure:2a ---------------------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
    
    -- 3. Error ('reporters/tests.R:24:3'): Error:1 --------------------------------
    Error in `eval(code, test_env)`: stop
    
    -- 4. Error ('reporters/tests.R:30:8'): errors get tracebacks ------------------
    Error in `h()`: !
    Backtrace:
        x
     1. \-f()
     2.   \-g()
     3.     \-h()
    
    == DONE ========================================================================

# can control maximum reports

    reporters/tests: 
    Successes: .
    Failures: 12
    Errors: 34
    Skips: SS
    Warnings: WS
    
    == Skipped =====================================================================
    1. explicit skips are reported ('reporters/tests.R:38:3') - Reason: skip
    
    2. empty tests are implicitly skipped ('reporters/tests.R:41:1') - Reason: empty test
    
    3. warnings get backtraces ('reporters/tests.R:45:1') - Reason: empty test
    
    == Warnings ====================================================================
    1. warnings get backtraces ('reporters/tests.R:47:5') - def
    
    == Failed ======================================================================
    -- 1. Failure ('reporters/tests.R:13:3'): Failure:1 ----------------------------
    Expected `x` to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    
    -- 2. Failure ('reporters/tests.R:17:8'): Failure:2a ---------------------------
    Expected FALSE to be TRUE.
    Differences:
    `actual`:   FALSE
    `expected`: TRUE 
    
    Backtrace:
        x
     1. \-f()
     2.   \-testthat::expect_true(FALSE)
      ... and 2 more
    
    
    Maximum number of 2 failures reached, some test results may be missing.
    
    == DONE ========================================================================

# source_file wraps error

    Code
      source_file(test_path("reporters/error-setup.R"), wrap = FALSE)
    Condition
      Error:
      ! Failed to evaluate 'reporters/error-setup.R'.
      Caused by error in `h()`:
      ! !

# checks its inputs

    Code
      source_file(1)
    Condition
      Error:
      ! `path` must be a single string, not the number 1.
    Code
      source_file("x")
    Condition
      Error:
      ! `path` does not exist.
    Code
      source_file(".", "x")
    Condition
      Error:
      ! `env` must be an environment, not the string "x".

# works on code like the describe() example

    Code
      filter_desc(code, c("math library", "division()", "can handle division by 0"))
    Condition
      Error:
      ! Failed to find test with description "can handle division by 0".

# preserve srcrefs

    Code
      filter_desc(code, "foo")
    Output
      expression(test_that("foo", {
            # this is a comment
          }))

# errors if zero or duplicate labels

    Code
      filter_desc(code, "baz")
    Condition
      Error:
      ! Found multiple tests with description "baz".
    Code
      filter_desc(code, "missing")
    Condition
      Error:
      ! Failed to find test with description "missing".

# generates actionable failure message

    Code
      expect_length(x, 2)
    Condition
      Error:
      ! Expected `x` to have length 2.
      Actual length: 10.

# expect_length validates its inputs

    Code
      expect_length(1:5, "a")
    Condition
      Error in `expect_length()`:
      ! `n` must be a whole number, not the string "a".

# dim compared correctly

    Code
      expect_shape(matrix(nrow = 6, ncol = 3), dim = c(6L, 2L))
    Condition
      Error:
      ! Expected `matrix(nrow = 6, ncol = 3)` to have dim (6, 2).
      Actual dim: (6, 3).

---

    Code
      expect_shape(matrix(nrow = 6, ncol = 3), dim = c(7L, 3L))
    Condition
      Error:
      ! Expected `matrix(nrow = 6, ncol = 3)` to have dim (7, 3).
      Actual dim: (6, 3).

---

    Code
      expect_shape(array(dim = 1:3), dim = 1:2)
    Condition
      Error:
      ! Expected `array(dim = 1:3)` to have 2 dimensions.
      Actual dimensions: 3.

---

    Code
      expect_shape(array(dim = 1:3), dim = 1:4)
    Condition
      Error:
      ! Expected `array(dim = 1:3)` to have 4 dimensions.
      Actual dimensions: 3.

# nrow compared correctly

    Code
      expect_shape(matrix(nrow = 5, ncol = 5), nrow = 6L)
    Condition
      Error:
      ! Expected `matrix(nrow = 5, ncol = 5)` to have 6 rows.
      Actual rows: 5.

---

    Code
      expect_shape(1, nrow = 1)
    Condition
      Error:
      ! Expected 1 to have dimensions.

# ncol compared correctly

    Code
      expect_shape(matrix(nrow = 5, ncol = 5), ncol = 7L)
    Condition
      Error:
      ! Expected `matrix(nrow = 5, ncol = 5)` to have 7 columns.
      Actual columns: 5.

---

    Code
      expect_shape(array(1), ncol = 1)
    Condition
      Error:
      ! Expected `array(1)` to have two or more dimensions.

---

    Code
      expect_shape(array(integer()), ncol = 0L)
    Condition
      Error:
      ! Expected `array(integer())` to have two or more dimensions.

# NA handling (e.g. dbplyr)

    Code
      expect_shape(x, nrow = 10L)
    Condition
      Error:
      ! Expected `x` to have 10 rows.
      Actual rows: NA.

---

    Code
      expect_shape(x, ncol = NA_integer_)
    Condition
      Error:
      ! Expected `x` to have NA columns.
      Actual columns: 10.

---

    Code
      expect_shape(x, dim = c(10L, NA_integer_))
    Condition
      Error:
      ! Expected `x` to have dim (10, NA).
      Actual dim: (NA, 10).

# checks inputs arguments, 

    Code
      expect_shape(1:10)
    Condition
      Error in `expect_shape()`:
      ! One of `nrow`, `ncol`, or `dim` must be supplied.
    Code
      expect_shape(1:10, nrow = 1L, ncol = 2L)
    Condition
      Error in `expect_shape()`:
      ! Exactly one of `nrow`, `ncol`, or `dim` must be supplied.
      x `nrow` and `ncol` were supplied together.
    Code
      expect_shape(1:10, 2)
    Condition
      Error in `expect_shape()`:
      ! `...` must be empty.
      x Problematic argument:
      * ..1 = 2
      i Did you forget to name an argument?
    Code
      expect_shape(array(1), nrow = "x")
    Condition
      Error in `expect_shape()`:
      ! `nrow` must be a whole number or `NA`, not the string "x".
    Code
      expect_shape(array(1), ncol = "x")
    Condition
      Error in `expect_shape()`:
      ! `ncol` must be a whole number or `NA`, not the string "x".
    Code
      expect_shape(array(1), dim = "x")
    Condition
      Error in `expect_shape()`:
      ! `dim` must be a numeric vector, not the string "x".

# validates its inputs

    Code
      expect_all_equal(mean, 1)
    Condition
      Error in `expect_all_equal()`:
      ! `object` must be a vector, not a function.
    Code
      expect_all_equal(logical(), 1)
    Condition
      Error in `expect_all_equal()`:
      ! `object` must not be empty.
    Code
      expect_all_equal(1:10, mean)
    Condition
      Error in `expect_all_equal()`:
      ! `expected` must be a vector, not a function.
    Code
      expect_all_equal(1:10, 1:2)
    Condition
      Error in `expect_all_equal()`:
      ! `expected` must be length 1.

# can compare atomic vectors

    Code
      expect_all_equal(x, TRUE)
    Condition
      Error:
      ! Expected every element of `x` to equal TRUE.
      Differences:
      `actual[2:8]`:   TRUE TRUE TRUE FALSE TRUE TRUE TRUE
      `expected[2:8]`: TRUE TRUE TRUE TRUE  TRUE TRUE TRUE

# can compare named lists

    Code
      expect_all_equal(x, list(1))
    Condition
      Error:
      ! Expected every element of `x` to equal `list(1)`.
      Differences:
        `actual$c`: 2.0
      `expected$c`: 1.0

# truncates very long differences

    Code
      expect_all_equal(x, FALSE)
    Condition
      Error:
      ! Expected every element of `x` to equal FALSE.
      Differences:
      `actual`:   TRUE  TRUE  TRUE  TRUE  TRUE  TRUE  TRUE  TRUE  TRUE  TRUE 
      `expected`: FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE FALSE

# basic comparisons work

    Code
      expect_lt(x, 10)
    Condition
      Error:
      ! Expected `x` < 10.
      Actual comparison: 10.0 >= 10.0
      Difference: 0.0 >= 0

---

    Code
      expect_gt(x, 10)
    Condition
      Error:
      ! Expected `x` > 10.
      Actual comparison: 10.0 <= 10.0
      Difference: 0.0 <= 0

# useful output when numbers are very small

    Code
      expect_lte(1.1 * x, x)
    Condition
      Error:
      ! Expected `1.1 * x` <= `x`.
      Actual comparison: 0.0000110 > 0.0000100
      Difference: 0.0000010 > 0

---

    Code
      expect_gt(x, 1.1 * x)
    Condition
      Error:
      ! Expected `x` > `1.1 * x`.
      Actual comparison: 0.0000100 <= 0.0000110
      Difference: -0.0000010 <= 0

# useful output when difference is zero

    Code
      expect_lt(x, 100)
    Condition
      Error:
      ! Expected `x` < 100.
      Actual comparison: 100.0 >= 100.0
      Difference: 0.0 >= 0

# useful output when differnce is large

    Code
      expect_lt(x, 0.001)
    Condition
      Error:
      ! Expected `x` < 0.001.
      Actual comparison: 100.000 >= 0.001
      Difference: 99.999 >= 0

# comparisons with Inf work

    Code
      expect_lt(x, Inf)
    Condition
      Error:
      ! Expected `x` < Inf.
      Actual comparison: Inf >= Inf

# comparisons with NA work

    Code
      expect_lt(x, 10)
    Condition
      Error:
      ! Expected `x` < 10.
      Actual comparison: NA >= 10.0

# comparisons with negative numbers work

    Code
      expect_gt(-5, -2)
    Condition
      Error:
      ! Expected `-5` > `-2`.
      Actual comparison: -5.0 <= -2.0
      Difference: -3.0 <= 0

# comparisons with POSIXct objects work

    Code
      expect_lt(time2, time)
    Condition
      Error:
      ! Expected `time2` < `time`.
      Actual comparison: "2020-01-01 01:00:01.5" >= "2020-01-01 01:00:00"
      Difference: 1.5 secs >= 0 secs

# comparisons with Date objects work

    Code
      expect_gt(date, date2)
    Condition
      Error:
      ! Expected `date` > `date2`.
      Actual comparison: "2020-01-01" <= "2020-01-02"
      Difference: -1.0 days <= 0 days

# comparisons with character objects work

    Code
      expect_lte("b", "a")
    Condition
      Error:
      ! Expected "b" <= "a".
      Actual comparison: "b" > "a"

# comparison must yield a single logical

    Code
      expect_lt(1:10, 5)
    Condition
      Error in `expect_lt()`:
      ! Result of comparison must be `TRUE`, `FALSE`, or `NA`

# basic properties upheld

    Code
      expect_vector(x, size = 5)
    Condition
      Error:
      ! `x` must have size 5, not size 10.

---

    Code
      expect_vector(y)
    Condition
      Error:
      ! `y` must be a vector, not `NULL`.

# expect_vector validates its inputs

    Code
      expect_vector(1:5, size = "large")
    Condition
      Error in `expect_vector()`:
      ! `size` must be a whole number or `NULL`, not the string "large".

# validates key inputs

    Code
      expect(1)
    Condition
      Error in `expect()`:
      ! `ok` must be `TRUE` or `FALSE`, not the number 1.
    Code
      expect(TRUE, 1)
    Condition
      Error in `expect()`:
      ! `failure_message` must be a character vector, not the number 1.

# tries multiple times

    Code
      result <- try_again(3, third_try())
    Message
      i Expectation failed; trying again (1)...
      i Expectation failed; trying again (2)...

---

    Code
      try_again(1, third_try())
    Message
      i Expectation failed; trying again (1)...
    Condition
      Error:
      ! Expected `i` to equal 0.
      Differences:
        `actual`: 1.0
      `expected`: 0.0

# handles errors

    Code
      result <- try_again(3, expect_equal(fails_twice(), 1))
    Message
      i Expectation errored; trying again (1)...
      i Expectation errored; trying again (2)...

# expect_type checks typeof

    Code
      expect_type(x, "double")
    Condition
      Error:
      ! Expected `x` to have type "double".
      Actual type: "integer"

# expect_type validates its inputs

    Code
      expect_type(1, c("integer", "double"))
    Condition
      Error in `expect_type()`:
      ! `type` must be a single string, not a character vector.

# expect_is checks class

    Code
      expect_is(factor("a"), "integer")
    Condition
      Error:
      ! Expected `factor("a")` to inherit from "character".
      Actual inheritance: "factor"

# expect_s3/s4_class fails if appropriate type

    Code
      expect_s3_class(x1, "double")
    Condition
      Error:
      ! Expected `x1` to be an S3 object.
      Actual OO type: none.

---

    Code
      expect_s3_class(x2, "double")
    Condition
      Error:
      ! Expected `x2` to be an S3 object.
      Actual OO type: S4.

---

    Code
      expect_s4_class(x3, "double")
    Condition
      Error:
      ! Expected `x3` to be an S4 object.
      Actual OO type: S3.

# expect_s[34]_class can check not S3/S4

    Code
      expect_s3_class(factor(), NA)
    Condition
      Error:
      ! Expected `factor()` not to be an S3 object.

---

    Code
      expect_s4_class(A(), NA)
    Condition
      Error:
      ! Expected `A()` not to be an S4 object.

# test_s4_class respects class hierarchy

    Code
      expect_s4_class(C(), "D")
    Condition
      Error:
      ! Expected `C()` to inherit from "D".
      Actual class: "C"/"A"/"B"/"list"/"vector".

# expect_s3_class validates its inputs

    Code
      expect_s3_class(factor("a"), 1)
    Condition
      Error in `expect_s3_class()`:
      ! `class` must be a character vector or NA, not the number 1.
    Code
      expect_s3_class(factor("a"), "factor", exact = "yes")
    Condition
      Error in `expect_s3_class()`:
      ! `exact` must be `TRUE` or `FALSE`, not the string "yes".

# test_s3_class respects class hierarchy

    Code
      expect_s3_class(x, "c")
    Condition
      Error:
      ! Expected `x` to inherit from "c".
      Actual class: "a"/"b".

---

    Code
      expect_s3_class(x, c("c", "d"))
    Condition
      Error:
      ! Expected `x` to inherit from "c"/"d".
      Actual class: "a"/"b".

# test_s3_class can request exact match

    Code
      expect_s3_class(x, "a", exact = TRUE)
    Condition
      Error:
      ! Expected `x` to have class "a".
      Actual class: "a"/"b".

# expect_s4_class validates its inputs

    Code
      expect_s4_class(factor("a"), 1)
    Condition
      Error in `expect_s4_class()`:
      ! `class` must be a character vector or NA, not the number 1.

# expect_r6_class generates useful failures

    Code
      expect_r6_class(x, "Student")
    Condition
      Error:
      ! Expected `x` to be an R6 object.
      Actual OO type: none.
    Code
      expect_r6_class(person, "Student")
    Condition
      Error:
      ! Expected `person` to inherit from "Student".
      Actual class: "Person"/"R6".

# expect_r6_class validates its inputs

    Code
      expect_r6_class(1, c("Person", "Student"))
    Condition
      Error in `expect_r6_class()`:
      ! `class` must be a single string, not a character vector.

# can check with actual class

    Code
      expect_s7_class(Foo(), class = Bar)
    Condition
      Error:
      ! Expected `Foo()` to inherit from <Bar>.
      Actual class: <Foo>.

---

    Code
      expect_s7_class(Baz(), class = Bar)
    Condition
      Error:
      ! Expected `Baz()` to inherit from <Bar>.
      Actual class: <Baz>/<Foo>.

# informative failure if not S7

    Code
      expect_s7_class(x, Foo)
    Condition
      Error:
      ! Expected `x` to be an S7 object.
      Actual OO type: S3.

# expect_s7_class validates its inputs

    Code
      expect_s7_class(1, 1)
    Condition
      Error in `expect_s7_class()`:
      ! `class` must be an S7 class object, not the number 1.

# warns if both inputs are named

    Code
      expect_setequal(c(a = 1), c(b = 1))
    Condition
      Warning:
      expect_setequal() ignores names

# checks inputs

    Code
      expect_setequal(sum, 1)
    Condition
      Error in `expect_setequal()`:
      ! `object` must be a vector, not a primitive function.
    Code
      expect_setequal(1, sum)
    Condition
      Error in `expect_setequal()`:
      ! `expected` must be a vector, not a primitive function.
    Code
      expect_setequal(!!fun, 1)
    Condition
      Error in `expect_setequal()`:
      ! `object` must be a vector, not a primitive function.
    Code
      expect_setequal(1, !!fun)
    Condition
      Error in `expect_setequal()`:
      ! `expected` must be a vector, not a primitive function.

# useful message on failure

    Code
      expect_setequal("actual", "expected")
    Condition
      Error:
      ! Expected "actual" to have the same values as "expected".
      Actual: "actual"
      Expected: "expected"
      Needs: "actual"
      Absent: "expected"

---

    Code
      expect_setequal(x, y)
    Condition
      Error:
      ! Expected `x` to have the same values as `y`.
      Actual: 1, 2
      Expected: 2
      Needs: 1

---

    Code
      expect_setequal(x, y)
    Condition
      Error:
      ! Expected `x` to have the same values as `y`.
      Actual: 2
      Expected: 2, 3
      Absent: 3

---

    Code
      expect_setequal(x, y)
    Condition
      Error:
      ! Expected `x` to have the same values as `y`.
      Actual: 1, 2
      Expected: 2, 3
      Needs: 1
      Absent: 3

---

    Code
      expect_setequal(x, y)
    Condition
      Error:
      ! Expected `x` to have the same values as `y`.
      Actual: "a", "a"
      Expected: "b", "b", "b"
      Needs: "a"
      Absent: "b"

---

    Code
      expect_setequal(x, c("a", "b", "c", "d"))
    Condition
      Error:
      ! Expected `x` to have the same values as `c("a", "b", "c", "d")`.
      Actual: "a", "b", "c"
      Expected: "a", "b", "c", "d"
      Absent: "d"

# truncates long vectors

    Code
      expect_setequal(x, y)
    Condition
      Error:
      ! Expected `x` to have the same values as `y`.
      Actual: 1, 2
      Expected: 1, 2, 3, 4, 5, 6, 7, 8, 9, ...
      Absent: 3, 4, 5, 6, 7, 8, 9, 10, 11, ...

# fails if names don't match

    Code
      expect_mapequal(x, y)
    Condition
      Error:
      ! Expected `x` to have the same names as `y`.
      Actual: "a", "b"
      Expected: "a"
      Needs: "b"

---

    Code
      expect_mapequal(y, x)
    Condition
      Error:
      ! Expected `y` to have the same names as `x`.
      Actual: "a"
      Expected: "a", "b"
      Absent: "b"

# fails if values don't match

    Code
      expect_mapequal(x, y)
    Condition
      Error:
      ! Expected `x` to contain the same values as `y`.
      Differences:
        `actual$b`: 2.0
      `expected$b`: 3.0

# warns if empty vector

    Code
      expect_success(expect_mapequal(list(), list()))
    Condition
      Warning:
      `object` and `expected` are empty vectors.

# validates its inputs

    Code
      expect_mapequal(sum, named)
    Condition
      Error in `expect_mapequal()`:
      ! `object` must be a vector, not a primitive function.
    Code
      expect_mapequal(named, sum)
    Condition
      Error in `expect_mapequal()`:
      ! `expected` must be a vector, not a primitive function.
    Code
      expect_mapequal(unnamed, named)
    Condition
      Error in `expect_mapequal()`:
      ! All elements in `object` must have names.
      x Empty names at position: 1
    Code
      expect_mapequal(named, unnamed)
    Condition
      Error in `expect_mapequal()`:
      ! All elements in `expected` must have names.
      x Empty names at position: 1
    Code
      expect_mapequal(named, duplicated)
    Condition
      Error in `expect_mapequal()`:
      ! All elements in `expected` must have unique names.
      x Duplicate names: "x"
    Code
      expect_mapequal(duplicated, named)
    Condition
      Error in `expect_mapequal()`:
      ! All elements in `object` must have unique names.
      x Duplicate names: "x"

# expect_contains() gives useful message on failure

    Code
      expect_contains(x1, x2)
    Condition
      Error:
      ! Expected `x1` to contain all values in `x2`.
      Actual: "a", "b", "c"
      Expected: "c", "d"
      Missing: "d"

---

    Code
      expect_contains(x1, x3)
    Condition
      Error:
      ! Expected `x1` to contain all values in `x3`.
      Actual: "a", "b", "c"
      Expected: "d", "e"
      Missing: "d", "e"

# expect_in() gives useful message on failure

    Code
      expect_in(x1, x2)
    Condition
      Error:
      ! Expected `x1` to only contain values from `x2`.
      Actual: "a", "b"
      Expected: "b", "c"
      Invalid: "a"

---

    Code
      expect_in(x1, x3)
    Condition
      Error:
      ! Expected `x1` to only contain values from `x3`.
      Actual: "a", "b"
      Expected: "d", "e"
      Invalid: "a", "b"

# expect_disjoint() gives useful message on failure

    Code
      expect_disjoint(x1, x2)
    Condition
      Error:
      ! Expected `x1` to be disjoint from `x2`.
      Actual: "a", "b", "c"
      Expected: None of "c", "d"
      Invalid: "c"

---

    Code
      expect_disjoint(x1, x3)
    Condition
      Error:
      ! Expected `x1` to be disjoint from `x3`.
      Actual: "a", "b", "c"
      Expected: None of "b", "c", "d"
      Invalid: "b", "c"

---

    Code
      expect_disjoint(NA, c("a", NA))
    Condition
      Error:
      ! Expected NA to be disjoint from `c("a", NA)`.
      Actual: NA
      Expected: None of "a", NA
      Invalid: NA

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-silent.R
\name{expect_silent}
\alias{expect_silent}
\title{Does code execute silently?}
\usage{
expect_silent(object)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}
}
\value{
The first argument, invisibly.
}
\description{
Checks that the code produces no output, messages, or warnings.
}
\examples{
expect_silent("123")

f <- function() {
  message("Hi!")
  warning("Hey!!")
  print("OY!!!")
}
\dontrun{
expect_silent(f())
}
}
\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/old-school.R
\name{takes_less_than}
\alias{takes_less_than}
\title{Does code take less than the expected amount of time to run?}
\usage{
takes_less_than(amount)
}
\arguments{
\item{amount}{maximum duration in seconds}
}
\description{
This is useful for performance regression testing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-no-condition.R
\name{expect_no_error}
\alias{expect_no_error}
\alias{expect_no_warning}
\alias{expect_no_message}
\alias{expect_no_condition}
\title{Do you expect the absence of errors, warnings, messages, or other conditions?}
\usage{
expect_no_error(object, ..., message = NULL, class = NULL)

expect_no_warning(object, ..., message = NULL, class = NULL)

expect_no_message(object, ..., message = NULL, class = NULL)

expect_no_condition(object, ..., message = NULL, class = NULL)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{...}{These dots are for future extensions and must be empty.}

\item{message, class}{The default, \verb{message = NULL, class = NULL},
will fail if there is any error/warning/message/condition.

In many cases, particularly when testing warnings and messages, you will
want to be more specific about the condition you are hoping \strong{not} to see,
i.e. the condition that motivated you to write the test.  Similar to
\code{expect_error()} and friends, you can specify the \code{message} (a regular
expression that the message of the condition must match) and/or the
\code{class} (a class the condition must inherit from). This ensures that
the message/warnings you don't want never recur, while allowing new
messages/warnings to bubble up for you to deal with.

Note that you should only use \code{message} with errors/warnings/messages
that you generate, or that base R generates (which tend to be stable).
Avoid tests that rely on the specific text generated by another package
since this can easily change. If you do need to test text generated by
another package, either protect the test with \code{skip_on_cran()} or
use \code{expect_snapshot()}.}
}
\description{
These expectations are the opposite of \code{\link[=expect_error]{expect_error()}},
\code{expect_warning()}, \code{expect_message()}, and \code{expect_condition()}. They
assert the absence of an error, warning, or message, respectively.
}
\examples{
expect_no_warning(1 + 1)

foo <- function(x) {
  warning("This is a problem!")
}

# warning doesn't match so bubbles up:
expect_no_warning(foo(), message = "bananas")

# warning does match so causes a failure:
try(expect_no_warning(foo(), message = "problem"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot-github.R
\name{snapshot_download_gh}
\alias{snapshot_download_gh}
\title{Download snapshots from GitHub}
\usage{
snapshot_download_gh(repository, run_id, dest_dir = ".")
}
\arguments{
\item{repository}{Repository owner/name, e.g. \code{"r-lib/testthat"}.}

\item{run_id}{Run ID, e.g. \code{"47905180716"}. You can find this in the action url.}

\item{dest_dir}{Package root directory. Defaults to the current directory.}
}
\description{
If your snapshots fail on GitHub, it can be a pain to figure out exactly
why, or to incorporate them into your local package. This function makes it
easy, only requiring you to interactively select which job you want to
take the artifacts from.

Note that you should not generally need to use this function manually;
instead copy and paste from the hint emitted on GitHub. This hint is only
emitted when running in a job named "R-CMD-check", since that's where the
testthat artifact is typically uploaded.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verify-output.R
\name{verify_output}
\alias{verify_output}
\title{Verify output}
\usage{
verify_output(
  path,
  code,
  width = 80,
  crayon = FALSE,
  unicode = FALSE,
  env = caller_env()
)
}
\arguments{
\item{path}{Path to record results.

This should usually be a call to \code{\link[=test_path]{test_path()}} in order to ensure that
the same path is used when run interactively (when the working directory
is typically the project root), and when run as an automated test (when
the working directory will be \code{tests/testthat}).}

\item{code}{Code to execute. This will usually be a multiline expression
contained within \code{{}} (similarly to \code{test_that()} calls).}

\item{width}{Width of console output}

\item{crayon}{Enable cli/crayon package colouring?}

\item{unicode}{Enable cli package UTF-8 symbols? If you set this to
\code{TRUE}, call \code{skip_if(!cli::is_utf8_output())} to disable the
test on your CI platforms that don't support UTF-8 (e.g. Windows).}

\item{env}{The environment to evaluate \code{code} in.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

This function is superseded in favour of \code{expect_snapshot()} and friends.

This is a regression test that records interwoven code and output into a
file, in a similar way to knitting an \code{.Rmd} file (but see caveats below).

\code{verify_output()} is designed particularly for testing print methods and error
messages, where the primary goal is to ensure that the output is helpful to
a human. Obviously, you can't test that with code, so the best you can do is
make the results explicit by saving them to a text file. This makes the output
easy to verify in code reviews, and ensures that you don't change the output
by accident.

\code{verify_output()} is designed to be used with git: to see what has changed
from the previous run, you'll need to use \verb{git diff} or similar.
}
\section{Syntax}{

\code{verify_output()} can only capture the abstract syntax tree, losing all
whitespace and comments. To mildly offset this limitation:
\itemize{
\item Strings are converted to R comments in the output.
\item Strings starting with \verb{# } are converted to headers in the output.
}
}

\section{CRAN}{

On CRAN, \code{verify_output()} will never fail, even if the output changes.
This avoids false positives because tests of print methods and error
messages are often fragile due to implicit dependencies on other packages,
and failure does not imply incorrect computation, just a change in
presentation.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-path.R
\name{test_path}
\alias{test_path}
\title{Locate a file in the testing directory}
\usage{
test_path(...)
}
\arguments{
\item{...}{Character vectors giving path components.}
}
\value{
A character vector giving the path.
}
\description{
Many tests require some external file (e.g. a \code{.csv} if you're testing a
data import function) but the working directory varies depending on the way
that you're running the test (e.g. interactively, with \code{devtools::test()},
or with \verb{R CMD check}). \code{test_path()} understands these variations and
automatically generates a path relative to \code{tests/testthat}, regardless of
where that directory might reside relative to the current working directory.
}
\examples{
\dontrun{
test_path("foo.csv")
test_path("data", "foo.csv")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-that.R
\name{test_that}
\alias{test_that}
\title{Run a test}
\usage{
test_that(desc, code)
}
\arguments{
\item{desc}{Test name. Names should be brief, but evocative. It's common to
write the description so that it reads like a natural sentence, e.g.
\code{test_that("multiplication works", { ... })}.}

\item{code}{Test code containing expectations. Braces (\code{{}}) should always
be used in order to get accurate location data for test failures.}
}
\value{
When run interactively, returns \code{invisible(TRUE)} if all tests
pass, otherwise throws an error.
}
\description{
A test encapsulates a series of expectations about a small, self-contained
unit of functionality. Each test contains one or more expectations, such as
\code{\link[=expect_equal]{expect_equal()}} or \code{\link[=expect_error]{expect_error()}}, and lives in a \verb{test/testhat/test*}
file, often together with other tests that relate to the same function or set
of functions.

Each test has its own execution environment, so an object created in a test
also dies with the test. Note that this cleanup does not happen automatically
for other aspects of global state, such as session options or filesystem
changes. Avoid changing global state, when possible, and reverse any changes
that you do make.
}
\examples{
test_that("trigonometric functions match identities", {
  expect_equal(sin(pi / 4), 1 / sqrt(2))
  expect_equal(cos(pi / 4), 1 / sqrt(2))
  expect_equal(tan(pi / 4), 1)
})

\dontrun{
test_that("trigonometric functions match identities", {
  expect_equal(sin(pi / 4), 1)
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-invisible.R
\name{expect_invisible}
\alias{expect_invisible}
\alias{expect_visible}
\title{Do you expect the result to be (in)visible?}
\usage{
expect_invisible(call, label = NULL)

expect_visible(call, label = NULL)
}
\arguments{
\item{call}{A function call.}

\item{label}{Used to customise failure messages. For expert use only.}
}
\value{
The evaluated \code{call}, invisibly.
}
\description{
Use this to test whether a function returns a visible or invisible
output. Typically you'll use this to check that functions called primarily
for their side-effects return their data argument invisibly.
}
\examples{
expect_invisible(x <- 10)
expect_visible(x)

# Typically you'll assign the result of the expectation so you can
# also check that the value is as you expect.
greet <- function(name) {
  message("Hi ", name)
  invisible(name)
}
out <- expect_invisible(greet("Hadley"))
expect_equal(out, "Hadley")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-summary.R
\name{SummaryReporter}
\alias{SummaryReporter}
\title{Report a summary of failures}
\description{
This is designed for interactive usage: it lets you know which tests have
run successfully and as well as fully reporting information about
failures and errors.

You can use the \code{max_reports} field to control the maximum number
of detailed reports produced by this reporter.

As an additional benefit, this reporter will praise you from time-to-time
if all your tests pass.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edition.R
\name{local_edition}
\alias{local_edition}
\alias{edition_get}
\title{Temporarily change the active testthat edition}
\usage{
local_edition(x, .env = parent.frame())

edition_get()
}
\arguments{
\item{x}{Edition Should be a single integer.}

\item{.env}{Environment that controls scope of changes. For expert use only.}
}
\description{
\code{local_edition()} allows you to temporarily (within a single test or
a single test file) change the active edition of testthat.
\code{edition_get()} allows you to retrieve the currently active edition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-vector.R
\name{expect_vector}
\alias{expect_vector}
\title{Does code return a vector with the expected size and/or prototype?}
\usage{
expect_vector(object, ptype = NULL, size = NULL)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{ptype}{(Optional) Vector prototype to test against. Should be a
size-0 (empty) generalised vector.}

\item{size}{(Optional) Size to check for.}
}
\description{
\code{expect_vector()} is a thin wrapper around \code{\link[vctrs:vec_assert]{vctrs::vec_assert()}}, converting
the results of that function in to the expectations used by testthat. This
means that it used the vctrs of \code{ptype} (prototype) and \code{size}. See
details in \url{https://vctrs.r-lib.org/articles/type-size.html}
}
\examples{
\dontshow{if (requireNamespace("vctrs")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
expect_vector(1:10, ptype = integer(), size = 10)
show_failure(expect_vector(1:10, ptype = integer(), size = 5))
show_failure(expect_vector(1:10, ptype = character(), size = 5))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot-file.R
\name{expect_snapshot_file}
\alias{expect_snapshot_file}
\alias{announce_snapshot_file}
\alias{compare_file_binary}
\alias{compare_file_text}
\title{Do you expect this code to create the same file as last time?}
\usage{
expect_snapshot_file(
  path,
  name = basename(path),
  binary = deprecated(),
  cran = FALSE,
  compare = NULL,
  transform = NULL,
  variant = NULL
)

announce_snapshot_file(path, name = basename(path))

compare_file_binary(old, new)

compare_file_text(old, new)
}
\arguments{
\item{path}{Path to file to snapshot. Optional for
\code{announce_snapshot_file()} if \code{name} is supplied.}

\item{name}{Snapshot name, taken from \code{path} by default.}

\item{binary}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use the
\code{compare} argument instead.}

\item{cran}{Should these expectations be verified on CRAN? By default,
they are not, because snapshot tests tend to be fragile because they
often rely on minor details of dependencies.}

\item{compare}{A function used to compare the snapshot files. It should take
two inputs, the paths to the \code{old} and \code{new} snapshot, and return either
\code{TRUE} or \code{FALSE}. This defaults to \code{compare_file_text} if \code{name} has
extension \code{.r}, \code{.R}, \code{.Rmd}, \code{.md}, or \code{.txt}, and otherwise uses
\code{compare_file_binary}.

\code{compare_file_binary()} compares byte-by-byte and
\code{compare_file_text()} compares lines-by-line, ignoring
the difference between Windows and Mac/Linux line endings.}

\item{transform}{Optionally, a function to scrub sensitive or stochastic
text from the output. Should take a character vector of lines as input
and return a modified character vector as output.}

\item{variant}{If not-\code{NULL}, results will be saved in
\verb{_snaps/\{variant\}/\{test\}/\{name\}}. This allows you to create
different snapshots for different scenarios, like different operating
systems or different R versions.

Note that there's no way to declare all possible variants up front which
means that as soon as you start using variants, you are responsible for
deleting snapshot variants that are no longer used. (testthat will still
delete all variants if you delete the test.)}

\item{old, new}{Paths to old and new snapshot files.}
}
\description{
Whole file snapshot testing is designed for testing objects that don't have
a convenient textual representation, with initial support for images
(\code{.png}, \code{.jpg}, \code{.svg}), data frames (\code{.csv}), and text files
(\code{.R}, \code{.txt}, \code{.json}, ...).

The first time \code{expect_snapshot_file()} is run, it will create
\verb{_snaps/\{test\}/\{name\}.\{ext\}} containing reference output. Future runs will
be compared to this reference: if different, the test will fail and the new
results will be saved in \verb{_snaps/\{test\}/\{name\}.new.\{ext\}}. To review
failures, call \code{\link[=snapshot_review]{snapshot_review()}}.

We generally expect this function to be used via a wrapper that takes care
of ensuring that output is as reproducible as possible, e.g. automatically
skipping tests where it's known that images can't be reproduced exactly.
}
\section{Announcing snapshots}{

testthat automatically detects dangling snapshots that have been
written to the \verb{_snaps} directory but which no longer have
corresponding R code to generate them. These dangling files are
automatically deleted so they don't clutter the snapshot
directory.

This can cause problems if your test is conditionally executed, either
because of an \code{if} statement or a \code{\link[=skip]{skip()}}. To avoid files being deleted in
this case, you can call \code{announce_snapshot_file()} before the conditional
code.

\if{html}{\out{<div class="sourceCode R">}}\preformatted{test_that("can save a file", \{
  if (!can_save()) \{
    announce_snapshot_file(name = "data.txt")
    skip("Can't save file")
  \}
  path <- withr::local_tempfile()
  expect_snapshot_file(save_file(path, mydata()), "data.txt")
\})
}\if{html}{\out{</div>}}
}

\examples{

# To use expect_snapshot_file() you'll typically need to start by writing
# a helper function that creates a file from your code, returning a path
save_png <- function(code, width = 400, height = 400) {
  path <- tempfile(fileext = ".png")
  png(path, width = width, height = height)
  on.exit(dev.off())
  code

  path
}
path <- save_png(plot(1:5))
path

\dontrun{
expect_snapshot_file(save_png(hist(mtcars$mpg)), "plot.png")
}

# You'd then also provide a helper that skips tests where you can't
# be sure of producing exactly the same output.
expect_snapshot_plot <- function(name, code) {
  # Announce the file before touching skips or running `code`. This way,
  # if the skips are active, testthat will not auto-delete the corresponding
  # snapshot file.
  name <- paste0(name, ".png")
  announce_snapshot_file(name = name)

  # Other packages might affect results
  skip_if_not_installed("ggplot2", "2.0.0")
  # Or maybe the output is different on some operating systems
  skip_on_os("windows")
  # You'll need to carefully think about and experiment with these skips

  path <- save_png(code)
  expect_snapshot_file(path, name)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/watcher.R
\name{safe_digest}
\alias{safe_digest}
\title{Compute a digest of a filename, returning NA if the file doesn't
exist.}
\usage{
safe_digest(path)
}
\arguments{
\item{filename}{filename to compute digest on}
}
\value{
a digest of the file, or NA if it doesn't exist.
}
\description{
Compute a digest of a filename, returning NA if the file doesn't
exist.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-known.R
\name{expect_known_output}
\alias{expect_known_output}
\alias{expect_known_value}
\alias{expect_equal_to_reference}
\alias{expect_known_hash}
\title{Expectations: is the output or the value equal to a known good value?}
\usage{
expect_known_output(
  object,
  file,
  update = TRUE,
  ...,
  info = NULL,
  label = NULL,
  print = FALSE,
  width = 80
)

expect_known_value(
  object,
  file,
  update = TRUE,
  ...,
  info = NULL,
  label = NULL,
  version = 2
)

expect_known_hash(object, hash = NULL)
}
\arguments{
\item{file}{File path where known value/output will be stored.}

\item{update}{Should the file be updated? Defaults to \code{TRUE}, with
the expectation that you'll notice changes because of the first failure,
and then see the modified files in git.}

\item{...}{Passed on to \code{\link[waldo:compare]{waldo::compare()}}.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{print}{If \code{TRUE} and the result of evaluating \code{code} is
visible, print the result using \code{testthat_print()}.}

\item{width}{Number of characters per line of output. This does not
inherit from \code{getOption("width")} so that tests always use the same
output width, minimising spurious differences.}

\item{version}{The serialization format version to use. The default, 2, was
the default format from R 1.4.0 to 3.5.3. Version 3 became the default from
R 3.6.0 and can only be read by R versions 3.5.0 and higher.}

\item{hash}{Known hash value. Leave empty and you'll be informed what
to use in the test output.}
}
\description{
For complex printed output and objects, it is often challenging to describe
exactly what you expect to see. \code{expect_known_value()} and
\code{expect_known_output()} provide a slightly weaker guarantee, simply
asserting that the values have not changed since the last time that you ran
them.
}
\details{
These expectations should be used in conjunction with git, as otherwise
there is no way to revert to previous values. Git is particularly useful
in conjunction with \code{expect_known_output()} as the diffs will show you
exactly what has changed.

Note that known values updates will only be updated when running tests
interactively. \verb{R CMD check} clones the package source so any changes to
the reference files will occur in a temporary directory, and will not be
synchronised back to the source package.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{expect_known_output()} and friends are deprecated in the 3rd edition;
please use \code{\link[=expect_snapshot_output]{expect_snapshot_output()}} and friends instead.
}

\examples{
tmp <- tempfile()

# The first run always succeeds
expect_known_output(mtcars[1:10, ], tmp, print = TRUE)

# Subsequent runs will succeed only if the file is unchanged
# This will succeed:
expect_known_output(mtcars[1:10, ], tmp, print = TRUE)

\dontrun{
# This will fail
expect_known_output(mtcars[1:9, ], tmp, print = TRUE)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-list.R
\name{testthat_results}
\alias{testthat_results}
\title{Create a \code{testthat_results} object from the test results
as stored in the ListReporter results field.}
\usage{
testthat_results(results)
}
\arguments{
\item{results}{a list as stored in ListReporter}
}
\value{
its list argument as a \code{testthat_results} object
}
\description{
Create a \code{testthat_results} object from the test results
as stored in the ListReporter results field.
}
\seealso{
ListReporter
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot-reporter.R
\name{local_snapshotter}
\alias{local_snapshotter}
\title{Instantiate local snapshotting context}
\usage{
local_snapshotter(
  snap_dir = "_snaps",
  reporter = SnapshotReporter,
  cleanup = FALSE,
  desc = NULL,
  fail_on_new = NULL,
  frame = caller_env()
)
}
\description{
Needed if you want to run snapshot tests outside of the usual testthat
framework For expert use only.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-constant.R
\name{logical-expectations}
\alias{logical-expectations}
\alias{expect_true}
\alias{expect_false}
\title{Does code return \code{TRUE} or \code{FALSE}?}
\usage{
expect_true(object, info = NULL, label = NULL)

expect_false(object, info = NULL, label = NULL)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}
}
\description{
These are fall-back expectations that you can use when none of the other
more specific expectations apply. The disadvantage is that you may get
a less informative error message.

Attributes are ignored.
}
\examples{
expect_true(2 == 2)
# Failed expectations will throw an error
\dontrun{
expect_true(2 != 2)
}
expect_true(!(2 != 2))
# or better:
expect_false(2 != 2)

a <- 1:3
expect_true(length(a) == 3)
# but better to use more specific expectation, if available
expect_equal(length(a), 3)
}
\seealso{
\code{\link[=is_false]{is_false()}} for complement

Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock2.R
\name{local_mocked_bindings}
\alias{local_mocked_bindings}
\alias{with_mocked_bindings}
\title{Mocking tools}
\usage{
local_mocked_bindings(..., .package = NULL, .env = caller_env())

with_mocked_bindings(code, ..., .package = NULL)
}
\arguments{
\item{...}{Name-value pairs providing new values (typically functions) to
temporarily replace the named bindings.}

\item{.package}{The name of the package where mocked functions should be
inserted. Generally, you should not supply this as it will be automatically
detected when whole package tests are run or when there's one package
under active development (i.e. loaded with \code{\link[pkgload:load_all]{pkgload::load_all()}}).
We don't recommend using this to mock functions in other packages,
as you should not modify namespaces that you don't own.}

\item{.env}{Environment that defines effect scope. For expert use only.}

\item{code}{Code to execute with specified bindings.}
}
\description{
\code{with_mocked_bindings()} and \code{local_mocked_bindings()} provide tools for
"mocking", temporarily redefining a function so that it behaves differently
during tests. This is helpful for testing functions that depend on external
state (i.e. reading a value from a file or a website, or pretending a package
is or isn't installed).

These functions represent a second attempt at bringing mocking to testthat,
incorporating what we've learned from the mockr, mockery, and mockthat
packages.
}
\section{Use}{
There are four places that the function you are trying to mock might
come from:
\itemize{
\item Internal to your package.
\item Imported from an external package via the \code{NAMESPACE}.
\item The base environment.
\item Called from an external package with \code{::}.
}

They are described in turn below.
\subsection{Internal & imported functions}{

You mock internal and imported functions the same way. For example, take
this code:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{some_function <- function() \{
  another_function()
\}
}\if{html}{\out{</div>}}

It doesn't matter whether \code{another_function()} is defined by your package
or you've imported it from a dependency with \verb{@import} or \verb{@importFrom},
you mock it the same way:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{local_mocked_bindings(
  another_function = function(...) "new_value"
)
}\if{html}{\out{</div>}}
}

\subsection{Base functions}{

To mock a function in the base package, you need to make sure that you
have a binding for this function in your package. It's easiest to do this
by binding the value to \code{NULL}. For example, if you wanted to mock
\code{interactive()} in your package, you'd need to include this code somewhere
in your package:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{interactive <- NULL
}\if{html}{\out{</div>}}

Why is this necessary? \code{with_mocked_bindings()} and \code{local_mocked_bindings()}
work by temporarily modifying the bindings within your package's namespace.
When these tests are running inside of \verb{R CMD check} the namespace is locked
which means it's not possible to create new bindings so you need to make sure
that the binding exists already.
}

\subsection{Namespaced calls}{

It's trickier to mock functions in other packages that you call with \code{::}.
For example, take this minor variation:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{some_function <- function() \{
  anotherpackage::another_function()
\}
}\if{html}{\out{</div>}}

To mock this function, you'd need to modify \code{another_function()} inside the
\code{anotherpackage} package. You \emph{can} do this by supplying the \code{.package}
argument to \code{local_mocked_bindings()} but we don't recommend it because
it will affect all calls to \code{anotherpackage::another_function()}, not just
the calls originating in your package. Instead, it's safer to either import
the function into your package, or make a wrapper that you can mock:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{some_function <- function() \{
  my_wrapper()
\}
my_wrapper <- function(...) \{
  anotherpackage::another_function(...)
\}

local_mocked_bindings(
  my_wrapper = function(...) "new_value"
)
}\if{html}{\out{</div>}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-oo.R
\name{local_mocked_r6_class}
\alias{local_mocked_r6_class}
\title{Mock an R6 class}
\usage{
local_mocked_r6_class(
  class,
  public = list(),
  private = list(),
  frame = caller_env()
)
}
\arguments{
\item{class}{An R6 class definition.}

\item{public, private}{A named list of public and private methods/data.}

\item{frame}{Calling frame which determines the scope of the mock.
Only needed when wrapping in another local helper.}
}
\description{
This function allows you to temporarily override an R6 class definition.
It works by creating a subclass then using \code{\link[=local_mocked_bindings]{local_mocked_bindings()}} to
temporarily replace the original definition. This means that it will not
affect subclasses of the original class; please file an issue if you need
this.

Learn more about mocking in \code{vignette("mocking")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-condition.R
\name{is_informative_error}
\alias{is_informative_error}
\title{Is an error informative?}
\usage{
is_informative_error(x, ...)
}
\arguments{
\item{x}{An error object.}

\item{...}{These dots are for future extensions and must be empty.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{is_informative_error()} is a generic predicate that indicates
whether testthat users should explicitly test for an error
class. Since we no longer recommend you do that, this generic
has been deprecated.
}
\details{
A few classes are hard-coded as uninformative:
\itemize{
\item \code{simpleError}
\item \code{rlang_error} unless a subclass is detected
\item \code{Rcpp::eval_error}
\item \code{Rcpp::exception}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/context.R
\name{context}
\alias{context}
\title{Describe the context of a set of tests.}
\usage{
context(desc)
}
\arguments{
\item{desc}{description of context.  Should start with a capital letter.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

Use of \code{context()} is no longer recommended. Instead omit it, and messages
will use the name of the file instead. This ensures that the context and
test file name are always in sync.

A context defines a set of tests that test related functionality.  Usually
you will have one context per file, but you may have multiple contexts
in a single file if you so choose.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{context()} is deprecated in the third edition, and the equivalent
information is instead recorded by the test file name.
}

\examples{
context("String processing")
context("Remote procedure calls")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-state.R
\name{set_state_inspector}
\alias{set_state_inspector}
\title{Check for global state changes}
\usage{
set_state_inspector(callback, tolerance = testthat_tolerance())
}
\arguments{
\item{callback}{Either a zero-argument function that returns an object
capturing global state that you're interested in, or \code{NULL}.}

\item{tolerance}{If non-\code{NULL}, used as threshold for ignoring small
floating point difference when comparing numeric vectors. Using any
non-\code{NULL} value will cause integer and double vectors to be compared
based on their values, not their types, and will ignore the difference
between \code{NaN} and \code{NA_real_}.

It uses the same algorithm as \code{\link[=all.equal]{all.equal()}}, i.e., first we generate
\code{x_diff} and \code{y_diff} by subsetting \code{x} and \code{y} to look only locations
with differences. Then we check that
\code{mean(abs(x_diff - y_diff)) / mean(abs(y_diff))} (or just
\code{mean(abs(x_diff - y_diff))} if \code{y_diff} is small) is less than
\code{tolerance}.}
}
\description{
One of the most pernicious challenges to debug is when a test runs fine
in your test suite, but fails when you run it interactively (or similarly,
it fails randomly when running your tests in parallel). One of the most
common causes of this problem is accidentally changing global state in a
previous test (e.g. changing an option, an environment variable, or the
working directory). This is hard to debug, because it's very hard to figure
out which test made the change.

Luckily testthat provides a tool to figure out if tests are changing global
state. You can register a state inspector with \code{set_state_inspector()} and
testthat will run it before and after each test, store the results, then
report if there are any differences. For example, if you wanted to see if
any of your tests were changing options or environment variables, you could
put this code in \code{tests/testthat/helper-state.R}:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{set_state_inspector(function() \{
  list(
    options = options(),
    envvars = Sys.getenv()
  )
\})
}\if{html}{\out{</div>}}

(You might discover other packages outside your control are changing
the global state, in which case you might want to modify this function
to ignore those values.)

Other problems that can be troublesome to resolve are CRAN check notes that
report things like connections being left open. You can easily debug
that problem with:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{set_state_inspector(function() \{
  getAllConnections()
\})
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-multi.R
\name{MultiReporter}
\alias{MultiReporter}
\title{Multi reporter: combine several reporters in one.}
\description{
This reporter is useful to use several reporters at the same time, e.g.
adding a custom reporter without removing the current one.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{expect_snapshot}
\alias{expect_snapshot}
\title{Snapshot testing}
\usage{
expect_snapshot(
  x,
  cran = FALSE,
  error = FALSE,
  transform = NULL,
  variant = NULL,
  cnd_class = FALSE
)
}
\arguments{
\item{x}{Code to evaluate.}

\item{cran}{Should these expectations be verified on CRAN? By default,
they are not, because snapshot tests tend to be fragile because they
often rely on minor details of dependencies.}

\item{error}{Do you expect the code to throw an error? The expectation
will fail (even on CRAN) if an unexpected error is thrown or the
expected error is not thrown.}

\item{transform}{Optionally, a function to scrub sensitive or stochastic
text from the output. Should take a character vector of lines as input
and return a modified character vector as output.}

\item{variant}{If non-\code{NULL}, results will be saved in
\verb{_snaps/\{variant\}/\{test.md\}}, so \code{variant} must be a single string
suitable for use as a directory name.

You can use variants to deal with cases where the snapshot output varies
and you want to capture and test the variations. Common use cases include
variations for operating system, R version, or version of key dependency.
Variants are an advanced feature. When you use them, you'll need to
carefully think about your testing strategy to ensure that all important
variants are covered by automated tests, and ensure that you have a way
to get snapshot changes out of your CI system and back into the repo.}

\item{cnd_class}{Whether to include the class of messages,
warnings, and errors in the snapshot. Only the most specific
class is included, i.e. the first element of \code{class(cnd)}.}
}
\description{
Snapshot tests (aka golden tests) are similar to unit tests except that the
expected result is stored in a separate file that is managed by testthat.
Snapshot tests are useful for when the expected value is large, or when
the intent of the code is something that can only be verified by a human
(e.g. this is a useful error message). Learn more in
\code{vignette("snapshotting")}.

\code{expect_snapshot()} runs code as if you had executed it at the console, and
records the results, including output, messages, warnings, and errors.
If you just want to compare the result, try \code{\link[=expect_snapshot_value]{expect_snapshot_value()}}.
}
\section{Workflow}{

The first time that you run a snapshot expectation it will run \code{x},
capture the results, and record them in \verb{tests/testthat/_snaps/\{test\}.md}.
Each test file gets its own snapshot file, e.g. \code{test-foo.R} will get
\verb{_snaps/foo.md}.

It's important to review the Markdown files and commit them to git. They are
designed to be human readable, and you should always review new additions
to ensure that the salient information has been captured. They should also
be carefully reviewed in pull requests, to make sure that snapshots have
updated in the expected way.

On subsequent runs, the result of \code{x} will be compared to the value stored
on disk. If it's different, the expectation will fail, and a new file
\verb{_snaps/\{test\}.new.md} will be created. If the change was deliberate,
you can approve the change with \code{\link[=snapshot_accept]{snapshot_accept()}} and then the tests will
pass the next time you run them.

Note that snapshotting can only work when executing a complete test file
(with \code{\link[=test_file]{test_file()}}, \code{\link[=test_dir]{test_dir()}}, or friends) because there's otherwise
no way to figure out the snapshot path. If you run snapshot tests
interactively, they'll just display the current value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/watcher.R
\name{compare_state}
\alias{compare_state}
\title{Compare two directory states.}
\usage{
compare_state(old, new)
}
\arguments{
\item{old}{previous state}

\item{new}{current state}
}
\value{
list containing number of changes and files which have been
\code{added}, \code{deleted} and \code{modified}
}
\description{
Compare two directory states.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-inheritance.R
\name{expect_is}
\alias{expect_is}
\title{Do you expect to inherit from this class?}
\usage{
expect_is(object, class, info = NULL, label = NULL)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{class}{Class name passed to \code{inherits()}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{expect_is()} is an older form that uses \code{\link[=inherits]{inherits()}} without checking
whether \code{x} is S3, S4, or neither. Instead, I'd recommend using
\code{\link[=expect_type]{expect_type()}}, \code{\link[=expect_s3_class]{expect_s3_class()}}, or \code{\link[=expect_s4_class]{expect_s4_class()}} to more clearly
convey your intent.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{expect_is()} is formally deprecated in the 3rd edition.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock2-helpers.R
\name{mock_output_sequence}
\alias{mock_output_sequence}
\title{Mock a sequence of output from a function}
\usage{
mock_output_sequence(..., recycle = FALSE)
}
\arguments{
\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}> Values to return in sequence.}

\item{recycle}{whether to recycle. If \code{TRUE}, once all values have been returned,
they will be returned again in sequence.}
}
\value{
A function that you can use within \code{local_mocked_bindings()} and
\code{with_mocked_bindings()}
}
\description{
Specify multiple return values for mocking
}
\examples{
# inside local_mocked_bindings()
\dontrun{
local_mocked_bindings(readline = mock_output_sequence("3", "This is a note", "n"))
}
# for understanding
mocked_sequence <- mock_output_sequence("3", "This is a note", "n")
mocked_sequence()
mocked_sequence()
mocked_sequence()
try(mocked_sequence())
recycled_mocked_sequence <- mock_output_sequence(
  "3", "This is a note", "n",
  recycle = TRUE
)
recycled_mocked_sequence()
recycled_mocked_sequence()
recycled_mocked_sequence()
recycled_mocked_sequence()
}
\seealso{
Other mocking: 
\code{\link{local_mocked_bindings}()}
}
\concept{mocking}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{testthat_tolerance}
\alias{testthat_tolerance}
\title{Default numeric tolerance}
\usage{
testthat_tolerance()
}
\description{
testthat's default numeric tolerance is 1.4901161 × 10\if{html}{\out{<sup>}}-8\if{html}{\out{</sup>}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-zzz.R
\name{find_reporter}
\alias{find_reporter}
\title{Find reporter object given name or object.}
\usage{
find_reporter(reporter)
}
\arguments{
\item{reporter}{name of reporter(s), or reporter object(s)}
}
\description{
If not found, will return informative error message.
Pass a character vector to create a \link{MultiReporter} composed
of individual reporters.
Will return null if given NULL.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-progress.R
\name{set_max_fails}
\alias{set_max_fails}
\title{Set maximum number of test failures allowed before aborting the run}
\usage{
set_max_fails(n)
}
\arguments{
\item{n}{Maximum number of failures allowed.}
}
\description{
This sets the \code{TESTTHAT_MAX_FAILS} env var which will affect both the
current R process and any processes launched from it.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teardown.R
\name{teardown}
\alias{teardown}
\alias{setup}
\title{Run code before/after tests}
\usage{
teardown(code, env = parent.frame())

setup(code, env = parent.frame())
}
\arguments{
\item{code}{Code to evaluate}

\item{env}{Environment in which code will be evaluated. For expert
use only.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

We no longer recommend using \code{setup()} and \code{teardown()}; instead
we think it's better practice to use a \strong{test fixture} as described in
\code{vignette("test-fixtures")}.

Code in a \code{setup()} block is run immediately in a clean environment.
Code in a \code{teardown()} block is run upon completion of a test file,
even if it exits with an error. Multiple calls to \code{teardown()} will be
executed in the order they were created.
}
\examples{
\dontrun{
# Old approach
tmp <- tempfile()
setup(writeLines("some test data", tmp))
teardown(unlink(tmp))
}

# Now recommended:
local_test_data <- function(env = parent.frame()) {
  tmp <- tempfile()
  writeLines("some test data", tmp)
  withr::defer(unlink(tmp), env)

  tmp
}
# Then call local_test_data() in your tests
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot-value.R
\name{expect_snapshot_value}
\alias{expect_snapshot_value}
\title{Snapshot testing for values}
\usage{
expect_snapshot_value(
  x,
  style = c("json", "json2", "deparse", "serialize"),
  cran = FALSE,
  tolerance = testthat_tolerance(),
  ...,
  variant = NULL
)
}
\arguments{
\item{x}{Code to evaluate.}

\item{style}{Serialization style to use:
\itemize{
\item \code{json} uses \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} and \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}. This
produces the simplest output but only works for relatively simple
objects.
\item \code{json2} uses \code{\link[jsonlite:serializeJSON]{jsonlite::serializeJSON()}} and \code{\link[jsonlite:serializeJSON]{jsonlite::unserializeJSON()}}
which are more verbose but work for a wider range of type.
\item \code{deparse} uses \code{\link[=deparse]{deparse()}}, which generates a depiction of the object
using R code.
\item \code{serialize()} produces a binary serialization of the object using
\code{\link[=serialize]{serialize()}}. This is all but guaranteed to work for any R object,
but produces a completely opaque serialization.
}}

\item{cran}{Should these expectations be verified on CRAN? By default,
they are not, because snapshot tests tend to be fragile because they
often rely on minor details of dependencies.}

\item{tolerance}{Numerical tolerance: any differences (in the sense of
\code{\link[base:all.equal]{base::all.equal()}}) smaller than this value will be ignored.

The default tolerance is \code{sqrt(.Machine$double.eps)}, unless long doubles
are not available, in which case the test is skipped.}

\item{...}{Passed on to \code{\link[waldo:compare]{waldo::compare()}} so you can control the details of
the comparison.}

\item{variant}{If non-\code{NULL}, results will be saved in
\verb{_snaps/\{variant\}/\{test.md\}}, so \code{variant} must be a single string
suitable for use as a directory name.

You can use variants to deal with cases where the snapshot output varies
and you want to capture and test the variations. Common use cases include
variations for operating system, R version, or version of key dependency.
Variants are an advanced feature. When you use them, you'll need to
carefully think about your testing strategy to ensure that all important
variants are covered by automated tests, and ensure that you have a way
to get snapshot changes out of your CI system and back into the repo.}
}
\description{
Captures the result of function, flexibly serializing it into a text
representation that's stored in a snapshot file. See \code{\link[=expect_snapshot]{expect_snapshot()}}
for more details on snapshot testing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testthat-package.R
\docType{package}
\name{testthat-package}
\alias{testthat}
\alias{testthat-package}
\title{An R package to make testing fun!}
\description{
Try the example below. Have a look at the references and learn more
from function documentation such as \code{\link[=test_that]{test_that()}}.
}
\section{Options}{

\itemize{
\item \code{testthat.use_colours}: Should the output be coloured? (Default: \code{TRUE}).
\item \code{testthat.summary.max_reports}: The maximum number of detailed test
reports printed for the summary reporter (default: 10).
\item \code{testthat.summary.omit_dots}: Omit progress dots in the summary reporter
(default: \code{FALSE}).
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://testthat.r-lib.org}
  \item \url{https://github.com/r-lib/testthat}
  \item Report bugs at \url{https://github.com/r-lib/testthat/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
  \item R Core team (Implementation of utils::recover()) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{extract_test}
\alias{extract_test}
\title{Extract a reprex from a failed expectation}
\usage{
extract_test(location, path = stdout(), package = Sys.getenv("TESTTHAT_PKG"))
}
\arguments{
\item{location}{A string giving the location in the form
\verb{FILE:LINE[:COLUMN]}.}

\item{path}{Path to write the reprex to. Defaults to \code{stdout()}.}

\item{package}{If supplied, will be used to construct a test environment
for the extracted code.}
}
\value{
This function is called for its side effect of rendering a
reprex to \code{path}. This function will never error: if extraction
fails, the error message will be written to \code{path}.
}
\description{
\code{extract_test()} creates a minimal reprex for a failed expectation.
It extracts all non-test code before the failed expectation as well as
all code inside the test up to and including the failed expectation.

This is particularly useful when you're debugging test failures in
someone else's package.
}
\examples{
# If you see a test failure like this:
# -- Failure (test-extract.R:46:3): errors if can't find test -------------
# Expected FALSE to be TRUE.
# Differences:
# `actual`:   FALSE
# `expected`: TRUE

# You can run this:
\dontrun{extract_test("test-extract.R:46:3")}
# to see just the code needed to reproduce the failure
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/source.R
\name{source_file}
\alias{source_file}
\alias{source_dir}
\alias{source_test_helpers}
\alias{source_test_setup}
\alias{source_test_teardown}
\title{Source a file, directory of files, or various important subsets}
\usage{
source_file(
  path,
  env = test_env(),
  chdir = TRUE,
  desc = NULL,
  wrap = TRUE,
  shuffle = FALSE,
  error_call = caller_env()
)

source_dir(
  path,
  pattern = "\\\\.[rR]$",
  env = test_env(),
  chdir = TRUE,
  wrap = TRUE,
  shuffle = FALSE
)

source_test_helpers(path = "tests/testthat", env = test_env())

source_test_setup(path = "tests/testthat", env = test_env())

source_test_teardown(path = "tests/testthat", env = test_env())
}
\arguments{
\item{path}{Path to files.}

\item{env}{Environment in which to evaluate code.}

\item{chdir}{Change working directory to \code{dirname(path)}?}

\item{desc}{A character vector used to filter tests. This is used to
(recursively) filter the content of the file, so that only the non-test
code up to and including the matching test is run.}

\item{wrap}{Automatically wrap all code within \code{\link[=test_that]{test_that()}}? This ensures
that all expectations are reported, even if outside a test block.}

\item{shuffle}{If \code{TRUE}, randomly reorder the top-level expressions
in the file.}

\item{pattern}{Regular expression used to filter files.}
}
\description{
These are used by \code{\link[=test_dir]{test_dir()}} and friends
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprec-condition.R
\name{capture_condition}
\alias{capture_condition}
\alias{capture_error}
\alias{capture_expectation}
\alias{capture_message}
\alias{capture_warning}
\alias{capture_messages}
\alias{capture_warnings}
\title{Capture conditions, including messages, warnings, expectations, and errors.}
\usage{
capture_condition(code, entrace = FALSE)

capture_error(code, entrace = FALSE)

capture_expectation(code, entrace = FALSE)

capture_message(code, entrace = FALSE)

capture_warning(code, entrace = FALSE)

capture_messages(code)

capture_warnings(code, ignore_deprecation = FALSE)
}
\arguments{
\item{code}{Code to evaluate}

\item{entrace}{Whether to add a \link[rlang:trace_back]{backtrace} to
the captured condition.}
}
\value{
Singular functions (\code{capture_condition}, \code{capture_expectation} etc)
return a condition object. \code{capture_messages()} and \code{capture_warnings}
return a character vector of message text.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

These functions allow you to capture the side-effects of a function call
including printed output, messages and warnings. We no longer recommend
that you use these functions, instead relying on the \code{\link[=expect_message]{expect_message()}}
and friends to bubble up unmatched conditions. If you just want to silence
unimportant warnings, use \code{\link[=suppressWarnings]{suppressWarnings()}}.
}
\examples{
f <- function() {
  message("First")
  warning("Second")
  message("Third")
}

capture_message(f())
capture_messages(f())

capture_warning(f())
capture_warnings(f())

# Condition will capture anything
capture_condition(f())
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-all.R
\name{expect_all_equal}
\alias{expect_all_equal}
\alias{expect_all_true}
\alias{expect_all_false}
\title{Do you expect every value in a vector to have this value?}
\usage{
expect_all_equal(object, expected)

expect_all_true(object)

expect_all_false(object)
}
\arguments{
\item{object, expected}{Computation and value to compare it to.

Both arguments supports limited unquoting to make it easier to generate
readable failures within a function or for loop. See \link{quasi_label} for
more details.}
}
\description{
These expectations are similar to \code{expect_true(all(x == "x"))},
\code{expect_true(all(x))} and \code{expect_true(all(!x))} but give more informative
failure messages if the expectations are not met.
}
\examples{
x1 <- c(1, 1, 1, 1, 1, 1)
expect_all_equal(x1, 1)

x2 <- c(1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2)
show_failure(expect_all_equal(x2, 1))

# expect_all_true() and expect_all_false() are helpers for common cases
set.seed(1016)
show_failure(expect_all_true(rpois(100, 10) < 20))
show_failure(expect_all_false(rpois(100, 10) > 20))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-stop.R
\name{StopReporter}
\alias{StopReporter}
\title{Error if any test fails}
\description{
The default reporter used when \code{\link[=expect_that]{expect_that()}} is run interactively.
It responds by displaying a summary of the number of successes and failures
and \code{\link[=stop]{stop()}}ping on if there are any failures.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-known.R
\name{expect_output_file}
\alias{expect_output_file}
\title{Do you expect the output/result to equal a known good value?}
\usage{
expect_output_file(
  object,
  file,
  update = TRUE,
  ...,
  info = NULL,
  label = NULL,
  print = FALSE,
  width = 80
)
}
\description{
\code{expect_output_file()} behaves identically to \code{\link[=expect_known_output]{expect_known_output()}}.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{expect_output_file()} is deprecated in the 3rd edition;
please use \code{\link[=expect_snapshot_output]{expect_snapshot_output()}} and friends instead.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-zzz.R
\name{reporter-accessors}
\alias{reporter-accessors}
\alias{set_reporter}
\alias{get_reporter}
\alias{with_reporter}
\title{Get and set active reporter.}
\usage{
set_reporter(reporter)

get_reporter()

with_reporter(reporter, code, start_end_reporter = TRUE)
}
\arguments{
\item{reporter}{Reporter to use to summarise output. Can be supplied
as a string (e.g. "summary") or as an R6 object
(e.g. \code{SummaryReporter$new()}).

See \link{Reporter} for more details and a list of built-in reporters.}

\item{code}{Code to execute.}

\item{start_end_reporter}{Should the reporters \code{start_reporter()} and
\code{end_reporter()} methods be called? For expert use only.}
}
\value{
\code{with_reporter()} invisible returns the reporter active when \code{code}
was evaluated.
}
\description{
\code{get_reporter()} and \code{set_reporter()} access and modify the current "active"
reporter. Generally, these functions should not be called directly; instead
use \code{with_reporter()} to temporarily change, then reset, the active reporter.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter.R
\name{Reporter}
\alias{Reporter}
\title{Manage test reporting}
\description{
The job of a reporter is to aggregate the results from files, tests, and
expectations and display them in an informative way. Every testthat function
that runs multiple tests provides a \code{reporter} argument which you can
use to override the default (which is selected by \code{\link[=default_reporter]{default_reporter()}}).
}
\details{
You only need to use this \code{Reporter} object directly if you are creating
a new reporter. Currently, creating new Reporters is undocumented,
so if you want to create your own, you'll need to make sure that you're
familiar with \href{https://adv-r.hadley.nz/R6.html}{R6} and then need read the
source code for a few.
}
\examples{
path <- testthat_example("success")

test_file(path)
# Override the default by supplying the name of a reporter
test_file(path, reporter = "minimal")
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{SilentReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-that.R
\name{not}
\alias{not}
\title{Negate an expectation}
\usage{
not(f)
}
\arguments{
\item{f}{an existing expectation function}
}
\description{
This negates an expectation, making it possible to express that you
want the opposite of a standard expectation. This function is deprecated
and will be removed in a future version.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quasi-label.R
\name{quasi_label}
\alias{quasi_label}
\title{Quasi-labelling}
\usage{
quasi_label(quo, label = NULL, arg = NULL)
}
\arguments{
\item{quo}{A quosure created by \code{rlang::enquo()}.}

\item{label}{An optional label to override the default. This is
only provided for internal usage. Modern expectations should not
include a \code{label} parameter.}

\item{arg}{Argument name shown in error message if \code{quo} is missing.}
}
\value{
A list containing two elements:
\item{val}{The evaluate value of \code{quo}}
\item{lab}{The quasiquoted label generated from \code{quo}}
}
\description{
The first argument to every \code{expect_} function can use unquoting to
construct better labels. This makes it easy to create informative labels when
expectations are used inside a function or a for loop. \code{quasi_label()} wraps
up the details, returning the expression and label.
}
\section{Limitations}{

Because all \code{expect_} function use unquoting to generate more informative
labels, you can not use unquoting for other purposes. Instead, you'll need
to perform all other unquoting outside of the expectation and only test
the results.
}

\examples{
f <- function(i) if (i > 3) i * 9 else i * 10
i <- 10

# This sort of expression commonly occurs inside a for loop or function
# And the failure isn't helpful because you can't see the value of i
# that caused the problem:
show_failure(expect_equal(f(i), i * 10))

# To overcome this issue, testthat allows you to unquote expressions using
# !!. This causes the failure message to show the value rather than the
# variable name
show_failure(expect_equal(f(!!i), !!(i * 10)))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-check.R
\name{CheckReporter}
\alias{CheckReporter}
\title{Report results for \verb{R CMD check}}
\description{
\verb{R CMD check} displays only the last 13 lines of the result, so this
report is designed to ensure that you see something useful there.
}
\seealso{
Other reporters: 
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-comparison.R
\name{comparison-expectations}
\alias{comparison-expectations}
\alias{expect_lt}
\alias{expect_lte}
\alias{expect_gt}
\alias{expect_gte}
\title{Do you expect a value bigger or smaller than this?}
\usage{
expect_lt(object, expected, label = NULL, expected.label = NULL)

expect_lte(object, expected, label = NULL, expected.label = NULL)

expect_gt(object, expected, label = NULL, expected.label = NULL)

expect_gte(object, expected, label = NULL, expected.label = NULL)
}
\arguments{
\item{object, expected}{A value to compare and its expected bound.}

\item{label, expected.label}{Used to customise failure messages. For expert
use only.}
}
\description{
These functions compare values of comparable data types, such as numbers,
dates, and times.
}
\examples{
a <- 9
expect_lt(a, 10)

\dontrun{
expect_lt(11, 10)
}

a <- 11
expect_gt(a, 10)
\dontrun{
expect_gt(9, 10)
}
}
\seealso{
Other expectations: 
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-condition.R
\name{expect_error}
\alias{expect_error}
\alias{expect_warning}
\alias{expect_message}
\alias{expect_condition}
\title{Does code throw an error, warning, message, or other condition?}
\usage{
expect_error(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  inherit = TRUE,
  info = NULL,
  label = NULL
)

expect_warning(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  inherit = TRUE,
  all = FALSE,
  info = NULL,
  label = NULL
)

expect_message(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  inherit = TRUE,
  all = FALSE,
  info = NULL,
  label = NULL
)

expect_condition(
  object,
  regexp = NULL,
  class = NULL,
  ...,
  inherit = TRUE,
  info = NULL,
  label = NULL
)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{regexp}{Regular expression to test against.
\itemize{
\item A character vector giving a regular expression that must match the
error message.
\item If \code{NULL}, the default, asserts that there should be an error,
but doesn't test for a specific value.
\item If \code{NA}, asserts that there should be no errors, but we now recommend
using \code{\link[=expect_no_error]{expect_no_error()}} and friends instead.
}

Note that you should only use \code{message} with errors/warnings/messages
that you generate. Avoid tests that rely on the specific text generated by
another package since this can easily change. If you do need to test text
generated by another package, either protect the test with \code{skip_on_cran()}
or use \code{expect_snapshot()}.}

\item{class}{Instead of supplying a regular expression, you can also supply
a class name. This is useful for "classed" conditions.}

\item{...}{
  Arguments passed on to \code{\link[=expect_match]{expect_match}}
  \describe{
    \item{\code{fixed}}{If \code{TRUE}, treats \code{regexp} as a string to be matched exactly
(not a regular expressions). Overrides \code{perl}.}
    \item{\code{perl}}{logical.  Should Perl-compatible regexps be used?}
  }}

\item{inherit}{Whether to match \code{regexp} and \code{class} across the
ancestry of chained errors.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}

\item{all}{\emph{DEPRECATED} If you need to test multiple warnings/messages
you now need to use multiple calls to \code{expect_message()}/
\code{expect_warning()}}
}
\value{
If \code{regexp = NA}, the value of the first argument; otherwise
the captured condition.
}
\description{
\code{expect_error()}, \code{expect_warning()}, \code{expect_message()}, and
\code{expect_condition()} check that code throws an error, warning, message,
or condition with a message that matches \code{regexp}, or a class that inherits
from \code{class}. See below for more details.

In the 3rd edition, these functions match (at most) a single condition. All
additional and non-matching (if \code{regexp} or \code{class} are used) conditions
will bubble up outside the expectation. If these additional conditions
are important you'll need to catch them with additional
\code{expect_message()}/\code{expect_warning()} calls; if they're unimportant you
can ignore with \code{\link[=suppressMessages]{suppressMessages()}}/\code{\link[=suppressWarnings]{suppressWarnings()}}.

It can be tricky to test for a combination of different conditions,
such as a message followed by an error. \code{\link[=expect_snapshot]{expect_snapshot()}} is
often an easier alternative for these more complex cases.
}
\section{Testing \code{message} vs \code{class}}{

When checking that code generates an error, it's important to check that the
error is the one you expect. There are two ways to do this. The first
way is the simplest: you just provide a \code{regexp} that match some fragment
of the error message. This is easy, but fragile, because the test will
fail if the error message changes (even if its the same error).

A more robust way is to test for the class of the error, if it has one.
You can learn more about custom conditions at
\url{https://adv-r.hadley.nz/conditions.html#custom-conditions}, but in
short, errors are S3 classes and you can generate a custom class and check
for it using \code{class} instead of \code{regexp}.

If you are using \code{expect_error()} to check that an error message is
formatted in such a way that it makes sense to a human, we recommend
using \code{\link[=expect_snapshot]{expect_snapshot()}} instead.
}

\examples{
# Errors ------------------------------------------------------------------
f <- function() stop("My error!")
expect_error(f())
expect_error(f(), "My error!")

# You can use the arguments of grepl to control the matching
expect_error(f(), "my error!", ignore.case = TRUE)

# Note that `expect_error()` returns the error object so you can test
# its components if needed
err <- expect_error(rlang::abort("a", n = 10))
expect_equal(err$n, 10)

# Warnings ------------------------------------------------------------------
f <- function(x) {
  if (x < 0) {
    warning("*x* is already negative")
    return(x)
  }
  -x
}
expect_warning(f(-1))
expect_warning(f(-1), "already negative")
expect_warning(f(1), NA)

# To test message and output, store results to a variable
expect_warning(out <- f(-1), "already negative")
expect_equal(out, -1)

# Messages ------------------------------------------------------------------
f <- function(x) {
  if (x < 0) {
    message("*x* is already negative")
    return(x)
  }

  -x
}
expect_message(f(-1))
expect_message(f(-1), "already negative")
expect_message(f(1), NA)
}
\seealso{
\code{\link[=expect_no_error]{expect_no_error()}}, \code{expect_no_warning()},
\code{expect_no_message()}, and \code{expect_no_condition()} to assert
that code runs without errors/warnings/messages/conditions.

Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make-expectation.R
\name{make_expectation}
\alias{make_expectation}
\title{Make an equality test.}
\usage{
make_expectation(x, expectation = "equals")
}
\arguments{
\item{x}{a vector of values}

\item{expectation}{the type of equality you want to test for
(\code{"equals"}, \code{"is_equivalent_to"}, \code{"is_identical_to"})}
}
\description{
This a convenience function to make a expectation that checks that
input stays the same.
}
\examples{
x <- 1:10
make_expectation(x)

make_expectation(mtcars$mpg)

df <- data.frame(x = 2)
make_expectation(df)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-package.R
\name{test_package}
\alias{test_package}
\alias{test_check}
\alias{test_local}
\title{Run all tests in a package}
\usage{
test_package(package, reporter = check_reporter(), ...)

test_check(package, reporter = check_reporter(), ...)

test_local(path = ".", reporter = NULL, ..., load_package = "source")
}
\arguments{
\item{package}{If these tests belong to a package, the name of the package.}

\item{reporter}{Reporter to use to summarise output. Can be supplied
as a string (e.g. "summary") or as an R6 object
(e.g. \code{SummaryReporter$new()}).

See \link{Reporter} for more details and a list of built-in reporters.}

\item{...}{Additional arguments passed to \code{\link[=test_dir]{test_dir()}}}

\item{path}{Path to directory containing tests.}

\item{load_package}{Strategy to use for load package code:
\itemize{
\item "none", the default, doesn't load the package.
\item "installed", uses \code{\link[=library]{library()}} to load an installed package.
\item "source", uses \code{\link[pkgload:load_all]{pkgload::load_all()}} to a source package.
To configure the arguments passed to \code{load_all()}, add this
field in your DESCRIPTION file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Config/testthat/load-all: list(export_all = FALSE, helpers = FALSE)
}\if{html}{\out{</div>}}
}}
}
\value{
A list (invisibly) containing data about the test results.
}
\description{
\itemize{
\item \code{test_local()} tests a local source package.
\item \code{test_package()} tests an installed package.
\item \code{test_check()} checks a package during \verb{R CMD check}.
}

See \code{vignette("special-files")} to learn about the various files that
testthat works with.
}
\section{\verb{R CMD check}}{

To run testthat automatically from \verb{R CMD check}, make sure you have
a \code{tests/testthat.R} that contains:

\if{html}{\out{<div class="sourceCode">}}\preformatted{library(testthat)
library(yourpackage)

test_check("yourpackage")
}\if{html}{\out{</div>}}
}

\section{Environments}{

Each test is run in a clean environment to keep tests as isolated as
possible. For package tests, that environment inherits from the package's
namespace environment, so that tests can access internal functions
and objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/watcher.R
\name{dir_state}
\alias{dir_state}
\title{Capture the state of a directory.}
\usage{
dir_state(path, pattern = NULL, hash = TRUE)
}
\arguments{
\item{path}{path to directory}

\item{pattern}{regular expression with which to filter files}

\item{hash}{use hash (slow but accurate) or time stamp (fast but less
accurate)}
}
\description{
Capture the state of a directory.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-that.R
\name{fail}
\alias{fail}
\alias{pass}
\title{Declare that an expectation either passes or fails}
\usage{
fail(
  message = "Failure has been forced",
  info = NULL,
  srcref = NULL,
  trace_env = caller_env(),
  trace = NULL
)

pass()
}
\arguments{
\item{message}{A character vector describing the failure. The
first element should describe the expected value, and the second (and
optionally subsequence) elements should describe what was actually seen.}

\item{info}{Character vector continuing additional information. Included
for backward compatibility only and new expectations should not use it.}

\item{srcref}{Location of the failure. Should only needed to be explicitly
supplied when you need to forward a srcref captured elsewhere.}

\item{trace_env}{If \code{trace} is not specified, this is used to generate an
informative traceback for failures. You should only need to set this if
you're calling \code{fail()} from a helper function; see
\code{vignette("custom-expectation")} for details.}

\item{trace}{An optional backtrace created by \code{\link[rlang:trace_back]{rlang::trace_back()}}.
When supplied, the expectation is displayed with the backtrace.
Expert use only.}
}
\description{
These are the primitives that you can use to implement your own expectations.
Every path through an expectation should either call \code{pass()}, \code{fail()},
or throw an error (e.g. if the arguments are invalid). Expectations should
always return \code{invisible(act$val)}.

Learn more about creating your own expectations in
\code{vignette("custom-expectation")}.
}
\examples{
expect_length <- function(object, n) {
  act <- quasi_label(rlang::enquo(object), arg = "object")

  act_n <- length(act$val)
  if (act_n != n) {
    fail(sprintf("\%s has length \%i, not length \%i.", act$lab, act_n, n))
  } else {
    pass()
  }

  invisible(act$val)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-llm.R
\name{LlmReporter}
\alias{LlmReporter}
\title{Report test progress for LLMs}
\description{
\code{LlmReporter} is designed for use with Large Language Models (LLMs).
It reports problems (warnings, skips, errors, and failures) as they
occur and the total number of successes at the end.

\code{LlmReporter} is used by default when tests are run by a coding agent.
Currently we detect Claude Code, Cursor, and Gemini CLI.
If using another tool, configure it to set env var \code{AGENT=1}.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-rstudio.R
\name{RStudioReporter}
\alias{RStudioReporter}
\title{Test reporter: RStudio}
\description{
This reporter is designed for output to RStudio. It produces results in
any easily parsed form.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-compiled-code.R
\name{expect_cpp_tests_pass}
\alias{expect_cpp_tests_pass}
\alias{run_cpp_tests}
\title{Do C++ tests past?}
\usage{
expect_cpp_tests_pass(package)

run_cpp_tests(package)
}
\arguments{
\item{package}{The name of the package to test.}
}
\description{
Test compiled code in the package \code{package}. A call to this function will
automatically be generated for you in \code{tests/testthat/test-cpp.R} after
calling \code{\link[=use_catch]{use_catch()}}; you should not need to manually call this expectation
yourself.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-junit.R
\name{JunitReporter}
\alias{JunitReporter}
\title{Report results in jUnit XML format}
\description{
This reporter includes detailed results about each test and summaries,
written to a file (or stdout) in jUnit XML format. This can be read by
the Jenkins Continuous Integration System to report on a dashboard etc.
Requires the \emph{xml2} package.

To fit into the jUnit structure, \code{context()} becomes the \verb{<testsuite>}
name as well as the base of the \verb{<testcase> classname}. The
\code{test_that()} name becomes the rest of the \verb{<testcase> classname}.
The deparsed \code{expect_that()} call becomes the \verb{<testcase>} name.
On failure, the message goes into the \verb{<failure>} node message
argument (first line only) and into its text content (full message).
Execution time and some other details are also recorded.

References for the jUnit XML format:
\url{https://github.com/testmoapp/junitxml}
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-progress.R
\name{ProgressReporter}
\alias{ProgressReporter}
\alias{CompactProgressReporter}
\alias{ParallelProgressReporter}
\title{Report progress interactively}
\description{
\code{ProgressReporter} is designed for interactive use. Its goal is to
give you actionable insights to help you understand the status of your
code. This reporter also praises you from time-to-time if all your tests
pass. It's the default reporter for \code{\link[=test_dir]{test_dir()}}.

\code{ParallelProgressReporter} is very similar to \code{ProgressReporter}, but
works better for packages that want parallel tests.

\code{CompactProgressReporter} is a minimal version of \code{ProgressReporter}
designed for use with single files. It's the default reporter for
\code{\link[=test_file]{test_file()}}.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expectation.R
\name{expect}
\alias{expect}
\title{The previous building block of all \code{expect_} functions}
\usage{
expect(
  ok,
  failure_message,
  info = NULL,
  srcref = NULL,
  trace = NULL,
  trace_env = caller_env()
)
}
\arguments{
\item{ok}{\code{TRUE} or \code{FALSE} indicating if the expectation was successful.}

\item{failure_message}{A character vector describing the failure. The
first element should describe the expected value, and the second (and
optionally subsequence) elements should describe what was actually seen.}

\item{info}{Character vector continuing additional information. Included
for backward compatibility only and new expectations should not use it.}

\item{srcref}{Location of the failure. Should only needed to be explicitly
supplied when you need to forward a srcref captured elsewhere.}

\item{trace}{An optional backtrace created by \code{\link[rlang:trace_back]{rlang::trace_back()}}.
When supplied, the expectation is displayed with the backtrace.
Expert use only.}

\item{trace_env}{If \code{trace} is not specified, this is used to generate an
informative traceback for failures. You should only need to set this if
you're calling \code{fail()} from a helper function; see
\code{vignette("custom-expectation")} for details.}
}
\value{
An expectation object from either \code{succeed()} or \code{fail()}.
with a \code{muffle_expectation} restart.
}
\description{
Previously, we recommended using \code{expect()} when writing your own
expectations. Now we instead recommend \code{\link[=pass]{pass()}} and \code{\link[=fail]{fail()}}. See
\code{vignette("custom-expectation")} for details.
}
\seealso{
\code{\link[=exp_signal]{exp_signal()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expectations-matches.R
\name{expect_match}
\alias{expect_match}
\alias{expect_no_match}
\title{Does a string match a regular expression?}
\usage{
expect_match(
  object,
  regexp,
  perl = FALSE,
  fixed = FALSE,
  ...,
  all = TRUE,
  info = NULL,
  label = NULL
)

expect_no_match(
  object,
  regexp,
  perl = FALSE,
  fixed = FALSE,
  ...,
  all = TRUE,
  info = NULL,
  label = NULL
)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{regexp}{Regular expression to test against.}

\item{perl}{logical.  Should Perl-compatible regexps be used?}

\item{fixed}{If \code{TRUE}, treats \code{regexp} as a string to be matched exactly
(not a regular expressions). Overrides \code{perl}.}

\item{...}{
  Arguments passed on to \code{\link[base:grep]{base::grepl}}
  \describe{
    \item{\code{ignore.case}}{if \code{FALSE}, the pattern matching is \emph{case
      sensitive} and if \code{TRUE}, case is ignored during matching.}
    \item{\code{useBytes}}{logical.  If \code{TRUE} the matching is done
    byte-by-byte rather than character-by-character.  See
    \sQuote{Details}.}
  }}

\item{all}{Should all elements of actual value match \code{regexp} (TRUE),
or does only one need to match (FALSE).}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}
}
\description{
Does a string match a regular expression?
}
\details{
\code{expect_match()} is a wrapper around \code{\link[=grepl]{grepl()}}. See its documentation for
more detail about the individual arguments. \code{expect_no_match()} provides
the complementary case, checking that a string \emph{does not} match a regular
expression.
}
\section{Functions}{
\itemize{
\item \code{expect_no_match()}: Check that a string doesn't match a regular
expression.

}}
\examples{
expect_match("Testing is fun", "fun")
expect_match("Testing is fun", "f.n")
expect_no_match("Testing is fun", "horrible")

\dontrun{
expect_match("Testing is fun", "horrible")

# Zero-length inputs always fail
expect_match(character(), ".")
}
}
\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-comparison.R
\name{expect_less_than}
\alias{expect_less_than}
\alias{expect_more_than}
\title{Deprecated numeric comparison functions}
\usage{
expect_less_than(...)

expect_more_than(...)
}
\arguments{
\item{...}{All arguments passed on to \code{expect_lt()}/\code{expect_gt()}.}
}
\description{
These functions have been deprecated in favour of the more concise
\code{\link[=expect_gt]{expect_gt()}} and \code{\link[=expect_lt]{expect_lt()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/watcher.R
\name{watch}
\alias{watch}
\title{Watch a directory for changes (additions, deletions & modifications).}
\usage{
watch(path, callback, pattern = NULL, hash = TRUE)
}
\arguments{
\item{path}{character vector of paths to watch.  Omit trailing backslash.}

\item{callback}{function called everytime a change occurs.  It should
have three parameters: added, deleted, modified, and should return
TRUE to keep watching, or FALSE to stop.}

\item{pattern}{file pattern passed to \code{\link[=dir]{dir()}}}

\item{hash}{hashes are more accurate at detecting changes, but are slower
for large files.  When FALSE, uses modification time stamps}
}
\description{
This is used to power the \code{\link[=auto_test]{auto_test()}} and
\code{\link[=auto_test_package]{auto_test_package()}} functions which are used to rerun tests
whenever source code changes.
}
\details{
Use Ctrl + break (windows), Esc (mac gui) or Ctrl + C (command line) to
stop the watcher.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local.R
\name{local_test_directory}
\alias{local_test_directory}
\title{Locally set test directory options}
\usage{
local_test_directory(path, package = NULL, .env = parent.frame())
}
\arguments{
\item{path}{Path to directory of files}

\item{package}{Optional package name, if known.}
}
\description{
For expert use only.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/context.R
\name{context_start_file}
\alias{context_start_file}
\title{Start test context from a file name}
\usage{
context_start_file(name)
}
\arguments{
\item{name}{file name}
}
\description{
For use in external reporters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter.R
\name{default_reporter}
\alias{default_reporter}
\alias{default_compact_reporter}
\alias{check_reporter}
\title{Determine default reporters}
\usage{
default_reporter(parallel = FALSE)

default_compact_reporter()

check_reporter()
}
\arguments{
\item{parallel}{If \code{TRUE}, return a reporter suitable for parallel testing.}
}
\description{
These three functions are used to determine the default reporters used
for \code{test_dir()}, \code{test_file()}, and \code{test_package()}:
\itemize{
\item \code{default_reporter()} returns the default reporter for \code{\link[=test_dir]{test_dir()}}.
If \code{parallel} is \code{TRUE}, it uses \link{ParallelProgressReporter}, which you
can override with option \code{testthat.default_parallel_reporter}.
If \code{parallel} is \code{FALSE}, it uses \link{ProgressReporter}, which you
can override with option \code{testthat.default_reporter}.
\item \code{default_compact_reporter()} returns the default reporter for
\code{\link[=test_file]{test_file()}}. It defaults to \link{CompactProgressReporter}, which you can
override with the \code{testthat.default_compact_reporter} option.
\item \code{check_reporter()} returns the default reporter for \code{\link[=test_package]{test_package()}}.
It defaults to \link{CheckReporter}, which you can override with the
\code{testthat.default_check_reporter} option.
}

Both \code{default_reporter()} and \code{default_compact_reporter()} will use
\link{LlmReporter} if it appears that the tests are being run by a coding agent.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-output.R
\name{expect_output}
\alias{expect_output}
\title{Do you expect printed output to match this pattern?}
\usage{
expect_output(
  object,
  regexp = NULL,
  ...,
  info = NULL,
  label = NULL,
  width = 80
)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{regexp}{Regular expression to test against.
\itemize{
\item A character vector giving a regular expression that must match the output.
\item If \code{NULL}, the default, asserts that there should output,
but doesn't check for a specific value.
\item If \code{NA}, asserts that there should be no output.
}}

\item{...}{
  Arguments passed on to \code{\link[=expect_match]{expect_match}}
  \describe{
    \item{\code{all}}{Should all elements of actual value match \code{regexp} (TRUE),
or does only one need to match (FALSE).}
    \item{\code{fixed}}{If \code{TRUE}, treats \code{regexp} as a string to be matched exactly
(not a regular expressions). Overrides \code{perl}.}
    \item{\code{perl}}{logical.  Should Perl-compatible regexps be used?}
  }}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}

\item{width}{Number of characters per line of output. This does not
inherit from \code{getOption("width")} so that tests always use the same
output width, minimising spurious differences.}
}
\value{
The first argument, invisibly.
}
\description{
Test for output produced by \code{print()} or \code{cat()}. This is best used for
very simple output; for more complex cases use \code{\link[=expect_snapshot]{expect_snapshot()}}.
}
\examples{
str(mtcars)
expect_output(str(mtcars), "32 obs")
expect_output(str(mtcars), "11 variables")

# You can use the arguments of grepl to control the matching
expect_output(str(mtcars), "11 VARIABLES", ignore.case = TRUE)
expect_output(str(mtcars), "$ mpg", fixed = TRUE)
}
\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-minimal.R
\name{MinimalReporter}
\alias{MinimalReporter}
\title{Report minimal results as compactly as possible}
\description{
The minimal test reporter provides the absolutely minimum amount of
information: whether each expectation has succeeded, failed or experienced
an error.  If you want to find out what the failures and errors actually
were, you'll need to run a more informative test reporter.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/describe.R
\name{describe}
\alias{describe}
\alias{it}
\title{describe: a BDD testing language}
\usage{
describe(description, code)

it(description, code = NULL)
}
\arguments{
\item{description}{description of the feature}

\item{code}{test code containing the specs}
}
\description{
A simple \href{https://en.wikipedia.org/wiki/Behavior-driven_development}{behavior-driven development (BDD)}
\href{https://en.wikipedia.org/wiki/Domain-specific_language}{domain-specific language}
for writing tests. The language is similar to \href{https://rspec.info/}{RSpec}
for Ruby or \href{https://mochajs.org/}{Mocha} for JavaScript. BDD tests read
like sentences and it should thus be easier to understand what the
specification of a function/component is.
}
\details{
Tests using the \code{describe} syntax not only verify the tested code, but
also document its intended behaviour. Each \code{describe} block specifies a
larger component or function and contains a set of specifications. A
specification is defined by an \code{it} block. Each \code{it} block
functions as a test and is evaluated in its own environment. You
can also have nested \code{describe} blocks.

This test syntax helps to test the intended behaviour of your code. For
example: you want to write a new function for your package. Try to describe
the specification first using \code{describe}, before your write any code.
After that, you start to implement the tests for each specification (i.e.
the \code{it} block).

Use \code{describe} to verify that you implement the right things and use
\code{\link[=test_that]{test_that()}} to ensure you do the things right.
}
\examples{
describe("matrix()", {
  it("can be multiplied by a scalar", {
    m1 <- matrix(1:4, 2, 2)
    m2 <- m1 * 2
    expect_equal(matrix(1:4 * 2, 2, 2), m2)
  })
  it("can have not yet tested specs")
})

# Nested specs:
## code
addition <- function(a, b) a + b
division <- function(a, b) a / b

## specs
describe("math library", {
  describe("addition()", {
    it("can add two numbers", {
      expect_equal(1 + 1, addition(1, 1))
    })
  })
  describe("division()", {
    it("can divide two numbers", {
      expect_equal(10 / 2, division(10, 2))
    })
    it("can handle division by 0") #not yet implemented
  })
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-teamcity.R
\name{TeamcityReporter}
\alias{TeamcityReporter}
\title{Test reporter: Teamcity format.}
\description{
This reporter will output results in the Teamcity message format.
For more information about Teamcity messages, see
http://confluence.jetbrains.com/display/TCD7/Build+Script+Interaction+with+TeamCity
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-inheritance.R
\name{inheritance-expectations}
\alias{inheritance-expectations}
\alias{expect_type}
\alias{expect_s3_class}
\alias{expect_s4_class}
\alias{expect_r6_class}
\alias{expect_s7_class}
\title{Do you expect an S3/S4/R6/S7 object that inherits from this class?}
\usage{
expect_type(object, type)

expect_s3_class(object, class, exact = FALSE)

expect_s4_class(object, class)

expect_r6_class(object, class)

expect_s7_class(object, class)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{type}{String giving base type (as returned by \code{\link[=typeof]{typeof()}}).}

\item{class}{The required type varies depending on the function:
\itemize{
\item \code{expect_type()}: a string.
\item \code{expect_s3_class()}: a string or character vector. The behaviour of
multiple values (i.e. a character vector) is controlled by the
\code{exact} argument.
\item \code{expect_s4_class()}: a string.
\item \code{expect_r6_class()}: a string.
\item \code{expect_s7_class()}: an \code{\link[S7:S7_class]{S7::S7_class()}} object.
}

For historical reasons, \code{expect_s3_class()} and \code{expect_s4_class()} also
take \code{NA} to assert that the \code{object} is not an S3 or S4 object.}

\item{exact}{If \code{FALSE}, the default, checks that \code{object} inherits
from any element of \code{class}. If \code{TRUE}, checks that object has a class
that exactly matches \code{class}.}
}
\description{
See \url{https://adv-r.hadley.nz/oo.html} for an overview of R's OO systems, and
the vocabulary used here.
\itemize{
\item \code{expect_type(x, type)} checks that \code{typeof(x)} is \code{type}.
\item \code{expect_s3_class(x, class)} checks that \code{x} is an S3 object that
\code{\link[=inherits]{inherits()}} from \code{class}
\item \code{expect_s3_class(x, NA)} checks that \code{x} isn't an S3 object.
\item \code{expect_s4_class(x, class)} checks that \code{x} is an S4 object that
\code{\link[=is]{is()}} \code{class}.
\item \code{expect_s4_class(x, NA)} checks that \code{x} isn't an S4 object.
\item \code{expect_r6_class(x, class)} checks that \code{x} an R6 object that
inherits from \code{class}.
\item \code{expect_s7_class(x, Class)} checks that \code{x} is an S7 object that
\code{\link[S7:S7_inherits]{S7::S7_inherits()}} from \code{Class}
}

See \code{\link[=expect_vector]{expect_vector()}} for testing properties of objects created by vctrs.
}
\examples{
x <- data.frame(x = 1:10, y = "x", stringsAsFactors = TRUE)
# A data frame is an S3 object with class data.frame
expect_s3_class(x, "data.frame")
show_failure(expect_s4_class(x, "data.frame"))
# A data frame is built from a list:
expect_type(x, "list")

f <- factor(c("a", "b", "c"))
o <- ordered(f)

# Using multiple class names tests if the object inherits from any of them
expect_s3_class(f, c("ordered", "factor"))
# Use exact = TRUE to test for exact match
show_failure(expect_s3_class(f, c("ordered", "factor"), exact = TRUE))
expect_s3_class(o, c("ordered", "factor"), exact = TRUE)

# An integer vector is an atomic vector of type "integer"
expect_type(x$x, "integer")
# It is not an S3 object
show_failure(expect_s3_class(x$x, "integer"))

# Above, we requested data.frame() converts strings to factors:
show_failure(expect_type(x$y, "character"))
expect_s3_class(x$y, "factor")
expect_type(x$y, "integer")
}
\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-oo.R
\name{local_mocked_s3_method}
\alias{local_mocked_s3_method}
\alias{local_mocked_s4_method}
\title{Mock S3 and S4 methods}
\usage{
local_mocked_s3_method(generic, signature, definition, frame = caller_env())

local_mocked_s4_method(generic, signature, definition, frame = caller_env())
}
\arguments{
\item{generic}{A string giving the name of the generic.}

\item{signature}{A character vector giving the signature of the method.}

\item{definition}{A function providing the method definition, or \code{NULL} to
temporarily remove the method.}

\item{frame}{Calling frame which determines the scope of the mock.
Only needed when wrapping in another local helper.}
}
\description{
These functions temporarily override S3 or S4 methods. They can mock
methods that don't already exist, or temporarily remove a method by setting
\code{definition = NULL}.

Learn more about mocking in \code{vignette("mocking")}.
}
\examples{
x <- as.POSIXlt(Sys.time())
local({
  local_mocked_s3_method("length", "POSIXlt", function(x) 42)
  length(x)
})

length(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-reference.R
\name{expect_reference}
\alias{expect_reference}
\title{Do you expect a reference to this object?}
\usage{
expect_reference(
  object,
  expected,
  info = NULL,
  label = NULL,
  expected.label = NULL
)
}
\arguments{
\item{object, expected}{Computation and value to compare it to.

Both arguments supports limited unquoting to make it easier to generate
readable failures within a function or for loop. See \link{quasi_label} for
more details.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label, expected.label}{Used to customise failure messages. For expert
use only.}
}
\description{
\code{expect_reference()} compares the underlying memory addresses of
two symbols. It is for expert use only.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{expect_reference()} is deprecated in the third edition. If you know what
you're doing, and you really need this behaviour, just use \code{is_reference()}
directly: \code{expect_true(rlang::is_reference(x, y))}.
}

\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare}
\alias{compare}
\alias{compare.default}
\alias{compare.character}
\alias{compare.numeric}
\alias{compare.POSIXt}
\title{Provide human-readable comparison of two objects}
\usage{
compare(x, y, ...)

\method{compare}{default}(x, y, ..., max_diffs = 9)

\method{compare}{character}(
  x,
  y,
  check.attributes = TRUE,
  ...,
  max_diffs = 5,
  max_lines = 5,
  width = cli::console_width()
)

\method{compare}{numeric}(
  x,
  y,
  tolerance = testthat_tolerance(),
  check.attributes = TRUE,
  ...,
  max_diffs = 9
)

\method{compare}{POSIXt}(x, y, tolerance = 0.001, ..., max_diffs = 9)
}
\arguments{
\item{x, y}{Objects to compare}

\item{...}{Additional arguments used to control specifics of comparison}

\item{max_diffs}{Maximum number of differences to show}

\item{check.attributes}{If \code{TRUE}, also checks values of attributes.}

\item{max_lines}{Maximum number of lines to show from each difference}

\item{width}{Width of output device}

\item{tolerance}{Numerical tolerance: any differences (in the sense of
\code{\link[base:all.equal]{base::all.equal()}}) smaller than this value will be ignored.

The default tolerance is \code{sqrt(.Machine$double.eps)}, unless long doubles
are not available, in which case the test is skipped.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{compare} is similar to \code{\link[base:all.equal]{base::all.equal()}}, but somewhat buggy in its
use of \code{tolerance}. Please use \href{https://waldo.r-lib.org/}{waldo} instead.
}
\examples{
# Character -----------------------------------------------------------------
x <- c("abc", "def", "jih")
compare(x, x)

y <- paste0(x, "y")
compare(x, y)

compare(letters, paste0(letters, "-"))

x <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Duis cursus
 tincidunt auctor. Vestibulum ac metus bibendum, facilisis nisi non, pulvinar
 dolor. Donec pretium iaculis nulla, ut interdum sapien ultricies a. "
y <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Duis cursus
 tincidunt auctor. Vestibulum ac metus1 bibendum, facilisis nisi non, pulvinar
 dolor. Donec pretium iaculis nulla, ut interdum sapien ultricies a. "
compare(x, y)
compare(c(x, x), c(y, y))

# Numeric -------------------------------------------------------------------

x <- y <- runif(100)
y[sample(100, 10)] <- 5
compare(x, y)

x <- y <- 1:10
x[5] <- NA
x[6] <- 6.5
compare(x, y)

# Compare ignores minor numeric differences in the same way
# as all.equal.
compare(x, x + 1e-9)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock.R
\name{with_mock}
\alias{with_mock}
\alias{local_mock}
\title{Mock functions in a package.}
\usage{
with_mock(..., .env = topenv())

local_mock(..., .env = topenv(), .local_envir = parent.frame())
}
\arguments{
\item{...}{named parameters redefine mocked functions, unnamed parameters
will be evaluated after mocking the functions}

\item{.env}{the environment in which to patch the functions,
defaults to the top-level environment.  A character is interpreted as
package name.}

\item{.local_envir}{Environment in which to add exit handler.
For expert use only.}
}
\value{
The result of the last unnamed parameter
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{with_mock()} and \code{local_mock()} are deprecated in favour of
\code{\link[=with_mocked_bindings]{with_mocked_bindings()}} and \code{\link[=local_mocked_bindings]{local_mocked_bindings()}}.

These functions worked by using some C code to temporarily modify the mocked
function \emph{in place}. This was an abuse of R's internals and it is no longer
permitted.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot-manage.R
\name{snapshot_accept}
\alias{snapshot_accept}
\alias{snapshot_review}
\title{Snapshot management}
\usage{
snapshot_accept(files = NULL, path = "tests/testthat")

snapshot_review(files = NULL, path = "tests/testthat")
}
\arguments{
\item{files}{Optionally, filter effects to snapshots from specified files.
This can be a snapshot name (e.g. \code{foo} or \code{foo.md}), a snapshot file name
(e.g. \code{testfile/foo.txt}), or a snapshot file directory (e.g. \verb{testfile/}).}

\item{path}{Path to tests.}
}
\description{
\itemize{
\item \code{snapshot_accept()} accepts all modified snapshots.
\item \code{snapshot_review()} opens a Shiny app that shows a visual diff of each
modified snapshot. This is particularly useful for whole file snapshots
created by \code{expect_snapshot_file()}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-example.R
\name{test_examples}
\alias{test_examples}
\alias{test_rd}
\alias{test_example}
\title{Test package examples}
\usage{
test_examples(path = "../..")

test_rd(rd, title = attr(rd, "Rdfile"))

test_example(path, title = path)
}
\arguments{
\item{path}{For \code{test_examples()}, path to directory containing Rd files.
For \code{test_example()}, path to a single Rd file. Remember the working
directory for tests is \code{tests/testthat}.}

\item{rd}{A parsed Rd object, obtained from \code{\link[tools:Rdutils]{tools::Rd_db()}} or otherwise.}

\item{title}{Test title to use}
}
\description{
These helper functions make it easier to test the examples in a package.
Each example counts as one test, and it succeeds if the code runs without
an error. Generally, this is redundant with R CMD check, and is not
recommended in routine practice.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-fail.R
\name{FailReporter}
\alias{FailReporter}
\title{Fail if any tests fail}
\description{
This reporter will simply throw an error if any of the tests failed. It is
best combined with another reporter, such as the
\link{SummaryReporter}.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-list.R
\name{ListReporter}
\alias{ListReporter}
\title{Capture test results and metadata}
\description{
This reporter gathers all results, adding additional information such as
test elapsed time, and test filename if available. Very useful for reporting.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-files.R
\name{teardown_env}
\alias{teardown_env}
\title{Run code after all test files}
\usage{
teardown_env()
}
\description{
This environment has no purpose other than as a handle for \code{\link[withr:defer]{withr::defer()}}:
use it when you want to run code after all tests have been run.
Typically, you'll use \code{withr::defer(cleanup(), teardown_env())}
immediately after you've made a mess in a \verb{setup-*.R} file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{simulate_test_env}
\alias{simulate_test_env}
\title{Simulate a test environment}
\usage{
simulate_test_env(package, path)
}
\arguments{
\item{package}{Name of installed package.}

\item{path}{Path to \code{tests/testthat}.}
}
\description{
This function is designed to allow you to simulate testthat's testing
environment in an interactive session. To undo it's affect, you
will need to restart your R session.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-location.R
\name{LocationReporter}
\alias{LocationReporter}
\title{Test reporter: location}
\description{
This reporter simply prints the location of every expectation and error.
This is useful if you're trying to figure out the source of a segfault,
or you want to figure out which code triggers a C/C++ breakpoint
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-that.R
\name{succeed}
\alias{succeed}
\title{Mark a test as successful}
\usage{
succeed(message = "Success has been forced", info = NULL)
}
\arguments{
\item{message}{A character vector describing the failure. The
first element should describe the expected value, and the second (and
optionally subsequence) elements should describe what was actually seen.}

\item{info}{Character vector continuing additional information. Included
for backward compatibility only and new expectations should not use it.}
}
\description{
This is an older version of \code{\link[=pass]{pass()}} that exists for backwards compatibility.
You should now use \code{pass()} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-files.R
\name{test_file}
\alias{test_file}
\title{Run tests in a single file}
\usage{
test_file(
  path,
  reporter = default_compact_reporter(),
  desc = NULL,
  package = NULL,
  ...
)
}
\arguments{
\item{path}{Path to file.}

\item{reporter}{Reporter to use to summarise output. Can be supplied
as a string (e.g. "summary") or as an R6 object
(e.g. \code{SummaryReporter$new()}).

See \link{Reporter} for more details and a list of built-in reporters.}

\item{desc}{Optionally, supply a string here to run only a single
test (\code{test_that()} or \code{describe()}) with this \code{desc}ription.}

\item{package}{If these tests belong to a package, the name of the package.}

\item{...}{Additional parameters passed on to \code{test_dir()}}
}
\value{
A list (invisibly) containing data about the test results.
}
\description{
Helper, setup, and teardown files located in the same directory as the
test will also be run. See \code{vignette("special-files")} for details.
}
\section{Environments}{

Each test is run in a clean environment to keep tests as isolated as
possible. For package tests, that environment inherits from the package's
namespace environment, so that tests can access internal functions
and objects.
}

\examples{
path <- testthat_example("success")
test_file(path)
test_file(path, desc = "some tests have warnings")
test_file(path, reporter = "minimal")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{expect_snapshot_output}
\alias{expect_snapshot_output}
\alias{expect_snapshot_error}
\alias{expect_snapshot_warning}
\title{Snapshot helpers}
\usage{
expect_snapshot_output(x, cran = FALSE, variant = NULL)

expect_snapshot_error(x, class = "error", cran = FALSE, variant = NULL)

expect_snapshot_warning(x, class = "warning", cran = FALSE, variant = NULL)
}
\arguments{
\item{x}{Code to evaluate.}

\item{cran}{Should these expectations be verified on CRAN? By default,
they are not, because snapshot tests tend to be fragile because they
often rely on minor details of dependencies.}

\item{variant}{If non-\code{NULL}, results will be saved in
\verb{_snaps/\{variant\}/\{test.md\}}, so \code{variant} must be a single string
suitable for use as a directory name.

You can use variants to deal with cases where the snapshot output varies
and you want to capture and test the variations. Common use cases include
variations for operating system, R version, or version of key dependency.
Variants are an advanced feature. When you use them, you'll need to
carefully think about your testing strategy to ensure that all important
variants are covered by automated tests, and ensure that you have a way
to get snapshot changes out of your CI system and back into the repo.

Note that there's no way to declare all possible variants up front which
means that as soon as you start using variants, you are responsible for
deleting snapshot variants that are no longer used. (testthat will still
delete all variants if you delete the test.)}

\item{class}{Class of expected error or warning. The expectation will
always fail (even on CRAN) if an error of this class isn't seen
when executing \code{x}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#questioning}{\figure{lifecycle-questioning.svg}{options: alt='[Questioning]'}}}{\strong{[Questioning]}}

These snapshotting functions are questioning because they were developed
before \code{\link[=expect_snapshot]{expect_snapshot()}} and we're not sure that they still have a
role to play.
\itemize{
\item \code{expect_snapshot_output()} captures just output printed to the console.
\item \code{expect_snapshot_error()} captures an error message and
optionally checks its class.
\item \code{expect_snapshot_warning()} captures a warning message and
optionally checks its class.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture-output.R
\name{capture_output}
\alias{capture_output}
\alias{capture_output_lines}
\alias{testthat_print}
\title{Capture output to console}
\usage{
capture_output(code, print = FALSE, width = 80)

capture_output_lines(code, print = FALSE, width = 80)

testthat_print(x)
}
\arguments{
\item{code}{Code to evaluate.}

\item{print}{If \code{TRUE} and the result of evaluating \code{code} is
visible, print the result using \code{testthat_print()}.}

\item{width}{Number of characters per line of output. This does not
inherit from \code{getOption("width")} so that tests always use the same
output width, minimising spurious differences.}
}
\value{
\code{capture_output()} returns a single string. \code{capture_output_lines()}
returns a character vector with one entry for each line
}
\description{
Evaluates \code{code} in a special context in which all output is captured,
similar to \code{\link[=capture.output]{capture.output()}}.
}
\details{
Results are printed using the \code{testthat_print()} generic, which defaults
to \code{print()}, giving you the ability to customise the printing of your
object in tests, if needed.
}
\examples{
capture_output({
  cat("Hi!\n")
  cat("Bye\n")
})

capture_output_lines({
  cat("Hi!\n")
  cat("Bye\n")
})

capture_output("Hi")
capture_output("Hi", print = TRUE)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluate-promise.R
\name{evaluate_promise}
\alias{evaluate_promise}
\title{Evaluate a promise, capturing all types of output.}
\usage{
evaluate_promise(code, print = FALSE)
}
\arguments{
\item{code}{Code to evaluate.}
}
\value{
A list containing
\item{result}{The result of the function}
\item{output}{A string containing all the output from the function}
\item{warnings}{A character vector containing the text from each warning}
\item{messages}{A character vector containing the text from each message}
}
\description{
Evaluate a promise, capturing all types of output.
}
\examples{
evaluate_promise({
  print("1")
  message("2")
  warning("3")
  4
})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-setequal.R
\name{expect_setequal}
\alias{expect_setequal}
\alias{expect_mapequal}
\alias{expect_contains}
\alias{expect_in}
\title{Does code return a vector containing the expected values?}
\usage{
expect_setequal(object, expected)

expect_mapequal(object, expected)

expect_contains(object, expected)

expect_in(object, expected)
}
\arguments{
\item{object, expected}{Computation and value to compare it to.

Both arguments supports limited unquoting to make it easier to generate
readable failures within a function or for loop. See \link{quasi_label} for
more details.}
}
\description{
\itemize{
\item \code{expect_setequal(x, y)} tests that every element of \code{x} occurs in \code{y},
and that every element of \code{y} occurs in \code{x}.
\item \code{expect_contains(x, y)} tests that \code{x} contains every element of \code{y}
(i.e. \code{y} is a subset of \code{x}).
\item \code{expect_in(x, y)} tests every element of \code{x} is in \code{y}
(i.e. \code{x} is a subset of \code{y}).
\item \code{expect_mapequal(x, y)} tests that \code{x} and \code{y} have the same names, and
that \code{x[names(y)]} equals \code{y}.
}
}
\details{
Note that \code{expect_setequal()} ignores names, and you will be warned if both
\code{object} and \code{expected} have them.
}
\examples{
expect_setequal(letters, rev(letters))
show_failure(expect_setequal(letters[-1], rev(letters)))

x <- list(b = 2, a = 1)
expect_mapequal(x, list(a = 1, b = 2))
show_failure(expect_mapequal(x, list(a = 1)))
show_failure(expect_mapequal(x, list(a = 1, b = "x")))
show_failure(expect_mapequal(x, list(a = 1, b = 2, c = 3)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-self-test.R
\name{expect_success}
\alias{expect_success}
\alias{expect_failure}
\alias{expect_snapshot_failure}
\alias{show_failure}
\title{Test your custom expectations}
\usage{
expect_success(expr)

expect_failure(expr, message = NULL, ...)

expect_snapshot_failure(expr)

show_failure(expr)
}
\arguments{
\item{expr}{Code to evaluate}

\item{message}{Check that the failure message matches this regexp.}

\item{...}{Other arguments passed on to \code{\link[=expect_match]{expect_match()}}.}
}
\description{
\code{expect_success()} checks that there's exactly one success and no failures;
\code{expect_failure()} checks that there's exactly one failure and no successes.
\code{expect_snapshot_failure()} records the failure message so that you can
manually check that it is informative.

Use \code{show_failure()} in examples to print the failure message without
throwing an error.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-tap.R
\name{TapReporter}
\alias{TapReporter}
\title{Report results in TAP format}
\description{
This reporter will output results in the Test Anything Protocol (TAP),
a simple text-based interface between testing modules in a test harness.
For more information about TAP, see http://testanything.org
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{SlowReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-slow.R, R/reporter-timing.R
\name{SlowReporter}
\alias{SlowReporter}
\title{Find slow tests}
\description{
\code{SlowReporter} is designed to identify slow tests. It reports the
execution time for each test and can optionally filter out tests that
run faster than a specified threshold (default: 1 second). This reporter
is useful for performance optimization and identifying tests that may
benefit from optimization or parallelization.

\code{SlowReporter} is designed to identify slow tests. It reports the
execution time for each test, ignoring tests faster than a specified
threshold (default: 0.5s).

The easiest way to run it over your package is with
\code{devtools::test(reporter = "slow")}.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}

Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LlmReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{SilentReporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-compiled-code.R
\name{use_catch}
\alias{use_catch}
\title{Use Catch for C++ Unit Testing}
\usage{
use_catch(dir = getwd())
}
\arguments{
\item{dir}{The directory containing an \R package.}
}
\description{
Add the necessary infrastructure to enable C++ unit testing
in \R packages with \href{https://github.com/catchorg/Catch2}{Catch} and
\code{testthat}.
}
\details{
Calling \code{use_catch()} will:
\enumerate{
\item Create a file \code{src/test-runner.cpp}, which ensures that the
\code{testthat} package will understand how to run your package's
unit tests,
\item Create an example test file \code{src/test-example.cpp}, which
showcases how you might use Catch to write a unit test,
\item Add a test file \code{tests/testthat/test-cpp.R}, which ensures that
\code{testthat} will run your compiled tests during invocations of
\code{devtools::test()} or \verb{R CMD check}, and
\item Create a file \code{R/catch-routine-registration.R}, which ensures that
\R will automatically register this routine when
\code{tools::package_native_routine_registration_skeleton()} is invoked.
}

You will also need to:
\itemize{
\item Add xml2 to Suggests, with e.g. \code{usethis::use_package("xml2", "Suggests")}
\item Add testthat to LinkingTo, with e.g.
\code{usethis::use_package("testthat", "LinkingTo")}
}

C++ unit tests can be added to C++ source files within the
\code{src} directory of your package, with a format similar
to \R code tested with \code{testthat}. Here's a simple example
of a unit test written with \code{testthat} + Catch:

\preformatted{
context("C++ Unit Test") {
  test_that("two plus two is four") {
    int result = 2 + 2;
    expect_true(result == 4);
  }
}
}

When your package is compiled, unit tests alongside a harness
for running these tests will be compiled into your \R package,
with the C entry point \code{run_testthat_tests()}. \code{testthat}
will use that entry point to run your unit tests when detected.
}
\section{Functions}{


All of the functions provided by Catch are
available with the \code{CATCH_} prefix -- see
\href{https://github.com/catchorg/Catch2/blob/master/docs/assertions.md}{here}
for a full list. \code{testthat} provides the
following wrappers, to conform with \code{testthat}'s
\R interface:

\tabular{lll}{
\strong{Function} \tab \strong{Catch} \tab \strong{Description} \cr
\code{context} \tab \code{CATCH_TEST_CASE} \tab The context of a set of tests. \cr
\code{test_that} \tab \code{CATCH_SECTION} \tab A test section. \cr
\code{expect_true} \tab \code{CATCH_CHECK} \tab Test that an expression evaluates to \code{true}. \cr
\code{expect_false} \tab \code{CATCH_CHECK_FALSE} \tab Test that an expression evalutes to \code{false}. \cr
\code{expect_error} \tab \code{CATCH_CHECK_THROWS} \tab Test that evaluation of an expression throws an exception. \cr
\code{expect_error_as} \tab \code{CATCH_CHECK_THROWS_AS} \tab Test that evaluation of an expression throws an exception of a specific class. \cr
}

In general, you should prefer using the \code{testthat}
wrappers, as \code{testthat} also does some work to
ensure that any unit tests within will not be compiled or
run when using the Solaris Studio compilers (as these are
currently unsupported by Catch). This should make it
easier to submit packages to CRAN that use Catch.
}

\section{Symbol Registration}{


If you've opted to disable dynamic symbol lookup in your
package, then you'll need to explicitly export a symbol
in your package that \code{testthat} can use to run your unit
tests. \code{testthat} will look for a routine with one of the names:

\preformatted{
    C_run_testthat_tests
    c_run_testthat_tests
    run_testthat_tests
}

See \href{https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Controlling-visibility}{Controlling Visibility}
and \href{https://cran.r-project.org/doc/manuals/r-release/R-exts.html#Registering-symbols}{Registering Symbols}
in the \strong{Writing R Extensions} manual for more information.
}

\section{Advanced Usage}{


If you'd like to write your own Catch test runner, you can
instead use the \code{testthat::catchSession()} object in a file
with the form:

\preformatted{
#define TESTTHAT_TEST_RUNNER
#include <testthat.h>

void run()
{
    Catch::Session& session = testthat::catchSession();
    // interact with the session object as desired
}
}

This can be useful if you'd like to run your unit tests
with custom arguments passed to the Catch session.
}

\section{Standalone Usage}{


If you'd like to use the C++ unit testing facilities provided
by Catch, but would prefer not to use the regular \code{testthat}
\R testing infrastructure, you can manually run the unit tests
by inserting a call to:

\preformatted{
.Call("run_testthat_tests", PACKAGE = <pkgName>)
}

as necessary within your unit test suite.
}

\seealso{
\href{https://github.com/catchorg/Catch2/blob/master/docs/assertions.md}{Catch},
the library used to enable C++ unit testing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skip.R
\name{skip_on_travis}
\alias{skip_on_travis}
\alias{skip_on_appveyor}
\title{Superseded skip functions}
\usage{
skip_on_travis()

skip_on_appveyor()
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}
\itemize{
\item \code{skip_on_travis()} and \code{skip_on_appveyor()} have been superseded by
\code{\link[=skip_on_ci]{skip_on_ci()}}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expectation.R
\name{expectation}
\alias{expectation}
\alias{new_expectation}
\alias{exp_signal}
\alias{is.expectation}
\title{Expectation conditions}
\usage{
expectation(type, message, ..., srcref = NULL, trace = NULL)

new_expectation(
  type,
  message,
  ...,
  srcref = NULL,
  trace = NULL,
  .subclass = NULL
)

exp_signal(exp)

is.expectation(x)
}
\arguments{
\item{type}{Expectation type. Must be one of "success", "failure", "error",
"skip", "warning".}

\item{message}{Message describing test failure}

\item{...}{Additional attributes for the expectation object.}

\item{srcref}{Optional \code{srcref} giving location of test.}

\item{trace}{An optional backtrace created by \code{\link[rlang:trace_back]{rlang::trace_back()}}.
When supplied, the expectation is displayed with the backtrace.
Expert use only.}

\item{.subclass}{An optional subclass for the expectation object.}

\item{exp}{An expectation object, as created by
\code{\link[=new_expectation]{new_expectation()}}.}

\item{x}{object to test for class membership}
}
\description{
\code{new_expectation()} creates an expectation condition object and
\code{exp_signal()} signals it. \code{expectation()} does both. \code{is.expectation()}
tests if a captured condition is a testthat expectation.

These functions are primarily for internal use. If you are creating your
own expectation, you do not need these functions are instead should use
\code{\link[=pass]{pass()}} or \code{\link[=fail]{fail()}}. See \code{vignette("custom-expectation")} for more
details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-files.R
\name{find_test_scripts}
\alias{find_test_scripts}
\title{Find test files}
\usage{
find_test_scripts(
  path,
  filter = NULL,
  invert = FALSE,
  ...,
  full.names = TRUE,
  start_first = NULL
)
}
\arguments{
\item{path}{path to tests}

\item{filter}{If not \code{NULL}, only tests with file names matching this
regular expression will be executed. Matching is performed on the file
name after it's stripped of \code{"test-"} and \code{".R"}.}

\item{invert}{If \code{TRUE} return files which \strong{don't} match.}

\item{...}{Additional arguments passed to \code{\link[=grepl]{grepl()}} to control filtering.}

\item{start_first}{A character vector of file patterns (globs, see
\code{\link[utils:glob2rx]{utils::glob2rx()}}). The patterns are for the file names (base names),
not for the whole paths. testthat starts the files matching the
first pattern first,  then the ones matching the second, etc. and then
the rest of the files, alphabetically. Parallel tests tend to finish
quicker if you start the slowest files first. \code{NULL} means alphabetical
order.}
}
\value{
A character vector of paths
}
\description{
Find test files
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporter-silent.R
\name{SilentReporter}
\alias{SilentReporter}
\title{Test reporter: gather all errors silently.}
\description{
This reporter quietly runs all tests, simply gathering all expectations.
This is helpful for programmatically inspecting errors after a test run.
You can retrieve the results with the \code{expectations()}
method.
}
\seealso{
Other reporters: 
\code{\link{CheckReporter}},
\code{\link{DebugReporter}},
\code{\link{FailReporter}},
\code{\link{JunitReporter}},
\code{\link{ListReporter}},
\code{\link{LocationReporter}},
\code{\link{MinimalReporter}},
\code{\link{MultiReporter}},
\code{\link{ProgressReporter}},
\code{\link{RStudioReporter}},
\code{\link{Reporter}},
\code{\link{StopReporter}},
\code{\link{SummaryReporter}},
\code{\link{TapReporter}},
\code{\link{TeamcityReporter}}
}
\concept{reporters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-equality.R
\name{expect_equivalent}
\alias{expect_equivalent}
\title{Is an object equal to the expected value, ignoring attributes?}
\usage{
expect_equivalent(
  object,
  expected,
  ...,
  info = NULL,
  label = NULL,
  expected.label = NULL
)
}
\arguments{
\item{object, expected}{Computation and value to compare it to.

Both arguments supports limited unquoting to make it easier to generate
readable failures within a function or for loop. See \link{quasi_label} for
more details.}

\item{...}{Passed on to \code{\link[=compare]{compare()}}.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label, expected.label}{Used to customise failure messages. For expert
use only.}
}
\description{
Compares \code{object} and \code{expected} using \code{\link[=all.equal]{all.equal()}} and
\code{check.attributes = FALSE}.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{expect_equivalent()} is deprecated in the 3rd edition. Instead use
\code{expect_equal(ignore_attr = TRUE)}.
}

\examples{
#' # expect_equivalent() ignores attributes
a <- b <- 1:3
names(b) <- letters[1:3]
\dontrun{
expect_equal(a, b)
}
expect_equivalent(a, b)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example.R
\name{testthat_examples}
\alias{testthat_examples}
\alias{testthat_example}
\title{Retrieve paths to built-in example test files}
\usage{
testthat_examples()

testthat_example(filename)
}
\arguments{
\item{filename}{Name of test file}
}
\description{
\code{testthat_examples()} retrieves path to directory of test files,
\code{testthat_example()} retrieves path to a single test file.
}
\examples{
dir(testthat_examples())
testthat_example("success")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-env.R
\name{is_testing}
\alias{is_testing}
\alias{is_parallel}
\alias{is_checking}
\alias{is_snapshot}
\alias{testing_package}
\title{Determine testing status}
\usage{
is_testing()

is_parallel()

is_checking()

is_snapshot()

testing_package()
}
\description{
These functions help you determine if you code is running in a particular
testing context:
\itemize{
\item \code{is_testing()} is \code{TRUE} inside a test.
\item \code{is_snapshot()} is \code{TRUE} inside a snapshot test
\item \code{is_checking()} is \code{TRUE} inside of \verb{R CMD check} (i.e. by \code{\link[=test_check]{test_check()}}).
\item \code{is_parallel()} is \code{TRUE} if the tests are run in parallel.
\item \code{testing_package()} gives name of the package being tested.
}

A common use of these functions is to compute a default value for a \code{quiet}
argument with \code{is_testing() && !is_snapshot()}. In this case, you'll
want to avoid an run-time dependency on testthat, in which case you should
just copy the implementation of these functions into a \code{utils.R} or similar.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skip.R
\name{skip}
\alias{skip}
\alias{skip_if_not}
\alias{skip_if}
\alias{skip_if_not_installed}
\alias{skip_unless_r}
\alias{skip_if_offline}
\alias{skip_on_cran}
\alias{local_on_cran}
\alias{skip_on_os}
\alias{skip_on_ci}
\alias{skip_on_covr}
\alias{skip_on_bioc}
\alias{skip_if_translated}
\title{Skip a test for various reasons}
\usage{
skip(message = "Skipping")

skip_if_not(condition, message = NULL)

skip_if(condition, message = NULL)

skip_if_not_installed(pkg, minimum_version = NULL)

skip_unless_r(spec)

skip_if_offline(host = "captive.apple.com")

skip_on_cran()

local_on_cran(on_cran = TRUE, frame = caller_env())

skip_on_os(os, arch = NULL)

skip_on_ci()

skip_on_covr()

skip_on_bioc()

skip_if_translated(msgid = "'\%s' not found")
}
\arguments{
\item{message}{A message describing why the test was skipped.}

\item{condition}{Boolean condition to check. \code{skip_if_not()} will skip if
\code{FALSE}, \code{skip_if()} will skip if \code{TRUE}.}

\item{pkg}{Name of package to check for}

\item{minimum_version}{Minimum required version for the package}

\item{spec}{A version specification like '>= 4.1.0' denoting that this test
should only be run on R versions 4.1.0 and later.}

\item{host}{A string with a hostname to lookup}

\item{on_cran}{Pretend we're on CRAN (\code{TRUE}) or not (\code{FALSE}).}

\item{frame}{Calling frame to tie change to; expect use only.}

\item{os}{Character vector of one or more operating systems to skip on.
Supported values are \code{"windows"}, \code{"mac"}, \code{"linux"}, \code{"solaris"},
and \code{"emscripten"}.}

\item{arch}{Character vector of one or more architectures to skip on.
Common values include \code{"i386"} (32 bit), \code{"x86_64"} (64 bit), and
\code{"aarch64"} (M1 mac). Supplying \code{arch} makes the test stricter; i.e. both
\code{os} and \code{arch} must match in order for the test to be skipped.}

\item{msgid}{R message identifier used to check for translation: the default
uses a message included in most translation packs. See the complete list in
\href{https://github.com/wch/r-source/blob/master/src/library/base/po/R-base.pot}{\code{R-base.pot}}.}
}
\description{
\code{skip_if()} and \code{skip_if_not()} allow you to skip tests, immediately
concluding a \code{\link[=test_that]{test_that()}} block without executing any further expectations.
This allows you to skip a test without failure, if for some reason it
can't be run (e.g. it depends on the feature of a specific operating system,
or it requires a specific version of a package).

See \code{vignette("skipping")} for more details.
}
\section{Helpers}{

\itemize{
\item \code{skip_if_not_installed("pkg")} skips tests if package "pkg" is not
installed or cannot be loaded (using \code{requireNamespace()}). Generally,
you can assume that suggested packages are installed, and you do not
need to check for them specifically, unless they are particularly
difficult to install.
\item \code{skip_if_offline()} skips if an internet connection is not available
(using \code{\link[curl:nslookup]{curl::nslookup()}}) or if the test is run on CRAN. Requires
\{curl\} to be installed and included in the dependencies of your package.
\item \code{skip_if_translated("msg")} skips tests if the "msg" is translated.
\item \code{skip_on_bioc()} skips on Bioconductor (using the \code{IS_BIOC_BUILD_MACHINE}
env var).
\item \code{skip_on_cran()} skips on CRAN (using the \code{NOT_CRAN} env var set by
devtools and friends). \code{local_on_cran()} gives you the ability to
easily simulate what will happen on CRAN.
\item \code{skip_on_covr()} skips when covr is running (using the \code{R_COVR} env var).
\item \code{skip_on_ci()} skips on continuous integration systems like GitHub Actions,
travis, and appveyor (using the \code{CI} env var).
\item \code{skip_on_os()} skips on the specified operating system(s) ("windows",
"mac", "linux", or "solaris").
}
}

\examples{
if (FALSE) skip("Some Important Requirement is not available")

test_that("skip example", {
  expect_equal(1, 1L)    # this expectation runs
  skip('skip')
  expect_equal(1, 2)     # this one skipped
  expect_equal(1, 3)     # this one is also skipped
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/old-school.R
\name{expect_that}
\alias{expect_that}
\title{Expect that a condition holds.}
\usage{
expect_that(object, condition, info = NULL, label = NULL)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{condition, }{a function that returns whether or not the condition
is met, and if not, an error message to display.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}
}
\value{
the (internal) expectation result as an invisible list
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

An old style of testing that's no longer encouraged.
}
\section{3rd edition}{

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This style of testing is formally deprecated as of the 3rd edition.
Use a more specific \code{expect_} function instead.
}

\examples{
expect_that(5 * 2, equals(10))
expect_that(sqrt(2) ^ 2, equals(2))
\dontrun{
expect_that(sqrt(2) ^ 2, is_identical_to(2))
}
}
\seealso{
\code{\link[=fail]{fail()}} for an expectation that always fails.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local.R
\name{local_test_context}
\alias{local_test_context}
\alias{local_reproducible_output}
\title{Temporarily set options for maximum reproducibility}
\usage{
local_test_context(.env = parent.frame())

local_reproducible_output(
  width = 80,
  crayon = FALSE,
  unicode = FALSE,
  rstudio = FALSE,
  hyperlinks = FALSE,
  lang = "C",
  .env = parent.frame()
)
}
\arguments{
\item{.env}{Environment to use for scoping; expert use only.}

\item{width}{Value of the \code{"width"} option.}

\item{crayon}{Determines whether or not crayon (now cli) colour
should be applied.}

\item{unicode}{Value of the \code{"cli.unicode"} option.
The test is skipped if \code{l10n_info()$`UTF-8`} is \code{FALSE}.}

\item{rstudio}{Should we pretend that we're inside of RStudio?}

\item{hyperlinks}{Should we use ANSI hyperlinks.}

\item{lang}{Optionally, supply a BCP47 language code to set the language
used for translating error messages. This is a lower case two letter
\href{https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes}{ISO 639 country code},
optionally followed by "_" or "-" and an upper case two letter
\href{https://en.wikipedia.org/wiki/ISO_3166-2}{ISO 3166 region code}.}
}
\description{
\code{local_test_context()} is run automatically by \code{test_that()} but you may
want to run it yourself if you want to replicate test results interactively.
If run inside a function, the effects are automatically reversed when the
function exits; if running in the global environment, use
\code{\link[withr:defer]{withr::deferred_run()}} to undo.

\code{local_reproducible_output()} is run automatically by \code{test_that()} in the
3rd edition. You might want to call it to override the the default settings
inside a test, if you want to test Unicode, coloured output, or a
non-standard width.
}
\details{
\code{local_test_context()} sets \code{TESTTHAT = "true"}, which ensures that
\code{\link[=is_testing]{is_testing()}} returns \code{TRUE} and allows code to tell if it is run by
testthat.

In the third edition, \code{local_test_context()} also calls
\code{local_reproducible_output()} which temporary sets the following options:
\itemize{
\item \code{cli.dynamic = FALSE} so that tests assume that they are not run in
a dynamic console (i.e. one where you can move the cursor around).
\item \code{cli.unicode} (default: \code{FALSE}) so that the cli package never generates
unicode output (normally cli uses unicode on Linux/Mac but not Windows).
Windows can't easily save unicode output to disk, so it must be set to
false for consistency.
\item \code{cli.condition_width = Inf} so that new lines introduced while
width-wrapping condition messages don't interfere with message matching.
\item \code{crayon.enabled} (default: \code{FALSE}) suppresses ANSI colours generated by
the cli and crayon packages (normally colours are used if cli detects
that you're in a terminal that supports colour).
\item \code{cli.num_colors} (default: \code{1L}) Same as the crayon option.
\item \code{lifecycle_verbosity = "warning"} so that every lifecycle problem always
generates a warning (otherwise deprecated functions don't generate a
warning every time).
\item \code{max.print = 99999} so the same number of values are printed.
\item \code{OutDec = "."} so numbers always uses \code{.} as the decimal point
(European users sometimes set \code{OutDec = ","}).
\item \code{rlang_interactive = FALSE} so that \code{\link[rlang:is_interactive]{rlang::is_interactive()}} returns
\code{FALSE}, and code that uses it pretends you're in a non-interactive
environment.
\item \code{useFancyQuotes = FALSE} so base R functions always use regular (straight)
quotes (otherwise the default is locale dependent, see \code{\link[=sQuote]{sQuote()}} for
details).
\item \code{width} (default: 80) to control the width of printed output (usually this
varies with the size of your console).
}

And modifies the following env vars:
\itemize{
\item Unsets \code{RSTUDIO}, which ensures that RStudio is never detected as running.
\item Sets \code{LANGUAGE = "en"}, which ensures that no message translation occurs.
}

Finally, it sets the collation locale to "C", which ensures that character
sorting the same regardless of system locale.
}
\examples{
local({
  local_test_context()
  cat(cli::col_blue("Text will not be colored"))
  cat(cli::symbol$ellipsis)
  cat("\n")
})
test_that("test ellipsis", {
  local_reproducible_output(unicode = FALSE)
  expect_equal(cli::symbol$ellipsis, "...")

  local_reproducible_output(unicode = TRUE)
  expect_equal(cli::symbol$ellipsis, "\u2026")
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto-test.R
\name{auto_test_package}
\alias{auto_test_package}
\title{Watches a package for changes, rerunning tests as appropriate.}
\usage{
auto_test_package(pkg = ".", reporter = default_reporter(), hash = TRUE)
}
\arguments{
\item{pkg}{path to package}

\item{reporter}{test reporter to use}

\item{hash}{Passed on to \code{\link[=watch]{watch()}}.  When FALSE, uses less accurate
modification time stamps, but those are faster for large files.}
}
\description{
Watches a package for changes, rerunning tests as appropriate.
}
\seealso{
\code{\link[=auto_test]{auto_test()}} for details on how method works
}
\keyword{debugging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/try-again.R
\name{try_again}
\alias{try_again}
\title{Evaluate an expectation multiple times until it succeeds}
\usage{
try_again(times, code)
}
\arguments{
\item{times}{Number of times to retry.}

\item{code}{Code to evaluate.}
}
\description{
If you have a flaky test, you can use \code{try_again()} to run it a few times
until it succeeds. In most cases, you are better fixing the underlying
cause of the flakeyness, but sometimes that's not possible.
}
\examples{
usually_return_1 <- function(i) {
  if (runif(1) < 0.1) 0 else 1
}

\dontrun{
# 10\% chance of failure:
expect_equal(usually_return_1(), 1)

# 1\% chance of failure:
try_again(1, expect_equal(usually_return_1(), 1))

# 0.1\% chance of failure:
try_again(2, expect_equal(usually_return_1(), 1))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/old-school.R
\name{oldskool}
\alias{oldskool}
\alias{is_null}
\alias{is_a}
\alias{is_true}
\alias{is_false}
\alias{has_names}
\alias{is_less_than}
\alias{is_more_than}
\alias{equals}
\alias{is_equivalent_to}
\alias{is_identical_to}
\alias{equals_reference}
\alias{shows_message}
\alias{gives_warning}
\alias{prints_text}
\alias{throws_error}
\alias{matches}
\title{Old-style expectations.}
\usage{
is_null()

is_a(class)

is_true()

is_false()

has_names(expected, ignore.order = FALSE, ignore.case = FALSE)

is_less_than(expected, label = NULL, ...)

is_more_than(expected, label = NULL, ...)

equals(expected, label = NULL, ...)

is_equivalent_to(expected, label = NULL)

is_identical_to(expected, label = NULL)

equals_reference(file, label = NULL, ...)

shows_message(regexp = NULL, all = FALSE, ...)

gives_warning(regexp = NULL, all = FALSE, ...)

prints_text(regexp = NULL, ...)

throws_error(regexp = NULL, ...)

matches(regexp, all = TRUE, ...)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

Initial testthat used a style of testing that looked like
\verb{expect_that(a, equals(b)))} this allowed expectations to read like
English sentences, but was verbose and a bit too cutesy. This style
will continue to work but has been soft-deprecated - it is no longer
documented, and new expectations will only use the new style
\code{expect_equal(a, b)}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-equality.R
\name{equality-expectations}
\alias{equality-expectations}
\alias{expect_equal}
\alias{expect_identical}
\title{Does code return the expected value?}
\usage{
expect_equal(
  object,
  expected,
  ...,
  tolerance = if (edition_get() >= 3) testthat_tolerance(),
  info = NULL,
  label = NULL,
  expected.label = NULL
)

expect_identical(
  object,
  expected,
  info = NULL,
  label = NULL,
  expected.label = NULL,
  ...
)
}
\arguments{
\item{object, expected}{Computation and value to compare it to.

Both arguments supports limited unquoting to make it easier to generate
readable failures within a function or for loop. See \link{quasi_label} for
more details.}

\item{...}{\strong{3e}: passed on to \code{\link[waldo:compare]{waldo::compare()}}. See its docs to see other
ways to control comparison.

\strong{2e}: passed on to \code{\link[=compare]{compare()}}/\code{\link[=identical]{identical()}}.}

\item{tolerance}{\strong{3e}: passed on to \code{\link[waldo:compare]{waldo::compare()}}. If non-\code{NULL}, will
ignore small floating point differences. It uses same algorithm as
\code{\link[=all.equal]{all.equal()}} so the tolerance is usually relative (i.e.
\verb{mean(abs(x - y) / mean(abs(y)) < tolerance}), except when the differences
are very small, when it becomes absolute (i.e. \verb{mean(abs(x - y) < tolerance}).
See waldo documentation for more details.

\strong{2e}: passed on to \code{\link[=compare]{compare()}}, if set. It's hard to
reason about exactly what tolerance means because depending on the precise
code path it could be either an absolute or relative tolerance.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label, expected.label}{Used to customise failure messages. For expert
use only.}
}
\description{
These functions provide two levels of strictness when comparing a
computation to a reference value. \code{expect_identical()} is the baseline;
\code{expect_equal()} relaxes the test to ignore small numeric differences.

In the 2nd edition, \code{expect_identical()} uses \code{\link[=identical]{identical()}} and
\code{expect_equal} uses \code{\link[=all.equal]{all.equal()}}. In the 3rd edition, both functions use
\href{https://github.com/r-lib/waldo}{waldo}. They differ only in that
\code{expect_equal()} sets \code{tolerance = testthat_tolerance()} so that small
floating point differences are ignored; this also implies that (e.g.) \code{1}
and \code{1L} are treated as equal.
}
\examples{
a <- 10
expect_equal(a, 10)

# Use expect_equal() when testing for numeric equality
\dontrun{
expect_identical(sqrt(2) ^ 2, 2)
}
expect_equal(sqrt(2) ^ 2, 2)
}
\seealso{
\itemize{
\item \code{\link[=expect_setequal]{expect_setequal()}}/\code{\link[=expect_mapequal]{expect_mapequal()}} to test for set equality.
\item \code{\link[=expect_reference]{expect_reference()}} to test if two names point to same memory address.
}

Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-self-test.R
\name{expect_no_success}
\alias{expect_no_success}
\alias{expect_no_failure}
\title{Test for absence of success or failure}
\usage{
expect_no_success(expr)

expect_no_failure(expr)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

These functions are deprecated because \code{\link[=expect_success]{expect_success()}} and
\code{\link[=expect_failure]{expect_failure()}} now test for exactly one success or no failures, and
exactly one failure and no successes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-constant.R
\name{expect_null}
\alias{expect_null}
\title{Does code return \code{NULL}?}
\usage{
expect_null(object, info = NULL, label = NULL)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}
}
\description{
This is a special case because \code{NULL} is a singleton so it's possible
check for it either with \code{expect_equal(x, NULL)} or \code{expect_type(x, "NULL")}.
}
\examples{
x <- NULL
y <- 10

expect_null(x)
show_failure(expect_null(y))
}
\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto-test.R
\name{auto_test}
\alias{auto_test}
\alias{auto_test_package}
\title{Watches code and tests for changes, rerunning tests as appropriate.}
\usage{
auto_test(
  code_path,
  test_path,
  reporter = default_reporter(),
  env = test_env(),
  hash = TRUE
)

auto_test_package(pkg = ".", reporter = default_reporter(), hash = TRUE)
}
\arguments{
\item{code_path}{path to directory containing code}

\item{test_path}{path to directory containing tests}

\item{reporter}{test reporter to use}

\item{env}{environment in which to execute test suite.}

\item{hash}{Passed on to \code{\link[=watch]{watch()}}.  When FALSE, uses less accurate
modification time stamps, but those are faster for large files.}

\item{pkg}{path to package}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

The idea behind \code{auto_test()} is that you just leave it running while
you develop your code.  Every time you save a file it will be automatically
tested and you can easily see if your changes have caused any test
failures.

The current strategy for rerunning tests is as follows:
\itemize{
\item if any code has changed, then those files are reloaded and all tests
rerun
\item otherwise, each new or modified test is run
}
}
\seealso{
\code{\link[=auto_test_package]{auto_test_package()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-shape.R
\name{expect_length}
\alias{expect_length}
\alias{expect_shape}
\title{Do you expect an object with this length or shape?}
\usage{
expect_length(object, n)

expect_shape(object, ..., nrow, ncol, dim)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{n}{Expected length.}

\item{...}{Not used; used to force naming of other arguments.}

\item{nrow, ncol}{Expected \code{\link[=nrow]{nrow()}}/\code{\link[=ncol]{ncol()}} of \code{object}.}

\item{dim}{Expected \code{\link[=dim]{dim()}} of \code{object}.}
}
\description{
\code{expect_length()} inspects the \code{\link[=length]{length()}} of an object; \code{expect_shape()}
inspects the "shape" (i.e. \code{\link[=nrow]{nrow()}}, \code{\link[=ncol]{ncol()}}, or \code{\link[=dim]{dim()}}) of
higher-dimensional objects like data.frames, matrices, and arrays.
}
\examples{
expect_length(1, 1)
expect_length(1:10, 10)
show_failure(expect_length(1:10, 1))

x <- matrix(1:9, nrow = 3)
expect_shape(x, nrow = 3)
show_failure(expect_shape(x, nrow = 4))
expect_shape(x, ncol = 3)
show_failure(expect_shape(x, ncol = 4))
expect_shape(x, dim = c(3, 3))
show_failure(expect_shape(x, dim = c(3, 4, 5)))
}
\seealso{
\code{\link[=expect_vector]{expect_vector()}} to make assertions about the "size" of a vector.

Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_match}()},
\code{\link{expect_named}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-env.R
\name{test_env}
\alias{test_env}
\title{Generate default testing environment.}
\usage{
test_env(package = NULL)
}
\description{
We use a new environment which inherits from \code{\link[=globalenv]{globalenv()}} or a package
namespace. In an ideal world, we'd avoid putting the global environment on
the search path for tests, but it's not currently possible without losing
the ability to load packages in tests.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-named.R
\name{expect_named}
\alias{expect_named}
\title{Do you expect a vector with (these) names?}
\usage{
expect_named(
  object,
  expected,
  ignore.order = FALSE,
  ignore.case = FALSE,
  info = NULL,
  label = NULL
)
}
\arguments{
\item{object}{Object to test.

Supports limited unquoting to make it easier to generate readable failures
within a function or for loop. See \link{quasi_label} for more details.}

\item{expected}{Character vector of expected names. Leave missing to
match any names. Use \code{NULL} to check for absence of names.}

\item{ignore.order}{If \code{TRUE}, sorts names before comparing to
ignore the effect of order.}

\item{ignore.case}{If \code{TRUE}, lowercases all names to ignore the
effect of case.}

\item{info}{Extra information to be included in the message. This argument
is soft-deprecated and should not be used in new code. Instead see
alternatives in \link{quasi_label}.}

\item{label}{Used to customise failure messages. For expert use only.}
}
\description{
You can either check for the presence of names (leaving \code{expected}
blank), specific names (by supplying a vector of names), or absence of
names (with \code{NULL}).
}
\examples{
x <- c(a = 1, b = 2, c = 3)
expect_named(x)
expect_named(x, c("a", "b", "c"))

# Use options to control sensitivity
expect_named(x, c("B", "C", "A"), ignore.order = TRUE, ignore.case = TRUE)

# Can also check for the absence of names with NULL
z <- 1:4
expect_named(z, NULL)
}
\seealso{
Other expectations: 
\code{\link{comparison-expectations}},
\code{\link{equality-expectations}},
\code{\link{expect_error}()},
\code{\link{expect_length}()},
\code{\link{expect_match}()},
\code{\link{expect_null}()},
\code{\link{expect_output}()},
\code{\link{expect_reference}()},
\code{\link{expect_silent}()},
\code{\link{inheritance-expectations}},
\code{\link{logical-expectations}}
}
\concept{expectations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-files.R
\name{test_dir}
\alias{test_dir}
\title{Run all tests in a directory}
\usage{
test_dir(
  path,
  filter = NULL,
  reporter = NULL,
  env = NULL,
  ...,
  load_helpers = TRUE,
  stop_on_failure = TRUE,
  stop_on_warning = FALSE,
  package = NULL,
  load_package = c("none", "installed", "source"),
  shuffle = FALSE
)
}
\arguments{
\item{path}{Path to directory containing tests.}

\item{filter}{If not \code{NULL}, only tests with file names matching this
regular expression will be executed. Matching is performed on the file
name after it's stripped of \code{"test-"} and \code{".R"}.}

\item{reporter}{Reporter to use to summarise output. Can be supplied
as a string (e.g. "summary") or as an R6 object
(e.g. \code{SummaryReporter$new()}).

See \link{Reporter} for more details and a list of built-in reporters.}

\item{env}{Environment in which to execute the tests. Expert use only.}

\item{...}{Additional arguments passed to \code{\link[=grepl]{grepl()}} to control filtering.}

\item{load_helpers}{Source helper files before running the tests?}

\item{stop_on_failure}{If \code{TRUE}, throw an error if any tests fail.}

\item{stop_on_warning}{If \code{TRUE}, throw an error if any tests generate
warnings.}

\item{package}{If these tests belong to a package, the name of the package.}

\item{load_package}{Strategy to use for load package code:
\itemize{
\item "none", the default, doesn't load the package.
\item "installed", uses \code{\link[=library]{library()}} to load an installed package.
\item "source", uses \code{\link[pkgload:load_all]{pkgload::load_all()}} to a source package.
To configure the arguments passed to \code{load_all()}, add this
field in your DESCRIPTION file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Config/testthat/load-all: list(export_all = FALSE, helpers = FALSE)
}\if{html}{\out{</div>}}
}}

\item{shuffle}{If \code{TRUE}, randomly reorder the top-level expressions
in the file.}
}
\value{
A list (invisibly) containing data about the test results.
}
\description{
This function is the low-level workhorse that powers \code{\link[=test_local]{test_local()}} and
\code{\link[=test_package]{test_package()}}. Generally, you should not call this function directly.
In particular, you are responsible for ensuring that the functions to test
are available in the test \code{env} (e.g. via \code{load_package}).

See \code{vignette("special-files")} to learn more about the conventions for test,
helper, and setup files that testthat uses, and what you might use each for.
}
\section{Environments}{

Each test is run in a clean environment to keep tests as isolated as
possible. For package tests, that environment inherits from the package's
namespace environment, so that tests can access internal functions
and objects.
}
